/*
 * @(#)Bodhidharma.java	1.2.2 07/15/05
 *
 * McGill University
 */

import	bodhidharma.ConfigureFileLocationsJFrame;
import	java.io.*;
import	java.util.LinkedList;
import	javax.swing.JFrame;
import	javax.swing.tree.DefaultMutableTreeNode;


/**
 * The root class of the Bodhidharma automatic musical classification system.
 * If no command line arguments are present, then running the main method of
 * this class from the command line instantiates the main 
 * <code>ConfigureFileLocationsJFrame</code> of the GUI, which allows
 * the user to choose appropriate file paths for the work he or she will be
 * performing. Command line arguments cause Bodhidharma to be run without
 * displaying the GUI with the given settings.
 *
 * <p>The command line action options are as follows:
 * -extract features: Extract features from MIDID files.
 * -tran <path>: Train on all currently loaded recordings and save to the given path.
 * -classify <path>: Classify all currently loaded recordings using the specified trained classifiers.
 *
 * <p>The command line settings are as follows:
 * -proj <path>: Specify the path of a project file to load.
 * -tax <path>: Specify the path of a taxonomy file to load.
 * -feat <path>: Specify the path of a feature settings file to load.
 * -rec <path>: Specify the path of a recordings file to load.
 * -clas <path>: Specify the path of a trained classifiers file to load.
 * -pref <path>: Specify the path of a preferences file to load.
 * -dir <path>: Specify the path of the directory to use as the current directory.
 * 
 * <p>Note that if a project file is specified at the command line, then this overrides
 * other specified files (e.g. taxonomy file, feature settings file, etc.)
 *
 * @author Cory McKay
 * @see	ConfigureFileLocationsJFrame
 */
public class Bodhidharma
{
	/**
	 * Instantiates the parent <code>JFrame</code> (a 
	 * <code>ConfigureFileLocationsJFrame</code>) of the Theophastus
	 * software.
	 *
	 * @param	com_arg	Command line arguments. None currently implemented.
	 * @see				ConfigureFileLocationsJFrame
     */
	public static void main (String com_arg[])
	{
		// Prepare the file paths panel
		ConfigureFileLocationsJFrame frame = new ConfigureFileLocationsJFrame();

		// Display the GUI if Bodhidharma was not given any additional command
		// line arguments
		if (com_arg.length == 0)
		{
			frame.pack();
			frame.show();
		}

		// Parse command line arguments if any
		else
		{
			// Display an error and quit if there are an odd number of command
			// line arguments
			if (com_arg.length % 2 != 0)
			{
				System.out.println("ERROR: Incorrect number of command line arguments.");
				System.exit(-1);
			}

			// Prepare to store the parameters entered through the command line
			String project_path = "";
			String taxonomy_path = "";
			String feature_settings_path = "";
			String recordings_path = "";
			String trained_classifiers_path = "";
			String preferences_path = "";
			String current_directory_path = "";
			String train_output_path = null;
			String classification_output_path = null;
			String extract_code = null;

			// Parse the command line parameters
			int i = 0;
			while (i < com_arg.length)
			{
				if (com_arg[i].equals("-proj"))
				{
					project_path = com_arg[i+1];
					i = i + 2;
				}
				else if (com_arg[i].equals("-tax"))
				{
					taxonomy_path = com_arg[i+1];
					i = i + 2;
				}
				else if (com_arg[i].equals("-feat"))
				{
					feature_settings_path = com_arg[i+1];
					i = i + 2;
				}
				else if (com_arg[i].equals("-rec"))
				{
					recordings_path = com_arg[i+1];
					i = i + 2;
				}
				else if (com_arg[i].equals("-clas"))
				{
					trained_classifiers_path = com_arg[i+1];
					i = i + 2;
				}
				else if (com_arg[i].equals("-pref"))
				{
					preferences_path = com_arg[i+1];
					i = i + 2;
				}
				else if (com_arg[i].equals("-dir"))
				{
					current_directory_path = com_arg[i+1];
					i = i + 2;
				}
				else if (com_arg[i].equals("-extract"))
				{
					extract_code = new String(com_arg[i+1]);
					i = i + 2;
				}
				else if (com_arg[i].equals("-train"))
				{
					train_output_path = new String(com_arg[i+1]);
					i = i + 2;
				}
				else if (com_arg[i].equals("-classify"))
				{
					classification_output_path = new String(com_arg[i+1]);
					i = i + 2;
				}
				else
				{
					System.out.println("ERROR: Unknown flag: " + com_arg[i]);
					System.exit(-1);
				}
			}

			// Display an error message and quit if both or neither of the
			// training and classifiying options are selected.
			if (train_output_path == null && classification_output_path == null && extract_code == null)
			{
				System.out.println("ERROR: Must either train or classify.");
				System.exit(-1);
			}
			if (train_output_path != null && classification_output_path != null)
			{
				System.out.println("ERROR: Must either train or classify, not both.");
				System.exit(-1);
			}

			// Display an error message and quit if a project file is specified
			// and other paths are also specified
			if ( !project_path.equals("") &&
				 ( 
				   !taxonomy_path.equals("") ||
				   !feature_settings_path.equals("") ||
				   !recordings_path.equals("") ||
				   !trained_classifiers_path.equals("") ||
				   !preferences_path.equals("") ||
				   !current_directory_path.equals("")
				 )
			   )
			{
				System.out.println("ERROR: May not override a set of project settings.");
				System.exit(-1);
			}

			// Update the stored set of file paths
			frame.setProjectPath(project_path);
			frame.setTaxonomyPath(taxonomy_path);
			frame.setFeaturesPath(feature_settings_path);
			frame.setRecordingsPath(recordings_path);
			frame.setClassifiersPath(trained_classifiers_path);
			frame.setPreferencesPath(preferences_path);
			frame.setCurrentDirectory(current_directory_path);

			// Set up the project if appropriate
			if ( !project_path.equals("") )
				frame.loadProject(project_path, true);

			// Extract features if appropriate
			if (extract_code != null)
				if (extract_code.equals("features"))
					frame.commenceFeatureExtraction();

			// Commence training if appropriate
			if (train_output_path != null)
				frame.commenceTraining(train_output_path);

			// Commence classification if appropriate
			if (classification_output_path != null)
				frame.commenceClassifiying(classification_output_path);

			// End execution
			System.exit(0);
		}
	}
}