/*
 * @(#)ConfigureFileLocationsJFrame.java	1.2.2 07/13/05
 *
 * McGill Univarsity
 */

package bodhidharma;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import bodhidharma.xml_parsing.*;
import bodhidharma.utilities.*;

/**
 * Part of the Bodhidharma automatic music classification
 * project. This is the first dialog box that users of the software
 * are presented with.
 *
 * <p>This dialog box allows one to choose the paths of the 
 * current project, the file containing the musical taxonomy used
 * to train classifiers, the list of recordings to train and/or
 * classify, the features of these recordings, the trained classifying
 * systm, the prefences file and the current directory.
 *
 * <p>These paths may be chosen by typing them in (except for 
 * the project field) or by pressing on<i>Brwose</i> buttons, which
 * will bring up a reusable <code>JFileChoose</code> dialog box.
 *
 * <p>The <i>Load Project</i> button may be pressed in order to
 * load a project name and its associated file paths into this
 * dialog box.
 *
 * <p>The <i>Clear Paths</i> button clears all text fields.
 *
 * <p>The <i>Restore Original Defaults</i> button loads the
 * hardcoded defaults into this dialog box.
 *
 * <p>The <i>Save As Startup Defaults</i> button saves the
 * contents of the dialog box as the defaults that will fill
 * the dialog box when it is instantiated. 
 *
 * <p>The <i>Start New Project</i> button brings up a dialog box that
 * allows the user to choose the name of a new project and where to save
 * it. The project XML file is then created at the given location with
 * the given name. A new directory with the same name (minus the .xml)
 * is created. The current directory is updated to this directory. The user
 * is asked if s/he would like to copy the files referred to by
 * the existing entries to the new directory. The text fields and 
 * fields are all updated.
 *
 * <p>The <i>Cancel</i> button hides this dialog box. Any changes
 * that were made are lost, except for the creation of new project
 * files.
 *
 * <p>The <i>OK</i> button hides this window and instantiates
 * the main window. If the main window has already been created,
 * it is replaced.
 *
 * <p>The path fields of this dialog box hold the paths of these
 * pieces of information after they have been chosen, and may be
 * referenced by other objects.
 *
 * <p>The <code>setTaxonomyPath</code>, <code>setFeaturesPath</code> and 
 * <code>setRecordingsPath</code> methods can be called to reset the
 * appropriate fields and text fields.
 *
 * <p>The browse methods of this class can also be called by other
 * objects to bring up the <code>JFileChoose</code> dialog box to change
 * these file paths individually. The newly updated path field(s) can
 * then be accessed by other objects.
 *
 * <p>Objects of this class also hold a <code>JFileChoose</code>
 * dialog box that can be called by other classes in the project.
 * Objects of this class also holdo a reference to the most recently
 * spawned main window.
 *
 * <p>The <code>loadProject</code> method may also be called by other objects
 * in order to load an entirely new <i>project_file</i> XML without going through
 * this dialog box manually.
 *
 * <p>The <code>saveFileExternally</code> method may also be called by
 * other objects in order to save a file, potentially with a 
 * <code>JFileChooser</code> dialog box, and, if appropriate, update
 * the path fields of this dialog box.
 *
 * <p>The <code>getFile</code> method may also be called by other
 * objects in order to get a File from a given path.
 *
 * <p>The <code>saveProjectFile</code> method may also be called by other
 * objects in order to save the current project paths in the current
 * <i>project_file</i>.
 *
 * <p>The <code>startNewProject</code> method may also be called by other
 * methods to bring up a dialog box allowing the creation of a new project.
 *
 * <p>The defaults that are loaded when an object of this class is
 * instantiated are set by the default_file_locations.xml file
 * found in the same directory as this class. The only field of this
 * file that is considered is the project file that it refers to, which
 * is then loaded. This default_file_locations file will be
 * automatically recreated if it is absent.
 *
 * @author Cory McKay
 * @see	Bodhidharma
 * @see MainJFrame
 * @see FileFilterXML
 */
public class ConfigureFileLocationsJFrame
	extends JFrame
	implements ActionListener
{
	/* FIELDS ******************************************************************/


	/**
	 * Path of the current project.
	 */
	public String			project_path;

	/**
	 * Path of the current musical taxonomy file.
	 */
	public String			taxonomy_path;

	/**
	 * Path of the current list of feature settings.
	 */
	public String			features_path;

	/**
	 * Path of the current list of recordings intended for
	 * training and/or testing.
	 */
	public String			recordings_list_path;

	/**
	 * Path of the file holding the current classification system.
	 */
	public String			classifiers_path;

	/**
	 * Path of the preferences file.
	 */
	public String			preferences_path;

	/**
	 * Path of the current directory
	 */
	public String			current_directory;

	/**
	 * The main window of the Bodhidharma GUI
	 */
	public MainJFrame		main_window;

	// Inicates whether or not the main window has been displayed
	private boolean			main_window_has_been_displayed;

	// Containers to hold Swing elements
	private Container		content_pane;
	private JPanel			project_panel;
	private JPanel			taxonomy_panel;
	private JPanel			features_panel;
	private JPanel			classifiers_panel;
	private JPanel			recordings_panel;
	private JPanel			preferences_panel;
	private JPanel			current_directory_panel;
	private JPanel			button_panel;

	// Text fields to display and allow editing of currently selected
	// file paths
	private JTextField		project_text_field;
	private JTextField		taxonomy_text_field;
	private JTextField		features_text_field;
	private JTextField		classifiers_text_field;
	private JTextField		recordings_text_field;
	private JTextField		preferences_text_field;
	private JTextField		current_directory_text_field;

	// Buttons to bring up JFileChooser dialog box to select file paths
	private JButton			taxonomy_browse_button;
	private JButton			features_browse_button;
	private JButton			classifiers_browse_button;
	private JButton			recordings_browse_button;
	private JButton			preferences_browse_button;
	private JButton			current_directory_browse_button;

	// Button to clear fields and text fields
	private JButton			clear_button;

	// Button to revert to create a new project
	private JButton			new_project_button;

	// Button to revert to pre-programmed paths
	private JButton			restore_defaults_button;

	// Button to store currently selected paths as defaults
	private JButton			save_as_defaults_button;

	// Button to load a project file containing all of these paths
	private JButton			load_project_button;

		// Button to finalize the paths entered
	private JButton			cancel_button;

	// Button to finalize the paths entered
	private JButton			ok_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. Configures the dialog box and loads the
	 * default paths from the default_file_locations.xml file.
	 * Instantiates and configures the browse_dialo JFileChooser.
	 */
	public ConfigureFileLocationsJFrame()
	{
		// Inisitialize path fields to empty
		project_path = new String("");
		taxonomy_path = new String("");
		recordings_list_path = new String("");
		features_path = new String("");
		classifiers_path = new String("");
		preferences_path = new String("");
		current_directory = new String("");

		// Configure overall window settings and instantiate related panels
		setTitle("Configure File Locations");
		int number_text_columns = 30; // number of columns in each text field
		int horizontal_gap = 6; // horizontal space between GUI elements
		int vertical_gap = 11; // horizontal space between GUI elements
		content_pane = getContentPane();
		content_pane.setLayout(new GridLayout(8, 1, horizontal_gap, vertical_gap));
		project_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		taxonomy_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		features_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		classifiers_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		recordings_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		preferences_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		current_directory_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		button_panel = new JPanel(new GridLayout(2, 3, horizontal_gap, vertical_gap));

		// Note that the main window of the GUI has not yet been displayed
		main_window_has_been_displayed = false;

		// Instantiate and configure text fields
		project_text_field = new JTextField(number_text_columns);
		project_text_field.setEditable(false);
		taxonomy_text_field = new JTextField(number_text_columns);
		features_text_field = new JTextField(number_text_columns);
		classifiers_text_field = new JTextField(number_text_columns);
		recordings_text_field = new JTextField(number_text_columns);
		preferences_text_field = new JTextField(number_text_columns);
		current_directory_text_field = new JTextField(number_text_columns);

		// Instantiate and configure buttons and their ActionListener
		load_project_button = new JButton("Load Project");
		taxonomy_browse_button = new JButton("Browse");
		features_browse_button = new JButton("Browse");
		classifiers_browse_button = new JButton("Browse");
		recordings_browse_button = new JButton("Browse");
		preferences_browse_button = new JButton("Browse");
		current_directory_browse_button = new JButton("Browse");
		clear_button = new JButton("Clear Paths");
		restore_defaults_button = new JButton("Restore Original Defaults");
		save_as_defaults_button = new JButton("Save As Startup Defaults");
		new_project_button = new JButton("Start New Project");
		cancel_button = new JButton("Cancel");
		ok_button = new JButton("OK");
		load_project_button.addActionListener(this);
		taxonomy_browse_button.addActionListener(this);
		features_browse_button.addActionListener(this);
		classifiers_browse_button.addActionListener(this);
		recordings_browse_button.addActionListener(this);
		preferences_browse_button.addActionListener(this);
		current_directory_browse_button.addActionListener(this);
		clear_button.addActionListener(this);
		restore_defaults_button.addActionListener(this);
		save_as_defaults_button.addActionListener(this);
		new_project_button.addActionListener(this);
		cancel_button.addActionListener(this);
		ok_button.addActionListener(this);

		// Load default paths
		initializeDefaults();

		// Add labels and previously created elements to panels and container
		project_panel.add(new JLabel("Project file:"), BorderLayout.NORTH);
		project_panel.add(project_text_field, BorderLayout.WEST);
		project_panel.add(load_project_button, BorderLayout.CENTER);
		taxonomy_panel.add(new JLabel("File holding musical taxonomy:"), BorderLayout.NORTH);
		taxonomy_panel.add(taxonomy_text_field, BorderLayout.WEST);
		taxonomy_panel.add(taxonomy_browse_button, BorderLayout.CENTER);
		features_panel.add(new JLabel("File holding feature settings:"), BorderLayout.NORTH);
		features_panel.add(features_text_field, BorderLayout.WEST);
		features_panel.add(features_browse_button, BorderLayout.CENTER);
		recordings_panel.add(new JLabel("File holding training and/or testing recordings:"), BorderLayout.NORTH);
		recordings_panel.add(recordings_text_field, BorderLayout.WEST);
		recordings_panel.add(recordings_browse_button, BorderLayout.CENTER);
		classifiers_panel.add(new JLabel("File holding trained classifiers:"), BorderLayout.NORTH);
		classifiers_panel.add(classifiers_text_field, BorderLayout.WEST);
		classifiers_panel.add(classifiers_browse_button, BorderLayout.CENTER);
		preferences_panel.add(new JLabel("Preferences file:"), BorderLayout.NORTH);
		preferences_panel.add(preferences_text_field, BorderLayout.WEST);
		preferences_panel.add(preferences_browse_button, BorderLayout.CENTER);
		current_directory_panel.add(new JLabel("Current directory:"), BorderLayout.NORTH);
		current_directory_panel.add(current_directory_text_field, BorderLayout.WEST);
		current_directory_panel.add(current_directory_browse_button, BorderLayout.CENTER);
		button_panel.add(clear_button); 
		button_panel.add(restore_defaults_button); 
		button_panel.add(save_as_defaults_button);
		button_panel.add(cancel_button);
		button_panel.add(new_project_button);
		button_panel.add(ok_button);
		content_pane.add(project_panel);
		content_pane.add(taxonomy_panel);
		content_pane.add(features_panel);
		content_pane.add(recordings_panel);
		content_pane.add(classifiers_panel);
		content_pane.add(preferences_panel);
		content_pane.add(current_directory_panel);
		content_pane.add(button_panel);

		// Cause program to react when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				cancel();
			}
		});
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the load_project_button
		if (event.getSource().equals(load_project_button))
			loadProject(null, false);

		// React to the taxonomy_browse_button
		else if (event.getSource().equals(taxonomy_browse_button))
			browseTaxonomyPath(false);

		// React to the recordings_browse_button
		else if (event.getSource().equals(features_browse_button))
			browseFeaturesPath(false);

		// React to the classifiers_browse_button
		else if (event.getSource().equals(classifiers_browse_button))
			browseClassifiersPath(false);

		// React to the recordings_browse_button
		else if (event.getSource().equals(recordings_browse_button))
			browseRecordingsPath(false);

		// React to the preferences_browse_button
		else if (event.getSource().equals(preferences_browse_button))
			browsePreferencesPath(false);

		// React to the current_directory_browse_button
		else if (event.getSource().equals(current_directory_browse_button))
			browseCurrentDirectoryPath(false);

		// React to the clear_button
		if (event.getSource().equals(clear_button))
			clearPaths();

		// React to the new_project_button
		else if (event.getSource().equals(new_project_button))
			startNewProject();

		// React to the restore_defaults_button 
		else if (event.getSource().equals(restore_defaults_button))
			restoreDefaultPaths();

		// React to the save_as_defaults_button by saving a project XML
		// file in the same directory as the Bodhidharma program.
		else if (event.getSource().equals(save_as_defaults_button))
			saveProjectFile("." + File.separator + "default_file_locations.xml", "This project_file is a dummy used to specify the defualt project and related file locations loaded when the Bodhidharma software is run. It is not truly a project itself.", true);

		// React to the cancel_button by hiding this dialog box.
		// Any changes to the paths are lost
		else if (event.getSource().equals(cancel_button))
			cancel();

		// React to the ok_button
		else if (event.getSource().equals(ok_button))
			ok();
	}

	
	/**
	 * Takes in the path of an XML project file (of type file_locations).
	 * Loads this file, parses it to verify its validity and sets the
	 * displays and fields of this dialog box to match its contents.
	 *
	 * <p> If the argument does not point to a valid project file of if
	 * <i>null</i> is passed as the argument, then choose the file using
	 * a <code>JFileChooser</code> dialog box.
	 *
	 * @param	project_path		Path of a valid file_locations XML file.
	 * @param	called_externally	True if called by an external object, false if called internally.
	 */
	public void loadProject(String project_path, boolean called_externally)
	{
		File selection = null;

		if (project_path != null)
		{
			try{selection = new File(project_path);}
			catch(Exception e) {JOptionPane.showMessageDialog(null, "Unable to open project file. Please proceed to choose your project file.", "ERROR", JOptionPane.ERROR_MESSAGE);}
		}

		if (selection == null)
		{
			JFileChooser browse_dialog = getNewJFileChooser();

			int browse_return = browse_dialog.showOpenDialog(ConfigureFileLocationsJFrame.this);
			if (browse_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				selection = browse_dialog.getSelectedFile();
				project_path = selection.getAbsolutePath();
			}
			else selection = null;
		}

		if (selection != null)
		{
		try
			{
				String[] parsed_file = (String[])(XMLDocumentParser.parseXMLDocument(project_path, "project_file"));
				for (int i = 0; i < parsed_file.length; i++)
				{
					project_text_field.setText(parsed_file[1]); // comment data is ignored
					taxonomy_text_field.setText(parsed_file[2]);
					recordings_text_field.setText(parsed_file[3]);
					features_text_field.setText(parsed_file[4]);
					classifiers_text_field.setText(parsed_file[5]);
					preferences_text_field.setText(parsed_file[6]);
					current_directory_text_field.setText(parsed_file[7]);
				}
			}
			catch(Exception e) {JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);}
		}

		if (called_externally) readTextFields();
	}


	/**
	 * Sets the <i>project_path</i> <i>project_text_field</i> fields to the value of the parameter.
	 *
	 * @param	new_path	The new path.
	 */
	public void setProjectPath(String new_path)
	{
		project_path = new_path;
		project_text_field.setText(new_path);
	}
	

	/**
	 * Sets the <i>taxonomy_path</i> <i>taxonomy_text_field</i> fields to the value of the parameter.
	 *
	 * @param	new_path	The new path.
	 */
	public void setTaxonomyPath(String new_path)
	{
		taxonomy_path = new_path;
		taxonomy_text_field.setText(new_path);
	}
	

	/**
	 * Sets the <i>features_path</i> <i>features_text_field</i> fields to the value of the parameter.
	 *
	 * @param	new_path	The new path.
	 */
	public void setFeaturesPath(String new_path)
	{
		features_path = new_path;
		features_text_field.setText(new_path);
	}
	

	/**
	 * Sets the <i>recordings_list_path</i> <i>recordings_text_field</i> fields to the value of the parameter.
	 *
	 * @param	new_path	The new path.
	 */
	public void setRecordingsPath(String new_path)
	{
		recordings_list_path = new_path;
		recordings_text_field.setText(new_path);
	}
	
	
	/**
	 * Sets the <i>classifiers_path</i> <i>classifiers_text_field</i> fields to the value of the parameter.
	 *
	 * @param	new_path	The new path.
	 */
	public void setClassifiersPath(String new_path)
	{
		classifiers_path = new_path;
		classifiers_text_field.setText(new_path);
	}
	

	/**
	 * Sets the <i>preferences_path</i> <i>preferences_text_field</i> fields to the value of the parameter.
	 *
	 * @param	new_path	The new path.
	 */
	public void setPreferencesPath(String new_path)
	{
		preferences_path = new_path;
		preferences_text_field.setText(new_path);
	}
	

	/**
	 * Sets the <i>current_directory</i> <i>current_directory_text_field</i> fields to the value of the parameter.
	 *
	 * @param	new_path	The new path.
	 */
	public void setCurrentDirectory(String new_path)
	{
		current_directory = new_path;
		current_directory_text_field.setText(new_path);
	}
	
	
	/**
	 * Opens a <code>JFileChooser</code> dialog box that only displays
	 * directories and .xml files. Allows user to choose the path for
	 * the XML file holding the musical taxonomy.
	 *
	 * @param	called_externally	True if called by an external object, false if called internally.
	 * @return	true if the OK button was pressed, false if the cancel button was pressed
	 * @see		FileFilterXML
	 */
	public boolean browseTaxonomyPath(boolean called_externally)
	{
		JFileChooser browse_dialog = getNewJFileChooser();
		int browse_return = browse_dialog.showOpenDialog(ConfigureFileLocationsJFrame.this);
		if (browse_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
		{
			File selection = browse_dialog.getSelectedFile();
			taxonomy_text_field.setText(selection.getPath());

			if (called_externally) readTextFields();
			
			return true;
		}
		else
			return false;
	}


	/**
	 * Opens a <code>JFileChooser</code> dialog box that only displays
	 * directories and .xml files. Allows user to choose the path for
	 * the XML file holding the recordings list.
	 *
	 * @param	called_externally	True if called by an external object, false if called internally.
	 * @return	true if the OK button was pressed, false if the cancel button was pressed
	 * @see FileFilterXML
	 */
	public boolean browseRecordingsPath(boolean called_externally)
	{
		JFileChooser browse_dialog = getNewJFileChooser();
		int browse_return = browse_dialog.showOpenDialog(ConfigureFileLocationsJFrame.this);
		if (browse_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
		{
			File selection = browse_dialog.getSelectedFile();
			recordings_text_field.setText(selection.getPath());

			if (called_externally) readTextFields();
			
			return true;
		}
		else
			return false;
	}


	/**
	 * Opens a <code>JFileChooser</code> dialog box that only displays
	 * directories and .xml files. Allows user to choose the path for
	 * the XML file holding the features for the given recording.
	 *
	 * @param	called_externally	True if called by an external object, false if called internally.
	 * @return	true if the OK button was pressed, false if the cancel button was pressed
	 * @see FileFilterXML
	 */
	public boolean browseFeaturesPath(boolean called_externally)
	{
		JFileChooser browse_dialog = getNewJFileChooser();
		int browse_return = browse_dialog.showOpenDialog(ConfigureFileLocationsJFrame.this);
		if (browse_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
		{
			File selection = browse_dialog.getSelectedFile();
			features_text_field.setText(selection.getPath());

			if (called_externally) readTextFields();
			
			return true;
		}
		else
			return false;
	}


	/**
	 * Opens a <code>JFileChooser</code> dialog box that only displays
	 * directories and .xml files. Allows user to choose the path for
	 * the XML file holding the classification system.
	 *
	 * @param	called_externally	True if called by an external object, false if called internally.
	 * @return	true if the OK button was pressed, false if the cancel button was pressed
	 * @see FileFilterXML
	 */
	public boolean browseClassifiersPath(boolean called_externally)
	{
		JFileChooser browse_dialog = getNewJFileChooser();
		int browse_return = browse_dialog.showOpenDialog(ConfigureFileLocationsJFrame.this);
		if (browse_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
		{
			File selection = browse_dialog.getSelectedFile();
			classifiers_text_field.setText(selection.getPath());

			if (called_externally) readTextFields();
			
			return true;
		}
		else
			return false;
	}


	/**
	 * Opens a <code>JFileChooser</code> dialog box that only displays
	 * directories and .xml files. Allows user to choose the path for
	 * the XML file holding the preferences.
	 *
	 * @param	called_externally	True if called by an external object, false if called internally.
	 * @return	true if the OK button was pressed, false if the cancel button was pressed
	 * @see FileFilterXML
	 */
	public boolean browsePreferencesPath(boolean called_externally)
	{
		JFileChooser browse_dialog = getNewJFileChooser();
		int browse_return = browse_dialog.showOpenDialog(ConfigureFileLocationsJFrame.this);
		if (browse_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
		{
			File selection = browse_dialog.getSelectedFile();
			preferences_text_field.setText(selection.getPath());

			if (called_externally) readTextFields();
			
			return true;
		}
		else
			return false;
	}

	
	/**
	 * Opens a <code>JFileChooser</code> dialog box that only displays
	 * directories. Allows user to choose the path for the current default
	 * directory. Sets the current directory to that value.
	 *
	 * @param	called_externally	True if called by an external object, false if called internally.
	 * @return	true if the OK button was pressed, false if the cancel button was pressed
	 */
	public boolean browseCurrentDirectoryPath(boolean called_externally)
	{
		JFileChooser browse_dialog = getNewJFileChooser();
		browse_dialog.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);

		int browse_return = browse_dialog.showOpenDialog(ConfigureFileLocationsJFrame.this);
		if (browse_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
		{
			File selection = browse_dialog.getSelectedFile();
			current_directory_text_field.setText(selection.getPath());
			browse_dialog.setCurrentDirectory(browse_dialog.getSelectedFile());
			
			if (called_externally) readTextFields();

			browse_dialog.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);

			return true;
		}
		else
		{
			browse_dialog.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
			return false;
		}
	}


	/**
	 * Opens a <code>JFileChooser</code> dialog box that allows the user to choose
	 * the path of the file to be saved. Makes sure can write to this path and 
	 * returns the corresponding File. Displays an error dialog box if there is a
	 * problem. Returns null if can or should not save.
	 * 
	 * <p>The <i>file_type</i> parameter specifies the type of file to be saved. The
	 * appropriate fields and and text fields are updated if the file type is recognized.
	 *
	 * <p>Automatically adds .xml as an extension if this is not already present.
	 *
	 * @param	can_erase		Whether or not existing files with the specified path should be overwritten without asking the user.
	 * @param	file_type		The code identifying the type of XML file.
	 * @return	A File if appropriate, null if nothing should be saved after all.
	 */
	public File saveFileExternally(boolean can_erase, String file_type)
	{
		int save_return; // what is returned by the JFileChooser object
		JFileChooser save_dialog = new JFileChooser(new File(current_directory));	
		save_return = save_dialog.showSaveDialog(this);
		File selection = null;
		if (save_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
		{
			// Get file and its path
			File temp_file = save_dialog.getSelectedFile();
			String path = temp_file.getPath();

			// Make sure has .xml extension
			String ext = bodhidharma.utilities.StringMethods.getExtension(path);
			if (ext == null)
			{
				path += ".xml";
				temp_file = new File(path);
			}
			else if (!ext.equals(".xml"))
			{
				path = bodhidharma.utilities.StringMethods.removeExtension(path) + ".xml";
				temp_file = new File(path);
			}

			// Update fields
			selection = getFile(path, can_erase);
			if (selection != null)
			{
				// Update the appropriate field and JTextField
				if (file_type.equals("taxonomy_file"))
				{
					taxonomy_path = path;
					taxonomy_text_field.setText(path);
				}
				else if (file_type.equals("recordings_file"))
				{
					recordings_list_path = path;
					recordings_text_field.setText(path);
				}
				else if (file_type.equals("features_file"))
				{
					features_path = path;
					features_text_field.setText(path);
				}
				else if (file_type.equals("classifiers_file"))
				{
					classifiers_path = path;
					classifiers_text_field.setText(path);
				}
				else if (file_type.equals("preferences_file"))
				{
					preferences_path = path;
					preferences_text_field.setText(path);
				}
			}
		}
		return selection;
	}


	/**
	 * Saves a <i>project_file</i> XML file in the specified location with the
	 * specified comments. This file is given the paths specified in the text
	 * boxes in this dialog box.
	 *
	 * @param	path		The path under which the file is to be saved.
	 * @param	contents	The contents of the file to save.
	 * @param	can_erase	True means user is not warned if user is overwrting a file, false means is warned.
	 */
	public void saveProjectFile(String path, String comments, boolean can_erase)
	{
		File to_file = getFile(path, can_erase);
		String contents = new String
		(
			"<?xml version=\"1.0\"?>\n" +
			"<!DOCTYPE project_file [\n" +
			"   <!ELEMENT project_file (project_comments, project_path, taxonomy_path, recordings_list_path, features_path, classifiers_path, preferences_path, current_directory)>\n" +
			"   <!ELEMENT project_comments (#PCDATA)>\n" +
			"   <!ELEMENT project_path (#PCDATA)>\n" +
			"   <!ELEMENT taxonomy_path (#PCDATA)>\n" +
			"   <!ELEMENT recordings_list_path (#PCDATA)>\n" +
			"   <!ELEMENT features_path (#PCDATA)>\n" +
			"   <!ELEMENT classifiers_path (#PCDATA)>\n" +
			"   <!ELEMENT preferences_path (#PCDATA)>\n" +
			"   <!ELEMENT current_directory (#PCDATA)>\n" +
			"]>\n" +
			"\n" +
			"<project_file>\n" +
			"   <project_comments>" + comments + "</project_comments>\n" +
			"   <project_path>" + project_text_field.getText() + "</project_path>\n" +
			"   <taxonomy_path>" + taxonomy_text_field.getText() + "</taxonomy_path>\n" +
			"   <recordings_list_path>" + recordings_text_field.getText() + "</recordings_list_path>\n" +
			"   <features_path>" + features_text_field.getText() + "</features_path>\n" +
			"   <classifiers_path>" + classifiers_text_field.getText() + "</classifiers_path>\n" +
			"   <preferences_path>" + preferences_text_field.getText() + "</preferences_path>\n" +
			"   <current_directory>" + current_directory_text_field.getText() + "</current_directory>\n" +
			"</project_file>"
		);

		try
		{
			FileOutputStream to = new FileOutputStream(to_file);
			DataOutputStream writer = new DataOutputStream(to);
			writer.writeBytes(contents);
			writer.close();
		}
		catch (Exception e)
		{
				JOptionPane.showMessageDialog(null, ("Unable to write file " + to_file.getName() + "."), "ERROR", JOptionPane.ERROR_MESSAGE);
				// e.printStackTrace();
		}
	}


	/**
	 * Brings up a dialog box that allows the user to choose the name
	 * of a new project and where to save it. The project XML file is
	 * then created at the given location with the given name. A new
	 * directory with the same name (minus the .xml) is created.
	 * The current directory is updated to this directory. The user
	 * is asked if s/he would like to copy the files referred to by
	 * the existing entries to the new directory. The text fields and 
	 * fields are all updated.
	 */
	public void startNewProject()
	{
		int save_return; // what is returned by the JFileChooser object
		JFileChooser save_dialog = new JFileChooser(new File("."));	
		save_dialog.setDialogTitle("Save New Project File"); 
		save_return = save_dialog.showSaveDialog(this);
		if (save_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
		{
			// Get name of selected project_file
			File selection = save_dialog.getSelectedFile();
			String new_project_path = selection.getPath();

			// Make sure project_file has .xml extension and give if doesn't
			String ext = bodhidharma.utilities.StringMethods.getExtension(new_project_path);
			if (ext == null)
			{
				new_project_path += ".xml";
				selection = new File(new_project_path);
			}
			else if (!ext.equals(".xml"))
			{
				new_project_path = bodhidharma.utilities.StringMethods.removeExtension(new_project_path) + ".xml";
				selection = new File(new_project_path);
			}

			try
			{
				// Make sure file doesn't already exist
				if (selection.exists())
					throw new Exception("Project file by same name already exists.");

				// Create new directory for project file and set as current directory
				String new_directory = bodhidharma.utilities.StringMethods.removeExtension(new_project_path) + File.separator;
				current_directory_text_field.setText(new_directory);
				current_directory = current_directory_text_field.getText();
				File new_directory_file = new File(new_directory);
				if (new_directory_file.exists())
					throw new Exception("Directory called " + new_directory + " already exists.");
				new_directory_file.mkdir();

				// Ask user if wishes to keep pre-existing paths
				int copy = JOptionPane.showConfirmDialog(null, "Do you wish to copy over previous configuration files?", "New Project", JOptionPane.YES_NO_OPTION);
				
				// Copy files over to new directory if this option selected
				if (copy == JOptionPane.YES_OPTION)
				{
					// Find new paths
					String new_taxonomy_path = new_directory + bodhidharma.utilities.StringMethods.convertFilePathToFileName(taxonomy_text_field.getText());
					String new_features_path = new_directory + bodhidharma.utilities.StringMethods.convertFilePathToFileName(features_text_field.getText());
					String new_preferences_path = new_directory + bodhidharma.utilities.StringMethods.convertFilePathToFileName(preferences_text_field.getText());
					String new_recordings_path = new_directory + bodhidharma.utilities.StringMethods.convertFilePathToFileName(recordings_text_field.getText());
					String new_classifiers_path = new_directory + bodhidharma.utilities.StringMethods.convertFilePathToFileName(classifiers_text_field.getText());

					// Copy files and store names of new files
					if (!taxonomy_text_field.getText().equals(""))
					{
						bodhidharma.utilities.FileMethods.copyFile(taxonomy_text_field.getText(), new_taxonomy_path);
						taxonomy_text_field.setText(new_taxonomy_path);
					}
					if (!features_text_field.getText().equals(""))
					{
						bodhidharma.utilities.FileMethods.copyFile(features_text_field.getText(), new_features_path);
						features_text_field.setText(new_features_path);
					}
					if (!preferences_text_field.getText().equals(""))
					{
						bodhidharma.utilities.FileMethods.copyFile(preferences_text_field.getText(), new_preferences_path);
						preferences_text_field.setText(new_preferences_path);
					}
					if (!recordings_text_field.getText().equals(""))
					{
						// Copy recordings file
						bodhidharma.utilities.FileMethods.copyFile(recordings_text_field.getText(), new_recordings_path);

						// Copy feature values file too, if exists
						String values_file_path = recordings_text_field.getText().substring(0, recordings_text_field.getText().length() - 4) +  "_feature_values.xml";
						File values_file = new File(values_file_path);
						if (values_file.exists())
						{
							String new_feature_values_path = new_directory + bodhidharma.utilities.StringMethods.convertFilePathToFileName(values_file_path);
							bodhidharma.utilities.FileMethods.copyFile(values_file_path, new_feature_values_path);
						}

						// Update text field
						recordings_text_field.setText(new_recordings_path);
					}
					if (!classifiers_text_field.getText().equals(""))
					{
						bodhidharma.utilities.FileMethods.copyFile(classifiers_text_field.getText(), new_classifiers_path);
						classifiers_text_field.setText(new_classifiers_path);
					}
				}
				
				// Set configuration files to none if this option selected
				else
				{
					taxonomy_text_field.setText("");
					features_text_field.setText("");
					classifiers_text_field.setText("");
					recordings_text_field.setText("");
					preferences_text_field.setText("");
				}

				// Update stored paths
				taxonomy_path = taxonomy_text_field.getText();
				features_path = features_text_field.getText();
				recordings_list_path = recordings_text_field.getText();
				classifiers_path = classifiers_text_field.getText();
				preferences_path = preferences_text_field.getText();

				// Save the actual project file
				project_text_field.setText(new_project_path);
				project_path = project_text_field.getText();
				saveProjectFile(new_project_path, "", false);
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			}
		}
	}


	/**
	 * Gets a File object based on the given path. If the can_erase
	 * parameter is false, then gives a warning message asking user if s/he
	 * wishes to overwrite the file. Returns null if choice is to not overwrite
	 * file. Attempts to write empty string and displays error message if cannot
	 * (also returns null in this case).
	 *
	 * @param	path		The path under which the file is to be saved.
	 * @param	can_erase	Whether or not the file should be automatically overwritten if it already exists.
	 */
	public File getFile(String path, boolean can_erase)
	{
		// Check to see if should overwrite a file
		boolean go_ahead = true;
		File to_file = new File(path);
		if (to_file.exists() && can_erase == false)
		{
			int response = JOptionPane.showConfirmDialog(ConfigureFileLocationsJFrame.this, "A file " +
				"with the specified path already exists. Do you wish to overwrite it?", "Warning",
				JOptionPane.OK_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE);
			if (response != JOptionPane.YES_OPTION)
				go_ahead = false;
		}

		// Check that can write to file and eturn appropriate value 
		if (go_ahead == true)
		{
			try
			{
				FileOutputStream to = new FileOutputStream(to_file);
				DataOutputStream writer = new DataOutputStream(to);
				writer.writeBytes("");
				return to_file;
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(null, "Unable to write file.", "ERROR", JOptionPane.ERROR_MESSAGE);
				// e.printStackTrace();
				return null;
			}
		}
		
		// Return indication that can or should not write to file
		return null;
	}


	/**
	 * Perform feature extraction and saves the results. Only performed
	 * if features not already extracted.
	 */
	public void commenceFeatureExtraction()
	{
		if (main_window == null)
			main_window = new MainJFrame(this);
		main_window.commenceFeatureExtraction();
	}
	

	/**
	 * Commences training on the recordings specified in this
	 * ConfigureFileLocationsJFrame. Saves the trained classifiers
	 * in the specified location (this overrides settings in this
	 * ConfigureFileLocationsJFrame). Uses the preferences and taxonomy
	 * specified in this ConfigureFileLocationsJFrame. This method is
	 * designed to be used with the command line interface of Bodhidharma.
	 *
	 * @param	train_save_path	The path to save the trained classifiers.
	 *							If null, then saves them to the path
	 *							specified in this ConfigureFileLocationsJFrame.
	 */
	public void commenceTraining(String train_save_path)
	{
		if (main_window == null)
			main_window = new MainJFrame(this);
		if (train_save_path != null)
			setClassifiersPath(train_save_path);
		main_window.commenceTraining();
	}


	/**
	 * Commences classifying the recordings specified in this
	 * ConfigureFileLocationsJFrame using the classifiers
	 * specified in the ConfigureFileLocationsJFrame and the
	 * preferences and taxonomy specified in this
	 * ConfigureFileLocationsJFrame. The classificatoins are saved
	 * to the path in the argument. This method is designed to be
	 * used with the command line interface of Bodhidharma.
	 *
	 * @param	output_classifications_path	The path to save the output
	 *										classifications.
	 */
	public void commenceClassifiying(String output_classifications_path)
	{
		if (main_window == null)
			main_window = new MainJFrame(this);
		main_window.commenceClassifiying(output_classifications_path);
	}

	
	/* PRIVATE METHODS *********************************************************/


	/**
	 * Checks to see if the default_file_locations.xml file exists
	 * in the directory that the software was run in. If it does,
	 * checks its validity and loads it into the dialog box. Then
	 * loads the project file it refers to (other fields ignored). If not, 
	 * generates it based on hardcoded defaults and loads it.
	 */
	private void initializeDefaults()
	{
		if ( !((new File("." + File.separator + "default_file_locations.xml")).exists()) )
		{
			restoreDefaultPaths();
			saveProjectFile("." + File.separator + "default_file_locations.xml", "This project_file is a dummy used to specify the defualt project and related file locations loaded when the Bodhidharma software is run. It is not truly a project itself.", true);
		}
		loadProject("." + File.separator + "default_file_locations.xml", false);
		if (!project_path.equals(""))
			loadProject(project_text_field.getText(), false);
	}


	// Clear all text fields
	private void clearPaths()
	{
		project_text_field.setText(new String(""));
		taxonomy_text_field.setText(new String(""));
		recordings_text_field.setText(new String(""));
		features_text_field.setText(new String(""));
		classifiers_text_field.setText(new String(""));
		preferences_text_field.setText(new String(""));
		current_directory_text_field.setText(new String(""));
	}
	
		
	// Sets the values of the text fields to their hard coded defaults.
	private void restoreDefaultPaths()
	{
		project_text_field.setText(new String("Projects" +
			File.separator + "T_9_Trained_With_Recordings.xml"));
		taxonomy_text_field.setText(new String("Projects" +
			File.separator + "T_9_Trained_With_Recordings" + File.separator + "T_9.xml"));
		recordings_text_field.setText(new String("Projects" +
			File.separator + "T_9_Trained_With_Recordings" + File.separator + "T_9_recordings.xml"));
		features_text_field.setText(new String("Projects" +
			File.separator + "T_9_Trained_With_Recordings" + File.separator + "all_features.xml"));
		classifiers_text_field.setText(new String("Projects" +
			File.separator + "T_9_Trained_With_Recordings" + File.separator + "T_9_classifiers.xml"));
		preferences_text_field.setText(new String("Projects" +
			File.separator + "T_9_Trained_With_Recordings" + File.separator + "T_9_preferences.xml"));
		current_directory_text_field.setText( new String("Projects" +
			File.separator + "T_9_Trained_With_Recordings" + File.separator));
	}


	// Cause the stored paths to correspond with the contents of the text
	// fields.
	private void readTextFields()
	{
		project_path = project_text_field.getText();
		taxonomy_path = taxonomy_text_field.getText();
		recordings_list_path = recordings_text_field.getText();
		features_path = features_text_field.getText();
		classifiers_path = classifiers_text_field.getText();
		preferences_path = preferences_text_field.getText();
		current_directory = current_directory_text_field.getText();
	}


	// Returns a new JFileChooser that only displays XML files and is in the current directory.
	private JFileChooser getNewJFileChooser()
	{
		JFileChooser dialog = new JFileChooser();
		dialog.setFileFilter(new FileFilterXML());
		if (current_directory.equals(""))
			dialog.setCurrentDirectory(new File("."));
		else
			dialog.setCurrentDirectory(new File(current_directory));

		return dialog;
	}


	// React to closing this window or pressing the cancel button. If OK has not
	// ever been pressed, then quit the Java Virtual Machine. If OK has been pressed,
	// then restore the text fields to their values before this window was shown
	// last and hide the window.
	private void cancel()
	{
		if (main_window_has_been_displayed == false)
			System.exit(0);
		else
		{
			project_text_field.setText(project_path);
			taxonomy_text_field.setText(taxonomy_path);
			recordings_text_field.setText(recordings_list_path);
			features_text_field.setText(features_path);
			classifiers_text_field.setText(classifiers_path);
			preferences_text_field.setText(preferences_path);
			current_directory_text_field.setText(current_directory);

			ConfigureFileLocationsJFrame.this.hide();
			main_window.show();
		}	
	}

	
	// Take the actions related to pressing the OK button.
	private void ok()
	{
		readTextFields();

		if (main_window_has_been_displayed)
		{
			int response = JOptionPane.showConfirmDialog(ConfigureFileLocationsJFrame.this,
				"This will erase any unsaved changes you have made. Do you wish to proceed?",
				"Warning", JOptionPane.OK_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE);
			if (response == JOptionPane.YES_OPTION)
			{
				main_window.hide();
				ConfigureFileLocationsJFrame.this.hide();
				main_window = new MainJFrame(this);
				main_window.show();
			}
		}
		else
		{
			ConfigureFileLocationsJFrame.this.hide();
			main_window_has_been_displayed = true;
			main_window = new MainJFrame(this);
			main_window.show();
		}
	}
}