/*
 * @(#)FeaturesPanel.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;
import java.io.*;
import bodhidharma.xml_parsing.*;
import bodhidharma.data_structures.*;
import bodhidharma.midi_parsing.*;
import bodhidharma.midi_parsing.multi_dimensional_features.*;
import bodhidharma.midi_parsing.one_dimensional_features.*;
import bodhidharma.utilities.*;

/**
 * Part of the Bodhidharma automatic music classification
 * project. This is the panel of the GUI that allows users
 * to see the available features and choose which ones are
 * to be extracted from recordings, which ones are to be
 * candidates for use in classification during feature selection
 * and whiether or not individual features are to be used
 * for classification in overide mode.
 *
 * <p>This class does not extract or store actual feature values.
 * These are both done by the <code>RecordingsPanel</code> class.
 * Objects of the <code>RecordingsPanel</code> class do, however,
 * access the <code>feature_settings</code> field of this class
 * in order to determine which features to extract.
 *
 * <p>The table displays all available features. These may be
 * sorted be clicking on the column heading that one wishes
 * to base the search on. A user may also see a description of
 * a feature by double clicking on it. The user may change the
 * extract, selection candidate and overide status of a feature
 * by clicking on an appropriate check box.
 *
 * <p><b>Important:</b> It should be noted that the table on this panel displays
 * ALL available features, not just the ones that may be loaded and saved
 * on <i>features_file</i> XML files. These files simply load the values
 * of the check boxes for features. Features not mentioned in these files are
 * simply set to defaults. The actual features available are hardwired into
 * the software.
 *
 * <p>It is also important to note that actual feature values are
 * not stored in objects of this class. They can be found in the <code>feature_set</code>
 * field of each <code>Recording</code> object in the <code>recordings_list</code>
 * field of the <code>Recording</code> panel.
 *
 * <p>The <i>Extract</i> setting controls whether or not a feature
 * is to be extracted from recordings the next time that an extraction 
 * is performed. The <i>Selection Candidate</i> setting controls
 * whether a feature is to be considered during feature
 * selection. The <i>Overide Status</i> setting controls whether or not
 * a feature is used for classification in overide classifications,
 * which is to say in classifications where feature selection is
 * not performed and the features given overide status (and only
 * these features) are used for classification.
 *
 * <p>The <i>Code</i> column gives the code of each feature. The
 * <i>Name</i> column gives the name of each feature. The
 * <i>Type</i> column states whether the feature is one-dimensional
 * or multi-dimensional.
 *
 * <p>The <i>View Descriptions</i> button gives a brief description
 * of all features that are highlighted.
 *
 * <p>The <i>Load Features</i> button loads a <i>features_file</i>
 * XML file into memory and displays it.
 *
 * <p>The <i>Save Features</i> button saves the current
 * contents of the <i>feature_settings</i> field into the path
 * referred to in the <code>ConfigureFileLocationsJFrame</code>
 * dialog box. The <i>Save As</i> button allows the user
 * to choose the path to which the file is to be saved and
 * updates the path in the <code>ConfigureFileLocationsJFrame</code>
 * dialog box.
 *
 * <p>The <code>getFeatureSettings</code> method can be called by
 * other objects to access the feature settings stored in objects
 * of this class.
 *
 * <p>The <code>saveFeaturesSettingsFile</code> method can be called
 * by other objects to save the contents of the <i>feature_settings</i>
 * field into the path referred to in the <code>ConfigureFileLocationsJFrame</code>
 * dialog box.
 *
 * @author Cory McKay
 * @see	ConfigureFileLocationsJFrame
 * @see MainJFrame
 * @see XMLDocumentParser
 * @see ParseFeaturesFileHandler
 * @see JTable
 * @see FeaturesTableModel
 */
public class FeaturesPanel
	extends JPanel
	implements ActionListener
{
	/* FIELDS ******************************************************************/


	// Indicates which features are to be used for which purposes
	private	FeatureSettings[]		feature_settings;

	// The order of in which to display the features_list
	private int[]					list_ordering;
	
	// Fields relating to the table used to display the contents of feature_settings
	private	JTable					features_table;
	private FeaturesTableModel		features_table_model;

	// Window that holds this JPanel
	private	MainJFrame				parent_window;

	// Display panels
	private JPanel					features_display_panel;
	private JPanel					button_panel;
	private JScrollPane				features_display_scroll_pane;

	// Displays general messages
	private JTextField				status_bar;

	// Buttons
	private JButton					view_feature_description_button;
	private JButton					load_features_settings_button;
	private JButton					save_features_settings_button;
	private JButton					save_features_settings_as_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. Configures the panel and loads the feature settings specified in the 
	 * <i>features_file</i> XML file specified in the <code>ConfigureFileLocationsJFrame</code> 
	 * object. Creates new feature_settings if a valid features file is not specified
	 * or if the specified <i>features_file</i> is invalid.
	 *
 	 * @param	parent	Name of the <code>MainJFrame</code> window that holds this panel.
	 * @see		ConfigureFileLocationsJFrame
	 * @see		MainJFrame
	 */
	public FeaturesPanel(MainJFrame parent)
	{
		// Store a pointer to the container that called this RecordingsPanel
		parent_window = parent;

		// Initialize layout settings
		setLayout(new BorderLayout());
		int horizontal_gap = 4; // horizontal space between GUI elements
		int vertical_gap = 4; // horizontal space between GUI elements

		// Set up status bar
		status_bar = new JTextField("");
		status_bar.setEditable(false);
		add(status_bar, BorderLayout.SOUTH);

		// Set up buttons
		button_panel = new JPanel(new GridLayout(18, 1, horizontal_gap, vertical_gap));
		view_feature_description_button = new JButton("View Descriptions");
		load_features_settings_button = new JButton("Load Features");
		save_features_settings_button = new JButton("Save Features");
		save_features_settings_as_button = new JButton("Save Features As");
		view_feature_description_button.addActionListener(this);
		load_features_settings_button.addActionListener(this);
		save_features_settings_button.addActionListener(this);
		save_features_settings_as_button.addActionListener(this);
		button_panel.add(view_feature_description_button);
		button_panel.add(new JLabel(""));
		button_panel.add(load_features_settings_button);
		button_panel.add(save_features_settings_button);
		button_panel.add(save_features_settings_as_button);
		add(button_panel, BorderLayout.EAST);

		// Read in the starting features settings, validate them and display them
		// If a blank string is given, then just start new features settings
		features_display_panel = null;
		newFeatureSettings(true);
		if (!(parent_window.file_locations.features_path.equals("")))
		{
			boolean worked = false;
			worked = parseFeatureSettingsFile(parent_window.file_locations.features_path);
			if (!worked)
				parent_window.file_locations.setFeaturesPath(""); // set path to blank if file incorrect
		}
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Retuan feature_settings after ensuring that it has been updated to
	 * include any changes entered on the table.
	 *
	 * @return	the contents of feature_settings
	 */
	public FeatureSettings[] getFeatureSettings()
	{
		updateFeaturesSettings();
		return feature_settings;
	}


	/**
	 * Saves a <i>features_file</i> XML file with the contents specified
	 * in the <code>feature_settings</code> field and the comments specified in
	 * the <code>comments</code> parameter.
	 *
	 * <p>If the <code>is_basic_save</code> parameter is true, then the
	 * file is saved at the path specified in the <code>ConfigureFileLocationsJFrame</code>
	 * object's <code>features_path</code> field (automatically overwriting
	 * any existing files with the same path). If the <code>is_basic_save</code>
	 * parameter is false, then the file is saved at a location chosen by the user
	 * with a <code>JFileChooser</code> dialog box.
	 *
	 * @param	comments		Any comments to be saved inside the comments element of the XML file.
	 * @param	is_basic_save	Whether this is a save (true) or save as (false) operation.
	 * @see		ConfigureFileLocationsJFrame
	 */
	public void saveFeaturesSettingsFile(String comments, boolean is_basic_save)
	{
		// Make sure that feature_settings is updated
		updateFeaturesSettings();

		// Get File to save to
		File to_save_to = null;
		if (is_basic_save && !parent_window.file_locations.features_path.equals("")) // perform a basic save operation
			to_save_to = parent_window.file_locations.getFile(parent_window.file_locations.features_path, true);
		else // perform a save as operation
		{
			if (parent_window.file_locations.features_path.equals(""))
				JOptionPane.showMessageDialog(null, "No file name is specified for saving the features.\nPlease enter one in the following dialog box.", "NOTE", JOptionPane.INFORMATION_MESSAGE);

			to_save_to = parent_window.file_locations.saveFileExternally(false, "features_file");
		}

		// Write the file to disk
		if (to_save_to != null)
		{
			try
			{
				// Prepare stream writer
				FileOutputStream to = new FileOutputStream(to_save_to);
				DataOutputStream writer = new DataOutputStream(to);

				// Write the header and the first element of the XML file
				String header = new String
				(
					"<?xml version=\"1.0\"?>\n" +
					"<!DOCTYPE features_file [\n" +
					"   <!ELEMENT features_file (comments, feature+)>\n" +
					"   <!ELEMENT comments (#PCDATA)>\n" +
					"   <!ELEMENT feature (name, extract, candidate, overide)>\n" +
					"   <!ELEMENT name (#PCDATA)>\n" +
					"   <!ELEMENT extract (#PCDATA)>\n" +
					"   <!ELEMENT candidate (#PCDATA)>\n" +
					"   <!ELEMENT overide (#PCDATA)>\n" +
					"]>\n\n" +
					"<features_file>\n\n" +
					"   <comments>" + comments + "</comments>\n\n"
				);
				writer.writeBytes(header);

				// Write the XML code to represent the contents of the feature_settings field
				for (int i = 0; i < feature_settings.length; i++)
				{
					writer.writeBytes("   <feature>\n");
					writer.writeBytes("      <name>" + feature_settings[i].getFeature().getName() + "</name>\n");
					writer.writeBytes("      <extract>" + String.valueOf(feature_settings[i].extract_feature) + "</extract>\n");
					writer.writeBytes("      <candidate>" + String.valueOf(feature_settings[i].candidate_feature) + "</candidate>\n");
					writer.writeBytes("      <overide>" + String.valueOf(feature_settings[i].overide_feature) + "</overide>\n");
					writer.writeBytes("   </feature>\n\n");
				}
				writer.writeBytes("</features_file>");

				// Close the output stream
				writer.close();
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(null, ("Unable to write file " + to_save_to.getName() + "."), "ERROR", JOptionPane.ERROR_MESSAGE);
				// e.printStackTrace();
			}
		}
	}


	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the view_feature_description_button
		if (event.getSource().equals(view_feature_description_button))
			displayFeatureDescriptions(features_table.getSelectedRows());

		// React to the load_features_settings_button
		else if (event.getSource().equals(load_features_settings_button))
		{
			boolean ok_pressed = parent_window.file_locations.browseFeaturesPath(true);
			if (ok_pressed)
			{
				String feature_settings_path = parent_window.file_locations.features_path;
				boolean worked = false;
				worked = parseFeatureSettingsFile(feature_settings_path);
				if (!worked)
					parent_window.file_locations.setFeaturesPath(""); // reset to blank path if file incorrect
			}
		}

		// React to the save_features_settings_button
		else if (event.getSource().equals(save_features_settings_button))
			saveFeaturesSettingsFile("", true);

		// React to the save_features_settings_as_button
		else if (event.getSource().equals(save_features_settings_as_button))
			saveFeaturesSettingsFile("", false);
	}

	
	/**
	 * If a column table is clicked on, this method sorts the display as well as the 
	 * feature_settings based on the column title that a user clicks on.
	 *
	 * <p>If a row is double clicked on, brings up a description of the corresponding
	 * feature.
	 */
	public void addTableMouseListener()
	{
		features_table.getTableHeader().addMouseListener(new MouseAdapter()
		{
			public void mouseClicked(MouseEvent event)
			{
				// Make sure that the feature_settings reflects any changes to the table
				updateFeaturesSettings();
					
				// Sort the table contents by calculating the correct values for list_ordering
				int column_clicked = features_table.columnAtPoint(event.getPoint());
				String[][] sorted_table = Sorter.sortTable(convertFeatureSettingsToStringTable(), column_clicked, list_ordering);
				
				// Update feature_settings
				FeatureSettings[] features_copy = new FeatureSettings[feature_settings.length];
				for (int i = 0; i < features_copy.length; i++)
					features_copy[i] = feature_settings[i];
				for (int i = 0; i < feature_settings.length; i++)
					feature_settings[i] = features_copy[list_ordering[i]];

				// Update the table display
				features_table_model.fillTable(feature_settings);
			}
		});

		features_table.addMouseListener(new MouseAdapter()
		{
			public void mouseClicked(MouseEvent event)
			{
				if (event.getClickCount() == 2)
				{
					int[] row_clicked = new int[1];
					row_clicked[0] = features_table.rowAtPoint(event.getPoint());
					displayFeatureDescriptions(row_clicked);
				}
			}
		});
	}


	/* PRIVATE METHODS *********************************************************/


	// Updates feature_settings so that any changes on the three check boxes
	// in the table are relfected in feature_settings
	private void updateFeaturesSettings()
	{
		for (int i = 0; i < feature_settings.length; i++)
		{
			feature_settings[i].extract_feature = ((Boolean) features_table_model.getValueAt(i, 0)).booleanValue();
			feature_settings[i].candidate_feature = ((Boolean) features_table_model.getValueAt(i, 1)).booleanValue();
			feature_settings[i].overide_feature = ((Boolean) features_table_model.getValueAt(i, 2)).booleanValue();
		}
	}


	// Generate a new feature_settings field with extract_feature and candidate_feature 
	// set to default_logic and overide_feature set to its opposite. Update the table
	// and the status bar.
	private void newFeatureSettings(boolean default_logic)
	{
		// The total number of features
		int number_of_features = FeatureMaker.getNumberPossibleFeatures();
		feature_settings = new FeatureSettings[number_of_features];

		// Initialize feature settings
		for (int feat = 0; feat < number_of_features; feat++)
		{
			try
			{
				Feature feature = FeatureMaker.makeFeature(Integer.toString(feat), null);
				feature_settings[feat] = new FeatureSettings(feature, default_logic, default_logic, !default_logic);
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			}
		}

		// Initialize the ordering of the list
		list_ordering = new int[feature_settings.length];
		for (int i = 0; i < feature_settings.length; i++)
			list_ordering[i] = i;

		// Update the table and the status bar
		displayFeaturesTable();
		update_status_bar();
	}


	// Updates the table and the status bar to reflect the contents of feature_settings.
	private void displayFeaturesTable()
	{
		// Remove anything on the left side of the panel
		if (features_display_panel != null)
			remove(features_display_panel);

		// Initialize features_table_model and features_table
		Object[] column_names = {new String("Extract"),
		                         new String("Selection Candidate"),
		                         new String("Overide Status"),
		                         new String("Code"),
		                         new String("Name"),
		                         new String("Type")};
		features_table_model = new FeaturesTableModel(column_names, feature_settings.length);
		features_table_model.fillTable(feature_settings);
		features_table = new JTable(features_table_model);
		addTableMouseListener();

		// Set up features_display_panel
		features_display_scroll_pane = new JScrollPane(features_table);
		features_display_panel = new JPanel(new GridLayout(1, 1));
		features_display_panel.add(features_display_scroll_pane);
		add(features_display_panel, BorderLayout.CENTER);

		// Display the table and update the status bar
		features_table_model.fireTableDataChanged();
		repaint();
		parent_window.repaint();
	}


	// Causes the status bar to display the number of features
	// that are available, including a breakdown by one-dimensional
	// and multi-dimensional characteristics.
	private void update_status_bar()
	{
		int number_one_D_features = 0;
		int number_multi_D_features = 0;

		for (int i = 0; i < feature_settings.length; i++)
		{
			if (feature_settings[i].one_dimensional_feature)
				number_one_D_features++;
			else
				number_multi_D_features++;
		}

		int total_number_features = number_one_D_features + number_multi_D_features;
		String status_text = new String(number_one_D_features + " one-dimensional " +
			"features and " + number_multi_D_features + " multi-dimensional " +
			"features available, for a total of " + total_number_features + ".");
		
		status_bar.setText(status_text);
	}


	// Parses the specified file and, if it is an appropriate features_file XML file,
	// updates feature_settings to reflect its contents and displays the results.
	// Displays an error message if it is not an appropriate XML file or if it refers
	// to unknown features. Returns true if the end of the procedure
	// is reached without an exception being thrown.
	private boolean parseFeatureSettingsFile(String file_path)
	{
		boolean succesful = false;
		
		try
		{
			// Parse the contents of the recordings_file into a FeatureSettings array
			// called settings_from_file
			Object[] results = (Object[]) XMLDocumentParser.parseXMLDocument(file_path, "features_file");
			FeatureSettings[] settings_from_file = new FeatureSettings[results.length];
			for (int i = 0; i < settings_from_file.length; i++)
				settings_from_file[i] = (FeatureSettings) results[i];

			// Ensure that all of the features referred to in settings_from_file are
			// also present in feature_settings. Throw an exception if an unknown
			// feature name is present or if a feature is null.
			for (int i = 0; i < settings_from_file.length; i++)
			{
				if (settings_from_file[i].getFeature() == null)
					throw new Exception("The file " + file_path + " contains an unknown or blank feature " +
										"name. \n\nPlease correct the file. No information from this file " +
					                    "has been kept. Feature settings will revert to defaults or " +
				                        "to previously loaded values.");
			}

			// Reset feature_settings so that any that aren't mentioned in parsed file
			// have defaults set to false
			newFeatureSettings(false);

			// Update feature_settings to reflect the data parsed from the file
			for (int i = 0; i < settings_from_file.length; i ++)
			{
				for (int j = 0; j < feature_settings.length; j++)
				{
					if ( settings_from_file[i].getFeature().getName().equals( feature_settings[j].getFeature().getName() ) )
					{
						feature_settings[j] = settings_from_file[i];
						j = feature_settings.length + 1; // exit loop
					}
				}
			}

			// Display the parsed results		
			displayFeaturesTable();

			// Note that parsing was succesful
			succesful = true;
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
		}

		return succesful;
	}


	// Return a String table ([rows][columns]) describing the contents of
	// feature_settings.
	private String[][] convertFeatureSettingsToStringTable()
	{
		String results[][] = new String[feature_settings.length][6];
		for (int row = 0; row < results.length; row++)
		{
			if (feature_settings[row].extract_feature) results[row][0] = "True";
			else results[row][0] = "False";

			if (feature_settings[row].candidate_feature) results[row][1] = "True";
			else results[row][1] = "False";

			if (feature_settings[row].overide_feature) results[row][2] = "True";
			else results[row][2] = "False";

			results[row][3] = feature_settings[row].getFeature().getCode();
			
			results[row][4] = feature_settings[row].getFeature().getName();

			if (feature_settings[row].one_dimensional_feature) results[row][5] = "One-Dimensional";
			else results[row][5] = "Multi-Dimensional";
		}

		return results;
	}


	// Display a dialog box describing the feature corresponding to each
	// of the rows in the input parameter
	private void displayFeatureDescriptions(int[] rows)
	{
		for (int i = 0; i < rows.length; i++)
			JOptionPane.showMessageDialog(null, feature_settings[rows[i]].getFeature().getDescription(), new String(feature_settings[rows[i]].getFeature().getName() + " Description"), JOptionPane.INFORMATION_MESSAGE);
	}
}