/*
 * @(#)PreferencesPanel.java.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma;

import java.io.*;
import java.util.LinkedList;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import bodhidharma.classifiers.feedforward_neural_networks.NeuralNetworkJFrame;
import bodhidharma.classifiers.genetic_algorithms.GeneticAlgorithmJFrame;
import bodhidharma.xml_parsing.*;

/**
 * Part of the Bodhidharma automatic music classification
 * project. This is the panel of the GUI that allows users
 * to set preferences.
 *
 * <p>The <i>Neural Net Preferences</i> button allows the user
 * to change the settings for neural networks. The
 * <i>Genetic Algorithm Preferences</i> button allows the user
 * to change the settings for genetic algorithms.
 *
 * <p>The <i>Load Preferences</i>, <i>Save Preferences</i> and
 * <i>Save Preferences As</i> buttons allow the user to save the
 * current preference settins.
 *
 * <p>See the descriptions of the methods below or the on-line
 * help in order to see what each preference does.
 *
 * <p>The <code>savePreferencesFile</code> method can be called
 * to save the preferences. The other public methods return
 * the values of preferences.
 *
 * <p>Note that these preferences only affect behaviour at
 * instantiation of objects, and subsequent changes will have
 * no effect on existing objects. Exception are the
 * <i>Classify Into Root Categories</i>, <i>Classify Into
 * Direct Descendants of Parents</i>, <i>Classify Into All
 * Leaf Categories</i> and <i>Classify Using Round Robin
 * Ensembles</i> preferences, which play a roll during
 * classification as well as training.
 *
 * @author Cory McKay
 * @see	BioKNearestNeighbour
 * @see	FeedForwardNerualNetwork
 * @see	ConfigureFileLocationsJFrame
 * @see MainJFrame
 */
public class PreferencesPanel
	extends JPanel
	implements ActionListener
{
	/* FIELDS ******************************************************************/


	/**
	 * The basic parameters of <i>new</i> neural networks that are to be created.
	 */
	public NeuralNetworkJFrame		network_settings;

	
	/**
	 * The basic parameters of <i>new</i> genetic algorithms that are to be created.
	 */
	public GeneticAlgorithmJFrame	ga_settings;

	
	// GUI preference elements for training. See their corresponding methods for explanations.
	private JTextArea				training_fraction_text_area;
	private	JCheckBox				use_overide_features_checkbox;
	private	JCheckBox				should_use_one_dimensional_features_in_training_checkbox;
	private	JCheckBox				should_use_multi_dimensional_features_in_training_checkbox;
	private	JCheckBox				should_BioKNNs_use_feature_selection_checkbox;
	private	JCheckBox				should_BioKNNs_use_feature_weighting_checkbox;
	private	JCheckBox				should_select_classifiers_in_each_ensemble_checkbox;
	private	JCheckBox				should_weight_classifiers_in_each_ensemble_checkbox;
	private	JTextArea				internal_training_fraction_text_area;
	private JCheckBox				should_classify_into_all_leaf_categories_checkbox;
	private JCheckBox				should_classify_into_root_categories_checkbox;
	private JCheckBox				should_classify_into_direct_descendants_of_parent_categories_checkbox;
	private JCheckBox				should_classify_using_round_robin_checkbox;

	// GUI preference elememnts for ending training
	private	JTextArea				number_GA_iterations_text_area;
	private	JTextArea				classifier_vs_feature_GA_multiplier_text_area;
	private	JCheckBox				auto_end_GA_training_checkbox;
	private	JTextArea				max_GA_training_error_for_early_ending_text_area;
	private	JTextArea				min_GA_training_iterations_for_early_ending_text_area;
	private	JTextArea				number_NN_iterations_text_area;
	private	JCheckBox				auto_NN_training_checkbox;
	private	JTextArea				max_NN_training_error_for_early_ending_text_area;
	private	JTextArea				min_NN_training_iterations_for_early_ending_text_area;

	// GUI preference elememnts for choosing classification criteria
	private	JTextArea				minimum_automatic_qualification_score_text_area;
	private	JTextArea				minimum_score_to_qualify_text_area;
	private	JTextArea				percentage_below_highest_to_qualify_text_area;
	private	JTextArea				minimum_score_to_be_second_choice_text_area;
	private	JTextArea				percentage_below_highest_to_be_second_choice_text_area;
	private	JTextArea				leaf_weighing_text_area;
	private	JTextArea				parent_weighting_text_area;
	private	JTextArea				round_robin_weighting_text_area;
	private JCheckBox				should_only_use_best_for_round_robin_checkbox;
	private	JTextArea				round_robin_divisor_text_area;
	private	JCheckBox				should_only_pick_one_winner_for_combined_classification_checkbox;

	// GUI preference elements for results reporting. See their corresponding methods for explanations.
	private	JTextArea				time_between_progress_bar_polls_text_area;
	private	JCheckBox				should_display_training_times_in_raw_seconds_checkbox;
	private	JCheckBox				should_display_classifier_settings_checkbox;
	private	JCheckBox				should_display_success_rates_for_classifiers_checkbox;
	private	JCheckBox				should_display_root_category_success_rates_checkbox;
	private	JCheckBox				should_display_confusion_matrix_checkbox;
	private	JCheckBox				should_display_list_of_elligible_categories_checkbox;
	private	JCheckBox				should_display_combined_classification_results_for_each_recording_checkbox;
	private	JCheckBox				should_display_individual_classifier_results_for_each_recording_checkbox;
	private	JCheckBox				should_report_scores_with_category_names_checkbox;
	private	JCheckBox				should_report_secondary_classification_choices_checkbox;
	private	JCheckBox				should_display_model_results_along_with_results_checkbox;
	private	JCheckBox				should_display_combined_classification_scores_for_each_recording_checkbox;
	private	JCheckBox				should_display_individual_classifier_scores_for_each_recording_checkbox;
	private	JCheckBox				should_display_model_results_for_each_recording_in_separate_table_checkbox;
	private	JTextArea				training_error_reporting_interval_text_area;
	private	JCheckBox				should_display_training_recording_names_checkbox;
	private	JCheckBox				should_display_feature_selection_weighting_training_error_checkbox;
	private	JCheckBox				should_display_multi_dim_classifier_training_error_checkbox;
	private	JCheckBox				should_display_classifier_selection_training_error_checkbox;
	private	JCheckBox				should_display_classifier_weighting_training_error_checkbox;
	private	JCheckBox				should_display_one_dimensional_features_selected_checkbox;
	private	JCheckBox				should_display_one_dimensional_feature_weightings_checkbox;
	private	JCheckBox				should_display_classifiers_selected_checkbox;
	private	JCheckBox				should_display_classifiers_weightings_checkbox;

	// Window that holds this JPanel
	public	MainJFrame				parent_window;

	// Display panels
	private JPanel					preferences_display;
	private JScrollPane				preferences_display_scroll_pane;
	private JPanel					button_panel;

	// Displays general messages
	private JTextField				status_bar;

	// Buttons
	private JButton					change_neural_net_button;
	private JButton					change_genetic_alogrithm_button;
	private JButton					load_preferences_button;
	private JButton					save_preferences_button;
	private JButton					save_preferences_as_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor that configures and instantiates the panel.
	 *
 	 * @param	parent	Name of the <code>MainJFrame</code> window that holds this panel.
	 * @see		ConfigureFileLocationsJFrame
	 * @see		MainJFrame
	 */
	public PreferencesPanel(MainJFrame parent)
	{
		// Store a pointer to the container that called this RecordingsPanel
		parent_window = parent;

		// Initialize layout settings
		setLayout(new BorderLayout());
		int horizontal_gap = 4; // horizontal space between GUI elements
		int vertical_gap = 4; // horizontal space between GUI elements

		// Set up status bar
		status_bar = new JTextField("");
		status_bar.setEditable(false);
		add(status_bar, BorderLayout.SOUTH);

		// Set up buttons
		button_panel = new JPanel(new GridLayout(18, 1, horizontal_gap, vertical_gap));
		change_neural_net_button = new JButton("Neural Net Preferences");
		change_genetic_alogrithm_button = new JButton("Genetic Algorithm Preferences");
		load_preferences_button = new JButton("Load Preferences");
		save_preferences_button = new JButton("Save Preferences");
		save_preferences_as_button = new JButton("Save Preferences As");
		change_neural_net_button.addActionListener(this);
		change_genetic_alogrithm_button.addActionListener(this);
		load_preferences_button.addActionListener(this);
		save_preferences_button.addActionListener(this);
		save_preferences_as_button.addActionListener(this);
		button_panel.add(change_neural_net_button);
		button_panel.add(change_genetic_alogrithm_button);
		button_panel.add(new JLabel());
		button_panel.add(load_preferences_button);
		button_panel.add(save_preferences_button);
		button_panel.add(save_preferences_as_button);
		add(button_panel, BorderLayout.EAST);

		// Set up training preferences selection elements
		int number_columns = 10;
		training_fraction_text_area =  new JTextArea("0.8", 1, number_columns);
		use_overide_features_checkbox = new JCheckBox();
		use_overide_features_checkbox.setSelected(false);
		should_use_one_dimensional_features_in_training_checkbox = new JCheckBox();
		should_use_one_dimensional_features_in_training_checkbox.setSelected(true);
		should_use_multi_dimensional_features_in_training_checkbox = new JCheckBox();
		should_use_multi_dimensional_features_in_training_checkbox.setSelected(true);
		should_BioKNNs_use_feature_selection_checkbox = new JCheckBox();
		should_BioKNNs_use_feature_selection_checkbox.setSelected(true);
		should_BioKNNs_use_feature_weighting_checkbox = new JCheckBox();
		should_BioKNNs_use_feature_weighting_checkbox.setSelected(false);
		should_select_classifiers_in_each_ensemble_checkbox = new JCheckBox();
		should_select_classifiers_in_each_ensemble_checkbox.setSelected(false);
		should_weight_classifiers_in_each_ensemble_checkbox = new JCheckBox();
		should_weight_classifiers_in_each_ensemble_checkbox.setSelected(true);
		internal_training_fraction_text_area = new JTextArea("0.75", 1, number_columns);
		should_classify_into_all_leaf_categories_checkbox = new JCheckBox();
		should_classify_into_all_leaf_categories_checkbox.setSelected(true);
		should_classify_into_root_categories_checkbox = new JCheckBox();
		should_classify_into_root_categories_checkbox.setSelected(true);
		should_classify_into_direct_descendants_of_parent_categories_checkbox = new JCheckBox();
		should_classify_into_direct_descendants_of_parent_categories_checkbox.setSelected(true);
		should_classify_using_round_robin_checkbox = new JCheckBox();
		should_classify_using_round_robin_checkbox.setSelected(true);

		// Setup preference for ending training
		number_GA_iterations_text_area = new JTextArea("50", 1, number_columns);
		classifier_vs_feature_GA_multiplier_text_area = new JTextArea("4.0", 1, number_columns);
		auto_end_GA_training_checkbox = new JCheckBox();
		auto_end_GA_training_checkbox.setSelected(true);
		max_GA_training_error_for_early_ending_text_area = new JTextArea("0.00001", 1, number_columns);
		min_GA_training_iterations_for_early_ending_text_area = new JTextArea("75", 1, number_columns);
		number_NN_iterations_text_area = new JTextArea("1000", 1, number_columns);
		auto_NN_training_checkbox = new JCheckBox();
		auto_NN_training_checkbox.setSelected(true);
		max_NN_training_error_for_early_ending_text_area = new JTextArea("0.0000001", 1, number_columns);
		min_NN_training_iterations_for_early_ending_text_area = new JTextArea("750", 1, number_columns);

		// Set up preference for choosing classification criteria
		minimum_automatic_qualification_score_text_area = new JTextArea("0.5", 1, number_columns);
		minimum_score_to_qualify_text_area = new JTextArea("0.25", 1, number_columns);
		percentage_below_highest_to_qualify_text_area = new JTextArea("20.0", 1, number_columns);
		minimum_score_to_be_second_choice_text_area = new JTextArea("0.2", 1, number_columns);
		percentage_below_highest_to_be_second_choice_text_area = new JTextArea("30.0", 1, number_columns);
		leaf_weighing_text_area =  new JTextArea("0.2", 1, number_columns);
		parent_weighting_text_area =  new JTextArea("0.6", 1, number_columns);
		round_robin_weighting_text_area =  new JTextArea("0.2", 1, number_columns);
		should_only_use_best_for_round_robin_checkbox = new JCheckBox();
		should_only_use_best_for_round_robin_checkbox.setSelected(true);
		round_robin_divisor_text_area = new JTextArea("1", 1, number_columns);
		should_only_pick_one_winner_for_combined_classification_checkbox = new JCheckBox();
		should_only_pick_one_winner_for_combined_classification_checkbox.setSelected(false);

		// Set up reporting preferences selection elements
		time_between_progress_bar_polls_text_area = new JTextArea("1.0", 1, number_columns);
		should_display_training_times_in_raw_seconds_checkbox = new JCheckBox();
		should_display_training_times_in_raw_seconds_checkbox.setSelected(true);
		should_display_classifier_settings_checkbox = new JCheckBox();
		should_display_classifier_settings_checkbox.setSelected(true);
		should_display_success_rates_for_classifiers_checkbox = new JCheckBox();
		should_display_success_rates_for_classifiers_checkbox.setSelected(true);
		should_display_root_category_success_rates_checkbox = new JCheckBox();
		should_display_root_category_success_rates_checkbox.setSelected(true);
		should_display_confusion_matrix_checkbox = new JCheckBox();
		should_display_confusion_matrix_checkbox.setSelected(true);
		should_display_list_of_elligible_categories_checkbox = new JCheckBox();
		should_display_list_of_elligible_categories_checkbox.setSelected(false);
		should_display_combined_classification_results_for_each_recording_checkbox = new JCheckBox();
		should_display_combined_classification_results_for_each_recording_checkbox.setSelected(true);
		should_display_individual_classifier_results_for_each_recording_checkbox = new JCheckBox();
		should_display_individual_classifier_results_for_each_recording_checkbox.setSelected(true);
		should_report_scores_with_category_names_checkbox = new JCheckBox();
		should_report_scores_with_category_names_checkbox.setSelected(true);
		should_report_secondary_classification_choices_checkbox = new JCheckBox();
		should_report_secondary_classification_choices_checkbox.setSelected(true);
		should_display_model_results_along_with_results_checkbox = new JCheckBox();
		should_display_model_results_along_with_results_checkbox.setSelected(true);
		should_display_combined_classification_scores_for_each_recording_checkbox = new JCheckBox();
		should_display_combined_classification_scores_for_each_recording_checkbox.setSelected(true);
		should_display_individual_classifier_scores_for_each_recording_checkbox = new JCheckBox();
		should_display_individual_classifier_scores_for_each_recording_checkbox.setSelected(true);
		should_display_model_results_for_each_recording_in_separate_table_checkbox = new JCheckBox();
		should_display_model_results_for_each_recording_in_separate_table_checkbox.setSelected(true);		
		training_error_reporting_interval_text_area =  new JTextArea("5", 1, number_columns);
		should_display_training_recording_names_checkbox = new JCheckBox();
		should_display_training_recording_names_checkbox.setSelected(true);
		should_display_feature_selection_weighting_training_error_checkbox = new JCheckBox();
		should_display_feature_selection_weighting_training_error_checkbox.setSelected(true);
		should_display_multi_dim_classifier_training_error_checkbox = new JCheckBox();
		should_display_multi_dim_classifier_training_error_checkbox.setSelected(true);
		should_display_classifier_selection_training_error_checkbox = new JCheckBox();
		should_display_classifier_selection_training_error_checkbox.setSelected(true);
		should_display_classifier_weighting_training_error_checkbox = new JCheckBox();
		should_display_classifier_weighting_training_error_checkbox.setSelected(true);
		should_display_one_dimensional_features_selected_checkbox = new JCheckBox();
		should_display_one_dimensional_features_selected_checkbox.setSelected(true);
		should_display_one_dimensional_feature_weightings_checkbox = new JCheckBox();
		should_display_one_dimensional_feature_weightings_checkbox.setSelected(true);
		should_display_classifiers_selected_checkbox = new JCheckBox();
		should_display_classifiers_selected_checkbox.setSelected(true);
		should_display_classifiers_weightings_checkbox = new JCheckBox();
		should_display_classifiers_weightings_checkbox.setSelected(true);

		// Add training preferences selection elements to preferences_display
		preferences_display = new JPanel(new GridLayout(65, 2, horizontal_gap, vertical_gap));
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel("TRAININING PREFERENCES"));
		preferences_display.add(new JLabel("Fraction of Recordings to use for Training:"));
		preferences_display.add(training_fraction_text_area);
		preferences_display.add(new JLabel("Train With Overide Features:"));
		preferences_display.add(use_overide_features_checkbox);
		preferences_display.add(new JLabel("Train With One-D Features:"));
		preferences_display.add(should_use_one_dimensional_features_in_training_checkbox);
		preferences_display.add(new JLabel("Train With Multi-D Features:"));
		preferences_display.add(should_use_multi_dimensional_features_in_training_checkbox);
		preferences_display.add(new JLabel("Find 1-D FeaturesSelection:"));
		preferences_display.add(should_BioKNNs_use_feature_selection_checkbox);
		preferences_display.add(new JLabel("Find 1-D Feature Weightings:"));
		preferences_display.add(should_BioKNNs_use_feature_weighting_checkbox);
		preferences_display.add(new JLabel("Find Classifier Selection:"));
		preferences_display.add(should_select_classifiers_in_each_ensemble_checkbox);
		preferences_display.add(new JLabel("Find Classifier Weightings:"));
		preferences_display.add(should_weight_classifiers_in_each_ensemble_checkbox);
		preferences_display.add(new JLabel("Selection/Weighting Training Fraction:"));
		preferences_display.add(internal_training_fraction_text_area);
		preferences_display.add(new JLabel("Classify Into Root Categories:"));
		preferences_display.add(should_classify_into_root_categories_checkbox);
		preferences_display.add(new JLabel("Classify Into Direct Descendants of Parents:"));
		preferences_display.add(should_classify_into_direct_descendants_of_parent_categories_checkbox);
		preferences_display.add(new JLabel("Classify Into All Leaf Categories:"));
		preferences_display.add(should_classify_into_all_leaf_categories_checkbox);
		preferences_display.add(new JLabel("Classify Using Round Robin Ensembles:"));
		preferences_display.add(should_classify_using_round_robin_checkbox);

		// Add preference for ending training to preferences_display
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel("TRAINING TERMINATION PREFERENCES"));
		preferences_display.add(new JLabel("Auto End GA Training:"));
		preferences_display.add(auto_end_GA_training_checkbox);
		preferences_display.add(new JLabel("Max Number GA Epochs:"));
		preferences_display.add(number_GA_iterations_text_area);
		preferences_display.add(new JLabel("Classifier vs. Feature Iter. Multiplier:"));
		preferences_display.add(classifier_vs_feature_GA_multiplier_text_area);
		preferences_display.add(new JLabel("Max Change in GA Error for Training Abort:"));
		preferences_display.add(max_GA_training_error_for_early_ending_text_area);
		preferences_display.add(new JLabel("Min Consec GA Epochs for Training Abort:"));
		preferences_display.add(min_GA_training_iterations_for_early_ending_text_area);
		preferences_display.add(new JLabel("Auto End NN Training:"));
		preferences_display.add(auto_NN_training_checkbox);
		preferences_display.add(new JLabel("Max Number NN Iterations:"));
		preferences_display.add(number_NN_iterations_text_area);
		preferences_display.add(new JLabel("Max Change in NN Error for Training Abort:"));
		preferences_display.add(max_NN_training_error_for_early_ending_text_area);
		preferences_display.add(new JLabel("Min Consec NN Iters for Training Abort:"));
		preferences_display.add(min_NN_training_iterations_for_early_ending_text_area);

		// Add preference for choosing classification criteria to preferences_display
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel("CLASSIFICATION PREFERENCES"));
		preferences_display.add(new JLabel("Hierarchal Relative Weighting:"));
		preferences_display.add(parent_weighting_text_area);
		preferences_display.add(new JLabel("All Leaf Category Relavie Weighting:"));
		preferences_display.add(leaf_weighing_text_area);
		preferences_display.add(new JLabel("Round Robin Relative Weightings:"));
		preferences_display.add(round_robin_weighting_text_area);
		preferences_display.add(new JLabel("Round Robin Score Divisor:"));
		preferences_display.add(round_robin_divisor_text_area);
		preferences_display.add(new JLabel("Should Use Winner Only For Round Robin:"));
		preferences_display.add(should_only_use_best_for_round_robin_checkbox);
		preferences_display.add(new JLabel("Min Score for Auto 1st Choice:"));
		preferences_display.add(minimum_automatic_qualification_score_text_area);
		preferences_display.add(new JLabel("Min Score to be 1st Choice:"));
		preferences_display.add(minimum_score_to_qualify_text_area);
		preferences_display.add(new JLabel("% Below Highest to be 1st Choice:"));
		preferences_display.add(percentage_below_highest_to_qualify_text_area);
		preferences_display.add(new JLabel("Min Score to be 2nd Choice:"));
		preferences_display.add(minimum_score_to_be_second_choice_text_area);
		preferences_display.add(new JLabel("% Below Highest to be 2nd Choice:"));
		preferences_display.add(percentage_below_highest_to_be_second_choice_text_area);
		preferences_display.add(new JLabel("Only Choose One Winner in Combined Classification:"));
		preferences_display.add(should_only_pick_one_winner_for_combined_classification_checkbox);

		// Add reporting preferences selection elements to preferences_display
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel(""));
		preferences_display.add(new JLabel("REPORTING PREFERENCES"));
		preferences_display.add(new JLabel("Seconds Between Progress Bar Polls:"));
		preferences_display.add(time_between_progress_bar_polls_text_area);
		preferences_display.add(new JLabel("Should Display Training Times in Raw Seconds:"));
		preferences_display.add(should_display_training_times_in_raw_seconds_checkbox);
		preferences_display.add(new JLabel("Report Classifier Settings:"));
		preferences_display.add(should_display_classifier_settings_checkbox);
		preferences_display.add(new JLabel("Report Ensemble Success Rates:"));
		preferences_display.add(should_display_success_rates_for_classifiers_checkbox);
		preferences_display.add(new JLabel("Report Root Category Success Rates:"));
		preferences_display.add(should_display_root_category_success_rates_checkbox);
		preferences_display.add(new JLabel("Report Confusion Matrix:"));
		preferences_display.add(should_display_confusion_matrix_checkbox);
		preferences_display.add(new JLabel("Report Ensemble Elligible Categories:"));
		preferences_display.add(should_display_list_of_elligible_categories_checkbox);
		preferences_display.add(new JLabel("Report Combined Ensemble Classifications:"));
		preferences_display.add(should_display_combined_classification_results_for_each_recording_checkbox);
		preferences_display.add(new JLabel("Report Individual Classifier Classifications:"));
		preferences_display.add(should_display_individual_classifier_results_for_each_recording_checkbox);
		preferences_display.add(new JLabel("Include Classification Scores w/ Category Names:"));
		preferences_display.add(should_report_scores_with_category_names_checkbox);
		preferences_display.add(new JLabel("Include Secondary Classification Results:"));
		preferences_display.add(should_report_secondary_classification_choices_checkbox);
		preferences_display.add(new JLabel("Include Model Results With Classifications:"));
		preferences_display.add(should_display_model_results_along_with_results_checkbox);
		preferences_display.add(new JLabel("Report Combined Ensemble Category Scores:"));
		preferences_display.add(should_display_combined_classification_scores_for_each_recording_checkbox);
		preferences_display.add(new JLabel("Report Individual Ensemble Category Scores:"));
		preferences_display.add(should_display_individual_classifier_scores_for_each_recording_checkbox);
		preferences_display.add(new JLabel("Report Model Results in Separate Table:"));
		preferences_display.add(should_display_model_results_for_each_recording_in_separate_table_checkbox);
		preferences_display.add(new JLabel("Lines to Skip When Reporting Training Errors:"));
		preferences_display.add(training_error_reporting_interval_text_area);
		preferences_display.add(new JLabel("Report Recordings Used to Train Ensemble:"));
		preferences_display.add(should_display_training_recording_names_checkbox);
		preferences_display.add(new JLabel("Report Ensemble 1-D Feature Training Error:"));
		preferences_display.add(should_display_feature_selection_weighting_training_error_checkbox);
		preferences_display.add(new JLabel("Report Ensemble Multi-D Training Error:"));
		preferences_display.add(should_display_multi_dim_classifier_training_error_checkbox);
		preferences_display.add(new JLabel("Report Ensemble Classifier Selection Error:"));
		preferences_display.add(should_display_classifier_selection_training_error_checkbox);
		preferences_display.add(new JLabel("Report Ensemble Classifier Weighting Error:"));
		preferences_display.add(should_display_classifier_weighting_training_error_checkbox);
		preferences_display.add(new JLabel("Report Ensemble 1-D Features Selected:"));
		preferences_display.add(should_display_one_dimensional_features_selected_checkbox);
		preferences_display.add(new JLabel("Report Ensemble 1-D Feature Weightings:"));
		preferences_display.add(should_display_one_dimensional_feature_weightings_checkbox);
		preferences_display.add(new JLabel("Report Ensemble Classifiers Selected:"));
		preferences_display.add(should_display_classifiers_selected_checkbox);
		preferences_display.add(new JLabel("Report Ensemble Classifier Weightings:"));
		preferences_display.add(should_display_classifiers_weightings_checkbox);

		// Set up results_display
		preferences_display_scroll_pane = new JScrollPane(preferences_display);
		add(preferences_display_scroll_pane, BorderLayout.CENTER);

		// Instantiate neural network and genetic algorithm settings dialog boxes
		network_settings = new NeuralNetworkJFrame(this);
		ga_settings = new GeneticAlgorithmJFrame(this);

		// Read in the starting preferences, validate them and display them
		if (!(parent_window.file_locations.preferences_path.equals("")))
			parsePreferencesFile();
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the change_neural_net_button by opening a NN dialog box
		// and making the main window inaccessible
		if (event.getSource().equals(change_neural_net_button))
		{
			network_settings.pack();
			network_settings.show();
			parent_window.setEnabled(false);
		}

		// React to the change_genetic_alogrithm_button by opening a GA dialog box
		// and making the main window inaccessible
		else if (event.getSource().equals(change_genetic_alogrithm_button))
		{
			ga_settings.pack();
			ga_settings.show();
			parent_window.setEnabled(false);
		}

		// React to the load_preferences_button
		else if (event.getSource().equals(load_preferences_button))
		{
			boolean ok_pressed = parent_window.file_locations.browsePreferencesPath(true);
			if (ok_pressed)
				parsePreferencesFile();
		}

		// React to the save_preferences_button
		else if (event.getSource().equals(save_preferences_button))
			savePreferencesFile("", true);

		// React to the save_preferences_as_button
		else if (event.getSource().equals(save_preferences_as_button))
			savePreferencesFile("", false);
	}


	/**
	 * Saves a <i>preferences_file</i> XML file with the contents specified
	 * in the current fields.
	 *
	 * <p>If the <code>is_basic_save</code> parameter is true, then the
	 * file is saved at the path specified in the <code>ConfigureFileLocationsJFrame</code>
	 * object's <code>preferences_path</code> field (automatically overwriting
	 * any existing files with the same path). If the <code>is_basic_save</code>
	 * parameter is false, then the file is saved at a location chosen by the user
	 * with a <code>JFileChooser</code> dialog box.
	 *
	 * @param	comments		Any comments to be saved inside the comments element of the XML file.
	 * @param	is_basic_save	Whether this is a save (true) or save as (false) operation.
	 * @see		ConfigureFileLocationsJFrame
	 */
	public void savePreferencesFile(String comments, boolean is_basic_save)
	{
		// Get File to save to
		File to_save_to = null;
		if (is_basic_save && !parent_window.file_locations.preferences_path.equals("")) // perform a basic save operation
			to_save_to = parent_window.file_locations.getFile(parent_window.file_locations.preferences_path, true);
		else // perform a save as operation
		{
			if (parent_window.file_locations.preferences_path.equals(""))
				JOptionPane.showMessageDialog(null, "No file name is specified for saving the preferences.\nPlease enter one in the following dialog box.", "NOTE", JOptionPane.INFORMATION_MESSAGE);

			to_save_to = parent_window.file_locations.saveFileExternally(false, "preferences_file");
		}

		// Write the file to disk
		if (to_save_to != null)
		{
			try
			{
				// Prepare stream writer
				FileOutputStream to = new FileOutputStream(to_save_to);
				DataOutputStream writer = new DataOutputStream(to);

				// Write the header and the first element of the XML file
				writer.writeBytes
				(
					"<?xml version=\"1.0\"?>\n" +
					"<!DOCTYPE preferences_file [\n" +
					"   <!ELEMENT preferences_file (comments, neural_nets, genetic_algorithms, training, end_training, classification, reports)>\n" +
					"   <!ELEMENT neural_nets (hidden_unit_coefficient, learning_rate, momentum, min_weight, max_weight)>\n" +
					"   <!ELEMENT genetic_algorithms (word_length, population_size, mating_rate, mutation_probability, elitism, number_villages, migration_delay, migration_size)>\n" +
					"   <!ELEMENT training (training_fraction_text_area, use_overide_features_checkbox, should_use_one_dimensional_features_in_training_checkbox, should_use_multi_dimensional_features_in_training_checkbox, should_BioKNNs_use_feature_selection_checkbox, should_BioKNNs_use_feature_weighting_checkbox, should_select_classifiers_in_each_ensemble_checkbox, should_weight_classifiers_in_each_ensemble_checkbox, internal_training_fraction_text_area, should_classify_into_all_leaf_categories_checkbox, should_classify_into_root_categories_checkbox, should_classify_into_direct_descendants_of_parent_categories_checkbox, should_classify_using_round_robin_checkbox)>\n" +
					"   <!ELEMENT end_training (auto_end_GA_training_checkbox, number_GA_iterations_text_area, classifier_vs_feature_GA_multiplier_text_area, max_GA_training_error_for_early_ending_text_area, min_GA_training_iterations_for_early_ending_text_area, auto_NN_training_checkbox, number_NN_iterations_text_area, max_NN_training_error_for_early_ending_text_area, min_NN_training_iterations_for_early_ending_text_area)>\n" +
					"   <!ELEMENT classification (parent_weighting_text_area, leaf_weighing_text_area, round_robin_weighting_text_area, round_robin_divisor_text_area, should_only_use_best_for_round_robin_checkbox, minimum_automatic_qualification_score_text_area, minimum_score_to_qualify_text_area, percentage_below_highest_to_qualify_text_area, minimum_score_to_be_second_choice_text_area, percentage_below_highest_to_be_second_choice_text_area, should_only_pick_one_winner_for_combined_classification_checkbox)>\n" +
					"   <!ELEMENT reports (time_between_progress_bar_polls_text_area, should_display_training_times_in_raw_seconds_checkbox, should_display_classifier_settings_checkbox, should_display_success_rates_for_classifiers_checkbox, should_report_scores_with_category_names_checkbox, should_report_secondary_classification_choices_checkbox, should_display_confusion_matrix_checkbox, should_display_root_category_success_rates_checkbox, should_display_list_of_elligible_categories_checkbox, should_display_combined_classification_results_for_each_recording_checkbox, should_display_individual_classifier_results_for_each_recording_checkbox, should_display_model_results_along_with_results_checkbox, should_display_combined_classification_scores_for_each_recording_checkbox, should_display_individual_classifier_scores_for_each_recording_checkbox, should_display_model_results_for_each_recording_in_separate_table_checkbox, training_error_reporting_interval_text_area,  should_display_training_recording_names_checkbox, should_display_feature_selection_weighting_training_error_checkbox, should_display_multi_dim_classifier_training_error_checkbox, should_display_classifier_selection_training_error_checkbox, should_display_classifier_weighting_training_error_checkbox, should_display_one_dimensional_features_selected_checkbox, should_display_one_dimensional_feature_weightings_checkbox, should_display_classifiers_selected_checkbox, should_display_classifiers_weightings_checkbox)>\n" +
					"   <!ELEMENT comments (#PCDATA)>\n" +
					"   <!ELEMENT hidden_unit_coefficient (#PCDATA)>\n" +
					"   <!ELEMENT learning_rate (#PCDATA)>\n" +
					"   <!ELEMENT momentum (#PCDATA)>\n" +
					"   <!ELEMENT min_weight (#PCDATA)>\n" +
					"   <!ELEMENT max_weight (#PCDATA)>\n" +
					"   <!ELEMENT word_length (#PCDATA)>\n" +
					"   <!ELEMENT population_size (#PCDATA)>\n" +
					"   <!ELEMENT mating_rate (#PCDATA)>\n" +
					"   <!ELEMENT mutation_probability (#PCDATA)>\n" +
					"   <!ELEMENT elitism (#PCDATA)>\n" +
					"   <!ELEMENT number_villages (#PCDATA)>\n" +
					"   <!ELEMENT migration_delay (#PCDATA)>\n" +
					"   <!ELEMENT migration_size (#PCDATA)>\n" +
					"   <!ELEMENT training_fraction_text_area (#PCDATA)>\n" +
					"   <!ELEMENT use_overide_features_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_use_one_dimensional_features_in_training_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_use_multi_dimensional_features_in_training_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_BioKNNs_use_feature_selection_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_BioKNNs_use_feature_weighting_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_select_classifiers_in_each_ensemble_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_weight_classifiers_in_each_ensemble_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT internal_training_fraction_text_area (#PCDATA)>\n" +
					"   <!ELEMENT should_classify_into_all_leaf_categories_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_classify_into_root_categories_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_classify_into_direct_descendants_of_parent_categories_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_classify_using_round_robin_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT auto_end_GA_training_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT number_GA_iterations_text_area (#PCDATA)>\n" +
					"   <!ELEMENT classifier_vs_feature_GA_multiplier_text_area (#PCDATA)>\n" +
					"   <!ELEMENT max_GA_training_error_for_early_ending_text_area (#PCDATA)>\n" +
					"   <!ELEMENT min_GA_training_iterations_for_early_ending_text_area (#PCDATA)>\n" +
					"   <!ELEMENT auto_NN_training_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT number_NN_iterations_text_area (#PCDATA)>\n" +
					"   <!ELEMENT max_NN_training_error_for_early_ending_text_area (#PCDATA)>\n" +
					"   <!ELEMENT min_NN_training_iterations_for_early_ending_text_area (#PCDATA)>\n" +
					"   <!ELEMENT parent_weighting_text_area (#PCDATA)>\n" +
					"   <!ELEMENT leaf_weighing_text_area (#PCDATA)>\n" +
					"   <!ELEMENT round_robin_weighting_text_area (#PCDATA)>\n" +
					"   <!ELEMENT round_robin_divisor_text_area (#PCDATA)>\n" +
					"   <!ELEMENT should_only_use_best_for_round_robin_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT minimum_automatic_qualification_score_text_area (#PCDATA)>\n" +
					"   <!ELEMENT minimum_score_to_qualify_text_area (#PCDATA)>\n" +
					"   <!ELEMENT percentage_below_highest_to_qualify_text_area (#PCDATA)>\n" +
					"   <!ELEMENT minimum_score_to_be_second_choice_text_area (#PCDATA)>\n" +
					"   <!ELEMENT percentage_below_highest_to_be_second_choice_text_area (#PCDATA)>\n" +
					"   <!ELEMENT should_only_pick_one_winner_for_combined_classification_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT time_between_progress_bar_polls_text_area (#PCDATA)>\n" +
					"   <!ELEMENT should_display_training_times_in_raw_seconds_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_classifier_settings_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_success_rates_for_classifiers_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_report_scores_with_category_names_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_report_secondary_classification_choices_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_confusion_matrix_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_root_category_success_rates_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_list_of_elligible_categories_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_combined_classification_results_for_each_recording_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_individual_classifier_results_for_each_recording_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_model_results_along_with_results_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_combined_classification_scores_for_each_recording_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_individual_classifier_scores_for_each_recording_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_model_results_for_each_recording_in_separate_table_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT training_error_reporting_interval_text_area (#PCDATA)>\n" +
					"   <!ELEMENT should_display_training_recording_names_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_feature_selection_weighting_training_error_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_multi_dim_classifier_training_error_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_classifier_selection_training_error_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_classifier_weighting_training_error_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_one_dimensional_features_selected_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_one_dimensional_feature_weightings_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_classifiers_selected_checkbox (#PCDATA)>\n" +
					"   <!ELEMENT should_display_classifiers_weightings_checkbox (#PCDATA)>\n" +
					"]>\n\n" +
					"<preferences_file>\n\n" +
					"   <comments>" + comments + "</comments>\n\n"	
				);

				// Write the preferences
				writer.writeBytes("<neural_nets>\n");
				writer.writeBytes("   <hidden_unit_coefficient>" + network_settings.getHiddenUnitCoefficient() + "</hidden_unit_coefficient>\n");
				writer.writeBytes("   <learning_rate>" + network_settings.getLearningRate() + "</learning_rate>\n");
				writer.writeBytes("   <momentum>" + network_settings.getMomentum() + "</momentum>\n");
				writer.writeBytes("   <min_weight>" + network_settings.getMinimumInitialWeight() + "</min_weight>\n");
				writer.writeBytes("   <max_weight>" + network_settings.getMaximumInitialWeight() + "</max_weight>\n");
				writer.writeBytes("</neural_nets>\n");
				writer.writeBytes("<genetic_algorithms>\n");
				writer.writeBytes("   <word_length>" + ga_settings.getWordLength() + "</word_length>\n");
				writer.writeBytes("   <population_size>" + ga_settings.getPopulationSize() + "</population_size>\n");
				writer.writeBytes("   <mating_rate>" + ga_settings.getMatingRate() + "</mating_rate>\n");
				writer.writeBytes("   <mutation_probability>" + ga_settings.getMutationProbability() + "</mutation_probability>\n");
				writer.writeBytes("   <elitism>" + ga_settings.getElitism() + "</elitism>\n");
				writer.writeBytes("   <number_villages>" + ga_settings.getNumberVillages() + "</number_villages>\n");
				writer.writeBytes("   <migration_delay>" + ga_settings.getMigrationDelay() + "</migration_delay>\n");
				writer.writeBytes("   <migration_size>" + ga_settings.getMigrationSize() + "</migration_size>\n");
				writer.writeBytes("</genetic_algorithms>\n");
				writer.writeBytes("<training>\n");
				writer.writeBytes("   <training_fraction_text_area>" + training_fraction_text_area.getText() + "</training_fraction_text_area>\n");
				writer.writeBytes("   <use_overide_features_checkbox>" + use_overide_features_checkbox.isSelected() + "</use_overide_features_checkbox>\n");
				writer.writeBytes("   <should_use_one_dimensional_features_in_training_checkbox>" + should_use_one_dimensional_features_in_training_checkbox.isSelected() + "</should_use_one_dimensional_features_in_training_checkbox>\n");
				writer.writeBytes("   <should_use_multi_dimensional_features_in_training_checkbox>" + should_use_multi_dimensional_features_in_training_checkbox.isSelected() + "</should_use_multi_dimensional_features_in_training_checkbox>\n");
				writer.writeBytes("   <should_BioKNNs_use_feature_selection_checkbox>" + should_BioKNNs_use_feature_selection_checkbox.isSelected() + "</should_BioKNNs_use_feature_selection_checkbox>\n");
				writer.writeBytes("   <should_BioKNNs_use_feature_weighting_checkbox>" + should_BioKNNs_use_feature_weighting_checkbox.isSelected() + "</should_BioKNNs_use_feature_weighting_checkbox>\n");
				writer.writeBytes("   <should_select_classifiers_in_each_ensemble_checkbox>" + should_select_classifiers_in_each_ensemble_checkbox.isSelected() + "</should_select_classifiers_in_each_ensemble_checkbox>\n");
				writer.writeBytes("   <should_weight_classifiers_in_each_ensemble_checkbox>" + should_weight_classifiers_in_each_ensemble_checkbox.isSelected() + "</should_weight_classifiers_in_each_ensemble_checkbox>\n");
				writer.writeBytes("   <internal_training_fraction_text_area>" + internal_training_fraction_text_area.getText() + "</internal_training_fraction_text_area>\n");
				writer.writeBytes("   <should_classify_into_all_leaf_categories_checkbox>" + should_classify_into_all_leaf_categories_checkbox.isSelected() + "</should_classify_into_all_leaf_categories_checkbox>\n");
				writer.writeBytes("   <should_classify_into_root_categories_checkbox>" + should_classify_into_root_categories_checkbox.isSelected() + "</should_classify_into_root_categories_checkbox>\n");
				writer.writeBytes("   <should_classify_into_direct_descendants_of_parent_categories_checkbox>" + should_classify_into_direct_descendants_of_parent_categories_checkbox.isSelected() + "</should_classify_into_direct_descendants_of_parent_categories_checkbox>\n");
				writer.writeBytes("   <should_classify_using_round_robin_checkbox>" + should_classify_using_round_robin_checkbox.isSelected() + "</should_classify_using_round_robin_checkbox>\n");
				writer.writeBytes("</training>\n");
				writer.writeBytes("<end_training>\n");
				writer.writeBytes("   <auto_end_GA_training_checkbox>" + auto_end_GA_training_checkbox.isSelected() + "</auto_end_GA_training_checkbox>\n");
				writer.writeBytes("   <number_GA_iterations_text_area>" + number_GA_iterations_text_area.getText() + "</number_GA_iterations_text_area>\n");
				writer.writeBytes("   <classifier_vs_feature_GA_multiplier_text_area>" + classifier_vs_feature_GA_multiplier_text_area.getText() + "</classifier_vs_feature_GA_multiplier_text_area>\n");
				writer.writeBytes("   <max_GA_training_error_for_early_ending_text_area>" + max_GA_training_error_for_early_ending_text_area.getText() + "</max_GA_training_error_for_early_ending_text_area>\n");
				writer.writeBytes("   <min_GA_training_iterations_for_early_ending_text_area>" + min_GA_training_iterations_for_early_ending_text_area.getText() + "</min_GA_training_iterations_for_early_ending_text_area>\n");
				writer.writeBytes("   <auto_NN_training_checkbox>" + auto_NN_training_checkbox.isSelected() + "</auto_NN_training_checkbox>\n");
				writer.writeBytes("   <number_NN_iterations_text_area>" + number_NN_iterations_text_area.getText() + "</number_NN_iterations_text_area>\n");
				writer.writeBytes("   <max_NN_training_error_for_early_ending_text_area>" + max_NN_training_error_for_early_ending_text_area.getText() + "</max_NN_training_error_for_early_ending_text_area>\n");
				writer.writeBytes("   <min_NN_training_iterations_for_early_ending_text_area>" + min_NN_training_iterations_for_early_ending_text_area.getText() + "</min_NN_training_iterations_for_early_ending_text_area>\n");
				writer.writeBytes("</end_training>\n");
				writer.writeBytes("<classification>\n");
				writer.writeBytes("   <parent_weighting_text_area>" + parent_weighting_text_area.getText() + "</parent_weighting_text_area>\n");
				writer.writeBytes("   <leaf_weighing_text_area>" + leaf_weighing_text_area.getText() + "</leaf_weighing_text_area>\n");
				writer.writeBytes("   <round_robin_weighting_text_area>" + round_robin_weighting_text_area.getText() + "</round_robin_weighting_text_area>\n");
				writer.writeBytes("   <round_robin_divisor_text_area>" + round_robin_divisor_text_area.getText() + "</round_robin_divisor_text_area>\n");
				writer.writeBytes("   <should_only_use_best_for_round_robin_checkbox>" + should_only_use_best_for_round_robin_checkbox.isSelected() + "</should_only_use_best_for_round_robin_checkbox>\n");
				writer.writeBytes("   <minimum_automatic_qualification_score_text_area>" + minimum_automatic_qualification_score_text_area.getText() + "</minimum_automatic_qualification_score_text_area>\n");
				writer.writeBytes("   <minimum_score_to_qualify_text_area>" + minimum_score_to_qualify_text_area.getText() + "</minimum_score_to_qualify_text_area>\n");
				writer.writeBytes("   <percentage_below_highest_to_qualify_text_area>" + percentage_below_highest_to_qualify_text_area.getText() + "</percentage_below_highest_to_qualify_text_area>\n");
				writer.writeBytes("   <minimum_score_to_be_second_choice_text_area>" + minimum_score_to_be_second_choice_text_area.getText() + "</minimum_score_to_be_second_choice_text_area>\n");
				writer.writeBytes("   <percentage_below_highest_to_be_second_choice_text_area>" + percentage_below_highest_to_be_second_choice_text_area.getText() + "</percentage_below_highest_to_be_second_choice_text_area>\n");
				writer.writeBytes("   <should_only_pick_one_winner_for_combined_classification_checkbox>" + should_only_pick_one_winner_for_combined_classification_checkbox.isSelected() + "</should_only_pick_one_winner_for_combined_classification_checkbox>\n");
				writer.writeBytes("</classification>\n");
				writer.writeBytes("<reports>\n");
				writer.writeBytes("   <time_between_progress_bar_polls_text_area>" + time_between_progress_bar_polls_text_area.getText() + "</time_between_progress_bar_polls_text_area>\n");
				writer.writeBytes("   <should_display_training_times_in_raw_seconds_checkbox>" + should_display_training_times_in_raw_seconds_checkbox.isSelected() + "</should_display_training_times_in_raw_seconds_checkbox>\n");
				writer.writeBytes("   <should_display_classifier_settings_checkbox>" + should_display_classifier_settings_checkbox.isSelected() + "</should_display_classifier_settings_checkbox>\n");
				writer.writeBytes("   <should_display_success_rates_for_classifiers_checkbox>" + should_display_success_rates_for_classifiers_checkbox.isSelected() + "</should_display_success_rates_for_classifiers_checkbox>\n");
				writer.writeBytes("   <should_report_scores_with_category_names_checkbox>" + should_report_scores_with_category_names_checkbox.isSelected() + "</should_report_scores_with_category_names_checkbox>\n");
				writer.writeBytes("   <should_report_secondary_classification_choices_checkbox>" + should_report_secondary_classification_choices_checkbox.isSelected() + "</should_report_secondary_classification_choices_checkbox>\n");
				writer.writeBytes("   <should_display_confusion_matrix_checkbox>" + should_display_confusion_matrix_checkbox.isSelected() + "</should_display_confusion_matrix_checkbox>\n");
				writer.writeBytes("   <should_display_root_category_success_rates_checkbox>" + should_display_root_category_success_rates_checkbox.isSelected() + "</should_display_root_category_success_rates_checkbox>\n");
				writer.writeBytes("   <should_display_list_of_elligible_categories_checkbox>" + should_display_list_of_elligible_categories_checkbox.isSelected() + "</should_display_list_of_elligible_categories_checkbox>\n");
				writer.writeBytes("   <should_display_combined_classification_results_for_each_recording_checkbox>" + should_display_combined_classification_results_for_each_recording_checkbox.isSelected() + "</should_display_combined_classification_results_for_each_recording_checkbox>\n");
				writer.writeBytes("   <should_display_individual_classifier_results_for_each_recording_checkbox>" + should_display_individual_classifier_results_for_each_recording_checkbox.isSelected() + "</should_display_individual_classifier_results_for_each_recording_checkbox>\n");
				writer.writeBytes("   <should_display_model_results_along_with_results_checkbox>" + should_display_model_results_along_with_results_checkbox.isSelected() + "</should_display_model_results_along_with_results_checkbox>\n");
				writer.writeBytes("   <should_display_combined_classification_scores_for_each_recording_checkbox>" + should_display_combined_classification_scores_for_each_recording_checkbox.isSelected() + "</should_display_combined_classification_scores_for_each_recording_checkbox>\n");
				writer.writeBytes("   <should_display_individual_classifier_scores_for_each_recording_checkbox>" + should_display_individual_classifier_scores_for_each_recording_checkbox.isSelected() + "</should_display_individual_classifier_scores_for_each_recording_checkbox>\n");
				writer.writeBytes("   <should_display_model_results_for_each_recording_in_separate_table_checkbox>" + should_display_model_results_for_each_recording_in_separate_table_checkbox.isSelected() + "</should_display_model_results_for_each_recording_in_separate_table_checkbox>\n");
				writer.writeBytes("   <training_error_reporting_interval_text_area>" + training_error_reporting_interval_text_area.getText() + "</training_error_reporting_interval_text_area>\n");
				writer.writeBytes("   <should_display_training_recording_names_checkbox>" + should_display_training_recording_names_checkbox.isSelected() + "</should_display_training_recording_names_checkbox>\n");
				writer.writeBytes("   <should_display_feature_selection_weighting_training_error_checkbox>" + should_display_feature_selection_weighting_training_error_checkbox.isSelected() + "</should_display_feature_selection_weighting_training_error_checkbox>\n");
				writer.writeBytes("   <should_display_multi_dim_classifier_training_error_checkbox>" + should_display_multi_dim_classifier_training_error_checkbox.isSelected() + "</should_display_multi_dim_classifier_training_error_checkbox>\n");
				writer.writeBytes("   <should_display_classifier_selection_training_error_checkbox>" + should_display_classifier_selection_training_error_checkbox.isSelected() + "</should_display_classifier_selection_training_error_checkbox>\n");
				writer.writeBytes("   <should_display_classifier_weighting_training_error_checkbox>" + should_display_classifier_weighting_training_error_checkbox.isSelected() + "</should_display_classifier_weighting_training_error_checkbox>\n");
				writer.writeBytes("   <should_display_one_dimensional_features_selected_checkbox>" + should_display_one_dimensional_features_selected_checkbox.isSelected() + "</should_display_one_dimensional_features_selected_checkbox>\n");
				writer.writeBytes("   <should_display_one_dimensional_feature_weightings_checkbox>" + should_display_one_dimensional_feature_weightings_checkbox.isSelected() + "</should_display_one_dimensional_feature_weightings_checkbox>\n");
				writer.writeBytes("   <should_display_classifiers_selected_checkbox>" + should_display_classifiers_selected_checkbox.isSelected() + "</should_display_classifiers_selected_checkbox>\n");
				writer.writeBytes("   <should_display_classifiers_weightings_checkbox>" + should_display_classifiers_weightings_checkbox.isSelected() + "</should_display_classifiers_weightings_checkbox>\n");
				writer.writeBytes("</reports>\n");

				// Close the output stream
				writer.writeBytes("</preferences_file>");
				writer.close();
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(null, ("Unable to write file " + to_save_to.getName() + "."), "ERROR", JOptionPane.ERROR_MESSAGE);
				// e.printStackTrace();
			}
		}
	}


	/**
	 * Returns the fraction of candidate recordings to use for training.
	 */
	public double getTrainingFraction()
		throws Exception
	{
		double contents = Double.parseDouble(training_fraction_text_area.getText());
		if (contents < 0.0 || contents > 1.0)
			throw new Exception("You have selected a Percentage of Recordings to use for Training of " +
			                    contents + ". This value must be greater than 0 and less than 1.");
		return contents;
	}


	/**
	 * Returns true if the user has selected the option of using features marked as <i>Selection
	 * Candidates</i> in the <code>FeaturesPanel</code> for feature selection and returns false
	 * if the user has instead selected the option of using the featues marked as having
	 * <i>Overide Status</i> in the <code>FeaturesPanel</code> for feature selection.
	 */
	public boolean shouldUseFeaturesMarkedAsSelectionCandidates()
	{
		return !(use_overide_features_checkbox.isSelected());
	}


	/**
	 * Returns true if the user has selected the option of using one-dimensional features (and
	 * therefore <code>BioKNearestNeighbour</code> classifiers in training.
	 */
	public boolean shouldUseOneDimensionalFeaturesInTraining()
	{
		return should_use_one_dimensional_features_in_training_checkbox.isSelected();
	}


	/**
	 * Returns true if the user has selected the option of using multi-dimensional features (and
	 * therefore <code>FeedForwardNetworks</code> classifiers) in training.
	 */
	public boolean shouldUseMultiDimensionalFeaturesInTraining()
	{
		return should_use_multi_dimensional_features_in_training_checkbox.isSelected();
	}


	/**
	 * Returns true if the user has selected the option of having <code>BioKNearestNeighbour</code>
	 * classifiers perform feature selection.
	 */
	public boolean shouldBioKNNsUseFeatureSelection()
	{
		return should_BioKNNs_use_feature_selection_checkbox.isSelected();
	}


	/**
	 * Returns true if the user has selected the option of having <code>BioKNearestNeighbour</code>
	 * classifiers perform feature weighting.
	 */
	public boolean shouldBioKNNsUseFeatureWeighting()
	{
		return should_BioKNNs_use_feature_weighting_checkbox.isSelected();
	}


	/**
	 * Returns true if the user has selected the option of having classifier selection
	 * automatically applied to classifiers in each <code>NN_BioKNN_Ensemble</code>.
	 */
	public boolean shouldSelectClassifiersInEachEnsemble()
	{
		return should_select_classifiers_in_each_ensemble_checkbox.isSelected();
	}


	/**
	 * Returns true if the user has selected the option of having classifier weightings
	 * automatically calculated internally in each <code>NN_BioKNN_Ensemble</code>.
	 */
	public boolean shouldWeightClassifiersInEachEnsemble()
	{
		return should_weight_classifiers_in_each_ensemble_checkbox.isSelected();
	}


	/**
	 * Returns the fraction of training samples that the user has selected that will actually
	 * be used for training when feature selection and weighting are calculated for one-dimensional
	 * features. The remainder are used for testing. This value must be between 0 and 1, and an
	 * exception is thrown if it is not.
	 */
	public double getInternalTrainingFraction()
		throws Exception
	{
		double contents = Double.parseDouble(internal_training_fraction_text_area.getText());
		if (contents < 0.0 || contents > 1.0)
			throw new Exception("You have selected a Selection/Weighting Training Fraction of " +
			                    contents + ". This value must be between 0 and 1.");
		return contents;
	}


	/**
	 * Returns whether or not training/classification should be done where recordings are
	 * classified into all leaf categories.
	 */
	public boolean shouldClassifyIntoAllLeafCategories()
	{
		return should_classify_into_all_leaf_categories_checkbox.isSelected();
	}


	/**
	 * Returns whether or not training/classification should be done where recordings are
	 * classified into the root categories.
	 */
	public boolean shouldClassifyIntoRootCategories()
	{
		return should_classify_into_root_categories_checkbox.isSelected();
	}


	/**
	 * Returns whether or not training/classification should be done where recordings are
	 * classified into direct descendants of each parent category.
	 */
	public boolean shouldClassifyIntoDirectDescendantsOfParentCategories()
	{
		return should_classify_into_direct_descendants_of_parent_categories_checkbox.isSelected();
	}


	/**
	 * Returns whether or not round robin ensembles should be trained / used during classification.
	 */
	public boolean shouldClassifyUsingRoundRobin()
	{
		return should_classify_using_round_robin_checkbox.isSelected();
	}


	/**
	 * Returns the relative weighting to be given to the leaf classifier when the results
	 * of multiple classifiers are combined. This weighting is relative to the Hierarchal
	 * and Round Robin weightings. Throws an exception if a negative value is entered.
	 */
	public double getAllLeafRelativeWeighing()
		throws Exception
	{
		double contents = Double.parseDouble(leaf_weighing_text_area.getText());
		if (contents < 0)
			throw new Exception("You have selected an All Leaf Category Relavie Weighting of " +
			                    contents + ". This value must be greater or equal to 0.");
		return contents;
	}


	/**
	 * Returns the relative weighting to be given to the hierarchal classifiers when the results
	 * of multiple classifiers are combined. This weighting is relative to the Leaf
	 * and Round Robin weightings. Throws an exception if a negative value is entered.
	 */
	public double getHierarchalRelativeWeighing()
		throws Exception
	{
		double contents = Double.parseDouble(parent_weighting_text_area.getText());
		if (contents < 0)
			throw new Exception("You have selected a Hierarchal Relative Weighting of " +
			                    contents + ". This value must be greater or equal to 0.");
		return contents;
	}


	/**
	 * Returns whether or not all but the highest round robin score should be set to 0.
	 */
	public boolean shouldOnlyUseBestForRoundRobin()
	{
		return should_only_use_best_for_round_robin_checkbox.isSelected();
	}


	/**
	 * Returns the divisor that is used to divide all round robin scores during classification.
	 */
	public double getRoundRobinDivisor()
		throws Exception
	{
		double contents = Double.parseDouble(round_robin_divisor_text_area.getText());
		if (contents < 1.0)
			throw new Exception("You have selected a Hierarchal Relative Weighting of " +
			                    contents + ". This value must be greater or equal to 1.");
		return contents;
	}


	/**
	 * Returns the relative weighting to be given to the round classifiers when the results
	 * of multiple classifiers are combined. This weighting is relative to the Leaf
	 * and Hierarchal weightings. Throws an exception if a negative value is entered.
	 */
	public double getRoundRobinRelativeWeighing()
		throws Exception
	{
		double contents = Double.parseDouble(round_robin_weighting_text_area.getText());
		if (contents < 0)
			throw new Exception("You have selected a Round Robin Relative Weighting of " +
			                    contents + ". This value must be greater or equal to 0.");
		return contents;
	}


	/**
	 * Returns the maximum number of generations that should be performed for feature (but not classifier)
	 * selections and weightings. Returns a negative value if the use has selected to automatically check
	 * if training should be stopped before this number has been reached.
	 */
	public int getNumberGAIterations()
		throws Exception
	{
		int contents = Math.abs( Integer.parseInt(number_GA_iterations_text_area.getText()) );
		if (contents < 1)
			throw new Exception("You have selected an Auto End GA Training of " +
			                    contents + ". This value must be 1 or greater.");
		if (auto_end_GA_training_checkbox.isSelected())
			contents = (-1) * contents;
		return contents;
	}


	/**
	 * Returns multiplier applied to the number of GA iterations for feature selections and weightings
	 * in order to calculate classifier selections and weightings. Also applied to the population.
	 */
	public double getClassifierVsFeatureGAMultiplier()
		throws Exception
	{
		double contents = Math.abs( Double.parseDouble(classifier_vs_feature_GA_multiplier_text_area.getText()) );
		if (contents < 1.0)
			throw new Exception("You have selected a Classifier vs. Feature Iter. Multiplier of " +
			                    contents + ". This value must be 1 or greater.");
		return contents;
	}


	/**
	 * Return the maximum acceptable change in error selected by the user for GA
	 * feature selection training error rates that will allow training to be stopped early. Throws an 
	 * exception if a negative value is entered.
	 */
	public double getMaxGATrainingErrorForEarlyEnding()
		throws Exception
	{
		double contents = Double.parseDouble(max_GA_training_error_for_early_ending_text_area.getText());
		if (contents < 0.0)
			throw new Exception("You have selected a Max Change In GA Error For Training Abort value of " +
			                    contents + ". This value must be greater than 0.");
		return contents;
	}


	/**
	 * Returns the minimum number of consecutive iterations which the user has selected must go by when
	 * feature selection training a GA where the error is below the maximum
	 * error rate in order for training to stop early. Throws an exception if a negative value is entered.
	 */
	public int getMinGATrainingIterationsForEarlyEnding()
		throws Exception
	{
		int contents = Integer.parseInt(min_GA_training_iterations_for_early_ending_text_area.getText());
		if (contents < 0.0)
			throw new Exception("You have selected a Min Consec GA Epochs for Training Abort of " +
			                    contents + ". This value must be positive.");
		return contents;
	}


	/**
	 * Returns the maximum number of generations that should be performed for <code>FeedForwardNeuralNetwork</code>
	 * Return a negative value if the use has selected to automatically check if training should be stopped
	 * before this number has been reached.
	 */
	public int getNumberNNIterations()
	{
		int contents = Math.abs( Integer.parseInt(number_NN_iterations_text_area.getText()) );
		if (auto_NN_training_checkbox.isSelected())
			contents = (-1) * contents;
		return contents;
	}


	/**
	 * Return the maximum acceptable error selected by the user for <code>FeedForwardNeuralNetwork</code>
	 * training error rates that will allow training to be stopped early. Throws an exception if a negative value
	 * is entered.
	 */
	public double getMaxNNTrainingErrorForEarlyEnding()
		throws Exception
	{
		double contents = Double.parseDouble(max_NN_training_error_for_early_ending_text_area.getText());
		if (contents < 0.0)
			throw new Exception("You have selected a Max Change in NN Error for Training Abort of " +
			                    contents + ". This value must be greater than 0.");
		return contents;
	}


	/**
	 * Returns the minimum number of consecutive iterations which the user has selected must go by when
	 * training a <code>FeedForwardNeuralNetwork</code> where the error is below the maximum
	 * error rate in order for training to stop early. Throws an exception if a negative value is entered.
	 */
	public int getMinNNTrainingIterationsForEarlyEnding()
		throws Exception
	{
		int contents = Integer.parseInt(min_NN_training_iterations_for_early_ending_text_area.getText());
		if (contents < 0.0)
			throw new Exception("You have selected a Min Consec NN Iters for Training Abort of " +
			                    contents + ". This value must be greater than 0.");
		return contents;
	}


	/**
	 * Returns the interval at which training errors are reported. Throws an exception if a negative
	 * value is entered.
	 */
	public int getLinesToSkipWhenReportingTrainingErrors()
		throws Exception
	{
		int contents = Integer.parseInt(training_error_reporting_interval_text_area.getText());
		if (contents < 0)
			throw new Exception("You have selected a Lines to Skip When Reporting Training Errors of " +
			                    contents + ". This value must be greater or equal to 0.");
		return contents + 1;
	}


	/**
	 * Returns the minimum score that a category must have to automatically be
	 * counted as a first choice during classification.
	 */
	public double getMinimumAutomaticQualificationScore()
		throws Exception
	{
		double contents = Double.parseDouble(minimum_automatic_qualification_score_text_area.getText());
		if (contents < 0.0 || contents > 1.0)
			throw new Exception("You have selected a Min Score for Auto 1st Choice of " +
			                    contents + ". This value must be greater than 0 and less than 1.");
		return contents;
	}


	/**
	 * Returns the minimum score that a recording must have to qualify as a first choice
	 * during classification.
	 */
	public double getMinimumScoreToQualify()
		throws Exception
	{
		double contents = Double.parseDouble(minimum_score_to_qualify_text_area.getText());
		if (contents < 0.0 || contents > 1.0)
			throw new Exception("You have selected a Min Score to be 1st Choice of " +
			                    contents + ". This value must be greater than 0 and less than 1.");
		return contents;
	}


	/**
	 * Returns the percentage below the score of the highest performer at which
	 * a category will be classified as a first choice.
	 */
	public double getPercentageBelowHighestToQualify()
		throws Exception
	{
		double contents = Double.parseDouble(percentage_below_highest_to_qualify_text_area.getText());
		if (contents < 0.0 || contents > 100.0)
			throw new Exception("You have selected a % Below Highest to be 1st Choice of " +
			                    contents + ". This value must be greater than 0 and less than 100.");
		return contents;
	}


	/**
	 * Returns the minimum score that a recording must have to qualify as a second choice
	 * during classification.
	 */
	public double getMinimumScoreToBeSecondChoice()
		throws Exception
	{
		double contents = Double.parseDouble(minimum_score_to_be_second_choice_text_area.getText());
		if (contents < 0.0 || contents > 1.0)
			throw new Exception("You have selected a Min Score to be 2nd Choice of " +
			                    contents + ". This value must be greater than 0 and less than 1.");
		return contents;
	}


	/**
	 * Returns the percentage below the score of the highest performer at which
	 * a category will be classified as a second choice.
	 */
	public double getPercentageBelowHighestToBeSecondChoice()
		throws Exception
	{
		double contents = Double.parseDouble(percentage_below_highest_to_be_second_choice_text_area.getText());
		if (contents < 0.0 || contents > 100.0)
			throw new Exception("You have selected a % Below Highest to be 2nd Choice of " +
			                    contents + ". This value must be greater than 0 and less than 100.");
		return contents;
	}


	/**
	 * Returns whether or not only one winning category is permitted per recorcing.
	 */
	public boolean shouldOnlyPickOneWinnerForCombinedClassification()
	{
		return should_only_pick_one_winner_for_combined_classification_checkbox.isSelected();
	}


	/**
	 * Returns the time in milliseconds between progress bar polls. Defaults to ten seconds if
	 * a negative value is entered.
	 */
	public double getTimeBetweenProgressBarPolls()
	{
		double contents = Double.parseDouble(time_between_progress_bar_polls_text_area.getText());
		if (contents < 0.0)
		{
			contents = 10.0;
			time_between_progress_bar_polls_text_area.setText("10.0");
		}
		return contents * 1000.0;
	}


	/**
	 * Returns whether or not training times should be reported as raw seconds.
	 */
	public boolean shouldDisplayTrainingTimesInRawSeconds()
	{
		return should_display_training_times_in_raw_seconds_checkbox.isSelected();
	}


	/**
	 * Returns whether or not classifier settings should be included in classification reports.
	 */
	public boolean shouldDisplayClassifierSettings()
	{
		return should_display_classifier_settings_checkbox.isSelected();
	}


	/**
	 * Returns whether or not classifier scores for each winning category are reported along with the
	 * categories.
	 */
	public boolean reportScoresWithWinningCategories()
	{
		return should_report_scores_with_category_names_checkbox.isSelected();
	}


	/**
	 * Returns whether or not classifier scores for secondary choices for winning category are reported
	 * along with the winning categories.
	 */
	public boolean reportSecondaryWinningCategories()
	{
		return should_report_secondary_classification_choices_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification report should include the successful
	 * classification statistics for the combined and individual classifiers
	 * for an ensemble of classifiers.
	 */
	public boolean shouldDisplaySuccessRatesForClassifiers()
	{
		return should_display_success_rates_for_classifiers_checkbox.isSelected();
	}



	/**
	 * Returns whether or not the classification report should include a 
	 * confusion matrix.
	 */
	public boolean shouldDisplayConfusionMatrix()
	{
		return should_display_confusion_matrix_checkbox.isSelected();
	}

	
	/**
	 * Returns whether or not to display root category success rates.
	 */
	public boolean shouldDisplayRootCategorySuccessRates()
	{
		return should_display_root_category_success_rates_checkbox.isSelected();
	}

	
	/**
	 * Returns whether or not the classification report should include a list
	 * of elligible categories into which recordings can be classified by
	 * an ensemble of classifiers.
	 */
	public boolean shouldDisplayListOfElligibleCategories()
	{
		return should_display_list_of_elligible_categories_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification report should include the combined classification
	 * results (i.e. name of resulting categories) for each recording by an ensemble of classifiers.
	 */
	public boolean shouldDisplayCombinedClassificationResultsForEachRecording()
	{
		return should_display_combined_classification_results_for_each_recording_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification report should include separately, for each classifier,
	 * the classification results (i.e. name of resulting categories) for each recording by an ensemble
	 * of classifiers.
	 */
	public boolean shouldDisplayIndividualClassifierResultsForEachRecording()
	{
		return should_display_individual_classifier_results_for_each_recording_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification reports of resulting categories should also include
	 * the correct classifications, if available.
	 */
	public boolean shouldDisplayModelResultsAlongWithResults()
	{
		return should_display_model_results_along_with_results_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification report should include the combined classification
	 * scores for each category of each recording by an ensemble of classifiers.
	 */
	public boolean shouldDisplayCombinedClassificationScoresForEachRecording()
	{
		return should_display_combined_classification_scores_for_each_recording_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification report should include, separately for each
	 * classifier, the classification scores for each category of each recording by an ensemble
	 * of classifiers.
	 */
	public boolean shouldDisplayIndividualClassifierScoresForEachRecording()
	{
		return should_display_individual_classifier_scores_for_each_recording_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification report should include, as a separate table,
	 * a list of the elligible classsification categories for an ensemble of classifiers,
	 * along with their code numbers.
	 */
	public boolean shouldDisplayModelResultsForEeachRecordingInSeparateTable()
	{
		return should_display_model_results_for_each_recording_in_separate_table_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the training report should include 
	 * the names and model leaf categories of the recordings used
	 * to train it.
	 */
	public boolean shouldDisplayTrainingRecordingNames()
	{
		return should_display_training_recording_names_checkbox.isSelected();
	}
	
	
	/**
	 * Returns whether or not the classification report should include the one-dimensional feature
	 * selection or weighting training error for an ensemble of classifiers.
	 */
	public boolean shouldDisplayFeatureSelectionWeightingTrainingError()
	{
		return should_display_feature_selection_weighting_training_error_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification report should include the training error
	 * for the multi-dimensional classifiers in an ensemble of classifiers.
	 */
	public boolean shouldDisplayMultiDimClassifierTrainingError()
	{
		return should_display_multi_dim_classifier_training_error_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification report should include the classifier
	 * selection training error for an ensemble of classifiers.
	 */
	public boolean shouldDisplayClassifierSelectionTrainingError()
	{
		return should_display_classifier_selection_training_error_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the classification report should include the classifier
	 * weighting training error for an ensemble of classifiers.
	 */
	public boolean shouldDisplayClassifierWeightingTrainingError()
	{
		return should_display_classifier_weighting_training_error_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the selection/weighting report should include 
	 * the one-dimensional features selected for an ensemble of classifiers.
	 */
	public boolean shouldDisplayOneDimensionalFeaturesSelected()
	{
		return should_display_one_dimensional_features_selected_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the selection/weighting report should include 
	 * the one-dimensional feature weightings for an ensemble of classifiers.
	 */
	public boolean shouldDisplayOneDimensionalFeatureWeightings()
	{
		return should_display_one_dimensional_feature_weightings_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the selection/weighting report should include 
	 * the classifiers selected for an ensemble of classifiers.
	 */
	public boolean shouldDisplayClassifiersSelected()
	{
		return should_display_classifiers_selected_checkbox.isSelected();
	}


	/**
	 * Returns whether or not the selection/weighting report should include 
	 * the classifier weightings for an ensemble of classifiers.
	 */
	public boolean shouldDisplayClassifiersWeightings()
	{
		return should_display_classifiers_weightings_checkbox.isSelected();
	}


	/* PRIVATE METHODS *********************************************************/


	// Parse the preferences file currently in the ConfigureFileLocations object, parse it and
	// store its preferences.
	private void parsePreferencesFile()
	{
		// Find the name of the preferences file to parse
		String preferences_path = parent_window.file_locations.preferences_path;

		try
		{
			// Find the names and values of the preferences
			LinkedList[] parse_results = (LinkedList[])(XMLDocumentParser.parseXMLDocument(preferences_path, "preferences_file"));
			Object[] preference_names_objects = parse_results[0].toArray();
			Object[] preference_values_objects = parse_results[1].toArray();
			String[] preference_names = new String[preference_names_objects.length];
			String[] preference_values = new String[preference_values_objects.length];
			for (int pref = 0; pref < preference_names_objects.length; pref++)
			{
				preference_names[pref] = (String) preference_names_objects[pref];
				preference_values[pref] = (String) preference_values_objects[pref];
			}

			// Set preferences one by one
			for (int pref = 0; pref < preference_names_objects.length; pref++)
			{
				// Fill in neural network preferences
				if (preference_names[pref].equals("hidden_unit_coefficient"))
					network_settings.setHiddenUnitCoefficient(preference_values[pref]);
				else if (preference_names[pref].equals("learning_rate"))
					network_settings.setLearningRate(preference_values[pref]);
				else if (preference_names[pref].equals("momentum"))
					network_settings.setMomentum(preference_values[pref]);
				else if (preference_names[pref].equals("min_weight"))
					network_settings.setMinimumInitialWeight(preference_values[pref]);
				else if (preference_names[pref].equals("max_weight"))
					network_settings.setMaximumInitialWeight(preference_values[pref]);

				// Fill in genetic algorithm preferences
				else if (preference_names[pref].equals("word_length"))
					ga_settings.setWordLength(preference_values[pref]);
				else if (preference_names[pref].equals("population_size"))
					ga_settings.setPopulationSize(preference_values[pref]);
				else if (preference_names[pref].equals("mating_rate"))
					ga_settings.setMatingRate(preference_values[pref]);
				else if (preference_names[pref].equals("mutation_probability"))
					ga_settings.setMutationProbability(preference_values[pref]);
				else if (preference_names[pref].equals("elitism"))
					ga_settings.setElitism(preference_values[pref]);
				else if (preference_names[pref].equals("number_villages"))
					ga_settings.setNumberVillages(preference_values[pref]);
				else if (preference_names[pref].equals("migration_delay"))
					ga_settings.setMigrationDelay(preference_values[pref]);
				else if (preference_names[pref].equals("migration_size"))
					ga_settings.setMigrationSize(preference_values[pref]);

				// Fill in check boxes
				else if (preference_names[pref].equals("use_overide_features_checkbox"))
					use_overide_features_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_use_one_dimensional_features_in_training_checkbox"))
					should_use_one_dimensional_features_in_training_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_use_multi_dimensional_features_in_training_checkbox"))
					should_use_multi_dimensional_features_in_training_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_BioKNNs_use_feature_selection_checkbox"))
					should_BioKNNs_use_feature_selection_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_BioKNNs_use_feature_weighting_checkbox"))
					should_BioKNNs_use_feature_weighting_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_select_classifiers_in_each_ensemble_checkbox"))
					should_select_classifiers_in_each_ensemble_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_weight_classifiers_in_each_ensemble_checkbox"))
					should_weight_classifiers_in_each_ensemble_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_classify_into_all_leaf_categories_checkbox"))
					should_classify_into_all_leaf_categories_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_classify_into_root_categories_checkbox"))
					should_classify_into_root_categories_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_classify_into_direct_descendants_of_parent_categories_checkbox"))
					should_classify_into_direct_descendants_of_parent_categories_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_classify_using_round_robin_checkbox"))
					should_classify_using_round_robin_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("auto_end_GA_training_checkbox"))
					auto_end_GA_training_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("auto_NN_training_checkbox"))
					auto_NN_training_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_only_use_best_for_round_robin_checkbox"))
					should_only_use_best_for_round_robin_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_only_pick_one_winner_for_combined_classification_checkbox"))
					should_only_pick_one_winner_for_combined_classification_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_training_times_in_raw_seconds_checkbox"))
					should_display_training_times_in_raw_seconds_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_classifier_settings_checkbox"))
					should_display_classifier_settings_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_success_rates_for_classifiers_checkbox"))
					should_display_success_rates_for_classifiers_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_root_category_success_rates_checkbox"))
					should_display_root_category_success_rates_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_confusion_matrix_checkbox"))
					should_display_confusion_matrix_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_list_of_elligible_categories_checkbox"))
					should_display_list_of_elligible_categories_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_combined_classification_results_for_each_recording_checkbox"))
					should_display_combined_classification_results_for_each_recording_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_individual_classifier_results_for_each_recording_checkbox"))
					should_display_individual_classifier_results_for_each_recording_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_report_scores_with_category_names_checkbox"))
					should_report_scores_with_category_names_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_report_secondary_classification_choices_checkbox"))
					should_report_secondary_classification_choices_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_model_results_along_with_results_checkbox"))
					should_display_model_results_along_with_results_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_combined_classification_scores_for_each_recording_checkbox"))
					should_display_combined_classification_scores_for_each_recording_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_individual_classifier_scores_for_each_recording_checkbox"))
					should_display_individual_classifier_scores_for_each_recording_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_model_results_for_each_recording_in_separate_table_checkbox"))
					should_display_model_results_for_each_recording_in_separate_table_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_training_recording_names_checkbox"))
					should_display_training_recording_names_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_feature_selection_weighting_training_error_checkbox"))
					should_display_feature_selection_weighting_training_error_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_multi_dim_classifier_training_error_checkbox"))
					should_display_multi_dim_classifier_training_error_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_classifier_selection_training_error_checkbox"))
					should_display_classifier_selection_training_error_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_classifier_weighting_training_error_checkbox"))
					should_display_classifier_weighting_training_error_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_one_dimensional_features_selected_checkbox"))
					should_display_one_dimensional_features_selected_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_one_dimensional_feature_weightings_checkbox"))
					should_display_one_dimensional_feature_weightings_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_classifiers_selected_checkbox"))
					should_display_classifiers_selected_checkbox.setSelected(interperetBooleanString(preference_values[pref]));
				else if (preference_names[pref].equals("should_display_classifiers_weightings_checkbox"))
					should_display_classifiers_weightings_checkbox.setSelected(interperetBooleanString(preference_values[pref]));

				// Fill in text areas
				else if (preference_names[pref].equals("training_fraction_text_area"))
					training_fraction_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("internal_training_fraction_text_area"))
					internal_training_fraction_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("number_GA_iterations_text_area"))
					number_GA_iterations_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("classifier_vs_feature_GA_multiplier_text_area"))
					classifier_vs_feature_GA_multiplier_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("max_GA_training_error_for_early_ending_text_area"))
					max_GA_training_error_for_early_ending_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("min_GA_training_iterations_for_early_ending_text_area"))
					min_GA_training_iterations_for_early_ending_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("number_NN_iterations_text_area"))
					number_NN_iterations_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("max_NN_training_error_for_early_ending_text_area"))
					max_NN_training_error_for_early_ending_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("min_NN_training_iterations_for_early_ending_text_area"))
					min_NN_training_iterations_for_early_ending_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("minimum_automatic_qualification_score_text_area"))
					minimum_automatic_qualification_score_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("minimum_score_to_qualify_text_area"))
					minimum_score_to_qualify_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("percentage_below_highest_to_qualify_text_area"))
					percentage_below_highest_to_qualify_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("minimum_score_to_be_second_choice_text_area"))
					minimum_score_to_be_second_choice_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("percentage_below_highest_to_be_second_choice_text_area"))
					percentage_below_highest_to_be_second_choice_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("leaf_weighing_text_area"))
					leaf_weighing_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("parent_weighting_text_area"))
					parent_weighting_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("round_robin_weighting_text_area"))
					round_robin_weighting_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("round_robin_divisor_text_area"))
					round_robin_divisor_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("time_between_progress_bar_polls_text_area"))
					time_between_progress_bar_polls_text_area.setText(preference_values[pref]);
				else if (preference_names[pref].equals("training_error_reporting_interval_text_area"))
					training_error_reporting_interval_text_area.setText(preference_values[pref]);
			}
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
		}
	}


	// Returns true if the given String is "true" and false otherwise
	private boolean interperetBooleanString(String boolean_string)
	{
		if (boolean_string.equals("true"))
			return true;
		else return false;
	}
}