/*
 * @(#)ProgressBarTaskTrainMonitor.java	1.1.1 08/10/04
 *
 * McGill University
 */

package bodhidharma;

import javax.swing.*;
import bodhidharma.utilities.*;
import bodhidharma.data_structures.*;

/**
 * An implementation of the <code>LargeProgressBarTaskMonitor</code> abstract class.
 * Trains ensembles of classifiers and provides information about training times.
 * Provides information to a <code>LargeProgressBarDialog</code> object that is used
 * to keep it updated.
 *
 * <p>Training is started by calling the <code>go</code> method. This will cause
 * the processing in <code>ActualTask</code> to occur. <code>ActualTask</code>
 * will keep the fields described in this class' superclass updated, which can
 * be accessed using the methods in this class' superclass. The <code>ActualTask</code>
 * simply makes use of a </code>ClassificationPanel</code>'s <code>performHierarchicalTraining</code>
 * method. The </code>performAFoldOfTrainingAndClassification</code> method of the
 * </code>ClassificationPanel</code> can also be used to perform multiple folds.
 * 
 * <p>The <code>setProgressDialog</code> method of an object of this class
 * must be called externally to cancel training.
 *
 * <p>The <code>startTopTask, startMiddleTask, startBottomTask, setBottomProgressValue, </code>
 * and <code>markTaskComplete</code> methods are called during training to keep
 * an object of this class updated.
 *
 * <p>The <code>isTrainingTimesAvailable, getMiddleTaskDurations, getBottomTaskDurations</code> and
 * <code>getOverallTaskTime</code> methods can be used after training to gain information about training times.
 *
 * <p>Note that the top, middle and bottom tasks are hierarchal, in the sense that
 * one the top task is complete when all middle sub-tasks are complete, and a middle
 * sub-task is complete when all its bottom sub-tasks are complete.
 *
 * @author	Cory McKay
 * @see		LargeProgressBarTaskMonitor
 * @see		ProgressBarDialog
 * @see		SwingWorker
 * @see		Recording
 * @see		FeatureSet
 * @see		RecordingsPanel
 */
public class ProgressBarTaskTrainMonitor
	extends LargeProgressBarTaskMonitor
{
	/* FIELDS ******************************************************************/


	// Fields from calling ClassificationPanel that are used to start training
	Recording[]					all_eligible_recordings;
	boolean[]					training_key;
	boolean						report_training_results;
	boolean[][]					multi_fold_training_keys;
	int							multi_fold_current_fold;

	// Fields used to effect change in other classes
	ClassificationPanel			classification_doer;
	LargeProgressBarDialog		progress_dialog;

	// A reference to this object
	ProgressBarTaskTrainMonitor	training_monitor;

	// Fields used to store information about times for different tasks
	String[]					top_event_names;
	long[]						top_event_durations;
	String[][]					middle_event_names;
	long[][]					middle_event_durations;
	int							current_top_indice;
	int							current_middle_indice;
	long						overall_start_time;
	long						overall_end_time;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. Same as superclass, and also sets fields needed to commence
	 * training.
	 *
	 * @param	all_eligible_recordings		Recordings that will be broken up into training and testing groups.
	 * @param	training_key				Which recordings are to be used for testing and which are to be used for training. Indice corresponds to that of the all_eligible_recordings parameter. Value of true means use for training. A value of null means that breakdown is randomly selected.
	 * @param	report_training_results		Whether or not training errors (i.e. classifier convergence) are reported. Overrides preferences settings.
	 * @param	classification_doer			ClassificationPanel used to perform training.
	 * @param	multi_fold_training_keys	Keys to be used in further folds. Null if only one fold in all is used.
	 * @param	multi_fold_current_fold		The current fold being processed if there are multiple folds. Value is irrelevant if there is only one fold in all.
	 */
	public ProgressBarTaskTrainMonitor( Recording[] all_eligible_recordings,
	                                    boolean[] training_key,
	                                    boolean report_training_results,
	                                    ClassificationPanel classification_doer,
	                                    boolean[][] multi_fold_training_keys,
	                                    int multi_fold_current_fold )
	{
		super();

		this.all_eligible_recordings = all_eligible_recordings;
	    this.training_key = training_key;
	    this.report_training_results = report_training_results;
	    this.classification_doer = classification_doer;
		this.multi_fold_training_keys = multi_fold_training_keys;
		this.multi_fold_current_fold = multi_fold_current_fold;
	    training_monitor = this;
		progress_dialog = null;

		top_event_names = null;
		top_event_durations = null;
		middle_event_names = null;
		middle_event_durations = null;
		current_top_indice = 0;
		current_middle_indice = 0;
		overall_start_time = 0;
		overall_end_time = 0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Called from classification_doer to set up a <code>SwingWorker</code> thread
	 * that will initiate training.
	 */
	public void go(  )
	{
		final SwingWorker worker = new SwingWorker()
		{
			public Object construct()
			{
				return new ActualTask(  );
			}
		};

		worker.start();
	}


	/**
	 * Set the progress box so that it can be cancelled if an error occurs.
	 */
	public void setProgressDialog(LargeProgressBarDialog training_indicator)
	{
		progress_dialog = training_indicator;
	}


	/**
	 * Begin keeping track of a new overall job with the given number of sub_tasks.
	 */
	public void startTopTask(int sub_tasks, String message)
	{
		top_current_amount_completed = -1;
		top_time_started = System.currentTimeMillis();

		top_progress_message = message;
		if (multi_fold_training_keys != null)
			top_progress_message += " Fold " + (multi_fold_current_fold + 1) + " of " + multi_fold_training_keys.length;
		top_length_of_task = sub_tasks;
		new_top_task_started = true;

		// Keep track of task times for later reference
		overall_start_time = top_time_started;
		top_event_names = new String[sub_tasks];
		top_event_durations = new long[sub_tasks];
		middle_event_names = new String[top_event_names.length][];
		middle_event_durations = new long[middle_event_names.length][];
	}


	/**
	 * Begin keeping track of a new middle-level job with the given number of sub_tasks.
	 */
	public void startMiddleTask(int sub_tasks, String message)
	{
		top_current_amount_completed++;
		
		middle_current_amount_completed = -1;
		middle_time_started = System.currentTimeMillis();
		
		middle_progress_message = message;
		middle_length_of_task = sub_tasks;
		new_middle_task_started = true;

		// Keep track of task times for later reference
		top_event_names[current_top_indice] = message;
		top_event_durations[current_top_indice] = middle_time_started;
		if (current_top_indice != 0)
			top_event_durations[current_top_indice - 1] = middle_time_started - top_event_durations[current_top_indice - 1];
		if (current_middle_indice == 0)
		{
			middle_event_names[current_top_indice] = new String[sub_tasks];
			middle_event_durations[current_top_indice] = new long[sub_tasks];
		}
	}


	/**
	 * Begin keeping track of a new bottom-level job with the given number of sub_tasks.
	 */
	public void startBottomTask(int sub_tasks, String message)
	{
		middle_current_amount_completed++;
		
		bottom_current_amount_completed = -1;
		bottom_time_started = System.currentTimeMillis();
	
		bottom_progress_message = message;
		bottom_length_of_task = sub_tasks;
		new_bottom_task_started = true;

		// Keep track of task times for later reference
		middle_event_names[current_top_indice][current_middle_indice] = message;
		middle_event_durations[current_top_indice][current_middle_indice] = bottom_time_started;
		if (current_middle_indice != 0)
			middle_event_durations[current_top_indice][current_middle_indice - 1] = bottom_time_started - middle_event_durations[current_top_indice][current_middle_indice - 1];
		current_middle_indice++;
		if( current_middle_indice == middle_event_durations[current_top_indice].length &&
		    current_top_indice != (top_event_durations.length - 1) )
		{
			middle_event_durations[current_top_indice][current_middle_indice - 1] = bottom_time_started - middle_event_durations[current_top_indice][current_middle_indice - 1];
			current_middle_indice = 0;
			current_top_indice++;
		}
	}


	/**
	 * Update the progress of the bottom progress bar.
	 */
	public void setBottomProgressValue(int progress_value)
	{
		bottom_current_amount_completed = progress_value;
	}


	/**
	 * Called if the overall task is complete. Stops processing and records
	 * final task times.
	 */
	public void markTaskComplete()
	{
		// Finalize end times of tasks for later reference.
		overall_end_time = System.currentTimeMillis();
		top_event_durations[current_top_indice] = overall_end_time - top_event_durations[current_top_indice];
		if (middle_event_durations[0].length != 0)
			middle_event_durations[current_top_indice][current_middle_indice - 1] = overall_end_time - middle_event_durations[current_top_indice][current_middle_indice - 1];

		// Mark overall task as complete
		super.markTaskComplete();
	}


	/**
	 * Returns whether or training times are available.
	 */
	public boolean isTrainingTimesAvailable()
	{
		if (middle_event_durations == null)
			return false;
		if (middle_event_durations[0].length == 0)
			return false;
		else
			return true;
	}


	/**
	 * Returns the names and durations of each of the tasks appearing in
	 * the middle progress text area. Last indice value of 0 for returned array corresponds to
	 * the name of a task and value of 2 to its duration.
	 *
	 * @param	want_raw_seconds	Task times reported in seconds if true, as formatted string if false.
	 */
	public String[][] getMiddleTaskDurations(boolean want_raw_seconds)
	{
		String[][] durations = new String[top_event_durations.length][2];
		for (int task = 0; task < durations.length; task++)
		{
			durations[task][0] = top_event_names[task];
			top_event_durations[task] = top_event_durations[task] / 1000;
			if (want_raw_seconds)
				durations[task][1] = String.valueOf(top_event_durations[task]);
			else
				durations[task][1] = findTimeWithProperUnits( (new Long(top_event_durations[task])).intValue() );
		}
		return durations;
	}


	/**
	 * Returns the names and durations of each of the tasks appearing in
	 * the bottom progress text area. First indice of returned array corresponds to
	 * the high level task, and the second to the lower level task for that high level
	 * task. A value of 0 for the third indice gives name of a task and the second its duration.
	 *
	 * @param	want_raw_seconds	Task times reported in seconds if true, as formatted string if false.
	 */
	public String[][][] getBottomTaskDurations(boolean want_raw_seconds)
	{
		String[][][] durations = new String[middle_event_names.length][][];
		for (int h_task = 0; h_task < durations.length; h_task++)
		{
			durations[h_task] = new String[middle_event_durations[h_task].length][2];
			for (int l_task = 0; l_task < durations[h_task].length; l_task++)
			{
				durations[h_task][l_task][0] = middle_event_names[h_task][l_task];
				middle_event_durations[h_task][l_task] = middle_event_durations[h_task][l_task] / 1000;
				if (want_raw_seconds)
					durations[h_task][l_task][1] = String.valueOf(middle_event_durations[h_task][l_task]);
				else
					durations[h_task][l_task][1] = findTimeWithProperUnits( (new Long(middle_event_durations[h_task][l_task])).intValue() );
			}
		}
		return durations;
	}


	/**
	 * Returns the overall duration of the outer task.
	 *
	 * @param	want_raw_seconds	Time reported in seconds if true, as formatted string if false.
	 */
	public String getOverallTaskTime(boolean want_raw_seconds)
	{
		long total_duration = overall_end_time - overall_start_time;
		total_duration = total_duration / 1000;
		if (want_raw_seconds)
			return String.valueOf(total_duration);
		else
			return findTimeWithProperUnits( (new Long(total_duration)).intValue() );
	}


	/* INTERNAL CLASS **********************************************************/


	/**
	 * The actual task that is performed. This runs in a <code>SwingWorker</code> thread.
	 * See the performHierarchicalTraining method of the ClassificationPanel to see how
	 * training is performed.
	 *
	 * Actions are also taken to keep track of the feature and classifier weightings for
	 * each fold if multi-fold training/classification is being performed.
	 *
	 * If the <code>task_canceled</code> field is set to true externally then processing is
	 * terminated.
	 */
	class ActualTask
	{
		ActualTask()
		{
			try
			{
				// Train
				classification_doer.performHierarchicalTraining( all_eligible_recordings,
																 training_key,
																 report_training_results,
																 training_monitor );

				// Continue to next fold if this option is selected
				if (multi_fold_training_keys != null)
				{
					// Record the weightings for the current fold
					classification_doer.recordWeightingsAccrossFolds(multi_fold_current_fold);

					// Do next fold
					multi_fold_current_fold++;
					int total_duration = (new Long(overall_end_time - overall_start_time)).intValue() / 1000;
					classification_doer.performAFoldOfTrainingAndClassification(multi_fold_training_keys, multi_fold_current_fold, total_duration);
				}
			}
			catch (Throwable t)
			{
				// React to the Java Runtime running out of memory
				if (t.toString().equals("java.lang.OutOfMemoryError"))
					JOptionPane.showMessageDialog(null, "The Java Runtime ran out of memory. Please rerun this program with a higher amount of memory assigned to the Java Runtime heap.", "ERROR", JOptionPane.ERROR_MESSAGE);
				else
					JOptionPane.showMessageDialog(null, t.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);

				// Note that an error occured
				classification_doer.results_display.append("COULD NOT COMPLETE TRAINING DUE TO AN ERROR.");

				// Clear all stored classifier ensembles and information about them
				classification_doer.clearStoredData();
				classification_doer.clearMultiFoldStoredData();

				// Display error details
				// t.printStackTrace();

				// Mark task as cancelled
				stop();
				progress_dialog.cancel();
			}
		}
	}
}