/*
 * @(#)TaxonomyPanel.java	1.2.2 07/15/05
 *
 * McGill Univarsity
 */

package bodhidharma;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import java.io.*;
import java.util.LinkedList;
import bodhidharma.xml_parsing.*;
import bodhidharma.utilities.*;

/**
 * Part of the Bodhidharma automatic music classification
 * project. This is the panel of the GUI that allows users
 * to see, edit, load and save taxonomies that will be used
 * for classification.
 *
 * <p>The currently loaded taxonomy can be seen in one of two
 * views. The first, called the Hierarchical View, allows users
 * to see a hierarchal tree reflecting the relationships of 
 * broad categories going down to narrower categories.
 * The names of these categories may be edited by clicking
 * on them (the return button must be pressed to register
 * changes) or by using the buttons described below.
 *
 * <p>The second view, called the Ancestry View, allows
 * users to see the taxonomy in the form of a list of leaf
 * categories. Children of leaf categories represent a list
 * of all ancestors of the given category. Leaf categories
 * that appear more than once in the <code>taxonomy</code>
 * field are combined into one under this view (the
 * <code>taxonomy</code> field itself is not changed) so
 * that all of the combined ancestors are listed together.
 * The taxonomy is not editable in this view.
 *
 * <p>Both of the trees in the above two views are sorted
 * alphebetically. The <i>Display Hierarchy</i> and
 * <i>Display Ancestry</i> buttons allow users to
 * switch between the two views respectively.
 *
 * <p>The <i>New Taxonomy</i> button deletes the
 * current taxonomy and creates a new empty taxonomy.
 *
 * <p>The <i>Load Taxonomy</i> button loads a
 * <i>taxonomy_file</i> XML file into memorey and displays it.
 *
 * <p>Note that it is also possible to load tab delimited text files in the MIREX 2005 format that
 * hold taxonomies an alternative to the XML files. This functionality is intended primarily
 * for use with the command line use of Bodhidharma.
 *
 * <p>The <i>Save Taxonomy</i> button saves the current
 * contents of the <i>taxonomy</i> field into the path
 * referred to in the <i>ConfigureFileLocationsJFrame</i>
 * dialog box. The <i>Save As</i> button allows the user
 * to choose the path to which the file is to be saved and
 * updates the path in the <i>ConfigureFileLocationsJFrame</i>
 * dialog box.
 *
 * <p>The <i>Add New Sibling Category</i> and the <i>Add New
 * Child</i> buttons add a new sibling or a new child category
 * to the taxonomy respectively, with a location based on the
 * node that is currently selected. The <i>Delete Branch</i>
 * button removes the selected category and its decendants.
 *
 * <p>The <code>taxonomy</code> field can be accessed by other
 * objects in order to gain access to the taxonomy.
 *
 * <p>The <code>getLeafsAndTheirParents</code> method can also be
 * used by other objects to obtain the information of the
 * Ancestry View in the form of a 2-D array of strings. Similarly,
 * the <code>getLeafLabels/code>, <code>getParentLabels</code> and
 * <code>getMiddleLabels</code> methods can be called to get arrays
 * holding the labels of all leaf categories in the <code>taxonomy</code>
 * field, the labels of all parent categories (i.e. categories at
 * the highest level of the taxonomy) in the <code>taxonomy</code>
 * field and the labels in the <code>taxonomy</code> field that are
 * neither leaves nor parent categories resepectively.
 *
 * <p>The <code>getAllParentsAndTheirDirectChildren</code> method
 * can be called to obtain the direct descendants of each parent
 * node (whether it be at the root or intermediate).
 *
 * <p>The <code>saveTaxonomyFile</code> method can be called by
 * other objects in order to save the contents of the
 * <code>taxonomy</code> field.
 *
 * <p>The <code>isTreeEmpty</code> method can be called by
 * other objects in order to check if the<code>taxonomy</code>
 * field is empty.
 * 
 * @author Cory McKay
 * @see	ConfigureFileLocationsJFrame
 * @see MainJFrame
 * @see XMLDocumentParser
 * @see ParseTaxonomyFileHandler
 * @see JTree
 * @see DefaultTreeModel
 * @see DefaultMutableTreeNode
 */
public class TaxonomyPanel
	extends JPanel
	implements ActionListener
{
	/* FIELDS ******************************************************************/


	/**
	 * The tree holding the taxonomy
	 */
	public	DefaultTreeModel	taxonomy;

	// JTree that holds the taxonomy DefaultTreeModel
	private JTree				taxonomy_display;

	// Window that holds this panel
	private MainJFrame			parent_window;

	// Display panels
	private JPanel				tree_display_panel;
	private JPanel				ancestry_display_panel;
	private JPanel				button_panel;
	private JScrollPane			taxonomy_display_scroll_pane;

	// Displays general messages
	private JTextField			status_bar;

	// Buttons
	private JButton				display_hierarchy_button;
	private JButton				display_ancestry_button;
	private JButton				new_taxonomy_button;
	private JButton				load_taxonomy_button;
	private JButton				save_taxonomy_button;
	private JButton				save_taxonomy_as_button;
	private JButton				add_new_sibling_category_button;
	private JButton				add_new_child_category_button;
	private JButton				delete_branch_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. Configures the panel and loads the taxonomy specified in the 
	 * <i>taxonomy_file</i> XML file specified in the <code>ConfigureFileLocationsJFrame</code> 
	 * object. Creates a new empty taxonomy if a valid taxonomy file is not specified
	 * there.
	 *
 	 * @param	parent	Name of the <code>MainJFrame</code> window that holds this panel.
	 * @see		ConfigureFileLocationsJFrame
	 * @see		MainJFrame
	 */
	public TaxonomyPanel(MainJFrame parent)
	{
		// Store a pointer to the container that called this TaxonomyPanel
		parent_window = parent;

		// Initialize layout settings
		setLayout(new BorderLayout());
		int horizontal_gap = 4; // horizontal space between GUI elements
		int vertical_gap = 4; // horizontal space between GUI elements

		// Set up status bar
		status_bar = new JTextField("");
		status_bar.setEditable(false);
		add(status_bar, BorderLayout.SOUTH);

		// Set up buttons
		button_panel = new JPanel(new GridLayout(18, 1, horizontal_gap, vertical_gap));
		display_hierarchy_button = new JButton("Display Hierarchy");
		display_hierarchy_button.setEnabled(false);
		display_ancestry_button = new JButton("Display Ancestry");
		new_taxonomy_button = new JButton("New Taxonomy");
		load_taxonomy_button = new JButton("Load Taxonomy");
		save_taxonomy_button = new JButton("Save Taxonomy");
		save_taxonomy_as_button = new JButton("Save Taxonomy As");
		add_new_sibling_category_button = new JButton("Add New Sibling Category");
		add_new_child_category_button = new JButton("Add New Child Category");
		delete_branch_button = new JButton("Delete Branch");
		display_hierarchy_button.addActionListener(this);
		display_ancestry_button.addActionListener(this);
		new_taxonomy_button.addActionListener(this);
		load_taxonomy_button.addActionListener(this);
		save_taxonomy_button.addActionListener(this);
		save_taxonomy_as_button.addActionListener(this);
		add_new_sibling_category_button.addActionListener(this);
		add_new_child_category_button.addActionListener(this);
		delete_branch_button.addActionListener(this);
		button_panel.add(display_hierarchy_button);
		button_panel.add(display_ancestry_button);
		button_panel.add(new JLabel(""));
		button_panel.add(new_taxonomy_button);
		button_panel.add(load_taxonomy_button);
		button_panel.add(save_taxonomy_button);
		button_panel.add(save_taxonomy_as_button);
		button_panel.add(new JLabel(""));
		button_panel.add(add_new_sibling_category_button);
		button_panel.add(add_new_child_category_button);
		button_panel.add(delete_branch_button);
		add(button_panel, BorderLayout.EAST);

		// Read in the starting taxonomy and display it
		// If a blank string is given, then just start a new taxonomy
		if (parent_window.file_locations.taxonomy_path.equals(""))
			newTaxonomy();
		else
		{
			boolean worked = false;
			worked = parseTaxonomyFileIntoJTree(parent_window.file_locations.taxonomy_path);
			if (!worked)
				parent_window.file_locations.setTaxonomyPath(""); // set path to blank if file incorrect
		}
		tree_display_panel = null;
		ancestry_display_panel = null;
		displayTaxonomyHierarchically();
	}


	/* PUBLIC METHODS **********************************************************/

	
	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the display_taxonomy_button
		if (event.getSource().equals(display_hierarchy_button))
		{
			displayTaxonomyHierarchically();

			display_hierarchy_button.setEnabled(false);
			display_ancestry_button.setEnabled(true);
			new_taxonomy_button.setEnabled(true);
			load_taxonomy_button.setEnabled(true);
			save_taxonomy_button.setEnabled(true);
			save_taxonomy_as_button.setEnabled(true);
			add_new_sibling_category_button.setEnabled(true);
			add_new_child_category_button.setEnabled(true);
			delete_branch_button.setEnabled(true);
		}

		// React to the display_ancestry_button
		if (event.getSource().equals(display_ancestry_button))
		{
			displayAncestryTree();

			display_hierarchy_button.setEnabled(true);
			display_ancestry_button.setEnabled(false);
			new_taxonomy_button.setEnabled(false);
			load_taxonomy_button.setEnabled(false);
			save_taxonomy_button.setEnabled(false);
			save_taxonomy_as_button.setEnabled(false);
			add_new_sibling_category_button.setEnabled(false);
			add_new_child_category_button.setEnabled(false);
			delete_branch_button.setEnabled(false);
		}

		// React to the new_taxonomy_button
		else if (event.getSource().equals(new_taxonomy_button))
			newTaxonomy();

		// React to the load_taxonomy_button
		else if (event.getSource().equals(load_taxonomy_button))
		{
			boolean ok_pressed = parent_window.file_locations.browseTaxonomyPath(true);
			if (ok_pressed)
			{
				String taxonomy_path = parent_window.file_locations.taxonomy_path;
				boolean worked = false;
				worked = parseTaxonomyFileIntoJTree(taxonomy_path);
				if (!worked)
					parent_window.file_locations.setTaxonomyPath(""); // reset to blank path if file incorrect
				displayTaxonomyHierarchically();
			}
		}

		// React to the save_taxonomy_button
		else if (event.getSource().equals(save_taxonomy_button))
			saveTaxonomyFile("", true);

		// React to the save_taxonomy_as_button
		else if (event.getSource().equals(save_taxonomy_as_button))
			saveTaxonomyFile("", false);

		// React to the add_new_sibling_category_button
		else if (event.getSource().equals(add_new_sibling_category_button))
		{
			DefaultMutableTreeNode selected_node = (DefaultMutableTreeNode) taxonomy_display.getLastSelectedPathComponent();
			if (selected_node == null) return;

			DefaultMutableTreeNode new_node = new DefaultMutableTreeNode("New Category");
			
			DefaultMutableTreeNode parent = (DefaultMutableTreeNode) selected_node.getParent();
			if (parent != null)
			{
				int selected_index = parent.getIndex(selected_node);
				taxonomy.insertNodeInto(new_node, parent, selected_index + 1);
			}

			TreeNode[] nodes = taxonomy.getPathToRoot(new_node);
			TreePath path = new TreePath(nodes);
			taxonomy_display.scrollPathToVisible(path);
			update_status_bar(null);
		}

		// React to the add_new_child_category_button
		else if (event.getSource().equals(add_new_child_category_button))
		{
			DefaultMutableTreeNode selected_node = (DefaultMutableTreeNode) taxonomy_display.getLastSelectedPathComponent();
			if (selected_node == null) return;

			DefaultMutableTreeNode new_node = new DefaultMutableTreeNode("New Category");
			
			taxonomy.insertNodeInto(new_node, selected_node, selected_node.getChildCount());
			
			TreeNode[] nodes = taxonomy.getPathToRoot(new_node);
			TreePath path = new TreePath(nodes);
			taxonomy_display.scrollPathToVisible(path);
			update_status_bar(null);
		}

		// React to the delete_branch_button
		else if (event.getSource().equals(delete_branch_button))
		{
			DefaultMutableTreeNode selected_node = (DefaultMutableTreeNode) taxonomy_display.getLastSelectedPathComponent();
			if (selected_node == null) return;
			if (selected_node.getParent() != null)
				taxonomy.removeNodeFromParent(selected_node);
			update_status_bar(null);
		}
	}
	

	/**
	 * Returns each leaf category in the <code>taxonomy</code> field and its ancestry.
	 * Duplicates are removed (i.e. if the same leaf category appears more than once
	 * in <code>taxonomy</code> then the duplicate is removed and the ancestors are
	 * combined).
	 *
	 * <p>This information is returned in the form of a 2-D array of Strings. The first
	 * indice specifies a particular leaf category and its ancestors. The second
	 * indice specifies the particular ancestor of a given leaf category or the leaf
	 * category itself. Indice 0 always corresponds to the name of the leaf category
	 * itself, but the order of the ancestors is not necessarily meaningful.
	 *
	 * @return	The leaf categories and their ancestors.
	 * @see		DefaultMutableTreeNode
	 */
	public String[][] getLeafsAndTheirParents()
	{
		// The root of the tree from which leaf categories and ancestors are to
		// be extracted
		DefaultMutableTreeNode root = (DefaultMutableTreeNode) taxonomy.getRoot();

		// The initial results (duplicates not removed)
		String[][] results = new String[root.getLeafCount()][];
		
		// Extract the leaf categories and their ancestors
		DefaultMutableTreeNode leaf = root.getFirstLeaf();
		String[] ancestor_names;
		int count = 0;
		while (count < root.getLeafCount())
		{
			// Get the ancestry of the leaf
			TreeNode[] ancestry_uncasted = taxonomy.getPathToRoot(leaf);
			DefaultMutableTreeNode[] ancestry = new DefaultMutableTreeNode[ancestry_uncasted.length];
			for (int i = 0; i < ancestry.length; i++)
				ancestry[i] = (DefaultMutableTreeNode) ancestry_uncasted[i];

			// Convert to string form
			ancestor_names = new String[ancestry.length - 1]; // 11 to avoid root
			for (int i = 0; i < ancestry.length - 1; i++)
				ancestor_names[i] = (String) ancestry[i + 1].getUserObject(); // +1 to avoid root
			results[count] = ancestor_names;

			// Prepare for next iteration
			count++;
			leaf = leaf.getNextLeaf();
		}

		// Reverse the contents of each string array so that leaf categories
		// come first
		String[][] reversed_results = new String[results.length][];
		for (int i = 0; i < results.length; i++)
		{
			reversed_results[i] = new String[results[i].length];
			int k = 0;
			for (int j = results[i].length - 1; j >= 0; j--)
			{
				reversed_results[i][k] = results[i][j];
				k++;
			}
			results[i] = reversed_results[i];
		}

		// If leaf categories appear more than once, combine into one
		// category with all ancestors combined
		int replacement_count = 0;
		for (int i = 0; i < results.length; i++)
		{
			for (int j = 0; j < i; j++)
			{
				if (results[i] != null && results[j] != null)
				{
					if (results[i][0].equals(results[j][0]))
					{
						String[] temp = new String[results[i].length - 1]; // holds ancestors of results[i]
						int m = 1;
						for (int k = 0; k < results[i].length - 1; k++)
						{
							temp[k] = results[i][m];
							m++;
						}

						String[] temp_2 = new String[results[j].length + temp.length]; // holds combined ancestors
						for (int k = 0; k < temp_2.length; k++)
						{
							if (k < results[j].length)
								temp_2[k] = results[j][k];
							else
								temp_2[k] = temp[k - results[j].length];
						}
						
						results[j] = temp_2; // List all ancestors in first occurence

						results[i] = null;
						replacement_count++;
					}
				}
			}
		}
		String[][] reduced_results = new String[results.length - replacement_count][];
		int m = 0;
		for (int i = 0; i < reduced_results.length; i++)
		{
			while (results[m] == null)
				m++;
			reduced_results[i] = results[m];
			m++;
		}
		
		// Remove ancestors that are listed more than once
		for (int i = 0; i < reduced_results.length; i++)
		{

			int repetition_count = 0;
			for (int j = 1; j < reduced_results[i].length; j++)
			{
				for (int k = 1; k < reduced_results[i].length; k++)
				{
					if (reduced_results[i][j] != null && j != k)
					{
						if (reduced_results[i][j].equals(reduced_results[i][k]))
						{
							repetition_count++;
							reduced_results[i][j] = null;
						}
					}
				}
			}
			if (repetition_count != 0)
			{
				String[] temp = new String[reduced_results.length];
				for (int j = 0; j < reduced_results[i].length; j++)
					temp[j] = reduced_results[i][j];
				reduced_results[i] = new String[reduced_results[i].length - repetition_count];
				reduced_results[i][0] = temp[0];
				int reduced_results_index = 1;
				for (int j = 1; j < temp.length; j++)
				{
					if (temp[j] != null)
					{
						reduced_results[i][reduced_results_index] = temp[j];
						reduced_results_index++;
					}
				}
			}
		}

		// Return the final results
		return reduced_results;
	}



	/**
	 * Returns an array holding the labels of all leaf categories in the
	 * <code>taxonomy</code> field. No duplicates are present.
	 *
	 * @return	 The labels of all leaf categories in the <code>taxonomy</code> field.
	 */
	public String[] getLeafLabels()
	{
		String[][] categories = getLeafsAndTheirParents();
		String[] leaf_categories = new String[categories.length];
		for (int i = 0; i < leaf_categories.length; i++)
			leaf_categories[i] = categories[i][0];
		return leaf_categories;
	}


	/**
	 * Returns an array holding the labels of all parent categories
	 * (i.e. categories at the highest level of the taxonomy) in the
	 * <code>taxonomy</code> field. No duplicates are present.
	 *
	 * @return	 The labels of all parent categories in the <code>taxonomy</code> field.
	 */
	public String[] getParentLabels()
	{
		DefaultMutableTreeNode root = (DefaultMutableTreeNode) taxonomy.getRoot();
		String[] parent_categories = new String[root.getChildCount()];
		DefaultMutableTreeNode this_parent = null;
		for (int i = 0; i < parent_categories.length; i++)
		{
			if (i == 0) this_parent = (DefaultMutableTreeNode) root.getFirstChild();
			else this_parent = this_parent.getNextSibling();

			parent_categories[i] = (String) this_parent.getUserObject();
		}
		return parent_categories;
	}


	/**
	 * Returns an array holding the labels in the <code>taxonomy</code> field
	 * that are neither leaves nor parent categories (i.e. categories at the
	 * highest level of the taxonomy). No duplicates are present.
	 *
	 * @return	 The labels of all mid-level categories in the <code>taxonomy</code> field.
	 */
	public String[] getMiddleLabels()
	{
		// Get the set of all categories and the set of parent categories
		String[][] categories = getLeafsAndTheirParents();
		String[] parent_categories = getParentLabels();

		// Find non-leaf categories that are not parent categories and add them to the
		// list if they have not been found already
		LinkedList categories_so_far = new LinkedList();
		for (int i = 0; i < categories.length; i++)
		{
			for (int j = 1; j < categories[i].length; j++)
			{
				// Check if the category in categories is in parent_categories
				boolean present = false;
				for (int k = 0; k < parent_categories.length; k++)
				{
					if (categories[i][j].equals(parent_categories[k]))
					{
						present = true;
						k = parent_categories.length + 1; // exit the loop
					}
				}

				// If not present, then check if it has already been added
				// to categories_so_far. If not, then add it.
				if (!present)
				{
					boolean present_2 = false;
					for (int k = 0; k < categories_so_far.size(); k++)
					{
						if (categories[i][j].equals((String) categories_so_far.get(k)))
						{
							present_2 = true;
							k = categories_so_far.size() + 1; // exit the loop
						}
					}
					if (!present_2)
					{
						categories_so_far.add(categories[i][j]);
					}
				}

			}
		}

		// Convert categories_so_far to an array of strings and return it
		String[] final_categories = new String[categories_so_far.size()];
		for (int i = 0; i < categories_so_far.size(); i++)
			final_categories[i] = (String) categories_so_far.get(i);
		return final_categories;
	}


	/**
	 * Returns the name of each parent category (root or intermediate) and their direct descendants.
	 * The first indice corresponds to the different parents. The second indice goes through the children.
	 * The first entry of each row is the name of the parent.
	 */
	public String[][] getAllParentsAndTheirDirectChildren()
	{
		// Prepare the list of parents and their direct descendants
		LinkedList parents_and_their_children = new LinkedList();
		
		// Recurse through the tree
		DefaultMutableTreeNode root = (DefaultMutableTreeNode) taxonomy.getRoot();
		int number_children = root.getChildCount();
		for (int child = 0; child < number_children; child++)
		{
			DefaultMutableTreeNode this_child = (DefaultMutableTreeNode) root.getChildAt(child);
			addChildrenToList(parents_and_their_children, this_child);
		}
		
		// Format the information to return
		Object[] list_as_object = parents_and_their_children.toArray();
		String[][] results = new String[list_as_object.length][];
		for (int i = 0; i < results.length; i++)
			results[i] = (String[]) list_as_object[i];

		// Return the results
		return results;
	}


	/**
	 * Saves a <i>taxonomy_file</i> XML file with the contents specified
	 * in the <code>taxonomy</code> field and the comments specified in
	 * the <code>comments</code> parameter.
	 *
	 * <p>If the <code>is_basic_save</code> parameter is true, then the
	 * file is saved at the path specified in the <code>ConfigureFileLocationsJFrame</code>
	 * object's <code>taxonomy_path</code> field (automatically overwriting
	 * any existing files with the same path). If the <code>is_basic_save</code>
	 * parameter is false, then the file is saved at a location chosen by the user
	 * with a <code>JFileChooser</code> dialog box.
	 *
	 * @param	comments		Any comments to be saved inside the comments element of the XML file.
	 * @param	is_basic_save	Whether this is a save (true) or save as (false) operation.
	 * @see		ConfigureFileLocationsJFrame
	 * @see		DefaultMutableTreeNode
	 */
	public void saveTaxonomyFile(String comments, boolean is_basic_save)
	{
		// Get File to save to
		File to_save_to = null;
		if (is_basic_save && !parent_window.file_locations.taxonomy_path.equals("")) // perform a basic save operation
			to_save_to = parent_window.file_locations.getFile(parent_window.file_locations.taxonomy_path, true);
		else // perform a save as operation
		{
			if (parent_window.file_locations.taxonomy_path.equals(""))
				JOptionPane.showMessageDialog(null, "No file name is specified for saving the taxonomy.\nPlease enter one in the following dialog box.", "NOTE", JOptionPane.INFORMATION_MESSAGE);

			to_save_to = parent_window.file_locations.saveFileExternally(false, "taxonomy_file");
			if (to_save_to != null)
			{
				resetRootName(parent_window.file_locations.taxonomy_path);
				displayTaxonomyHierarchically();
			}
		}

		// Write the file to disk
		if (to_save_to != null)
		{
			try
			{
				// Prepare stream writer
				FileOutputStream to = new FileOutputStream(to_save_to);
				DataOutputStream writer = new DataOutputStream(to);

				// Write the header and the first element of the XML file
				String pre_tree_part = new String
				(
					"<?xml version=\"1.0\"?>\n" +
					"<!DOCTYPE taxonomy_file [\n" +
					"   <!ELEMENT taxonomy_file (taxonomy_comments, parent_category+)>\n" +
					"   <!ELEMENT taxonomy_comments (#PCDATA)>\n" +
					"   <!ELEMENT parent_category (category_name, sub_category+)>\n" +
					"   <!ELEMENT category_name (#PCDATA)>\n" +
					"   <!ELEMENT sub_category (category_name, sub_category*)>\n" +
					"]>\n\n" +
					"<taxonomy_file>\n\n" +
					"   <taxonomy_comments>" + comments + "</taxonomy_comments>\n\n"	
				);
				writer.writeBytes(pre_tree_part);

				// Write the XML code to represent the contents of the taxonomy field
				DefaultMutableTreeNode taxonomy_node = (DefaultMutableTreeNode) taxonomy.getRoot();
				for (int i = 0; i < taxonomy_node.getChildCount() ; i++)
				{
					writer.writeBytes("   <parent_category>\n");
					DefaultMutableTreeNode child = (DefaultMutableTreeNode) taxonomy_node.getChildAt(i);
					writer.writeBytes(assembleTreeXMLElements(child, 1));
					writer.writeBytes("   </parent_category>\n\n");
				}
				writer.writeBytes("</taxonomy_file>");



				// Close the output stream
				writer.close();
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(null, ("Unable to write file " + to_save_to.getName() + "."), "ERROR", JOptionPane.ERROR_MESSAGE);
				// e.printStackTrace();
			}
		}
	}


	/**
	 * Returns true if the <i>taxonomy</i> field is empty and false if it is not.
	 *
	 * @return	Whether or not the <i>taxonomy</i> field is empty.
	 */
	public boolean isTreeEmpty()
	{
		DefaultMutableTreeNode root = (DefaultMutableTreeNode) taxonomy.getRoot();
		if (root.getChildCount() == 0)
			return true;
		else
			return false;
	}


	/**
	 * Recursively assembles the XML code to represent the contents of the
	 * given tree structure.
	 *
	 * @param	node	The root node of the tree.
	 * @param	depth	The depth of the tree.
	 */
	public static String assembleTreeXMLElements(DefaultMutableTreeNode node, int depth)
	{
		String XML_string = new String("");

		if (depth != 1)
		{
			for (int j = 0; j <= depth; j++)
				XML_string += "   ";
			XML_string += "<sub_category>\n";	
		}

		if (node.isLeaf())
		{
			for (int j = 0; j <= depth + 1; j++)
				XML_string += "   ";
			XML_string += "<category_name>" + StringMethods.giveStringXMLEscapeCharacters((String) node.getUserObject()) + "</category_name>\n";
		}
		else
		{
			for (int i = 0; i < node.getChildCount(); i++)
			{
				if (i == 0)
				{
					for (int j = 0; j <= depth + 1; j++)
						XML_string += "   ";
					XML_string += "<category_name>" + StringMethods.giveStringXMLEscapeCharacters((String) node.getUserObject()) + "</category_name>\n";
				}
				DefaultMutableTreeNode child = (DefaultMutableTreeNode) node.getChildAt(i);
				XML_string += assembleTreeXMLElements(child, depth + 1);
			}
		}

		if (depth != 1)
		{
			for (int j = 0; j <= depth; j++)
				XML_string += "   ";
			XML_string += "</sub_category>\n";
		}

		return XML_string;
	}

	
	/* PRIVATE METHODS *********************************************************/


	// Displays the <code>taxonomy</code> field in the left section of the panel
	// in the form of a hierarchal tree reflecting the relationships of 
	// broad categories going down to narrower categories.
	// Sorts the tree categories alphebetically by name.
	// Refreshes the display.
	// Updates the status bar
	private void displayTaxonomyHierarchically()
	{
		// Remove anything on the left side of the panel
		if (tree_display_panel != null)
			remove(tree_display_panel);
		if (ancestry_display_panel != null)
			remove(ancestry_display_panel);

		// Sort tree alphabetically, make it editable and configure its display
		taxonomy = new DefaultTreeModel (Sorter.sortTree((DefaultMutableTreeNode) taxonomy.getRoot()));
		taxonomy_display = new JTree(taxonomy);
		setTreeIcons(taxonomy_display, new String("." + File.separator + "ProgramFiles" + File.separator + "bluebull.gif"),
		                               new String("." + File.separator + "ProgramFiles" + File.separator + "beigebull.gif"));
		taxonomy_display.setEditable(true);

		// Update the status bar
		update_status_bar(null);

		// Refressh the display
		taxonomy_display_scroll_pane = new JScrollPane(taxonomy_display);
		tree_display_panel = new JPanel(new GridLayout(1, 1));
		tree_display_panel.add(taxonomy_display_scroll_pane);
		add(tree_display_panel, BorderLayout.CENTER);
		repaint();
		parent_window.repaint();
	}


	// Displays the <code>taxonomy</code> field in the left section of the panel
	// in the form of a list of leaf categories. Children of leaf categories represent
	// a list of all ancestors of the given category.
	// Sorts the tree categories alphebetically by name.
	// Refreshes the display.
	// Updates the status bar
	private void displayAncestryTree()
	{
		// Remove anything on the left side of the panel
		if (tree_display_panel != null)
			remove(tree_display_panel);
		if (ancestry_display_panel != null)
			remove(ancestry_display_panel);
		
		 // Do nothing if no children
		if (((DefaultMutableTreeNode) (taxonomy.getRoot())).getChildCount() != 0)
		{
			// Construct ancestor tree, sort it alphabetically, 
			// make it editable and configure its display
			DefaultTreeModel ancestry = getAncestorTree();
			ancestry = new DefaultTreeModel(Sorter.sortTree((DefaultMutableTreeNode) ancestry.getRoot()));
			JTree anestry_display = new JTree(ancestry);
			setTreeIcons(anestry_display, new String("." + File.separator + "ProgramFiles" + File.separator + "bluebull.gif"),
										  new String("." + File.separator + "ProgramFiles" + File.separator + "beigebull.gif"));

			// Update the status bar
			update_status_bar((DefaultMutableTreeNode) ancestry.getRoot());

			// Refressh the display
			JScrollPane ancesty_display_scroll_pane = new JScrollPane(anestry_display);
			ancestry_display_panel = new JPanel(new GridLayout(1, 1));
			ancestry_display_panel.add(ancesty_display_scroll_pane);
			add(ancestry_display_panel, BorderLayout.CENTER);
			repaint();
			parent_window.repaint();
		}
	}
	
	
	// Parses the specified file and, if it is an appropriate taxonomy_file XML file,
	// stores it in taxonomy and displays it. Otherwise gives an error message and
	// creates and displays a new blank tree. Returns true if the end of the procedure
	// is reached without an exception being thrown.
	//
	// Note that if the extension of the file is not XML or xml, then it is assumed that
	// the file is a tab delimited text file of the MIREX 2005 format, and is parsed as
	// such. This functionality is intended primarily for use with the command line
	// use of Bodhidharma.
	private boolean parseTaxonomyFileIntoJTree(String file_path)
	{
		boolean succesful = false;
		
		try
		{
			// Check if the file is an XML file by looking at its extension
			boolean using_xml_file = true;
			String extension = StringMethods.getExtension(file_path);
			if (extension == null)
				using_xml_file = false;
			else if (!extension.equals(".xml") && !extension.equals(".XML"))
				using_xml_file = false;

			// Holds the results of parsing
			DefaultTreeModel[] parse_results;

			// Parse the contents of the taxonomy_file into parse_results from an XML file
			if (using_xml_file)
				parse_results = (DefaultTreeModel[])(XMLDocumentParser.parseXMLDocument(file_path, "taxonomy_file"));

			// Parse the contents of the taxonomy_file into parse_results from a text file
			else
			{
				// Read in and tokenize the file
				String[][] file_contents = bodhidharma.utilities.FileMethods.tokenizeTabDelFile(file_path);

				// Generate a tree holding the taxonomy
				DefaultMutableTreeNode root_node = new DefaultMutableTreeNode("Taxonomy");
				for (int leaf = 0; leaf < file_contents.length; leaf++)
					root_node = addToTree(root_node, file_contents[leaf]);

				// Prepare the results
				parse_results = new DefaultTreeModel[1];
				parse_results[0] = new DefaultTreeModel(root_node.getRoot());
			}
			
			// Organize the results
			taxonomy = parse_results[0];
			resetRootName(file_path);
			taxonomy_display = new JTree(taxonomy);
			taxonomy_display.setEditable(true);
			setTreeIcons(taxonomy_display, new String("." + File.separator + "ProgramFiles" + File.separator + "bluebull.gif"),
			                               new String("." + File.separator + "ProgramFiles" + File.separator + "beigebull.gif"));
			succesful = true;
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			newTaxonomy();
		}

		return succesful;
	}


	// Constructs a tree representation of the <code>taxonomy</code> field whereby
	// the field is represented by a list of leaf categories. Children of leaf categories
	// represent the ancestors of the corresponding category. Returns the root of the
	// resulting tree.
	private DefaultTreeModel getAncestorTree()
	{
		DefaultMutableTreeNode root = new DefaultMutableTreeNode(parent_window.file_locations.taxonomy_path);
		String[][] ancestry = getLeafsAndTheirParents();
		for (int i = 0; i < ancestry.length; i++)
		{
			DefaultMutableTreeNode parent = new DefaultMutableTreeNode(ancestry[i][0]);
			for (int j = 1; j < ancestry[i].length; j++)
				parent.add(new DefaultMutableTreeNode(ancestry[i][j]));
			root.add(parent);
		}
		return new DefaultTreeModel(root);
	}


	// Creates and displays a new empty tree. Also resets the RecordingsPanel.
	// Sets the parent taxonomy file path to empty.
	private void newTaxonomy()
	{
		parent_window.file_locations.setTaxonomyPath("");

		if (parent_window.recordings_panel != null)
			parent_window.recordings_panel.newRecordingsList();
		
		DefaultMutableTreeNode new_node = new DefaultMutableTreeNode("New Taxonomy");
		taxonomy = new DefaultTreeModel(new_node);
		taxonomy_display = new JTree(taxonomy);
		setTreeIcons(taxonomy_display, new String("." + File.separator + "ProgramFiles" + File.separator + "bluebull.gif"),
		                               new String("." + File.separator + "ProgramFiles" + File.separator + "beigebull.gif"));
		taxonomy_display.setEditable(true);
		
		TreeNode[] nodes = taxonomy.getPathToRoot(new_node);
		
		TreePath path = new TreePath(nodes);
		taxonomy_display.setSelectionPath(path);
		
		displayTaxonomyHierarchically();
	}


	// Adds the name of node to list as well as its direct decendants. Recurses
	// through the children of node, applying the same procedure. Does nothing
	// if node is a leaf.
	private void addChildrenToList(LinkedList list, DefaultMutableTreeNode node)
	{
		if (!node.isLeaf())
		{
			// Find the number of children of node
			int number_children = node.getChildCount();
			
			// Find the direct descendants of node
			String[] direct_descendants = new String[number_children];
			for (int child = 0; child < number_children; child++)
			{
				DefaultMutableTreeNode this_child = (DefaultMutableTreeNode) node.getChildAt(child);
				direct_descendants[child] = (String) this_child.getUserObject();
			}

			// Add the name of node and its direct descendants to the list
			String[] results = new String[1 + number_children];
			results[0] = (String) node.getUserObject();
			for (int child = 0; child < number_children; child++)
				results[child + 1] = direct_descendants[child];
			list.add(results);

			// Recursively apply this process to non-leaf children
			for (int child = 0; child < number_children; child++)
			{
				DefaultMutableTreeNode this_child = (DefaultMutableTreeNode) node.getChildAt(child);
				if (!this_child.isLeaf())
					addChildrenToList(list, this_child);
			}
		}
	}

	
	// Gives the root of the taxonomy tree the given name
	private void resetRootName(String name)
	{
		DefaultMutableTreeNode root_node = (DefaultMutableTreeNode) taxonomy.getRoot();
		root_node.setUserObject(name);
	}

	
	// Causes the status bar to display the number of parent categories
	// in the taxonomy field if the root parameter is null.
	// If the root paratmeter is not null, then sets the status bar to
	// display the number of leaf categories in the taxonomy field.
	private void update_status_bar(DefaultMutableTreeNode root)
	{
		if (root == null) // hierarchy display
		{
			int number_parent_categories = ((DefaultMutableTreeNode)(taxonomy.getRoot())).getChildCount();
			status_bar.setText(new String(number_parent_categories + " parent categories"));
		}
		else
		{
			int number_parent_categories = root.getChildCount();
			status_bar.setText(new String(number_parent_categories + " leaf categories"));
		}
	}


	// Sets the icons for the given tree to the images specified by the given
	// paths. Open and closed icons are displayed identically.
	private void setTreeIcons(JTree tree, String leaf_icon_path, String non_leaf_icon_path)
	{
		DefaultTreeCellRenderer renderer = new DefaultTreeCellRenderer();
		renderer.setLeafIcon(new ImageIcon(leaf_icon_path));
		renderer.setOpenIcon(new ImageIcon(non_leaf_icon_path));
		renderer.setClosedIcon(new ImageIcon(non_leaf_icon_path));
		tree.setCellRenderer(renderer);
	}


	/**
	 * Recursively constructs a taxonomy tree by adding nodes one by one with a given
	 * ancestry.
	 *
	 * @param	root_node	The node to add children to.
	 * @param	ancestry	The ancestry of the node to add as a child. The first entry
	 *						is the leaf and the ancestry proceeds from youngest to oldest.
	 * @return				The root node with the child added to it.
	 */
	private static DefaultMutableTreeNode addToTree(DefaultMutableTreeNode root_node, String[] ancestry)
	{
		if (ancestry.length > 1)
		{
			int par = ancestry.length - 1;
			boolean found = false;
			int number_nodes = root_node.getChildCount();
			for (int child = 0; child < number_nodes; child++)
			{
				DefaultMutableTreeNode this_node = (DefaultMutableTreeNode) root_node.getChildAt(child);
				String this_label = (String) this_node.getUserObject();
				if (this_label.equals(ancestry[par]))
				{
					String[] reduced_ancestry = new String[ancestry.length - 1];
					for (int i = 0; i < reduced_ancestry.length; i++)
						reduced_ancestry[i] = ancestry[i];
					this_node = addToTree(this_node, reduced_ancestry);
					child = number_nodes;
					par = 1;
					found = true;
				}
			}
			if (!found)
			{
				DefaultMutableTreeNode new_node = new DefaultMutableTreeNode(ancestry[par]);
				root_node.add(new_node);
				String[] reduced_ancestry = new String[ancestry.length - 1];
				for (int i = 0; i < reduced_ancestry.length; i++)
					reduced_ancestry[i] = ancestry[i];
				new_node = addToTree(new_node, reduced_ancestry);
			}
		}
		else
		{
			root_node.add(new DefaultMutableTreeNode(ancestry[0]));
		}
		return root_node;
	}
}