/*
 * @(#)BioKNearestNeighbour.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers.bio_k_nearest_neighbour;

import bodhidharma.classifiers.*;
import bodhidharma.classifiers.genetic_algorithms.*;
import bodhidharma.xml_parsing.*;
import java.util.LinkedList;
import java.io.*;

/**
 * An interface for using the k-nearest neighbour algorithm with feature selection
 * and/or weighting performed with genetic alogrithms.
 * 
 * <p>This classifier is trained by having the genetic algorithm choose the feature
 * selection and/or weighting. The k-nearest neighbour algorithm only needs one
 * iteration of training, of course. The "iterations" discussed in this class'
 * description actually refer to generations of the genetic algorithm.
 *
 * <p>This classifier also calclulates scaling factors for all of the input feature values.
 * This is done to ensure that all features fall into roughly the same range, so that
 * they have roughly the same "weighting" before the genetic algorithm goes to work.
 * The scaling factors are calculated during the first iteration of training and stored.
 * They are then stored and applied to future unknown features that are input. They are
 * not altered unless further training is performed.
 *
 * <p><b>WARNING:</b> the scaling factors are only calculated the first time that this
 * object is trained. Additional training will not cause recalculation of scaling factors.
 * To do this, a new BioKNearestNeighbour must be instantiated.
 *
 * <p>This system uses exemplar based learning to classify arbitrary feature sets after
 * training. This implementation makes it possible to assign more than one label to a
 * single feature set.
 *
 * <p>Feature sets are fed into the classifier as arrays of doubles. Categories
 * are specified as arrays of Strings.
 * 
 * <p>Use the <code>train</code> method to train the classifier (the
 * <code>getFeatureNames</code> method is useful for deteriming the features that
 * can be fed to the classifer). This also calculates the feature value scaling
 * factors.
 *
 * <p>One constructor is provided for creating a new network and one is provided
 * for parsing XML code and using it to reconstruct a trained network.
 *
 * <p>Use the <code>classify</code> method to classify feature sets once training
 * has been completed (the <code>getCategories</code> method is useful for
 * determining what categories feature sets can be classified into and for
 * determining the order of the categories when parsing classification results).
 *
 * <p>Use the <code>save</code> method to save the classifier and its current
 * state to disk.
 * 
 * <p>Use the <code>getClassifierName</code> and <code>getClassifierParameters</code>
 * methods to obtain information about the classifier.
 *
 * <p>Use the <code>getClassifierIdentifier</code> method to get a name or code that
 * was given to an instantiation of the classifier when it was constructed. This
 * identifier can be used by external classes to identify the instantiation.
 *
 * <p>Use the <code>getScaledFeatureValues</code> method to find out what the values
 * of a set of feature values would be after scaling.
 *
 * @see KNN
 * @see Breeder
 * @see FeatureSelectionEvaluator
 * @see FeeatureWeightingEvaluator
 *
 * @author Cory McKay
 */
public class BioKNearestNeighbour
	extends SupervisedClassifier
{
	/* FIELDS ******************************************************************/

	
	// The KNN classifier that exists after training. Set to null if untrained.
	private	KNN			knn;

	// Contain the minimum and maximum acceptable values for each feature after training,
	// where the feature is identified by the indice (corresponding to the labels_of_features
	// field). New feature values are cut off so that they fall between these values by
	// the getScaledFeatures method. These values are set by the calculateScalingInformation
	// method.
	private	double[]	min_feature_value_cutoffs;
	private double[]	max_feature_value_cutoffs;

	// Feature selection and weighting parameters
	private	boolean		using_feature_selection;	// whether or not to calculate and use feature selection
	private	boolean		using_feature_weighting;	// whether or not to calculate and use feature weighting
	private	double		fsw_training_fraction;		// fraction of training samples to actually be used for training when feature selection and weighting are calculated
	private ClassificationResultsInterpereter results_interpereter;	 // used to calculate fitnesses with genetic algorithms. It may be null if the user wants to use alternative fitness measures not based on actual classifications.

	// Genetic algorithm parameters
	private int		word_length;			// number of bits used to represent weighting values
	private	int		population_size;		// total number of chromosomes in the population
	private	double	mating_rate;			// fraction of population allowed to reproduce (0 to 1)
	private	double	mutation_probability;	// probability of mutation in an offspring
	private	boolean	allow_elite;			// controls whether best chromosome per village is cloned exactly
	private	int		number_of_villages;		// number of vilages that segregate chromosomes
	private	int		migration_size;			// number of chromosomes that move during migration
	private	int		migration_delay;		// number of epochs between migrations between villages


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Generate a BioKNearestNeighbour with the given parameters and randomly.
	 *
	 * @param	feature_names			The names of the features that classifications will be based on.
	 * @param	categories				The names of categories that correspond to the possible classification results.
	 * @param	identifier				An identifier that can be associated with the classifier so that outside classes can identify it.
	 * @param	using_feature_selection	Whether or not feature selection will be performed.
	 * @param	using_feature_weighting	Whether or not variable feature weighting will be performed.
	 * @param	gen_alg_settings		Initialization settings for the genetic algorithm.
	 * @param	fsw_training_fraction	Fraction of training samples to actually be used for training when feature selection and weighting are calculated. Must be between 0 and 1.
	 * @param	training_monitor		Used to monitor training progress. It may be null if it is not used.
	 * @param	results_interpereter	Used to calculate fitnesses with genetic algorithms. It may be null if the user wants to use alternative fitness measures not based on actual classifications.
	 */
	public BioKNearestNeighbour(String[] feature_names, String[] categories,
	                            String identifier, boolean using_feature_selection,
	                            boolean using_feature_weighting, GeneticAlgorithmJFrame gen_alg_settings,
								double fsw_training_fraction, ClassificationResultsInterpereter results_interpereter)
	{
		// Set superclass' fields
		this.feature_names = feature_names;
		this.categories = categories;
		this.identifier = identifier;

		// Set private fields
		knn = null;
		min_feature_value_cutoffs = null;
		max_feature_value_cutoffs = null;
		this.using_feature_selection = using_feature_selection;
		this.using_feature_weighting = using_feature_weighting;
		this.fsw_training_fraction = fsw_training_fraction;
		this.results_interpereter = results_interpereter;

		// Set genetic algorithm parameters
		word_length = gen_alg_settings.getWordLength();
		population_size = gen_alg_settings.getPopulationSize();
		mating_rate = gen_alg_settings.getMatingRate();
		mutation_probability = gen_alg_settings.getMutationProbability();
		allow_elite = gen_alg_settings.getElitism();
		number_of_villages = gen_alg_settings.getNumberVillages();
		migration_size = gen_alg_settings.getMigrationSize();
		migration_delay = gen_alg_settings.getMigrationDelay();

		// Set training_monitor
		training_monitor = null;
	}


	/**
	 * Parse the file specified by the given file path to recreate the specificed trained
	 * classifier. Throws an exception if problems occur during parsing.
	 *
	 * @param	file_path			The file path of a BioKNearestNeighbour_file that this instantiation is to be based on.
	 */
	public BioKNearestNeighbour(String file_path)
		throws Exception
	{
		// Parse file
		Object[] results = (Object[]) XMLDocumentParser.parseXMLDocument(file_path, "BioKNearestNeighbour_file");
		String[] fixed_fields = (String[]) results[0];
		
		// Set non-classifier fields
		identifier = fixed_fields[0];
		categories = (String[]) results[1];
		feature_names = (String[]) results[2];
		min_feature_value_cutoffs = (double[]) results[3];
		max_feature_value_cutoffs = (double[]) results[4];
		using_feature_selection = (Boolean.valueOf(fixed_fields[2])).booleanValue();
		using_feature_weighting = (Boolean.valueOf(fixed_fields[3])).booleanValue();
		results_interpereter = null; // doesn't matter, because a new BioKNearestNeighbour needs to instantiated for training, and this only is important during training
		fsw_training_fraction = (Double.valueOf(fixed_fields[4])).doubleValue();
		word_length = (Integer.valueOf(fixed_fields[5])).intValue();
		population_size = (Integer.valueOf(fixed_fields[6])).intValue();
		mating_rate = (Double.valueOf(fixed_fields[7])).doubleValue();
		mutation_probability = (Double.valueOf(fixed_fields[8])).doubleValue();
		allow_elite = (Boolean.valueOf(fixed_fields[9])).booleanValue();
		number_of_villages = (Integer.valueOf(fixed_fields[10])).intValue();
		migration_size = (Integer.valueOf(fixed_fields[11])).intValue();
		migration_delay = (Integer.valueOf(fixed_fields[12])).intValue();

		// Set up knn
		knn = new KNN(fixed_fields[1]);

		// Set training_monitor
		training_monitor = null;
	}


	/* PUBLIC METHODS **********************************************************/

	
	/**
	 * Returns the name of the type of classifier. This information can be used
	 * by external classes to identify the type of classifier in situations such
	 * as when the configuration of a trained classifier is being read from a file.
	 */
	public String getClassifierName()
	{
		return new String("BioKNearestNeighbour");
	}


	/**
	 * Returns a String describing the parameters of the classifier. This includes
	 * word length, population size, mating rate, mutation probability, whether
	 * or not elitism is allowed, number of villages, migration size, migration
	 * delay, whether or not feature selection is enabled, whether or not feature
	 * weighting is enabled and the value of K.
	 */
	public String getClassifierParameters()
	{
		String parameters = new String();
		parameters += "Word Length: " + word_length + "\n";
		parameters += "Population Size: " + population_size + "\n";
		parameters += "Mating Rate: " + mating_rate + "\n";
		parameters += "Mutation Probability: " + mutation_probability + "\n";
		parameters += "Allow Elite: " + allow_elite + "\n";
		parameters += "Number of Villages: " + number_of_villages + "\n";
		parameters += "Migration Size: " + migration_size + "\n";
		parameters += "Migration Delay: " + migration_delay + "\n";
		parameters += "Using Feature Selection: " + using_feature_selection + "\n";
		parameters += "Using Feature Weighting: " + using_feature_weighting + "\n";
		parameters += "K: " + knn.getK() + "\n";

		return parameters;
	}


	/**
	 * Returns a copy of the the array indicating whether or not a given feature is to be
	 * used for classification. Indices correspond to those of the <i>feature_names</i>
	 * field. These values are set during the feature selection portion of the
	 * <code>train</code> method. True means that the feature is to be used and false
	 * that it isn't. A value of null is returned if feature selection has not been
	 * performed, and all features are to be used.
	 */
	public boolean[] getFeatureSelection()
	{
		if (knn.features_to_use == null)
			return null;
		boolean[] selected = new boolean[knn.features_to_use.length];
		for (int i = 0; i < selected.length; i++)
			selected[i] = knn.features_to_use[i];
		return selected;
	}
	

	/**
	 * Returns a copy of the the array indicating the feature weights that are to be
	 * used for classification. Indices correspond to those of the <i>feature_names</i>
	 * field. These values are set during the feature weighting portion of the
	 * <code>train</code> method. Values are between 0 and 1, with 1 meaning that the
	 * feature is very important relative to others. A value of null is returned if
	 * feature weighting has not been performed, and all weights are defaulted to 1.
	 */
	public double[] getFeatureWeights()
	{
		if (knn.feature_weights == null)
			return null;
		double[] weights = new double[knn.feature_weights.length];
		for (int i = 0; i < weights.length; i++)
			weights[i] = knn.feature_weights[i];
		return weights;
	}
	

	/**
	 * Saves all of the fields to the given file. A separate file is created
	 * for the knn object. Throws an exception if a problem occurs.
	 */
	public void	save(File place_to_save)
		throws Exception
	{
		// Create file for knn
		String knn_file_name;
		if (knn == null)
			knn_file_name = "null";
		else
		{
			knn_file_name = bodhidharma.utilities.StringMethods.getDirectoryName(place_to_save.getPath()) +
		                    "knn.xml";
			File knn_file = new File(knn_file_name);
			knn.save(knn_file);
		}

		// Save the BioKNearestNeighbour file
		FileOutputStream to = new FileOutputStream(place_to_save);
		DataOutputStream writer = new DataOutputStream(to);
		String header = new String
		(
			"<?xml version=\"1.0\"?>\n" +
			"<!DOCTYPE BioKNearestNeighbour_file [\n" +
			"   <!ELEMENT BioKNearestNeighbour_file (comments, identifier, categories, feature_names, knn_file_path, min_feature_value_cutoffs, max_feature_value_cutoffs, using_feature_selection, using_feature_weighting, fsw_training_fraction, word_length, population_size, mating_rate, mutation_probability, allow_elite, number_of_villages, migration_size, migration_delay)>\n" +
			"   <!ELEMENT comments (#PCDATA)>\n" +
			"   <!ELEMENT identifier (#PCDATA)>\n" +
			"   <!ELEMENT categories (cat+)>\n" +
			"   <!ELEMENT cat (#PCDATA)>\n" +
			"   <!ELEMENT feature_names (feat+)>\n" +
			"   <!ELEMENT feat (#PCDATA)>\n" +
			"   <!ELEMENT knn_file_path (#PCDATA)>\n" +
			"   <!ELEMENT min_feature_value_cutoffs (val*)>\n" +
			"   <!ELEMENT val (#PCDATA)>\n" +
			"   <!ELEMENT max_feature_value_cutoffs (val*)>\n" +
			"   <!ELEMENT using_feature_selection (#PCDATA)>\n" +
			"   <!ELEMENT using_feature_weighting (#PCDATA)>\n" +
			"   <!ELEMENT fsw_training_fraction (#PCDATA)>\n" +
			"   <!ELEMENT word_length (#PCDATA)>\n" +
			"   <!ELEMENT population_size (#PCDATA)>\n" +
			"   <!ELEMENT mating_rate (#PCDATA)>\n" +
			"   <!ELEMENT mutation_probability (#PCDATA)>\n" +
			"   <!ELEMENT allow_elite (#PCDATA)>\n" +
			"   <!ELEMENT number_of_villages (#PCDATA)>\n" +
			"   <!ELEMENT migration_size (#PCDATA)>\n" +
			"   <!ELEMENT migration_delay (#PCDATA)>\n" +
			"]>\n\n" +
			"<BioKNearestNeighbour_file>\n\n" +
			"   <comments></comments>\n\n"
		);
		writer.writeBytes(header);
		writer.writeBytes("   <identifier>" + identifier + "</identifier>\n");
		writer.writeBytes("   <categories>\n");
		for (int i = 0; i < categories.length; i++)
			writer.writeBytes("      <cat>" + categories[i] + "</cat>\n");
		writer.writeBytes("   </categories>\n");
		writer.writeBytes("   <feature_names>\n");
		for (int i = 0; i < feature_names.length; i++)
			writer.writeBytes("      <feat>" + feature_names[i] + "</feat>\n");
		writer.writeBytes("   </feature_names>\n");
		writer.writeBytes("   <knn_file_path>" + knn_file_name + "</knn_file_path>\n");
		writer.writeBytes("   <min_feature_value_cutoffs>\n");
		if (min_feature_value_cutoffs != null)
			for (int i = 0; i < min_feature_value_cutoffs.length; i++)
				writer.writeBytes("      <val>" + min_feature_value_cutoffs[i] + "</val>\n");
		else
			writer.writeBytes("      <use>null</use>\n");
		writer.writeBytes("   </min_feature_value_cutoffs>\n");
		writer.writeBytes("   <max_feature_value_cutoffs>\n");
		if (max_feature_value_cutoffs != null)
			for (int i = 0; i < max_feature_value_cutoffs.length; i++)
				writer.writeBytes("      <val>" + max_feature_value_cutoffs[i] + "</val>\n");
		else
			writer.writeBytes("      <use>null</use>\n");
		writer.writeBytes("   </max_feature_value_cutoffs>\n");
		writer.writeBytes("   <using_feature_selection>" + using_feature_selection + "</using_feature_selection>\n");
		writer.writeBytes("   <using_feature_weighting>" + using_feature_weighting + "</using_feature_weighting>\n");
		writer.writeBytes("   <fsw_training_fraction>" + fsw_training_fraction + "</fsw_training_fraction>\n");
		writer.writeBytes("   <word_length>" + word_length + "</word_length>\n");
		writer.writeBytes("   <population_size>" + population_size + "</population_size>\n");
		writer.writeBytes("   <mating_rate>" + mating_rate + "</mating_rate>\n");
		writer.writeBytes("   <mutation_probability>" + mutation_probability + "</mutation_probability>\n");
		writer.writeBytes("   <allow_elite>" + allow_elite + "</allow_elite>\n");
		writer.writeBytes("   <number_of_villages>" + number_of_villages + "</number_of_villages>\n");
		writer.writeBytes("   <migration_size>" + migration_size + "</migration_size>\n");
		writer.writeBytes("   <migration_delay>" + migration_delay + "</migration_delay>\n");
		writer.writeBytes("</BioKNearestNeighbour_file>");
		writer.close();
	}

	
	/**
	 * Trains the <code>BioKNearestNeighbour</code> using the given feature sets.
 	 * The training done by this mehtod consists of storing the training points
	 * in the <code>KNN</code> classifier and finding good feature selection and
	 * feature weighting settings (if these two options were set in the constructor).
	 * 
	 * <p><b>NOTE:</b> It should be noted that feature weighting is done using the 
	 * features selection calculated during the feature selection portion of training,
	 * if the feature selection option is selected.
	 *
	 * <p>The first indice of the <i>feature_sets</i> parameter corresponds to different
	 * feature sets. The second indice corresponds to different features in the given featue set.
	 * It should be noted that all feature sets must use the same features in
	 * the same order as given in the </i>feature_labels</i> parameter.
	 *
	 * <p>The <i>feature_labels</i> parameter specifies the names of each of the features
	 * in the <i>feature_sets</i> parameter. The features in the <i>feature_sets</i> parameter
	 * will automatically be matched to the features in the <i>feature_names</i> field based on
	 * the content of the <i>feature_labels</i> parameter unless a value of <i>null</i> is passed
	 * to the <i>feature_labels</i>. In this case, the feature values in the <i>feature_sets</i>
	 * parameter will simply be fed into the classifier in the order that they occur.
	 *
	 * <p>The <i>model_categories</i> parameter gives the categories of each of
	 * the given feature sets. The first indice corresponds to the feature set.
	 * The second indice corresponds to different model categories for the given
	 * feature set. Only categories to which the feature set belongs should be included.
	 *
	 * <p>The <i>iterations</i> parameter specifies the number of training iterations
	 * performed.  If a negative value is passed here, then the number of iterations to perform
	 * is calculated automatically based on the <i>acceptable_threshold</i> parameter, which 
	 * specifies the absolute rate of change of the training error below which training will
	 * stop, and the <i>consecutive_iterations</i> parameter, which specifies the number of consecutive
	 * iterations for which the rate of change must be below this threshold in order for
	 * training to stop. The number of iterations that go by will never exceed the absolute
	 * valud of the <i>iterations</i> value, irregardless of the other parameters.
	 *
	 * <p>For example, if a value of 1000 is given for <i>iterations</i>, then 1000 iterations
	 * will be performed irregardless of the other parameters. If a value of -1000 is given,
	 * then training will automatically stop if the absolute value of the rate of change
	 * of the training error from one sample to the next falls below the
	 * <i>acceptable_threshold</i> parameter for <i>consecutive_iterations</i> iterations, but
	 * no more than 1000 iterations will be performed in any case.
	 *
	 * <p>The returned double is an average error after training iterations. The indice
	 * of the returned array corresponds to the iteration of training that the error is associated
	 * with. This is the data for feature weighting if this option was selected, the data for
	 * feature selection if this option was selected but feature weighting was not and null
	 * otherwise (since basic KNN does not need training iteration.
	 *
	 * <p>This method also calclulates scaling factors for all of the input feature values.
	 * This is done to ensure that all features fall into roughly the same range, so that
	 * they have roughly the same "weighting" before the genetic algorithm goes to work.
	 * The scaling factors are calculated during the first iteration of training and stored.
	 * They are then stored and applied to future unknown features that are input. They are
	 * not altered unless further training is performed.
	 *
	 * <p><b>WARNING:</b> the scaling factors are only calculated the first time that this
	 * object is trained. Additional training will not cause recalculation of scaling factors.
	 * To do this, a new BioKNearestNeighbour must be instantiated. Also, each retraining
	 * will reset feature selection and feature weighting settings, if these have been
	 * found during earlier trainings.
	 *
	 * <p>An exception if thrown if the <i>feature_labels</i> do not contain the same names as
	 * <i>feature_names</i> (although a different ordering is permitted) or if any of the
	 * feature sets in <i>feature_sets</i> have a different number of features than
	 * <i>feature_names</i>. An exception is also thrown if <i>feature_sets</i> and
	 * <i>model_categories</i> have different sizes in regard to their first parameters.
	 * An exception is also thrown if the <i>given_results</i> parameter contains a name not
	 * present in the <i>categories</i> field or if it contains the same category more than once.
	 * An exception is also thrown if there are problems during evolution.
	 */
	public double[] train(double[][] feature_sets, String[] feature_labels,
	                      String[][] model_categories, int iterations,
	                      double acceptable_threshold, int consecutive_iterations)
		throws Exception
	{
		// Ensure that model_categories and feature_sets have the same size with regards
		// to their first parameters.
		if (feature_sets.length != model_categories.length)
			throw new Exception("A total of " + feature_sets.length + " feature sets were " +
			                    "provided to the classifier, but " + model_categories.length +
			                    " model classifications were provided. These two numbers must" +
			                    " match.");

		// Ensure that the features in feature_labels match feature_names and order
		// the features in training_data to match the order in feature_names. An exception
		// is thrown if any problems occur.
		double[][] training_data;
		if (feature_labels != null)
			training_data = getOrderedFeatureSets(feature_sets, feature_labels);
		else
		{
			training_data = new double[feature_sets.length][];
			for (int rec = 0; rec < training_data.length; rec++)
			{
				training_data[rec] = new double[feature_sets[rec].length];
				for (int feat = 0; feat < training_data[rec].length; feat++)
					training_data[rec][feat] = feature_sets[rec][feat];
			}
		}

		// Generate the model results and throw an exception if there is a problem with
		// model_categories. Entries are set to true if a given feature set should belong
		// to a particular category and to false otherwise.
		boolean[][] boolean_model_results = getModelResults(model_categories);
		int[][] int_model_results = new int[boolean_model_results.length][];
		for (int rec = 0; rec < int_model_results.length; rec++)
		{
			// Find the number of categories that the recording belongs to
			int number_categories_recording_belongs_to = 0;
			for (int cat = 0; cat < boolean_model_results[rec].length; cat++)
				if (boolean_model_results[rec][cat])
					number_categories_recording_belongs_to++;

			// Store the indices of the categories that the recording belongs to
			// in int_model_results
			int_model_results[rec] = new int[number_categories_recording_belongs_to];
			int hits_so_far = 0;
			for (int cat = 0; cat < boolean_model_results[rec].length; cat++)
				if (boolean_model_results[rec][cat])
			{
				int_model_results[rec][hits_so_far] = cat;
				hits_so_far++;
			}
		}

		// Calculate feature scaling factors if they haven't been calculated before
		if (min_feature_value_cutoffs == null)
			calculateScalingInformation(training_data);

		// Scale the training data based on scaling factors
		training_data = getScaledFeatures(training_data);

		// Find the feature selection and/or weighting settings, if appropriate
		boolean[] features_to_use = null;
		double[] feature_selection_error_rates = null;
		double[] feature_weights = null;
		double[] feature_weighting_error_rates = null;
		if (using_feature_selection || using_feature_weighting)
		{
			// Find the subset of the training data to use for training and testing
			// for the puropses of feature selection and/or weighting
			boolean[] reduced_training_samples = getReducedTrainingSamples(boolean_model_results, fsw_training_fraction);

			// Find reduced number of training and testing data
			int number_reduced_training_samples = 0;
			for (int i = 0; i < reduced_training_samples.length; i++)
				if (reduced_training_samples[i])
					number_reduced_training_samples++;
			int number_reduced_testing_samples = reduced_training_samples.length - number_reduced_training_samples;
			
			// Generate the reduced training and testing data and their models
			double[][] reduced_training_data = new double[number_reduced_training_samples][];
			int[][] reduced_int_model_training_results = new int[number_reduced_training_samples][];
			int training_count = 0;
			double[][] reduced_testing_data = new double[number_reduced_testing_samples][];
			int[][] reduced_int_model_testing_results = new int[number_reduced_testing_samples][];
			int testing_count = 0;
			for (int i = 0; i < training_data.length; i++)
			{
				if (reduced_training_samples[i])
				{
					reduced_training_data[training_count] = training_data[i];
					reduced_int_model_training_results[training_count] = int_model_results[i];
					training_count++;
				}
				else
				{
					reduced_testing_data[testing_count] = training_data[i];
					reduced_int_model_testing_results[testing_count] = int_model_results[i];
					testing_count++;
				}
			}

			// Train a test knn with the reduced training and testing data
			KNN temp_knn = new KNN(categories.length);
			temp_knn.addTrainingSamples(reduced_training_data, reduced_int_model_training_results);

			// Find a good feature selection
			if (using_feature_selection)
			{
				// Update the training_monitor
				if (training_monitor != null)
					training_monitor.startBottomTask(Math.abs(iterations), "Training 1-D feature selection.");

				// Set up the evaluator to test performance of chromosomes
				FeatureSelectionEvaluator feature_selection_evaluator =
					new FeatureSelectionEvaluator(temp_knn, reduced_testing_data, reduced_int_model_testing_results, results_interpereter);

				// Set up a breeder to oversee evolution
				Breeder breeder = new Breeder(feature_selection_evaluator, population_size, mutation_probability,
				                              mating_rate, allow_elite, number_of_villages, migration_delay, migration_size);

				// Carry out evolution and find the best performing feature selection
				short[] best_feature_selection = new short[feature_selection_evaluator.getBitStringLength()];
				feature_selection_error_rates = evolve(iterations, acceptable_threshold, consecutive_iterations,
				                                       breeder, best_feature_selection);

				// Set features_to_use to the settings of the best performing bit string
				Object[] features_to_use_objects = feature_selection_evaluator.parseBitString(best_feature_selection);
				features_to_use = new boolean[features_to_use_objects.length];
				for (int feat = 0; feat < features_to_use_objects.length; feat++)
					features_to_use[feat] = ((Boolean) features_to_use_objects[feat]).booleanValue();
			}

			// Find a good feature weighting
			if (using_feature_weighting)
			{
				// Update the training_monitor
				if (training_monitor != null)
					training_monitor.startBottomTask(Math.abs(iterations), "Training 1-D feature weighting.");

				// Set the temp_knn's feature selection to the best one found during
				// feature selection (or null if feature selection wasn't performed)
				temp_knn.features_to_use = features_to_use;

				// Set up the evaluator to test performance of chromosomes
				FeatureWeightingEvaluator feature_weighting_evaluator =
					new FeatureWeightingEvaluator(temp_knn, reduced_testing_data, reduced_int_model_testing_results, word_length, results_interpereter);

				// Set up a breeder to oversee evolution
				Breeder breeder = new Breeder(feature_weighting_evaluator, population_size, mutation_probability,
				                              mating_rate, allow_elite, number_of_villages, migration_delay, migration_size);

				// Carry out evolution and find the best performing feature weightings
				short[] best_feature_weightings = new short[feature_weighting_evaluator.getBitStringLength()];
				feature_weighting_error_rates = evolve(iterations, acceptable_threshold, consecutive_iterations,
				                                       breeder, best_feature_weightings);

				// Set feature_weights to the settings of the best performing bit string
				Object[] features_weights_objects = feature_weighting_evaluator.parseBitString(best_feature_weightings);
				feature_weights = new double[features_weights_objects.length];
				for (int feat = 0; feat < features_weights_objects.length; feat++)
					feature_weights[feat] = ((Double) features_weights_objects[feat]).doubleValue();
			}
		}

		// Train the knn classifier with the complete given feature_sets
		if (knn == null)
			knn = new KNN(categories.length);
		knn.addTrainingSamples(training_data, int_model_results);
		knn.features_to_use = features_to_use;
		knn.feature_weights = feature_weights;

		// Return the average error for each epoch. This is the data for feature weighting if this,
		// option was selected, the data for feature selection if this option was selected but
		// feature weighting was not and null otherwise (since basic KNN does not need training
		// iteration.
		if (using_feature_weighting)
			return feature_weighting_error_rates;
		else if (using_feature_selection)
			return feature_selection_error_rates;
		else
			return null;
	}


	/**
	 * Returns the relative scores of each of the possible categories when the given sets
	 * of features are classified. There is one entry in the returned array for each of the
	 * entries in the <i>categories</i> field, and they appear in the same order. Higher scores
	 * in the returned array correspond to a greater certainty that the feature set should
	 * have the corresponding label. Scores fall in the range between 0.0 and 1.0. The
	 * first indice of the returned array corresponds to the feature set and the second
	 * corresponds to the category. The order of the categories is the same as in the
	 * categories field and the order of the feature sets is the same as the order in which
	 * they were passed to the <i>feature_sets</i> parameter.
	 *
	 * <p>The <i>feature_sets</i> parameter specifies the feature sets to be classified. The
	 * first indice corresponds to different feature sets. The second indice corresponds to
	 * different features in the given featue set. It should be noted that all feature sets
	 * must use the same features in the same order as given in the </i>feature_labels</i>
	 * parameter.
	 *
	 * <p>The <i>feature_labels</i> parameter specifies the names of each of the features
	 * in the <i>feature_sets</i> parameter. The features in the <i>feature_sets</i> parameter
	 * will automatically matched to the features in the <i>feature_names</i> field based on
	 * the content of the <i>feature_labels</i> parameter unless a value of <i>null</i> is passed
	 * to the <i>feature_labels</i>. In this case, the feature values in the <i>feature_sets</i>
	 * parameter will simply be fed into the classifier in the order that they occur.
	 *
	 * <p>An exception if thrown if the <i>feature_labels</i> do not contain the same names as
	 * <i>feature_names</i> (although a different ordering is permitted) or if any of the
	 * feature sets in <i>feature_sets</i> have a different number of features than
	 * <i>feature_names</i>. Also throws an exception if the KNN classifier is untrained.
	 */
	public double[][] classify(double[][] feature_sets, String[] feature_labels)
		throws Exception
	{
		// Throw an exception if the KNN classifier is untrained.
		if (knn == null)
			throw new Exception("KNN classifier is untrained");

		// Ensure that the features in feature_labels match feature_names and order
		// the features in classification_data to match the order in feature_names. An exception
		// is thrown if any problems occur.
		double[][] classification_data;
		if (feature_labels != null)
			classification_data = getOrderedFeatureSets(feature_sets, feature_labels);
		else
		{
			classification_data = new double[feature_sets.length][];
			for (int rec = 0; rec < classification_data.length; rec++)
			{
				classification_data[rec] = new double[feature_sets[rec].length];
				for (int feat = 0; feat < classification_data[rec].length; feat++)
					classification_data[rec][feat] = feature_sets[rec][feat];
			}
		}

		// Scale the classification_data based on scaling factors
		classification_data = getScaledFeatures(classification_data);

		// Find the number of training samples that are in the k nearest samples around the test
		// point for each category
		double[][] results = new double[classification_data.length][categories.length];
		for (int rec = 0; rec < classification_data.length; rec++)
			results[rec] = knn.classifyTestSample(classification_data[rec]);

		// Normalize the results of each recording in order to get the fraction of the
		// training samples within k that fell within each category
		results = bodhidharma.utilities.Statistics.normalize(results);

		// Return the classification scores
		return results;
	}


	/**
	 * Returns an array that represents the features_to_scale parameter with values scaled
	 * to fall between 0 and 1 based on previous training of the BioKNearestNeighbour.
	 * The first indice of both the <i>feature_sets</i> parameter and the return array
	 * identifies the recording and the second identifies the feature. The order of features
	 * in the returned array is the same as the order in the <i>feature_labels</i> parameter.
	 *
	 * <p>The <i>feature_labels</i> parameter specifies the names of each of the features
	 * in the <i>feature_sets</i> parameter. The features in the <i>feature_sets</i> parameter
	 * will automatically matched to the features in the <i>feature_names</i> field based on
	 * the content of the <i>feature_labels</i> parameter.
	 *
	 * <p>The acual values of the <i>feature_sets</i> parameter are not themselves changed.
	 *
	 * <p>An exception if thrown if the <i>feature_labels</i> do not contain the same names as
	 * <i>feature_names</i> (although a different ordering is permitted) or if any of the
	 * feature sets in <i>feature_sets</i> have a different number of features than
	 * <i>feature_names</i>. Also throws an exception if the scaling factors have not yet
	 * been calculated.
	 */
	public double[][] getScaledFeatureValues(double[][] feature_sets, String[] feature_labels)
		throws Exception
	{
		// Throw an exception if the scaling factors have not yet been calculated.
		if (min_feature_value_cutoffs == null)
			throw new Exception("Scaling factors have not yet been calculated.\n" +
			                    "Ond-dimensional feature classifier must be trained first.");

		// Ensure that the features in feature_labels match feature_names and order
		// the features in feature_values to match the order in feature_names. An exception
		// is thrown if any problems occur.
		double[][] feature_values = getOrderedFeatureSets(feature_sets, feature_labels);

		// Calculate the scaled values of the features
		double[][] scaled_values = getScaledFeatures(feature_values);

		// Reorder the scaled values to match the order of the features in feature_labels
		double[][] ordered_scaled_values = new double[scaled_values.length][feature_labels.length];
		int[] proper_indices = new int[feature_names.length]; // will give the indice to which features should be moved
		for (int i = 0; i < proper_indices.length; i++)
			proper_indices[i] = -1;
		for (int i = 0; i < feature_labels.length; i++)
			for (int j = 0; j < feature_names.length; j++)
				if (feature_labels[i].equals(feature_names[j]))
					proper_indices[i] = j;
		for (int set = 0; set < feature_sets.length; set++)
			for (int feat = 0; feat < feature_sets[set].length; feat++)
				ordered_scaled_values[set][ proper_indices[feat] ] = scaled_values[set][feat];

		// Return the scaled values in the same feature order that was provided
		return ordered_scaled_values;
	}
	

	/**
	 * Returns the minimum allowable value for each of the features. Features below
	 * their respective values are rounded up to it. Features occur in the same order
	 * as in the <i>feature_names</i> field.
	 */
	public double[] getMinFeatureValueCutoffs()
		throws Exception
	{
		// Throw an exception if the scaling factors have not yet been calculated.
		if (min_feature_value_cutoffs == null)
			throw new Exception("Minimum cutoffs have not yet been calculated.\n" +
			                    "Ond-dimensional feature classifier must be trained first.");
		
		return min_feature_value_cutoffs;
	}


	/**
	 * Returns the maximum allowable value for each of the features. Features above
	 * their respective values are rounded down to it.
	 */
	public double[] getMaxFeatureValueCutoffs()
		throws Exception
	{
		// Throw an exception if the scaling factors have not yet been calculated.
		if (max_feature_value_cutoffs == null)
			throw new Exception("Maximum cutoffs have not yet been calculated.\n" +
			                    "Ond-dimensional feature classifier must be trained first.");

		return max_feature_value_cutoffs;
	}


	/* PRIVATE METHODS *********************************************************/


	// Calculates the values of the min_feature_value_cutoffs and max_feature_value_cutoffs
	// fields based on the contents of the features_to_base_scaling_on parameter. The
	// first indice of this parameter identifies the recording and the second indice identifies
	// the feature.
	// The min_feature_value_cutoffs and max_feature_value_cutoffs values are calculated
	// to be 2 standard deviations away from the average feature values. Minima below 0 are raised to 0.
	// The features_to_base_scaling_on parameter is not altered.
	private void calculateScalingInformation(double[][] features_to_base_scaling_on)
	{
		// Assemble arrays containing the values of the features for each recording
		double[][] features = new double [feature_names.length][features_to_base_scaling_on.length];
		for (int feat = 0; feat < features.length; feat++)
			for (int rec = 0; rec < features[feat].length; rec++)
				features[feat][rec] = features_to_base_scaling_on[rec][feat];

		// Calculate the means for each of the features
		double[] means = new double[feature_names.length];
		for (int feat = 0; feat < features.length; feat++)
			means[feat] = bodhidharma.utilities.Statistics.getAverage(features[feat]);

		// Calculate the standard deviations for each of the features
		double[] standard_deviations = new double[feature_names.length];
		for (int feat = 0; feat < features.length; feat++)
			standard_deviations[feat] = bodhidharma.utilities.Statistics.getStandardDeviation(features[feat]);

		// Calculate the min and max cutoff values for each feature by setting so all data
		// will fall within 2 standard deviations of the mean
		max_feature_value_cutoffs = new double[features.length];
		min_feature_value_cutoffs = new double[features.length];
		for (int feat = 0; feat < feature_names.length; feat++)
		{
			double twice_sd = 2.0 * standard_deviations[feat];
			max_feature_value_cutoffs[feat] = means[feat] + twice_sd;
			min_feature_value_cutoffs[feat] = means[feat] - twice_sd;
		}

		// Raise min_feature_value_cutoffs to 0 if it is negative
		for (int feat = 0; feat < feature_names.length; feat++)
			if (min_feature_value_cutoffs[feat] < 0.0)
				min_feature_value_cutoffs[feat] = 0;
	}


	// Return an array that represents the features_to_scale parameter with values scaled
	// to fall between 0 and 1 based on previous training of the BioKNearestNeighbour.
	// The first indice of both the parameter and the return array identifies the recording
	// and the second identifies the feature. All feature values above the max_feature_value_cutoffs
	// or below the min_feature_value_cutoffs for each feature are set to either the min or the max,
	// as the case may be. All values are then scaled so that a value of min_feature_value_cutoffs
	// for a feature has a value of 0 and a value of max_feature_value_cutoffs for a feature has a
	// value of 1. If max_feature_value_cutoffs and min_feature_value_cutoffs are the same, then
	// the value is set to 0. The original features_to_scale parameter is not changed.
	private double[][] getScaledFeatures(double[][] features_to_scale)
	{
		// The scaled features that will be returned
		double[][] scaled_features = new double[features_to_scale.length][feature_names.length];
		for (int rec = 0; rec < features_to_scale.length; rec++)
			for (int feat = 0; feat < feature_names.length; feat++)
				scaled_features[rec][feat] = features_to_scale[rec][feat];

		// Cut off all feature values over two standard deviations away from the mean
		// by setting them equal to two standard deviations awary from the mean
		for (int feat = 0; feat < feature_names.length; feat++)
		{
			for (int rec = 0; rec < features_to_scale.length; rec++)
			{
				if (scaled_features[rec][feat] > max_feature_value_cutoffs[feat])
					scaled_features[rec][feat] = max_feature_value_cutoffs[feat];
				if (scaled_features[rec][feat] < min_feature_value_cutoffs[feat])
					scaled_features[rec][feat] = min_feature_value_cutoffs[feat];
			}
		}

		// Subtract the min_feature_value_cutoffs from all feature values
		for (int feat = 0; feat < feature_names.length; feat++)
			for (int rec = 0; rec < features_to_scale.length; rec++)
				scaled_features[rec][feat] -= min_feature_value_cutoffs[feat];

		// Divide all feature values by the differnce between its max_feature_value_cutoffs
		// and min_feature_value_cutoffs
		for (int feat = 0; feat < feature_names.length; feat++)
		{
			double difference = max_feature_value_cutoffs[feat] - min_feature_value_cutoffs[feat];
			for (int rec = 0; rec < features_to_scale.length; rec++)
			{
				if (difference == 0.0)
					scaled_features[rec][feat] = 0.0;
				else
					scaled_features[rec][feat] = scaled_features[rec][feat] / difference;
			}
		}

		// Returned the scaled version of features_to_scale
		return scaled_features;
	}


	// Returns an array whose indice identifies the sample (in the same order
	// as the first indice of the model_results parameter). Each entry is set
	// to true if the sample is to be used for training and to false if it
	// is to be used for testing.
	// 
	// The first indice of the model_results parameter identifies the recording
	// and the second identifies the category. An entry of true means that the
	// given recording belongs to the given category.
	//
	// The training_fraction identifies the fraction of all samples that should
	// be used for training. This value must be between 0 and 1.
	//
	// The fraction of samples belonging to each category given by training_sample
	// is used for training. In other words, steps are taken to ensure that the
	// number of samples belonging to each category used for training is proportionately 
	// the same as the fraction of the total number of samples belonging to each category.
	// Samples that do not belong to any known category are also divided up using training_fraction.
	private boolean[] getReducedTrainingSamples(boolean[][] model_results, double training_fraction)
		throws Exception
	{
		// The total number of recordings
		int num_recordings = model_results.length;

		// The total number of categories
		int num_categories = model_results[0].length;
		
		// Identifies which recordings are to be used for training
		boolean[] reduced_training_list = new boolean[num_recordings];
		for (int samp = 0; samp < reduced_training_list.length; samp++)
			reduced_training_list[samp] = false;
		
		// Choose the training samples that belong to one of the accepted categories
		for (int cat = 0; cat < num_categories; cat++)
		{
			// Find the number of samples belonging to this category that
			// have not already been selected for training
			int number_belonging_to_cat = 0;
			for (int samp = 0; samp < num_recordings; samp++)
				if (model_results[samp][cat] && !reduced_training_list[samp])
					number_belonging_to_cat++;

			// Store the sample indices of samples belonging to this category that
			// have not already been selected for training
			int[] indices_belonging_to_category = new int[number_belonging_to_cat];
			int current_sample = 0;
			for (int samp = 0; samp < num_recordings; samp++)
				if (model_results[samp][cat] && !reduced_training_list[samp])
				{
					indices_belonging_to_category[current_sample] = samp;
					current_sample++;
				}
			
			// Randomly select training_fraction of the category for training by randomly
			// generating a number for each sample that belongs to cat and has not already
			// been chosen, and choosing the highest number (as set by training_fraction).
			int random_numbers[] = new int[indices_belonging_to_category.length];
			for (int i = 0; i < indices_belonging_to_category.length; i++)
				random_numbers[i] = bodhidharma.utilities.Statistics.generateRandomNumber(indices_belonging_to_category.length * 3);
			int num_to_train = (int) (((double) number_belonging_to_cat) * training_fraction);
			for (int i = 0; i < num_to_train; i++)
			{
				int index_of_largest = bodhidharma.utilities.Statistics.getIndexOfLargest(random_numbers);
				random_numbers[index_of_largest] = 0; // to avoid double counting
				reduced_training_list[ indices_belonging_to_category[index_of_largest] ] = true;
			}
		}

		// Choose additional training samples that do not belong to any of the accepted categories
		boolean[] non_belongers = new boolean[num_recordings];
		for (int rec = 0; rec < num_recordings; rec++)
			non_belongers[rec] = true;
		for (int rec = 0 ; rec < num_recordings; rec++)
			for (int cat = 0; cat < num_categories; cat++)
				if (model_results[rec][cat])
					non_belongers[rec] = false;
		int number_non_belongers = 0;
		for (int rec = 0; rec < num_recordings; rec++)
			if (non_belongers[rec])
				number_non_belongers++;
		int[] indices_belonging_to_no_category = new int[number_non_belongers];
		int current_rec = 0;
		for (int rec = 0; rec < num_recordings; rec++)
			if (non_belongers[rec])
			{
				indices_belonging_to_no_category[current_rec] = rec;
				current_rec++;
			}
		int random_numbers[] = new int[indices_belonging_to_no_category.length];
		for (int i = 0; i < indices_belonging_to_no_category.length; i++)
			random_numbers[i] = bodhidharma.utilities.Statistics.generateRandomNumber(indices_belonging_to_no_category.length * 3);
		int num_to_train = (int) (((double) number_non_belongers) * training_fraction);
		for (int i = 0; i < num_to_train; i++)
		{
			int index_of_largest = bodhidharma.utilities.Statistics.getIndexOfLargest(random_numbers);
			random_numbers[index_of_largest] = 0; // to avoid double counting
			reduced_training_list[ indices_belonging_to_no_category[index_of_largest] ] = true;
		}

		// Return the information identifying which samples should be used for training
		return reduced_training_list;
	}


	// Causes the given breeder to keep evolving chromosomes until either
	// the given conditions (seec below) are met or the given maximum number of epochs
	// is exceeded. Also sets the contents of the best_DNA parameter (whose initial
	// contents are irrelevant) to the bit string of the best performing chromosome
	// of the last epoch.
	//
	// The returned array of doubles gives the value of (1 - fitness) for the fitness of
	// the best performing chromosome for each epoch. A value of 0 in this array is
	// indicative of perfect performance of the chromosome and an error of 1 indicates
	// complete failure.
	// 
	// The max_epochs parameter specifies the number of training iterations performed.
	// If a negative value is passed here, then the number of iterations to perform
	// is calculated automatically based on the acceptable_threshold parameter, which 
	// specifies the absolute rate of change of the fitness of the best performing
	// chromosome below which training will stop, and the consecutive_epochs
	// parameter, which specifies the number of consecutive epochs for which the rate
	// of change must be below this threshold in order for training to stop. The number
	// of epochs that go by will never exceed the absolute value of the max_epochs
	// value, irregardless of the other parameters.
	//
	// Throws an exception if there is a problem during evolution.
	private double[] evolve(int max_epochs, double acceptable_threshold, 
	                        int consecutive_epochs, Breeder breeder,
	                        short[] best_DNA)
		throws Exception
	{
		// The error for each epoch that went by during evolution.
		// Calculated by subtracting the fitness of the best chromosome from 1
		double[] reported_errors = new double[Math.abs(max_epochs)];
		for (int i = 0; i < reported_errors.length; i++)
			reported_errors[i] = -1;

		// Perform the evolution and end evolution when appropriate
		int maximum_epochs = Math.abs(max_epochs);
		int below_threshold_so_far = 0; // number of consecutive iterations with error below acceptable_threshold
		for (int epoch = 0; epoch < maximum_epochs; epoch++)
		{
			// Update the training_monitor
			if (training_monitor != null)
			{
				training_monitor.setBottomProgressValue(epoch);
				if (training_monitor.isCancelled())
					throw new Exception("Training cancelled.");
			}

			// Evolve over one epoch
			Object[] evolution_report = breeder.evolve();

			// Store the error of the best performing chromosome for each epoch
			reported_errors[epoch] = 1 - (((double[]) evolution_report[2])[0]);

			// Store the DNA (bit string) of the current best performer
			short[] temp_DNA = (short[]) evolution_report[3];
			for (int bit = 0; bit < best_DNA.length; bit++)
				best_DNA[bit] = temp_DNA[bit];

			// Check if should end evolution
			if (max_epochs < 0)
			{
				// Check for convergence
				if (epoch > 0)
				{
					if ( Math.abs(reported_errors[epoch] - reported_errors[epoch - 1]) < acceptable_threshold)
					{
						below_threshold_so_far++;
						if (below_threshold_so_far == consecutive_epochs)
						{
							double[] temp = new double[epoch + 1];
							for (int i = 0; i < temp.length; i++)
								temp[i] = reported_errors[i];
							reported_errors = temp;
							epoch = maximum_epochs + 1; // exit the loop
						}
					}
					else
						below_threshold_so_far = 0;
				}
			}
		}

		// Return the error measurements
		return reported_errors;
	}
}