/*
 * @(#)FeedForwardNeuralNetwork.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers.feedforward_neural_networks;

import bodhidharma.classifiers.SupervisedClassifier;
import java.io.*;
import bodhidharma.xml_parsing.*;

/**
 * An interface for using a backpropagation feedforward neural network that is
 * actually implemented as a <code>LinkedNodes</code> object. The neural networks
 * used here are limited to perceptrons and networks with one layer of hidden nodes.
 * The learning rate, the momentum and the range of initial seed weights are user
 * definable. The number of hidden nodes is set to a user definable coefficient
 * multiplied by the square root of the sum of the number of input and output
 * nodes.
 * 
 * <p>One input node is assigned for each feature and one output node is assigned
 * for each category. The sigmoidal activation function is used, and all values
 * of output nodes are trained to fall between 0.2 and 0.8. Output nodes
 * are trained to 0.8 if a feature set belongs to the corresponding feature
 * set and to 0.2 if it does not. However, these are scaled by the
 * <code>classify</code> method to fall between 0.0 and 1.0, so this detail
 * is hidden from the user.
 *
 * <p>This classifier can be trained using exemplar based learning to classify
 * arbitrary feature sets. This implementation makes it possible to assign more
 * than one label to a single feature set.
 *
 * <p>Feature sets are fed into the classifier as arrays of doubles. Categories
 * are specified as arrays of Strings.
 * 
 * <p>Use the <code>train</code> method to train the classifier (the
 * <code>getFeatureNames</code> method is useful for deteriming the features that
 * can be fed to the classifer).
 *
 * <p>One constructor is provided for creating a new network and one is provided
 * for parsing XML code and using it to reconstruct a trained network.
 *
 * <p>Use the <code>classify</code> method to classify feature sets once training
 * has been completed (the <code>getCategories</code> method is useful for
 * determining what categories feature sets can be classified into and for
 * determining the order of the categories when parsing classification results).
 *
 * <p>Use the <code>save</code> method to save the classifier and its current
 * state to disk.
 * 
 * <p>Use the <code>getClassifierName</code> and <code>getClassifierParameters</code>
 * methods to obtain information about the classifier.
 *
 * <p>Use the <code>getClassifierIdentifier</code> method to get a name or code that
 * was given to an instantiation of the classifier when it was constructed. This
 * identifier can be used by external classes to identify the instantiation.
 *
 * @author Cory McKay
 */
public class FeedForwardNeuralNetwork
	extends SupervisedClassifier
{
	/* FIELDS ******************************************************************/

	
	// The actual nodes of the neural network
	private LinkedNodes	network;


	/* CONSTRUCTOR *************************************************************/

	
	/**
	 * Generate a neural network with the given parameters and randomly generated weights.
	 * Sets the number of hidden units to the coefficient specified in the <i>network_settings</i>
	 * parameter multiplied by the square root of the sum of the number of input and output
	 * nodes.
	 *
	 * @param	labels_of_features		The names of the features that will be fed into the input nodes.
	 * @param	labels_of_categories	The names of categories that correspond to the output nodes.
	 * @param	identification_string	An identifier that can be associated with the network so that outside classes can identify it.
	 * @param	network_settings		Contains network initialization settings.
	 */
	public FeedForwardNeuralNetwork(String[] labels_of_features, String[] labels_of_categories,
	                                String identification_string, NeuralNetworkJFrame network_settings)
	{
		// Set superclass' fields
		feature_names = labels_of_features;
		categories = labels_of_categories;
		identifier = identification_string;

		// Find initialization settings
		double hidden_unit_coefficient = network_settings.getHiddenUnitCoefficient();
		double learning_rate = network_settings.getLearningRate();
		double momentum = network_settings.getMomentum();
		double min_initial_weight = network_settings.getMinimumInitialWeight();
		double max_initial_weight = network_settings.getMaximumInitialWeight();

		// Calculate additional initialization settings
		int number_input_units = labels_of_features.length;
		int number_output_units = labels_of_categories.length;
		int number_hidden_units = (int) (hidden_unit_coefficient * Math.sqrt( (double) (number_input_units + number_output_units) ));
		
		// Instantiate the network itself
		network = new LinkedNodes(learning_rate, momentum, number_input_units,
		                          number_hidden_units, number_output_units,
		                          min_initial_weight, max_initial_weight);

		// Set training_monitor
		training_monitor = null;
	}


	/**
	 * Parse the file specified by the given file path to recreate the specificed trained
	 * classifier. Throws an exception if problems occur during parsing.
	 *
	 * @param	file_path			The file path of a FeedForwardNeuralNetwork_file that this instantiation is to be based on.
	 */
	public FeedForwardNeuralNetwork(String file_path)
		throws Exception
	{
		Object[] results = (Object[]) XMLDocumentParser.parseXMLDocument(file_path, "FeedForwardNeuralNetwork_file");
		String[] fixed_fields = (String[]) results[0];
		
		identifier = fixed_fields[5];
		categories = (String[]) results[1];
		feature_names = (String[]) results[2];
		
		double learning_rate = (Double.valueOf(fixed_fields[0])).doubleValue();
		double momentum = (Double.valueOf(fixed_fields[1])).doubleValue();
		int number_input_nodes = (Integer.valueOf(fixed_fields[2])).intValue();
		int number_hidden_nodes = (Integer.valueOf(fixed_fields[3])).intValue();
		int number_output_nodes = (Integer.valueOf(fixed_fields[4])).intValue();
		double[] weights = (double[]) results[3];
		network = new LinkedNodes(learning_rate, momentum, number_input_nodes,
		                          number_hidden_nodes, number_output_nodes, weights);

		training_monitor = null;
	}


	/* PUBLIC METHODS **********************************************************/

	
	/**
	 * Returns the name of the type of classifier. This information can be used
	 * by external classes to identify the type of classifier in situations such
	 * as when the configuration of a trained classifier is being read from a file.
	 */
	public String getClassifierName()
	{
		return new String("FeedForwardNeuralNetwork");
	}


	/**
	 * Returns a String describing the parameters of the classifier. This includes
	 * classifier name, learning rate, momentum and number of hidden nodes.
	 */
	public String getClassifierParameters()
	{
		String parameters = new String();
		parameters += "Classifier type: " + getClassifierName() + "\n";
		parameters += "Learning rate: " + network.getLearningRate() + "\n";
		parameters += "Momentum: " + network.getMomentum() + "\n";
		parameters += "Number of hidden nodes: " + network.getHiddenNodesCount();
		return parameters;
	}


	/**
	 * Trains the network using the given feature sets. The first indice of the
	 * <i>unscrambled_feature_sets</i> parameter corresponds to different feature sets. The
	 * second indice corresponds to different features in the given featue set.
	 * It should be noted that all feature sets must use the same features in
	 * the same order as given in the </i>feature_labels</i> parameter.
	 *
	 * <p>The <i>feature_labels</i> parameter specifies the names of each of the features
	 * in the <i>unscrambled_feature_sets</i> parameter. The features in the <i>unscrambled_feature_sets</i>
	 * parameter will automatically matched to the features in the <i>feature_names</i> field based on
	 * the content of the <i>feature_labels</i> parameter unless a value of <i>null</i> is passed
	 * to the <i>feature_labels</i>. In this case, the feature values in the <i>unscrambled_feature_sets</i>
	 * parameter will simply be fed into the classifier in the order that they occur.
	 *
	 * <p>The <i>unscrambled_model_categories</i> parameter gives the categories of each of
	 * the given feature sets. The first indice corresponds to the feature set.
	 * The second indice corresponds to different model categories for the given
	 * feature set. Only categories to which the feature set belongs should be included.
	 *
	 * <p>The <i>iterations</i> parameter specifies the number of training iterations
	 * performed.  If a negative value is passed here, then the number of iterations to perform
	 * is calculated automatically based on the <i>acceptable_threshold</i> parameter, which 
	 * specifies the absolute rate of change of the sum of squared error below which training will
	 * stop, and the <i>consecutive_iterations</i> parameter, which specifies the number of consecutive
	 * iterations for which the rate of change must be below this threshold in order for
	 * training to stop. The number of iterations that go by will never exceed the absolute
	 * valud of the <i>iterations</i> value, irregardless of the other parameters.
	 *
	 * <p>For example, if a value of 1000 is given for <i>iterations</i>, then 1000 iterations
	 * will be performed irregardless of the other parameters. If a value of -1000 is given,
	 * then training will automatically stop if the absolute value of the rate of change
	 * of the sum of squared error from one sample to the next falls below the
	 * <i>acceptable_threshold</i> parameter for <i>consecutive_iterations</i> iterations, but
	 * no more than 1000 iterations will be performed in any case.
	 *
	 * <p>The returned double is an average sum of squared error after training iterations. The indice
	 * of the returned array corresponds to the iteration of training that the error is associated
	 * with.
	 *
	 * <p>An exception if thrown if the <i>feature_labels</i> do not contain the same names as
	 * <i>feature_names</i> (although a different ordering is permitted) or if any of the
	 * feature sets in <i>unscrambled_feature_sets</i> have a different number of features than
	 * <i>feature_names</i>. An exception is also thrown if <i>unscrambled_feature_sets</i> and
	 * <i>unscrambled_model_categories</i> have different sizes in regard to their first parameters.
	 * An exception is thrown if the <i>given_results</i> parameter contains a name not
	 * present in the <i>categories</i> field or if it contains the same category more than once.
	 */
	public double[] train(double[][] unscrambled_feature_sets, String[] feature_labels,
	                      String[][] unscrambled_model_categories, int iterations,
	                      double acceptable_threshold, int consecutive_iterations)
		throws Exception
	{
		// Ensure that model_categories and feature_sets have the same size with regards
		// to their first parameters.
		if (unscrambled_feature_sets.length != unscrambled_model_categories.length)
			throw new Exception("A total of " + unscrambled_feature_sets.length + " feature sets were " +
			                    "provided to the classifier, but " + unscrambled_model_categories.length +
			                    " model classifications were provided. These two numbers must" +
			                    " match.");

		// Find a random ordering of the recordings in unscrambled_feature_sets and 
		// unscrambled_model_categories so that they are in a random ordering. This facilitates
		// learning by the networks.
		int[] scrambled_indices = bodhidharma.utilities.Statistics.getRandomOrdering(unscrambled_feature_sets.length);
		double[][] feature_sets = new double[unscrambled_feature_sets.length][];
		String[][] model_categories = new String[unscrambled_model_categories.length][];
		for (int rec = 0; rec < unscrambled_feature_sets.length; rec++)
		{
			feature_sets[rec] = unscrambled_feature_sets[ scrambled_indices[rec] ];
			model_categories[rec] = unscrambled_model_categories[ scrambled_indices[rec] ];
		}

		// Ensure that the features in feature_labels match feature_names and order
		// the features in training_data to match the order in feature_names. An exception
		// is thrown if any problems occur.
		double[][] training_data;
		if (feature_labels != null)
			training_data = getOrderedFeatureSets(feature_sets, feature_labels);
		else
		{
			training_data = new double[feature_sets.length][];
			for (int rec = 0; rec < training_data.length; rec++)
			{
				training_data[rec] = new double[feature_sets[rec].length];
				for (int feat = 0; feat < training_data[rec].length; feat++)
					training_data[rec][feat] = feature_sets[rec][feat];
			}
		}

		// Generate the model results and throw an exception if there is a problem with
		// model_categories. Entries are set to true if a given feature set should belong
		// to a particular category and to false otherwise.
		boolean[][] model_results = getModelResults(model_categories);

		// Prepare the expected output values for the neural nets (setting yes to 0.8
		// and no to 0.2 because of the sigmoidal activation function)
		double[][] model_results_values = new double[model_results.length][];
		for (int i = 0; i < model_results_values.length; i++)
		{
			model_results_values[i] = new double[model_results[i].length];
			for (int j = 0; j < model_results_values[i].length; j++)
			{
				if (model_results[i][j])
					model_results_values[i][j] = 0.8;
				else
					model_results_values[i][j] = 0.2;
			}
		}

		// The maximum number of iterations to do
		int max_iterations = Math.abs(iterations);

		// The average sum-of-squared-error for each training iteration
		double[] average_error = new double[max_iterations];

		// Train the network, keeping track of the average error for each iteration
		if (iterations > 0) // case where a fixed number of iterations are to be performed
		{
			for (int iteration = 0; iteration < max_iterations; iteration++)
			{
				// Update the training_monitor and check if job cancelled
				if (training_monitor != null)
				{
					training_monitor.setBottomProgressValue(iteration);
					if (training_monitor.isCancelled())
						throw new Exception("Training cancelled.");
				}

				// Perform training
				double error_sum = 0;
				for (int set = 0; set < training_data.length; set++)
				{
					double error = network.training_iteration(training_data[set], model_results_values[set]);
					error_sum += error;
				}
				average_error[iteration] = error_sum / (double) training_data.length;
			}
		}
		else // case where training is ended automatically in case of convergence of sum of squared error
		{
			int below_threshold_so_far = 0; // number of consecutive iterations with error below acceptable_threshold
			for (int iteration = 0; iteration < max_iterations; iteration++)
			{
				// Update the training_monitor
				if (training_monitor != null)
				{
					training_monitor.setBottomProgressValue(iteration);
					if (training_monitor.isCancelled())
						throw new Exception("Training cancelled.");
				}

				// Perform a training iteration
				double error_sum = 0;
				for (int set = 0; set < training_data.length; set++)
				{
					double error = network.training_iteration(training_data[set], model_results_values[set]);
					error_sum += error;
				}
				average_error[iteration] = error_sum / (double) training_data.length;

				// Check for convergence
				if (iteration > 0)
				{
					if ( Math.abs(average_error[iteration] - average_error[iteration - 1]) < acceptable_threshold)
					{
						below_threshold_so_far++;
						if (below_threshold_so_far == consecutive_iterations)
						{
							double[] temp = new double[iteration + 1];
							for (int i = 0; i < temp.length; i++)
								temp[i] = average_error[i];
							average_error = temp;
							iteration = max_iterations + 1; // exit the loop
						}
					}
					else
						below_threshold_so_far = 0;
				}
			}
		}

		// Return the average error for each iteration
		return average_error;
	}

	
	/**
	 * Returns the relative scores of each of the possible categories when the given sets
	 * of features are classified. There is one entry in the returned array for each of the
	 * entries in the <i>categories</i> field, and they appear in the same order. Higher scores
	 * in the returned array correspond to a greater certainty that the feature set should
	 * have the corresponding label. Scores fall in the range between 0.0 and 1.0. The
	 * first indice of the returned array corresponds to the feature set and the second
	 * corresponds to the category. The order of the categories is the same as in the
	 * categories field and the order of the feature sets is the same as the order in which
	 * they were passed to the <i>feature_sets</i> parameter.
	 *
	 * <p>The <i>feature_sets</i> parameter specifies the feature sets to be classified. The
	 * first indice corresponds to different feature sets. The second indice corresponds to
	 * different features in the given featue set. It should be noted that all feature sets
	 * must use the same features in the same order as given in the </i>feature_labels</i>
	 * parameter.
	 *
	 * <p>The <i>feature_labels</i> parameter specifies the names of each of the features
	 * in the <i>feature_sets</i> parameter. The features in the <i>feature_sets</i> parameter
	 * will automatically matched to the features in the <i>feature_names</i> field based on
	 * the content of the <i>feature_labels</i> parameter unless a value of <i>null</i> is passed
	 * to the <i>feature_labels</i>. In this case, the feature values in the <i>feature_sets</i>
	 * parameter will simply be fed into the classifier in the order that they occur.
	 *
	 * <p>An exception if thrown if the <i>feature_labels</i> do not contain the same names as
	 * <i>feature_names</i> (although a different ordering is permitted) or if any of the
	 * feature sets in <i>feature_sets</i> have a different number of features than
	 * <i>feature_names</i>.
	 */
	public double[][] classify(double[][] feature_sets, String[] feature_labels)
		throws Exception
	{
		// Ensure that the features in feature_labels match feature_names and order
		// the features in classification_data to match the order in feature_names. An exception
		// is thrown if any problems occur.
		double[][] classification_data;
		if (feature_labels != null)
			classification_data = getOrderedFeatureSets(feature_sets, feature_labels);
		else
		{
			classification_data = new double[feature_sets.length][];
			for (int rec = 0; rec < classification_data.length; rec++)
			{
				classification_data[rec] = new double[feature_sets[rec].length];
				for (int feat = 0; feat < classification_data[rec].length; feat++)
					classification_data[rec][feat] = feature_sets[rec][feat];
			}
		}

		// Collect the output of the network for each of the feature sets
		double[][] network_output = new double[feature_sets.length][];
		for (int set = 0; set < network_output.length; set++)
			network_output[set] = network.get_output(classification_data[set]);
	
		// Scale values to fall between 0.0 and 1.0 in (remember that network was trained
		// to 0.2 for false and to 0.8 for true because of sigmoidal activation function)
		for (int i = 0; i < network_output.length; i++)
			for (int j = 0; j < network_output[i].length; j++)
			{
				if (network_output[i][j] < 0.2)
					network_output[i][j] = 0.2;
				else if (network_output[i][j] > 0.8)
					network_output[i][j] = 0.8;

				network_output[i][j] = (network_output[i][j] - 0.2) * (1.6666667);
			}

		// Return the classification scores
		return network_output;
	}


	/**
	 * Saves all of the fields to the given file. Throws an exception if a problem occurs.
	 */
	public void	save(File place_to_save)
		throws Exception
	{
		FileOutputStream to = new FileOutputStream(place_to_save);
		DataOutputStream writer = new DataOutputStream(to);
		String header = new String
		(
			"<?xml version=\"1.0\"?>\n" +
			"<!DOCTYPE FeedForwardNeuralNetwork_file [\n" +
			"   <!ELEMENT FeedForwardNeuralNetwork_file (comments, identifier, categories, feature_names, learning_rate, momentum, number_input_nodes, number_hidden_nodes, number_output_nodes, weights)>\n" +
			"   <!ELEMENT comments (#PCDATA)>\n" +
			"   <!ELEMENT identifier (#PCDATA)>\n" +
			"   <!ELEMENT categories (cat+)>\n" +
			"   <!ELEMENT cat (#PCDATA)>\n" +
			"   <!ELEMENT feature_names (feat+)>\n" +
			"   <!ELEMENT feat (#PCDATA)>\n" +
			"   <!ELEMENT learning_rate (#PCDATA)>\n" +
			"   <!ELEMENT momentum (#PCDATA)>\n" +
			"   <!ELEMENT number_input_nodes (#PCDATA)>\n" +
			"   <!ELEMENT number_hidden_nodes (#PCDATA)>\n" +
			"   <!ELEMENT number_output_nodes (#PCDATA)>\n" +
			"   <!ELEMENT weights (w+)>\n" +
			"   <!ELEMENT w (#PCDATA)>\n" +
			"]>\n\n" +
			"<FeedForwardNeuralNetwork_file>\n\n" +
			"   <comments></comments>\n\n"
		);
		writer.writeBytes(header);
		writer.writeBytes("   <identifier>" + identifier + "</identifier>\n");
		writer.writeBytes("   <categories>\n");
		for (int i = 0; i < categories.length; i++)
			writer.writeBytes("      <cat>" + categories[i] + "</cat>\n");
		writer.writeBytes("   </categories>\n");
		writer.writeBytes("   <feature_names>\n");
		for (int i = 0; i < feature_names.length; i++)
			writer.writeBytes("      <feat>" + feature_names[i] + "</feat>\n");
		writer.writeBytes("   </feature_names>\n");
		writer.writeBytes("   <learning_rate>" + network.getLearningRateValue() + "</learning_rate>\n");
		writer.writeBytes("   <momentum>" + network.getMomentumValue() + "</momentum>\n");
		writer.writeBytes("   <number_input_nodes>" + network.getNumberOfInputUnits() + "</number_input_nodes>\n");
		writer.writeBytes("   <number_hidden_nodes>" + network.getNumberOfHiddenUnits() + "</number_hidden_nodes>\n");
		writer.writeBytes("   <number_output_nodes>" + network.getNumberOfOutputUnits() + "</number_output_nodes>\n");
		double[] weights = network.getWeights();
		writer.writeBytes("   <weights>\n");
		if (weights != null)
			for (int i = 0; i < weights.length; i++)
				writer.writeBytes("      <w>" + weights[i] + "</w>\n");
		writer.writeBytes("   </weights>\n");
		writer.writeBytes("</FeedForwardNeuralNetwork_file>");
		writer.close();
	}
}