/*
 * @(#)NeuralNetworkJFrame.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma.classifiers.feedforward_neural_networks;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import bodhidharma.PreferencesPanel;

/**
 * Dialog box used to hold and edit the basic parameters of new feedforward neural
 * networks.
 *
 * <p><b>IMPORTANT:<b> These parameters are used when <b>new</b> neural network systems
 * are created. They do not affect existing neural network systems.
 * 
 * <p>The <i>Hidden Unit Coef</i> field sets the coefficient that is used to multiply
 * the value that is used to calculate the number of hidden units in a network. A
 * value between 0 and 1 will decrease the number of hidden units and a value above 1
 * will increase it.
 * 
 * <p>The <i>Learning Rate</i> field sets the learning rate of the neural network.
 * This affects iterations until convergence and probability of falling into a local
 * minimum in error space.
 * 
 * <p>The <i>Momentum</i> field sets the momentum of the neural network.
 * This affects iterations until convergence and probability of falling into a local
 * minimum in error space.
 * 
 * <p>The <i>Min Initial Weight</i> field sets the minimum value of weights between
 * network units when these weights are initially randomly generated.
 *
 * <p>The <i>Max Initial Weight</i> field sets the maximum value of weights between
 * network units when these weights are initially randomly generated.
 * 
 * <p>It should be noted that error messages are given if these parameters are incorrectly
 * set.
 *
 * <p>The various get methods of this class can be used by external objects to access
 * the stored parameters.
 *
 * <p>The <i>OK</i> button stores the entered settings. An error message is displayed
 * if there is a problem with the parameters.
 *
 * <p>The <i>Cancel</i> button cancels the entered settings and restores the previous
 * values.
 *
 * @see	FeedForwardNeuralNetwork
 * @see Breeder
 * @see PreferencesPanel
 *
 * @author Cory McKay
 */
 public class NeuralNetworkJFrame
	extends JFrame
	implements ActionListener
{
	/* FIELDS ******************************************************************/

	// Window that instantiated this window
	private PreferencesPanel	parent_window;
	
	// Containers to hold Swing elements
	private Container		content_pane;
	private JPanel			settings_panel;
	private JPanel			button_panel;
	
	// Genetic algorithm parameters
	private double			hidden_unit_coefficient;
	private double			learning_rate;
	private double			momentum;
	private double			min_initial_weight;
	private double			max_initial_weight;

	// Genetic algorithm parameter editors
	private JTextArea		hidden_unit_coefficient_text_area;
	private JTextArea		learning_rate_text_area;
	private JTextArea		momentum_text_area;
	private JTextArea		min_initial_weight_text_area;
	private JTextArea		max_initial_weight_text_area;

	// Buttons
	private JButton			ok_button;
	private JButton			cancel_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. Configures the panel and its fields. Removes access to the
	 * parent <code>JFrame</code>.
	 *
 	 * @param	parent	Name of the <code>JFrame</code> window that spawned this dialog box.
	 */
	public NeuralNetworkJFrame(PreferencesPanel parent)
	{
		// Store parent recording
		parent_window = parent;

		// Configure overall window settings
		setTitle("Feedforward Neural Network Initial Settings");
		int horizontal_gap = 6; // horizontal space between GUI elements
		int vertical_gap = 11; // horizontal space between GUI elements
		content_pane = getContentPane();
		content_pane.setLayout(new BorderLayout(horizontal_gap, vertical_gap));
		settings_panel = new JPanel(new GridLayout(5, 2, horizontal_gap, vertical_gap));
		button_panel = new JPanel(new FlowLayout());

		// Configure the settings graphical elements
		int number_columns = 7;
		hidden_unit_coefficient_text_area = new JTextArea("1.0", 1, number_columns);
		learning_rate_text_area = new  JTextArea("0.1", 1, number_columns);
		momentum_text_area = new  JTextArea("0.1", 1, number_columns);
		min_initial_weight_text_area = new  JTextArea("0.1", 1, number_columns);
		max_initial_weight_text_area = new  JTextArea("0.7", 1, number_columns);

		// Instantiate and configure buttons
		cancel_button = new JButton("Cancel");
		ok_button = new JButton("OK");
		cancel_button.addActionListener(this);
		ok_button.addActionListener(this);
		button_panel.add(cancel_button);
		button_panel.add(ok_button);

		// Add labels and previously created elements to panels and container
		settings_panel.add(new JLabel("Hidden Unit Coef:"));
		settings_panel.add(hidden_unit_coefficient_text_area);
		settings_panel.add(new JLabel("Learning Rate:"));
		settings_panel.add(learning_rate_text_area);
		settings_panel.add(new JLabel("Momentum:"));
		settings_panel.add(momentum_text_area);
		settings_panel.add(new JLabel("Min Initial Weight:"));
		settings_panel.add(min_initial_weight_text_area);
		settings_panel.add(new JLabel("Max Initial Weight:"));
		settings_panel.add(max_initial_weight_text_area);

		// Set up top display
		content_pane.add(settings_panel, BorderLayout.CENTER);
		content_pane.add(button_panel, BorderLayout.SOUTH);

		// Cause program to react when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				cancel();
			}
		});

		// Set neural netwrork parameters equal to defaults and show an error
		// message if there is a problem
		try { readGraphicalSettings(); }
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			this.show();
			this.pack();
			parent_window.parent_window.setEnabled(false);
		}
	}


	/* PUBLIC METHODS **********************************************************/

	
	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the cancel_button
		if (event.getSource().equals(cancel_button))
			cancel();

		// React to the ok_button
		else if (event.getSource().equals(ok_button))
			ok();
	}

	
	/**
	 * Return the value of the hidden_unit_coefficient field
	 */
	public double getHiddenUnitCoefficient()
	{
		return hidden_unit_coefficient;
	}


	/**
	 * Return the value of the learning_rate field
	 */
	public double getLearningRate()
	{
		return learning_rate;
	}


	/**
	 * Return the value of the momentum field
	 */
	public double getMomentum()
	{
		return momentum;
	}


	/**
	 * Return the value of the min_initial_weight field
	 */
	public double getMinimumInitialWeight()
	{
		return min_initial_weight;
	}


	/**
	 * Return the value of the elitism field
	 */
	public double getMaximumInitialWeight()
	{
		return max_initial_weight;
	}


	/**
	 * Set the value of the hidden_unit_coefficient and hidden_unit_coefficient_text_area fields
	 */
	public void setHiddenUnitCoefficient(String value)
		throws Exception
	{
		hidden_unit_coefficient_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the learning_rate and learning_rate_text_area fields
	 */
	public void setLearningRate(String value)
		throws Exception
	{
		learning_rate_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the momentum and momentum_text_area fields
	 */
	public void setMomentum(String value)
		throws Exception
	{
		momentum_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the min_initial_weight and min_initial_weight_text_area fields
	 */
	public void setMinimumInitialWeight(String value)
		throws Exception
	{
		min_initial_weight_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the elitism and max_initial_weight_text_area fields
	 */
	public void setMaximumInitialWeight(String value)
		throws Exception
	{
		max_initial_weight_text_area.setText(value);
		readGraphicalSettings();
	}


	/* PRIVATE METHODS *********************************************************/


	// Read settings in GUI elements into neural network fields.
	// Throws an exception if improper parameters are entered.
	private void readGraphicalSettings()
		throws Exception
	{
		// Check for exceptions
		if (Double.parseDouble(hidden_unit_coefficient_text_area.getText()) < 0.0)
			throw new Exception("You have entered a negative value for the Hidden Unit Coef.\n" +
								"Please adjust this parameter.");
		else if (Double.parseDouble(learning_rate_text_area.getText()) < 0.0)
			throw new Exception("You have entered a negative value for the Learning Rate.\n" +
								"Please adjust this parameter.");
		else if (Double.parseDouble(momentum_text_area.getText()) < 0.0)
			throw new Exception("You have entered a negative value for the Momentum.\n" +
								"Please adjust this parameter.");
		else if (Double.parseDouble(min_initial_weight_text_area.getText()) < 0.0)
			throw new Exception("You have entered a negative value for the Min Initial Weight.\n" +
								"Please adjust this parameter.");
		else if (Double.parseDouble(max_initial_weight_text_area.getText()) < 0.0)
			throw new Exception("You have entered a negative value for the Max Initial Weight.\n" +
								"Please adjust this parameter.");
		
		// Set the genetic algorithm parameters
		hidden_unit_coefficient = Double.parseDouble(hidden_unit_coefficient_text_area.getText());
		learning_rate = Double.parseDouble(learning_rate_text_area.getText());
		momentum = Double.parseDouble(momentum_text_area.getText());
		min_initial_weight = Double.parseDouble(min_initial_weight_text_area.getText());
		max_initial_weight = Double.parseDouble(max_initial_weight_text_area.getText());
	}


	// Close the dialog box without registering any changes in the
	// neural network fields. Reset graphical elements to hold
	// contents of neural network fields.
	private void cancel()
	{
		restoreGraphicalValuesToStoredValues();
		parent_window.parent_window.setEnabled(true);
		this.hide();
	}


	// Close the dialog box and update the neural network fields
	// to reflect any changes. Displays an error message if there
	// are any problems
	private void ok()
	{
		// Update the neural network parameter fields and display
		// an error message if necessary
		try 
		{
			readGraphicalSettings(); 
		
			// Reactivate the parent window and hide this dialog box
			parent_window.parent_window.setEnabled(true);
			this.hide();
		}
		catch (Exception e) { JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE); }
	}


	// Reset graphical elements to hold contents of neural network fields.
	private void restoreGraphicalValuesToStoredValues()
	{
		hidden_unit_coefficient_text_area.setText(Double.toString(hidden_unit_coefficient));
		learning_rate_text_area.setText(Double.toString(learning_rate));
		momentum_text_area.setText(Double.toString(momentum));
		min_initial_weight_text_area.setText(Double.toString(min_initial_weight));
		max_initial_weight_text_area.setText(Double.toString(max_initial_weight));
	}
}