/*
 * @(#)Breeder.java	1.0	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers.genetic_algorithms;

/**
 * Objects of this class control the evolution of <code>Chromosome</code> objects based on their fitness
 * as judged by an <code>Evaluator</code> object and on the parameters specified when objects of this
 * class are instantiated.
 *
 * <p>The <code>evovle</code> method of this class can be called to carry out one epoch of evolution at a
 * time. Chromosomes may be divided into villages that evolve independantly, but exchange a given portion 
 * of their population after a number of epochs. The population of each village stays constant in each generation.
 * Elitism is also possible, whereby the best performing members of the population are cloned, and therefore
 * passed on to the next generation unaltered. Crossover occurs at random locations in the parents.
 
 * <p>A modified wheel of fortune algortihm is used to select parents that will engage in crossover during
 *  reproduction. This parent selection algorithm works as follows:
 *
 * <p>1) The number of parents per village is found by multiplying the number of chromosomes in a village by
 * the mating rate.
 * <br>2) Parents are selected in each village until the number of parents is equal to the number determined
 * in step one. Each chromosome has a probability of being selected based on its fitness relative to the
 * fitness of other chromosomes in its village. Each parent may only be selected once at this stage (although
 * a single parent may reproduce polygamously multiple times in step 3).
 * <br>3) All chromosomes that were not selected as parents in step 2 are replaced by offspring from the chromosomes
 * selected as parents in the same village in step 2. Which parents pair up for crossover is randomly determined.
 * <br>4) Chromosomes selected in step two are randomly paired up in each village. Their offspring replace them.
 *
 * <p>The advantage of this approach over the traditional wheel of fortune is that it allows the degree of genetic
 * diversity to be controlled with the mating rate parameter. A low mating rate causes only a few parents to
 * produce all of the offspring of the next generation, whereas a high mating rate causes the offspring to
 * be descendants of many parents. The first case may increase performance, but specialisation may be a problem,
 * since genetic diversity will be limited. The second case helps guard against specialisation, but weakly performing
 * chromosomes are much more likely to breed, thus perpetuating poor genetic material. Allowing the user to control
 * the mating rate parameter gives better insights into how the evolution works and allows for interesting experimentation.
 *
 * <p>After the crossover phase is complete, mutation may occur, based on the probability of mutation parameter. The
 * probability is the likelihood that a bit will flip in the DNA of each chromosome. This is a good way of introducing
 * new genetic variations.
 *
 * <p>After this phase, a clone of the best performing chromosome of the parent generation in each village will replace
 * a randomly selected member of each village if the elitism option is selected. This ensures the perpetuation of the
 * best genes so far.
 *
 * <p>The use of villages allows separate pools of chromosomes to develop independantly. Some chromosomes will periodically
 * exchange with other villages to introduce any positive developments that they have evolved. Which villages exchange is 
 * randomly determined each time that an exchange occurs.
 *
 * <p><code>Breeder</code> objects to not contain any assumptions about the meaning of bit strings.
 *
 *
 * <p>The <code>evolve</code> method of this class can be called to carry out one epoch of evolution.
 *
 * @see	Evaluator
 * @see	Breeder
 */
public class Breeder
{
	/* FIELDS ******************************************************************/

	
	private int epoch;					// the number of the current epoch
	private Evaluator evaluator;		// evaluates the fitness of a given chromosome
	private int population_size;		// number of chromosomes in the population
	private int DNA_length;				// length of the bit string that represents DNA
	private double mutation_prob;		// probability of mutation in an offspring
	private double mating_rate;			// fraction of population allowed to reproduce (0 to 1)
	private boolean allow_elite;		// controls whether best chromosome per village is cloned exactly
	private int number_villages;		// number of vilages that segregate chromosomes
	private int village_size;			// number of chromosomes in each village
	private int migration_delay;		// number of epochs between migrations between villages
	private int migration_size;			// number of chromosomes that move during migration
	private int reproduce_per_vil;		// number chromosomes that will be parents per village
	private Chromosome[][] population;	// the Chromosomes that make up the population (organized by village)
	private double[][] temp;			// used to store temporary markers regarding chromosomes


	/* CONSTRUCTORS ************************************************************/


	/**
	 * Instantiates a new <code>Breeder</code> object to control the evolutions of <code>Chromosomes</code>.
	 *
	 * <p><b>IMPORTANT:</b>The population size must be a multiple of the number of villages. There must be more
	 * than one elligible parent per village. Migration size must be smaller than village size.
	 *
	 * <p>An exception is thrown if there is a problem instantiating <i>Chromosomes</i>.
	 *
	 * @param	eval			An <code>Evaluator</code> object that evaluates the fitness of individual <code>Chromosomes</code>
	 * @param	pop_size		Total number of chromosomes in the population.
	 * @param	mut_prob		Probability per bit of mutation in an offspring <code>Chromosome</code>.
	 * @param	mate_rate		Fraction of population allowed to reproduce (between 0 to 1).
	 * @param	elite_allowed	Whether or not the best chromosome per village is cloned exactly during evolution.
	 * @param	num_villages	Number of vilages that segregate chromosomes.
	 * @param	mig_delay		Number of epochs between migrations between villages.
	 * @param	mig_size		Number of chromosomes exchanged between villages when exchange occurs.
	 */
	public Breeder (Evaluator eval, int pop_size, double mut_prob, double mate_rate,
	         boolean elite_allowed, int num_villages, int mig_delay, int mig_size)
		throws Exception
	{
		epoch = 0;
		evaluator = eval;
		population_size = pop_size;
		DNA_length = eval.getBitStringLength();
		mutation_prob = mut_prob;
		mating_rate = mate_rate;
		allow_elite = elite_allowed;
		number_villages = num_villages;
		village_size = population_size / number_villages;
		migration_delay = mig_delay;
		migration_size = mig_size;
		reproduce_per_vil = (int) (mating_rate * village_size);
		population = new Chromosome[number_villages][village_size];
		temp = new double[number_villages][village_size];

		// Populate population array with randomly generated chromosomes
		for (int i = 0; i < number_villages; i++)
		{
			for (int j = 0; j < village_size; j++)
				population[i][j] = new Chromosome(DNA_length, evaluator);
		}

		// Ensure that an even number of parents are selected as parents per village
		if (reproduce_per_vil % 2 != 0)
			reproduce_per_vil--;

		// Initialize temp
		resetTemp();
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Carry out one epoch of evolution and return an array of objects containing the
	 * following sets of information for the previous generation:
	 *
	 *  <p>- array of doubles representing the average fitness of each village
	 *  <br>- array of doubles representing the average fitness of the members selected
	 *  to reproduce for each village
	 *  <br>- the fitness of the best chromosome in the entire population
	 *  <br>- the bit string making up the DNA of the best chromosome in the entire population
	 */
	public Object[] evolve()
		throws Exception
	{
		// Fill temp with the normalised probability that each chromosome will breed
		calculateProbability();
		
		// Enter codes into temp to indicate which chromosomes will breed and which are
		// the fittest in each village
		chooseParents();

		// Generate information for parent generation
		Object[] generation_data = generateInfo();

		// Reproduce using crossover
		Chromosome best_chromosomes[] = reproduce();
		resetTemp();

		// Mutate offspring if appropriate
		if (mutation_prob > 0.0) mutate();

		// Clone elite parents if appropriate
		if (allow_elite == true) clone(best_chromosomes);

		// Exchange chromosomes between villages if appropriate
		if (epoch % migration_delay == 0 && epoch != 0 && number_villages > 1) migrate();
		
		// Keep count of the number of epochs processed
		epoch++;

		return generation_data;
	}


	/* PRIVATE METHODS *********************************************************/


	// Fill temp with the normalised probability that each Chromosome will breed,
	// based on the fitness of each Chromosome relative to the other Chromosomes
	// in its village
	// NOTE: assumes fitness can never be negative
	private void calculateProbability()
	{
		double sum = 0.0; // sum of fitness for each village
		for (int i = 0; i < number_villages; i++)
		{
			sum = 0.0;
			for (int j = 0; j < village_size; j++)
				sum += population[i][j].getFitness();
			for (int j = 0; j < village_size; j++)
				temp[i][j] = population[i][j].getFitness() / sum;
		}
	}


	// Select which chromosomes will reproduce in each village. Each chromosome
	// has a liklihood of being a parent based on its fitness relative to other
	// chromosomes in its village (wheel of fortune approach).
	// reproduce_per_vil parents are selected per village to breed polygamously
	// each epoch.
	// Fills temp array with 3's or -3's (3 if chromosome has been chosen to
	// reproduce, -3 if it hasn't) for best chromosome per village, 2's for other
	// chromosomes selected to reproduce and -1's for chromosomes that will die
	// childless.
	// NOTE: assumes that calculateProbability() was called first
	private void chooseParents()
	{
		// Find best performing chromosome per village
		int[] indices_of_best = new int[number_villages];
		double max_so_far;
		int index_of_max_so_far;
		for (int i = 0; i < number_villages; i++)
		{
			max_so_far = 0.0;
			index_of_max_so_far = 0;
			for (int j = 0; j < village_size; j++)
			{
				if (temp[i][j] > max_so_far)
				{
					index_of_max_so_far = j;
					max_so_far = temp[i][j];
				}
			}
			indices_of_best[i] = index_of_max_so_far;
		}

		// Prepare temp for wheel of fortune by ranking probabilities sequentially
		double sum_so_far;
		for (int i = 0; i < number_villages; i++)
		{
			sum_so_far = 0.0;
			for (int j = 0; j < village_size; j++)
			{
				temp[i][j] += sum_so_far;
				sum_so_far = temp[i][j];
			}
		}

		// Mark parent chromosomes with 2's in temp
		int number_selected_so_far; // number parents found
		double random_number; // used to determine parents
		int placeholder; // index of a selected parent
		double lower_bound; // lower probability bound
		short found_one; // indicates a parent found
		boolean[] selected = new boolean[village_size]; // indicates chosen parents for a village
		for (int i = 0; i < number_villages; i++)
		{
			number_selected_so_far = 0;
			for (int j = 0; j < village_size; j++)
				selected[j] = false;
			
			int consecutive_non_succesful_iters = 0;
			while (number_selected_so_far < reproduce_per_vil)
			{
				found_one = 0;
				random_number = Math.random();
				lower_bound = 0.0;
				placeholder = -1;

				for (int j = 0; j < village_size; j++)
				{
					if (random_number >= lower_bound && random_number < temp[i][j] )
					{
						if(!selected[j])
						{
							placeholder = j;
							found_one = 1;
						}
						else if (consecutive_non_succesful_iters > village_size)// if keep hitting ones that have already been selected
						{
							for (int z = j + 1; z <= temp[i].length; z++)
							{
								// Wrap aroung
								if (z ==  temp[i].length)
									z = 0;

								// See if is selected
								if(!selected[z])
								{
									placeholder = z;
									found_one = 1;
									z = temp[i].length + 2; // exit loop
								}
							}
						}
						else
							consecutive_non_succesful_iters++;
					}
					lower_bound = temp[i][j];
				}
				if (found_one == 1)
				{
					selected[placeholder] = true;
					number_selected_so_far++;
					consecutive_non_succesful_iters = 0;
				}
			}
			for (int j = 0; j < village_size; j++)
			{
				if (selected[j])
					temp[i][j] = 2.0;
			}
		}

		// Mark best chromosomes per village with 3's in temp if they're reproducing
		// and -3's if they're not
		for (int i = 0; i < number_villages; i++)
		{
			if (temp[i][ indices_of_best[i] ] > 1.5) temp[i][ indices_of_best[i] ] = 3.0;
			else temp[i][ indices_of_best[i] ] = -3.0;
		}

		// Mark all other (non-reproducing) chromosomes with -1
		for (int i = 0; i < number_villages; i++)
			for (int j = 0; j < village_size; j++)
				if (temp[i][j] < 1.5 && temp[i][j] > -1.0)
					temp[i][j] = -1.0;

	}


	// Return an array of objects containing the following sets of information:
	//  - array of doubles representing the average fitness of each village
	//  - array of doubles representing the average fitness of the members selected
	//    to reproduce for each village
	//  - single cell array of double containing the fitness of the best chromosome in the
	//    entire population
	//  - the bit string making up the DNA of the best chromosome in the entire population
	// NOTE: assumes that chooseParents was called first
	private Object[] generateInfo()
	{
		Object[] data = new Object[4];
		double[] overall_average = new double[number_villages];
		double[] reproducing_average = new double[number_villages];
		double[] best_fitness = new double[1];
		int x_index_of_best_fitness = 0;
		int y_index_of_best_fitness = 0;
		short[] best_DNA = new short[DNA_length];
		double overall_sum;
		double reproducing_sum;
		int reproducing_count;

		best_fitness[0] = 0.0;
		for (int i = 0; i < number_villages; i++)
		{
			overall_sum = 0.0;
			reproducing_sum = 0.0;
			reproducing_count = 0;
			for (int j = 0; j < village_size; j++)
			{
				overall_sum += population[i][j].getFitness();
				if (temp[i][j] > 0.0)
				{
					reproducing_count++;
					reproducing_sum += population[i][j].getFitness();
				}
				if (population[i][j].getFitness() > best_fitness[0])
				{
					best_fitness[0] = population[i][j].getFitness();
					x_index_of_best_fitness = i;
					y_index_of_best_fitness = j;
				}
			}
			overall_average[i] = overall_sum / ((double) village_size);
			reproducing_average[i] = reproducing_sum / ((double) reproducing_count);
		}
	
		data[0] = overall_average;
		data[1] = reproducing_average;
		data[2] = best_fitness;
		data[3] = population[x_index_of_best_fitness][y_index_of_best_fitness].getDNA();
		return data;
	}


	// Cause all chromosomes marked to reproduce to do so using crossover at a random
	// point.
	// First each chromosomes marked to die childless (i.e. -1 or -3 in temp) is replaced by 
	// an offspring from randomly selected parents eligible parents (i.e. 2 or 3 in temp).
	// Then each chromosome marked to reproduce (i.e. 2 or 3 in temp) pairs itself with another
	// chromosome selected to reproduce and they replace themselves with their offspring.
	// Children belonging to the next generation are marked with a 10.0 in temp.
	// An array holding the fittest members of each village from the previous generation
	// (marked -3 or 3 in temp) is returned in case elitism (cloning) is implemented.
	// An exception is thrown if there is a problem instantiating Chromosomes.
	// NOTE: assumes that chooseParents was called first
	private Chromosome[] reproduce()
		throws Exception
	{
		// Copy best chromosomes for possible cloning
		Chromosome[] best_chromosomes = new Chromosome[number_villages];
		for (int i = 0; i < number_villages; i++)
		{
			for (int j = 0; j < village_size; j++)
				if (temp[i][j] > 2.5 || temp[i][j] < -2.5)
					best_chromosomes[i] = population[i][j];
		}
		
		// Replace childless chromosomes with offspring of eligible parents
		int first_parent_index = -1;
		int second_parent_index = -1;
		for (int i = 0; i < number_villages; i++)
		{
			for (int j = 0; j < village_size; j++)
			{
				if (temp[i][j] < 0.0)
				{
					first_parent_index = findPartner(-1, i); // find any eligible parent
					second_parent_index = findPartner(first_parent_index, i);
					population[i][j] = crossover(population[i][first_parent_index],
					                             population[i][second_parent_index])[0];
					temp[i][j] = 10.0;
				}
			}
		}

		// Replace parent chromosomes with offspring
		first_parent_index = -1;
		second_parent_index = -1;
		Chromosome[] offspring;
		for (int i = 0; i < number_villages; i++)
		{
			for (int j = 0; j < village_size; j++)
			{
				if (temp[i][j] > 0.0 && temp[i][j] < 9.5)
				{
					first_parent_index = j;
					second_parent_index = findPartner(first_parent_index, i);
					offspring = crossover(population[i][first_parent_index],
					                      population[i][second_parent_index]);
					population[i][first_parent_index] = offspring[0];
					population[i][second_parent_index] = offspring[1];
					temp[i][first_parent_index] = 10.0;
					temp[i][second_parent_index] = 10.0;
				}
			}
		}

		// Return best chromosomes for possible cloning
		return best_chromosomes;
	}


	// Cause all chromosomes in the population to mutate based on the mutation_prob
	private void mutate()
	{
		for (int i = 0; i < number_villages; i++)
		{
			for (int j = 0; j < village_size; j++)
				population[i][j].mutate(mutation_prob);
		}
	}


	// Replace a random member of each village with an exact clone of the best member
	// of the previous generation
	private void clone(Chromosome[] best_chromosomes)
	{
		for (int i = 0; i < number_villages; i++)
			population[i][ generateRandomNumber(village_size) ] = best_chromosomes[i];
	}


	// Cause a fraction of Chromosomes in each village to migrate to another village
	private void migrate()
	{
		int migrations_so_far = 0; // number of exchanges between villages that have occured
		int max_migrations = number_villages / 2;
		
		int[] status = new int[number_villages]; // marked with 1 if migration complete, 0 if not
		for (int i = 0; i < number_villages; i++)
			status[i] = 0;
		
		int first_village;
		int second_village;
		while (migrations_so_far < max_migrations)
		{
			// Find a pair of villages that haven't migrated yet
			first_village = generateRandomNumber(number_villages);
			while (status[first_village] != 0)
				first_village = generateRandomNumber(number_villages);
			status[first_village] = 1;
			second_village = generateRandomNumber(number_villages);
			while (status[second_village] != 0)
				second_village = generateRandomNumber(number_villages);
			status[second_village] = 1;

			// Randomly generate the set of villagers that will migrate from each village
			int[] first_villagers = getRandomArray(migration_size, village_size);
			int[] second_villagers = getRandomArray(migration_size, village_size);

			// Switch villagers between the two villagers
			Chromosome migrant;
			int second_index = 0;
			for (int first_index = 0; first_index < village_size; first_index++)
			{
				if (first_villagers[first_index] == 1)
				{
					migrant = population[first_village][first_index];
					while (second_villagers[second_index] != 1)
						second_index++;
					population[first_village][first_index] = population[second_village][second_index];
					population[second_village][second_index] = migrant;
					second_index++;
				}
			}

			// Increment number of exchanges
			migrations_so_far++;
		}
	}


	// Returns the index of an eligible (authourised to reproduce) partner for the
	// given parent in the given village
	private int findPartner(int parent_index, int village_index)
	{
		int found_one = 0;
		int test_index = 0;

		while (found_one == 0)
		{
			test_index = generateRandomNumber(village_size);
			if (temp[village_index][test_index] > 0.0
			    && temp[village_index][test_index] < 9.5 && test_index != parent_index)
				found_one = 1;
		}

		return test_index;
	}


	// Returns the two children of the given parents using crossover with a randomly
	// determined split point. An exception is thrown if there is a problem instantiating
	// Chromosomes.
	private Chromosome[] crossover(Chromosome first_parent, Chromosome second_parent)
		throws Exception
	{
		Chromosome[] children = new Chromosome[2];
		int split_point = generateRandomNumber(DNA_length);
		if (split_point == 0) split_point = 1;
		children[0] = new Chromosome(first_parent.getLeftDNA(split_point),
		                             second_parent.getRightDNA(split_point), evaluator);
		children[1] = new Chromosome(second_parent.getLeftDNA(split_point),
		                             first_parent.getRightDNA(split_point), evaluator);
		return children;
	}


	// Generates an array filled with 0's and a randomly placed number of 1's
	private int[] getRandomArray(int number_of_1s, int size)
	{
		int[] random_array = new int[size];
		for (int i = 0; i < size; i++)
			random_array[i] = 0;
		
		int changes = 0; // number of 1's placed so far
		int place;

		while (changes < number_of_1s)
		{
			place = generateRandomNumber(size);
			if (random_array[place] != 1)
			{
				random_array[place] = 1;
				changes++;
			}
		}

		return random_array;
	}


	// Returns a random integer from 0 to max - 1, based on the uniform distribution.
	// WARNING: max must be lest than 10 000
	private int generateRandomNumber(int max) 
	{
		int random_number = (int) (10000.0 * Math.random());
		return (random_number % max);
	}


	// Set all cells of temp to 0.0
	private void resetTemp()
	{
		for (int i = 0; i < number_villages; i++)
		{
			for (int j = 0; j < village_size; j++)
				temp[i][j] = 0.0;
		}
	}


	// Print the contents of temp to standard out. Used for testing purposes.
	private void printTemp()
	{
		for (int i = 0; i < number_villages; i++)
		{
			for (int j = 0; j < village_size; j++)
				System.out.print(temp[i][j] + " ");
			System.out.println();
		}
		System.out.println();
	}
}