/*
 * @(#)ClassifierSelectionEvaluator	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers.genetic_algorithms;

import bodhidharma.classifiers.bio_k_nearest_neighbour.*;
import bodhidharma.classifiers.ClassificationResultsInterpereter;
import java.util.LinkedList;

/**
 * Objects of this class evaluate and interpret the meaning of bit strings stored in
 * genetic algorithm <code>Chromosome</code> objects. Each bit in the bit string
 * represents the on (1) or off (0) setting of a classifier. Classifiers with an on
 * setting are used for final classification and classifiers with an off are not used
 * in final classification. This <code>Evaluator</code> is used for classifier selection.
 *
 * The fitness value is calculated by performing a combined classification using equal
 * weighting for all of the classifiers.
 *
 * @see	KNN
 * @see	Chromosome
 * @see Breeder
 */
public class ClassifierSelectionEvaluator
	extends Evaluator
{
	/* FIELDS ******************************************************************/


	// Test classifier result values that will be classified to test performance. First indice
	// indicates classifier, second indice indicates recording and third indice indicates category.
	// Value of 1 for first indice corresponds to single-dimensional feature classifier. Is null if
	// this classifier not used.
	private	double[][][] test_classifier_results;

	// Model classifications of test_classifier_results. First indice indicates recording.
	// Entries indicate the indices of the categories that are correct classifications.
	private	int[][]	correct_classifications;

	// The possible number of categories into which recordings can be classified.
	private int number_categories;

	// Used to perform actual classifications so that fitnesses can be performed. Default fitness will be used if null.
	private	ClassificationResultsInterpereter results_interpereter;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor.
	 *
	 * @param	test_classifier_results		Test classifier result values that will be classified to test performance. First indice indicates classifier, second indice indicates recording and third indice indicates category. Value of 1 for first indice corresponds to single-dimensional feature classifier. Is null if this classifier not used.
	 * @param	correct_classifications		Model classifications of test_classifier_results. First indice indicates recording. Entries indicate the indices of the categories that are correct classifications.
	 * @param	number_categories			The possible number of categories into which recordings can be classified.
	 * @param	results_interpereter		Used to perform actual classifications so that fitnesses can be performed. Default fitness will be used if null.
	 */
	public ClassifierSelectionEvaluator( double[][][] test_classifier_results,
	                                     int[][] correct_classifications,
	                                     int number_categories,
	                                     ClassificationResultsInterpereter results_interpereter )
	{
		bit_string_length = test_classifier_results.length;
		this.test_classifier_results = test_classifier_results;
		this.correct_classifications = correct_classifications;
		this.number_categories = number_categories;
		this.results_interpereter = results_interpereter;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Return the fitness of a chromosome given its DNA (bit string). The fitness
	 * should vary between 0 (least fit possible) and 1 (most fit possible). This
	 * need not be normalized. Bit strings passed to this method must be
	 * of the same length as the length specified in the constructor of this
	 * object.
	 *
	 * <p>If all classifiers in the bit string are set to off (0), then one of them
	 * is randomly chosen to be set to 1
	 *
	 * <p>The fitness is calculated for each recording and category, and then averaged
	 * over all recordings. 
	 *
	 * <p>There are two ways of calculating fitness. The first is used if the results_interpereter
	 * field is not null, and the second if it is null. The first way gives a fitness
	 * based on the fraction of correct classifications based on the given bit_string.
	 *
	 * <p>The alternative fitness for a category of a recording is set equal to the
	 * difference between the average score of the classifiers and 1 if the recording
	 * does belong to the given category and 0 if it does not.
	 *
	 * <p>Throws an exception if there is a problem with the classification or if
	 * the length of the <i>bit_string</i> is not equal to the length of the
	 * <i>bit_string_length</i> field.
	 */
	public double getFitness(short[] bit_string)
		throws Exception
	{
		// Throw an exception if the length of the bit_string is not equal to the
		// length of the bit_string_length> field
		if (bit_string.length != bit_string_length)
			throw new Exception("The given bit string has " + bit_string.length +
			                    " bits, but it should have " + bit_string_length + ".");

		// Find the number of classifiers used with this bit string
		double number_used_classifiers = 0.0;
		for (int clas = 0; clas < test_classifier_results.length; clas++)
		{
			if (test_classifier_results[clas] != null && bit_string[clas] != 0)
				number_used_classifiers++;
		}

		// If all entries in the bit string are set to off (0), then randomly set
		// one of them to 1
		if (number_used_classifiers == 0)
		{
			// Find the indices of the elligible classifiers
			LinkedList indices_of_elligible_classifiers_list = new LinkedList();
			for (int clas = 0; clas < test_classifier_results.length; clas++)
				if (test_classifier_results[clas] != null)
					indices_of_elligible_classifiers_list.add(new Integer(clas));
			Object[] temp_objs = indices_of_elligible_classifiers_list.toArray();
			int[] indices_of_elligible_classifiers = new int[temp_objs.length];
			for (int clas = 0; clas < indices_of_elligible_classifiers.length; clas++)
				indices_of_elligible_classifiers[clas] = ((Integer) temp_objs[clas]).intValue();

			// Choose one to set as selected
			int ran = bodhidharma.utilities.Statistics.generateRandomNumber(indices_of_elligible_classifiers.length);
			int selected_indice_to_activate = indices_of_elligible_classifiers[ran];

			// Set the selected classier and update number_used_classifiers
			bit_string[selected_indice_to_activate] = 1;
			number_used_classifiers = 1;
		}

		// Calculate the combined ensebmle results using the appropriate feature selection
		int number_recordings = correct_classifications.length;
		double[][] combined_results = new double[number_recordings][number_categories];
		for (int rec = 0; rec < number_recordings; rec++)
		{
			for (int cat = 0; cat < number_categories; cat++)
			{
				combined_results[rec][cat] = 0.0;
				for (int clas = 0; clas < test_classifier_results.length; clas++)
				{
					if (test_classifier_results[clas] != null && bit_string[clas] != 0)
						combined_results[rec][cat] += test_classifier_results[clas][rec][cat];
				}
				combined_results[rec][cat] = combined_results[rec][cat] / (double) number_used_classifiers;
			}
		}

		// The fitness
		double fitness;

		// Base fitness on basic scores
		if (results_interpereter == null)
		{
			// Calculate the model results in score form
			double[][] model_scores = new double[number_recordings][number_categories];
			for (int rec = 0; rec < number_recordings; rec++)
				for (int cat = 0; cat < number_categories; cat++)
					model_scores[rec][cat] = 0.0;
			for (int rec = 0; rec < number_recordings; rec++)
				for (int i = 0; i < correct_classifications[rec].length; i++)
					model_scores[rec][ correct_classifications[rec][i] ] = 1.0;

			// Calculate the errors of the ensemble
			double errors = 0.0;
			for (int rec = 0; rec < number_recordings; rec++)
				for (int cat = 0; cat < number_categories; cat++)
					errors += Math.abs(model_scores[rec][cat] - combined_results[rec][cat]);
			errors = errors / ((double) (number_recordings * number_categories));
	
			// Calculate the average fitness for all recordings
			fitness = 1.0 - errors;
		}

		// Base fitness on classifications
		else
		{
			// Calculate the model classifications
			boolean[][] model_winners = new boolean[number_recordings][number_categories];
			for (int rec = 0; rec < number_recordings; rec++)
				for (int cat = 0; cat < number_categories; cat++)
					model_winners[rec][cat] = false;
			for (int rec = 0; rec < number_recordings; rec++)
				for (int i = 0; i < correct_classifications[rec].length; i++)
					model_winners[rec][ correct_classifications[rec][i] ] = true;

			// Find the winners based on the given bit_string and classifier results
			boolean[][] winners = results_interpereter.getWinners(combined_results);

			// Find the correct fraction of categories found per recording
			int[] corrects = new int[number_recordings];
			for (int rec = 0; rec < corrects.length; rec++)
				corrects[rec] = correct_classifications[rec].length;
			double[] percentage_right_per_rec = new double[number_recordings];
			for (int rec = 0; rec < corrects.length; rec++)
			{
				int found_so_far = 0;
				for (int cat = 0; cat < winners[rec].length; cat++)
				{
					if (winners[rec][cat] && model_winners[rec][cat])
						found_so_far++;
				}
				percentage_right_per_rec[rec] = ((double) found_so_far) / ((double) corrects[rec]);
			}

			// Set the fitness to the fraction correct
			fitness = bodhidharma.utilities.Statistics.getAverage(percentage_right_per_rec);
		}		

		// Return the fitness
		return fitness;
	}


	/**
	 * Returns an array of Booleans holding the meaning of the given bit string
	 * of a <code>Chromosome</code>. Values of 1 in the given <i>bit_string</i>
	 * are set to true and values of 0 are set to false. There is one element
	 * in the returned array for each bit in the <i>bit_string</i>. Bit strings
	 * passed to this method should be of the same length as the length specified
	 * in the constructor of this object.
	 *
	 * <p>Automatically sets the first entry to false if test_classifier_results[0]
	 * is null.
	 *
	 * <p>Throws an exception if the length of the <i>bit_string</i> is not equal
	 * to the length of the bit_string_length> field.
	 */
	public Object[] parseBitString(short[] bit_string)
		throws Exception
	{
		// Throw an exception if the length of the bit_string is not equal to the
		// length of the bit_string_length> field
		if (bit_string.length != bit_string_length)
			throw new Exception("The given bit string has " + bit_string.length +
			                    " bits, but it should have " + bit_string_length + ".");

		Boolean[] classifiers_to_use = new Boolean[bit_string.length];
		for (int bit = 0; bit < classifiers_to_use.length; bit++)
		{
			if (bit_string[bit] == 0)
				classifiers_to_use[bit] = new Boolean(false);
			else
				classifiers_to_use[bit] = new Boolean(true);
		}

		// Set one-dimensional classifier to false if it is not used
		if (test_classifier_results[0] == null)
			classifiers_to_use[0] = new Boolean(false);

		return classifiers_to_use;
	}
}