/*
 * @(#)FeatureSelectionEvaluator	1.1.2 09/14/04
 *
 * McGill University
 */

package bodhidharma.classifiers.genetic_algorithms;

import bodhidharma.classifiers.ClassificationResultsInterpereter;
import bodhidharma.classifiers.bio_k_nearest_neighbour.*;

/**
 * Objects of this class evaluate and interpret the meaning of bit strings stored in
 * genetic algorithm <code>Chromosome</code> objects. Each bit in the bit string
 * represents the on (1) or off (0) setting of a feature. Features with an on
 * setting are used for classification and feautres with an off are not used in
 * classification. This <code>Evaluator</code> is used for feature selection.
 *
 * The fitness value is calculated by performing a classification using a
 * <code>KNN</code> classifier.
 *
 * @see	KNN
 * @see	Chromosome
 * @see Breeder
 */
public class FeatureSelectionEvaluator
	extends Evaluator
{
	/* FIELDS ******************************************************************/


	// The trained KNN classifier used to evaluate the fitness of DNA
	private	KNN	classifier;

	// Test feature values that will be classified to test performance. First indice
	// indicates recording and second indicates feature.
	private	double[][]	test_feature_values;

	// Model classifications of feature_values. First indice indicates recording.
	// Entries indicate the indices of the categories that are correct classifications.
	private	int[][]	correct_classifications;

	// Used to perform actual classifications so that fitnesses can be performed. Default fitness will be used if null.
	private	ClassificationResultsInterpereter results_interpereter;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. The <i>knn_classifier</i> must be trained with training points,
	 * but its <i>features_to_use</i> and <i>feature_weights</i> parameters should be
	 * untrained (i.e. null). The order of the features and categories passed in the
	 * <i>feature_values</i> and <i>classifications</i> parameters must be consistent
	 * with those that were used to train the <i>knn_classifier</i> externally.
	 *
	 * <p>Throws an exception if the <i>feature_values</i> or <i>classifications</i>
	 * parameters are of improper sizes.
	 *
	 * @param	knn_classifier			Trained KNN classifier used to evaluate the fitness of DNA.
	 * @param	feature_values			Test feature values that will be classified to test performance. First indice indicates recording and second indicates feature.
	 * @param	classifications			Model classifications of feature_values. First indice indicates recording. Entries indicate the indices of the categories that are correct classifications.
	 * @param	results_interpereter	Used to perform actual classifications so that fitnesses can be performed. Default fitness will be used if null.
	 */
	public FeatureSelectionEvaluator(KNN knn_classifier,
	                                 double[][] feature_values,
	                                 int[][] classifications, 
	                                 ClassificationResultsInterpereter results_interpereter )
		throws Exception
	{
		bit_string_length = feature_values[0].length;
		classifier = knn_classifier;
		test_feature_values = feature_values;
		correct_classifications = classifications;
		this.results_interpereter = results_interpereter;

		if (feature_values.length != classifications.length)
			throw new Exception("The given feature values and their classifications have " +
			                    "different sizes.");
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Return the fitness of a chromosome given its DNA (bit string). The fitness
	 * should vary between 0 (least fit possible) and 1 (most fit possible). This
	 * need not be normalized. Bit strings passed to this method must be
	 * of the same length as the length specified in the constructor of this
	 * object.
	 *
	 * <p>The fitness is calculated for each recording, and then averaged over all
	 * recordings. 
	 *
	 * <p>There are two ways of calculating fitness. The first is used if the results_interpereter
	 * field is not null, and the second if it is null. The first way gives a fitness
	 * based on the fraction of correct classifications based on the given bit_string.
	 *
	 * <p>The alternative fitness for each recording is set equal to the sum of the
	 * normalized scores returned by the classifier for those categories to which
	 * the recording belongs.
	 *
	 * <p>Throws an exception if there is a problem with the classification or if
	 * the length of the <i>bit_string</i> is not equal to the length of the
	 * <i>bit_string_length</i> field.
	 */
	public double getFitness(short[] bit_string)
		throws Exception
	{
		// Throw an exception if the length of the bit_string is not equal to the
		// length of the bit_string_length> field
		if (bit_string.length != bit_string_length)
			throw new Exception("The given bit string has " + bit_string.length +
			                    " bits, but it should have " + bit_string_length + ".");

		// Set the features_to_use field of classifier based on the value of the
		// passed bit_string
		Boolean[] interpereted_bit_string = (Boolean[]) parseBitString(bit_string);
		boolean[] features_to_use = new boolean[interpereted_bit_string.length];
		for (int feat = 0; feat < features_to_use.length; feat++)
			features_to_use[feat] = interpereted_bit_string[feat].booleanValue();
		classifier.features_to_use = features_to_use;

		// Classify the test_feature_values field
		double[][] results = new double[test_feature_values.length][];
		for (int rec = 0; rec < results.length; rec++)
			results[rec] = classifier.classifyTestSample(test_feature_values[rec]);

		// Normalize the results of each recording in order to get the fraction of the
		// training samples within k that fell within each category
		double[] totals = new double[results.length];
		for (int rec = 0; rec < results.length; rec++)
		{
			totals[rec] = 0.0;
			for (int cat = 0; cat < results[rec].length; cat++)
				totals[rec] += results[rec][cat];
		}
		for (int rec = 0; rec < results.length; rec++)
			for (int cat = 0; cat < results[rec].length; cat++)
				results[rec][cat] = results[rec][cat] / totals[rec];

		// The fitness
		double fitness;

		// Base fitness on basic scores
		if (results_interpereter == null)
		{
			// Calculate the fitnesses for each recording in test_feature_values
			double fitnesses[] = new double[results.length];
			for (int rec = 0; rec < results.length; rec++)
			{
				fitnesses[rec] = 0.0;
				for (int i = 0; i < correct_classifications[rec].length; i++)
					fitnesses[rec] += results[rec][ correct_classifications[rec][i] ];
			}

			// Calculate the average fitness for all recordings
			fitness = bodhidharma.utilities.Statistics.getAverage(fitnesses);
		}

		// Base fitness on classifications
		else
		{
			int number_recordings = correct_classifications.length;
			int number_categories = classifier.getNumberCategories();
			// 	int number_categories = correct_classifications.length; // erronious line from version 1.1.1

			// Calculate the model classifications
			boolean[][] model_winners = new boolean[number_recordings][number_categories];
			for (int rec = 0; rec < number_recordings; rec++)
				for (int cat = 0; cat < number_categories; cat++)
					model_winners[rec][cat] = false;
			for (int rec = 0; rec < number_recordings; rec++)
				for (int i = 0; i < correct_classifications[rec].length; i++)
					model_winners[rec][ correct_classifications[rec][i] ] = true;

			// Find the winners based on the given bit_string and classifier results
			boolean[][] winners = results_interpereter.getWinners(results);

			// Find the correct fraction of categories found per recording
			int[] corrects = new int[number_recordings];
			for (int rec = 0; rec < corrects.length; rec++)
				corrects[rec] = correct_classifications[rec].length;
			double[] percentage_right_per_rec = new double[number_recordings];
			for (int rec = 0; rec < corrects.length; rec++)
			{
				int found_so_far = 0;
				for (int cat = 0; cat < winners[rec].length; cat++)
				{
					if (winners[rec][cat] && model_winners[rec][cat])
						found_so_far++;
				}
				percentage_right_per_rec[rec] = ((double) found_so_far) / ((double) corrects[rec]);
			}

			// Set the fitness to the fraction correct
			fitness = bodhidharma.utilities.Statistics.getAverage(percentage_right_per_rec);
		}		

		// Return the fitness
		return fitness;
	}


	/**
	 * Returns an array of Booleans holding the meaning of the given bit string
	 * of a <code>Chromosome</code>. Values of 1 in the given <i>bit_string</i>
	 * are set to true and values of 0 are set to false. There is one element
	 * in the returned array for each bit in the <i>bit_string</i>. Bit strings
	 * passed to this method should be of the same length as the length specified
	 * in the constructor of this object.
	 *
	 * <p>Throws an exception if the length of the <i>bit_string</i> is not equal
	 * to the length of the bit_string_length> field.
	 */
	public Object[] parseBitString(short[] bit_string)
		throws Exception
	{
		// Throw an exception if the length of the bit_string is not equal to the
		// length of the bit_string_length> field
		if (bit_string.length != bit_string_length)
			throw new Exception("The given bit string has " + bit_string.length +
			                    " bits, but it should have " + bit_string_length + ".");

		Boolean[] features_to_use = new Boolean[bit_string.length];
		for (int bit = 0; bit < features_to_use.length; bit++)
		{
			if (bit_string[bit] == 0)
				features_to_use[bit] = new Boolean(false);
			else
				features_to_use[bit] = new Boolean(true);
		}
		return features_to_use;
	}
}