/*
 * @(#)FeatureWeightingEvaluator	1.1.2 09/14/04
 *
 * McGill University
 */

package bodhidharma.classifiers.genetic_algorithms;

import bodhidharma.classifiers.ClassificationResultsInterpereter;
import bodhidharma.classifiers.bio_k_nearest_neighbour.*;

/**
 * Objects of this class evaluate and interpret the meaning of bit strings stored in
 * genetic algorithm <code>Chromosome</code> objects. Each bit in the bit string
 * represents part of the representation of a double. Each double represents the
 * weighting for a feature. This <code>Evaluator</code> is used for finding a good
 * feature weighting.
 *
 * The fitness value is calculated by performing a classification using a
 * <code>KNN</code> classifier.
 *
 * @see	KNN
 * @see	Chromosome
 * @see Breeder
 */
public class FeatureWeightingEvaluator
	extends Evaluator
{
	/* FIELDS ******************************************************************/


	// The trained KNN classifier used to evaluate the fitness of DNA
	private	KNN	classifier;

	// Test feature values that will be classified to test performance. First indice
	// indicates recording and second indicates feature.
	private	double[][]	test_feature_values;

	// Model classifications of feature_values. First indice indicates recording.
	// Entries indicate the indices of the categories that are correct classifications.
	private	int[][]	correct_classifications;

	// Number of bits used to represent each weighting. More bits means a greater
	// variety of possible values.
	private	int		bits_per_weight;

	// Used to perform actual classifications so that fitnesses can be performed. Default fitness will be used if null.
	private	ClassificationResultsInterpereter results_interpereter;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. The <i>knn_classifier</i> must be trained with training points,
	 * but its <i>feature_weights</i> parameter should be untrained (i.e. null). The order
	 * of the features and categories passed in the <i>feature_values</i> and
	 * <i>classifications</i> parameters must be consistent with those that were used to
	 * train the <i>knn_classifier</i> externally.
	 *
	 * <p>Throws an exception if the <i>feature_values</i> or <i>classifications</i>
	 * parameters are of improper sizes.
	 *
	 * @param	knn_classifier		Trained KNN classifier used to evaluate the fitness of DNA.
	 * @param	feature_values		Test feature values that will be classified to test performance. First indice indicates recording and second indicates feature.
	 * @param	classifications		Model classifications of feature_values. First indice indicates recording. Entries indicate the indices of the categories that are correct classifications.
	 * @param	bits_per_value		Number of bits used to represent each weighting. More bits means a greater variety of possible values.
	 * @param	results_interpereter	Used to perform actual classifications so that fitnesses can be performed. Default fitness will be used if null.
	 */
	public FeatureWeightingEvaluator(KNN knn_classifier,
	                                 double[][] feature_values,
	                                 int[][] classifications,
	                                 int bits_per_value,
	                                 ClassificationResultsInterpereter results_interpereter )
		throws Exception
	{
		bit_string_length = feature_values[0].length * bits_per_value;
		classifier = knn_classifier;
		test_feature_values = feature_values;
		correct_classifications = classifications;
		bits_per_weight = bits_per_value;

		if (feature_values.length != classifications.length)
			throw new Exception("The given feature values and their classifications have " +
			                    "different sizes.");
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Return the fitness of a chromosome given its DNA (bit string). The fitness
	 * should vary between 0 (least fit possible) and 1 (most fit possible). This
	 * need not be normalized. Bit strings passed to this method must be
	 * of the same length as the length specified in the constructor of this
	 * object.
	 *
	 * <p>The fitness is calculated for each recording, and then averaged over all
	 * recordings. 
	 *
	 * <p>There are two ways of calculating fitness. The first is used if the results_interpereter
	 * field is not null, and the second if it is null. The first way gives a fitness
	 * based on the fraction of correct classifications based on the given bit_string.
	 *
	 * <p>The alternative fitness for each recording is set equal to the sum of the
	 * normalized scores returned by the classifier for those categories to which
	 * the recording belongs.
	 *
	 * <p>If previous feature selection set a feature to not be used, its weighting
	 * is interepereted as 0, irregardless of what it actually is, and its value is
	 * changed to 0 in the bit string. This is done by <code>parseBitString</code> method.
	 *
	 * <p>Throws an exception if there is a problem with the classification or if
	 * the length of the <i>bit_string</i> is not equal to the length of the
	 * <i>bit_string_length</i> field.
	 */
	public double getFitness(short[] bit_string)
		throws Exception
	{
		// Throw an exception if the length of the bit_string is not equal to the
		// length of the bit_string_length field
		if (bit_string.length != bit_string_length)
			throw new Exception("The given bit string has " + bit_string.length +
			                    " bits, but it should have " + bit_string_length + ".");

		// Set the feature_weights field of classifier based on the value of the
		// passed bit_string
		Double[] interpereted_bit_string = (Double[]) parseBitString(bit_string);
		double[] feature_weightings = new double[interpereted_bit_string.length];
		for (int feat = 0; feat < feature_weightings.length; feat++)
			feature_weightings[feat] = interpereted_bit_string[feat].doubleValue();
		classifier.feature_weights = feature_weightings;

		// Classify the test_feature_values field
		double[][] results = new double[test_feature_values.length][];
		for (int rec = 0; rec < results.length; rec++)
			results[rec] = classifier.classifyTestSample(test_feature_values[rec]);

		// Normalize the results of each recording in order to get the fraction of the
		// training samples within k that fell within each category
		double[] totals = new double[results.length];
		for (int rec = 0; rec < results.length; rec++)
		{
			totals[rec] = 0.0;
			for (int cat = 0; cat < results[rec].length; cat++)
				totals[rec] += results[rec][cat];
		}
		for (int rec = 0; rec < results.length; rec++)
			for (int cat = 0; cat < results[rec].length; cat++)
				results[rec][cat] = results[rec][cat] / totals[rec];

		// The fitness
		double fitness;

		// Base fitness on basic scores
		if (results_interpereter == null)
		{
			// Calculate the fitnesses for each recording in test_feature_values
			double fitnesses[] = new double[results.length];
			for (int rec = 0; rec < results.length; rec++)
			{
				fitnesses[rec] = 0.0;
				for (int i = 0; i < correct_classifications[rec].length; i++)
					fitnesses[rec] += results[rec][ correct_classifications[rec][i] ];
			}
			fitness = bodhidharma.utilities.Statistics.getAverage(fitnesses);
		}

		// Base fitness on classifications
		else
		{
			int number_recordings = correct_classifications.length;
			int number_categories = classifier.getNumberCategories();
			// 	int number_categories = correct_classifications.length; // erronious line from version 1.1.1
			
			// Calculate the model classifications
			boolean[][] model_winners = new boolean[number_recordings][number_categories];
			for (int rec = 0; rec < number_recordings; rec++)
				for (int cat = 0; cat < number_categories; cat++)
					model_winners[rec][cat] = false;
			for (int rec = 0; rec < number_recordings; rec++)
				for (int i = 0; i < correct_classifications[rec].length; i++)
					model_winners[rec][ correct_classifications[rec][i] ] = true;

			// Find the winners based on the given bit_string and classifier results
			boolean[][] winners = results_interpereter.getWinners(results);

			// Find the correct fraction of categories found per recording
			int[] corrects = new int[number_recordings];
			for (int rec = 0; rec < corrects.length; rec++)
				corrects[rec] = correct_classifications[rec].length;
			double[] percentage_right_per_rec = new double[number_recordings];
			for (int rec = 0; rec < corrects.length; rec++)
			{
				int found_so_far = 0;
				for (int cat = 0; cat < winners[rec].length; cat++)
				{
					if (winners[rec][cat] && model_winners[rec][cat])
						found_so_far++;
				}
				percentage_right_per_rec[rec] = ((double) found_so_far) / ((double) corrects[rec]);
			}

			// Set the fitness to the fraction correct
			fitness = bodhidharma.utilities.Statistics.getAverage(percentage_right_per_rec);
		}		

		// Calculate the average fitness for all recordings
		return fitness;
	}


	/**
	 * Returns an array of Doubles holding the meaning of the given bit string
	 * of a <code>Chromosome</code>. These values represent feature weightings
	 * and fall between 0 and 1. If previous feature selection set a feature to not
	 * be used, its weighting is interepereted as 0, irregardless of what it actually
	 * is, and its value is changed to 0 in the bit string.
	 *
	 * <p>Throws an exception if the length of the <i>bit_string</i> is not equal
	 * to the length of the <i>bit_string_length</i> field or if it is not evenly
	 * divisible by the word length.
	 */
	public Object[] parseBitString(short[] bit_string)
		throws Exception
	{
		// Throw an exception if the length of the bit_string is not equal to the
		// length of the bit_string_length> field
		if (bit_string.length != bit_string_length)
			throw new Exception("The given bit string has " + bit_string.length +
			                    " bits, but it should have " + bit_string_length + ".");

		// Throw an exception if the length of the bit_string is not evenly divisible
		// by the word length
		if (bit_string_length % bits_per_weight != 0)
			throw new Exception("The given bit string has is not evenly divisible by the " +
			                    "word length.");

		Double[] features_weightings = new Double[bit_string_length / bits_per_weight];
		for (int word = 0; word < features_weightings.length; word++)
		{
			// The starting indice of the word in bit_string
			int starting_indice = word * bits_per_weight;

			// Check if weighting should be set to 0 because of feature selection
			// and do so if necessary
			if (classifier.features_to_use != null)
				if (!classifier.features_to_use[word])
					for (int bit = 0; bit < bits_per_weight; bit++)
						bit_string[ starting_indice + bit ] = 0;

			// Construct each word representing a weighting
			short[] this_word = new short[bits_per_weight];
			for (int bit = 0; bit < bits_per_weight; bit++)
				this_word[bit] = bit_string[ starting_indice + bit ];

			// Find the value of the word
			features_weightings[word] = new Double( translateABitWord(this_word) );
		}

		return features_weightings;
	}

	
	/* PRIVATE METHODS *********************************************************/


	// Returns the value of a given segment of a bit string
	// Bit strings of length bits are converted to double form using the formula:
	//	value = ((2^0*string(0) + 2^1*string(1) + ... + 2^(bits-1)*string(bits-1)) * 200 / 2^bits ) - 100
	// This results values between 0 and 1
	// An exception is thrown if the bit string is of an inappropriate length
	private double translateABitWord(short[] bit_string_segment)
		throws Exception
	{
		// Make sure that bit_string_segment is of the appropriate length
		if (bit_string_segment.length != bits_per_weight)
			throw new Exception("Inappropriate bit string size.");

		// Calculate the value of the word
		double value = 0.0;
		for (int i = 0; i < bits_per_weight; i++)
			value += Math.pow(2,i) * bit_string_segment[i];
		value = value / ((double) (Math.pow(2,bits_per_weight) - 1));
		
		// Return the value of the word
		return value;
	}
}