/*
 * @(#)GeneticAlgorithmJFrame.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma.classifiers.genetic_algorithms;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import bodhidharma.PreferencesPanel;

/**
 * Dialog box used to hold and edit the basic parameters of new genetic algorithms.
 * 
 * <p><b>IMPORTANT:<b> These parameters are used when <b>new</b> genetic algorithm systems
 * are created. They do not affect existing genetic algorithm systems.
 * 
 * <p>The <i>Bits Per Word</i> field sets the number of bits used for each word in the
 * DNA bit string of <code>Chromosome</code> objects. Some implementations of genetic
 * algorithms ignore this field (in cases where the word size is fixed).
 * 
 * <p>The <i>Population Size</i> field sets the total number of chromosomes in one
 * generation in all villages put together.
 * 
 * <p>The <i>Mating Rate</i> field sets the fraction of the population that can serve
 * as parents during each generation. Must be between 0 and 1.
 * 
 * <p>The <i>Mutation Probability</i> field sets the probability that any bit in the
 * DNA bit string of <code>Chromosome</code> object will flip after crossover.  Must
 * be between 0 and 1.
 * 
 * <p>The <i>Elitism Allowed</i> field sets whether the best performing chromosome
 * is cloned to the next generation.
 * 
 * <p>The <i>Number of Villages</i> field sets the number of isolated villages in
 * which <code>Chromosome</code> objects evolve. There is interchange between villages
 * periodicallly (see bottom two parameters).
 * 
 * <p>The <i>Migration Delay</i> field sets the number of generations which pass between
 * chromosome interchanges between villages. This field has no effect if there is only
 * one village.
 * 
 * <p>The <i>Migration Size</i> field sets the number of chromosomes that are exchanged
 * between villages during interchanges. This field has no effect if there is only
 * one village.
 * 
 * <p>It should be noted that error messages are given if these parameters are incorrectly
 * set.
 *
 * <p>The various get methods of this class can be used by external objects to access
 * the stored parameters.
 *
 * <p>The <i>OK</i> button stores the entered settings. An error message is displayed
 * if there is a problem with the parameters.
 *
 * <p>The <i>Cancel</i> button cancels the entered settings and restores the previous
 * values.
 *
 * @see	BioKNearestNeighbour
 * @see Breeder
 * @see PreferencesPanel
 *
 * @author Cory McKay
 */
 public class GeneticAlgorithmJFrame
	extends JFrame
	implements ActionListener
{
	/* FIELDS ******************************************************************/

	// Window that instantiated this window
	private PreferencesPanel	parent_window;
	
	// Containers to hold Swing elements
	private Container		content_pane;
	private JPanel			settings_panel;
	private JPanel			button_panel;
	
	// Genetic algorithm parameters
	private int				word_length;
	private int				population_size;
	private double			mating_rate;
	private double			mutation_probability;
	private boolean			elitism;
	private int				number_villages;
	private int				migration_delay;
	private int				migration_size;

	// Genetic algorithm parameter editors
	private JTextArea		word_length_text_area;
	private JTextArea		population_size_text_area;
	private JTextArea		mating_rate_text_area;
	private JTextArea		mutation_probability_text_area;
	private JCheckBox		elitism_checkbox;
	private JTextArea		number_villages_text_area;
	private JTextArea		migration_delay_text_area;
	private JTextArea		migration_size_text_area;

	// Buttons
	private JButton			ok_button;
	private JButton			cancel_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. Configures the panel and its fields. Removes access to the
	 * parent <code>JFrame</code>.
	 *
 	 * @param	parent	Name of the <code>JFrame</code> window that spawned this dialog box.
	 */
	public GeneticAlgorithmJFrame(PreferencesPanel parent)
	{
		// Store parent recording
		parent_window = parent;

		// Configure overall window settings
		setTitle("Genetic Algorithm Initial Settings");
		int horizontal_gap = 6; // horizontal space between GUI elements
		int vertical_gap = 11; // horizontal space between GUI elements
		content_pane = getContentPane();
		content_pane.setLayout(new BorderLayout(horizontal_gap, vertical_gap));
		settings_panel = new JPanel(new GridLayout(8, 2, horizontal_gap, vertical_gap));
		button_panel = new JPanel(new FlowLayout());

		// Configure the settings graphical elements
		int number_columns = 7;
		word_length_text_area = new JTextArea("5", 1, number_columns);
		population_size_text_area = new  JTextArea("125", 1, number_columns);
		mating_rate_text_area = new  JTextArea("0.4", 1, number_columns);
		mutation_probability_text_area = new  JTextArea("0.02", 1, number_columns);
		elitism_checkbox = new JCheckBox();
		elitism_checkbox.setSelected(true);
		number_villages_text_area = new  JTextArea("1", 1, number_columns);
		migration_delay_text_area = new  JTextArea("1", 1, number_columns);
		migration_size_text_area = new  JTextArea("0", 1, number_columns);

		// Instantiate and configure buttons
		cancel_button = new JButton("Cancel");
		ok_button = new JButton("OK");
		cancel_button.addActionListener(this);
		ok_button.addActionListener(this);
		button_panel.add(cancel_button);
		button_panel.add(ok_button);

		// Add labels and previously created elements to panels and container
		settings_panel.add(new JLabel("Bits Per Word:"));
		settings_panel.add(word_length_text_area);
		settings_panel.add(new JLabel("Population Size:"));
		settings_panel.add(population_size_text_area);
		settings_panel.add(new JLabel("Mating Rate:"));
		settings_panel.add(mating_rate_text_area);
		settings_panel.add(new JLabel("Mutation Probability:"));
		settings_panel.add(mutation_probability_text_area);
		settings_panel.add(new JLabel("Elitism Allowed:"));
		settings_panel.add(elitism_checkbox);
		settings_panel.add(new JLabel("Number of Villages:"));
		settings_panel.add(number_villages_text_area);
		settings_panel.add(new JLabel("Migration Delay:"));
		settings_panel.add(migration_delay_text_area);
		settings_panel.add(new JLabel("Migration Size:"));
		settings_panel.add(migration_size_text_area);

		// Set up top display
		content_pane.add(settings_panel, BorderLayout.CENTER);
		content_pane.add(button_panel, BorderLayout.SOUTH);

		// Cause program to react when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				cancel();
			}
		});

		// Set genetic algorithm parameters equal to defaults and show an error
		// message if there is a problem
		try { readGraphicalSettings(); }
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			this.show();
			this.pack();
			parent_window.parent_window.setEnabled(false);
		}
	}


	/* PUBLIC METHODS **********************************************************/

	
	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the cancel_button
		if (event.getSource().equals(cancel_button))
			cancel();

		// React to the ok_button
		else if (event.getSource().equals(ok_button))
			ok();
	}

	
	/**
	 * Return the value of the word_length field
	 */
	public int getWordLength()
	{
		return word_length;
	}


	/**
	 * Return the value of the population_size field
	 */
	public int getPopulationSize()
	{
		return population_size;
	}


	/**
	 * Return the value of the mating_rate field
	 */
	public double getMatingRate()
	{
		return mating_rate;
	}


	/**
	 * Return the value of the mutation_probability field
	 */
	public double getMutationProbability()
	{
		return mutation_probability;
	}


	/**
	 * Return the value of the elitism field
	 */
	public boolean getElitism()
	{
		return elitism;
	}


	/**
	 * Return the value of the number_villages field
	 */
	public int getNumberVillages()
	{
		return number_villages;
	}


	/**
	 * Return the value of the migration_delay field
	 */
	public int getMigrationDelay()
	{
		return migration_delay;
	}


	/**
	 * Return the value of the migration_size field
	 */
	public int getMigrationSize()
	{
		return migration_size;
	}


	/**
	 * Set the value of the word_length and word_length fields
	 */
	public void setWordLength(String value)
		throws Exception
	{
		word_length_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the population_size and population_size_text_area fields
	 */
	public void setPopulationSize(String value)
		throws Exception
	{
		population_size_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the mating_rate and mating_rate_text_area fields
	 */
	public void setMatingRate(String value)
		throws Exception
	{
		mating_rate_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the mutation_probability and mutation_probability_text_area fields
	 */
	public void setMutationProbability(String value)
		throws Exception
	{
		mutation_probability_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the elitism and elitism_checkbox fields
	 */
	public void setElitism(String value)
		throws Exception
	{
		if (value.equals("true"))
			elitism_checkbox.setSelected(true);
		else
			elitism_checkbox.setSelected(false);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the number_villages and number_villages_text_area fields
	 */
	public void setNumberVillages(String value)
		throws Exception
	{
		number_villages_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the migration_delay and migration_delay_text_area fields
	 */
	public void setMigrationDelay(String value)
		throws Exception
	{
		migration_delay_text_area.setText(value);
		readGraphicalSettings();
	}


	/**
	 * Set the value of the migration_size and migration_size_text_area fields
	 */
	public void setMigrationSize(String value)
		throws Exception
	{
		migration_size_text_area.setText(value);
		readGraphicalSettings();
	}


	/* PRIVATE METHODS *********************************************************/


	// Read settings in GUI elements into genetic algorithm fields.
	// Throws an exception if improper parameters are entered.
	private void readGraphicalSettings()
		throws Exception
	{
		// Check for exceptions
		int village_size = Integer.parseInt(population_size_text_area.getText()) / Integer.parseInt(number_villages_text_area.getText());
		if (Integer.parseInt(word_length_text_area.getText()) < 1)
			throw new Exception("You have entered a negative value for the Bits Per Word setting.\n" +
								"Please adjust this parameter.");
		else if (Integer.parseInt(population_size_text_area.getText()) < 2)
			throw new Exception("Population Size must be greater than 1.\n" +
								"Please adjust this parameter.");
		else if (Double.parseDouble(mating_rate_text_area.getText()) > 1.0 ||
		         Double.parseDouble(mating_rate_text_area.getText()) < 0.0)
			throw new Exception("Mating Rate must be greater than 0 and less than 1.\n" +
								"Please adjust this parameter.");
		else if (Double.parseDouble(mutation_probability_text_area.getText()) > 1.0 ||
		         Double.parseDouble(mutation_probability_text_area.getText()) < 0.0)
			throw new Exception("Mutation Probability must be greater than 0 and less than 1.\n" +
								"Please adjust this parameter.");
		else if (Integer.parseInt(number_villages_text_area.getText()) < 1)
			throw new Exception("Number of Villages must be greater than 0.\n" +
								"Please adjust this parameter.");
		else if (Integer.parseInt(migration_delay_text_area.getText()) < 1)
			throw new Exception("Migration Delay must be greater than 0.\n" +
								"Please adjust this parameter.");
		else if (Integer.parseInt(migration_size_text_area.getText()) < 0)
			throw new Exception("Migration Size must be at least 0.\n" +
								"Please adjust this parameter.");
		else if (Integer.parseInt(population_size_text_area.getText()) %
		         Integer.parseInt(number_villages_text_area.getText()) != 0)
			throw new Exception("The Population Size must be a multiple of the Number of " +
								"Villages in order to ensure a uniform village size.\n" +
								"Please adjust the parameters.");
		else if ( (Integer.parseInt(population_size_text_area.getText()) /
		           Integer.parseInt(number_villages_text_area.getText())) <=
			       Integer.parseInt(migration_size_text_area.getText()) )
			throw new Exception("Migration Size must be smaller than village size. You currently " +
								"have a village size of " + village_size + 
								" and a Migration Size of " + Integer.parseInt(migration_size_text_area.getText()) + ".\n" +
								"Please adjust the Population Size, Number of Villages and/or Migration Size " +
								"so that this requirement is fulfilled.");
		else if ( ((int) (Double.parseDouble(mating_rate_text_area.getText()) * village_size)) < 2)
			throw new Exception("There must be at least two elligible parents per village.\n" +
							    "Your current settings for Population Size, village size and Mating Rate " + 
								"result in " + ((int) (Double.parseDouble(mating_rate_text_area.getText()) * village_size)) +
								" parents being selected per village.\nPlease adjust these parameters.");
		
		// Set the genetic algorithm parameters
		word_length = Integer.parseInt(word_length_text_area.getText());
		population_size = Integer.parseInt(population_size_text_area.getText());
		mating_rate = Double.parseDouble(mating_rate_text_area.getText());
		mutation_probability = Double.parseDouble(mutation_probability_text_area.getText());
		elitism = elitism_checkbox.isSelected();
		number_villages = Integer.parseInt(number_villages_text_area.getText());
		migration_delay = Integer.parseInt(migration_delay_text_area.getText());
		migration_size = Integer.parseInt(migration_size_text_area.getText());
	}


	// Close the dialog box without registering any changes in the
	// genetic algorithm fields. Reset graphical elements to hold
	// contents of genetic algorithm fields.
	private void cancel()
	{
		restoreGraphicalValuesToStoredValues();
		parent_window.parent_window.setEnabled(true);
		this.hide();
	}


	// Close the dialog box and update the genetic algorithm fields
	// to reflect any changes. Displays an error message if there
	// are any problems
	private void ok()
	{
		// Update the genetic algorithm parameter fields and display
		// an error message if necessary
		try 
		{
			readGraphicalSettings(); 
		
			// Reactivate the parent window and hide this dialog box
			parent_window.parent_window.setEnabled(true);
			this.hide();
		}
		catch (Exception e) { JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE); }
	}


	// Reset graphical elements to hold contents of genetic algorithm fields.
	private void restoreGraphicalValuesToStoredValues()
	{
		word_length_text_area.setText(Integer.toString(word_length));
		population_size_text_area.setText(Integer.toString(population_size));
		mating_rate_text_area.setText(Double.toString(mating_rate));
		mutation_probability_text_area.setText(Double.toString(mutation_probability));
		elitism_checkbox.setSelected(elitism);
		number_villages_text_area.setText(Integer.toString(number_villages));
		migration_delay_text_area.setText(Integer.toString(migration_delay));
		migration_size_text_area.setText(Integer.toString(migration_size));
	}
}