/*
 * @(#)AverageNumberOfIndependentVoicesFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma. midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class AverageNumberOfIndependentVoicesFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public AverageNumberOfIndependentVoicesFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// Instantiate of the variable holding the number of voices sounding at each tick
			int[] number_sounding = new int[recording_info.channel_tick_map.length];
			for (int i = 0; i < number_sounding.length; i++)
				number_sounding[i] = 0;

			// Find the number of voices sounding at each tick
			int rest_count = 0;
			for (int tick = 0; tick < recording_info.channel_tick_map.length; tick++)
			{
				for (int chan = 0; chan < recording_info.channel_tick_map[tick].length; chan++)
					if (recording_info.channel_tick_map[tick][chan])
						number_sounding[tick]++;

				// Keep track of number of ticks with no notes sounding
				if (number_sounding[tick] == 0 )
					rest_count++;
			}

			// Only count the ticks where at least one note was sounding
			double[] final_number_sounding = new double[number_sounding.length - rest_count];
			int count = 0;
			for (int i = 0; i < number_sounding.length; i++)
				if (number_sounding[i] > 0.5)
				{
					final_number_sounding[count] = (double) number_sounding[i];
					count++;
				}

			// Calculate the average
			value = bodhidharma.utilities.Statistics.getAverage(final_number_sounding);
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Average Number of Independent Voices");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Average number of different channels in which notes have sounded\n" +
		                  "simultaneously. Rests are not included in this calculation.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("T-2");
	}
}