/*
 * @(#)AverageVariabilityOfTimeBetweenAttacksForEachVoiceFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma.midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class AverageVariabilityOfTimeBetweenAttacksForEachVoiceFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public AverageVariabilityOfTimeBetweenAttacksForEachVoiceFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			double[][] intervals = new double[16][];
			
			// Record the intervals
			for (int chan = 0; chan < 16; chan++)
			{
				// Mark with null if no notes contained in channel
				if (recording_info.channel_statistics[chan][0] == 0)
					intervals[chan] = null;

				// If notes are contained in channel
				else
				{
					// Find number of intervals
					int number_of_intervals = -1;
					for (int tick = 0; tick < recording_info.note_beginnings_map.length; tick++)
						if (recording_info.note_beginnings_map[tick][chan])
							number_of_intervals++;

					// Fill in the array of intervals
					intervals[chan] = new double[number_of_intervals];
					number_of_intervals = 0;
					double time_so_far = 0.0;
					int tick_of_last_attack  = -1;
					for (int tick = 0; tick < recording_info.note_beginnings_map.length; tick++)
					{
						// Check if an attack occured on this tick
						boolean attack = false;
						if (recording_info.note_beginnings_map[tick][chan])
							attack = true;
						
						if (!attack)
							time_so_far += recording_info.seconds_per_tick[tick];
						else
						{
							if (tick_of_last_attack != -1)
							{
								intervals[chan][number_of_intervals] = time_so_far;
								number_of_intervals++;
							}

							time_so_far = 0.0;
							tick_of_last_attack = tick;
						}
					}
				}
			}

			// Find the standard deviations
			int number_channels_with_notes = 16;
			for (int chan = 0; chan < 16; chan++)
				if (intervals[chan] == null)
					number_channels_with_notes--;
			double[] sds = new double[number_channels_with_notes];
			int count = 0;
			for (int chan = 0; chan < 16; chan++)
			{
				if (intervals[chan] != null)
				{
					sds[count] = bodhidharma.utilities.Statistics.getStandardDeviation(intervals[chan]);
					count++;
				}
			}

			// Calculate average of standard deviations
			value = bodhidharma.utilities.Statistics.getAverage(sds);
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Average Variability of Time Between Attacks For Each Voice");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Average standard deviation, in seconds, of time between Note On\n" +
		                  "events on individual channels that contain at least one note.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("R-25");
	}
}