/*
 * @(#)DominantSpreadFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma. midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class DominantSpreadFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public DominantSpreadFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// Check all streaks
			int max_count = 0;
			for (int bin = 0; bin < recording_info.fifths_pitch_histogram.length; bin++)
			{
				if (recording_info.fifths_pitch_histogram[bin] >= 0.09)
				{
					boolean done = false;
					int count = 1;
					int i = bin + 1;
					while (!done)
					{
						// Wrap around
						if (i == recording_info.fifths_pitch_histogram.length)
							i = 0;

						// If rereach starting point
						if (i == bin)
							done = true;

						// Increment if sufficient number
						else if (recording_info.fifths_pitch_histogram[i] >= 0.09)
						{
							count++;
							i++;
						}

						// End if not sufficient number
						else
							done = true;
					}

					// Record the largest streak
					if (count > max_count)
						max_count = count;
				}
			}

			// Calculate the value
			value = (double) max_count;
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Dominant Spread");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Largest number of consecutive pitch classes separated by perfect\n" +
		                  "5ths that accounted for at least 9% each of the notes.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("P-17");
	}
}