/*
 * @(#)PolyrhythmsFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma. midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class PolyrhythmsFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public PolyrhythmsFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// Find the number of sufficiently large peaks
			int count = 0;
			for (int bin = 0; bin < recording_info.rhythmic_histogram_table.length; bin++)
				if (recording_info.rhythmic_histogram_table[bin][2] > 0.001)
					count++;

			// Store the peak bins
			int[] peak_bins = new int[count];
			int so_far = 0;
			for (int bin = 0; bin < recording_info.rhythmic_histogram_table.length; bin++)
				if (recording_info.rhythmic_histogram_table[bin][2] > 0.001)
				{
					peak_bins[so_far] = bin;
					so_far++;
				}

			// Find the highest peak
			int highest_index = 0;
			double max_so_far = 0.0;
			for (int bin = 0; bin < peak_bins.length; bin++)
				if (recording_info.rhythmic_histogram_table[ peak_bins[bin] ][2] > max_so_far)
				{
					max_so_far = recording_info.rhythmic_histogram_table[ peak_bins[bin] ][2];
					highest_index = peak_bins[bin];
				}

			// Find the number of peak bins which are multiples or factors of the highest bin
			int hits = 0;
			for (int i = 0; i < peak_bins.length; i++)
			{
				int left_limit = peak_bins[i] - 3;
				if (left_limit < 0)
					left_limit = 0;

				int right_limit = peak_bins[i] + 4;
				if (right_limit > recording_info.rhythmic_histogram_table.length)
					right_limit = recording_info.rhythmic_histogram_table.length;

				int[] multipliers = {1, 2, 3, 4, 6, 8};
				for (int j = left_limit; j < right_limit; j++)
				{
					if (bodhidharma.utilities.Statistics.isFactorOrMultiple(j, highest_index, multipliers))
					{
						hits++;
						left_limit = right_limit + 1; // exit loop
					}
				}
			}
			
			// Calculate the value
			value = (double) hits / (double) peak_bins.length;
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Polyrhythms");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Number of beat peaks with frequencies at least 30% of the highest\n" +
		                  "frequency whose bin labels are not integer multiples or factors\n" +
		                  "(using only multipliers of 1, 2, 3, 4, 6 and 8) (with an accepted error\n" +
		                  "of +/- 3 bins) of the bin label of the peak with the highest frequency.\n" +
		                  "This number is then divided by the total number of beat bins with\n" +
		                  "frequencies over 30% of the highest frequency.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("R-12");
	}
}