/*
 * @(#)RelativeNoteDensityOfHighestLineFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma. midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class RelativeNoteDensityOfHighestLineFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public RelativeNoteDensityOfHighestLineFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// Find the channel with the highest average pitch
			int max_so_far = 0;
			int highest_chan = 0;
			for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
				if (recording_info.channel_statistics[chan][6] != 0 && chan != (10 - 1))
					if (recording_info.channel_statistics[chan][6] > max_so_far)
					{
						max_so_far = recording_info.channel_statistics[chan][6];
						highest_chan = chan;
					}

			// Find the number of channels with no note ons (or that is channel 10)
			int silent_count = 0;
			for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
				if (recording_info.channel_statistics[chan][0] == 0 || chan == (10 - 1))
					silent_count++;

			// Find the average number of notes in each channel
			double[] number_of_notes = new double[recording_info.channel_statistics.length - silent_count];
			int count = 0;
			for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
				if (recording_info.channel_statistics[chan][0] != 0 && chan != (10 - 1))
				{
					number_of_notes[count] = (double) recording_info.channel_statistics[chan][0];
					count++;
				}
			double total_average = bodhidharma.utilities.Statistics.getAverage(number_of_notes);
			
			// Set value
			value = ((double) max_so_far) / ((double) total_average);
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Relative Note Density of Highest Line");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Number of Note Ons in the channel with the highest average pitch\n" +
		                  "divided by the average number of Note Ons in all channels that\n" +
		                  "contain at least one note.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("T-13");
	}
}