/*
 * @(#)RhythmicLoosenessFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma. midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class RhythmicLoosenessFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public RhythmicLoosenessFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// Find the number of sufficiently large peaks
			int count = 0;
			for (int bin = 0; bin < recording_info.rhythmic_histogram_table.length; bin++)
				if (recording_info.rhythmic_histogram_table[bin][2] > 0.001)
					count++;

			// Store the peak bins
			int[] peak_bins = new int[count];
			int so_far = 0;
			for (int bin = 0; bin < recording_info.rhythmic_histogram_table.length; bin++)
				if (recording_info.rhythmic_histogram_table[bin][2] > 0.001)
				{
					peak_bins[so_far] = bin;
					so_far++;
				}

			// Find the width of each peak
			double[] widths = new double[peak_bins.length];
			for (int peak = 0; peak < peak_bins.length; peak++)
			{
				// 30% of this peak
				double limit_value = 0.3 * recording_info.rhythmic_histogram[ peak_bins[peak] ];

				// Find left limit
				int i = peak_bins[peak];
				int left_index = 0;
				while (i >= 0 )
				{
					if (recording_info.rhythmic_histogram[i] < limit_value)
						i = -1;
					else
					{
						left_index = i;
						i--;
					}
				}

				// Find right limit
				i = peak_bins[peak];
				int right_index = 0;
				while (i < recording_info.rhythmic_histogram.length)
				{
					if (recording_info.rhythmic_histogram[i] < limit_value)
						i = recording_info.rhythmic_histogram.length + 1;
					else
					{
						right_index = i;
						i++;
					}
				}

				// Calculate width
				widths[peak] = (double) right_index - (double) left_index;
			}

			// Calculate the value
			value = bodhidharma.utilities.Statistics.getAverage(widths);
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Rhythmic Looseness");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Average width of beat histogram peaks (in beats per minute).\n" +
		                  "Width is measured for all peaks with frequencies at least 30%\n" +
		                  "as high as the highest peak, and is defined by the distance between the\n" +
		                  "points on the peak in question that are 30% of the height of the peak.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("R-11");
	}
}