/*
 * @(#)VariabilityOfNotePrevalenceOfPitchedInstrumentsFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma.midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class VariabilityOfNotePrevalenceOfPitchedInstrumentsFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public VariabilityOfNotePrevalenceOfPitchedInstrumentsFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// Find the number of pitched instruments used to play at least one note
			int instruments_present = 0;
			for (int instrument = 0; instrument < recording_info.pitched_instrumentation_frequencies.length; instrument++)
				if (recording_info.pitched_instrumentation_frequencies[instrument][0] != 0)
					instruments_present++;

			// Calculate the feature value
			double[] instrument_frequencies = new double[instruments_present];
			int count = 0;
			for (int instrument = 0; instrument < recording_info.pitched_instrumentation_frequencies.length; instrument++)
				if (recording_info.pitched_instrumentation_frequencies[instrument][0] != 0)
				{
					instrument_frequencies[count] = (double) recording_info.pitched_instrumentation_frequencies[instrument][0];
					count++;
				}
			value = bodhidharma.utilities.Statistics.getStandardDeviation(instrument_frequencies);
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Variability of Note Prevalence of Pitched Instruments");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Standard deviation of the fraction of Note Ons played by\n" +
		                  "each (pitched) General MIDI instrument that is used to play\n" +
		                  "at least one note");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("I-6");
	}
}