/*
 * @(#)VibratoPrevalenceFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma.midi_parsing.*;
import java.util.LinkedList;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class VibratoPrevalenceFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public VibratoPrevalenceFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// If there are no pitch bends
			if (recording_info.pitch_bends_list.size() == 0)
				value = 0.0;

			else
			{
				// Generate array of pitch bends
				Object[] notes_objects = recording_info.pitch_bends_list.toArray();
				LinkedList[] notes = new LinkedList[notes_objects.length];
				for (int i = 0; i < notes.length; i++)
					notes[i] = (LinkedList) notes_objects[i];
				int[][] pitch_bends = new int[notes.length][];
				for (int i = 0; i < notes.length; i++)
				{
					Object[] this_note_pitch_bends_objects = notes[i].toArray();
					pitch_bends[i] = new int[this_note_pitch_bends_objects.length];
					for (int j = 0; j < pitch_bends[i].length; j++)
						pitch_bends[i][j] = ((Integer) this_note_pitch_bends_objects[j]).intValue();
				}

				// Find the number of changes of direction of bend bend for each note
				int notes_with_vibrato = 0;;
				for (int note = 0; note < pitch_bends.length; note++)
				{
					int changes = 0;
					int last_value = pitch_bends[note][0];
					int direction = 0;
					for (int bend = 0; bend < pitch_bends[note].length; bend++)
					{
						if (pitch_bends[note][bend] > last_value)
						{
							if (direction == -1)
								changes++;
							direction = 1;
						}
						else if (pitch_bends[note][bend] < last_value)
						{
							if (direction == 1)
								changes++;
							direction = -1;
						}
					}

					if (changes > 2)
						notes_with_vibrato++;
				}
	
	

				// Calculate the value
				value = (double) notes_with_vibrato / (double) pitch_bends.length;
			}
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Vibrato Prevalence");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Number of notes for which Pitch Bend messages change direction\n" +
		                  "at least twice divided by total number of notes that have Pitch\n" +
		                  "Bend messages associated with them.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("P-25");
	}
}