/*
 * @(#)VoiceSeparationFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma. midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class VoiceSeparationFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public VoiceSeparationFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// Find the number of channels with no note ons (or that is channel 10)
			int silent_count = 0;
			for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
				if (recording_info.channel_statistics[chan][0] == 0 || chan == (10 - 1))
					silent_count++;

			// If there's only one voice
			if (silent_count > 14)
			{
				value = 0.0;
			}
			else
			{
				// Store the average melodic interval of notes in the other channels
				double[] average_pitches = new double[recording_info.channel_statistics.length - silent_count];
				int count = 0;
				for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
					if (recording_info.channel_statistics[chan][0] != 0 && chan != (10 - 1))
					{
						average_pitches[count] = (double) recording_info.channel_statistics[chan][6];
						count++;
					}

				// Sort the average pitches
				average_pitches = bodhidharma.utilities.Sorter.sortDoubleArray(average_pitches);

				// Find the intervals
				double[] intervals = new double[average_pitches.length - 1];
				for (int i = 0; i < average_pitches.length - 1; i++)
					intervals[i] = average_pitches[i+1] - average_pitches[i];

				// Find the average interval
				value = bodhidharma.utilities.Statistics.getAverage(intervals);
			}
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Voice Separation");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Average separation in semi-tones between the average pitches of consecutive\n" +
		                  "channels (after sorting based on average pitch) that contain at least one note.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("T-20");
	}
}