/*
 * @(#)ProgressBarDialog.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.utilities;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * A general purpose window that displays the progress of a long process as
 * it occurs.
 *
 * <p>The user may cancel processing, either during or after processing, by
 * pressing on the <i>Cancel</i> button. The results of processing are not actually implemented
 * implemented until the user presses on the <i>Accept</i> button, which cannot
 * be done until processing is finished. The effects of pressing the <i>Accept</i>
 * button are controlled by the <code>acceptProcessingResults</i> method of the
 * <i>results_storer</i> parameter of the constructor.
 *
 * <p>The actual processing is done by the <code>ProgressBarTaskMonitor</code>
 * object that is passed to the constructor. This object is also provides the
 * information needed to keep this window updated.
 *
 * <p>The parent <coed>JFrame</code> which instantiated this object is made
 * inaccessible to the user during processing in order to avoid concurrency problems.
 *
 * <p>This window includes a progress bar as well as a listing of the time
 * elapsed as well as the estimated processing time remaining. There is also
 * a Progress area that displays any progress messages that are generated
 * during processing. There is also an Errors Encountered section that is
 * used to report any non-fatal errors as they occur.
 *
 * @author	Cory McKay
 * @see		ProgressBarTaskMonitor
 * @see		Timer
 * @see		JProgressBar
 */
 public class ProgressBarDialog
	extends JFrame
	implements ActionListener
{
	/* FIELDS ******************************************************************/


	// Performs and monitors task that is to be monitored by this JFrame
    private	ProgressBarTaskMonitor	task_monitor;

	// Timer that queries task_monitor to keep this JFrame up to date
    private	Timer					timer;

	// Place where results of succesful processing can be stored
	private	ProcessResultsReceiver	results_receiver;

	// Window that is involved in instantiation of this window
	private JFrame					parent_window;
	
	// Components to display progress
	private	JProgressBar			progress_bar;
	private JTextArea				time_report_text_area;
	private	JTextArea				progress_report_text_area;
	private	JTextArea				error_report_text_area;

	// Containers to hold Swing elements
	private Container				content_pane;
	private JPanel					progress_bar_panel;
	private JPanel					text_panel;
	private JPanel					progress_report_panel;
	private JPanel					error_report_panel;
	private JPanel					button_panel;
	private JScrollPane				progress_report_scroll_pane;
	private JScrollPane				error_report_scroll_pane;

	// Buttons
	private JButton					cancel_button;
	private	JButton					done_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * A constructor that configures and displays the window, triggers processing
	 * by the given <code>ProgressBarTaskMonitor</code> and keeps the window updated
	 * during processing.
	 *
	 * <p>The <i>parent</i> parameter refers to the <code>JFrame</code> that instantiated
	 * this object and which needs to be made non-accessible. The <i>task_checker_and_performer</i>
	 * parameter refers to the <code>ProgressBarTaskMonitor</code> which performs the
	 * actual task to be performed and which is queried to keep this window updated.
	 * The <i>number_tasks</i> parameter gives the number of sub-tasks that need to be
	 * performed in the overall processing. This is used in order to calculate how
	 * much work has been done proportionately. It is assumed that all sub-tasks will
	 * take an equal amount of time which, while incorrect in most cases, is nonetheless
	 * close enough for most applications.
	 *
  	 * @param	parent	Name of the <code>JFrame</code> window that instantiated this window.
  	 * @param	task_checker_and_performer	Performs processing and provides information needed for updates to this window.	
   	 * @param	number_tasks	Number of sub-tasks needing to be performed for the overall process to be completed.	
  	 * @param	time_between_polls			Time in milliseconds between polling of task_checker_and_performer.	
	 * @see		ProgressBarTaskMonitor
	 * @see		Timer
	 * @see		JProgressBar
	 */
	public ProgressBarDialog(JFrame parent, ProcessResultsReceiver results_storer, ProgressBarTaskMonitor task_checker_and_performer, int number_tasks, double time_between_polls)
	{
		// Disable the parent window
		parent_window = parent;
		parent_window.setEnabled(false);

		// Configure overall window settings and instantiate related panels
		results_receiver = results_storer;
		setTitle("Progress");
		int horizontal_gap = 6; // horizontal space between GUI elements
		int vertical_gap = 4; // horizontal space between GUI elements
		content_pane = getContentPane();
		content_pane.setLayout(new BorderLayout(horizontal_gap, vertical_gap));
		progress_bar_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		text_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		progress_report_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		error_report_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		button_panel = new JPanel();

		// Configure progress bar		
		progress_bar = new JProgressBar(0, number_tasks);
		progress_bar.setValue(0);
		progress_bar.setStringPainted(true);

		// Configure text areas
		int number_text_columns = 50;
		time_report_text_area = new JTextArea(1, number_text_columns);
		time_report_text_area.setEditable(false);
		time_report_text_area.setLineWrap(false);
		time_report_text_area.setCursor(null); //inherit the panel's cursor (see Java bug 4851758)
		progress_report_text_area = new JTextArea(10, number_text_columns);
		progress_report_text_area.setEditable(false);
		progress_report_text_area.setLineWrap(true);
		progress_report_text_area.setWrapStyleWord(true);
		progress_report_text_area.setCursor(null); //inherit the panel's cursor (see Java bug 4851758)
		error_report_text_area = new JTextArea(6, number_text_columns);
		error_report_text_area.setEditable(false);
		error_report_text_area.setLineWrap(true);
		error_report_text_area.setWrapStyleWord(true);
		error_report_text_area.setCursor(null); //inherit the panel's cursor (see Java bug 4851758)

		// Configure buttons
		cancel_button = new JButton("Cancel");
		cancel_button.addActionListener(this);
		done_button = new JButton("Accept");
		done_button.addActionListener(this);

		// Add labels and previously created elements to panels and container
		progress_report_scroll_pane = new JScrollPane(progress_report_text_area);
		error_report_scroll_pane = new JScrollPane(error_report_text_area);
		progress_bar_panel.add(progress_bar, BorderLayout.NORTH);
		progress_bar_panel.add(time_report_text_area, BorderLayout.SOUTH);
		progress_report_panel.add(new JLabel("Progress:"), BorderLayout.NORTH);
		progress_report_panel.add(progress_report_scroll_pane, BorderLayout.SOUTH);
		error_report_panel.add(new JLabel("Errors Encountered:"), BorderLayout.NORTH);
		error_report_panel.add(error_report_scroll_pane, BorderLayout.SOUTH);
		text_panel.add(progress_report_panel, BorderLayout.NORTH);
		text_panel.add(error_report_panel, BorderLayout.SOUTH);
		button_panel.add(cancel_button);
		button_panel.add(done_button);
		content_pane.add(progress_bar_panel, BorderLayout.NORTH);
		content_pane.add(text_panel, BorderLayout.CENTER);
		content_pane.add(button_panel, BorderLayout.SOUTH);

		// Cause program to react when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				cancel();
			}
		});

		// Set up task_monitor
		task_monitor = task_checker_and_performer;

		// Create a timer to check progress every time_between_polls and update
		// this window accordingly. Also takes appropriate actions when
		// processing is finished.
		timer = new Timer((new Double(time_between_polls)).intValue(), new ActionListener()
		{
			public void actionPerformed(ActionEvent evt)
			{
				progress_bar.setValue(task_monitor.getCurrentAmountCompleted());
				time_report_text_area.setText(task_monitor.getTimeMessage());
				time_report_text_area.setCaretPosition(time_report_text_area.getDocument().getLength());
				String progress_message = task_monitor.getProgressMessage();
				if (progress_message != null)
				{
					progress_report_text_area.append(progress_message);
					// progress_report_text_area.setCaretPosition(progress_report_text_area.getDocument().getLength());
				}
				String error_message = task_monitor.getErrorMessage();
				if (error_message != null)
				{
					error_report_text_area.append(error_message);
					// error_report_text_area.setCaretPosition(progress_report_text_area.getDocument().getLength());
				}
				if (task_monitor.isDone())
				{
					Toolkit.getDefaultToolkit().beep(); // beep when done
					timer.stop(); // stop the timer
					progress_report_text_area.append("TASK COMPLETE"); // note that finishted
					setCursor(null); //turn off the wait cursor
					done_button.setEnabled(true); // make it possible to accept results of task
				}
			}
		});

		// Begin the task and monitor it
		done_button.setEnabled(false);
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		task_monitor.go();
		timer.start();

		// Dispay this window
		pack();
		show();
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the cancel_button
		if (event.getSource().equals(cancel_button))
			cancel();

		// React to the done_button
		else if (event.getSource().equals(done_button))
			done();
	}


	/* PRIVATE METHODS *********************************************************/

	
	// React to the close button or the cancel button.
	// Stop the process, reenable the parent_window and hide this window without
	// taking any actions to enact results of processing.
	private void cancel()
	{
		timer.stop();
		task_monitor.stop();
		parent_window.setEnabled(true);
		this.hide();		
	}


	/**
	 * React to the accept button.
	 * Reenable the parent_window and hide this window after causing the results
	 * of processing to be enacted.
	 */
	public void done()
	{
		results_receiver.acceptProcessingResults();
		parent_window.setEnabled(true);
		this.hide();		
	}
}