/*
 * @(#)ProgressBarTaskMonitor.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.utilities;

/**
 * An abstract class that is used by <code>ProgressBarDialog</code> windows
 * in order to perform the processing that is measured by the window as well
 * as keep the window updated.
 *
 * <p>All parts of this class are implemented here except for the <code>go</code>
 * method. This method shoulud use a <code>SwingWorker</code> object to
 * start and perform the task to be performed. It should also keep the
 * <i>current_amount_completed</i>,  <i>time_started</i>, <i>task_completed</i>,
 * <i>task_completed</i>, <i>task_canceled</i>, <i>progress_message</i> and
 * <i>error_message</i> fields updated so that they can be queried by the calling
 * <code>ProgressBarDialog</code>.
 *
 * <p>The part of the class implemented here essentially provides an interface
 * for the calling <code>ProgressBarDialog</code> and calculates the time elapsed
 * and the estimated processing time remaining. The get and is methods can be called to
 * get updates on how the process is progressing. The <code>stop</code> method
 * can be called in order to stop processing if the process is to be cancelled.
 *
 * @author	Cory McKay
 * @see		ProgressBarDialog
 * @see		SwingWorker
 */
public abstract class ProgressBarTaskMonitor
{
	/* FIELDS ******************************************************************/


	// The number of sub-tasks to be completed before entire task is done
	protected	int				length_of_task;

	// The current number of sub-tasks completed
	protected	int				current_amount_completed;
	
	// The time when the overall task began
	protected	long			time_started;

	// Whether or not the process has been completed or cancelled
	protected	boolean			task_completed;
	protected	boolean			task_canceled;
	
	// Messages to be displayed on the ProgressBarDialog
	protected	String			progress_message;
	protected	String			error_message;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor that sets the overall task to complete as well as the
	 * total number of sub-tasks to be completed in the overall task. Also
	 * sets other private fields to defaluts.
	 *
	 * <p>NOTE: All sub-tasks are assumed to be of equal length in terms of processing
	 * duration.
	 *
	 * @param	number_of_sub_tasks	Number of sub-tasks that will need to be completed in all.
	 */
	public ProgressBarTaskMonitor(int number_of_sub_tasks)
	{
		length_of_task = number_of_sub_tasks;

		current_amount_completed = 0;
		task_completed = false;
		task_canceled = false;
		progress_message = null;
		error_message = null;
    }


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Called from ProgressBarDialog to start the overall task. Sets up a SwingWorker
	 * object to carry out the thread and sets time_started. Keeps the appropriate
	 * fields updated (see above).
	 */
	public abstract void go();


	/**
	 * Called from a ProgressBarDialog to find out how many sub-tasks remain
	 * to be done.
	 */
	public int getLengthOfTask()
	{
		return length_of_task;
	}


	/**
	 * Called from a ProgressBarDialog to find out how many sub-tasks have been done.
	 */
	public int getCurrentAmountCompleted()
	{
		return current_amount_completed;
	}


	/**
	 * Called from a ProgressBarDialog to find out if the overall task has completed.
	 */
	public boolean isDone()
	{
		return task_completed;
	}


	/**
	 * Called from a ProgressBarDialog to find out if the overall task has been cancelled.
	 */
	public boolean isCancelled()
	{
		return task_canceled;
	}


	/**
	 * Returns a <code>String</code> indicating how much time has already passed and the
	 * estimated time remaining.
	 */
	public String getTimeMessage()
	{
		// Return empty if nothing ready yet
		if (length_of_task == 0 || current_amount_completed == 0)
			return new String("");

		// Calculate the estimated time remaining before task is completed
		double fraction_completed = ((double) current_amount_completed) / ((double) length_of_task);
		double time_already_passed_sec = (double) ((System.currentTimeMillis() - time_started) / 1000);
		double time_total_sec = time_already_passed_sec / fraction_completed;
		double time_remaining_sec = time_total_sec - time_already_passed_sec;
		int time_already_passed = (int) time_already_passed_sec;
		int time_remaining = (int) time_remaining_sec;
		if (time_remaining < 0)
			time_remaining = 0;

		// Devise the string to return
		String time_passed_message = findTimeWithProperUnits(time_already_passed);
		String time_remaining_message = findTimeWithProperUnits(time_remaining);
		return new String("Time elapsed: " + time_passed_message +
						  "        Estimated time remaining: " + time_remaining_message);
	}


	/**
	 * Returns the most recent progress message(s), or null if there is no 
	 * current progress message. Sets progress_message to null if it is
	 * not already.
	 */
	public String getProgressMessage()
	{
		if (progress_message != null)
		{
			String temp = progress_message;
			progress_message = null;
			return temp;
		}
		else
			return null;
	}


	/**
	 * Returns the most recent error message(s), or null if there is no 
	 * current error message. Sets error_message to null if it is
	 * not already.
	 */
	public String getErrorMessage()
	{
		if (error_message != null)
		{
			String temp = error_message;
			error_message = null;
			return temp;
		}
		else
			return null;
	}


	/**
	 * Called if the overall task is cancelled. Stops processing.
	 */
	public void stop()
	{
		task_canceled = true;
		progress_message = null;
	}


	/* PRIVATE METHODS *********************************************************/

	
	// Takes in a number of seconds and returns the appropriate amount of time
	// corresponding to this based on how many seconds there are. Some rounding
	// takes place.
	private String findTimeWithProperUnits(int seconds)
	{
		if (seconds > 3601)
		{
			int hours = seconds / 3600;
			int minutes = (seconds - (hours * 3600)) / 60;
			return new String(hours + " hours and " + minutes + " minutes");
		}
		else if (seconds > 120)
		{
			int minutes = seconds / 60;
			return new String(minutes + " minutes");
		}
		else
			return new String(seconds + " seconds");
	}
}