/*
 * @(#)ParseClassifiersFileHandler.java	1.1 06/25/04
 *
 * McGill Univarsity
 */


package bodhidharma.xml_parsing;

import org.xml.sax.*;
import java.util.LinkedList;


/**
 * An extension of the <i>Xerces</i> <code>XML DefaultHandler</code>
 * class that implements the <i>SAX</i> <code>ContentHandler</code>.
 * The methods of this class are called by an instance of an 
 * <code>XMLReader</code> while it is parsing an XML document.
 *
 * <p>This particular implementation is custom designed to parse
 * XML files of the <i>classifiers_file</i> type used by the Bodhidharma
 * music classification system. A custom exceptions is thrown 
 * if the file is not of this type. At the end of parsing,
 * the contents of the files elements are stored in the
 * <i>parsed_file_contents</i> field.
 *
 * @author Cory McKay
 * @see ParseFileHandler
 * @see	Bodhidharma
 * @see XMLDocumentParser
 * @see DefaultHandler
 * @see ContentHandler
 * @see XMLReader
 */
public class ParseClassifiersFileHandler extends ParseFileHandler
{
	/* FIELDS ******************************************************************/


	/* NOTE: String[] parsed_file_contents:
	 * Holds the data extracted from the XML file.
	 */
	
	// Counts how many elements have been encountered and read
	private int count;

	// Identifies what tag has been found
	private	int tag_identifier;

	// Stores the classifier_sub_folder_path, testing_recordings_file_path and model_testing_classifications_file_path
	private String[] set_paths;

	// Stores the NN_BioKNN_Ensemble_file_path
	private LinkedList ensemble_file_paths;


	/* PUBLIC METHODS **********************************************************/


	/**
	 * This method is called when the start of the XML file to be
	 * parsed is reached. Initializes the parsed_file_contents
	 * field to an array of empty strings and initializes the
	 * count to 0.
	 */
	public void startDocument()
	{
		set_paths = new String[4];
		for (int i = 0; i < set_paths.length; i++)
			set_paths[i] = new String("");
		ensemble_file_paths = new LinkedList();
		count = 0;
	}

	/**
	 * This method is called when the start of an XML element
	 * is encountered. Increments the count when this occurs.
	 *
	 * @param	name			Name of the element that is encountered.
	 * @throws	SAXException	Exception thrown if is wrong type of XML file.
	 */
	public void startElement(String namespace, String name, String qName, Attributes atts)
		throws SAXException
	{
		// Make sure is correct file type
		if (count == 0)
			if (!name.equals("classifiers_file"))
				throw new SAXException(".\n\nIt is in reality of the type " + name + ".");
		count++;

		// Identify the type of tag
		tag_identifier = 0;
		if (name.equals("classifier_sub_folder_path"))
			tag_identifier = 1;
		else if (name.equals("testing_recordings_file_path"))
			tag_identifier = 2;
		else if (name.equals("model_testing_classifications_file_path"))
			tag_identifier = 3;
		else if (name.equals("NN_BioKNN_Ensemble_file_path"))
			tag_identifier = 4;
		else if (name.equals("stored_training_times_report_file_path"))
			tag_identifier = 5;
	}

	/**
	 * This stores the contents of elements in the appropriate
	 * element of the parsed_file_contents array.
	 */
	public void characters(char[] ch, int start, int length)
	{
		String contents = new String(ch, start, length);

		if (tag_identifier == 1)
			set_paths[0] = contents;
		else if (tag_identifier == 2)
			set_paths[1] = contents;
		else if (tag_identifier == 3)
			set_paths[2] = contents;
		else if (tag_identifier == 4)
			ensemble_file_paths.add(contents);
		else if (tag_identifier == 5)
			set_paths[3] = contents;
	}

	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Fills the <i>parsed_file_contents</i>
	 * field with the <i>set_paths</i> and <code>ensemble_file_paths</code>.
	 *
	 * <p>After this mehtod is called, the <i>parsed_file_contents</i> field
	 * will hold information in the following order: classifier_sub_folder_path, 
	 * testing_recordings_file_path, model_testing_classifications_file_path
	 * and each of the NN_BioKNN_Ensemble_file_paths.
	 */
	public void endDocument()
	{
		Object[] ensemble_file_paths_object = ensemble_file_paths.toArray();
		String[] ensemble_file_paths_string = new String[ensemble_file_paths_object.length];
		for (int i = 0; i < ensemble_file_paths_string.length; i++)
			ensemble_file_paths_string[i] = (String) ensemble_file_paths_object[i];
		parsed_file_contents = new String[set_paths.length + ensemble_file_paths_string.length];
		int current = 0;
		for (int i = 0; i < set_paths.length; i++)
		{
			parsed_file_contents[current] = set_paths[i];
			current++;
		}
		for (int i = 0; i < ensemble_file_paths_string.length; i++)
		{
			parsed_file_contents[current] = ensemble_file_paths_string[i];
			current++;
		}
	}
}