/*
 * @(#)ParseFeatureValuesFileHandler.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma.xml_parsing;

import org.xml.sax.*;
import java.util.LinkedList;
import java.io.File;
import bodhidharma.*;
import bodhidharma.data_structures.*;
import bodhidharma.midi_parsing.*;
import bodhidharma.midi_parsing.multi_dimensional_features.*;
import bodhidharma.midi_parsing.one_dimensional_features.*;

/**
 * An extension of the <i>Xerces</i> <code>XML DefaultHandler</code>
 * class that implements the <i>SAX</i> <code>ContentHandler</code>.
 * The methods of this class are called by an instance of an 
 * <code>XMLReader</code> while it is parsing an XML document.
 *
 * <p>This particular implementation is custom designed to parse
 * XML files of the <i>feature_values_file</i> type used by the Bodhidharma
 * music classification system. A custom exceptions is thrown 
 * if the file is not of this type. At the end of parsing,
 * the contents of the files elements are stored in the
 * <i>parsed_file_contents</i> field.
 *
 * <p>After parsing, the <i>parsed_file_contents</i> field will hold an
 * array of objects. All of the objects will be <code>Recording</code> objects
 * with only the <code>file_path</code> and <code>feature_set</code> fields filled
 * (based on the tags in the parsed document) and all others set to null. 
 *
 * @author Cory McKay
 * @see	Bodhidharma
 * @see Recording
 * @see RecordingsPanel
 * @see FeatureSet
 * @see Feature
 * @see XMLDocumentParser
 * @see DefaultHandler
 * @see ContentHandler
 * @see XMLReader
 * @see LinkedList
 */
public class ParseFeatureValuesFileHandler
	extends ParseFileHandler
{

	
	/* FIELDS ******************************************************************/


	/* NOTE: public Recording[] parsed_file_contents:
	 * Holds the data extracted from the XML file.
	 */
	
	// Holds the Recording objects before they are finally converted to an array
	private	LinkedList	recordings_list;

	// The recording currently being filled
	private	Recording	current_recording;

	// The features that have been found for a given recording
	private	LinkedList	features;

	// The name of the type of Feature currently being parsed
	private String		feature_name;

	// The number of values remaining to be parsed for the given feature
	private int			values_remaining;

	// The values of a feature
	private double[]	values;

	// Identifies what tag has been found
	private	int			tag_identifier;

	// A count of the number of start elements encountered
	private	int			count;


	/* PUBLIC METHODS **********************************************************/


	/**
	 * This method is called when the start of the XML file to be
	 * parsed is reached. Instantiates the <i>recordings_list</i>
	 * field and sets  count to 0.
	 */
	public void startDocument()
	{
		recordings_list = new LinkedList();
		count = 0;
	}


	/**
	 * This method is called when the start of an XML element
	 * is encountered. Instantiates a new Recording and a new list
	 * of features if a <i>recording</i> tag was encountered.
	 * Otherwise lets the <code>characters</code> method know what
	 * kind of action to take.
	 *
	 * @param	name			Name of the element that is encountered.
	 * @throws	SAXException	Exception thrown if is wrong type of XML file.
	 */
	public void startElement(String namespace, String name, String qName, Attributes atts)
		throws SAXException
	{
		// Make sure is correct file type
		if (count == 0)
			if (!name.equals("feature_values_file"))
				throw new SAXException("\n\nIt is in reality of the type " + name + ".");
		count++;

		// Identify the type of tag
		tag_identifier = 0;
		if (name.equals("recording"))
		{
			current_recording = new Recording();
			features = new LinkedList();
		}
		else if (name.equals("file_path"))
			tag_identifier = 1;
		else if (name.equals("name"))
			tag_identifier = 2;
		else if (name.equals("dimensions"))
			tag_identifier = 3;
		else if (name.equals("value"))
			tag_identifier = 4;
	}


	/**
	 * This method responds to the contents of tags in a way
	 * determined by the name of the tag (as determined by the
	 * <code>startElement</code> method).
	 * 
	 * <p>Bases action performed on the tag_identifier set up by the
	 * <code>startElement</code> method.
	 */
	public void characters(char[] ch, int start, int length)
	{
		String contents = new String(ch, start, length);

		if (tag_identifier == 1)
		{
			if (current_recording.file_path == null)
				current_recording.file_path = contents;
			else
				current_recording.file_path += contents;
		}
		else if (tag_identifier == 2)
			feature_name = contents;
		else if (tag_identifier == 3)
		{
			values_remaining = (Integer.valueOf(contents)).intValue();
			values = new double[values_remaining];
		}
		else if (tag_identifier == 4)
		{
			values[values.length - values_remaining] = (Double.valueOf(contents)).doubleValue();
			values_remaining--;
		}
	}


	/**
	 * This method is called when the end tag of an XML element
	 * is encountered.
	 *
	 * @param	name			Name of the element that is encountered.
	 */
	public void endElement(String namespace, String name, String qName)
	{
		if (name.equals("recording"))
		{
			Object[] temp = features.toArray();
			Feature[] features_array = new Feature[temp.length];
			for (int i = 0; i < features_array.length; i++)
				features_array[i] = (Feature) temp[i];
			current_recording.feature_set = new FeatureSet(features_array);
			recordings_list.add(current_recording);
		}
		else if (name.equals("feature"))
		{
			try
			{
				Feature current_feature = FeatureMaker.makeFeature(feature_name, null);
				if (values.length == 1) // one dimensional feature
					((FeatureOneDimensional) current_feature).setValue(values[0]);
				else // multi dimensional feature
					((FeatureMultiDimensional) current_feature).setValue(values);
				features.add(current_feature);
			}
			catch (Exception e) {}
		}
	}

	
	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Fills the <i>parsed_file_contents</i>
	 * field with the <i>recordings_list</i> <code>LinkedList</code>.
	 *
	 * <p>If a lear category is not specified, fills it in with "Unknown".
	 */
	public void endDocument()
	{
		parsed_file_contents = recordings_list.toArray();
	}
}