/*
 * @(#)ParseFeedForwardNeuralNetworkFileHandler.java	1.1 06/25/04
 *
 * McGill Univarsity
 */


package bodhidharma.xml_parsing;

import org.xml.sax.*;
import java.util.LinkedList;


/**
 * An extension of the <i>Xerces</i> <code>XML DefaultHandler</code>
 * class that implements the <i>SAX</i> <code>ContentHandler</code>.
 * The methods of this class are called by an instance of an 
 * <code>XMLReader</code> while it is parsing an XML document.
 *
 * <p>This particular implementation is custom designed to parse
 * XML files of the <i>FeedForwardNeuralNetwork_file</i> type used by the Bodhidharma
 * music classification system. A custom exceptions is thrown 
 * if the file is not of this type. At the end of parsing,
 * the contents of the files elements are stored in the
 * <i>parsed_file_contents</i> field.
 *
 * @author Cory McKay
 * @see ParseFileHandler
 * @see	Bodhidharma
 * @see XMLDocumentParser
 * @see DefaultHandler
 * @see ContentHandler
 * @see XMLReader
 */
public class ParseFeedForwardNeuralNetworkFileHandler extends ParseFileHandler
{
	/* FIELDS ******************************************************************/


	/* NOTE: Object[] parsed_file_contents:
	 * Holds the data extracted from the XML file.
	 */
	
	// Counts how many elements have been encountered and read
	private int count;

	// Identifies what tag has been found
	private	int tag_identifier;

	// Stores items that are sure to occur once and only once
	private String[] fixed_fields;

	// Stores items of unknown length
	private LinkedList current_list;


	/* PUBLIC METHODS **********************************************************/


	/**
	 * This method is called when the start of the XML file to be
	 * parsed is reached. Initializes the fixed_fields
	 * field to an array of empty strings and initializes the
	 * count to 0. Initializes parsed_file_contents.
	 */
	public void startDocument()
	{
		fixed_fields = new String[6];
		for (int i = 0; i < fixed_fields.length; i++)
			fixed_fields[i] = new String("");
		count = 0;

		parsed_file_contents = new Object[4];
	}

	/**
	 * This method is called when the start of an XML element
	 * is encountered. Increments the count when this occurs.
	 *
	 * @param	name			Name of the element that is encountered.
	 * @throws	SAXException	Exception thrown if is wrong type of XML file.
	 */
	public void startElement(String namespace, String name, String qName, Attributes atts)
		throws SAXException
	{
		// Make sure is correct file type
		if (count == 0)
			if (!name.equals("FeedForwardNeuralNetwork_file"))
				throw new SAXException(".\n\nIt is in reality of the type " + name + ".");
		count++;

		// Identify the type of tag
		tag_identifier = 0;

		// Identify the tags corresponding to fixed_fields
		if (name.equals("learning_rate"))
			tag_identifier = 1;
		else if (name.equals("momentum"))
			tag_identifier = 2;
		else if (name.equals("number_input_nodes"))
			tag_identifier = 3;
		else if (name.equals("number_hidden_nodes"))
			tag_identifier = 4;
		else if (name.equals("number_output_nodes"))
			tag_identifier = 5;
		else if (name.equals("identifier"))
			tag_identifier = 6;

		// Identify the tags corresponding to the starts of new linked lists
		// and instantiate the new LinkedList
		else if ( name.equals("weights") ||
		          name.equals("categories") ||
		          name.equals("feature_names") )
			current_list = new LinkedList();

		// Identify the tags corresponding to adding items to linked lists
		else if ( name.equals("w") ||
		          name.equals("cat") ||
		          name.equals("feat") )
			tag_identifier = 7;
	}

	/**
	 * This method is called when the end tag of an XML element
	 * is encountered.
	 *
	 * @param	name			Name of the element that is encountered.
	 */
	public void endElement(String namespace, String name, String qName)
	{
		if (name.equals("categories"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[1] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				parsed_file_contents[1] = temp;
			}
		}
		else if (name.equals("feature_names"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[2] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				parsed_file_contents[2] = temp;
			}
		}
		else if (name.equals("weights"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[3] = null;
			else
			{
				double[] temp = new double[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (Double.valueOf(((String) temp_obj[i]))).doubleValue();
				parsed_file_contents[3] = temp;
			}
		}
	}

	/**
	 * This stores the contents of elements in the appropriate
	 * element of the fixed_fields array or in current_list.
	 */
	public void characters(char[] ch, int start, int length)
	{
		String contents = new String(ch, start, length);
		if (contents.equals("null"))
			contents = null;

		// Store items in fixed_fields, if appropriate
		if (tag_identifier == 1)
			fixed_fields[0] = contents;
		else if (tag_identifier == 2)
			fixed_fields[1] = contents;
		else if (tag_identifier == 3)
			fixed_fields[2] = contents;
		else if (tag_identifier == 4)
			fixed_fields[3] = contents;
		else if (tag_identifier == 5)
			fixed_fields[4] = contents;
		else if (tag_identifier == 6)
			fixed_fields[5] = contents;
		
		// Add items to current_list
		else if (tag_identifier == 7)
			current_list.add(contents);
	}

	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Fills the first entry of the <i>parsed_file_contents</i>
	 * with the <i>fixed_fields</i> array. Other entries of <i>parsed_file_contents</i>
	 * have already been filled.
	 *
	 * <p>After this mehtod is called, entry 0 of the <i>parsed_file_contents</i>
	 * field will hold information in the following order: learning_rate, momentum,
	 * number_input_nodes, number_hidden_nodes and number_output_nodes, identifier. Entry 1
	 * will hold an array of Strings representing categories into which
	 * samples can be classified. Entry 2 will hold an array of Strings representing
	 * names of features which classifications can be based on. Entry 3 will hold an
	 * array of doubles representing weights.
	 */
	public void endDocument()
	{
		parsed_file_contents[0] = fixed_fields;
	}
}