/*
 * @(#)ParseKNNFileHandler.java	1.1 06/25/04
 *
 * McGill Univarsity
 */


package bodhidharma.xml_parsing;

import org.xml.sax.*;
import java.util.LinkedList;


/**
 * An extension of the <i>Xerces</i> <code>XML DefaultHandler</code>
 * class that implements the <i>SAX</i> <code>ContentHandler</code>.
 * The methods of this class are called by an instance of an 
 * <code>XMLReader</code> while it is parsing an XML document.
 *
 * <p>This particular implementation is custom designed to parse
 * XML files of the <i>BioKNearestNeighbour_file</i> type used by the Bodhidharma
 * music classification system. A custom exceptions is thrown 
 * if the file is not of this type. At the end of parsing,
 * the contents of the files elements are stored in the
 * <i>parsed_file_contents</i> field.
 *
 * @author Cory McKay
 * @see ParseFileHandler
 * @see	Bodhidharma
 * @see XMLDocumentParser
 * @see DefaultHandler
 * @see ContentHandler
 * @see XMLReader
 */
public class ParseKNNFileHandler extends ParseFileHandler
{
	/* FIELDS ******************************************************************/


	/* NOTE: Object[] parsed_file_contents:
	 * Holds the data extracted from the XML file.
	 */
	
	// Counts how many elements have been encountered and read
	private int count;

	// Identifies what tag has been found
	private	int tag_identifier;

	// Stores items that are sure to occur once and only once
	private String[] fixed_fields;

	// Stores items of unknown length
	private LinkedList current_list;

	// Stores items of unknown length that are in current_list
	private LinkedList current_deeper_list;


	/* PUBLIC METHODS **********************************************************/


	/**
	 * This method is called when the start of the XML file to be
	 * parsed is reached. Initializes the fixed_fields
	 * field to an array of empty strings and initializes the
	 * count to 0. Initializes parsed_file_contents.
	 */
	public void startDocument()
	{
		fixed_fields = new String[3];
		for (int i = 0; i < fixed_fields.length; i++)
			fixed_fields[i] = new String("");
		count = 0;

		parsed_file_contents = new Object[5];
	}

	/**
	 * This method is called when the start of an XML element
	 * is encountered. Increments the count when this occurs.
	 *
	 * @param	name			Name of the element that is encountered.
	 * @throws	SAXException	Exception thrown if is wrong type of XML file.
	 */
	public void startElement(String namespace, String name, String qName, Attributes atts)
		throws SAXException
	{
		// Make sure is correct file type
		if (count == 0)
			if (!name.equals("KNN_file"))
				throw new SAXException(".\n\nIt is in reality of the type " + name + ".");
		count++;

		// Identify the type of tag
		tag_identifier = 0;

		// Identify the tags corresponding to fixed_fields
		if (name.equals("number_of_categories"))
			tag_identifier = 1;
		else if (name.equals("k"))
			tag_identifier = 2;
		else if (name.equals("n"))
			tag_identifier = 3;
		
		// Identify the tags corresponding to the starts of new current_list
		// and instantiate the new LinkedList
		else if ( name.equals("features_to_use") ||
		          name.equals("feature_weights") ||
		          name.equals("training_samples") ||
		          name.equals("training_sample_categories") )
			current_list = new LinkedList();

		// Identify the tags corresponding to the starts of new current_deeper_list
		// and instantiate the new LinkedList
		else if (name.equals("s"))
			current_deeper_list = new LinkedList();

		// Identify the tags corresponding to adding items to current_list
		else if ( name.equals("use") ||
		          name.equals("weight") )
			tag_identifier = 4;

		// Identify the tags corresponding to adding items to current_deeper_list
		else if ( name.equals("f") ||
		          name.equals("c") )
			tag_identifier = 5;
	}

	/**
	 * This method is called when the end tag of an XML element
	 * is encountered.
	 *
	 * @param	name			Name of the element that is encountered.
	 */
	public void endElement(String namespace, String name, String qName)
	{
		if (name.equals("features_to_use"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[1] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				boolean[] bools = new boolean[temp.length];
				for (int i = 0; i < temp.length; i++)
					bools[i] = (Boolean.valueOf(temp[i])).booleanValue();
				parsed_file_contents[1] = bools;
			}
		}
		else if (name.equals("feature_weights"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[2] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				double[] doubs = new double[temp.length];
				for (int i = 0; i < temp.length; i++)
					doubs[i] = (Double.valueOf(temp[i])).doubleValue();
				parsed_file_contents[2] = doubs;
			}
		}
		else if (name.equals("s"))
			current_list.add(current_deeper_list);
		else if (name.equals("training_samples"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[3] = null;
			else
			{
				LinkedList[] temp_outer = new LinkedList[temp_obj.length];
				for (int i = 0; i < temp_outer.length; i++)
					temp_outer[i] = (LinkedList) temp_obj[i];

				String[][] string_results = new String[temp_outer.length][];
				for (int i = 0; i < temp_outer.length; i++)
				{
					temp_obj = temp_outer[i].toArray();
					String[] temp_str = new String[temp_obj.length];
					for (int j = 0; j < temp_str.length; j++)
						temp_str[j] = (String) temp_obj[j];
					string_results[i] = temp_str;
				}

				if (temp_obj[0] == null)
					parsed_file_contents[3] = null;
				else
				{
					double[][] double_results = new double[string_results.length][];
					for (int i = 0; i < string_results.length; i++)
					{
						double_results[i] = new double[string_results[i].length];
						for (int j = 0; j < string_results[i].length; j++)
							double_results[i][j] = (Double.valueOf(string_results[i][j])).doubleValue();
					}
					parsed_file_contents[3] = double_results;
				}
			}
		}
		else if (name.equals("training_sample_categories"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[4] = null;
			else
			{
				LinkedList[] temp_outer = new LinkedList[temp_obj.length];
				for (int i = 0; i < temp_outer.length; i++)
					temp_outer[i] = (LinkedList) temp_obj[i];

				String[][] string_results = new String[temp_outer.length][];
				for (int i = 0; i < temp_outer.length; i++)
				{
					temp_obj = temp_outer[i].toArray();
					String[] temp_str = new String[temp_obj.length];
					for (int j = 0; j < temp_str.length; j++)
						temp_str[j] = (String) temp_obj[j];
					string_results[i] = temp_str;
				}

				if (temp_obj[0] == null)
					parsed_file_contents[4] = null;
				else
				{
					int[][] int_results = new int[string_results.length][];
					for (int i = 0; i < string_results.length; i++)
					{
						int_results[i] = new int[string_results[i].length];
						for (int j = 0; j < string_results[i].length; j++)
							int_results[i][j] = (Integer.valueOf(string_results[i][j])).intValue();
					}
					parsed_file_contents[4] = int_results;
				}
			}
		}
	}

	/**
	 * This stores the contents of elements in the appropriate
	 * element of the fixed_fields array or in current_list.
	 */
	public void characters(char[] ch, int start, int length)
	{
		String contents = new String(ch, start, length);
		if (contents.equals("null"))
			contents = null;

		// Store items in fixed_fields, if appropriate
		if (tag_identifier == 1)
			fixed_fields[0] = contents;
		else if (tag_identifier == 2)
			fixed_fields[1] = contents;
		else if (tag_identifier == 3)
			fixed_fields[2] = contents;
		
		// Add items to current_list
		else if (tag_identifier == 4)
			current_list.add(contents);

		// Add items to current_list
		else if (tag_identifier == 5)
			current_deeper_list.add(contents);
	}

	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Fills the first entry of the <i>parsed_file_contents</i>
	 * with the <i>fixed_fields</i> array. Other entries of <i>parsed_file_contents</i>
	 * have already been filled.
	 *
	 * <p>After this mehtod is called, entry 0 of the <i>parsed_file_contents</i>
	 * field will hold information in this order: number_of_categories, k, n.
	 * Entry 1 will hold an array of booleans representing features_to_use.
	 * Entry 2 will hold an array of doubles representing feature_weights.
	 * Entry 3 will hold a 2-D array of doubles representing the training_samples.
	 * Entry 4 will hold a 2-D array of ints representing the training_sample_categories.
	 */
	public void endDocument()
	{
		parsed_file_contents[0] = fixed_fields;
	}
}