/*
 * @(#)ParsePreferencesFileHandler.java	1.1 06/25/04
 *
 * McGill Univarsity
 */


package bodhidharma.xml_parsing;

import java.util.LinkedList;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;


/**
 * An extension of the <i>Xerces</i> <code>XML DefaultHandler</code>
 * class that implements the <i>SAX</i> <code>ContentHandler</code>.
 * The methods of this class are called by an instance of an 
 * <code>XMLReader</code> while it is parsing an XML document.
 *
 * <p>This particular implementation is custom designed to parse
 * XML files of the <i>preferences_files</i> type used by the Bodhidharma
 * music classification system. A custom exceptions is thrown 
 * if the file is not of this type. At the end of parsing,
 * the contents of the files elements are stored in the
 * <i>parsed_file_contents</i> field.
 *
 * @author Cory McKay
 * @see ParseFileHandler
 * @see	Bodhidharma
 * @see XMLDocumentParser
 * @see DefaultHandler
 * @see ContentHandler
 * @see XMLReader
 */
public class ParsePreferencesFileHandler extends ParseFileHandler
{
	/* FIELDS ******************************************************************/


	/* NOTE: LinkedList[] parsed_file_contents:
	 * Holds the data extracted from the XML file. First entry is a list of
	 * preference names, seconde is a list of corresponding preference
	 * values. All in string form.
	 */
	
	// Holds the preference labels (entries correspond to preference_values)
	// Holds strings.
	private	LinkedList	preference_labels;

	// Holds the preference values (entries correspond to preference_labels)
	// Holds strings.
	private	LinkedList	preference_values;

	// Whether or not to add the next value to preference_values
	private boolean		ready_add_value;

	// Counts how many elements have been encountered and read
	private int			count;


	/* PUBLIC METHODS **********************************************************/


	/**
	 * This method is called when the start of the XML file to be
	 * parsed is reached. Initializes the parsed_file_contents
	 * field to an array of empty strings and initializes the
	 * count to 0. Initializes LinkdedLists.
	 */
	public void startDocument()
	{
		parsed_file_contents = new String[8];
		for (int i = 0; i < 8; i++)
			parsed_file_contents[i] = new String("");;
		count = 0;

		preference_labels = new LinkedList();
		preference_values = new LinkedList();
	}

	/**
	 * This method is called when the start of an XML element
	 * is encountered. Increments the count when this occurs.
	 *
	 * @param	name			Name of the element that is encountered.
	 * @throws	SAXException	Exception thrown if is wrong type of XML file.
	 */
	public void startElement(String namespace, String name, String qName, Attributes atts)
		throws SAXException
	{
		// Make sure is correct type of file
		if (count == 0)
			if (!name.equals("preferences_file"))
				throw new SAXException(".\n\nIt is in reality of the type " + name + ".");
		count++;

		// Store labels
		if ( !name.equals("preferences_file") &&  !name.equals("comments") &&
			 !name.equals("neural_nets") &&  !name.equals("genetic_algorithms") &&
			 !name.equals("training") &&  !name.equals("end_training") &&
			 !name.equals("classification") &&  !name.equals("reports"))
		{
			preference_labels.add(name);
			ready_add_value = true;
		}
		else
			ready_add_value = false;
	}


	/**
	 * This stores the contents of elements in the appropriate
	 * element of the parsed_file_contents array.
	 */
	public void characters(char[] ch, int start, int length)
	{
		// Store the contents of the element
		String contents = new String(ch, start, length);

		// Add the contents to the list if appropriate
		if (ready_add_value)
			preference_values.add(contents);
	}


	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Fills the <i>parsed_file_contents</i>
	 * field with the <code>LinkedLists</code>.
	 */
	public void endDocument()
	{
		parsed_file_contents = new LinkedList[2];
		parsed_file_contents[0] = preference_labels;
		parsed_file_contents[1] = preference_values;
	}
}