/*
 * @(#)EditRecordingJFrame.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.StringTokenizer;
import bodhidharma.data_structures.*;
import bodhidharma.utilities.*;

/**
 * Part of the Bodhidharma automatic music classification
 * project. This dialog box allows the user to edit the meta-data
 * associated with recordings that are stored in <code>Recording</code>
 * objects.
 *
 * <p>The <i>Title</i> and <i>Artist</i> fields can be edited directly.
 * The <i>File Path</i> field is non-editable. The <i>Role</i> combo box
 * allows the user to choose between designating the recording for
 * training, testing or classification.
 *
 * <p>The <i>Leaf Categories</i> and <i>Ancestor Categories</i> text boxes
 * display the categories assigned to the recording based on taxonomy
 * that is currently loaded. Recordings that have not been assigned labels
 * are automatically designated as "Unknown". Neither of these text boxes
 * may be edited directly, but they can be changed useing the <i>Add Category</i>
 * and <i>Clear Categories</i> buttons. The <i>Add Category</i> button will add
 * the leaf category that is selected in the combo box to the right of the button
 * to the <i>Leaf Categories</i> text box. The appropriate ancestor categories,
 * as defined by the currently loaded taxonomy, will be automatically added to the
 * <i>Ancestor Categories</i> text box. The <i>Clear Categories</i> button
 * removes all leaf and ancestor categories. The <i>Play</i> button will
 * open up an external application to play the recording with the path
 * referred to in the <i><Filte Path</i> field. It should be noted that the
 * <i>Play</i> button only works in Windows, as running external applications
 * requires system specific calls, and the system must have a player installed
 * for the given recording type.
 *
 * It should be noted that this dialog box is modal, and no changes will be
 * allowed to the main window until it is closed. The <i>cancel</i> button
 * closes the dialog box without registering any changes to the recording.
 * The OK button registers the changes and closes the dialog box.
 *
 * @author Cory McKay
 * @see	Recording
 * @see RecordingsPanel
 * @see MainJFrame
 */
 public class EditRecordingJFrame
	extends JFrame
	implements ActionListener
{
	/* FIELDS ******************************************************************/

	// The Recording object that is being edited
	private Recording		recording;

	// Indicates whether the recording is being created or edited
	private boolean			adding_rather_than_editing;

	// The number of EditRecordingJFrames to cycle through between returning access to
	// the MainJFrame
	private int[]			windows_to_cycle_through;

	// Window that instantiated this window
	private RecordingsPanel	parent_window;
	
	// Containers to hold Swing elements
	private Container		content_pane;
	private JPanel			title_panel;
	private JPanel			artist_panel;
	private JPanel			file_path_panel;
	private JPanel			leaf_categories_panel;
	private JPanel			ancestor_categories_panel;
	private JPanel			role_panel;
	private JPanel			button_panel;
	private JPanel			top_panel;
	private JPanel			middle_panel;

	// Components to display and/or edit the selected recording
	private JTextArea		title_text_area;
	private JTextArea		artist_text_area;
	private JTextArea		file_path_text_area;
	private JTextArea		leaf_categories_text_area;
	private JComboBox		leaf_categories_combo_box;
	private JTextArea		ancestor_categories_text_area;
	private JComboBox		role_combo;

	// Buttons
	private JButton			add_leaf_category_button;
	private JButton			clear_leaf_categories_button;
	private JButton			ok_button;
	private JButton			cancel_button;
	private JButton			play_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. Configures the panel and its fields. Removes access to the
	 * parent MainJFrame.
	 *
 	 * @param	parent						Name of the <code>MainJFrame</code> window that spawned this dialog box.
 	 * @param	given_recording				Name of the <code>Recording</code> that this dialog box is editing.
 	 * @param	number_windows_opened		Number of EditRecordingJFrame dialog boxes opened by the parent. Is an array in order to allow multiple instantiations to access and edit it.
 	 * @param	adding_instead_of_editing	Whether the <code>Recording</code> being edited edited is being newly created or already existed.
	 * @see		MainJFrame
	 * @see		Recording
	 */
	public EditRecordingJFrame(RecordingsPanel parent, Recording given_recording, int[] number_windows_opened, boolean adding_instead_of_editing)
	{
		// Configure overall window settings and instantiate related panels
		recording = given_recording;
		adding_rather_than_editing = adding_instead_of_editing;
		windows_to_cycle_through = number_windows_opened;
		parent_window = parent;
		setTitle("Edit Recording Properties");
		int number_text_columns = 35; // number of columns in each text field
		int horizontal_gap = 6; // horizontal space between GUI elements
		int vertical_gap = 11; // horizontal space between GUI elements
		content_pane = getContentPane();
		content_pane.setLayout(new BorderLayout(horizontal_gap, vertical_gap));
		top_panel = new JPanel(new GridLayout(4, 1, horizontal_gap, vertical_gap));
		title_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		artist_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		file_path_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		role_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		middle_panel = new JPanel(new GridLayout(2, 1, horizontal_gap, vertical_gap));
		leaf_categories_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		ancestor_categories_panel = new JPanel(new BorderLayout(horizontal_gap, vertical_gap));
		button_panel = new JPanel(new FlowLayout());

		// Instantiate and configure components for displaying and/or editing
		// the selected recording and control which components can be edited
		// directly.
		title_text_area = new JTextArea(1, number_text_columns);
		artist_text_area = new JTextArea(1, number_text_columns);
		file_path_text_area = new JTextArea(1, number_text_columns);
		file_path_text_area.setEditable(false);
		leaf_categories_text_area = new JTextArea(3, number_text_columns);
		leaf_categories_text_area.setEditable(false);
		leaf_categories_text_area.setLineWrap(true);
		leaf_categories_text_area.setWrapStyleWord(true);
		String[][] categories = parent_window.parent_window.taxonomy_panel.getLeafsAndTheirParents();
		categories = Sorter.sortTable(categories, 0, new int[categories.length]);
		String[] leaf_categories = new String[categories.length];
		for (int i = 0; i < categories.length; i++)
			leaf_categories[i] = categories[i][0];
		leaf_categories_combo_box = new JComboBox(leaf_categories);
		ancestor_categories_text_area = new JTextArea(5, number_text_columns);
		ancestor_categories_text_area.setLineWrap(true);
		ancestor_categories_text_area.setWrapStyleWord(true);
		ancestor_categories_text_area.setEditable(false);
		String[] role_options = {"training", "testing", "classification"};
		role_combo = new JComboBox(role_options);
		
		// Instantiate and configure buttons and their ActionListener
		add_leaf_category_button = new JButton("Add Category:");
		clear_leaf_categories_button = new JButton("Clear Categories");
		cancel_button = new JButton("Cancel");
		ok_button = new JButton("OK");
		play_button = new JButton("Play");
		add_leaf_category_button.addActionListener(this);
		clear_leaf_categories_button.addActionListener(this);
		cancel_button.addActionListener(this);
		ok_button.addActionListener(this);
		play_button.addActionListener(this);

		// Add labels and previously created elements to panels and container
		title_panel.add(new JLabel("Title:"), BorderLayout.NORTH);
		title_panel.add(title_text_area, BorderLayout.CENTER);
		artist_panel.add(new JLabel("Artist:"), BorderLayout.NORTH);
		artist_panel.add(artist_text_area, BorderLayout.CENTER);
		file_path_panel.add(new JLabel("File Path:"), BorderLayout.NORTH);
		file_path_panel.add(file_path_text_area, BorderLayout.CENTER);
		role_panel.add(new JLabel("Role:"), BorderLayout.NORTH);
		role_panel.add(role_combo, BorderLayout.CENTER);
		leaf_categories_panel.add(new JLabel("Leaf Categories:"), BorderLayout.NORTH);
		leaf_categories_panel.add(clear_leaf_categories_button, BorderLayout.EAST);
		leaf_categories_panel.add(add_leaf_category_button, BorderLayout.WEST);
		leaf_categories_panel.add(leaf_categories_combo_box, BorderLayout.CENTER);
		leaf_categories_panel.add(leaf_categories_text_area, BorderLayout.SOUTH);
		ancestor_categories_panel.add(new JLabel("Ancestor Categories:"), BorderLayout.NORTH);
		ancestor_categories_panel.add(ancestor_categories_text_area, BorderLayout.CENTER);
		button_panel.add(play_button, BorderLayout.WEST);
		button_panel.add(cancel_button, BorderLayout.CENTER);
		button_panel.add(ok_button, BorderLayout.EAST);
		top_panel.add(title_panel);
		top_panel.add(artist_panel);
		top_panel.add(file_path_panel);
		top_panel.add(role_panel);
		middle_panel.add(leaf_categories_panel);
		middle_panel.add(ancestor_categories_panel);
		content_pane.add(top_panel, BorderLayout.NORTH);
		content_pane.add(middle_panel, BorderLayout.CENTER);
		content_pane.add(button_panel, BorderLayout.SOUTH);

		// Cause program to react when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				cancel();
			}
		});

		// Display contents of the given recording
		displayRecordingContents();

		// Disable the MainJFrame
		parent_window.parent_window.setEnabled(false);
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the add_leaf_category_button by updating both the leaf
		// and ancestor fields
		if (event.getSource().equals(add_leaf_category_button))
		{
			// Update leaf categories
			String[] leaves;
			String new_category = (String) leaf_categories_combo_box.getSelectedItem();
			if (leaf_categories_text_area.getText().equals("Unknown"))
			{
				leaf_categories_text_area.setText(new_category);
				leaves = new String[1];
				leaves[0] = new_category;
			}
			else
			{
				leaf_categories_text_area.append(";  " + new_category);
				leaves = parseLeafCategories();
				leaves = removeDoubles(leaves);
				String leaf_categories = new String("");
				for (int j = 0; j < leaves.length; j++)
				{
					if (!leaf_categories.equals(""))
						leaf_categories += ";  ";
					leaf_categories += leaves[j];
				}
				leaf_categories_text_area.setText(leaf_categories);
			}

			// Update ancestor categories
			String[] ancestors = getAncestors(leaves);
			String ancestor_categories = new String("");
			for (int j = 0; j < ancestors.length; j++)
			{
				if (!ancestor_categories.equals(""))
					ancestor_categories += ";  ";
				ancestor_categories += ancestors[j];
			}
			ancestor_categories_text_area.setText(ancestor_categories);
		}

		// React to the clear_leaf_categories_button
		else if (event.getSource().equals(clear_leaf_categories_button))
		{
			leaf_categories_text_area.setText("");
			ancestor_categories_text_area.setText("");
		}

		// React to the play_button
		else if (event.getSource().equals(play_button))
		{
			Runtime rt = Runtime.getRuntime();
			try{ Process p = rt.exec ("rundll32 SHELL32.DLL,ShellExec_RunDLL " + file_path_text_area.getText()); }
			catch (Exception e) {JOptionPane.showMessageDialog(null, "Unable to access specified file.", "ERROR", JOptionPane.ERROR_MESSAGE);}
		}

		// React to the cancel_button
		else if (event.getSource().equals(cancel_button))
			cancel();

		// React to the ok_button
		else if (event.getSource().equals(ok_button))
			ok();
	}

	
	/* PRIVATE METHODS *********************************************************/


	// Close the dialog box without registering any changes.
	private void cancel()
	{
		closeThisInstantiation(false);
	}


	// Close the dialog box and update the recording to reflect any changes.
	// Cause the RecordingsPanel to display the changes.
	private void ok()
	{
		// Record any changes in title, artist or path
		recording.title = title_text_area.getText();
		recording.artist = artist_text_area.getText();
		recording.file_path = file_path_text_area.getText();

		// Record the proper role boolean values
		if (((String) role_combo.getSelectedItem()).equals("training"))
		{
			recording.use_for_training = true;
			recording.use_for_testing = false;
			recording.use_for_classification = false;
		}
		else if ( ((String) role_combo.getSelectedItem()).equals("testing"))
		{
			recording.use_for_training = false;
			recording.use_for_testing = true;
			recording.use_for_classification = false;
		}
		else
		{
			recording.use_for_training = false;
			recording.use_for_testing = false;
			recording.use_for_classification = true;
		}

		// Parse the leaf categories and record them
		if (leaf_categories_text_area.getText().equals("") ||
			leaf_categories_text_area.getText().equals("Unknown"))
		{
			String[] unknown_id = new String[1];
			unknown_id[0] = "Unknown";
			recording.leaf_categories = unknown_id;
		}
		else
			recording.leaf_categories = parseLeafCategories();

		// Record the revised ancestor genres
		if (ancestor_categories_text_area.getText().equals("") ||
			ancestor_categories_text_area.getText().equals("Unknown"))
		{
			String[] unknown_id = new String[1];
			unknown_id[0] = "Unknown";
			recording.ancestor_categories = unknown_id;
		}
		else
			recording.ancestor_categories = getAncestors(recording.leaf_categories);

		// Hide this dialog box and reenable the parent window if all editing dialog boxes
		// have been cycled through
		closeThisInstantiation(true);
	}


	// Hide this dialog box and reenable the parent window if all editing dialog boxes
	// have been cycled through. The should_register_changes parameter indicates whether any
	// changes that were made should be stored.
	private void closeThisInstantiation(boolean should_register_changes)
	{
		if (adding_rather_than_editing && should_register_changes)
			parent_window.updateAfterAddedRecording(recording);
		windows_to_cycle_through[0]--;
		if (windows_to_cycle_through[0] == 0)
		{
			parent_window.displayRecordingsTable();
			parent_window.parent_window.setEnabled(true);
		}
		this.hide();		
	}


	// Display the contents of recording in this dialog box
	private void displayRecordingContents()
	{
		title_text_area.setText(recording.title);
		artist_text_area.setText(recording.artist);
		file_path_text_area.setText(recording.file_path);

		String role;			
		if (recording.use_for_training)
			role = new String("training");		
		else if (recording.use_for_testing)
			role = new String("testing");		
		else
			role = new String("classification");	
		role_combo.setSelectedItem(role);

		String leaf_categories = new String("");
		for (int j = 0; j < recording.leaf_categories.length; j++)
		{
			if (!leaf_categories.equals(""))
				leaf_categories += ";  ";
			leaf_categories += recording.leaf_categories[j];
		}
		leaf_categories_text_area.setText(leaf_categories);

		String ancestor_categories = new String("");
		for (int j = 0; j < recording.ancestor_categories.length; j++)
		{
			if (!ancestor_categories.equals(""))
				ancestor_categories += ";  ";
			ancestor_categories += recording.ancestor_categories[j];
		}
		ancestor_categories_text_area.setText(ancestor_categories);
	}
	

	// Return an array of strings indicating leaf categories as parsed from leaf_categories_text_area
	private String[] parseLeafCategories()
	{
		StringTokenizer tokenizer = new StringTokenizer(leaf_categories_text_area.getText(), ";", false);
		int number_leaf_categories = 0;
		while (tokenizer.hasMoreTokens())
		{
			tokenizer.nextToken();
			number_leaf_categories++;
		}
		tokenizer = new StringTokenizer(leaf_categories_text_area.getText(), ";", false);
		String return_leaves[] = new String[number_leaf_categories];
		for (int i = 0; i < return_leaves.length; i++)
		{
			String category = new String(tokenizer.nextToken());
			category = removeLeadingWhiteSpace(category);
			return_leaves[i] = category;
		}
		return return_leaves;
	}


	// Return an array of strings denoting ancestor genres for the given leaf categories
	private String[] getAncestors (String[] leaves)
	{
		String[][] leaves_and_ancestors = parent_window.parent_window.taxonomy_panel.getLeafsAndTheirParents();
		String[] return_ancestors = null;
		int j = 0;
		while(j < leaves.length)
		{
			// Finde ancestors
			int k = 0;
			while(k < leaves_and_ancestors.length)
			{
				if (leaves[j].equals(leaves_and_ancestors[k][0]))
				{
					String[] ancestors = new String[leaves_and_ancestors[k].length - 1];
					for (int m = 0; m < ancestors.length; m++)
						ancestors[m] = leaves_and_ancestors[k][m + 1];
					
					// To ensure that ancestor categories aren't overwritten
					if (return_ancestors != null)
					{
						String[] existing_ancestors = return_ancestors;
						String[] combined_ancestors = new String[existing_ancestors.length + ancestors.length];
						for (int m = 0; m < existing_ancestors.length; m++)
							combined_ancestors[m] = existing_ancestors[m];
						for (int m = existing_ancestors.length; m < existing_ancestors.length + ancestors.length; m++)
							combined_ancestors[m] = ancestors[m - existing_ancestors.length];
						return_ancestors = combined_ancestors;
					}
					else
						return_ancestors = ancestors;
				}
				k++;
			}
			j++;
		}

		// Remove ancestors that are listed more than once
		return_ancestors = removeDoubles(return_ancestors);

		// Return results
		return return_ancestors;
	}


	// Return an array of strings that is the same as the input except
	// that any duplicate entries have been removed
	private String[] removeDoubles(String[] input)
	{
		String[] output = new String[input.length];
		for (int i = 0; i < input.length; i++)
			output[i] = input[i];
		int repetition_count = 0;
		for (int i = 0; i < output.length; i++)
		{
			for (int k = 0; k < output.length; k++)
			{
				if (output[i] != null && i != k)
				{
					if (output[i].equals(output[k]))
					{
						repetition_count++;
						output[i] = null;
					}
				}
			}
		}
	if (repetition_count != 0)
		{
			String[] temp = new String[output.length];
			for (int i = 0; i < output.length; i++)
				temp[i] = output[i];
			output = new String[output.length - repetition_count];
			int output_index = 0;
			for (int i = 0; i < temp.length; i++)
			{
				if (temp[i] != null)
				{
					output[output_index] = temp[i];
					output_index++;
				}
			}
		}
		return output;
	}


	// Return a string with any leading spaces removed
	private String removeLeadingWhiteSpace(String input)
	{
		String return_string = input;
		while (return_string.startsWith(" "))
			return_string = return_string.substring(1);
		return return_string;
	}
}