/*
 * @(#)FeaturesTableModel.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma;

import javax.swing.table.DefaultTableModel;
import bodhidharma.data_structures.*;
import bodhidharma.midi_parsing.one_dimensional_features.*;
import bodhidharma.midi_parsing.multi_dimensional_features.*;

/**
 * Part of the Bodhidharma automatic music classification
 * project. This is the table model used by the
 * <code>FeaturesPanel</code> frame.
 *
 * <p>Provides methods to fill a table row by row or delete
 * everything on it. Makes all cells non-editable except those
 * in the first three columns.
 *
 * @author Cory McKay
 * @see	FeaturessPanel
 * @see FeatureSettings
 * @see Feature
 * @see DefaultTableModel
 * @see JTable
 */
 public class FeaturesTableModel
	extends DefaultTableModel
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Same constructor as <code>DefaultTableModel</code> 
	 */
	FeaturesTableModel(Object[] columnNames, int rows)
	{
		super(columnNames, rows);
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns false for all cells but those in the first three columns.
	 * Overrides the <code>DefaultTableModel</code> method.
	 */
	public boolean isCellEditable(int row, int column)
	{
		if (column == 0 || column == 1 || column == 2)
			return true;
		else
			return false;
	}


	/**
	 * Deletes everything in the table and then fills it up one row at a time
	 * based on the given <code>FeatureSettings</code> array.
	 *
	 * @param	feature_settings	Data to place in the table.
	 */
	public void fillTable(FeatureSettings[] feature_settings)
	{
		// Remove the contents of the table
		clearTable();

		// Populate each row one by one
		for (int i = 0; i < feature_settings.length; i++)
		{
			// Set up row
			Object[] row_contents = new Object[6];
			row_contents[0] = new Boolean(feature_settings[i].extract_feature);
			row_contents[1] = new Boolean(feature_settings[i].candidate_feature);
			row_contents[2] = new Boolean(feature_settings[i].overide_feature);
			row_contents[3] = feature_settings[i].getFeature().getCode();
			row_contents[4] = feature_settings[i].getFeature().getName();
			if (feature_settings[i].getFeature() instanceof FeatureOneDimensional)
				row_contents[5] = "one-dimensional";
			else
				row_contents[5] = "multi-dimensional";

			// Add the row
			addRow(row_contents);
		}
	}


	/**
	 * Returns the type of class used for each column.
	 * Necessary in order for text boxes to be properly displayed.
	 *
	 * @param	column	Column to check.
	 */
	public Class getColumnClass(int column)
	{
		return getValueAt(0, column).getClass();
	}


	/* PRIVATE METHODS *********************************************************/


	/**
	 * Removes all contents of the table.
	 */
	private void clearTable()
	// Remove all data on the table
	{
		while (getRowCount() != 0)
		{
			removeRow(0);
		}
	}
}