/*
 * @(#)MainJFrame.java	1.2 06/29/05
 *
 * McGill Univarsity
 */

package bodhidharma;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.File;
import java.io.IOException;

/**
 * Part of the Bodhidharma automatic music classification
 * project. This is the main GUI container. It holds five
 * tabbed panels and two menus.
 * 
 * <p>The <i>File</i> menu allows the user to change the
 * paths in the <code>ConfigureFileLocationsJFrame</code>
 * and load new projects (<i>Set Project Paths</i>), save
 * the current project file (<i>Save Project File</i>),
 * save the current project file and all related configurations
 * <i>Save Project</i> or quit (<i>Quit</i>).
 *
 * <p>The <i>Help</i> menus allows the user to view basic
 * information about the the software (<i>About</i>) or
 * view the detailed on-line help (<i>Check Manual</i>).
 *
 * <p>The public fields allow the user to access the related
 * tabbed panels.
 *
 * <p>The <code>focusOnClassifierPane</code> method can be
 * called to switch the focus of the window to the <i>classification_panel</i>.
 *
 * @author Cory McKay
 * @see	ConfigureFileLocationsJFrame
 */
public class MainJFrame
	extends JFrame
	implements ActionListener
{
	/* FIELDS ******************************************************************/


	/**
	 * The panel used to hold and edit the current taxonomy.
	 */
	public	TaxonomyPanel		taxonomy_panel;

	/**
	 * The panel used to hold and edit recording meta-data and
	 * features.
	 */
	public	RecordingsPanel		recordings_panel;

	/**
	 * The panel used to hold and edit feature settings.
	 */
	public	FeaturesPanel		features_panel;

	/**
	 * The panel used to hold and edit preferences.
	 */
	public	PreferencesPanel	preferences_panel;

	/**
	 * The panel used to train and classify recordings as
	 * well as view related reports.
	 */
	public	ClassificationPanel	classification_panel;

	/**
	 * JFrame that stores paths of XML files used in current project.
	 */
	public	ConfigureFileLocationsJFrame	file_locations;

	// GUI components
	private JTabbedPane	tabbed_pane;
	private Container	content_pane;
	
	// Menu fields
	private	JMenuBar			menu_bar;
	private JMenu				file_menu;
	private JMenu				help_menu;
	private	JMenuItem			set_project_paths_menu_item;
	private	JMenuItem			save_project_file_menu_item;
	private	JMenuItem			save_project_menu_item;
	private	JMenuItem			quit_menu_item;
	private	JMenuItem			about_menu_item;
	private	JMenuItem			manual_menu_item;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic contructor used to set up the GUI and instantiate related objects.
 	 *
	 * @param	file_paths	Name of the <code>ConfigureFileLocationsJFrame</code> window that holds configuration file paths.
	 */
	public MainJFrame(ConfigureFileLocationsJFrame file_paths)
	{
		// Set up storage of file paths
		file_locations = file_paths;

		// Initialize GUI
		setSize(800, 600);
		setTitle("Bodhidharma           " + file_locations.project_path);
		content_pane = getContentPane();
		tabbed_pane = new JTabbedPane();

		// Set up menus
		menu_bar = new JMenuBar();
		file_menu = new JMenu("File");
		set_project_paths_menu_item = new JMenuItem("Set Project Paths");
		set_project_paths_menu_item.addActionListener(this);
		save_project_file_menu_item = new JMenuItem("Save Project File");
		save_project_file_menu_item.addActionListener(this);
		save_project_menu_item = new JMenuItem("Save Project");
		save_project_menu_item.addActionListener(this);
		quit_menu_item = new JMenuItem("Quit");
		quit_menu_item.addActionListener(this);
		file_menu.setMnemonic('F');
		set_project_paths_menu_item.setMnemonic('p');
		save_project_file_menu_item.setMnemonic('i');
		save_project_menu_item.setMnemonic('s');
		quit_menu_item.setMnemonic('q');
		file_menu.add(set_project_paths_menu_item);
		file_menu.add(save_project_file_menu_item);
		file_menu.add(save_project_menu_item);
		file_menu.add(quit_menu_item);
		menu_bar.add(file_menu);
		help_menu = new JMenu("Help");
		about_menu_item = new JMenuItem("About");
		about_menu_item.addActionListener(this);
		manual_menu_item = new JMenuItem("Check Manual");
		manual_menu_item.addActionListener(this);
		help_menu.setMnemonic('H');
		about_menu_item.setMnemonic('A');
		manual_menu_item.setMnemonic('M');
		help_menu.add(about_menu_item);
		help_menu.add(manual_menu_item);
		menu_bar.add(help_menu);
		setJMenuBar(menu_bar);
		
		// Set up tabbed pains
		taxonomy_panel = new TaxonomyPanel(this);
		recordings_panel = null;
		features_panel = new FeaturesPanel(this);
		recordings_panel = new RecordingsPanel(this);
		preferences_panel = new PreferencesPanel(this);
		classification_panel = new ClassificationPanel(this);
		tabbed_pane.addTab("Model Taxonomy", taxonomy_panel);
		tabbed_pane.addTab("Feature Settings", features_panel);
		tabbed_pane.addTab("Preferences", preferences_panel);
		tabbed_pane.addTab("Recordings", recordings_panel);
		tabbed_pane.addTab("Classification", classification_panel);
		content_pane.add(tabbed_pane);

		// Cause program to quit when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				System.exit(0);
			}
		});
	}


	/* PUBLIC METHODS **********************************************************/

	
	/**
	 * Listen for and react to user actions.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the set_project_paths_menu_item
		if (event.getSource().equals(set_project_paths_menu_item))
			file_locations.show();

		// React to the save_project_file_menu_item
		// Save the current project file
		if (event.getSource().equals(save_project_file_menu_item))
			file_locations.saveProjectFile(file_locations.project_path, "", true);

		// React to the save_project_menu_item
		// Save the current project file and any other appropriate XML files
		if (event.getSource().equals(save_project_menu_item))
		{
			taxonomy_panel.saveTaxonomyFile("", true);
			features_panel.saveFeaturesSettingsFile("", true);
			preferences_panel.savePreferencesFile("", true);
			recordings_panel.saveRecordingsFile("", true);
			classification_panel.saveClassifierFiles("", true);
			file_locations.saveProjectFile(file_locations.project_path, "", true);
		}

		// React to the quit_menu_item
		if (event.getSource().equals(quit_menu_item))
			System.exit(0);

		// React to the about_menu_item
		if (event.getSource().equals(about_menu_item))
			(new AboutDialog(this)).show();

		// React to the manual_menu_item
		if (event.getSource().equals(manual_menu_item))
			(new HelpDialog()).show();
	}


	/**
	 * Changet the focus of the window to the <i>classification_panel</i>.
	 */
	public void focusOnClassifierPane()
	{
		tabbed_pane.setSelectedComponent(classification_panel);
	}


	/**
	 * Perform feature extraction and saves the results. Only performed
	 * if features not already extracted. This method is designed
	 * to be used with the command line interface of Bodhidharma.
	 */
	public void commenceFeatureExtraction()
	{
		recordings_panel.extract_features_button.doClick();
		ProgressBarTaskFeatureExtractionMonitor	monitor_extractor = recordings_panel.monitor_extractor;
		if (monitor_extractor != null)
		{
			while (!monitor_extractor.isDone() && !monitor_extractor.isCancelled())
			{
			}
			if (monitor_extractor.isDone())
			{
				recordings_panel.progress_indicator.done();
				recordings_panel.saveFeatureValues();
			}
		}
	}
	

	/**
	 * Commences training on the recordings specified in
	 * file_locations. Saves the trained classifiers
	 * This method is designed to be used with the command
	 * line interface of Bodhidharma.
	 */
	public void commenceTraining()
	{
		classification_panel.trainFromCommandLine();
	}


	/**
	 * Commences classifying the recordings specified in 
	 * file_locations and prints the results to the specified
	 * file path. This method is designed to be used with the command line
	 * interface of Bodhidharma.
	 *
	 * @param	output_classifications_path	The path to save the output
	 *										classifications.
	 */
	public void commenceClassifiying(String output_classifications_path)
	{
		try
		{
			classification_panel.classifyFromCommandLine(output_classifications_path);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
		}
	}
}