/*
 * @(#)ProgressBarTaskFeatureExtractionMonitor.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma;

import bodhidharma.utilities.*;
import bodhidharma.data_structures.*;

/**
 * An implementation of the <code>ProgressBarTaskMonitor</code> abstract class.
 * Parses MIDI files and extracts features from them and provides information to
 * a <code>ProgressBarDialog</code> object that is used to keep it updated.
 *
 * <p>Parsing is started by calling the <code>go</code> method. This will cause
 * the processing in <code>ActualTask</code> to occur. <code>ActualTask</code>
 * will keep the fields described in this class' superclass updated, which can
 * be accessed using the methods in this class' superclass.
 *
 * <p>The parsed contents of the MIDI files can be accessed with the
 * <code>getParsedFeatures</code> method, which returns a <code>FeatureSet</code>
 * array with entries corresponding to those of the <i>recordings</i> parameter
 * passed to the constructor of this class. Entries are set to null if the
 * corresponding file was not set to be parsed or was not succesfully parsed.
 *
 * @author	Cory McKay
 * @see		ProgressBarTaskMonitor
 * @see		ProgressBarDialog
 * @see		SwingWorker
 * @see		Recording
 * @see		FeatureSet
 * @see		RecordingsPanel
 */
public class ProgressBarTaskFeatureExtractionMonitor
	extends ProgressBarTaskMonitor
{
	/* FIELDS ******************************************************************/


	// The features parsed from the files. Set to null if the file is not set to
	// be parsed in recordings_to_extract or if it could not be succesfully parsed.
	private FeatureSet[]	parsed_features;

	// Fields used to parse MIDI files
	private	Recording[]		recordings_list;
	private	boolean[]		recordings_to_extract;
	private	String[]		required_feature_names;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Same constructor as the superclass. Also stores the given parameters and instantiates
	 * parsed_features with all entries set to null.
	 *
	 * @param	number_of_sub_tasks			Number of sub-tasks that will need to be completed in all.
	 * @param	recordings					The Recording objects containing the names of the files that are to be parsed.
	 * @param	recordings_to_be_extracted	Set to true if the file referred to in the corresponding entry of recordings is to be parsed.
	 * @param	names_of_required_features	The names of the features that are to be extracted.
	 */
	public ProgressBarTaskFeatureExtractionMonitor( int number_of_sub_tasks,
		                                            Recording[] recordings,
	                                                boolean[] recordings_to_be_extracted,
	                                                String[] names_of_required_features )
	{
		super(number_of_sub_tasks);

		recordings_list = recordings;
		recordings_to_extract = recordings_to_be_extracted;
		required_feature_names = names_of_required_features;

		parsed_features = new FeatureSet[recordings.length];
		for (int i = 0; i < parsed_features.length; i++)
			parsed_features[i] = null;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Called from a <code>ProgressBarDialog</code> object to the task to be performed.
	 * Sets up a <code>SwingWorker</code> thread to carry out the thread and sets time_started.
	 */
	public void go()
	{
		final SwingWorker worker = new SwingWorker()
		{
			public Object construct()
			{
				current_amount_completed = 0;
				task_completed = false;
				task_canceled = false;
				progress_message = null;
				error_message = null;

				return new ActualTask();
			}
		};
		
		time_started = System.currentTimeMillis();

		worker.start();
	}


	/**
	 * Returns value parsed_features, the results of parsing. Entries correspond to
	 * those of the <i>recordings_list</i> field. Entries are set to null if the corresponding
	 * file was not set to be parsed or was not succesfully parsed.
	 */
	public FeatureSet[] getParsedFeatures()
	{
		return parsed_features;
	}


	/* PRIVATE METHODS *********************************************************/


	// Append the second parameter to the first parameter and return results. Replace the first
	// parameter with an empty string if it is null.
	private String appendWithoutNull(String string_to_append_to, String string_to_append)
	{
		if (string_to_append_to == null)
			string_to_append_to = new String("");
		string_to_append_to += string_to_append;
		return string_to_append_to;
	}


	/* INTERNAL CLASS **********************************************************/


	/**
	 * The actual task that is performed. This runs in a <code>SwingWorker</code> thread.
	 * Extract features from all recordings indicated in recordings_to_extract and store 
	 * them in the <code>parsed_features</code> field. Entries are set to null if the
	 * corresponding file was not set to be parsed or was not succesfully parsed.
	 *
	 * The <code>progress_message</code> and <code>error_message</code> fields are also
	 * kept updated with the names of files being processed and any errors that occur,
	 * respectively.
	 *
	 * If the <code>task_canceled</code> field is set to true externally then processing is
	 * terminated.
	 */
	class ActualTask
	{
		ActualTask()
		{
			// Initialize progress_message
			progress_message = "EXTRACTING FEATURES FROM " + length_of_task + " FILES...\n";

			// Extract features from recordings one file at a time
			for (int rec = 0; rec < recordings_list.length; rec++)
			{
				if (task_canceled)
					rec = recordings_list.length + 1; // stop processing if task cancelled
				else
				{
					// Only extract from recordings marked in recordings_to_extract
					if (recordings_to_extract[rec])
					{
						// Note the name of the file currently being processed
						progress_message = appendWithoutNull(progress_message, (">> " + StringMethods.convertFilePathToFileName(recordings_list[rec].file_path) + "..."));

						// Extract features from a file and note when and if completed succesfully
						try 
						{
							parsed_features[rec] = new FeatureSet(recordings_list[rec].file_path, required_feature_names);
							progress_message = appendWithoutNull(progress_message, " extraction succesful\n");
						}

						// React to any errors by setting the parsed_features entry to null and providing
						// a notification to the user
						catch (Throwable t)
						{
							// Update the progress_message
							progress_message = appendWithoutNull(progress_message, " extraction FAILED (see Errors Encountered below)\n");

							// React to the Java Runtime running out of memory
							if (t.toString().equals("java.lang.OutOfMemoryError"))
								error_message = appendWithoutNull(error_message, ("Eror processing " + StringMethods.convertFilePathToFileName(recordings_list[rec].file_path) +
								                                  "\nThe Java Runtime ran out of memory. Please rerun this program with a higher amount of memory assigned to the Java Runtime heap.\n\n"));
							
							// React to any other error
							else
							{
								error_message = appendWithoutNull(error_message, ("Eror processing " + StringMethods.convertFilePathToFileName(recordings_list[rec].file_path) +
								                                  "\n" +  t.getMessage() + "\n\n"));

								/* Print the filename where the error occured and show the error stack trace
								System.out.println("\n\n--> " + recordings_list[rec].file_path + "\n");
								t.printStackTrace();
								System.exit(-1);
								*/
							}

							// Make sure that the features are set to null for the file where the error was encountered
							parsed_features[rec] = null;
						}
					}
				}
			
				// Update the number of files processed
				current_amount_completed++;
			}

			// Mark the task as completed
			task_completed = true;
		}
	}
}