/*
 * @(#)RecordingsPanel.java	1.2.2 07/15/05
 *
 * McGill Univarsity
 */

package bodhidharma;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;
import java.io.*;
import java.util.LinkedList;
import bodhidharma.xml_parsing.*;
import bodhidharma.data_structures.*;
import bodhidharma.midi_parsing.*;
import bodhidharma.midi_parsing.one_dimensional_features.*;
import bodhidharma.midi_parsing.multi_dimensional_features.*;
import bodhidharma.utilities.*;

/**
 * Part of the Bodhidharma automatic music classification project. Allows the user to create,
 * view and edit lists of recordings, listen to recordings, edit the names and model categories
 * of recordings and extract features from recordings. Holds the <i>recordings_list</i> field
 * which can be accessed by other objects in order to gain information about what recordings
 * are currently loaded and their characteristics and feature values. The recordings in the list
 * are the ones that will be used for training and/or testing.
 *
 * <p>There are three possible views available in this <code>JPanel</code>. The user may move
 * between them by pressing on the <i>Display Recordings Info</i>, <i>Display Feature Values</i>
 * and <i>Display Scaled Feature Values</i> buttons. The first view is described in the following paragraphs.
 * The second view is only available if featuers have been extracted from at least one of the loaded recordings.
 * This second view displays the values of all features that have been extracted from the loaded
 * recordings. Multi-dimensional features are marked "MD" in this table. Double clicking on the MD
 * for a particular recording will bring up a window displaying each dimension of the feature
 * of the selected recording. Individual features that have not been extracted for a given recording
 * are marked "--". No buttons are available in this view other than the other two display buttons.
 * The third view is also only available if features have been extracted from the
 * loaded recordings, and a trained one-dimensional classifier with calculated scaling parameters must
 * also be loaded. This third view displays the scaled values of the one-dimensional (only) features for each
 * recording as well as the minimum and maximum value cutoffs for each feature, as defined by the classifier.
 * The following paragraphs discuss the view and options related to the view
 * brought up by pressing the <i>Display Recordings Info</i> button.
 *
 * <p>The table displayes the Title, Artist, File Name, Leaf Categories, Ancestor Categories
 * and roles of each recording. The column widths can be adjusted by clicking on the column
 * borders and dragging. The recordings may be sorted based on the contents of any
 * column by clicking on the appropriate column header.
 * 
 * <p>The contents of any particular recording or recordings may be viewed in greater detail 
 * and/or edited by selecting the appropriate row(s) and clicking on the <i>Edit Selections</i>
 * button. Double clicking on a single row has the equivalent effect. In either case, an
 * <code>EditRecordingJFrame</code> dialog box will be brought up for each of the selected
 * recordings. This will allow the user to view the recording in more detail and change its
 * title, artist, role and model category fields. The role of a recording may also be changed
 * by clicking on the appropriate row under the <i>Role</i> column and selecting the desired
 * Role from the resultant combo box.
 *
 * <p>Recordings may be deleted by selecting the appropriate rows and clicking on the 
 * <i>Delete Selections</i> button. A new recording may also be added by clicking on the
 * <i>Add New Recording</i> button. This will bring up a <code>JFileChooser</code> dialog
 * box that will allow the user to choose the appropriate file. It should be noted that the
 * <code>JFileChooser</code> dialog box only displays files of the appropriate recording type
 * that are not currently in the list. All recordings must be in the same directory, so if at
 * least one recording is loaded, then the user will only be permitted to see files in that
 * directory. Multiple files may be chosen from this dialog box to be added at once. A
 * <code>EditRecordingJFrame</code> dialog box will then be brought up to allow the user to
 * configure each new recording's meta-data.
 *
 * <p>Features are NOT automatically extracted from a recording when it is added by pressing
 * the <i>Add New Recordings</i> button. In order to extract features, one must press the
 * <i>Extract Features</i> button. This will bring up a <code>ProgressBarDialog</code>
 * that will extract the features marked for extraction in the <code>feature_settings</code>
 * field of the <code>FeaturesPanel</code> from all currently loaded recordings that have
 * not already had these features extracted. The user may cancel the extraction at any time
 * by pressing on the <code>ProgressBarDialog</code>'s <i>Cancel</i> button at any time, but
 * none of the extracted features will be recorded if this is done. When extraction is complete
 * the user has the option of pressing the <code>ProgressBarDialog</code>'s <i>Accept</i> button.
 * This will cause all the extracted features to be saved in the <code>recordings_list</code>
 * field. It will also cause the system to cycle through ALL loaded recordings ane delete any
 * features that were extracted previously but are not currently marked for extraction in the
 * <code>feature_settings</code> field of the <code>FeaturesPanel</code>. Pressing the <i>Accept</i>
 * button also causes all recordings that could not have features extracted from them due to
 * some error (which will have been reported in the <i>Errors Encountered</i> area of the 
 * <code>ProgressBarDialog</code>) to be deleted from the <code>recordings_list</code> field.
 *
 * <p>Pressing on the <i>Forget Extracted Features</i> button will cause the system to erase
 * all extracted feature values from the <code>recordings_list</code> field.
 *
 * <p>A new set of recordings can be loaded from a <code>recordings_list</code> file by
 * clicking on the <i>Load Recordings List</i> button. This will erase the currently loaded
 * recordings list, if any, and replace it with the contents of this file. Unspecified genres,
 * titles or artists will be displayed as "Unknown". Note that  <code>recordings_list</code> files
 * only include leaf categories, and that the contents of the Ancestor Categories column are
 * dynamically generated, and depond on the contents of the <i>taxonomy</i> field of the
 * <code>TaxonomyPanel</code>. Upon loading a <code>recordings_list</code> file, the system
 * checks to see if a <code>feature_values_file<code> file containing feature values for the
 * recordings in the <code>recordings_list</code> file is available, and loads it if it is. This
 * file is always in the same directory and has the same name as the <code>recordings_list</code>
 * file, but with the string "_feature_values" appended to the end of the file name.
 *
 * <p>Note that it is also possible to load tab delimited text files in the MIREX 2005 format that
 * hold recording lists as an alternative to the XML files. This functionality is intended primarily
 * for use with the command line use of Bodhidharma.
 *
 * <p>The <i>Verify Files</i> button causes the system to check that all of the files referred
 * to in the list of recordings actually exist. A report window is generated ro report any problems
 * encountered.
 *
 * <p><b>IMPORTANT:</b> alll recordings files referred to in a <code>recordings_list</code> file or
 * added to the list with the <i>Add New Recording</i> button must be in the same folder. It is
 * for this reason that the <code>JFileChooser</code> dialog box does not display directories
 * unless no recordings are currently in the list. If the user wishes to move the recordings to
 * a different directory, then s/he must press on the <i>Change Recordings Path</i> button
 * and select the new directory where all recordings referred to in the <i>recordings_list</i>
 * field can be found.
 *
 * <p>The current contents of the <i>recordings_list</i> field can be saved using the
 * <i>Save Recordings List</i> or <i>Save Recordings List As</i> buttons. If features have been
 * extracted for the recordings in the <code>recordings_list</code> field, then a 
 * <code>feature_values_file<code> file containing feature values for the recordings is also saved.
 * This file is saved in the same directory and has the same name as the <code>recordings_list</code>
 * file, but with the string "_feature_values" appended to the end of the file name. Features are
 * only saved for recordings that have had features extracted from them and no checking is performed
 * to see if all recordings have the same features extracted from them.
 *
 * <p>The <i>New Recordings List</i> button erases the current list of recordings and replaces
 * it with a blank list to which new recordings can be added.
 * 
 * <p>Statistics on the categories of the loaded recordings can be seen by pressing on the
 * <i>Display Category Count</i> button. This generates a window showing the number of
 * recordings falling into each category. Categories are divided into three groups:
 * leaf categories, parent categories (i.e. categories at the highest level of the taxonomy)
 * and middle categories (i.e. categoriesthat are neither leaves nor parent categories. The
 * categories in each of these groups are sorted alphabetically. The number of recordings which 
 * have not been assigned model categories is also reported, as are the sum of leaf categories
 * beyond the first for each recording and the number of recordings in all.
 *
 * <p>The <i>Classify Selections</i> button classifies the highlighted recordings and switches 
 * to the <code>ClassificationPanel</code> to show the results. Replaces any stored test recordings
 * <code>ClassificationPanel</code> with the selected recordings. Any model categories are ignored.
 * Note that the recordings that are classified are the ones that are highlighted, and that the
 * <i>role</i> field has no effect.
 *
 * <p>The <code>saveRecordingsFile</code> method can be called to save the contents of the
 * <code>recordings_list</code> field in a <code>recordings_list</code> file and a
 * <code>feature_values_file<code> file.
 *
 * <p>The <code>displayRecordingsTable</code> method can be called to update the displayed
 * table of recordings, the status bar and which buttons are enabled.
 *
 * <p>The <code>updateAfterAddedRecording</code> method can be called to update the 
 * <code>recordings_list</code> field and the display to reflect the addition of a new recording.
 *
 * <p>The <code>acceptProcessingResults</code> method can be called to cause the system to reflect
 * the results of feature extraction performed by a
 * <code>ProgressBarTaskFeatureExtractionMonitor</code>.
 *
 * <p>The <code>newRecordingsList</code> method can be called to clear the <code>recordings_list</code>
 * field and update the display.
 *
 * @author Cory McKay
 * @see ProcessResultsReceiver
 * @see	ConfigureFileLocationsJFrame
 * @see MainJFrame
 * @see XMLDocumentParser
 * @see ParseRecordingsFileHandler
 * @see ProgressBarTaskFeatureExtractionMonitor
 * @see ProgressBarTaskMonitor
 * @see ProgressBarDialog
 * @see JTable
 * @see RecordingsTableModel
 * @see Recording
 * @see FeatureSet
 * @see FeatureSettings
 */
public class RecordingsPanel
	extends ProcessResultsReceiver
	implements ActionListener
{
	/* FIELDS ******************************************************************/


	/**
	 * The array of recordings to be used for testing, training
	 * and/or classification.
	 */
	public	Recording[]								recordings_list;

	/**
	 * Window that holds this JPanel
	 */
	public MainJFrame								parent_window;

	/**
	 * Indicates whether features (of any kind) have been extracted
	 * from recordings in recordings_list
	 */
	public	boolean									features_extracted;

	/**
	 * Monitors feature extraction and holds results temporarily until
	 * acknowledged by user
	 */
	public	ProgressBarTaskFeatureExtractionMonitor	monitor_extractor;

	/**
	 * Dialog box displaying status of feature extraction.
	 */
	public	ProgressBarDialog						progress_indicator;

	// The order of in which to display the recording_list
	private int[]									list_ordering;
	
	// Notes which recordings need to have their features updated
	private	boolean[]								recordings_to_extract;

	// Names of the set of features to be extracted
	private	String[]								required_feature_names;

	// Fields relating to the table used to display the contents of recording_list
	private	JTable									recordings_table;
	private RecordingsTableModel					recordings_table_model;

	// Fields relating to the table used to display feature values of the contents of
	// recordings_list
	private	JTable									features_value_table;
	private NonEditableTableModel					features_value_table_model;

	// Fields relating to the table used to display scaled feature values of the contents of
	// recordings_list
	private	JTable									scaled_features_value_table;
	private NonEditableTableModel					scaled_features_value_table_model;

	// Display panels
	private JPanel									recordings_display_panel;
	private JPanel									button_panel;
	private JScrollPane								recordings_display_scroll_pane;
	private JScrollPane								feature_value_diplay_scroll_pane;
	private JScrollPane								scaled_feature_value_diplay_scroll_pane;

	// Displays general messages
	private JTextField								status_bar;

	// Buttons
	private JButton									display_recordings_button;
	private JButton									display_feature_values_button;
	private JButton									display_scaled_feature_values_button;
	public	JButton									extract_features_button;
	private JButton									classify_recordings_button;
	private JButton									new_recordings_list_button;
	private JButton									change_recordings_path_button;
	private JButton									load_recordings_list_button;
	private JButton									save_recordings_list_button;
	private JButton									save_recordings_list_as_button;
	private JButton									edit_recording_button;
	private JButton									add_new_recordings_button;
	private JButton									delete_recordings_button;
	private JButton									count_categories_button;
	private JButton									verify_files_button;
	private JButton									delete_saved_features_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor. Configures the panel and loads the recordings specified in the 
	 * <i>recordings_file</i> XML file specified in the <code>ConfigureFileLocationsJFrame</code> 
	 * object. Creates a new empty recordings list if a valid taxonomy file is not specified
	 * there or if the specified <i>recordings_file</i> is invalid or contains categories
	 * not in the currently loaded taxonomy.
	 *
 	 * @param	parent	Name of the <code>MainJFrame</code> window that holds this panel.
	 * @see		ConfigureFileLocationsJFrame
	 * @see		MainJFrame
	 */
	public RecordingsPanel(MainJFrame parent)
	{
		// Store a pointer to the container that called this RecordingsPanel
		parent_window = parent;

		// Initialize layout settings
		setLayout(new BorderLayout());
		int horizontal_gap = 4; // horizontal space between GUI elements
		int vertical_gap = 4; // horizontal space between GUI elements

		// Set up status bar
		status_bar = new JTextField("");
		status_bar.setEditable(false);
		add(status_bar, BorderLayout.SOUTH);

		// Set up buttons
		button_panel = new JPanel(new GridLayout(18, 1, horizontal_gap, vertical_gap));
		display_recordings_button = new JButton("Display Recordings Info");
		display_feature_values_button = new JButton("Display Feature Values");
		display_scaled_feature_values_button = new JButton("Display Scaled Feature Values");
		extract_features_button = new JButton("Extract Features");
		delete_saved_features_button = new JButton("Forget Extracted Features");
		classify_recordings_button = new JButton("Classify Selections");
		new_recordings_list_button = new JButton("New Recordings List");
		change_recordings_path_button = new JButton("Change Recordings Path");
		load_recordings_list_button = new JButton("Load Recordings List");
		save_recordings_list_button = new JButton("Save Recordings List");
		save_recordings_list_as_button = new JButton("Save Recordings List As");
		edit_recording_button = new JButton("Edit Selections");
		add_new_recordings_button = new JButton("Add New Recordings");
		count_categories_button = new JButton("Display Category Count");
		verify_files_button = new JButton("Verify Files");
		delete_recordings_button = new JButton("Delete Selections");
		display_recordings_button.addActionListener(this);
		display_feature_values_button.addActionListener(this);
		display_scaled_feature_values_button.addActionListener(this);
		extract_features_button.addActionListener(this);
		classify_recordings_button.addActionListener(this);
		new_recordings_list_button.addActionListener(this);
		change_recordings_path_button.addActionListener(this);
		load_recordings_list_button.addActionListener(this);
		save_recordings_list_button.addActionListener(this);
		save_recordings_list_as_button.addActionListener(this);
		edit_recording_button.addActionListener(this);
		add_new_recordings_button.addActionListener(this);
		delete_recordings_button.addActionListener(this);
		count_categories_button.addActionListener(this);
		verify_files_button.addActionListener(this);
		delete_saved_features_button.addActionListener(this);
		button_panel.add(display_recordings_button);
		button_panel.add(display_feature_values_button);
		button_panel.add(display_scaled_feature_values_button);
		button_panel.add(count_categories_button);
		button_panel.add(verify_files_button);
		button_panel.add(change_recordings_path_button);
		button_panel.add(new JLabel());
		button_panel.add(add_new_recordings_button);
		button_panel.add(edit_recording_button);
		button_panel.add(delete_recordings_button);
		button_panel.add(extract_features_button);
		button_panel.add(delete_saved_features_button);
		button_panel.add(classify_recordings_button);
		button_panel.add(new JLabel());
		button_panel.add(new_recordings_list_button);
		button_panel.add(load_recordings_list_button);
		button_panel.add(save_recordings_list_button);
		button_panel.add(save_recordings_list_as_button);
		add(button_panel, BorderLayout.EAST);

		// Read in the starting recordings, validate it and display it
		// If a blank string is given or if the taxonomy is empty,
		// then just start a new recordings list.
		recordings_list = null;
		if (parent_window.file_locations.recordings_list_path.equals(""))
			newRecordingsList();
		else if (parent_window.taxonomy_panel.isTreeEmpty())
			newRecordingsList();
		else
		{
			boolean worked = false;
			worked = parseRecordingsFile(parent_window.file_locations.recordings_list_path);
			if (!worked)
				parent_window.file_locations.setRecordingsPath(""); // set path to blank if file incorrect
		}
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the display_feature_values_button
		if (event.getSource().equals(display_recordings_button))
			displayRecordingsTable();

		// React to the display_feature_values_button
		else if (event.getSource().equals(display_feature_values_button))
			displayFeatureValues();

		// React to the display_scaled_feature_values_button
		else if (event.getSource().equals(display_scaled_feature_values_button))
		{
			try { displayScaledFeatureValues(); }
			catch (Exception e) { JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE); }
		}

		// React to the extract_features_button
		else if (event.getSource().equals(extract_features_button))
		{
			// Check if any features are marked to be extracted on the FeaturesPanel.
			// If so, extract them, if not, then give a message.
			FeatureSettings[] feature_settings = parent_window.features_panel.getFeatureSettings();
			int count = 0;
			for (int feat = 0; feat < feature_settings.length; feat++)
				if (feature_settings[feat].extract_feature)
					count++;
			if (count == 0)
				JOptionPane.showMessageDialog(null, "No features are marked to be extracted in the Features Panel.", "ERROR", JOptionPane.ERROR_MESSAGE);
			else
				extractFeaturesFromFiles();
		}

		// React to the delete_saved_features_button
		else if (event.getSource().equals(delete_saved_features_button))
		{
			// Remove all extracted recordings from recordings_list
			for (int rec = 0; rec < recordings_list.length; rec++)
				recordings_list[rec].feature_set = null;

			// Note that no features are stored
			features_extracted = false;
			updateStatusBar();
			changeButtonStatus();
		}

		// React to the classify_recordings_button
		if (event.getSource().equals(classify_recordings_button))
			classifyRecordings(recordings_table.getSelectedRows());

		// React to the count_categories_button
		else if (event.getSource().equals(count_categories_button))
			reportCategoryCount();

		// React to the verify_files_button
		else if (event.getSource().equals(verify_files_button))
			verifyFiles();

		// React to the new_recordings_list_button
		else if (event.getSource().equals(new_recordings_list_button))
		{
			int response = JOptionPane.showConfirmDialog(RecordingsPanel.this, "This will " +
				"erase the currently loaded recordings list.\nDo you wish to proceed?", "Warning",
				JOptionPane.OK_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE);
			if (response == JOptionPane.YES_OPTION)
				newRecordingsList();
		}

		// React to the change_recordings_path_button
		else if (event.getSource().equals(change_recordings_path_button))
		{
			// Allow the user to select a directory that should contain the recordings
			// referred to in recordings_list as well as any other recordings that the
			// user may wish to add
			JFileChooser chooser = new JFileChooser();
			chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
			chooser.setCurrentDirectory(new File(parent_window.file_locations.current_directory));
			int browse_return = chooser.showOpenDialog(RecordingsPanel.this);

			// If the user presses Open, then reset the file_path fields for all
			// Recordings in recordings_list based on the selection
			if (browse_return == JFileChooser.APPROVE_OPTION)
			{
				String selection = (chooser.getSelectedFile()).getPath();
				for (int rec = 0; rec < recordings_list.length; rec++)
				{
					String file_name = StringMethods.convertFilePathToFileName(recordings_list[rec].file_path);
					recordings_list[rec].file_path = selection + File.separator + file_name;
				}
			}
		}

		// React to the load_recordings_list_button
		else if (event.getSource().equals(load_recordings_list_button))
		{
			// Give an error if the taxonomy is empty
			if (parent_window.taxonomy_panel.isTreeEmpty())
				JOptionPane.showMessageDialog(null, "The currently loaded taxonomy is empty, so " +
					"recordings may not be loaded.", "ERROR", JOptionPane.ERROR_MESSAGE);
			else
			{
				boolean ok_pressed = parent_window.file_locations.browseRecordingsPath(true);
				if (ok_pressed)
				{
					String recordings_path = parent_window.file_locations.recordings_list_path;
					boolean worked = false;
					worked = parseRecordingsFile(recordings_path);
					if (!worked)
						parent_window.file_locations.setRecordingsPath(""); // reset to blank path if file incorrect
					changeButtonStatus();
				}
			}
		}

		// React to the save_recordings_list_button
		else if (event.getSource().equals(save_recordings_list_button))
			saveRecordingsFile("", true);
		
		// React to the save_recordings_list_as_button
		else if (event.getSource().equals(save_recordings_list_as_button))
			saveRecordingsFile("", false);

		// React to the edit_recording_button
		else if (event.getSource().equals(edit_recording_button))
			editRecordings(recordings_table.getSelectedRows());

		// React to the add_new_recordings_button by adding new recordings to the recording_list
		// based on the chosen files. EditRecordingJFrames are brought up to allow the user
		// to configure the new selections. Only MIDI files that have not already been loaded
		// are displayed. Directories are only displayed if recordings_list is null.
		else if (event.getSource().equals(add_new_recordings_button))
		{
			// Give an error if the taxonomy is empty
			if (parent_window.taxonomy_panel.isTreeEmpty())
				JOptionPane.showMessageDialog(null, "The currently loaded taxonomy is empty, so " +
					"recordings may not be added.", "ERROR", JOptionPane.ERROR_MESSAGE);
			else
			{
				// Set up the browsing dialog
				JFileChooser add_recording_chooser = new JFileChooser();
				add_recording_chooser.setMultiSelectionEnabled(true);
				add_recording_chooser.setApproveButtonText("Add");
				add_recording_chooser.setDialogTitle("Add New Recordings"); 

				// Start in the current directory if no recordings are loaded, start in the current
				// directory otherwise
				String recording_dir = new String();
				if (recordings_list != null)
					recording_dir = recordings_list[0].file_path.substring(0, recordings_list[0].file_path.lastIndexOf(File.separator) + 1);
				else
					recording_dir = parent_window.file_locations.current_directory;
				add_recording_chooser.setCurrentDirectory(new File(recording_dir));

				// Ensure that only MIDI files that have not yet been loaded are displayed
				String[] filename_list;
				if (recordings_list == null)
					filename_list = null;
				else
				{
					filename_list = new String[recordings_list.length];
					for (int i = 0; i < recordings_list.length; i++)
						filename_list[i] = recordings_list[i].file_path;
				}
				add_recording_chooser.setFileFilter(new FileFilterRecordings(filename_list));
				add_recording_chooser.setAcceptAllFileFilterUsed(false);
				
				// Bring up edit dialog box for any selected files
				int browse_return = add_recording_chooser.showOpenDialog(RecordingsPanel.this);
				if (browse_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
				{
					File[] selections = add_recording_chooser.getSelectedFiles();
					EditRecordingJFrame[] new_editors = new EditRecordingJFrame[selections.length];
					int[] windows_to_cycle_through = new int[1];
					windows_to_cycle_through[0] = selections.length;
					for (int i = 0; i < selections.length; i++)
					{
						String[] unknown_leaf_id = new String[1];
						String[] unknown_ancestor_id = new String[1];
						unknown_leaf_id[0] = "Unknown";
						unknown_ancestor_id[0] = "Unknown";
						Recording new_recording = new Recording("Untitled", "Anonymous", selections[i].getPath(), 
																unknown_leaf_id, unknown_ancestor_id, 
																false, false, true, null);
						new_editors[i] = new EditRecordingJFrame(this, new_recording, windows_to_cycle_through, true);
						new_editors[i].pack();
						new_editors[i].show();
					}
				}
			}
		}

		// React to the delete_recordings_button
		else if (event.getSource().equals(delete_recordings_button))
		{
			int selections[] = recordings_table.getSelectedRows();
			if (selections.length == recordings_list.length)
				newRecordingsList();
			else
			{
				// Copy arrays
				Recording[] temp_recordings = new Recording[recordings_list.length];
				for (int i = 0; i < recordings_list.length; i++)
					temp_recordings[i] = recordings_list[i];
				int[] temp_ordering = new int[list_ordering.length];
				for (int i = 0; i < list_ordering.length; i++)
					temp_ordering[i] = list_ordering[i];
				
				// Mark deleted entries
				for (int i = 0; i < temp_recordings.length; i++)
				{
					for (int j = 0; j < selections.length; j++)
					{
						if (i == selections[j])
						{
							temp_recordings[i] = null;
							j = selections.length; // exit loop
						}
					}
				}

				// Finalize
				recordings_list = new Recording[recordings_list.length - selections.length];
				list_ordering = new int[list_ordering.length - selections.length];
				int new_count = 0;
				for (int i = 0; i < temp_recordings.length; i++)
				{
					if (temp_recordings[i] != null)
					{
						recordings_list[new_count] = temp_recordings[i];
						list_ordering[new_count] = temp_ordering[i];
						new_count++;
					}
				}

				// Display changes
				displayRecordingsTable();
			}
		}
	}


	/**
	 * Causes edit dialog to appear if double click on a recording.
	 * Also enables sorting by clicking on a column heading.
	 */
	public void addRecordingsTableMouseListener()
	{
		recordings_table.getTableHeader().addMouseListener(new MouseAdapter()
		{
			public void mouseClicked(MouseEvent event)
			{
				// Sort and cause the display to show sorted data
				int column_clicked = recordings_table.columnAtPoint(event.getPoint());
				String[][] sorted_table = Sorter.sortTable(convertRecordingsToStringTable(), column_clicked, list_ordering);
				recordings_table_model.fillTable(sorted_table);

				// Sort the recording_list
				Recording[] recordings_copy = new Recording[recordings_list.length];
				for (int i = 0; i < recordings_list.length; i++)
					recordings_copy[i] = recordings_list[i];
				for (int i = 0; i < recordings_list.length; i++)
					recordings_list[i] = recordings_copy[list_ordering[i]];
			}
		});

		recordings_table.addMouseListener(new MouseAdapter()
		{
			public void mouseClicked(MouseEvent event)
			{
				if (event.getClickCount() == 2)
				{
					int[] row_clicked = new int[1];
					row_clicked[0] = recordings_table.rowAtPoint(event.getPoint());
					editRecordings(row_clicked);
				}
			}
		});
	}


	/**
	 * Causes a dialog box to appear listing the values of a multi-dimensional feature
	 * of a recording if the user double clicks on a cell shown in recordings_table marked MD.
	 */
	public void addFeatureValueTableMouseListener()
	{
		features_value_table.addMouseListener(new MouseAdapter()
		{
			public void mouseClicked(MouseEvent event)
			{
				if (event.getClickCount() == 2)
				{
					int row = features_value_table.rowAtPoint(event.getPoint());
					int column = features_value_table.columnAtPoint(event.getPoint());

					// Make sure that it's a mult-dimensional feature
					if (((String) features_value_table.getValueAt(row, column)).equals("MD"))
					{
						FeatureMultiDimensional feature = (FeatureMultiDimensional) recordings_list[row].feature_set.features[column - 2];
						String dialog_title =  feature.getName() + " values for " + StringMethods.convertFilePathToFileName(recordings_list[row].file_path);
						String dialog_info = new String("DIMENSION\tVALUE\n");
						for (int i = 0; i < feature.getValue().length; i++)
							dialog_info += String.valueOf(i) + "\t" + String.valueOf(feature.getValue()[i]) + "\n";
						InformationDialog multi_dimensional_feature_reporter = new InformationDialog(dialog_title, dialog_info);
					}
				}
			}
		});
	}


	/**
	 * Saves a <i>recordings_file</i> XML file with the contents specified
	 * in the <code>recordings_list</code> field and the comments specified in
	 * the <code>comments</code> parameter.
	 *
	 * <p>If the <code>is_basic_save</code> parameter is true, then the
	 * file is saved at the path specified in the <code>ConfigureFileLocationsJFrame</code>
	 * object's <code>recordings_list_path</code> field (automatically overwriting
	 * any existing files with the same path). If the <code>is_basic_save</code>
	 * parameter is false, then the file is saved at a location chosen by the user
	 * with a <code>JFileChooser</code> dialog box.
	 *
	 * <p>Also, if feature values are available for the currently loaded recordings then,
	 * calls a private method to save the feature values in a separate file.
	 * These are stored in a feature_values_file XML file in the same directory as the
	 * currently loaded recordings_file. The name of the file is the same as that of the
	 * recordings_file, but with "_feature_values" appended to the end of the file name.
	 * Features are only stored for recordings that have had features extracted from them.
	 * No checking is performed to see if all recordings have the same features extracted
	 * from them.
	 *
	 * @param	comments		Any comments to be saved inside the comments element of the XML file.
	 * @param	is_basic_save	Whether this is a save (true) or save as (false) operation.
	 * @see		ConfigureFileLocationsJFrame
	 */
	public void saveRecordingsFile(String comments, boolean is_basic_save)
	{
		if (recordings_list != null)
		{
			// Update any changes to the roles column that may have been entered by the user
			updateRoles();

			// Get File to save to
			File to_save_to = null;
			if (is_basic_save && !parent_window.file_locations.recordings_list_path.equals("")) // // perform a basic save operation
				to_save_to = parent_window.file_locations.getFile(parent_window.file_locations.recordings_list_path, true);
			else // perform a save as operation
			{
				if (parent_window.file_locations.recordings_list_path.equals(""))
					JOptionPane.showMessageDialog(null, "No file name is specified for saving the recordings list.\nPlease enter one in the following dialog box.", "NOTE", JOptionPane.INFORMATION_MESSAGE);

				to_save_to = parent_window.file_locations.saveFileExternally(false, "recordings_file");
			}

			// Write the file to disk
			if (to_save_to != null)
			{
				try
				{
					// Prepare stream writer
					FileOutputStream to = new FileOutputStream(to_save_to);
					DataOutputStream writer = new DataOutputStream(to);

					// Write the header and the first element of the XML file
					String header = new String
					(
						"<?xml version=\"1.0\"?>\n" +
						"<!DOCTYPE recordings_file [\n" +
						"   <!ELEMENT recordings_file (comments, file_path, recording+)>\n" +
						"   <!ELEMENT comments (#PCDATA)>\n" +
						"   <!ELEMENT file_path (#PCDATA)>\n" +
						"   <!ELEMENT recording (filename, role, model_genres, title, artist)>\n" +
						"   <!ELEMENT filename (#PCDATA)>\n" +
						"   <!ELEMENT role (#PCDATA)>\n" +
						"   <!ELEMENT model_genres (genre*)>\n" +
						"   <!ELEMENT genre (#PCDATA)>\n" +
						"   <!ELEMENT title (#PCDATA)>\n" +
						"   <!ELEMENT artist (#PCDATA)>\n" +
						"]>\n\n" +
						"<recordings_file>\n\n" +
						"   <comments>" + StringMethods.giveStringXMLEscapeCharacters(comments) + "</comments>\n\n" + 	
						"   <file_path>" + recordings_list[0].file_path.substring(0, recordings_list[0].file_path.lastIndexOf(File.separator)) + "</file_path>\n\n"
					);
					writer.writeBytes(header);
					
					// Write the XML code to represent the contents of the recordings_list field
					for (int i = 0; i < recordings_list.length; i++)
					{
						writer.writeBytes("   <recording>\n");
						String filename = StringMethods.giveStringXMLEscapeCharacters(StringMethods.convertFilePathToFileName(recordings_list[i].file_path));
						writer.writeBytes("      <filename>" + filename + "</filename>\n");
						String role;
						if (recordings_list[i].use_for_training)
							role = new String("training");		
						else if (recordings_list[i].use_for_testing)
							role = new String("testing");		
						else
							role = new String("classification");	
						writer.writeBytes("      <role>" + role + "</role>\n");
						writer.writeBytes("      <model_genres>\n");
						for (int j = 0; j < recordings_list[i].leaf_categories.length; j++)
							writer.writeBytes("         <genre>" + StringMethods.giveStringXMLEscapeCharacters(recordings_list[i].leaf_categories[j]) + "</genre>\n");
						writer.writeBytes("      </model_genres>\n");
						writer.writeBytes("      <title>" + StringMethods.giveStringXMLEscapeCharacters(recordings_list[i].title) + "</title>\n");
						writer.writeBytes("      <artist>" + StringMethods.giveStringXMLEscapeCharacters(recordings_list[i].artist) + "</artist>\n");
						writer.writeBytes("   </recording>\n\n");
					}
					writer.writeBytes("</recordings_file>");
					
					// Close the output stream
					writer.close();

					// Save feature values. Automatically overwrite existing file if present.
					saveFeatureValues();
				}
				catch (Exception e)
				{
					JOptionPane.showMessageDialog(null, ("Unable to write file " + to_save_to.getName() + "."), "ERROR", JOptionPane.ERROR_MESSAGE);
					// e.printStackTrace();
				}
			}
		}
	}


	/**
	 * Returns the recordings marked with a <i>role</i> of <i>classification</i>.
	 * Returns null if there are no recordings marked in this fashion.
	 */
	public Recording[] getRecordingsMarkedForClassification()
	{
		// Return null if recordings_list is empty
		if (recordings_list == null)
			return null;
		
		// Update any changes to the rolls that may have been added by the user
		updateRoles();

		// Find the number of recordings marked for classification
		int count = 0;
		for (int rec = 0; rec < recordings_list.length; rec++)
			if (recordings_list[rec].use_for_classification)
				count++;
		
		// Return null if no marked recordings.
		if (count == 0)
			return null;

		// Find the marked recordings
		Recording[] to_classify = new Recording[count];
		count = 0;
		for (int rec = 0; rec < recordings_list.length; rec++)
			if (recordings_list[rec].use_for_classification)
			{
				to_classify[count] = recordings_list[rec];
				count++;
			}

		// Return the result
		return to_classify;
	}


	/**
	 * Returns the recordings marked with a <i>role</i> of <i>testing</i>.
	 * Returns null if there are no recordings marked in this fashion.
	 */
	public Recording[] getRecordingsMarkedForTesting()
	{
		// Return null if recordings_list is empty
		if (recordings_list == null)
			return null;
		
		// Update any changes to the rolls that may have been added by the user
		updateRoles();

		// Find the number of recordings marked for testing
		int count = 0;
		for (int rec = 0; rec < recordings_list.length; rec++)
			if (recordings_list[rec].use_for_testing)
				count++;
		
		// Return null if no marked recordings.
		if (count == 0)
			return null;

		// Find the marked recordings
		Recording[] to_test = new Recording[count];
		count = 0;
		for (int rec = 0; rec < recordings_list.length; rec++)
			if (recordings_list[rec].use_for_testing)
			{
				to_test[count] = recordings_list[rec];
				count++;
			}

		// Return the result
		return to_test;
	}


	/**
	 *  Read the roles from the table and update training_list.
	 */
	public void updateRoles()
	{
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			String new_role = (String) recordings_table.getValueAt(rec, 5);

			if (new_role.equals("training"))
			{
				recordings_list[rec].use_for_training = true;
				recordings_list[rec].use_for_testing = false;
				recordings_list[rec].use_for_classification = false;
			}
			if (new_role.equals("testing"))
			{
				recordings_list[rec].use_for_training = false;
				recordings_list[rec].use_for_testing = true;
				recordings_list[rec].use_for_classification = false;
			}
			if (new_role.equals("classification"))
			{
				recordings_list[rec].use_for_training = false;
				recordings_list[rec].use_for_testing = false;
				recordings_list[rec].use_for_classification = true;
			}
		}
	}


	/**
	 * Updates the table and the status bar to reflect the contents of recordings_list.
	 * Updates which buttons are enabled.
	 */
	public void displayRecordingsTable()
	{
		// Remove anything on the left side of the panel
		if (recordings_display_panel != null)
			remove(recordings_display_panel);
		if (feature_value_diplay_scroll_pane != null)
			remove(feature_value_diplay_scroll_pane);
		if (scaled_feature_value_diplay_scroll_pane != null)
			remove(scaled_feature_value_diplay_scroll_pane);

		Object[] column_names = {new String("Title"),
		                         new String("Artist"),
		                         new String("File Name"),
		                         new String("Leaf Categories"),
		                         new String("Ancestor Categories"),
		                         new String("Role")};

		recordings_table_model = new RecordingsTableModel(column_names, 0);
		recordings_table = new JTable(recordings_table_model);
		addRecordingsTableMouseListener();

		if (recordings_list != null)
		{
			String[] role_options = {"training", "testing", "classification"};
			JComboBox role_selector = new JComboBox(role_options);
			recordings_table.getColumnModel().getColumn(5).setCellEditor(new DefaultCellEditor(role_selector));
			
			recordings_table_model.fillTable(convertRecordingsToStringTable());

			recordings_display_scroll_pane = new JScrollPane(recordings_table);
			recordings_display_panel = new JPanel(new GridLayout(1, 1));
			recordings_display_panel.add(recordings_display_scroll_pane);
			add(recordings_display_panel, BorderLayout.CENTER);
		}

		updateStatusBar();
		
		display_recordings_button.setEnabled(false);
		new_recordings_list_button.setEnabled(true);
		load_recordings_list_button.setEnabled(true);
		add_new_recordings_button.setEnabled(true);
		changeButtonStatus();

		recordings_table_model.fireTableDataChanged();
		repaint();
		parent_window.repaint();
	}


	/**
	 * Updates the recordings_list and the display to reflect the addition
	 * of the new recording specified in the parameter of this method.
	 *
	 * @param	new_recording	The new recording to add to recordings_list.
	 * @see		Recording
	 */
	public void updateAfterAddedRecording(Recording new_recording)
	{
		if (recordings_list == null)
		{
			changeButtonStatus();
			recordings_list = new Recording[1];
			recordings_list[0] = new_recording;
			list_ordering = new int[1];
			list_ordering[0] = 0;
			features_extracted = false;
		}
		else
		{
			Recording[] temp = new Recording[recordings_list.length + 1];
			for (int i = 0; i < recordings_list.length; i++)
				temp[i] = recordings_list[i];
			temp[recordings_list.length] = new_recording;
			recordings_list = temp;
			int[] temp_ordering = new int[list_ordering.length + 1];
			for (int i = 0; i < list_ordering.length; i++)
				temp_ordering[i] = list_ordering[i];
			temp_ordering[list_ordering.length] = list_ordering.length;
			list_ordering = temp_ordering;
		}
	}


	/**
	 * Make the results of parsing be implemented. Triggered by the user pressing the
	 * <i>Accept</i> button on a <code>ProgressBarDialog</code> window.
	 *
	 * <p>Cycles through all recordings that were not marked to have features extracted
	 * and deletes any unrequired features in them. Also deletes any recordings from
	 * the <code>recordings_list</code> field that could not be extracted due to an error.
	 * and updates the display. Also stores the new features in <code>recordings_list</code>.
	 */
	public void acceptProcessingResults()
	{
		// Cycle through all recordings that were not marked to have features extracted
		// and delete any unrequired features in them
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			// Only check if features need to be removed if features were marked not
			// to be extracted for this file
			if (!recordings_to_extract[rec])
			{
				// The reduced set of feature values
				Feature[] reduced_features = new Feature[required_feature_names.length];
				for (int i = 0; i < reduced_features.length; i++)
					reduced_features[i] = null;

				// Cycle through and search for all required features, updating
				// reduced_features as go
				for (int feat = 0; feat < reduced_features.length; feat++)
				{
					// Features already extracted in the current recording
					Feature[] rec_features = recordings_list[rec].feature_set.features;
					
					// Find the correct feature
					for (int rec_feat = 0; rec_feat < rec_features.length; rec_feat++)
					{
						if (rec_features[rec_feat].getName().equals( required_feature_names[feat] ))
						{
							reduced_features[feat] = rec_features[rec_feat];
							rec_feat = rec_features.length + 1; // exit loop
						}
					}
				}

				// Set the Recording to only have the reduced set of feature values
				recordings_list[rec].feature_set.features = reduced_features;
			}
		}

		// Store parsed features in recordings_list and mark recordings where errors occured during
		// feature extraction for future deletion by setting recordings_to_extract to null for their
		// entry
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			if (recordings_to_extract[rec])
			{
				// Mark recordings where errors occured during feature extraction for future deletion
				if (monitor_extractor.getParsedFeatures()[rec] == null)
					recordings_list[rec] = null;

				// Copy over succesfully extracted features to recordings_list
				else
					recordings_list[rec].feature_set = monitor_extractor.getParsedFeatures()[rec];
			}
		}

		// Since some elements of recordings_list may have been marked as null because they could
		// not be properly parsed, remake recording_list so that these elements are deleted
		// and update the window.
		int number_to_delete = 0;
		for (int i = 0; i < recordings_list.length; i++)
			if (recordings_list[i] == null)
				number_to_delete++;
		if (number_to_delete != 0)
		{
			Recording[] old_recordings_list = recordings_list;
			recordings_list = new Recording[old_recordings_list.length - number_to_delete];
			int recordings_list_index = 0;
			for (int i = 0; i < old_recordings_list.length; i++)
			{
				if (old_recordings_list[i] != null)
				{
					recordings_list[recordings_list_index] = old_recordings_list[i];
					recordings_list_index++;
				}
			}
			displayRecordingsTable();
		}

		// Note that features have been extracted, if they have, and update the status bar
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			if (recordings_list[rec].feature_set != null)
			{
				features_extracted = true;
				rec = recordings_list.length + 1; // exit loop
			}
		}
		updateStatusBar();

		// Ensure that the button status is updated
		changeButtonStatus();
	}


	/**
	 * Reset everything and clear screen and deactivate appropriate buttons.
	 * Reset the stored recording list file path to blank.
	 */
	public void newRecordingsList()
	{
		parent_window.file_locations.setRecordingsPath("");

		recordings_list = null;
		list_ordering = null;
		features_extracted = false;
		
		recordings_table_model = null;
		recordings_table = null;

		updateStatusBar();
		changeButtonStatus();

		if (recordings_display_panel != null)
			remove(recordings_display_panel);
		repaint();
		parent_window.repaint();
	}


	/**
	 * Saves the currently stored feature values for the currently loaded set or recordings.
	 * These are stored in a feature_values_file XML file in the same directory as the
	 * currently loaded recordings_file. The name of the file is the same as that of the
	 * recordings_file, but with "_feature_values" appended to the end of the file name.
	 * Features are only stored for recordings that have had features extracted from them.
	 * No checking is performed to see if all recordings have the same features extracted
	 * from them.
	 */
	public void saveFeatureValues()
	{

		// Get File to save to
		File to_save_to = null;
		String recordings_list_path = parent_window.file_locations.recordings_list_path;
		String feature_values_path = recordings_list_path.substring(0, recordings_list_path.length() - 4); // strip of .xml of recordings_list_path
		feature_values_path += "_feature_values.xml";
		to_save_to = parent_window.file_locations.getFile(feature_values_path, true);

		// Write the file to disk
		if (to_save_to != null)
		{
			try
			{
				// Prepare stream writer
				FileOutputStream to = new FileOutputStream(to_save_to);
				DataOutputStream writer = new DataOutputStream(to);

				// Write the header and the first element of the XML file
				String header = new String
				(
					"<?xml version=\"1.0\"?>\n" +
					"<!DOCTYPE feature_values_file [\n" +
					"   <!ELEMENT feature_values_file (comments, recording*)>\n" +
					"   <!ELEMENT comments (#PCDATA)>\n" +
					"   <!ELEMENT recording (file_path, feature+)>\n" +
					"   <!ELEMENT file_path (#PCDATA)>\n" +
					"   <!ELEMENT feature (name, dimensions, value+)>\n" +
					"   <!ELEMENT name (#PCDATA)>\n" +
					"   <!ELEMENT dimensions (#PCDATA)>\n" +
					"   <!ELEMENT value (#PCDATA)>\n" +
					"]>\n\n" +
					"<feature_values_file>\n\n" +
					"   <comments></comments>\n\n"
				);
				writer.writeBytes(header);

				// Write the XML code to represent the contents of the feature_values_file field
				for (int rec = 0; rec < recordings_list.length; rec++)
				{
					if (recordings_list[rec].feature_set != null )
					{
						writer.writeBytes("   <recording>\n");
						writer.writeBytes("      <file_path>" + StringMethods.giveStringXMLEscapeCharacters(recordings_list[rec].file_path) + "</file_path>\n");

						for (int feat = 0; feat < recordings_list[rec].feature_set.features.length; feat++)
						{
							writer.writeBytes("      <feature>\n");
							writer.writeBytes("         <name>" + recordings_list[rec].feature_set.features[feat].getName() + "</name>\n");
							if (recordings_list[rec].feature_set.features[feat] instanceof FeatureOneDimensional)
							{
								writer.writeBytes("         <dimensions>1</dimensions>\n");
								writer.writeBytes("         <value>" + ((FeatureOneDimensional) recordings_list[rec].feature_set.features[feat]).getValue() + "</value>\n");
							}
							else if (recordings_list[rec].feature_set.features[feat] instanceof FeatureMultiDimensional)
							{
								int number_dimensions = ((FeatureMultiDimensional) recordings_list[rec].feature_set.features[feat]).getValue().length;
								writer.writeBytes("         <dimensions>" + number_dimensions + "</dimensions>\n");
								for (int val = 0; val < number_dimensions; val++)
									writer.writeBytes("         <value>" + ((FeatureMultiDimensional) recordings_list[rec].feature_set.features[feat]).getValue()[val] + "</value>\n");
							}
							writer.writeBytes("      </feature>\n");
						}
						writer.writeBytes("   </recording>\n\n");
					}
				}
				writer.writeBytes("</feature_values_file>");

				// Close the output stream
				writer.close();
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(null, ("Unable to write file " + to_save_to.getName() + "."), "ERROR", JOptionPane.ERROR_MESSAGE);
				// e.printStackTrace();
			}
		}
	}

	
	/* PRIVATE METHODS *********************************************************/


	// Causes the status bar to display the number of recordings currently
	// loaded as well as whether or not features have been extracted.
	private void updateStatusBar()
	{
		String status_message = new String("");
		if (recordings_list == null)
			status_message += "No recordings currently loaded.";
		else
		{
			status_message += new String(recordings_list.length + " recordings are currently loaded.");
			if (features_extracted)
				status_message += " Features have been extracted.";
			else
				status_message += " No features have been extracted.";
		}

		status_bar.setText(status_message);
	}

	
	// Enable or disable the buttons that require at least one recording to
	// be loaded based on whether at least one recording is loaded and whether
	// or not features have been extracted.
	private void changeButtonStatus()
	{
		boolean enabled_value = false;
		if (recordings_list != null)
			if (recordings_list.length > 0)
				enabled_value = true;
			
		if (enabled_value && features_extracted)
		{
			display_feature_values_button.setEnabled(true);
			display_scaled_feature_values_button.setEnabled(true);
			delete_saved_features_button.setEnabled(true);
			classify_recordings_button.setEnabled(true);
		}
		else
		{
			display_feature_values_button.setEnabled(false);
			display_scaled_feature_values_button.setEnabled(false);
			delete_saved_features_button.setEnabled(false);
			classify_recordings_button.setEnabled(false);
		}
		if (enabled_value)
			verify_files_button.setEnabled(true);
		else
			verify_files_button.setEnabled(false);
		extract_features_button.setEnabled(enabled_value);
		change_recordings_path_button.setEnabled(enabled_value);
		save_recordings_list_button.setEnabled(enabled_value);
		save_recordings_list_as_button.setEnabled(enabled_value);
		edit_recording_button.setEnabled(enabled_value);
		delete_recordings_button.setEnabled(enabled_value);
		count_categories_button.setEnabled(enabled_value);
	}


	// Classifies the highlighted recordings and switches to the ClassificationPanel to show the
	// results. Replaces any stored test recordings on the ClassificationPanel.
	private void classifyRecordings(int[] rows)
	{
		try
		{
			// Give an error if the taxonomy is empty
			if (parent_window.taxonomy_panel.isTreeEmpty())
				throw new Exception("The currently loaded taxonomy is empty, so recordings may not be classified.");

			// Give an error if no rows are selected
			if (rows == null)
				throw new Exception("No recordings specified for classification.");

			// Give error if features not extracted yet
			if (!features_extracted)
				throw new Exception( "Cannot classify recordings because features have not\n" +
				                     "been extracted from them yet." );
			for (int rec = 0; rec < rows.length; rec++)
				if (recordings_list[rows[rec]].feature_set == null)
					throw new Exception( "Cannot classify recordings because features have not\n" +
					                     "been extracted from " + recordings_list[rows[rec]].title + " yet." );

			// Collect the recordings to classify
			Recording[] recordings_to_classify = new Recording[rows.length];
			for (int rec = 0; rec < recordings_to_classify.length; rec++)
				recordings_to_classify[rec] = recordings_list[rows[rec]];

			// Classify the recordings
			parent_window.classification_panel.classifyRecordings(recordings_to_classify);

			// Set view to classifier pane
			parent_window.focusOnClassifierPane();
			
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}


	// Parses the specified file and, if it is an appropriate recordings_file XML file,
	// stores it in recordings_list. Otherwise gives an error message and
	// creates and displays a new blank tree. Also gives an error message if the file
	// contains leaf categories that are not in the currently loaded taxonomy or if a
	// file_name field is blank. Replaces blank titles and artists with "Unknown".
	// Also populates the ancestor categories. Displays the parsed results. Returns true
	// if the end of the procedureis reached without an exception being thrown.
	// Also checks to see if feature values are available, parses them if they are
	// and updates features_extracted.
	//
	// Note that if the extension of the file is not XML or xml, then it is assumed that
	// the file is a tab delimited text file of the MIREX 2005 format, and is parsed as
	// such. This functionality is intended primarily for use with the command line
	// use of Bodhidharma.
	private boolean parseRecordingsFile(String file_path)
	{
		boolean succesful = false;
		
		try
		{
			// Check if the file is an XML file by looking at its extension
			boolean using_xml_file = true;
			String extension = StringMethods.getExtension(file_path);
			if (extension == null)
				using_xml_file = false;
			else if (!extension.equals(".xml") && !extension.equals(".XML"))
				using_xml_file = false;

			// Parse the contents of the recordings_file into recordings_list from an XML file
			if (using_xml_file)
			{
				Object[] results = (Object[]) XMLDocumentParser.parseXMLDocument(file_path, "recordings_file");
				recordings_list = new Recording[results.length];
				for (int i = 0; i < recordings_list.length; i++)
					recordings_list[i] = (Recording) results[i];
			}

			// Parse the contents of the recordings_file into recordings_list from a text file
			else
			{
				String[][] file_contents = FileMethods.tokenizeTabDelFile(file_path);
				recordings_list = new Recording[file_contents.length];
				for (int rec = 0; rec < recordings_list.length; rec++)
				{
					recordings_list[rec] = new Recording();
					recordings_list[rec].title = file_contents[rec][0];
					recordings_list[rec].file_path = file_contents[rec][0];
					recordings_list[rec].use_for_testing = true;
					if (file_contents[rec].length == 1)
					{
						String temp = file_contents[rec][0];
						file_contents[rec] = new String[2];
						file_contents[rec][0] = temp;
						file_contents[rec][1] = "Unknown";
					}
					String[] leaf_categories = new String[file_contents[rec].length - 1];
					for (int cat = 0; cat < file_contents[rec].length - 1; cat++)
						leaf_categories[cat] = file_contents[rec][cat + 1];
					recordings_list[rec].leaf_categories = leaf_categories;				
				}
			}

			// Give any empty titles or artist names designations of "Unknown"
			// Throw an exception if the file_path is empty.
			for (int i = 0; i < recordings_list.length; i++)
			{
				if (recordings_list[i].title == null)
					recordings_list[i].title = "Unknown";
				if (recordings_list[i].artist == null)
					recordings_list[i].artist = "Unknown";
				if (recordings_list[i].file_path == null)
					throw new Exception("The file " + file_path + " contains a recording " +
										"entitled " + recordings_list[i].title +
										". This recording does not have a file associated " +
										"with it.\n\nPlease correct the recordings" +
										" file. An empty list of recordings will now be generated.");
			}

			// Acquire all of the possible leaf categories and their ancestors in the current taxonomy
			String[][] leaves_and_ancestors = parent_window.taxonomy_panel.getLeafsAndTheirParents();

			// Make sure that all of the leaf categories found in the recordings are in the current taxonomy
			// Thow an exception if they are not
			for (int i = 0; i < recordings_list.length; i++)
			{
				if (!recordings_list[i].leaf_categories[0].equals("Unknown"))
				{
					for(int j = 0; j < recordings_list[i].leaf_categories.length; j++)
					{
						int k = 0;
						while (k <= leaves_and_ancestors.length)
						{
							if (k == leaves_and_ancestors.length)
								throw new Exception("The file " + file_path + " contains the leaf " +
													"category " + recordings_list[i].leaf_categories[j] +
													". This category is not a part of the currently " +
													"loaded taxonomy.\n\nPlease correct the recordings" +
													" file or adjust your taxonomy. An empty list of " +
													" recordings will now be generated.");
							else if (recordings_list[i].leaf_categories[j].equals(leaves_and_ancestors[k][0]))
								k = leaves_and_ancestors.length + 1;
							k++;
						}
					}
				}
			}

			// Fill the recordings_list with ancestor genres based on the currently loaded taxonomy
			for (int i = 0; i < recordings_list.length; i++)
			{
				if (recordings_list[i].leaf_categories[0].equals("Unknown"))
				{
					String[] unknown_id = new String[1];
					unknown_id[0] = "Unknown";
					recordings_list[i].ancestor_categories = unknown_id;
				}
				else
				{
					int j = 0;
					while(j < recordings_list[i].leaf_categories.length)
					{
						int k = 0;
						while(k < leaves_and_ancestors.length)
						{
							if (recordings_list[i].leaf_categories[j].equals(leaves_and_ancestors[k][0]))
							{
								String[] ancestors = new String[leaves_and_ancestors[k].length - 1];
								for (int m = 0; m < ancestors.length; m++)
									ancestors[m] = leaves_and_ancestors[k][m + 1];

								
								// To ensure that ancestor categories aren't overwritten
								if (recordings_list[i].ancestor_categories != null)
								{
									// Get the ancestors already found
									String[] existing_ancestors = recordings_list[i].ancestor_categories;

									// Remove any categories from ancestors that are already in
									// the listed ancestors for recording i
									int repetition_count = 0;
									for (int a = 0; a < ancestors.length; a++)
									{
										for (int b = 0; b < existing_ancestors.length; b++)
										{
											if (ancestors[a] != null)
											{
											if (ancestors[a].equals(existing_ancestors[b]))
												{
												repetition_count++;
												ancestors[a] = null;
												}
											}
										}
									}
									if (repetition_count != 0)
									{
										String[] temp = new String[ancestors.length];
										for (int a = 0; a < ancestors.length; a++)
											temp[a] = ancestors[a];
										ancestors = new String[ancestors.length - repetition_count];
										int ancestors_index = 0;
										for (int a = 0; a < temp.length; a++)
										{
											if (temp[a] != null)
											{
												ancestors[ancestors_index] = temp[a];
												ancestors_index++;
											}
										}
									}

									// Combine newly found ancestors with previously found ancestors
									String[] combined_ancestors = new String[existing_ancestors.length + ancestors.length];
									for (int m = 0; m < existing_ancestors.length; m++)
										combined_ancestors[m] = existing_ancestors[m];
									for (int m = existing_ancestors.length; m < existing_ancestors.length + ancestors.length; m++)
										combined_ancestors[m] = ancestors[m - existing_ancestors.length];
									recordings_list[i].ancestor_categories = combined_ancestors;
								}
								else
									recordings_list[i].ancestor_categories = ancestors;
							}
							k++;
						}
						j++;
					}
				}
			}

			// Initialize the ordering of the list
			list_ordering = new int[recordings_list.length];
			for (int i = 0; i < recordings_list.length; i++)
				list_ordering[i] = i;

			// Check to see if feature values are available, parse them if they are
			// and update features_extracted. Display an error message if appropriate.
			try
			{
				// Parse the dummy Recording objects that contain feature values for recordings_list
				String values_file_path = file_path.substring(0, file_path.length() - 4) +  "_feature_values.xml";
				Object[] temp = (Object[]) XMLDocumentParser.parseXMLDocument(values_file_path, "feature_values_file");
				Recording[] feature_value_holders = new Recording[temp.length];
				for (int i = 0; i < feature_value_holders.length; i++)
					feature_value_holders[i] = (Recording) temp[i];

				// Add the features into the appropriate entries of recordings_list
				boolean at_least_one_value_found = false;
				for (int i = 0; i < recordings_list.length; i++)
				{
					for (int j = 0; j < feature_value_holders.length; j++)
					{
						if ((recordings_list[i].file_path).equals(feature_value_holders[j].file_path))
						{
							recordings_list[i].feature_set = new FeatureSet(feature_value_holders[j].feature_set.features);
							at_least_one_value_found = true;
							j = feature_value_holders.length + 1; // exit the loop
						}
					}
				}

				// Note that extracted features are available
				if (at_least_one_value_found)
					features_extracted = true;
				else
					features_extracted = false;
			}
			catch (Exception e) // do not throw exception if problem is only that file has not been found
			{
				if ( !(e.getMessage().substring(0, 18).equals("The specified path")) )
					JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			}

			// Display the parsed results		
			displayRecordingsTable();

			// Note that parsing was succesful
			succesful = true;
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			newRecordingsList();
		}

		return succesful;
	}


	// Opens an EditRecordings JFrame for the recordings ferfered to in rows
	private void editRecordings(int[] rows)
	{
		// Give an error if the taxonomy is empty
		if (parent_window.taxonomy_panel.isTreeEmpty())
			JOptionPane.showMessageDialog(null, "The currently loaded taxonomy is empty, so " +
				"recordings may not be edited.", "ERROR", JOptionPane.ERROR_MESSAGE);
		else
		{
			EditRecordingJFrame[] recording_editors = new EditRecordingJFrame[rows.length];
			int[] windows_to_cycle_through = new int[1]; // keeps track of when to reallow access to this window
			windows_to_cycle_through[0] = rows.length;
			for (int i = 0; i < rows.length; i++)
			{
				recording_editors[i] = new EditRecordingJFrame(this, recordings_list[rows[i]], windows_to_cycle_through, false);
				recording_editors[i].pack();
				recording_editors[i].show();
			}
		}
	}


	// Parses the MIDI files present in recordings_list and extracts the features
	// that have the extract_feature field set to true in the FeatureSettings objects
	// held in the feature_settings field of the FeaturesPanel. Results are not actually
	// stored in recordings_list by this method, however (see acceptProcessingResults
	// method for how this is done).
	// Features are only extracted for recordings that have not already had all of the
	// required features extracted. 
	private void extractFeaturesFromFiles()
	{
		// Indicates which recordings need to have their features updated
		recordings_to_extract = new boolean[recordings_list.length];

		// Contains which features should be extracted
		FeatureSettings[] feature_settings = parent_window.features_panel.getFeatureSettings();

		// Construct array holding the names of the set of features to be extracted
		LinkedList required_feature_names_list = new LinkedList();
		for (int feat = 0; feat < feature_settings.length; feat++)
			if (feature_settings[feat].extract_feature)
				required_feature_names_list.add( feature_settings[feat].getFeature().getName() );
		Object[] required_feature_name_objects = required_feature_names_list.toArray();
		required_feature_names = new String[required_feature_name_objects.length];
		for (int i = 0; i < required_feature_names.length; i++)
			required_feature_names[i] = (String) required_feature_name_objects[i];

		// Check all the recordings in recordings_list to see which have all of the
		// required features extracted already. Mark recordings_to_extract as true
		// for all recordings that have one or more missing features.
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			// Automatically set a recording to have features extracted if none have
			// been extracted from it yet
			if (recordings_list[rec].feature_set == null)
				recordings_to_extract[rec] = true;

			// For recordings that have had some features extracted, make sure that
			// all required recordings are present
			else
			{
				// Whether or not all required features have been found
				boolean all_there = true;
				
				// Cycle through all required features
				for (int feat = 0; feat < required_feature_names.length; feat++)
				{
					// Features already extracted in the current recording
					Feature[] rec_features = recordings_list[rec].feature_set.features;

					// Whether or not the required feature has been found
					boolean found = false;

					// Check if the feature has already been extracted
					for (int rec_feat = 0; rec_feat < rec_features.length; rec_feat++)
					{
						// If required feature has already been extracted
						if (rec_features[rec_feat].getName().equals( required_feature_names[feat] ))
						{
							found = true;
							rec_feat = rec_features.length + 1; // exit loop
						}
					}

					// If at least one required feature has not already been extracted
					if (!found)
					{
						all_there = false;
						feat = feature_settings.length + 1; // exit loop
					}
				}

				// Set the recording to be parsed if one or more required features were
				// not found
				if (!all_there)
					recordings_to_extract[rec] = true;
				else
					recordings_to_extract[rec] = false;
			}
		}

		// Find the number of recordings that need to be parsed
		int number_to_parse = 0;
		for (int i = 0; i < recordings_to_extract.length; i++)
			if (recordings_to_extract[i])
				number_to_parse++;

		// Only process if some features need to be extracted
		if (number_to_parse == 0)
			JOptionPane.showMessageDialog(null, "All features marked for extraction have already been extracted from the loaded recordings.", "NOTE", JOptionPane.INFORMATION_MESSAGE);
		else
		{
			// Instantiate the ProgressBarTaskFeatureExtractionMonitor that will perform
			// extraction of features and keep the progress bar dialog updated
			monitor_extractor = new ProgressBarTaskFeatureExtractionMonitor(number_to_parse, recordings_list, recordings_to_extract, required_feature_names);
			
			// Instantiate the ProgressBarDialog that will monitor the extraction of features.
			// This automatically begins feature extraction.
			progress_indicator = new ProgressBarDialog(parent_window, RecordingsPanel.this, monitor_extractor, number_to_parse, parent_window.preferences_panel.getTimeBetweenProgressBarPolls());
		}
	}


	// Constructs and displays a table showing extracted feature values (features_value_table).
	// Multi-dimensional features are marked MD. Double clicking on an MD will bring
	// up a dialog box giving the values of the features. This table canNOT be sorted
	// by clicking on column headers. Recordings that have not had corresponding features
	// extracted are marked with "--".
	private void displayFeatureValues()
	{
		// Remove anything on the left side of the panel
		// Remove anything on the left side of the panel
		if (recordings_display_panel != null)
			remove(recordings_display_panel);
		if (feature_value_diplay_scroll_pane != null)
			remove(feature_value_diplay_scroll_pane);
		if (scaled_feature_value_diplay_scroll_pane != null)
			remove(scaled_feature_value_diplay_scroll_pane);

		// Control which buttons are accessible
		display_recordings_button.setEnabled(true);
		display_feature_values_button.setEnabled(false);
		display_scaled_feature_values_button.setEnabled(true);
		extract_features_button.setEnabled(false);
		classify_recordings_button.setEnabled(false);
		verify_files_button.setEnabled(false);
		new_recordings_list_button.setEnabled(false);
		change_recordings_path_button.setEnabled(false);
		load_recordings_list_button.setEnabled(false);
		save_recordings_list_button.setEnabled(false);
		save_recordings_list_as_button.setEnabled(false);
		edit_recording_button.setEnabled(false);
		add_new_recordings_button.setEnabled(false);
		delete_recordings_button.setEnabled(false);
		delete_saved_features_button.setEnabled(false);
		count_categories_button.setEnabled(false);

		// Find the titles and names of all recordings
		String[][] recording_info = new String[recordings_list.length][2];
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			recording_info[rec][0] = recordings_list[rec].title;
			recording_info[rec][1] = StringMethods.convertFilePathToFileName(recordings_list[rec].file_path);
		}

		// Determine the names of all currently extracted features in one or more
		// recordings and place them in feature_names
		LinkedList feature_names_list = new LinkedList();
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			if (recordings_list[rec].feature_set != null)
			{
				for (int feat = 0; feat < recordings_list[rec].feature_set.features.length; feat++)
				{
					// Add the feature if the list is empty
					if (feature_names_list.size() == 0)
						feature_names_list.add(recordings_list[rec].feature_set.features[feat].getName());

					// See if the list contains the current feature and add it if it
					// does not
					boolean found = false;
					for (int list_entry = 0; list_entry < feature_names_list.size(); list_entry++)
					{
						String current_entry = (String) feature_names_list.get(list_entry);
						if (current_entry.equals(recordings_list[rec].feature_set.features[feat].getName()))
						{
							found = true;
							list_entry = feature_names_list.size() + 1; // exit loop
						}
					}
					if (!found)
						feature_names_list.add(recordings_list[rec].feature_set.features[feat].getName());
				}
			}
		}
		Object[] feature_names = feature_names_list.toArray();

		// Generate String[][] feature_values filled with feature values for each recording.
		// Give value of "--" if feature not extracted for a particular array.
		String[][] feature_values = new String[recordings_list.length][feature_names.length];
		for (int rec = 0; rec < recordings_list.length; rec++)
			for (int feat = 0; feat < feature_names.length; feat++)
				feature_values[rec][feat] = new String("--");
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			for (int feat = 0; feat < feature_names.length; feat++)
			{
				if (recordings_list[rec].feature_set != null)
				{
					for (int i = 0; i < recordings_list[rec].feature_set.features.length; i++)
					{
						if (recordings_list[rec].feature_set.features[feat].getName().equals(feature_names[feat]))
						{
							if (recordings_list[rec].feature_set.features[feat] instanceof FeatureMultiDimensional)
								feature_values[rec][feat] = new String("MD");
							else if (recordings_list[rec].feature_set.features[feat] instanceof FeatureOneDimensional)
								feature_values[rec][feat] = String.valueOf( ((FeatureOneDimensional) (recordings_list[rec].feature_set.features[feat])).getValue() );
							i = recordings_list[rec].feature_set.features.length + 1; // exit loop
						}
					}
				}
			}
		}

		// Set up column headings of table
		Object[] column_headings = new Object[2 + feature_names.length];
		column_headings[0] = new String("Title");
		column_headings[1] = new String("File Name");
		for (int i = 0; i < feature_names.length; i++)
			column_headings[i + 2] = feature_names[i];

		// Set up contents of table
		String[][] table_contents = new String[recording_info.length][feature_values[0].length + 2];
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			table_contents[rec][0] = recording_info[rec][0];
			table_contents[rec][1] = recording_info[rec][1];
			
			for (int feat = 0; feat < feature_values[rec].length; feat++)
				table_contents[rec][feat + 2] = feature_values[rec][feat];
		}

		// Set up table and its display components
		features_value_table_model = new NonEditableTableModel(column_headings, 0);
		features_value_table = new JTable(features_value_table_model);
		makeTableScrollable(features_value_table, 95);
		addFeatureValueTableMouseListener();
		features_value_table_model.fillTable(table_contents);
		feature_value_diplay_scroll_pane = new JScrollPane(features_value_table);

		// Display table
		add(feature_value_diplay_scroll_pane, BorderLayout.CENTER);
		features_value_table_model.fireTableDataChanged();
		repaint();
		parent_window.repaint();
	}


	// Constructs and displays a table showing extracted one-dimensional (only) scaled
	// feature values (scaled_features_value_table). The scaling is calculated by the
	// currently loaded classifier one-dimensional classifier. The minimum and maximum
	// value cutoffs for each feature are also displayed, as defined by the classifier.
	// Feature values below or above these values are rounded appropriately. All values
	// are scaled to fall between 0 and 1.This table canNOT be sorted by clicking on
	// column headers. Throw an exception if scaling has not been trained or there is
	// a problem when scaling is attempted.
	private void displayScaledFeatureValues()
		throws Exception
	{
		// Get the column names and the scaled values.
		// Exceptions thrown by these functions if there are any problems
		String[] column_names = parent_window.classification_panel.getOneDimFeatureNames(recordings_list);
		double[][] scaled_values = parent_window.classification_panel.getScaledFeatureValues(recordings_list);
		double[] min_feature_cutoffs = parent_window.classification_panel.getMinFeatureValueCutoffs();
		double[] max_feature_cutoffs = parent_window.classification_panel.getMaxFeatureValueCutoffs();

		// Remove anything on the left side of the panel
		if (recordings_display_panel != null)
			remove(recordings_display_panel);
		if (feature_value_diplay_scroll_pane != null)
			remove(feature_value_diplay_scroll_pane);
		if (scaled_feature_value_diplay_scroll_pane != null)
			remove(scaled_feature_value_diplay_scroll_pane);

		// Remove anything on the left side of the panel
		if (recordings_display_panel != null)
			remove(recordings_display_panel);

		// Control which buttons are accessible
		display_recordings_button.setEnabled(true);
		display_feature_values_button.setEnabled(true);
		display_scaled_feature_values_button.setEnabled(false);
		extract_features_button.setEnabled(false);
		classify_recordings_button.setEnabled(false);
		verify_files_button.setEnabled(false);
		new_recordings_list_button.setEnabled(false);
		change_recordings_path_button.setEnabled(false);
		load_recordings_list_button.setEnabled(false);
		save_recordings_list_button.setEnabled(false);
		save_recordings_list_as_button.setEnabled(false);
		edit_recording_button.setEnabled(false);
		add_new_recordings_button.setEnabled(false);
		delete_recordings_button.setEnabled(false);
		delete_saved_features_button.setEnabled(false);
		count_categories_button.setEnabled(false);

		// Find the contents of the first two columns, other than the first two rows.
		String[][] first_two_columns = new String[recordings_list.length][2];
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			first_two_columns[rec][0] = recordings_list[rec].title;
			first_two_columns[rec][1] = StringMethods.convertFilePathToFileName(recordings_list[rec].file_path);
		}

		// Find the contents of the first two rows
		String[][] first_two_rows = new String[2][column_names.length + 2];
		first_two_rows[0][0] = "Low Cutoff Value";
		first_two_rows[0][1] = "--";
		first_two_rows[1][0] = "High Cutoff Value";
		first_two_rows[1][1] = "--";
		for (int feat = 0; feat < column_names.length; feat++)
		{
			first_two_rows[0][feat + 2] = bodhidharma.utilities.StringMethods.getRoundedDouble(min_feature_cutoffs[feat], 4);
			first_two_rows[1][feat + 2] = bodhidharma.utilities.StringMethods.getRoundedDouble(max_feature_cutoffs[feat], 4);
		}

		// Set up column headings of table
		Object[] column_headings = new Object[2 + column_names.length];
		column_headings[0] = new String("Title");
		column_headings[1] = new String("File Name");
		for (int i = 0; i < column_names.length; i++)
			column_headings[i + 2] = column_names[i];

		// Set up contents of table
		String[][] table_contents = new String[recordings_list.length + 2][column_names.length + 2];
		table_contents[0] = first_two_rows[0];
		table_contents[1] = first_two_rows[1];
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			table_contents[rec + 2][0] = first_two_columns[rec][0];
			table_contents[rec + 2][1] = first_two_columns[rec][1];
			for (int feat = 0; feat < column_names.length; feat++)
				table_contents[rec + 2][feat + 2] = bodhidharma.utilities.StringMethods.getRoundedDouble(scaled_values[rec][feat], 4);
		}

		// Set up table and its display components
		scaled_features_value_table_model = new NonEditableTableModel(column_headings, 0);
		scaled_features_value_table = new JTable(scaled_features_value_table_model);
		makeTableScrollable(scaled_features_value_table, 95);
		scaled_features_value_table_model.fillTable(table_contents);
		scaled_feature_value_diplay_scroll_pane = new JScrollPane(scaled_features_value_table);

		// Display table
		add(scaled_feature_value_diplay_scroll_pane, BorderLayout.CENTER);
		scaled_features_value_table_model.fireTableDataChanged();
		repaint();
		parent_window.repaint();
	}


	// Check all file paths held by recordings_list to ensure that they exist.
	// Display the results in an InformationDialog.
	private void verifyFiles()
	{
		String results = MIDIFileVerifier.verifyMIDIFiles(recordings_list);
		InformationDialog results_displayer = new InformationDialog("Verification Results", results);
	}


	// Generates a window showing the number of recordings falling into each category. Categories
	// are divided into three groups: leaf categories, parent categories (i.e. categories at
	// the highest level of the taxonomy) and middle categories (i.e. categoriesthat are
	// neither leaves nor parent categories. The categories in each of these groups are
	// sorted alphabetically. The number of recordings which have not been assigned model
	// categories is also reported, as are the sum of leaf categories beyond the first for each
	// recording and the number of recordings in all.
	private void reportCategoryCount()
	{
		// Find the labels of each type of category
		String[] leaf_categories = parent_window.taxonomy_panel.getLeafLabels();
		String[] parent_categories = parent_window.taxonomy_panel.getParentLabels();
		String[] middle_categories = parent_window.taxonomy_panel.getMiddleLabels();

		// Sort the labels alphabetically
		String[][] temp_leaf_categories = new String[leaf_categories.length][1];
		for (int i = 0; i < temp_leaf_categories.length; i++)
			temp_leaf_categories[i][0] = leaf_categories[i];
		temp_leaf_categories = Sorter.sortTable(temp_leaf_categories, 0, new int[temp_leaf_categories.length]);
		for (int i = 0; i < leaf_categories.length; i++)
			leaf_categories[i] = temp_leaf_categories[i][0];
		String[][] temp_parent_categories = new String[parent_categories.length][1];
		for (int i = 0; i < temp_parent_categories.length; i++)
			temp_parent_categories[i][0] = parent_categories[i];
		temp_parent_categories = Sorter.sortTable(temp_parent_categories, 0, new int[temp_parent_categories.length]);
		for (int i = 0; i < parent_categories.length; i++)
			parent_categories[i] = temp_parent_categories[i][0];
		String[][] temp_middle_categories = new String[middle_categories.length][1];
		for (int i = 0; i < temp_middle_categories.length; i++)
			temp_middle_categories[i][0] = middle_categories[i];
		temp_middle_categories = Sorter.sortTable(temp_middle_categories, 0, new int[temp_middle_categories.length]);
		for (int i = 0; i < middle_categories.length; i++)
			middle_categories[i] = temp_middle_categories[i][0];

		// Prepare variables for counting the number of double counts of recordings, the
		// number of recordings which do not include a model classification and the
		// number of recordings belonging to each category
		int unknown_count = 0;
		int leaf_double_counts = 0;
		int[] leaf_category_count = new int[leaf_categories.length];
		for (int i = 0; i < leaf_category_count.length; i++)
			leaf_category_count[i] = 0;
		int[] parent_category_count = new int[parent_categories.length];
		for (int i = 0; i < parent_category_count.length; i++)
			parent_category_count[i] = 0;
		int[] middle_category_count = new int[middle_categories.length];
		for (int i = 0; i < middle_category_count.length; i++)
			middle_category_count[i] = 0;

		// Perform the tally
		for (int rec = 0; rec < recordings_list.length; rec++)
		{
			if (recordings_list[rec].leaf_categories[0].equals("Unknown"))
				unknown_count++;
			
			else
			{
				// Tally leaf categories
				for (int i = 0; i < recordings_list[rec].leaf_categories.length; i++)
				{
					for (int j = 0; j < leaf_categories.length; j++)
					{
						if (recordings_list[rec].leaf_categories[i].equals(leaf_categories[j]))
						{
							leaf_category_count[j]++;
							if (i > 0)
								leaf_double_counts++; // note the double count
							j = leaf_categories.length + 1; // exit the loop
						}
					}
				}

				// Tally parent categories
				for (int i = 0; i < recordings_list[rec].ancestor_categories.length; i++)
				{
					for (int j = 0; j < parent_categories.length; j++)
					{
						if (recordings_list[rec].ancestor_categories[i].equals(parent_categories[j]))
						{
							parent_category_count[j]++;
							j = parent_categories.length + 1; // exit the loop
						}
					}
				}

				// Tally middle categories
				for (int i = 0; i < recordings_list[rec].ancestor_categories.length; i++)
				{
					for (int j = 0; j < middle_categories.length; j++)
					{
						if (recordings_list[rec].ancestor_categories[i].equals(middle_categories[j]))
						{
							middle_category_count[j]++;
							j = middle_categories.length + 1; // exit the loop
						}
					}
				}
			}
		}

		// Prepare the data to display
		String data_to_display = "LEAF CATEGORIES:\n";
		for (int i = 0; i < leaf_categories.length; i++)
			data_to_display += leaf_categories[i] + "    " + leaf_category_count[i] + "\n";
		data_to_display += "\nROOT CATEGORIES:\n";
		for (int i = 0; i < parent_categories.length; i++)
			data_to_display += parent_categories[i] + "    " + parent_category_count[i] + "\n";
		data_to_display += "\nMIDDLE CATEGORIES:\n";
		for (int i = 0; i < middle_categories.length; i++)
			data_to_display += middle_categories[i] + "    " + middle_category_count[i] + "\n";
		data_to_display += "\nUNKNOWN RECORDINGS:    " + unknown_count + "\n";
		data_to_display += "\nLEAF DOUBLE COUNTS:    " + leaf_double_counts + "\n";
		data_to_display += "\nNUMBER RECORDINGS:    " + recordings_list.length + "\n";

		// Display the results
		InformationDialog results = new InformationDialog("Number of Recordings Belonging to Each Category",
														  data_to_display);
	}


	// Returns a string representing the contents that are to be displayed in
	// the table. File paths are shortened to file names.
	private String[][] convertRecordingsToStringTable()
	{
		String recordings_table[][] = new String[recordings_list.length][6];

		for (int i = 0; i < recordings_list.length; i++)
		{
			// Extract the file name from the file path
			String file_name = StringMethods.convertFilePathToFileName(recordings_list[i].file_path);

			// Generate the list of leaf categories and ancestor categories
			String leaf_categories = new String("");
			for (int j = 0; j < recordings_list[i].leaf_categories.length; j++)
			{
				if (!leaf_categories.equals(""))
					leaf_categories += ";  ";
				leaf_categories += recordings_list[i].leaf_categories[j];
			}
			String ancestor_categories = new String("");
			for (int j = 0; j < recordings_list[i].ancestor_categories.length; j++)
			{
				if (!ancestor_categories.equals(""))
					ancestor_categories += ";  ";
				ancestor_categories += recordings_list[i].ancestor_categories[j];
			}

			// Generate the role column contents. Only allows one to be selected, classification is default.
			String role;			
			if (recordings_list[i].use_for_training)
				role = new String("training");		
			else if (recordings_list[i].use_for_testing)
				role = new String("testing");		
			else
				role = new String("classification");	

			// Populate a row of the column
			recordings_table[i][0] = recordings_list[i].title;
			recordings_table[i][1] = recordings_list[i].artist;
			recordings_table[i][2] = file_name;
			recordings_table[i][3] = leaf_categories;
			recordings_table[i][4] = ancestor_categories;
			recordings_table[i][5] = role;
		}

		return recordings_table;
	}
	

	// Set up the given table so that it will default to the given width,
	// will not allow resizing of columns to a smaller width and allows
	// the user to scroll (if the table is placed in a JScrollBar) accross
	// the columns.
	private void makeTableScrollable(JTable table, int min_width)
	{
		table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		TableColumn column = null;
		for (int i = 0; i < table.getColumnCount(); i++)
		{
			column = table.getColumnModel().getColumn(i);
			column.setMinWidth(min_width);
		}
	}
}