/*
 * @(#)ClassificationResultsInterpereter.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma.classifiers;

import bodhidharma.PreferencesPanel;

/**
 * Part of the Bodhidharma automatic music classification project.
 * This class is used to interperet the results of classifications
 * mase by a <code>ClassificationPanel</code> object. How the results
 * are calculated are based in part on the contents of the 
 * <code>PreferencesPanel</code> given to the constructor.
 * 
 * @author Cory McKay
 * @see	ClassificationPanel
 * @see PreferencesPanel
 */
public class ClassificationResultsInterpereter 
{
	/* FIELDS ******************************************************************/


	// Preferences used to determine how results are interpereted.
	private	PreferencesPanel	preferences;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor that sets the preferences used for calculation of results.
	 */
	public ClassificationResultsInterpereter(PreferencesPanel preferences)
	{
		this.preferences = preferences;
	}


	/* PUBLIC METHODS **********************************************************/

	
	/**
	 * Returns a confusion matrix showing how recordings were classified compared to how they should have been
	 * classified. The criteria used to judge winners are given in the <code>getWinningScores</code>
	 * method's description. The indice of the returned array indicates the recording, and corresponds to the
	 * first indice of the <i>classifier_scores</i> array.
	 *
	 * <p>The rows (first indice) of the returned array correspond to how the recordings were classified, and
	 * the columns (second indice) to how they should have been classified. The order of the labels of both the
	 * rows and the columns corresponds to that of the possible_categories parameter. One extra row (the last)
	 * is included for unknown classifications.
	 *
	 * <p>Classifications (and models) consisting of multiple categories are spread over the matrix fractionally
	 * for each category, so that every recording has the same weighting (1 in total).
	 *
	 * <p>An exception is thrown if the model_classifications contains a category not in possible_categories.
	 *
	 * @param	classifier_scores		Classification scores in the form of relative scores for each of the possible categories for each recording (not normalized). Scores should fall in the range between 0.0 and 1.0 and higher scores correspond to a greater certainty that a recording has a corresponding label. The first indice corresponds to recording and the second corresponds to the category.
	 * @param	possible_categories		The names of the categories into which <i>classifier_scores</i> can be classified. Indice of this and second indice of <i>classifier_scores</i> correspond.
	 * @param	model_classifications	The names of the correct categories for each recording. Returns null if null passed here. First indicie indicates recording. There is one entry for each correct category for each recording.
	 */
	public double[][] getConfusionMatrix(double[][] classifier_scores, String[][] model_classifications, String[] possible_categories)
		throws Exception
	{
		// Return null if model_classifications is null
		if (model_classifications == null)
			return null;

		// Prepare confusion matrix. Columns are true classifications, rows are performed classifications.
		double[][] confusion_matrix = new double[possible_categories.length + 1][possible_categories.length];
		for (int i = 0; i < confusion_matrix.length; i++)
			for (int j = 0; j < confusion_matrix[i].length; j++)
				confusion_matrix[i][j] = 0.0;

		// Find the correct classifications for each recording in boolean form.
		// Also find the number of recordings belonging to each of the possible_catetories in the model_classifications.
		// If a recording in the model_classifications belongs to more than one category, the appropriate fraction
		// is assigned to the category rather than just 1 to each category.
		boolean[][] should_belong_to_category = new boolean[classifier_scores.length][possible_categories.length];
		for (int rec = 0; rec < should_belong_to_category.length; rec++)
			for (int cat = 0; cat < should_belong_to_category[rec].length; cat++)
				should_belong_to_category[rec][cat] = false;
		double[] num_recs_belonging_each_cat = new double[possible_categories.length];
		for (int cat = 0; cat < num_recs_belonging_each_cat.length; cat++)
			num_recs_belonging_each_cat[cat] = 0.0;
		for (int rec = 0; rec < model_classifications.length; rec++)
			for (int g_cat = 0; g_cat < model_classifications[rec].length; g_cat++)
			{
				boolean found = false;
				for (int m_cat = 0; m_cat < possible_categories.length; m_cat++)
				{
					if (model_classifications[rec][g_cat].equals( possible_categories[m_cat] ))
					{
						should_belong_to_category[rec][m_cat] = true;
						num_recs_belonging_each_cat[m_cat] += (1.0 / ((double) model_classifications[rec].length));
						found = true;
						m_cat = possible_categories.length + 1; // exit the loop
					}
				}
				if (!found)
					throw new Exception ("Cannot form confusion matrix because missing category " + model_classifications[rec][g_cat] + ".");
			}

		// Find out how each recording was classified in boolean form.
		// Also find the number of categories each recording was classified as belonging to.
		boolean[][] classified_as_belonging_to_category = new boolean[classifier_scores.length][possible_categories.length];
		for (int rec = 0; rec < classified_as_belonging_to_category.length; rec++)
			for (int cat = 0; cat < classified_as_belonging_to_category[rec].length; cat++)
				classified_as_belonging_to_category[rec][cat] = false;
		int[] num_cats_for_rec = new int[classifier_scores.length];
		for (int rec = 0; rec < num_cats_for_rec.length; rec++)
			num_cats_for_rec[rec] = 0;
		float[][] winning_scores = getWinningScores(classifier_scores);
		for (int rec = 0; rec < winning_scores.length; rec++)
			for (int cat = 0; cat < winning_scores[rec].length; cat++)
				if (winning_scores[rec][cat] > 0)
				{
					classified_as_belonging_to_category[rec][cat] = true;
					num_cats_for_rec[rec]++;
				}
		
		// Form the confusion matrix. Columns are true classifications, rows are performed classifications.
		// If a recording is classified as belonging to more than one category, the appropriate fraction
		// is assigned to the category rather than just 1 to each category.
		for (int rec = 0; rec < classified_as_belonging_to_category.length; rec++)
		{
			double this_rec_value;
			if (num_cats_for_rec[rec] == 0)
				this_rec_value = 1.0;
			else
				this_rec_value = 1.0 / (double) num_cats_for_rec[rec];
			if (num_cats_for_rec[rec] == 0)
				for (int m_cat = 0; m_cat < should_belong_to_category[rec].length; m_cat++)
				{
					if (should_belong_to_category[rec][m_cat])
						confusion_matrix[confusion_matrix.length - 1][m_cat] += this_rec_value;
				}
			else
				for (int c_cat = 0; c_cat < classified_as_belonging_to_category[rec].length; c_cat++)
				{
					if (classified_as_belonging_to_category[rec][c_cat])
					{
						if (should_belong_to_category[rec][c_cat])
							confusion_matrix[c_cat][c_cat] += this_rec_value;
						else
							for (int m_cat = 0; m_cat < should_belong_to_category[rec].length; m_cat++)
								if (should_belong_to_category[rec][m_cat])
									confusion_matrix[c_cat][m_cat] += this_rec_value;
					}
				}
		}

		// Normalize columns of confusion matrix
		double[] column_totals = new double[possible_categories.length];
		for (int i = 0; i < possible_categories.length; i++)
		{	
			column_totals[i] = 0.0;
			for (int j = 0; j < confusion_matrix.length; j++)
				column_totals[i] += confusion_matrix[j][i];
		}
		for (int i = 0; i < confusion_matrix.length; i++)
			for (int j = 0; j < possible_categories.length; j++)
				confusion_matrix[i][j] = 100.0 * confusion_matrix[i][j] / column_totals[j];

		// Return the confusion matrix
		return confusion_matrix;
	}


	/**
	 * Returns a formatted set of strings containing the categories which are judged to be winners by the classification
	 * system for each recording. The criteria used to judge this are given in the <code>getWinningScores</code>
	 * method's description. The indice of the returned array indicates the recording, and corresponds to the
	 * first indice of the <i>classifier_scores</i> array.
	 *
	 * <p>Additional information that can be included in the report for each recording includes model classifications
	 * (null can be sent to the <i>model_classifications</i> parameter if these are not known, the classifier's
	 * secondsary choices and actual scores for each chosen category.
	 *
	 * <p>One, more than one or no categories may be selected for each recording. 
	 *
	 * <p>An exception is thrown if the user preferences have inappropriate values.
	 *
	 * @param	classifier_scores		Classification scores in the form of relative scores for each of the possible categories for each recording (not normalized). Scores should fall in the range between 0.0 and 1.0 and higher scores correspond to a greater certainty that a recording has a corresponding label. The first indice corresponds to recording and the second corresponds to the category.
	 * @param	possible_categories		The names of the categories into which <i>classifier_scores</i> can be classified. Indice of this and second indice of <i>classifier_scores</i> correspond.
	 * @param	model_classifications	The names of the correct categories for each recording. Ignored if null is passed here. First indicie indicates recording. There is one entry for each correct category for each recording.
	 * @param	report_scores			Whether or not to include reports of score values along with winning categories.
	 * @param	report_second_choices	Whether or not to report second choices along with winning categories.
	 */
	public String[] getWinnerLabels( double[][] classifier_scores,
	                                 String[] possible_categories,
	                                 String[][] model_classifications,
	                                 boolean report_scores,
	                                 boolean report_second_choices )
		throws Exception
	{
		// The categories of the winners, the second choice choice categories and their scores
		String[] winner_names = new String[classifier_scores.length];

		// Sort classifier_scores into winners, second placers and losers
		float[][] winning_scores = getWinningScores(classifier_scores);

		// Go through recordings one by one
		for (int rec = 0; rec < winning_scores.length; rec++)
		{
			// The report for this recording
			winner_names[rec] = new String("");
			
			// Count the number of winners
			int number_winners = 0;
			for (int cat = 0; cat < winning_scores[rec].length; cat++)
				if (winning_scores[rec][cat] > (float) 0.0)
					number_winners++;

			// Deal with case where there are no winners
			if (number_winners == 0)
				winner_names[rec] += "Unknown";
			
			// Deal with case where there is at least one winner. Check categories for winners 
			// and generate formatted list of winning categories and their scores
			else
			{
				for (int cat = 0; cat < winning_scores[rec].length; cat++)
				{
					int number_reported = 0;
					if (winning_scores[rec][cat] > (float) 0.0)
					{
						winner_names[rec] += possible_categories[cat];
						if (report_scores)
							winner_names[rec] += "(" + String.valueOf(winning_scores[rec][cat]) + ")";
						number_reported++;
						if (number_reported < number_winners)
							winner_names[rec] += ",  ";
					}
				}
			}

			// Generate formatted list of model classifications
			if (model_classifications != null)
				if (model_classifications[rec] != null)
					for (int cat = 0; cat < model_classifications[rec].length; cat++)
					{	
						if (cat == 0)
							winner_names[rec] += "     MODELS: ";
						winner_names[rec] += model_classifications[rec][cat];
						if (cat < (model_classifications[rec].length - 1))
							winner_names[rec] += ",  ";
					}

			// Check categories for second choicers and generate formatted list of
			// these categories and their scores
			if (report_second_choices)
			{
				// Find the number of second choices
				int number_second_choices = 0;
				for (int cat = 0; cat < winning_scores[rec].length; cat++)
					if (winning_scores[rec][cat] < (float) 0.0 && winning_scores[rec][cat] > (float) -10.0)
						number_second_choices++;

				// Report second choices
				if (number_second_choices > 0)
				{
					winner_names[rec] += "     2nd CHOICES: ";
					int number_reported = 0;
					for (int cat = 0; cat < winning_scores[rec].length; cat++)
					{
						if (winning_scores[rec][cat] < (float) 0.0 && winning_scores[rec][cat] > (float) -1.0)
						{
							winner_names[rec] += possible_categories[cat];
							if (report_scores)
								winner_names[rec] += "(" + String.valueOf(Math.abs(winning_scores[rec][cat])) + ")";
							number_reported++;
							if (number_reported < number_second_choices)
								winner_names[rec] += ",  ";
						}
					}
				}
			}
		}

		// Return results
		return winner_names;
	}


	/**
	 * Returns the following three statistics about the success of the given classifier results based
	 * on the given model classifications:
	 *
	 * <p>- The percentage of recordings where the correct categories were found in the first choice.
	 * The percentage for a single recording is the number of correct first choices for a category
	 * divided by the number of model categories for that recording, and the average of this is found
	 * accross recordings.
	 * <p>- The percentage of recordings where at least one category that was missed in the first
	 * choices was present in the second choices.
	 * <p>- The percentage of recordings where at least one extra first choice was present
	 * that was not in the model categories.
	 *
	 * <p>The returned array has these values in entries 0, 1 and 2 respectively.
	 *
	 * <p>A value of null is returned if model_classifications is null.
	 *
	 * <p>An exception is thrown if the user preferences have inappropriate values.
	 *
	 * @param	classifier_results		Classification scores in the form of relative scores for each of the possible categories for each recording (not normalized). Scores should fall in the range between 0.0 and 1.0 and higher scores correspond to a greater certainty that a recording has a corresponding label. The first indice corresponds to recording and the second corresponds to the category.
	 * @param	model_classifications	The correct category(ies) for each of the recordings. The indice identifies the recording  that the category(ies) correspond to. A value of null indicates that model results should not be printed.
	 * @param	classifier_categories	The categories into which the classifiers classified test rerordings.
	 */
	public double[] getSuccessRate(double[][] classifier_results, String[][] model_classifications, String[] classifier_categories)
		throws Exception
	{
		// Return null if no model classifications are given
		if (model_classifications == null)
			return null;

		// Get the winner and secondary winner information
		float[][] winning_scores = getWinningScores(classifier_results);

		// Store the names first and second choices of the classification results
		String[][] first_choices = new String[winning_scores.length][];
		String[][] second_choices = new String[winning_scores.length][];
		for (int rec = 0; rec < winning_scores.length; rec++)
		{
			// Find the number of first and second choices
			int number_first_choices = 0;
			int nubmer_second_choices = 0;
			for (int cat = 0; cat < winning_scores[rec].length; cat++)
			{
				if (winning_scores[rec][cat] > (float) 0.0)
					number_first_choices++;
				else if (winning_scores[rec][cat] < (float) 0.0 && winning_scores[rec][cat] > (float) -1.0)
					nubmer_second_choices++;
			}

			// Record the category names of the first and second choices
			first_choices[rec] = new String[number_first_choices];
			second_choices[rec] = new String[nubmer_second_choices];
			int current_first = 0;
			int current_second = 0;
			for (int cat = 0; cat < winning_scores[rec].length; cat++)
			{
				if (winning_scores[rec][cat] > (float) 0.0)
				{
					first_choices[rec][current_first] = classifier_categories[cat];
					current_first++;
				}
				else if (winning_scores[rec][cat] < (float) 0.0 && winning_scores[rec][cat] > (float) -1.0)
				{
					second_choices[rec][current_second] = classifier_categories[cat];
					current_second++;
				}
			}
		}

		// Record the number of correct first and second classifications for each recording
		int[] corrects_firsts_per_rec = new int[winning_scores.length];
		int[] corrects_seconds_per_rec = new int[winning_scores.length];
		for (int rec = 0; rec < classifier_results.length; rec++)
		{
			corrects_firsts_per_rec[rec] = 0;
			for (int fcat = 0; fcat < first_choices[rec].length; fcat++)
				for (int mcat = 0; mcat < model_classifications[rec].length; mcat++)
					if ( first_choices[rec][fcat].equals( model_classifications[rec][mcat] ) )
						corrects_firsts_per_rec[rec]++;

			corrects_seconds_per_rec[rec] = 0;
			for (int scat = 0; scat < second_choices[rec].length; scat++)
				for (int mcat = 0; mcat < model_classifications[rec].length; mcat++)
					if ( second_choices[rec][scat].equals( model_classifications[rec][mcat] ) )
						corrects_seconds_per_rec[rec]++;
		}

		// Find the percentage of recordings where the correct categories were found in the first choice.
		// The percentage for a single recording is the number of correct first choices for a category
		// divided by the number of model categories for that recording, and the average of this is found.
		double[] first_choice_success_rates = new double[corrects_firsts_per_rec.length];
		for (int rec = 0; rec < first_choice_success_rates.length; rec++)
			first_choice_success_rates[rec] = 100.0 * ((double) corrects_firsts_per_rec[rec]) / ((double) model_classifications[rec].length);
		int number_with_no_classifications = 0;
		for (int rec = 0; rec < model_classifications.length; rec++)
			if (model_classifications[rec].length == 0)
				number_with_no_classifications++;
		if (number_with_no_classifications != 0) // so that ones with model classifications are included in percentage
		{
			double[] temp = new double[first_choice_success_rates.length - number_with_no_classifications];
			int current_index = 0;
			for (int rec = 0; rec < first_choice_success_rates.length; rec++)
			{
				if (model_classifications[rec].length != 0)
				{
					temp[current_index] = first_choice_success_rates[rec];
					current_index++;
				}
			}
			first_choice_success_rates = temp;
		}
		double first_choice_success_percentage = bodhidharma.utilities.Statistics.getAverage(first_choice_success_rates);

		// Find the percentage of recordings where at least one category that was missed in the first
		// choices was present in the second choices
		double was_in_second_choice_percentage = 0.0;
		int number_correct_seconds = 0;
		for (int rec = 0; rec < corrects_seconds_per_rec.length; rec++)
			if (corrects_seconds_per_rec[rec] != 0)
				number_correct_seconds++;
		int number_incorrect_firsts = 0;
		for (int rec = 0; rec < corrects_firsts_per_rec.length; rec++)
			if (corrects_firsts_per_rec[rec] < model_classifications[rec].length)
				number_incorrect_firsts++;
		if (number_incorrect_firsts != 0)
			was_in_second_choice_percentage = 100.0 * ((double) number_correct_seconds) / ((double) number_incorrect_firsts);

		// Find the percentage of recordings where at least one extra first choice was present
		// that was not in the model categories
		double number_first_choice_overuns = 0.0;
		double number_recordings_with_classifications = 0.0;
		for (int rec = 0; rec < first_choices.length; rec++)
			if (model_classifications[rec].length != 0)
			{
				number_recordings_with_classifications++;
				if (first_choices[rec].length > model_classifications[rec].length)
					number_first_choice_overuns++;
			}
		double overun_percentage = 100.0 * number_first_choice_overuns / number_recordings_with_classifications;

		// Format and return the results
		double[] results = new double[3];
		results[0] = first_choice_success_percentage;
		results[1] = was_in_second_choice_percentage;
		results[2] = overun_percentage;
		return results;
	}


	/**
	 * Returns true for the winning categories for each recording and false otherwise.
	 * See the <code>getWinningScores</code> method for details on how this is determined.
	 *
	 * @param	classifier_scores	Classification scores in the form of relative scores for each of the possible categories for each recording (not normalized). Scores should fall in the range between 0.0 and 1.0 and higher scores correspond to a greater certainty that a recording has a corresponding label. The first indice corresponds to recording and the second corresponds to the category.
	 * @return	The winners. First indice indicates recording and second indicates category.
	 */
	public boolean[][] getWinners(double[][] classifier_scores)
		throws Exception
	{
		float[][] numeric_results = getWinningScores(classifier_scores);
		boolean[][] results = new boolean[numeric_results.length][numeric_results[0].length];
		for (int rec = 0; rec < results.length; rec++)
			for (int cat = 0; cat < results[rec].length; cat++)
			{
				if (numeric_results[rec][cat] > 0)
					results[rec][cat] = true;
				else
					results[rec][cat] = false;
			}
		return results;
	}


	/**
	 * Determines which categories are winners, which are second choices and which are losers
	 * given a set of classifier scores for a set of recordings. See below for a description
	 * of the returned data. A given recording can have no, one or many winners.
	 * 
	 * <p>All categories that meet one of the two following conditions for a recording are
	 * considered winners (NOTE THAT BELOW SPECIFIC VALUES ARE DEFAULTS. ACTUAL VALUES
	 * ARE SET BY USER PREFERENCES):
	 *
	 * <p>- Have a score over 0.5
	 * <br>- Have scores within 20% of the highest score for the recording, and must have a
	 * score of at least 0.25
	 *
	 * <p>All categories that meet one of the two following conditions for a recording are
	 * considered second choices:
	 *
	 * <p>- Are not winners
	 * <br>- Have scores within 30% of the highest score for the recording, and must have a
	 * score of at least 0.2
	 *
	 *<p>An exception is thrown if the user preferences have inappropriate values.
	 *
	 * @param	classifier_scores	Classification scores in the form of relative scores for each of the possible categories for each recording (not normalized). Scores should fall in the range between 0.0 and 1.0 and higher scores correspond to a greater certainty that a recording has a corresponding label. The first indice corresponds to recording and the second corresponds to the category.
	 * @return	The classifier scores of winners and second choices. First indice indicates recording and second indicates category. A score of -10 means that the recording is not classified as belonging to the category, a positive number means that the category is a first choice (the number is the classifier's score) and a negative number (other than -10) means that the category is a second choice for the recording.
	 */
	public float[][] getWinningScores(double[][] classifier_scores)
		throws Exception
	{
		// Some useful constants
		double minimum_automatic_qualification_score = preferences.getMinimumAutomaticQualificationScore();
		double minimum_score_to_qualify = preferences.getMinimumScoreToQualify();
		double percentage_below_highest_to_qualify = preferences.getPercentageBelowHighestToQualify();
		double minimum_score_to_be_second_choice = preferences.getMinimumScoreToBeSecondChoice();
		double percentage_below_highest_to_be_second_choice = preferences.getPercentageBelowHighestToBeSecondChoice();

		// Initialize winning scores. Set all categories to loser as default for all recordings.
		float[][] winning_scores = new float[classifier_scores.length][classifier_scores[0].length];
		for (int rec = 0; rec < winning_scores.length; rec++)
			for (int cat = 0; cat < winning_scores[rec].length; cat++)
				winning_scores[rec][cat] = (float) -10.0;

		// Go through recordings one by one, recording the scores of the winners
		for (int rec = 0; rec < classifier_scores.length; rec++)
		{
			// Mark all recordings with scores over minimum_automatic_qualification_score as winners
			for (int cat = 0; cat < classifier_scores[rec].length; cat++)
				if (classifier_scores[rec][cat] > minimum_automatic_qualification_score)
					winning_scores[rec][cat] = (float) classifier_scores[rec][cat];

			// Mark the category with the highest scores as well as all categories that have
			// scores percentage_below_highest_to_qualify lower than it as winners. Only
			// scores over minimimum_score_to_qualify can qualify as winners, however.
			// Also note second choices.
			int index_of_highest = bodhidharma.utilities.Statistics.getIndexOfLargest(classifier_scores[rec]);
			double threshold_to_win = classifier_scores[rec][index_of_highest] -
			                          (classifier_scores[rec][index_of_highest] * percentage_below_highest_to_qualify / 100.0);
			double threshold_to_be_second_choice = classifier_scores[rec][index_of_highest] -
			                                       (classifier_scores[rec][index_of_highest] * percentage_below_highest_to_be_second_choice / 100.0);
			for (int cat = 0; cat < classifier_scores[rec].length; cat++)
			{
				if (classifier_scores[rec][cat] > threshold_to_win && classifier_scores[rec][cat] > minimum_score_to_qualify)
					winning_scores[rec][cat] = (float) classifier_scores[rec][cat];
				else if (classifier_scores[rec][cat] > threshold_to_be_second_choice && classifier_scores[rec][cat] > minimum_score_to_be_second_choice)
					winning_scores[rec][cat] = (float) (-1.0 * classifier_scores[rec][cat]);
			}
		}

		// Only pick one winner if this option is selected in preferences
		if (preferences.shouldOnlyPickOneWinnerForCombinedClassification())
		{
			for (int rec = 0; rec < winning_scores.length; rec++)
			{
				int winning_index = bodhidharma.utilities.Statistics.getIndexOfLargest(winning_scores[rec]);
				for (int cat = 0; cat < winning_scores[rec].length; cat++)
					if (cat != winning_index && winning_scores[rec][cat] > 0)
						winning_scores[rec][cat] = (float) -1.0 * winning_scores[rec][cat];
			}
		}

		// Return the results
		return winning_scores;
	}
}