/*
 * @(#)NN_BioKNN_Ensemble.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers;

import bodhidharma.data_structures.*;
import bodhidharma.classifiers.genetic_algorithms.*;
import bodhidharma.classifiers.feedforward_neural_networks.*;
import bodhidharma.classifiers.bio_k_nearest_neighbour.BioKNearestNeighbour;
import bodhidharma.midi_parsing.multi_dimensional_features.FeatureMultiDimensional;
import bodhidharma.midi_parsing.one_dimensional_features.FeatureOneDimensional;
import bodhidharma.ProgressBarTaskTrainMonitor;
import java.io.*;
import bodhidharma.xml_parsing.*;

/**
 * An interface for using a combination of classifiers to result in a final
 * classification. The ensemble consists of one <code>BioKNearestNeighbour</code>
 * that is fed all <code>FeatureOneDimensionals</code> and one
 * <code>FeedForwardNeuralNetwork</code> for each <code>FeatureMultiDimensional</code>.
 * It is possible to configure <code>NN_BioKNN_Ensemble</code> objects to only use one
 * of these two types of classifiers.
 *
 * <p>Classifier selection can be performed on the different classifiers
 * using genetic algorighms. Weightings can also be evolved for each of the 
 * <code>FeedForwardNeuralNetworks</code> and the <code>BioKNearestNeighbour</code>.
 * Alternatively, half of the weighting can be automatically assigned to the
 * <code>BioKNearestNeighbour</code> classifier and the remaining weighting can be
 * distributed equally among the <code>FeedForwardNeuralNetwork</code>.
 *
 * <p>This implementation makes it possible to assign more than one label to a single
 * feature set.
 *
 * <p>Use the <code>train</code> method to train the classifier (the
 * <code>getFeatureNames</code> method is useful for deteriming the features that
 * can be fed to the classifer).
 *
 * <p><b>WARNING:</b> due to the nature of the classifiers used, it is best to only
 * train objects of this class once. In order to ensure this, the <code>train</code>
 * method throws an exception if it is called and the classifier has already been
 * trained.
 *
 * <p>Use the <code>classify</code> method to classify feature sets once training
 * has been completed (the <code>getCategories</code> method is useful for
 * determining what categories feature sets can be classified into and for
 * determining the order of the categories when parsing classification results).
 *
 * <p>Use the <code>save</code> method to save the classifier and its current
 * state to disk.
 * 
 * <p>Use the <code>getClassifierName</code> and <code>getClassifierParameters</code>
 * methods to obtain information about the classifier.
 *
 * <p>Use the <code>getClassifierIdentifier</code> method to get a name or code that
 * was given to an instantiation of the classifier when it was constructed. This
 * identifier can be used by external classes to identify the instantiation.
 *
 * <p>Use the <code>getClassifierParameters</code> method to get information about
 * the settings stored in the instantiation of an object of this class.
 *
 * <p>Use the <code>isTrained</code> method to find out if the ensemble has been trained
 * yet.
 *
 * <p>Use the <code>getEnsembleSelection</code> method to see the set of classifiers which
 * survived feature selection. The names of the corresponding classifiers can be accessed
 * with the <code>getClassifierNames</code> method.
 *
 * @see SupervisedClassifier
 * @see BioKNearestNeighbour
 * @see FeedForwardNeuralNetwork
 * @see FeatureOneDimensionals
 * @see FeatureMultiDimensional
 * @see FeatureSet
 * @see Recording
 * @author Cory McKay
 */
public class NN_BioKNN_Ensemble
{
	/* FIELDS ******************************************************************/


	// An identifier that can be associated with the ensemble so that outside classes
	// can identify it.
	private	String						identifier;

	// The classifiers used in this ensemble. Values of null mean that they have
	// not been trained yet.
	private	BioKNearestNeighbour		bio_knn_classifier;
	private	FeedForwardNeuralNetwork[]	neural_network_classifiers;

	// Whether or not particular classifiers are to be used. The first entry always
	// corresponds to the bio_knn_classifier, even if  the ensemble_using_classifier_selection
	// field is false (value of false in this array in this case). Following entries correspond
	// to the neural_network_classifiers.
	private boolean[]					ensemble_selection;

	// The weighting to be used for each classifier. The first entry always corresponds 
	// to the bio_knn_classifier, even if  the ensemble_using_classifier_selection
	// field is false. Following entries correspond to the neural_network_classifiers.
	// Should be normalized. Values of 0.0 should be present if the corresponding entry
	// of the ensemble_selection field is false.
	private double[]					ensemble_weightings;

	// Whether or not the classifier has already been trained
	private	boolean						already_trained;

	// Identifies what candidate categories there are
	private	String[]					possible_taxonomy_labels;

	// The names of the possible features that should be used for training
	// and testing
	private String[]					one_dim_feature_labels;
	private String[]					multi_dim_feature_labels;

	// Ensemble settings (see costructor for descriptions)
	private	boolean						ensemble_using_classifier_selection;
	private	boolean						ensemble_using_classifier_weighting;
	private	double						csw_training_fraction;

	// Whether or not one-dimensional and/or multi-dimensional feature extraction is
	// performed
	private	boolean						using_one_dimensional_features;
	private boolean						using_multi_dimensional_features;

	// Sub-classifier settings (see constructor for descriptions)
	private	GeneticAlgorithmJFrame		ga_settings;
	private	boolean						bioknn_using_feature_selection;
	private	boolean						bioknn_using_feature_weighting;
	private int							bioknn_iterations;
	private	double						classifier_vs_feature_multiplier;
	private double						bioknn_acceptable_threshold;
	private int							bioknn_consecutive_iterations;
	private	NeuralNetworkJFrame			network_settings;
	private int							network_iterations;
	private double						network_acceptable_threshold;
	private int							network_consecutive_iterations;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Sets the basic fields of the <code>NN_BioKNN_Ensemble</code>.
	 *
	 * <p><b>NOTE:</b> Some parameters may be ignored, based on the contents of the
	 * <i>using_one_dimensional_features</i> and <i>using_multi_dimensional_features</i>
	 * fields.
	 *
	 * <p>Throws an exception if both the <i>use_one_dimensional</i> and
	 * <i>use_multi_dimensional</i> parameters are false, if no taxonomy labels are specified,
	 * or if no features of the appropriate type are present in the <i>feature_settings</i>
	 * parameter.
	 *
	 * @param	identifier							An identifier that can be associated with the ensemble of classifiers so that outside classes can identify it.
	 * @param	possible_taxonomy_labels			The system will be trained to classify samples into one or more of these labels.
	 * @param	feature_settings					The features contained in the <code>FeaturesPanel</code>. The ones marked as Selection Candidates are used if the use_selection_candidates parameter is true and the ones marked as Overide Status are used otherwise.
	 * @param	use_selection_candidates			See <i>feature_settings</i> parameter.
	 * @param	ensemble_using_classifier_selection	Whether or not selection will be performed on the classifiers.		
	 * @param	ensemble_using_classifier_weighting	Whether or not classifier weighting will be performed.
	 * @param	csw_training_fraction				Fraction of training samples to actually be used for training when performing feature selection and weighting of one-dimensional features. Must be between 0 and 1.
	 * @param	using_one_dimensional_features		Whether or not training and classification is performed using one-dimensional features (and therefore <code>BioKNearestNeighbour</code> classifiers.
	 * @param	using_multi_dimensional_features	Whether or not training and classification is performed using multi-dimensional features (and therefore <code>FeedForwardNeuralNetwork</code> classifiers.
	 * @param	ga_settings							Initialization settings for all genetic algorithms.
	 * @param	bioknn_using_feature_selection		Whether or not feature selection will be performed on one-dimensional features.
	 * @param	bioknn_using_feature_weighting		Whether or not variable feature weighting will be performed on one-dimensional features.
	 * @param	bioknn_iterations					The absolute value of this parameter indicates the maximum number of training iterations that the <code>BioKNearestNeighbour</code> will perform. A negative value means that the system will automaticall check after each iteratoin, based on the below two parameters, whether training should continue. A positive value means that the below two parameters are ignored.
	 * @param	classifier_vs_feature_multiplier	The value used to multiply the bioknn_iterations and population when using GAS for classifier selection or weighting rather than feature selection or weighting.
	 * @param	bioknn_acceptable_threshold			The error threshold below which <code>BioKNearestNeighbour</code> objects will cease training automatically.
	 * @param	bioknn_consecutive_iterations		The number of consecutiove iterations which must go by for which the error is below the above threshold for <code>BioKNearestNeighbour</code> objects in order for training to stop.
	 * @param	network_settings					Initialization settings for all neural networks.
	 * @param	network_iterations					The absolute value of this parameter indicates the maximum number of training iterations that the <code>FeedForwardNeuralNetworks</code> will perform. A negative value means that the system will automaticall check after each iteratoin, based on the below two parameters, whether training should continue. A positive value means that the below two parameters are ignored.
	 * @param	network_acceptable_threshold		The error threshold below which <code>FeedForwardNeuralNetworks</code> objects will cease training automatically.
	 * @param	network_consecutive_iterations		The number of consecutiove iterations which must go by for which the error is below the above threshold for <code>FeedForwardNeuralNetworks</code> objects in order for training to automatically stop.
	 */
	public NN_BioKNN_Ensemble( String identifier,
	                           String[] possible_taxonomy_labels,
                               FeatureSettings[] feature_settings,
                               boolean use_selection_candidates,
	                           boolean ensemble_using_classifier_selection,
	                           boolean ensemble_using_classifier_weighting,
			                   double csw_training_fraction,
	                           boolean using_one_dimensional_features,
	                           boolean using_multi_dimensional_features,
	                           GeneticAlgorithmJFrame ga_settings,
	                           boolean bioknn_using_feature_selection,
	                           boolean bioknn_using_feature_weighting,
			                   int bioknn_iterations,
	                           double classifier_vs_feature_multiplier,
			                   double bioknn_acceptable_threshold,
			                   int bioknn_consecutive_iterations,
	                           NeuralNetworkJFrame network_settings,
			                   int	network_iterations,
			                   double network_acceptable_threshold,
			                   int network_consecutive_iterations )
		throws Exception
	{
		// Make sure that something actually can be trained
		if (!using_one_dimensional_features && !using_multi_dimensional_features)
			throw new Exception("You have configured a classifier ensemble to consider neither\n" +
			                    "one-dimensional nor multi-dimensional features. All ensembles\n" +
		                        "must consider at least one feature.");

		// Make sure that taxonomy labels are available
		if (possible_taxonomy_labels.length == 0)
			throw new Exception("No taxonomy labels have been specified for training by this set of " +
			                    "classifiers.");
		
		// Set identifier field
		this.identifier = identifier;

		// Set candidate categories
		this.possible_taxonomy_labels = possible_taxonomy_labels;

		// Set ensemble fields
		this.ensemble_using_classifier_selection = ensemble_using_classifier_selection;
		this.ensemble_using_classifier_weighting = ensemble_using_classifier_weighting;
		this.csw_training_fraction = csw_training_fraction;

		// Set what kind of classification are being performed
		this.using_one_dimensional_features = using_one_dimensional_features;
		this.using_multi_dimensional_features = using_multi_dimensional_features;

		// Set sub-classifier settings
		this.ga_settings = ga_settings;
		this.bioknn_using_feature_selection = bioknn_using_feature_selection;
		this.bioknn_using_feature_weighting = bioknn_using_feature_weighting;
		this.bioknn_iterations = bioknn_iterations;
		this.classifier_vs_feature_multiplier = classifier_vs_feature_multiplier;
		this.bioknn_acceptable_threshold = bioknn_acceptable_threshold;
		this.bioknn_consecutive_iterations = bioknn_consecutive_iterations;
		this.network_settings = network_settings;
		this.network_iterations = network_iterations ;
		this.network_acceptable_threshold = network_acceptable_threshold;
		this.network_consecutive_iterations = network_consecutive_iterations;

		// Set classifiers to not being trained
		already_trained = false;
		neural_network_classifiers = null;
		bio_knn_classifier = null;
		ensemble_selection = null;
		ensemble_weightings = null;

		// Find the appropriate feature names to use
		String[][] feature_names;
		if (use_selection_candidates)
			feature_names = getCandidateFeatures(feature_settings);
		else 
			feature_names = getOverideFeatures(feature_settings);
		one_dim_feature_labels = feature_names[0];
		multi_dim_feature_labels = feature_names[1];

		// Throw exception if asked to create a classifier for which no features could be available.
		if (using_one_dimensional_features == true && one_dim_feature_labels.length == 0)
			throw new Exception("The system is configured to use one-dimensional features, but " +
			                    "no one-dimensional features are set to be used for classification.");
		if (using_multi_dimensional_features == true && multi_dim_feature_labels.length == 0)
			throw new Exception("The system is configured to use multi-dimensional features, but " +
			                    "no multi-dimensional features are set to be used for classification.");
	}


	/**
	 * Parse the file specified by the given file path to recreate the specificed trained
	 * classifier. Throws an exception if problems occur during parsing.
	 *
	 * @param	file_path			The file path of a NN_BioKNN_Ensemble_file that this instantiation is to be based on.
	 * @param	ga_settings			Initialization settings for all genetic algorithms.
	 * @param	network_settings	Initialization settings for all neural networks.
	 */
	public NN_BioKNN_Ensemble(String file_path, GeneticAlgorithmJFrame ga_settings, NeuralNetworkJFrame network_settings)
		throws Exception
	{
		// Set the JFrame fields
		this.ga_settings = ga_settings;
		this.network_settings = network_settings;

		// Parse file
		Object[] results = (Object[]) XMLDocumentParser.parseXMLDocument(file_path, "NN_BioKNN_Ensemble_file");
		String[] fixed_fields = (String[]) results[0];
		
		// Set remaining non-classifier fields
		identifier = fixed_fields[0];
		ensemble_selection =  (boolean[]) results[2];
		ensemble_weightings = (double[]) results[3];
		if (fixed_fields[2].equals("true"))
			already_trained = true;
		else
			already_trained = false;
		possible_taxonomy_labels = (String[]) results[4];
		one_dim_feature_labels = (String[]) results[5];
		multi_dim_feature_labels = (String[]) results[6];
		if (fixed_fields[3].equals("true"))
			ensemble_using_classifier_selection = true;
		else
			ensemble_using_classifier_selection = false;
		if (fixed_fields[4].equals("true"))
			ensemble_using_classifier_weighting = true;
		else
			ensemble_using_classifier_weighting = false;
		csw_training_fraction = (Double.valueOf(fixed_fields[5])).doubleValue();
		if (fixed_fields[6].equals("true"))
			using_one_dimensional_features = true;
		else
			using_one_dimensional_features = false;
		if (fixed_fields[7].equals("true"))
			using_multi_dimensional_features = true;
		else
			using_multi_dimensional_features = false;
		if (fixed_fields[8].equals("true"))
			bioknn_using_feature_selection = true;
		else
			bioknn_using_feature_selection = false;
		if (fixed_fields[9].equals("true"))
			bioknn_using_feature_weighting = true;
		else
			bioknn_using_feature_weighting = false;
		bioknn_iterations = (Integer.valueOf(fixed_fields[10])).intValue();
		classifier_vs_feature_multiplier = 1.0; // because doesn't matter because would only use for training, which would require instantiation of new NN_BioKNN_Ensemble
		bioknn_acceptable_threshold = (Double.valueOf(fixed_fields[11])).doubleValue();
		bioknn_consecutive_iterations = (Integer.valueOf(fixed_fields[12])).intValue();
		network_iterations = (Integer.valueOf(fixed_fields[13])).intValue();
		network_acceptable_threshold = (Double.valueOf(fixed_fields[14])).doubleValue();
		network_consecutive_iterations = (Integer.valueOf(fixed_fields[15])).intValue();

		// Set classifiers
		String bio_knn_classifier_file_path = fixed_fields[1];
		if (bio_knn_classifier_file_path == null)
			bio_knn_classifier = null;
		else
			bio_knn_classifier = new BioKNearestNeighbour(bio_knn_classifier_file_path);
		String[] neural_network_classifiers_file_paths = (String[]) results[1];
		if (neural_network_classifiers_file_paths == null)
			neural_network_classifiers = null;
		else
		{
			neural_network_classifiers = new FeedForwardNeuralNetwork[neural_network_classifiers_file_paths.length];
			for (int i = 0; i < neural_network_classifiers_file_paths.length; i++)
				neural_network_classifiers[i] = new FeedForwardNeuralNetwork(neural_network_classifiers_file_paths[i]);
		}
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the type of classifier. This information can be used
	 * by external classes to identify the type of classifier in situations such
	 * as when the configuration of a trained classifier is being read from a file.
	 */
	public String getClassifierName()
	{
		return identifier;
	}


	/**
	 * Returns a String describing the parameters of the classifier. This does not
	 * include training values, however.
	 */
	public String getClassifierParameters()
	{
// IMPLELEMENT
		return new String("This feature not yet implemented.");
	}


	/**
	 * Returns the categories into which this ensemble classifies test samples.
	 */
	public String[] getCategories()
	{
		return possible_taxonomy_labels;
	}
	

	/**
	 * Returns the number of multi-dimensional features present.
	 */
	 public int getNumberMultiDimFeatures()
	{
		return multi_dim_feature_labels.length;
	}


	/**
	 * Returns whether or not the ensemble has been trained.
	 */
	public boolean isTrained()
		throws Exception
	{
		return already_trained;
	}


	/**
	 * Returns the set of classifiers which survived feature selection (true).
	 * The names of the corresponding classifiers can be accessed with the
	 * getClassifierNames method.
	 *
	 * <p>An exception is thrown if the classifiers have not yet been trained.
	 */
	public boolean[] getEnsembleSelection()
		throws Exception
	{
		if (!already_trained)
			throw new Exception ("A needed ensemble of classifiers has not yet been trained.\n" +
			                     "All necessary ensembles of classifiers must be trained\n" +
			                     "before classifier selection can be accessed.");

		return ensemble_selection;
	}


	/**
	 * Returns the weights for the set of classifiers after feature weighting.
	 * The names of the corresponding classifiers can be accessed with the
	 * getClassifierNames method.
	 *
	 * <p>An exception is thrown if the classifiers have not yet been trained.
	 */
	public double[] getEnsembleWeightings()
		throws Exception
	{
		if (!already_trained)
			throw new Exception ("A needed ensemble of classifiers has not yet been trained.\n" +
			                     "All necessary ensembles of classifiers must be trained\n" +
			                     "before classifier weightings can be accessed.");

		return ensemble_weightings;
	}


	/**
	 * Returns the names of the classifiers stored in this ensemble.
	 *
	 * <p>An exception is thrown if the classifiers have not yet been trained.
	 */
	public String[] getClassifierNames()
		throws Exception
	{
		if (!already_trained)
			throw new Exception ("A needed ensemble of classifiers has not yet been trained.\n" +
			                     "All necessary ensembles of classifiers must be trained\n" +
			                     "before their names can be accessed.");

		String[] classifier_names = new String[ensemble_selection.length];
		for (int i = 0; i < classifier_names.length; i++)
		{
			if (i == 0)
				classifier_names[i] = "One Dimensional Feature Classifier";
			else
				classifier_names[i] = neural_network_classifiers[i - 1].getClassifierIdentifier();
		}

		return classifier_names;
	}


	/**
	 * Trains the ensemble of classifiers using the given recordings and their model classifications.
	 * More specifically, the contents of the <i>bio_knn_classifier</i> and the <i>neural_network_classifiers</i>
	 * fields are instantiated and filled with the appropriate trained classifier(s). The following
	 * classifiers are produced and trained:
	 *
	 * <p>- A <code>BioKNearestNeighbour</code> that is fed all of the eligible one-dimensional
	 * features. This classifier is only created if the <i>using_one_dimensional_features</i>
	 * field is set to true.
	 * <br>- One <code>FeedForwardNeuralNet</code> for each multi-dimension feature. These classifiers
	 * are only created if the <i>using_multi_dimensional_features</i> field is set to true.
	 * These classifiers are created in the order that the elligible mult-dimensional features appear
	 * in the <code>FeatureSettings</code> that were passed to this object dureing instantiation.
	 * The networks are each given an identifier that consists of the name of the feature that they are
	 * associated with.
	 *
	 * <p>The <i>training_recordings</i> parameters should contain the recordings, with features
	 * already extracted, that will be used to train the classifier ensemble.
	 *
	 * <p>The <i>model_classifications</i> parameter specifies the model classifications for
	 * each of the recordings in the <i>training_recordings</i> parameter. The first indice 
	 * identifies the related recording, and the second indice corresponds to the category(is)
	 * to which this recording should belong. These categories must be a subset of the
	 * <i>possible_taxonomy_labels</i> field, of course. Although the recordings do contain the
	 * category(ies) they belong to, this parameter is nonetheless needed in order to facilitate
	 * classifications at different levels of the taxonomical hierarchy.
	 *
	 * <p>The <i></i> and <i></i> fields are also calculated (i.e. classifier selection and
	 * classifying weighting are performed, if the user has selected these options.
	 *
	 * <p>The returned 2-D array of doubles is a set of classification errors after training iterations.
	 * The first indice indicates the classifier. Indice 0 corresponds to the
	 * <code>BioKNearestNeighbour</code> (null if not used) and the others, except for the last
	 * two, correspond to the <code>FeedForwardNeuralNetworks</code> in the order that they occur in the 
	 * <i>multi_dim_feature_labels</i> field (not present if multi-dimensional features
	 * are not used). The second to last indice value gives the error for classification selection and
	 * the last indice value gives the error for classifier weighting (even if the
	 * <i>ensemble_using_classifier_selection</i> and <i>ensemble_using_classifier_weighting</i>
	 * fields are false). The second indice of the returned array corresponds to the iteration of
	 * training that the error is associated with.
	 *
	 * <p>The <i>training_monitor</i> parameter is used to monitor training progress. It may
	 * be null if it is not used.
	 *
	 * <p> The <i>results_interpereter</i> parameter is used to calculate some fitnesses with
	 * genetic algorithms. It may be null if the user wants to use alternative fitness measures
	 * not based on actual classifications.
	 *
	 * <p>A variety of informative exceptions are thrown if incompatible information
	 * is being used.
	 *
	 * <p><b>WARNING:</b> due to the nature of the classifiers used, it is best to only
	 * train objects of this class once. In order to ensure this, the <code>train</code>
	 * method throws an exception if it is called and the classifier has already been
	 * trained.
	 */
	public double[][] train( Recording[] training_recordings,
		                     String[][] model_classifications,
	                         ProgressBarTaskTrainMonitor training_monitor,
	                         ClassificationResultsInterpereter results_interpereter )
		throws Exception
	{
		// Throw an exception if the ensemble has already been trained
		if (already_trained)
			throw new Exception ("Ensemble of classifiers has already been trained.\n" +
			                     "Retraining is not permitted. Please create a new\n" +
			                     "ensemble of classifiers if you wish to train with\n" +
			                     "new data.");
		
		// Throws an exception if no recordings are specified.
		if (training_recordings.length == 0)
			throw new Exception("No recordings have been specified for use in training this set of " +
			                    "classifiers.");

		// Train BioKNearestNeighbour, if one-dimensional features are set to be used
		// Store the training error
		double[] one_dimensional_training_errors = null;
		if (using_one_dimensional_features)
		{
			// Find the set of one dimensional feature labels used in the training_recordings
			String[] labels_of_features = getFeatureLabelsFromMultipleRecordings(training_recordings);

			// Instantiate the BioKNearestNeighbour. An exception is thrown if all of the recordings
			// do not contain the same one-dimensional features in the same order
			bio_knn_classifier = new BioKNearestNeighbour( labels_of_features,
			                                               possible_taxonomy_labels,
			                                               "AllOneDimensionalFeatures",
			                                               bioknn_using_feature_selection,
			                                               bioknn_using_feature_weighting,
			                                               ga_settings,
			                                               csw_training_fraction,
			                                               results_interpereter );
			bio_knn_classifier.setTrainingMonitor(training_monitor);

			// Construct the formatted training data
			double[][] training_data = getFeatureSetsOfOneDimensionalFeatures(training_recordings, labels_of_features.length);

			// Train the BioKNearestNeighbour. Note that the getFeatureLabelsFromMultipleRecordings
			// method has already ensured that the recordings hold the same one-dimensional features
			// in the same order.
			one_dimensional_training_errors = bio_knn_classifier.train( training_data,
			                                                            labels_of_features,
			                                                            model_classifications,
			                                                            bioknn_iterations,
			                                                            bioknn_acceptable_threshold,
			                                                            bioknn_consecutive_iterations );
		}

		// Train FeedForwardNeuralNets, if multi-dimensional features are set to be used.
		// Store the training errors
		double[][] multi_dimensional_training_errors = null;
		if (using_multi_dimensional_features)
		{
			// Prepare classifiers
			neural_network_classifiers = new FeedForwardNeuralNetwork[multi_dim_feature_labels.length];

			// Prepare the error report
			multi_dimensional_training_errors = new double[multi_dim_feature_labels.length][];

			// Go through the multi-dimensional features one by one
			for (int mdfeat = 0; mdfeat < multi_dim_feature_labels.length; mdfeat++)
			{
				// Instantiate the network
				neural_network_classifiers[mdfeat] = generateFeedForwardNeuralNetwork( multi_dim_feature_labels[mdfeat],
				                                                                       possible_taxonomy_labels,
				                                                                       training_recordings[0] );
				neural_network_classifiers[mdfeat].setTrainingMonitor(training_monitor);

				// Update the training_monitor
				if (training_monitor != null)
					training_monitor.startBottomTask(Math.abs(network_iterations), "Training NN: " + multi_dim_feature_labels[mdfeat] + ".");
			
				// Train the network
				multi_dimensional_training_errors[mdfeat] = trainFeedForwardNeuralNetwork( neural_network_classifiers[mdfeat],
				                                                                           training_recordings,
				                                                                           model_classifications );
			}
		}

		// Note that classification is complete
		already_trained = true;

		// Calculate the appropriate values for the ensemble_selection and
		// ensemble_weightings fields
		double[][] ensemble_errors = setClassifierSelectionAndWeightings(training_recordings, model_classifications, training_monitor, results_interpereter);

		// Construct the error report information to return
		int number_reports = 3;
		if (using_multi_dimensional_features)
			number_reports += multi_dimensional_training_errors.length;
		double[][] combined_training_errors = new double[number_reports][];
		combined_training_errors[0] = one_dimensional_training_errors;
		if (using_multi_dimensional_features)
			for (int i = 0; i < multi_dimensional_training_errors.length; i++)
				combined_training_errors[i+1] = multi_dimensional_training_errors[i];
		combined_training_errors[ combined_training_errors.length - 2 ] = ensemble_errors[0];
		combined_training_errors[ combined_training_errors.length - 1 ] = ensemble_errors[1];

		// Return the combined training errors
		return combined_training_errors;
	}


	/**
	 * Returns the relative scores of each of the possible categories when the given recordings
	 * are classified. There is one entry in the returned array for each of the
	 * entries in the <i>possible_taxonomy_labels</i> field, and they appear in the same order.
	 * Higher scores in the returned array correspond to a greater certainty that the feature set
	 * should have the corresponding label. Scores should fall in the range between 0.0 and 1.0.
	 * The first indice of the returned array corresponds to the feature set and the second
	 * corresponds to the category. The order of the categories is the same as in the 
	 * <i>possible_taxonomy_labels</i> field. These can be accessed with the <code>getCategories</code>
	 * method. The given scores are after classifier selection and weighting have been applied.
	 *
	 * <p>The <i>test_recordings</i> parameter specifies the recordings to be classified.
	 *
	 * <p>The <i>individual_classifier_results</i> parameter is not actually needed for classification.
	 * If the user wishes to see the scores arrived at by each classifier, then s/he should pass
	 * the following individual_classifier_results to this parameter:
	 *
	 * <p>double[][][] individual_classifier_results = new double[(classifier_ensemble.getEnsembleSelection()).length][][];
	 *
	 * <p>where classifier_ensemble is an object of this class. This method will then fill this
	 * array to hold the uncombined results of each classifier in the ensemble. The first indice
	 * specifies the classifier, the second specifies the recording and the third specifies the
	 * category. The entry for a classifier will be null if it has been eliminated during feature
	 * selection or if it is the one dimensional classifier and is not set to be used. The
	 * order of the classifiers can be accessed with the <code>getClassifierNames</code> method.
	 * A vaule of null can be passed to this parameter if the user does not care about individual
	 * classifier results.
	 *
	 * <p>An exception if thrown if the test_recordings are missing needed features or if the
	 * classifiers have not yet been trained.
	 */
	public double[][] classify(Recording[] test_recordings, double[][][] individual_classifier_results)
		throws Exception
	{
		if (!already_trained)
			throw new Exception ("A needed ensemble of classifiers has not yet been trained.\n" +
			                     "All necessary ensembles of classifiers must be trained\n" +
			                     "before classification can proceed.");

		// The raw results for each classifier
		if (individual_classifier_results == null)
			individual_classifier_results = new double[ensemble_selection.length][][];

		// Obtain classification using one_dimensional features if this option is set
		// Set to classifier results null if classifier not used.
		if (ensemble_selection[0])
		{
			// Find the names of all the extracted one-dimensional features in the recordings.
			// An exception is thrown if the recordings do not all contain the same extracted
			// features in the same order.
			String[] feature_labels = getFeatureLabelsFromMultipleRecordings(test_recordings);

			// Construct the set of extracted feature values in each of the recordings for each
			// of the features in feature_labels (i.e. all one-dimensional features). Note that
			// the getFeatureLabelsFromMultipleRecordings method has already ensured that
			// the same one-dimensional features are stored in the same order in all of the
			// recordings.
			double[][] feature_sets = getFeatureSetsOfOneDimensionalFeatures(test_recordings, feature_labels.length);

			// Feed the feature sets into the classifier
			individual_classifier_results[0] = bio_knn_classifier.classify(feature_sets, feature_labels);
		}
		else
			individual_classifier_results[0] = null;

		// Obtain classification using multi-dimensional features if this option is set
		// Set to classifier results null if classifier not used.
		if (using_multi_dimensional_features)
		{
			for (int clas = 0; clas < neural_network_classifiers.length; clas++)
			{
				if (!ensemble_selection[clas + 1])
					individual_classifier_results[clas + 1] = null;
				else
				{
					// Find the name of the feature that this network is built for
					String feature_label = neural_network_classifiers[clas].getClassifierIdentifier();

					// Get the properly formatted feature sets for the network
					double[][] feature_sets = getFeatureValues(feature_label, test_recordings);

					// Feed the feature sets into the classifier
					individual_classifier_results[clas + 1] = neural_network_classifiers[clas].classify(feature_sets, null);
				}
			}
		}
		
		// Calculate the combined ensebmle results using the appropriate classifier weightings
		double[][] combined_results = new double[test_recordings.length][possible_taxonomy_labels.length];
		for (int rec = 0; rec < test_recordings.length; rec++)
		{
			for (int cat = 0; cat < possible_taxonomy_labels.length; cat++)
			{
				combined_results[rec][cat] = 0.0;
				for (int clas = 0; clas < individual_classifier_results.length; clas++)
				{
					if (individual_classifier_results[clas] != null)
						combined_results[rec][cat] += individual_classifier_results[clas][rec][cat] * ensemble_weightings[clas];
				}
			}
		}
	
		// Return the raw results of the classification
		return combined_results;
	}


	/**
	 * Returns an array that represents the values of the one-dimensional features in the
	 * recordings passed in the <i>recordings</i> parameter after having been scaled
	 * to fall between 0 and 1 based on previous training of the <code>BioKNearestNeighbour</code>.
	 * 
	 * <p>The first indice of the returned array identifies the recording and the second identifies
	 * the feature. The order of the features in the returned array is the same as the order of the
	 * features in the <i>recordings</i>
	 *
	 * <p>The acual values of the features in <i>recordings</i> parameter are not themselves changed.
	 *
	 * <p>An exception if thrown if the <i>recordings</i> parameter contains unknown features
	 * or if the <code>BioKNearestNeighbour</code> has not yet been trained.
	 */
	public double[][] getScaledOneDimensionalFeatureValues(Recording[] recordings)
		throws Exception
	{
		// Return exception if classifier has not been trained
		if (bio_knn_classifier == null)
			throw new Exception("Cannot get scaled feature values because one-dimensional" +
			                    "classifier has not been trained.");

		// Find the names of all the extracted one-dimensional features in the recordings.
		// An exception is thrown if the recordings do not all contain the same extracted
		// features in the same order.
		String[] feature_labels = getFeatureLabelsFromMultipleRecordings(recordings);

		// Construct the set of extracted feature values in each of the recordings for each
		// of the features in feature_labels (i.e. all one-dimensional features). Note that
		// the getFeatureLabelsFromMultipleRecordings method has already ensured that
		// the same one-dimensional features are stored in the same order in all of the
		// recordings.
		double[][] feature_sets = getFeatureSetsOfOneDimensionalFeatures(recordings, feature_labels.length);

		// Return exception if no available one-dimensional features
		if (feature_sets == null)
			throw new Exception("No one-dimensional features specified for scaling.");

		// Find the scaled feature values
		double[][] scaled_values = bio_knn_classifier.getScaledFeatureValues(feature_sets, feature_labels);
		
		// Return the results
		return scaled_values;
	}


	/**
	 * Returns the minimum allowable value for each of the one-dimensional features
	 * in the trained <code>BioKNearestNeighbour</code>. Features below
	 * their respective values are rounded up to it. Throws an exception if the
	 * <code>BioKNearestNeighbour</code> has not been trained yet.
	 */
	public double[] getMinFeatureValueCutoffs()
		throws Exception
	{
		// Return exception if classifier has not been trained
		if (bio_knn_classifier == null)
			throw new Exception("Cannot get scaled feature values because one-dimensional" +
			                    "classifier has not been trained.");

		return bio_knn_classifier.getMinFeatureValueCutoffs();
	}


	/**
	 * Returns the maximum allowable value for each of the one-dimensional features
	 * in the trained <code>BioKNearestNeighbour</code>. Features above
	 * their respective values are rounded down to it. Throws an exception if the
	 * <code>BioKNearestNeighbour</code> has not been trained yet.
	 */
	public double[] getMaxFeatureValueCutoffs()
		throws Exception
	{
		// Return exception if classifier has not been trained
		if (bio_knn_classifier == null)
			throw new Exception("Cannot get scaled feature values because one-dimensional" +
			                    "classifier has not been trained.");

		return bio_knn_classifier.getMaxFeatureValueCutoffs();
	}


	/**
	 * Returns the features selected for use by the <code>BioKNearestNeighbour</code> classifier.
	 * A value of true means that the feature was selected and a value of false means that it was not.
	 * The <code>getNamesOfFeaturesUsedByOneDimensionalClassifier</code> method should
	 * be called to obtain the names of the features that the indices refer to.
	 */
	public boolean[] getOneDimensionalFeatureSelection()
	{
		if (bio_knn_classifier == null)
			return null;
		else
			return bio_knn_classifier.getFeatureSelection();
	}


	/**
	 * Returns the features weightings used by the <code>BioKNearestNeighbour</code> classifier.
	 * The <code>getNamesOfFeaturesUsedByOneDimensionalClassifier</code> method should
	 * be called to obtain the names of the features that the indices refer to.
	 */
	public double[] getOneDimensionalFeatureWeightings()
	{
		if (bio_knn_classifier == null)
			return null;
		else
			return bio_knn_classifier.getFeatureWeights();
	}


	/**
	 * Returns the names of the features used by the <code>BioKNearestNeighbour</code> classifier.
	 * in the order that they are used
	 */
	public String[] getNamesOfFeaturesUsedByOneDimensionalClassifier()
	{
		if (bio_knn_classifier == null)
			return null;
		else
			return bio_knn_classifier.getFeatureNames();
	}


	/**
	 * Saves all of the fields except ga_settings and network_settings to the given File. A separate file is created
	 * for the bio_knn_classifier and a subfolder with separate files are created for neural_network_classifiers.
	 */
	public void	save(File place_to_save)
		throws Exception
	{
		// Create sub-folder for ensemble
		String sub_folder_path = bodhidharma.utilities.StringMethods.removeExtension(place_to_save.getPath()) + File.separator;
		File sub_folder = new File(sub_folder_path);
		if (sub_folder.isFile() == false && sub_folder.isDirectory() == false)
			sub_folder.mkdir();

		// Create file for bio_knn_classifier
		String bio_knn_classifier_file_name;
		if (bio_knn_classifier == null)
			bio_knn_classifier_file_name = "null";
		else
		{
			bio_knn_classifier_file_name = sub_folder_path + "bio_knn_classifier.xml";
			File bio_knn_classifier_file = new File(bio_knn_classifier_file_name);
			bio_knn_classifier.save(bio_knn_classifier_file);
		}
		
		// Create files for neural_network_classifiers
		String[] neural_network_classifiers_file_names;
		if (neural_network_classifiers == null)
		{
			neural_network_classifiers_file_names = new String[1];
			neural_network_classifiers_file_names[0] = "null";
		}
		else
		{
			neural_network_classifiers_file_names = new String[neural_network_classifiers.length];
			for (int i = 0; i < neural_network_classifiers_file_names.length; i++)
			{
				neural_network_classifiers_file_names[i] = sub_folder_path + "feed_forward_neural_network_" + i + ".xml";
				File neural_network_classifiers_file = new File(neural_network_classifiers_file_names[i]);
				neural_network_classifiers[i].save(neural_network_classifiers_file);
			}
		}

		// Save the NN_BioKNN_Ensemble file
		FileOutputStream to = new FileOutputStream(place_to_save);
		DataOutputStream writer = new DataOutputStream(to);
		String header = new String
		(
			"<?xml version=\"1.0\"?>\n" +
			"<!DOCTYPE NN_BioKNN_Ensemble_file [\n" +
			"   <!ELEMENT NN_BioKNN_Ensemble_file (comments, identifier, BioKNearestNeighbour_file_path, FeedForwardNeuralNetwork_file_paths, ensemble_selection, ensemble_weightings, already_trained, possible_taxonomy_labels, one_dim_feature_labels, multi_dim_feature_labels, ensemble_using_classifier_selection, ensemble_using_classifier_weighting, csw_training_fraction, using_one_dimensional_features, using_multi_dimensional_features, bioknn_using_feature_selection, bioknn_using_feature_weighting, bioknn_iterations, bioknn_acceptable_threshold, bioknn_consecutive_iterations, network_iterations, network_acceptable_threshold, network_consecutive_iterations)>\n" +
			"   <!ELEMENT comments (#PCDATA)>\n" +
			"   <!ELEMENT identifier (#PCDATA)>\n" +
			"   <!ELEMENT BioKNearestNeighbour_file_path (#PCDATA)>\n" +
			"   <!ELEMENT FeedForwardNeuralNetwork_file_paths (paths*)>\n" +
			"   <!ELEMENT paths (#PCDATA)>\n" +
			"   <!ELEMENT ensemble_selection (used*)>\n" +
			"   <!ELEMENT used (#PCDATA)>\n" +
			"   <!ELEMENT ensemble_weightings (weight*)>\n" +
			"   <!ELEMENT weight (#PCDATA)>\n" +
			"   <!ELEMENT already_trained (#PCDATA)>\n" +
			"   <!ELEMENT possible_taxonomy_labels (label*)>\n" +
			"   <!ELEMENT label (#PCDATA)>\n" +
			"   <!ELEMENT one_dim_feature_labels (label*)>\n" +
			"   <!ELEMENT multi_dim_feature_labels (label*)>\n" +
			"   <!ELEMENT ensemble_using_classifier_selection (#PCDATA)>\n" +
			"   <!ELEMENT ensemble_using_classifier_weighting (#PCDATA)>\n" +
			"   <!ELEMENT csw_training_fraction (#PCDATA)>\n" +
			"   <!ELEMENT using_one_dimensional_features (#PCDATA)>\n" +
			"   <!ELEMENT using_multi_dimensional_features (#PCDATA)>\n" +
			"   <!ELEMENT bioknn_using_feature_selection (#PCDATA)>\n" +
			"   <!ELEMENT bioknn_using_feature_weighting (#PCDATA)>\n" +
			"   <!ELEMENT bioknn_iterations (#PCDATA)>\n" +
			"   <!ELEMENT bioknn_acceptable_threshold (#PCDATA)>\n" +
			"   <!ELEMENT bioknn_consecutive_iterations (#PCDATA)>\n" +
			"   <!ELEMENT network_iterations (#PCDATA)>\n" +
			"   <!ELEMENT network_acceptable_threshold (#PCDATA)>\n" +
			"   <!ELEMENT network_consecutive_iterations (#PCDATA)>\n" +
			"]>\n\n" +
			"<NN_BioKNN_Ensemble_file>\n\n" +
			"   <comments></comments>\n\n"
		);
		writer.writeBytes(header);
		writer.writeBytes("   <identifier>" + identifier + "</identifier>\n");
		writer.writeBytes("   <BioKNearestNeighbour_file_path>" + bio_knn_classifier_file_name + "</BioKNearestNeighbour_file_path>\n");
		writer.writeBytes("   <FeedForwardNeuralNetwork_file_paths>\n");
		if (neural_network_classifiers_file_names != null)
			for (int i = 0; i < neural_network_classifiers_file_names.length; i++)
				writer.writeBytes("      <paths>" + neural_network_classifiers_file_names[i] + "</paths>\n");
		writer.writeBytes("   </FeedForwardNeuralNetwork_file_paths>\n");
		writer.writeBytes("   <ensemble_selection>\n");
		if (ensemble_selection != null)
			for (int i = 0; i < ensemble_selection.length; i++)
				writer.writeBytes("      <used>" + ensemble_selection[i] + "</used>\n");
		writer.writeBytes("   </ensemble_selection>\n");
		writer.writeBytes("   <ensemble_weightings>\n");
		if (ensemble_weightings != null)
			for (int i = 0; i < ensemble_weightings.length; i++)
				writer.writeBytes("      <weight>" + ensemble_weightings[i] + "</weight>\n");
		writer.writeBytes("   </ensemble_weightings>\n");
		writer.writeBytes("   <already_trained>" + already_trained + "</already_trained>\n");
		writer.writeBytes("   <possible_taxonomy_labels>\n");
		if (possible_taxonomy_labels != null)
			for (int i = 0; i < possible_taxonomy_labels.length; i++)
				writer.writeBytes("      <label>" + possible_taxonomy_labels[i] + "</label>\n");
		writer.writeBytes("   </possible_taxonomy_labels>\n");
		writer.writeBytes("   <one_dim_feature_labels>\n");
		if (one_dim_feature_labels != null)
			for (int i = 0; i < one_dim_feature_labels.length; i++)
				writer.writeBytes("      <label>" + one_dim_feature_labels[i] + "</label>\n");
		writer.writeBytes("   </one_dim_feature_labels>\n");
		writer.writeBytes("   <multi_dim_feature_labels>\n");
		if (multi_dim_feature_labels != null)
			for (int i = 0; i < multi_dim_feature_labels.length; i++)
				writer.writeBytes("      <label>" + multi_dim_feature_labels[i] + "</label>\n");
		writer.writeBytes("   </multi_dim_feature_labels>\n");
		writer.writeBytes("   <ensemble_using_classifier_selection>" + ensemble_using_classifier_selection + "</ensemble_using_classifier_selection>\n");
		writer.writeBytes("   <ensemble_using_classifier_weighting>" + ensemble_using_classifier_weighting + "</ensemble_using_classifier_weighting>\n");
		writer.writeBytes("   <csw_training_fraction>" + csw_training_fraction + "</csw_training_fraction>\n");
		writer.writeBytes("   <using_one_dimensional_features>" + using_one_dimensional_features + "</using_one_dimensional_features>\n");
		writer.writeBytes("   <using_multi_dimensional_features>" + using_multi_dimensional_features + "</using_multi_dimensional_features>\n");
		writer.writeBytes("   <bioknn_using_feature_selection>" + bioknn_using_feature_selection + "</bioknn_using_feature_selection>\n");
		writer.writeBytes("   <bioknn_using_feature_weighting>" + bioknn_using_feature_weighting + "</bioknn_using_feature_weighting>\n");
		writer.writeBytes("   <bioknn_iterations>" + bioknn_iterations + "</bioknn_iterations>\n");
		writer.writeBytes("   <bioknn_acceptable_threshold>" + bioknn_acceptable_threshold + "</bioknn_acceptable_threshold>\n");
		writer.writeBytes("   <bioknn_consecutive_iterations>" + bioknn_consecutive_iterations + "</bioknn_consecutive_iterations>\n");
		writer.writeBytes("   <network_iterations>" + network_iterations + "</network_iterations>\n");
		writer.writeBytes("   <network_acceptable_threshold>" + network_acceptable_threshold + "</network_acceptable_threshold>\n");
		writer.writeBytes("   <network_consecutive_iterations>" + network_consecutive_iterations + "</network_consecutive_iterations>\n");
		writer.writeBytes("</NN_BioKNN_Ensemble_file>");
		writer.close();
	}

	
	/* PRIVATE METHODS *********************************************************/


	// Return a 2-D String array that holds the features that are candidates for
	// feature selection. The first indice is either 0 or 1. 0 Corresponds to one-
	// dimensional features and 1 corresponds to multi-dimensional features.
	// The second indice specifies the particular feature. The candidate features
	// are derived from the user's settings in the FeaturesPanel.
	private String[][] getCandidateFeatures(FeatureSettings[] feature_settings)
	{
		// Find the number of each kind of feature that are candidates for feature
		// selection
		int number_one_dim_features = 0;
		int number_multi_dim_features = 0;
		for (int i = 0; i < feature_settings.length; i++)
		{
			if (feature_settings[i].candidate_feature)
			{
				if (feature_settings[i].getFeature() instanceof FeatureOneDimensional)
					number_one_dim_features++;
				else if (feature_settings[i].getFeature() instanceof FeatureMultiDimensional)
					number_multi_dim_features++;
			}
		}

		// Find the names of each feature that is a candidate for feature selection
		String[][] results = new String[2][];
		results[0] = new String[number_one_dim_features];
		results[1] = new String[number_multi_dim_features];
		int current_one = 0;
		int current_multi = 0;
		for (int i = 0; i < feature_settings.length; i++)
		{
			if (feature_settings[i].candidate_feature)
			{
				if (feature_settings[i].getFeature() instanceof FeatureOneDimensional)
				{
					results[0][current_one] = feature_settings[i].getFeature().getName();
					current_one++;
				}
				else if (feature_settings[i].getFeature() instanceof FeatureMultiDimensional)
				{
					results[1][current_multi] = feature_settings[i].getFeature().getName();
					current_multi++;
				}
			}
		}

		// Return the candidate feature names
		return results;
	}


	// Return a 2-D String array that holds the features that are set for use
	// in overide mode (see description of FeaturesPanel class for explanation of
	// overide mode). The first indice is either 0 or 1. 0 Corresponds to one-
	// dimensional features and 1 corresponds to multi-dimensional features.
	// The second indice specifies the particular feature. The candidate features
	// are derived from the user's settings in the FeaturesPanel.
	private String[][] getOverideFeatures(FeatureSettings[] feature_settings)
	{
		// Find the number of each kind of feature that is used in overide mode
		int number_one_dim_features = 0;
		int number_multi_dim_features = 0;
		for (int i = 0; i < feature_settings.length; i++)
		{
			if (feature_settings[i].overide_feature)
			{
				if (feature_settings[i].getFeature() instanceof FeatureOneDimensional)
					number_one_dim_features++;
				else if (feature_settings[i].getFeature() instanceof FeatureMultiDimensional)
					number_multi_dim_features++;
			}
		}

		// Find the names of each feature that is used in overide mode
		String[][] results = new String[2][];
		results[0] = new String[number_one_dim_features];
		results[1] = new String[number_multi_dim_features];
		int current_one = 0;
		int current_multi = 0;
		for (int i = 0; i < feature_settings.length; i++)
		{
			if (feature_settings[i].overide_feature)
			{
				if (feature_settings[i].getFeature() instanceof FeatureOneDimensional)
				{
					results[0][current_one] = feature_settings[i].getFeature().getName();
					current_one++;
				}
				else if (feature_settings[i].getFeature() instanceof FeatureMultiDimensional)
				{
					results[1][current_multi] = feature_settings[i].getFeature().getName();
					current_multi++;
				}
			}
		}

		// Return the overide feature names
		return results;
	}


	// Return an array holding the names of the one dimensional features that are contained
	// in any one the given recordings in the order that they are contained. All of the 
	// recordings must hold the same set of extracted features in the same order. An exception
	// is thrown if they do not. An exception is also thrown if any of the recordings do not
	// hold any extracted features or if there are no recordings present.
	private	String[] getFeatureLabelsFromMultipleRecordings(Recording[] recordings)
		throws Exception
	{
		// Throw an exception if the recordings parameter is null or empty
		if (recordings == null)
			throw new Exception("Cannot process features from an empty set of recordings.");
		else if (recordings.length == 0)
			throw new Exception("Cannot process features from an empty set of recordings.");

		// Find the set of one-dimensional feature labels found in the first recording
		String[] feature_labels = getFeatureLabelsFromOneRecording(recordings[0]);

		// Ensure that all recordings possess the same one-dimensional features in the
		// same order and throw an exception if they do not
		for (int rec = 0; rec < recordings.length; rec++)
		{
			String current_recording[] = getFeatureLabelsFromOneRecording(recordings[rec]);
			if (current_recording == null)
				throw new Exception("The recording " + recordings[rec].title + " does not contain any " +
				                    "extracted features.");
			if (current_recording.length != feature_labels.length)
				throw new Exception("The recording " + recordings[0].title + " has " +
				                    recordings[0].feature_set.features.length + " extracted features " +
				                    "but the recording " + recordings[rec].title + " has " +
				                    recordings[0].feature_set.features.length + "extracted features. " +
				                    " All recordings must have the same number of extracted features.");
			for (int feat = 0; feat < feature_labels.length; feat++)
			{
				if ( !current_recording[feat].equals(feature_labels[feat]) )
					throw new Exception("The recording " + recordings[0].title + " has different features " +
										"extracted or features extracted in a different order than the " +
					                    "recordings" + recordings[rec].title + ". All recordings must have " +
					                    "the same features in the same order.");
			}
		}

		// Return the set of feature labels
		return feature_labels;
	}


	// Return an array of Strings containing the names of all of the one-dimensional features
	// stored in the given Recording in the order that they are stored. Null is returned
	// if no features are held by the given recording.
	private String[] getFeatureLabelsFromOneRecording(Recording recording)
	{
		// Return null if no features are held by the recording
		if (recording.feature_set.features.length == 0)
			return null;
		
		// Find the number of one-dimensional features
		int number_one_dimensional_features = 0;
		for (int feat = 0; feat < recording.feature_set.features.length; feat++)
			if (recording.feature_set.features[feat] instanceof FeatureOneDimensional)
				number_one_dimensional_features++;
		
		String[] feature_labels = new String[number_one_dimensional_features];
		int current = 0;
		for (int feat = 0; feat < recording.feature_set.features.length; feat++)
			if (recording.feature_set.features[feat] instanceof FeatureOneDimensional)
			{
				feature_labels[current] = recording.feature_set.features[feat].getName();
				current++;
			}
		return feature_labels;
	}


	// Return the first number_of_features extracted one-dimensional features in each of the recordings
	// in recordings. The first indice of the returned array identifies the recording and the second 
	// identifies the features. Note that it must be ensured outside of this method that all recordings
	// have at least number_of_features extracted one-dimensional recordings.
	private double[][] getFeatureSetsOfOneDimensionalFeatures(Recording[] recordings, int number_features)
	{
		double[][] feature_sets = new double[recordings.length][number_features];
		for (int rec = 0; rec < recordings.length; rec++)
		{
			int count = 0;
			for (int feat = 0; feat < recordings[rec].feature_set.features.length; feat++)
			{
				if (recordings[rec].feature_set.features[feat] instanceof FeatureOneDimensional)
				{
					feature_sets[rec][count] = ((FeatureOneDimensional) recordings[rec].feature_set.features[feat]).getValue();
					count++;
				}
			}
		}
		return feature_sets;
	}


	// Generates and returns an untrained FeedForwardNeuralNetwork using:
	//    - The multi-dimensional feature named in the feature_label parameter.
	//    - The taxonomy labels specified in the possible_taxonomy_labels parameter. The
	//      network will be trained to classify samples into one or more of these labels.
	//    - A sample recording specified in the sample_recording parameter. This recording
	//      is only used to determine the number of dimensions of the feature.
	// The FeedForwardNeuralNetwork that is returned is untrained and its identifeier is
	//      labelled with the name given in the feature_label parameter. The network is
	//      configured to accept each of the dimensions of a feature of this type as inputs
	//      and output the category(ies) that the sample is classified as.
	// Throws a descriptive exception if:
	//   - Any of the recordings do not contain a required feature.
	private FeedForwardNeuralNetwork generateFeedForwardNeuralNetwork(String feature_label,
	                                                                  String[] possible_taxonomy_labels,
	                                                                  Recording sample_recording)
		throws Exception
	{
		// Find the number of dimensions of the feature and generate generic names
		// for the network's feature labels
		int number_dimensions = 0;
		Feature[] sample_recording_features = sample_recording.feature_set.features;
		for (int feat = 0; feat < sample_recording_features.length; feat++)
			if ( (sample_recording_features[feat].getName()).equals( feature_label ) )
				number_dimensions = ((FeatureMultiDimensional) sample_recording_features[feat]).getValue().length;
		if (number_dimensions < 1)
			throw new Exception("The recording " + sample_recording.title + " does not have " +
								"the required feature " + feature_label + ".");
		String[] labels_of_features = new String[number_dimensions];
		for (int i = 0; i < labels_of_features.length; i++)
			labels_of_features[i] = "Dimension " + String.valueOf(i);

		// Instantiate the network
		FeedForwardNeuralNetwork network = new FeedForwardNeuralNetwork( labels_of_features,
		                                                                 possible_taxonomy_labels,
		                                                                 feature_label,
		                                                                 network_settings );
		
		// Return the untrained network
		return network;
	}



	// Trains an existing FeedForwardNeuralNetwork using:
	//    - The network passed to the network parameter. This network must have its
	//      identifier field set to the multi-dimensional feature that was used to train it.
	//    - The recordings specified in the recordings parameter. These are used to train
	//      the clasifiers.
	//    - The model categories, as specified in the model_classifications parameter,
	//      for each of the recordings specified in the recordings parameter. The first
	//      indice identifies the related recording, and the second indice corresponds
	//      to the category(is) to which this recording should belong. These categories
	//      must be a subset of the possible_taxonomy_labels parameter labels, of course.
	//      Although the recordings do contain the category(ies) they belong to, this
	//      parameter is nonetheless needed in order to enable classifications at different
	//      levels of the hierarchy.
	// An array of doubles is returned containing the average sum of squared error after
	//    training iterations. The indice of the returned array corresponds to the iteration
	//    of training that the error is associated with.
	// Throws a descriptive exception if:
	//   - Any of the recordings do not contain the required feature.
	//   - The dimensionality of the feature is inconsistent accross recordings.
	private double[] trainFeedForwardNeuralNetwork(FeedForwardNeuralNetwork network,
	                                               Recording[] recordings,
	                                               String[][] model_classifications)
		throws Exception
	{
		// Find the name of the feature that this network is built for
		String feature_label = network.getClassifierIdentifier();

		// Get the properly formatted training values for the network
		double[][] training_data = getFeatureValues(feature_label, recordings);

		// Set the training parameters
		int iterations = 200;
		double acceptable_threshold = 0.001;
		int consecutive_iterations = 10;

		// Train the network
		double[] errors = network.train( training_data,
		                                 null,
		                                 model_classifications,
		                                 network_iterations,
		                                 network_acceptable_threshold, 
		                                 network_consecutive_iterations );

		// Return the set of sum of squared errors for each iteration
		return errors;
	}


	// Construct a 2-D array of doubles containing the value(s) of the feature named
	// in the <i>feature_label</i> parameter from the recordings found in the
	// <i>recording</i> parameter. The first indice of the returned array corresponds
	// to the related recording and the second indice identifies the values
	// of the different dimensions of the feature for the related recording.
	// Throws a descriptive exception if:
	//   - Any of the recordings do not contain the required feature.
	//   - The dimensionality of the feature is inconsistent accross recordings.
	private double[][] getFeatureValues(String feature_label, Recording[] recordings)
		throws Exception
	{
		// Find the feature values for  by going through the recordings one
		// by one and storing each dimension of the feature
		double[][] feature_values = new double[recordings.length][];
		for (int rec = 0; rec < recordings.length; rec++)
		{
			// Find the correct feature
			Feature[] recording_features = recordings[rec].feature_set.features;
			double[] values = null;
			for (int feat = 0; feat < recording_features.length; feat++)
			{
				if ( (recording_features[feat].getName()).equals( feature_label ) )
				{
					values = ((FeatureMultiDimensional) recording_features[feat]).getValue();
					feat = recording_features.length + 1; // exit the loop
				}
			}

			// Throw an exception if the feature is not found in the given recording
			if (feature_values == null)
				throw new Exception("The recording " + recordings[rec].title + " does not have " +
				                    "the required feature " + feature_label + ".");

			// Store the feature's values in feature_values
			feature_values[rec] = values;
		}

		// Make sure that the feature for each recording has the same size and throw
		// an exception if this is not the case
		int dimensions = feature_values[0].length;
		for (int rec = 0; rec < feature_values.length; rec++)
			if (feature_values[rec].length != dimensions)
				throw new Exception("The " + feature_label + " feature of the " + recordings[rec].title +
				                    " recording has a different number of dimensions than one or more " +
				                    " other recordings.");

		// Return the feature values
		return feature_values;
	}


	// Sets the ensemble_selection and ensemble_weightings fields based on the
	// given recordings and their model_classifications. Genetic algorithms
	// are used to do this. Alternatively, based on the values of the
	// ensemble_using_classifier_selection and ensemble_using_classifier_weighting
	// fields, all of the classifiers will be used by deafault and half of the
	// weighting will be automatically assigned to the BioKNearestNeighbour
	// classifier and the remaining weighting can be distributed equally among
	// FeedForwardNeuralNetworks.
	//
	// The ensemble_weightings array is normalized.
	//
	// The returned array contains the error for each epoch of the training.
	// The first indice indicates selection or weighting (0 and 1 respectively)
	// and the second indice indicates epoch.
	//
	// The training_monitor parameter is used to monitor training progress. It may
	// be null if it is not used.
	//
	// The <i>results_interpereter</i> parameter is used to calculate some fitnesses with
	// genetic algorithms. It may be null if the user wants to use alternative fitness measures
	// not based on actual classifications.
	private double[][] setClassifierSelectionAndWeightings( Recording[] training_recordings,
	                                                        String[][] model_classifications,
	                                                        ProgressBarTaskTrainMonitor training_monitor,
	                                                        ClassificationResultsInterpereter results_interpereter )
		throws Exception
	{
		// Make sure classifiers have been trained
		if (!already_trained)
			throw new Exception ("A needed ensemble of classifiers has not yet been trained.\n" +
			                     "All necessary ensembles of classifiers must be trained\n" +
			                     "before classifier selection and weighting can be performed.");

		// Find the total number of classifiers to find settings for
		int number_classifiers = 1 ;
		if (neural_network_classifiers != null)
			number_classifiers += neural_network_classifiers.length;

		// Instantiate ensemble_selection and ensemble_weightings
		ensemble_selection = new boolean[number_classifiers];
		ensemble_weightings = new double[number_classifiers];

		// Get the classification results and properly formatted model results for each of the classifiers
		// Entry 0 is set to null if one dimensional features are not used
		double[][][] individual_classifier_results = new double[number_classifiers][][];
		int[][]	int_model_results = new int[model_classifications.length][];
		if ( ensemble_using_classifier_selection || ensemble_using_classifier_weighting )
		{
			// Find classifier results
			if (using_one_dimensional_features)
			{
				String[] feature_labels = getFeatureLabelsFromMultipleRecordings(training_recordings);
				double[][] feature_sets = getFeatureSetsOfOneDimensionalFeatures(training_recordings, feature_labels.length);
				individual_classifier_results[0] = bio_knn_classifier.classify(feature_sets, feature_labels);
			}
			else
				individual_classifier_results[0] = null;
			if (using_multi_dimensional_features)
			{
				for (int clas = 0; clas < neural_network_classifiers.length; clas++)
				{
					String feature_label = neural_network_classifiers[clas].getClassifierIdentifier();
					double[][] feature_sets = getFeatureValues(feature_label, training_recordings);
					individual_classifier_results[clas + 1] = neural_network_classifiers[clas].classify(feature_sets, null);
				}
			}

			// Format model results in int form
			int_model_results = getIntModelResults(model_classifications);
		}

		// Set up ensemble_selection and keep track of error
		double[] selection_error = null;
		if (ensemble_using_classifier_selection)
		{
			if (!using_multi_dimensional_features)
			{
				ensemble_selection = new boolean[1];
				ensemble_selection[0] = true;
				selection_error = new double[1];
				selection_error[0] = 1.0;
			}
			else
			{
				// Update the training_monitor
				if (training_monitor != null)
					training_monitor.startBottomTask(((int) ( ((double) Math.abs(bioknn_iterations)) * classifier_vs_feature_multiplier)) , "Training classifier selection.");

				// Set up genetic algorithm
				ClassifierSelectionEvaluator evaluator = new ClassifierSelectionEvaluator( individual_classifier_results,
																						   int_model_results, 
																						   possible_taxonomy_labels.length,
				                                                                           results_interpereter );
				Breeder breeder = new Breeder( evaluator,
											   ((int) (((double) ga_settings.getPopulationSize()) * classifier_vs_feature_multiplier)),
											   ga_settings.getMutationProbability(),
											   ga_settings.getMatingRate(),
											   ga_settings.getElitism(),
											   ga_settings.getNumberVillages(),
											   ga_settings.getMigrationDelay(),
											   ga_settings.getMigrationSize() );

				// Carry out evolution and find the best performing classifier selection
				short[] best_classifier_selection = new short[evaluator.getBitStringLength()];
				selection_error = evolve(breeder, best_classifier_selection, training_monitor);

				// Set features_to_use to the settings of the best performing bit string
				Object[] ensemble_selection_objects = evaluator.parseBitString(best_classifier_selection);
				for (int clas = 0; clas < ensemble_selection_objects.length; clas++)
					ensemble_selection[clas] = ((Boolean) ensemble_selection_objects[clas]).booleanValue();
			}
		}
		else
		{
			for (int i = 0; i < ensemble_selection.length; i++)
				ensemble_selection[i] = true;
			if (!using_one_dimensional_features)
				ensemble_selection[0] = false;
			selection_error = new double[1];
			selection_error[0] = 1.0;
		}

		// Set up ensemble_weightings and keep track of error
		double[] weighting_error = null;
		if (ensemble_using_classifier_weighting)
		{
			if (!using_multi_dimensional_features)
			{
				ensemble_weightings = new double[1];
				ensemble_weightings[0] = 1.0;
				weighting_error = new double[1];
				weighting_error[0] = 1.0;
			}
			else
			{
				// Update the training_monitor
				if (training_monitor != null)
					training_monitor.startBottomTask(((int) (((double) Math.abs(bioknn_iterations)) * classifier_vs_feature_multiplier)), "Training classifier weighting.");

				// Set up genetic algorithm
				ClassifierWeightingEvaluator evaluator = new ClassifierWeightingEvaluator( individual_classifier_results,
																						   int_model_results, 
				                                                                           ensemble_selection,
																						   possible_taxonomy_labels.length,
				                                                                           ga_settings.getWordLength(),
				                                                                           results_interpereter );
				Breeder breeder = new Breeder( evaluator,
											   ((int) (((double) ga_settings.getPopulationSize()) * classifier_vs_feature_multiplier)),
											   ga_settings.getMutationProbability(),
											   ga_settings.getMatingRate(),
											   ga_settings.getElitism(),
											   ga_settings.getNumberVillages(),
											   ga_settings.getMigrationDelay(),
											   ga_settings.getMigrationSize() );

				// Carry out evolution and find the best performing classifier selection
				short[] best_classifier_selection = new short[evaluator.getBitStringLength()];
				weighting_error = evolve(breeder, best_classifier_selection, training_monitor);

				// Set features_to_use to the settings of the best performing bit string
				Object[] ensemble_weighting_objects = evaluator.parseBitString(best_classifier_selection);
				for (int clas = 0; clas < ensemble_weighting_objects.length; clas++)
					ensemble_weightings[clas] = ((Double) ensemble_weighting_objects[clas]).doubleValue();
			}
		}
		else
		{
			int number_multi_dims_selected_out = 0;
			for (int i = 1; i < ensemble_selection.length; i++)
				if (!ensemble_selection[i])
					number_multi_dims_selected_out--;
			double equal_weighting;
			if (using_one_dimensional_features && ensemble_selection[0])
			{
				ensemble_weightings[0] = 0.5;
				equal_weighting = 0.5 / ((double) (number_classifiers - 1 + number_multi_dims_selected_out));
			}
			else
			{
				ensemble_weightings[0] = 0.0;
				equal_weighting = 1.0 / ((double) (number_classifiers - 1 + number_multi_dims_selected_out));
			}
			if (using_one_dimensional_features && ensemble_selection[0] && using_multi_dimensional_features)
			{
				boolean other_than_one_d_selected = false;
				for (int i = 1; i < ensemble_selection.length; i++)
					if (ensemble_selection[i])
						other_than_one_d_selected = true;

				if (!other_than_one_d_selected)
				{
					ensemble_weightings[0] = 1.0;
					equal_weighting = 0.0;
				}
			}
			for (int i = 1; i < ensemble_weightings.length; i++)
			{
				if (ensemble_selection[i])
					ensemble_weightings[i] = equal_weighting;
				else
					ensemble_weightings[i] = 0;
			}
			if (!using_multi_dimensional_features)
				ensemble_weightings[0] = 1.0;
			weighting_error = new double[1];
			weighting_error[0] = 1.0;
		}

		// Set up and return errors
		double[][] combined_errors = new double[2][];
		combined_errors[0] = selection_error;
		combined_errors[1] = weighting_error;
		return combined_errors;
	}


	/**
	 * Returns a 2-D array whose first indice indicates recording and whose entries indicate the indices
	 * of the category(ies)in <i>possible_taxonomy_labels</i> that each recording belongs to in the
	 * <i>given_results</i> parameter.
	 *
	 * <p>An exception is thrown if the <i>given_results</i> parameter contains a name not
	 * present in the <i>possible_taxonomy_labels</i> field or if it contains the same category more than once.
	 */
	private int[][] getIntModelResults(String[][] given_results)
		throws Exception
	{
		// Instantiate the model results in boolean form
		boolean[][] model_results = new boolean[given_results.length][possible_taxonomy_labels.length];
		for (int i = 0; i < model_results.length; i++)
			for (int j = 0; j < model_results[i].length; j++)
				model_results[i][j] = false;

		// Match feature set categories with classifier categories
		for (int set = 0; set < given_results.length; set++)
		{
			for (int gcat = 0; gcat < given_results[set].length; gcat++)
			{
				boolean found = false;
				for (int mcat = 0; mcat < possible_taxonomy_labels.length; mcat++)
				{
					if ( given_results[set][gcat].equals( possible_taxonomy_labels[mcat] ) )
					{
						if (found)
							throw new Exception("The given model results contain the same " +
							                    "model category more than once for the same " +
							                    "feature set.");
						model_results[set][mcat] = true;
						found = true;
					}
				}
				if (!found)
					throw new Exception("The given model results contain a category called " +
					                    given_results[set][gcat] + " that is not accounted " +
					                    "for in the classifier.");
			}
		}

		// Find the model results in int form
		int[][] int_model_results = new int[model_results.length][];
		for (int rec = 0; rec < int_model_results.length; rec++)
		{
			// Find the number of categories that the recording belongs to
			int number_categories_recording_belongs_to = 0;
			for (int cat = 0; cat < model_results[rec].length; cat++)
				if (model_results[rec][cat])
					number_categories_recording_belongs_to++;

			// Store the indices of the categories that the recording belongs to
			// in int_model_results
			int_model_results[rec] = new int[number_categories_recording_belongs_to];
			int hits_so_far = 0;
			for (int cat = 0; cat < model_results[rec].length; cat++)
				if (model_results[rec][cat])
			{
				int_model_results[rec][hits_so_far] = cat;
				hits_so_far++;
			}
		}

		// Return the model results in int form
		return int_model_results;
	}


	// Causes the given breeder to keep evolving chromosomes until either
	// the given conditions (seec below) are met or the given maximum number of epochs
	// is exceeded. Also sets the contents of the best_DNA parameter (whose initial
	// contents are irrelevant) to the bit string of the best performing chromosome
	// of the last epoch.
	//
	// The returned array of doubles gives the value of (1 - fitness) for the fitness of
	// the best performing chromosome for each epoch. A value of 0 in this array is
	// indicative of perfect performance of the chromosome and an error of 1 indicates
	// complete failure.
	// 
	// The bioknn_iterations field controls the number of training iterations performed.
	// If a negative value is present here, then the number of iterations to perform
	// is calculated automatically based on the bioknn_acceptable_threshold field, which 
	// specifies the absolute rate of change of the fitness of the best performing
	// chromosome below which training will stop, and the bioknn_consecutive_iterations
	// field, which specifies the number of consecutive epochs for which the rate
	// of change must be below this threshold in order for training to stop. The number
	// of epochs that go by will never exceed the absolute value of the bioknn_iterations
	// field, irregardless of the other parameters.
	//
	// Throws an exception if there is a problem during evolution.
	//
	// The training_monitor parameter is used to monitor training progress. It may
	// be null if it is not used.
	private double[] evolve(Breeder breeder, short[] best_DNA, ProgressBarTaskTrainMonitor training_monitor)
		throws Exception
	{
		// The error for each epoch that went by during evolution.
		// Calculated by subtracting the fitness of the best chromosome from 1
		double[] reported_errors = new double[(int) (((double) Math.abs(bioknn_iterations)) * classifier_vs_feature_multiplier)];
		for (int i = 0; i < reported_errors.length; i++)
			reported_errors[i] = -1;

		// Perform the evolution and end evolution when appropriate
		int maximum_epochs = (int) (((double) Math.abs(bioknn_iterations)) * classifier_vs_feature_multiplier);
		int below_threshold_so_far = 0; // number of consecutive iterations with error below bioknn_acceptable_threshold
		for (int epoch = 0; epoch < maximum_epochs; epoch++)
		{
			// Update the training_monitor
			if (training_monitor != null)
			{
				training_monitor.setBottomProgressValue(epoch);
				if (training_monitor.isCancelled())
					throw new Exception("Training cancelled.");
			}

			// Evolve over one epoch
			Object[] evolution_report = breeder.evolve();

			// Store the error of the best performing chromosome for each epoch
			reported_errors[epoch] = 1 - (((double[]) evolution_report[2])[0]);

			// Store the DNA (bit string) of the current best performer
			short[] temp_DNA = (short[]) evolution_report[3];
			for (int bit = 0; bit < best_DNA.length; bit++)
				best_DNA[bit] = temp_DNA[bit];

			// Check if should end evolution
			if (bioknn_iterations < 0)
			{
				// Check for convergence
				if (epoch > 0)
				{
					if ( Math.abs(reported_errors[epoch] - reported_errors[epoch - 1]) < bioknn_acceptable_threshold)
					{
						below_threshold_so_far++;
						if (below_threshold_so_far == bioknn_consecutive_iterations)
						{
							double[] temp = new double[epoch + 1];
							for (int i = 0; i < temp.length; i++)
								temp[i] = reported_errors[i];
							reported_errors = temp;
							epoch = maximum_epochs + 1; // exit the loop
						}
					}
					else
						below_threshold_so_far = 0;
				}
			}
		}

		// Return the error measurements
		return reported_errors;
	}
}