/*
 * @(#)SupervisedClassifier.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers;
import bodhidharma.ProgressBarTaskTrainMonitor;

import java.io.*;

/**
 * An abstract class for designing a classifier that used supervised learning
 * in order to classify arbitrary feature sets. This implementation
 * makes it possible to assign more than one label to a single feature set.
 *
 * <p>Feature sets are fed into the classifier as arrays of doubles. Categories
 * are specified as arrays of Strings.
 *
 * <p>All classes implementing this interface should include a constructor that
 * allows the parameters of particular classifiers to be set. This constructor
 * should also set the <i>categories</i>, <i>feature_names</i> and <i>identifier</i>
 * fields.
 *
 * <p>All classes implementing this interface should also include a constructor
 * that only takes in a String holding the file path of an XML file holding the
 * information necessary to reconstruct a trained classifier.
 *
 * <p>Use the <code>train</code> method to train the classifier (the
 * <code>getFeatureNames</code> method is useful for deteriming the features that
 * can be fed to the classifer).
 *
 * <p>Use the <code>classify</code> method to classify feature sets once training
 * has been completed (the <code>getCategories</code> method is useful for
 * determining what categories feature sets can be classified into and for
 * determining the order of the categories when parsing classification results).
 *
 * <p>Use the <code>save</code> method to save the classifier and its current
 * state to disk.
 * 
 * <p>Use the <code>getClassifierName</code> and <code>getClassifierParameters</code>
 * methods to obtain information about the classifier.
 *
 * <p>Use the <code>getClassifierIdentifier</code> method to get a name or code that
 * was given to an instantiation of the classifier when it was constructed. This
 * identifier can be used by external classes to identify the instantiation.
 *
 * @author Cory McKay
 */
public abstract class SupervisedClassifier
{
	/* FIELDS ******************************************************************/

	
	/**
	 * The possible categories into which feature sets can be classified.
	 */
	protected	String[]	categories;


	/**
	 * The names of the different features which are used to perform classifications.
	 */
	protected	String[]	feature_names;


	/**
	 * An identifier that can be associated with the classifier so that outside classes
	 * can identify it.
	 */
	protected	String		identifier;


	/**
	 * Used to monitor training progress. It may be null if it is not used.
	 */
	protected	ProgressBarTaskTrainMonitor	training_monitor;


	/* PUBLIC METHODS **********************************************************/

	
	/**
	 * Returns the contents of the <i>categories</i> field.
	 */
	public	String[]	getCategories()
	{
		return categories;
	}
	

	/**
	 * Returns the contents of the <i>feature_names</i> field.
	 */
	public	String[]	getFeatureNames()
	{
		return feature_names;
	}
	

	/**
	 * Returns a name or code that was given to this instantiation of the classifier when
	 * it was constructed. This identifier can be used by external classes to identify
	 * this instantiation.
	 */
	public String getClassifierIdentifier()
	{
		return identifier;
	}


	/**
	 * Returns the name of the type of classifier. This information can be used
	 * by external classes to identify the type of classifier in situations such
	 * as when the configuration of a trained classifier is being read from a file.
	 */
	public	abstract	String	getClassifierName();


	/**
	 * Returns a String describing the parameters of the classifier. This does not
	 * include training values, however.
	 */
	public	abstract	String	getClassifierParameters();


	/**
	 * Sets the training_monitor parameter to the given value.
	 */
	public void setTrainingMonitor(ProgressBarTaskTrainMonitor monitor)
	{
		training_monitor = monitor;
	}


	/**
	 * Trains the SupervisedClassifier using the given feature sets. The first
	 * indice of the <i>feature_sets</i> parameter corresponds to different feature
	 * sets. The second indice corresponds to different features in the given
	 * featue set. It should be noted that all feature sets must use the same features
	 * in the same order as given in the </i>feature_labels</i> parameter.
	 *
	 * <p>The <i>feature_labels</i> parameter specifies the names of each of the features
	 * in the <i>feature_sets</i> parameter. The features in the <i>feature_sets</i> parameter
	 * will automatically matched to the features in the <i>feature_names</i> field based on
	 * the content of the <i>feature_labels</i> parameter unless a value of <i>null</i> is passed
	 * to the <i>feature_labels</i>. In this case, the feature values in the <i>feature_sets</i>
	 * parameter will simply be fed into the classifier in the order that they occur.
	 *
	 * <p>The <i>model_categories</i> parameter gives the categorie(s) of each of the given
	 * feature sets. The first indice corresponds to the feature set. The second indice
	 * corresponds to different possible categories for the given feature set. Only categories
	 * to which the feature set belongs should be included.
	 *
	 * <p>The <i>iterations</i> parameter specifies the number of training iterations
	 * performed.  If a negative value is passed here, then the number of iterations to perform
	 * is calculated automatically based on the <i>acceptable_threshold</i> parameter, which 
	 * specifies the absolute rate of change of the classification error below which training will
	 * stop, and the <i>consecutive_iterations</i> parameter, which specifies the number of consecutive
	 * iterations for which the rate of change must be below this threshold in order for
	 * training to stop. The number of iterations that go by will never exceed the absolute
	 * valud of the <i>iterations</i> value, irregardless of the other parameters.
	 *
	 * <p>For example, if a value of 1000 is given for <i>iterations</i>, then 1000 iterations
	 * will be performed irregardless of the other parameters. If a value of -1000 is given,
	 * then training will automatically stop if the absolute value of the rate of change
	 * of the classification error from one sample to the next falls below the
	 * <i>acceptable_threshold</i> parameter for <i>consecutive_iterations</i> iterations, but
	 * no more than 1000 iterations will be performed in any case.
	 *
	 * <p><b>NOTE:</b> the last three parameters are ignored if a type of classifier is used
	 * that does not use more than one iteration.
	 *
	 * <p><b>NOTE:</b> the way in which the classification error is calculated varies from implementation
	 * to implementation, but 0 is always perfect performance, and rising levels indicate poorer
	 * performance.
	 *
	 * <p>The returned double is a set of classification error after training iterations. The indice of
	 * the returned array corresponds to the iteration of training that the error is associated
	 * with. A value of null is returned if the classifier does not provide this information. 
	 *
	 * <p>An exception if thrown if the <i>feature_labels</i> do not contain the same names as
	 * <i>feature_names</i> (although a different ordering is permitted) or if any of the
	 * feature sets in <i>feature_sets</i> have a different number of features than
	 * <i>feature_names</i>. An exception is also thrown if <i>feature_sets</i> and
	 * <i>model_categories</i> have different sizes in regard to their first parameters.
	 * An exception is also thrown if the <i>given_results</i> parameter contains a name not
	 * present in the <i>categories</i> field or if it contains the same category more than once.
	 */
	public	abstract	double[]	train(double[][] feature_sets, String[] feature_labels,
	                                      String[][] model_categories, int iterations,
	                                      double acceptable_threshold, int consecutive_iterations)
		throws Exception;


	/**
	 * Returns the relative scores of each of the possible categories when the given sets
	 * of features are classified. There is one entry in the returned array for each of the
	 * entries in the <i>categories</i> field, and they appear in the same order. Higher scores
	 * in the returned array correspond to a greater certainty that the feature set should
	 * have the corresponding label. Scores should fall in the range between 0.0 and 1.0.
	 * The first indice of the returned array corresponds to the feature set and the second
	 * corresponds to the category. The order of the categories is the same as in the categories
	 * field and the order of the feature sets is the same as the order in which they were passed
	 * to the <i>feature_sets</i> parameter.
	 *
	 * <p>The <i>feature_sets</i> parameter specifies the feature sets to be classified. The
	 * first indice corresponds to different feature sets. The second indice corresponds to
	 * different features in the given featue set. It should be noted that all feature sets
	 * must use the same features in the same order as given in the </i>feature_labels</i>
	 * parameter.
	 *
	 * <p>The <i>feature_labels</i> parameter specifies the names of each of the features
	 * in the <i>feature_sets</i> parameter. The features in the <i>feature_sets</i> parameter
	 * will automatically matched to the features in the <i>feature_names</i> field based on
	 * the content of the <i>feature_labels</i> parameter unless a value of <i>null</i> is passed
	 * to the <i>feature_labels</i>. In this case, the feature values in the <i>feature_sets</i>
	 * parameter will simply be fed into the classifier in the order that they occur.
	 *
	 * <p>An exception if thrown if the <i>feature_labels</i> do not contain the same names as
	 * <i>feature_names</i> (although a different ordering is permitted) or if any of the
	 * feature sets in <i>feature_sets</i> have a different number of features than
	 * <i>feature_names</i>.
	 */
	public	abstract	double[][]	classify(double[][] feature_sets, String[] feature_labels)
		throws Exception;


	/**
	 * Writes the <code>SupervisedClassifier</code> and its current state to the given save_file.
	 * If the particular <code>SupervisedClassifier</code> needs to write more than one file, then
	 * a directory should be passed to this method's parameter, where the appropriate files will
	 * be written. Throws an exception if a problem occurs during saving.
	 */
	public	abstract	void	save(File place_to_save)
		throws Exception;


	/* PROTECTED METHODS *******************************************************/


	/**
	 * Returns a 2-D array of doubles that consists of the contents of the feature_sets
	 * parameter after having been reordered so that the order of the features (as specified
	 * in the feature_labels parameter) are the same as in the feature_names field. In
	 * other words, this method reorders the the contents referred to by the second
	 * indice of feature_sets in a copy of feature_sets. The order of the feature sets
	 * themselves (first parameter) is not changed.
	 *
	 * <p>An exception is thrown if the feature_labels parameter contains any labels 
	 * not found in the feature_names field or if there are any labels found in the
	 * feature_names field that are not in the feature_labels parameter. The thrown
	 * exception contains a description of the problem encountered. An exception is
	 * also thrown if any of the feature sets contain a number of features different
	 * from the number of feature labels.
	 *
	 * <p>This method is intended for use by the <code>train</code> and <code>classify</code>
	 * methods.
	 *
	 * <p>The arrays passed as parameters are not altered.
	 */
	protected double[][] getOrderedFeatureSets(double[][] feature_sets, String[] feature_labels)
		throws Exception
	{
		// Copy feature_sets into feature_sets_copy
		double[][] feature_sets_copy = new double[feature_sets.length][];
		for (int i = 0; i < feature_sets_copy.length; i++)
		{
			feature_sets_copy[i] = new double[feature_sets[i].length];
			for (int j = 0; j < feature_sets_copy[i].length; j++)
				feature_sets_copy[i][j] = feature_sets[i][j];
		}

		// Find the correct order that features should appear in feature_labels to match the
		// order in feature_names. Throw an exception if feature_names and feature_labels do
		// not contain the same strings.
		if (feature_labels.length != feature_names.length)
			throw new Exception("There are " + feature_labels.length + " features in the " +
		                        "given feature set, but " + feature_names.length + "features " +
			                    "in the classifier. These two numbers must match.");
		int[] proper_indices = new int[feature_names.length]; // will give the indice to which features should be moved
		for (int i = 0; i < proper_indices.length; i++)
			proper_indices[i] = -1;
		for (int i = 0; i < feature_labels.length; i++)
		{
			boolean found = false; // used to check for repeated features
			for (int j = 0; j < feature_names.length; j++)
			{
				if (feature_labels[i].equals(feature_names[j]))
				{
					if (found)
						throw new Exception("The feature " + feature_labels[i] + " occurs more " +
						                    "than once in the given feature set. Each feature " +
						                    "should only occur once.");
					proper_indices[i] = j;
					found = true;
				}
			}

			// Throw exception if a feature not found
			if (proper_indices[i] == -1)
				throw new Exception("The given features contain a feature named " + feature_labels[i] +
				                    "that is not accounted for in the classifier.");
		}

		// Reorder feature_sets_copy so that the order of the features matches the order in
		// feature_names
		for (int set = 0; set < feature_sets.length; set++)
		{
			// Throws an exception if any of the feature sets contain a number of features
			// different from the number of feature labels
			if (feature_sets[set].length != feature_names.length)
				throw new Exception("One of the given feature sets does not contain the appropriate " + 
				                    "number of features. It contains " + feature_sets[set].length +
				                    " features and should contain " + feature_names.length + ".");

			// Perform the reordering
			for (int feat = 0; feat < feature_sets[set].length; feat++)
			{
				feature_sets_copy[set][ proper_indices[feat] ] = feature_sets[set][feat];
			}
		}

		// Return the feature sets with the features properly ordered
		return feature_sets_copy;
	}


	/**
	 * Returns a 2-D array whose first indice corresponds to the feature sets specified by
	 * the first indice of the <i>given_results</i> parameter and whose second indice corresponds
	 * to each of the categories in the <i>categories</i> field. Entries are set to true if
	 * the given feature set belongs to the given category and to false if they do not.
	 *
	 * <p>This method is intended for use by the <code>train</method> method to process its
	 * <i>model_categories</i> parameter, which is passed to the <i>given_results</i> parameter
	 * of this method. The <i>given_results</i> parameter consists of feature sets (first
	 * indice) and the category names (second parameter) that each feature set belongs to.
	 *
	 * <p>An exception is thrown if the <i>given_results</i> parameter contains a name not
	 * present in the <i>categories</i> field or if it contains the same category more than once.
	 */
	protected boolean[][] getModelResults(String[][] given_results)
		throws Exception
	{
		// Instantiate the model results
		boolean[][] model_results = new boolean[given_results.length][categories.length];
		for (int i = 0; i < model_results.length; i++)
			for (int j = 0; j < model_results[i].length; j++)
				model_results[i][j] = false;

		// Match feature set categories with classifier categories
		for (int set = 0; set < given_results.length; set++)
		{
			for (int gcat = 0; gcat < given_results[set].length; gcat++)
			{
				boolean found = false;
				for (int mcat = 0; mcat < categories.length; mcat++)
				{
					if ( given_results[set][gcat].equals( categories[mcat] ) )
					{
						if (found)
							throw new Exception("The given model results contain the same " +
							                    "model category more than once for the same " +
							                    "feature set.");
						model_results[set][mcat] = true;
						found = true;
					}
				}
				if (!found)
					throw new Exception("The given model results contain a category called " +
					                    given_results[set][gcat] + " that is not accounted " +
					                    "for in the classifier.");
			}
		}

		// Return the model training results
		return model_results;
	}
}