/*
 * @(#)KNN.java	1.1.2 09/14/04
 *
 * McGill University
 */

package bodhidharma.classifiers.bio_k_nearest_neighbour;

import bodhidharma.xml_parsing.*;
import java.io.*;

/**
 * Uses the k-nearest neighbour algorithm to classify feature sets. k is automatically set as the
 * square root of the number of training samples.
 *
 * <p>The <i>features_to_use</i> and <i>feature_weights</i> fields can be set by external functions
 * in order to set feature selection and feature weighting settings respectively.
 *
 * <p>The <i>addTrainingSamples</i> method is used to train the classifier and the <i>classifyTestSample</i>
 * method is used to classify test points. The <i>getK</i> method reveals the k used.
 *
 * <p>The <i>save</i> method saves the current state of a KNN object to disk.
 *
 * <p>Training examples can be used that belong to more than one category.
 *
 * @author Cory McKay
 */
public class KNN
{
	/* FIELDS ******************************************************************/

	
	/**
	 * Whether or not a given feature is to be used when calculating the distance metric.
	 * Indices correspond to the second indices of the training samples field. A value of
	 * null indicates that all features should be used. Used for feature selection.
	 */
	public	boolean[]	features_to_use;


	/**
	 * The weights used for calculating the distance for each feature. Between 0 and 1.
	 * Indices correspond to the second indices of the training samples field. A value of
	 * null means that all features have a weighting of 1.
	 */
	public	double[]	feature_weights;


	// The data points that the classifier has been trained with. The first indice identifies
	// the sample and the second indice identifies the feature. A value of null means that
	// no training samples have been provided.
	private	double[][]	training_samples;


	// The number of categories into which test sample can be classified.
	private int			number_of_categories;


	// The categories to which each of the training_samples belong. The first indice identifies
	// the sample, and the second indice identifies the number of the category. The entry identifies
	// the number assigned to each category. The values of this field are set in the addTrainingSamples
	// method and used by the classifyTestSets method. The objects that call these methods
	// must know the meaning of these codes.
	private	int[][]		training_sample_categories;


	// The number of training samples to find around each test point.
	private	int			k;


	// The cumulative number of training samples that have been used to train the classifier
	// so far
	private int			n;



	/* CONSTRUCTOR *************************************************************/

	
	/**
	 * Basic constructor. Initializes system so that all features will be used in
	 * classification with an equal weighting of 1.
	 *
	 * <p>The parameter identifies how many categories the system will be categorizing test
	 * points into.
	 */
	public KNN(int how_many_categories)
	{
		// Set superclass' fields
		training_samples = null;
		number_of_categories = how_many_categories;
		feature_weights = null;
		features_to_use = null;
		n = 0;
	}


	/**
	 * Parse the file specified by the given file path to recreate the specificed trained
	 * classifier. Throws an exception if problems occur during parsing.
	 *
	 * @param	file_path			The file path of a KNN_file that this instantiation is to be based on.
	 */
	public KNN(String file_path)
		throws Exception
	{
		// Parse file
		Object[] results = (Object[]) XMLDocumentParser.parseXMLDocument(file_path, "KNN_file");
		String[] fixed_fields = (String[]) results[0];
		
		// Set fields
		features_to_use = (boolean[]) results[1];
		feature_weights = (double[]) results[2];
		training_samples = (double[][]) results[3];
		number_of_categories = (Integer.valueOf(fixed_fields[0])).intValue();
		training_sample_categories = (int[][]) results[4];
		k = (Integer.valueOf(fixed_fields[1])).intValue();
		n = (Integer.valueOf(fixed_fields[2])).intValue();
	}


	/* PUBLIC METHODS **********************************************************/

	
	/**
	 * Stores the provided feature sets as training data. The first indice of the training_data
	 * parameter identifies the sample and the second indice identifies the feature. The first
	 * indice of the model_categories parameter identifies the training sample and the second
	 * indice identifies the number of the category. The entries of model_categories corresponds
	 * to the code assigned to the category(ies) that each sample belongs to. This code must
	 * be 0 or above and less than the how_many_categories number passed to the constructor. An
	 * exception is thrown if these conditions are not fulfilled. If this object has previously
	 * been trained, it is the calling object's responsibility that the categories used in the
	 * past and current trainings are the same.
	 *
	 * <p>Training examples can be used that belong to more than one category.
	 *
	 * <p>It is assumed that all of the samples in the training_data parameter have the same
	 * features in the same order. The calling object must ensure that this is true. An exception
	 * is thrown if any of the features in training_data have a different number of features,
	 * either with each other or with any existing training samples.
	 *
	 * <p>If this object has previously been trained with feature sets, these new feature sets
	 * will be added to the existing ones.
	 *
	 * <p>k is automatically set to the square root of the cumulative number of training samples
	 * that have been used to train the classifier so far.
	 *
	 * <p><b>WARNING:</b> Care should be taken not to include samples that have previously been
	 * provided to the classifier, as this will cause the sample to be double counted if it falls
	 * within k.
	 */
	public void addTrainingSamples(double[][] training_data, int[][] model_categories)
		throws Exception
	{	
		// Throw an exception if any of the features in training_data have a different number
		// of features, either with each other or with any existing training samples
		if (training_samples != null)
		{
			if (training_samples[0].length != training_data[0].length)
				throw new Exception("Training samples added to existing samples with a different " +
				                    "number of features.");
		}
		int correct_number_features = training_data[0].length;
		for (int i = 0; i < training_data.length; i++)
		{
			if (training_data[i].length != correct_number_features)
				throw new Exception("Training samples added that have an inconsistent number of " +
				                    "features.");
		}

		// Throw an exception if an improper model_categories code is used.
		for (int i = 0; i < model_categories.length; i++)
			for (int j = 0; j < model_categories[i].length; j++)
				if (model_categories[i][j] < 0 || model_categories[i][j] >= number_of_categories)
					throw new Exception("Improper model classification code specified.");

		// Store model_categories in training_sample_categories
		training_sample_categories = model_categories;

		// Store training_data if no training_data has been previously provided
		if (training_samples == null)
		{
			training_samples = new double[training_data.length][];
			for (int rec = 0; rec < training_samples.length; rec++)
			{
				training_samples[rec] = new double[training_data[rec].length];
				for (int feat = 0; feat < training_samples[rec].length; feat++)
					training_samples[rec][feat] = training_data[rec][feat];				
			}
		}

		// Add new training samples to existing training samples if training samples
		// have been previously provided
		else
		{
			double[][] combined_samples = new double[ (training_data.length + training_samples.length) ][];
			for (int i = 0; i < training_samples.length; i++)
				combined_samples[i] = training_samples[i];
			for (int i = training_samples.length; i < training_samples.length + training_data.length; i++)
			{
				combined_samples[i] = new double[training_data[i - training_samples.length].length];
				for (int j = 0; j < combined_samples[i].length; j++)
					combined_samples[i][j] = training_data[i - training_samples.length][j];
			}
			training_samples = combined_samples;
		}

		// Update n and calculate k
		n += training_data.length;
		k = (int) Math.sqrt(n);
	}


	/**
	 * Returns an array of integers indicating the number of training samples that are included
	 * in the k points closest to the test_point for each category that the system has been
	 * trained with. The indice of the returned array identifies the category and the corresponding
	 * entry gives the number of points included in k for that category.  Note that dividing
	 * by k will not necessarily result in normalized results, since training categories can belong
	 * to more than one category.
	 *
	 * <p>The test_point parameter specifies the features of a sample that is to be classified.
	 * The features stored in the test_point and their order must correspond to that used during
	 * training with the addTrainingSamples method.
	 *
	 * <p>Uses the last features selection and/or feature weightings that have been specified (the
	 * default is to use all features with equal weightings of 1.
	 *
	 * <p>An exception is thrown if the test_point has a different number of features than the
	 * data that was used to train the classifier.
	 */
	public double[] classifyTestSample(double[] test_point)
		throws Exception
	{
		// Throw an exception if the test point has a differnt number of features than the first
		// training point.
		if (test_point.length != training_samples[0].length)
			throw new Exception("Test point has a different number of features than data that was " +
			                    "used to train the classifier");

		// Find the distances of the test point from each training point
		double[] training_sample_feature_distances = new double[training_samples.length];
		for (int rec = 0; rec < training_samples.length; rec++)
			training_sample_feature_distances[rec] =
				calculateEuclideanDistanceConsideringFeatureSelectionsAndWeightings(test_point, training_samples[rec]);

		// Find the points to include in k
		int[] ks = new int[k];
		double greatest_distance = training_sample_feature_distances[ bodhidharma.utilities.Statistics.getIndexOfLargest(training_sample_feature_distances) ];
		for (int i = 0; i < ks.length; i++)
		{
			ks[i] = bodhidharma.utilities.Statistics.getIndexOfSmallest(training_sample_feature_distances);
			training_sample_feature_distances[ ks[i] ] = greatest_distance; // so won't be double counted
		}

		// Initialize the number of memberships in k of each category to 0
		double[] categorization_results = new double[number_of_categories];
		for (int cat = 0; cat < categorization_results.length; cat++)
			categorization_results[cat] = 0.0;

		// Store the number of hits for each category in categorization_results
		for (int i = 0; i < ks.length; i++)
		{
			int[] categories_of_this_k = training_sample_categories[ ks[i] ];
			for (int j = 0; j < categories_of_this_k.length; j++)
				categorization_results[ categories_of_this_k[j] ]++;
		}

		// Return the classified results
		return categorization_results;	
	}


	/**
	 * Get the value of k used by the KNN classifier.
	 */
	public int getK()
	{
		return k;
	}


	/**
	 * Get the number of categories into which test sample can be classified.
	 */
	public int getNumberCategories()
	{
		return number_of_categories;
	}


	/**
	 * Saves all of the fields to the given file. Throws an exception if a problem occurs.
	 */
	public void	save(File place_to_save)
		throws Exception
	{
		FileOutputStream to = new FileOutputStream(place_to_save);
		DataOutputStream writer = new DataOutputStream(to);
		String header = new String
		(
			"<?xml version=\"1.0\"?>\n" +
			"<!DOCTYPE KNN_file [\n" +
			"   <!ELEMENT KNN_file (comments, features_to_use, feature_weights, training_samples, number_of_categories, training_sample_categories, k, n)>\n" +
			"   <!ELEMENT comments (#PCDATA)>\n" +
			"   <!ELEMENT features_to_use (use*)>\n" +
			"   <!ELEMENT use (#PCDATA)>\n" +
			"   <!ELEMENT feature_weights (weight*)>\n" +
			"   <!ELEMENT weight (#PCDATA)>\n" +
			"   <!ELEMENT training_samples (s*)>\n" +
			"   <!ELEMENT s (f* | c*)>\n" +
			"   <!ELEMENT f (#PCDATA)>\n" +
			"   <!ELEMENT number_of_categories (#PCDATA)>\n" +
			"   <!ELEMENT training_sample_categories (s*)>\n" +
			"   <!ELEMENT c (#PCDATA)>\n" +
			"   <!ELEMENT k (#PCDATA)>\n" +
			"   <!ELEMENT n (#PCDATA)>\n" +
			"]>\n\n" +
			"<KNN_file>\n\n" +
			"   <comments></comments>\n\n"
		);
		writer.writeBytes(header);
		writer.writeBytes("   <features_to_use>\n");
		if (features_to_use != null)
			for (int i = 0; i < features_to_use.length; i++)
				writer.writeBytes("      <use>" + features_to_use[i] + "</use>\n");
		else
			writer.writeBytes("      <use>null</use>\n");
		writer.writeBytes("   </features_to_use>\n");
		writer.writeBytes("   <feature_weights>\n");
		if (feature_weights != null)
			for (int i = 0; i < feature_weights.length; i++)
				writer.writeBytes("      <weight>" + feature_weights[i] + "</weight>\n");
		else
			writer.writeBytes("      <weight>null</weight>\n");
		writer.writeBytes("   </feature_weights>\n");
		writer.writeBytes("   <training_samples>\n");
		if (training_samples != null)
			for (int i = 0; i < training_samples.length; i++)
			{
				writer.writeBytes("      <s>\n");
				for (int j = 0; j < training_samples[i].length; j++)
					writer.writeBytes("         <f>" + training_samples[i][j] + "</f>\n");
				writer.writeBytes("      </s>\n");
			}
		writer.writeBytes("   </training_samples>\n");
		writer.writeBytes("   <number_of_categories>" + number_of_categories + "</number_of_categories>\n");
		writer.writeBytes("   <training_sample_categories>\n");
		if (training_sample_categories != null)
			for (int i = 0; i < training_sample_categories.length; i++)
			{
				writer.writeBytes("      <s>\n");
				for (int j = 0; j < training_sample_categories[i].length; j++)
					writer.writeBytes("         <c>" + training_sample_categories[i][j] + "</c>\n");
				writer.writeBytes("      </s>\n");
			}
		writer.writeBytes("   </training_sample_categories>\n");
		writer.writeBytes("   <k>" + k + "</k>\n");
		writer.writeBytes("   <n>" + n + "</n>\n");
		writer.writeBytes("</KNN_file>");
		writer.close();
	}


	/* PRIVATE METHODS *********************************************************/


	/**
	 * Returns the Euclidian distance between the feature points x and y.
	 * Checks to see if features have been selected for use in features_to_use
	 * and uses feature weights in feature_weights. Throws an exception if x
	 * and y have different sizes.
	 */
	private double calculateEuclideanDistanceConsideringFeatureSelectionsAndWeightings(double[] x, double[] y)
		throws Exception
	{
		if (x.length != y.length)
			throw new Exception("The two given arrays have different sizes.");

		// The total unrooted distance between the feature points
		double total = 0.0;

		if (features_to_use != null)
		{
			// Case where feature selection has been performed
			// and feature weighting has also been performed
			if (feature_weights != null)
			{
				for (int feat = 0; feat < x.length; feat++)
					if (features_to_use[feat])
						total += (feature_weights[feat] * (Math.pow( (x[feat] - y[feat]), 2 )));
			}
			
			// Case where feature selection has been performed
			// but feature weighting has not been performed
			else
			{
				for (int feat = 0; feat < x.length; feat++)
					if (features_to_use[feat])
						total += Math.pow( (x[feat] - y[feat]), 2 );
			}
		}
		else
		{
			// Case where feature selection has not been performed
			// but feature weighting has been performed
			if (feature_weights != null)
			{
				for (int feat = 0; feat < x.length; feat++)
					total += feature_weights[feat] * (Math.pow( (x[feat] - y[feat]), 2 ));
			}

			// Case where feature selection has not been performed
			// and feature weighting has not been performed
			else
			{
				for (int feat = 0; feat < x.length; feat++)
					total += Math.pow( (x[feat] - y[feat]), 2 );
			}
		}

		return Math.sqrt(total);
	}
}