/*
 * @(#)LinkedNodes.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers.feedforward_neural_networks;

import java.util.LinkedList;

/**
 * Objects of this class represent complete backprop neural networks. LinkedNodes objects
 * link the node objects together and allow them to interact with one another. LinkedNodes
 * objects are controlled by FeedForwardNeuralNetwork objects. Adjustable learning rates and 
 * momentums are permitted. An arbitrary number of input nodes, hidden nodes and output nodes
 * are permitted.
 *
 * <p>Only allows perceptrons or neural networks with one layer of hidden nodes. Only limited
 * error checking is implemented. It is assumed that correct input is passed to the methods.
 * 
 * @author Cory McKay
 */
public class LinkedNodes
{
	/* FIELDS ******************************************************************/


	private int num_input_nodes;		// the number of input nodes
	private int num_hidden_nodes;		// the number of hidden nodes
	private int num_output_nodes;		// the number of output nodes
	private InputNode[] input_nodes;	// the input nodes
	private HiddenNode[] hidden_nodes;	// the hidden nodes
	private OutputNode[] output_nodes;	// the output nodes


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Use this constructor when generating a network from scratch.
	 *
	 * @param	learning_rate	The learning rate neural network parameter used during training.
	 * @param	momentum		The momentum neural network parameter used during training.
	 * @param	inputs			The number of input nodes.
	 * @param	hidden			The number of hidden nodes.
	 * @param	outputs			The number of output nodes.
	 * @param	min_w			The minimum possible value for the randomly generated initial weights.
	 * @param	max_w			The maximum possible value for the randomly generated initial weights.
	 */
	public LinkedNodes(double learning_rate, double momentum, int inputs, int hidden,
	                     int outputs, double min_w, double max_w)
	{
		// Initialize fields for number of nodes
		num_input_nodes = inputs;
		num_hidden_nodes = hidden;
		num_output_nodes = outputs;

		// Initialize arrays of nodes
		input_nodes = new InputNode[num_input_nodes];
		output_nodes = new OutputNode[num_output_nodes];
		if (num_hidden_nodes > 0)
			hidden_nodes = new HiddenNode[num_hidden_nodes];
		else
			hidden_nodes = null;

		// Initialize input nodes
		for (int i = 0; i < num_input_nodes; i++)
			input_nodes[i] = new InputNode(0);

		// Initialize hidden and output nodes with randomly generated weights
		if (num_hidden_nodes > 0) // case where there is a hidden layer
		{
			for (int j = 0; j < num_hidden_nodes; j++)
			{
				hidden_nodes[j] = new HiddenNode(learning_rate, momentum, num_input_nodes + 1); // + 1 is because of bias
				for (int i = 0; i <= num_input_nodes; i++) // <= is because of bias
					hidden_nodes[j].set_weight(i, get_initial_weight(min_w, max_w));
			}
			for (int k = 0; k < num_output_nodes; k++)
			{
				output_nodes[k] = new OutputNode(learning_rate, momentum, num_hidden_nodes + 1); // + 1 is because of bias
				for (int j = 0; j <= num_hidden_nodes; j++) // <= is because of bias
					output_nodes[k].set_weight(j, get_initial_weight(min_w, max_w));
			}
		}
		else // case where there is no hidden layer
		{
			for (int k = 0; k < num_output_nodes; k++)
			{
				output_nodes[k] = new OutputNode(learning_rate, momentum, num_input_nodes + 1); // + 1 is because of bias
				for (int i = 0; i <= num_input_nodes; i++) // <= is because of bias
					output_nodes[k].set_weight(i, get_initial_weight(min_w, max_w));
			}
		}
	}


	/**
	 * Use this constructor when loading a pre-existing network.
	 *
	 * @param	learning_rate	The learning rate neural network parameter used during training.
	 * @param	momentum		The momentum neural network parameter used during training.
	 * @param	inputs			The number of input nodes.
	 * @param	hidden			The number of hidden nodes.
	 * @param	outputs			The number of output nodes.
	 * @param	weights			Contains the weights of the hidden nodes (if any) followed by those of the output nodes.
	 */
	public LinkedNodes(double learning_rate, double momentum, int inputs, int hidden,
	                     int outputs, double[] weights)
	{
		// Initialize fields for number of nodes
		num_input_nodes = inputs;
		num_hidden_nodes = hidden;
		num_output_nodes = outputs;

		// Initialize arrays of nodes
		input_nodes = new InputNode[num_input_nodes];
		output_nodes = new OutputNode[num_output_nodes];
		if (num_hidden_nodes > 0)
			hidden_nodes = new HiddenNode[num_hidden_nodes];
		else
			hidden_nodes = null;

		// Initialize input nodes
		for (int i = 0; i < num_input_nodes; i++)
			input_nodes[i] = new InputNode(0);

		// Initialize hidden and output nodes with given weights
		int weight_index = 0; // index of weight that has been addded so far
		if (num_hidden_nodes > 0) // case where there is a hidden layer
		{
			for (int j = 0; j < num_hidden_nodes; j++)
			{
				hidden_nodes[j] = new HiddenNode(learning_rate, momentum, num_input_nodes + 1); // + 1 is because of bias
				for (int i = 0; i <= num_input_nodes; i++) // <= is because of bias
				{
					hidden_nodes[j].set_weight(i, weights[weight_index]);
					weight_index++;
				}
			}
			for (int k = 0; k < num_output_nodes; k++)
			{
				output_nodes[k] = new OutputNode(learning_rate, momentum, num_hidden_nodes + 1); // + 1 is because of bias
				for (int j = 0; j <= num_hidden_nodes; j++) // <= is because of bias
				{
					output_nodes[k].set_weight(j, weights[weight_index]);
					weight_index++;
				}
			}
		}
		else // case where there is no hidden layer
		{
			for (int k = 0; k < num_output_nodes; k++)
			{
				output_nodes[k] = new OutputNode(learning_rate, momentum, num_input_nodes + 1); // + 1 is because of bias
				for (int i = 0; i <= num_input_nodes; i++) // <= is because of bias
				{
					output_nodes[k].set_weight(i, weights[weight_index]);
					weight_index++;
				}
			}
		}
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the array of values output at the output nodes corresponding to input_values
	 */
	public double[] get_output(double[] input_values)
	{
		double[] output_values = new double[num_output_nodes]; // the output of the network corresponding to
		                                                       // the given input

		for (int i = 0; i < num_input_nodes; i++)
			input_nodes[i].set_input(input_values[i]);
		propagate();
		for (int i = 0; i < num_output_nodes; i++)
			output_values[i] = output_nodes[i].get_current_value();

		return output_values;
	}


	/**
	 * Goes through one iteration of training to update all weights once, given the desired
	 * output values for the given input_values. Returns the sum-of-squared-error between the
	 * given desired output values and the results of the network AFTER the training iteration
	 * has been completed.
	 */
	public double training_iteration(double[] input_values, double[] desired_values)
	{

		// Send input values through network
		for (int i = 0; i < num_input_nodes; i++)
			input_nodes[i].set_input(input_values[i]);
		propagate();

		// Update weights leading into output layer
		for (int k = 0; k < num_output_nodes; k++)
			output_nodes[k].update_weights(desired_values[k]);

		// Update weights leading into hidden layer
		for (int j = 0; j < num_hidden_nodes; j++)
		{
			double error_sum = 0;
			for (int k = 0; k < num_output_nodes; k++)
				error_sum = error_sum + output_nodes[k].get_error_signal_times_weight(j);
			hidden_nodes[j].update_weights(error_sum);
		}

		// Calculate sum-of-squared-error
		double sum_of_squares_error = 0.0;
		double[] output_values = get_output(input_values);
		for (int i = 0; i < desired_values.length; i++)
			sum_of_squares_error = sum_of_squares_error + Math.pow( (desired_values[i] - output_values[i]), 2 );
		sum_of_squares_error = sum_of_squares_error * 0.5;

		// Return the sum-of-squared-error
		return sum_of_squares_error;
	}


	/**
	 *  Returns the sum-of-squared-error for given data set
	 */
	public double get_error(double[] sample_input, double[] corresponding_output)
	{
		double sum_of_squares_error = 0.0;
		double[] output_values = get_output(sample_input);
		for (int i = 0; i < corresponding_output.length; i++)
			sum_of_squares_error = sum_of_squares_error + Math.pow( (corresponding_output[i] - output_values[i]), 2 );
		sum_of_squares_error = sum_of_squares_error * 0.5;

		return sum_of_squares_error;
	}


	/**
	 * Returns the learning rate used by the network
	 */
	public double getLearningRateValue()
	{
		return output_nodes[0].learning_rate;
	}


	/**
	 * Returns the momentum used by the network
	 */
	public double getMomentumValue()
	{
		return output_nodes[0].momentum;
	}


	/**
	 * Returns the number of input units in the network
	 */
	public int getNumberOfInputUnits()
	{
		return num_input_nodes;
	}


	/**
	 * Returns the number of hidden units in the network
	 */
	public int getNumberOfHiddenUnits()
	{
		return num_hidden_nodes;
	}


	/**
	 * Returns the number of output units in the network
	 */
	public int getNumberOfOutputUnits()
	{
		return num_output_nodes;
	}


	/**
	 * Returns the weights of all of the links in the network, starting with the
	 * hidden nodes and movin on to the output nodes.
	 */
	public double[] getWeights()
	{
		LinkedList list = new LinkedList();
		if (num_hidden_nodes != 0) // in case there's no hidden layer
			for (int i = 0; i < num_hidden_nodes; i++)
				list.add(hidden_nodes[i].get_weight_values());
		for (int i = 0; i < num_output_nodes; i++)
			list.add(output_nodes[i].get_weight_values());
		
		Object[] object_list = list.toArray();
		double[][] double_double_list = new double[object_list.length][];
		int number_weights = 0;
		for (int i = 0; i < double_double_list.length; i++)
		{
			double_double_list[i] = (double[]) object_list[i];
			number_weights += double_double_list[i].length;
		}

		double[] results = new double[number_weights];
		int current = 0;
		for (int i = 0; i < double_double_list.length; i++)
			for (int j = 0; j < double_double_list[i].length; j++)
			{
				results[current] = double_double_list[i][j];
				current++;
			}

		return results;
	}
		

	/**
	 * Returns a string containting the current inputs, weights and values of the network nodes.
	 * Used during debugging
	 */
	public String print_network()
	{
		String network_structure = new String();
		
		for (int i = 0; i < num_input_nodes; i++)
		{
			network_structure += "INPUT NODE " + i + "\n";
			network_structure += input_nodes[i].output_values();
		}

		for (int i = 0; i < num_hidden_nodes; i++)
		{
			network_structure += "HIDDEN NODE " + i + "\n";
			network_structure += hidden_nodes[i].output_values();
		}

		for (int i = 0; i < num_output_nodes; i++)
		{
			network_structure += "OUTPUT NODE " + i + "\n";
			network_structure += output_nodes[i].output_values();
		}

		return network_structure;
	}


	/**
	 * Returns a string contating a vertical list of learning rate, momentum, number of
	 * input nodes, number of hidden nodes, number of output nodes and all weights (hidden
	 * nodes followed by output nodes).
	 */
	public String get_weight_string()
	{
		String weight_string = new String();

		// Collect network configuration information
		weight_string = output_nodes[0].learning_rate + "\n" + output_nodes[0].momentum +
						 "\n" + num_input_nodes + "\n" + num_hidden_nodes + "\n" +
		                 num_output_nodes + "\n";

		// Collect weights
		if (num_hidden_nodes != 0) // in case there's no hidden layer
		{
			for (int i = 0; i < num_hidden_nodes; i++)
				weight_string += hidden_nodes[i].get_weights();
		}
		for (int i = 0; i < num_output_nodes; i++)
			weight_string += output_nodes[i].get_weights();

		return weight_string;
	}


	/**
	 * Return the learning rate
	 */
	public String getLearningRate()
	{
		return String.valueOf(output_nodes[0].learning_rate);
	}


	/**
	 * Return the momentum
	 */
	public String getMomentum()
	{
		return String.valueOf(output_nodes[0].momentum);
	}


	/**
	 * Return the number of hidden nodes
	 */
	public String getHiddenNodesCount()
	{
		return String.valueOf(num_output_nodes);
	}

	
	/* PRIVATE METHODS *********************************************************/


	private double get_initial_weight(double minimum, double maximum)
	// Returns a random double between minimum and maximum
	{
		double random_number = Math.random();
		double difference = maximum - minimum;
		double proportion = random_number * difference;
		return minimum + proportion;
	}


	private void propagate()
	// Causes the network to perform all calculations to generate the values at the output
	// nodes corresponding to the values currently at the input nodes
	{
		double input_out; // the value of an input node
		double hidden_out; // the value of a hidden node

		if (num_hidden_nodes > 0) // case where hidden layer is present
		{
			for (int i = 0; i < num_input_nodes; i++)
			{
				input_out = input_nodes[i].get_current_value();
				for (int j = 0; j < num_hidden_nodes; j++)
					hidden_nodes[j].set_input(i, input_out);
			}
			
			for (int j = 0; j < num_hidden_nodes; j++)
			{
				hidden_nodes[j].calculate_value();
				hidden_out = hidden_nodes[j].get_current_value();
				for (int k = 0; k < num_output_nodes; k++)
					output_nodes[k].set_input(j, hidden_out);
			}

		}
		else // case where there is no hidden later
		{
			for (int i = 0; i < num_input_nodes; i++)
			{
				input_out = input_nodes[i].get_current_value();
				for (int k = 0; k < num_hidden_nodes; k++)
					output_nodes[k].set_input(i, input_out);
			}
		}

		for (int k = 0; k < num_output_nodes; k++)
			output_nodes[k].calculate_value();
	}
}