/*
 * @(#)OutputNode.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers.feedforward_neural_networks;

/**
 * Objects of this class represent the outputs of a neural network. OutputNodes can update
 * themselves during learning. Adjustable learning rates and momentums are permitted.The
 * sigmoidal activation funciton is used.
 * 
 * @author Cory McKay
 */
public class OutputNode
{
	/* FIELDS ******************************************************************/


	/**
	 * The learning rate of the node used during learning
	 */
	public double learning_rate;

	
	/**
	 * The momentum co-efficient used during learning
	 */
	public double momentum;

	
	protected int number_inputs;			// the number of links leading into the node
	protected double[] input_values;		// the current values at the inputs to the node
	protected double[] weights;				// the current weights of the inputs to the node
	protected double current_value;			// the current activation level of the node
	protected double error_signal;			// the current error signal (used during learning)
	protected double[] previous_update;		// the value of the change in weight of each link 
											//    during the last training iteration

	/* CONSTRUCTOR *************************************************************/


	/**
	 * Instantiate the OutputNode with the given parameters. inputs includes the bias node,
	 * which is the last one
	 */
	public OutputNode(double n, double alpha, int inputs)
	{
		// Set up fields
		learning_rate = n;
		momentum = alpha;
		number_inputs = inputs;
		input_values = new double[number_inputs];
		weights = new double[number_inputs];
		error_signal = 0.0;
		previous_update = new double[number_inputs];
		for (int i = 0; i < number_inputs; i++)
			previous_update[i] = 0.0;

		// Give matrices default values
		for (int i =0; i < number_inputs; i++)
		{
			input_values[i] = 1;
			weights[i] = 1;
		}

		// Set bias
		input_values[number_inputs - 1] = -1;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Set the current input value of the appropriate link to the given input.
	 */
	public void set_input(int link, double input)
	
	{
		input_values[link] = input;
	}


	/**
	 *  Set the current weight of the appropriate link to the given input.
	 */
	public void set_weight(int link, double input)
	{
		weights[link] = input;
	}


	/**
	 * Get the current value of the node
	 */
	public double get_current_value()
	{
		return current_value;
	}

	/**
	 * Return the weights
	 */
	public double[] get_weight_values()
	{
		return weights;
	}

	/**
	 * Return a vertical list of weights
	 */
	public String get_weights()
	{
		String weight_string = new String();
		for (int i = 0; i < weights.length; i++)
			weight_string += (weights[i] + "\n");
		return weight_string;
	}

	/**
	 * Perform calculations to set the current value of the node based on its current
	 * inputs and weights
	 */
	public void calculate_value()
	{
		double sum = 0.0;

		for (int i = 0; i < number_inputs; i++)
			sum = sum + input_values[i] * weights[i];
		
		current_value = activation_function(sum);
	}


	/**
	 * Update the weights of the node during a training iteration
	 */
	public void update_weights(double desired_output)
	{
		double old_weight; // the weight of a link prior to updating. Used to calculate the change in weight for
		                   // the momentum term

		calculate_error_signal(desired_output);

		for (int i = 0; i < number_inputs; i++)
		{
			old_weight = weights[i];
			weights[i] = weights[i] + learning_rate * error_signal * input_values[i] + momentum * previous_update[i];
			previous_update[i] = weights[i] - old_weight;
		}
	}


	/**
	 * Returns the error signal times the weight of the link back to the given hidden node
	 * Is used by a hidden node to calculate its own error signal
	 * update_weights should be called first on each iteration to ensure that error_signal is updated
	 */
	public double get_error_signal_times_weight(int link_index)
	{
		double value = weights[link_index] * error_signal;
		return value;
	}


	/**
	 * Returns the current inputs, weights and values of the node
	 * Used during debugging
	 */
	public String output_values()
	{
		String return_string = new String();
		for (int i = 0; i < number_inputs; i++)
			return_string += "In " + i + ": " + input_values[i] + "  Wt " + i + ": " + weights[i] +
			                     "  Val: " + current_value + "\n";
		return return_string;
	}


	/* PRIVATE METHODS *********************************************************/


	private double activation_function(double x)
	// Apply the sigmoidal activation function
	{
		double value = 1 / ( 1 + (Math.exp(-1 * x)) );
		return value;
	}


	private void calculate_error_signal(double desired_output)
	// Calculate the error signal (assuming sigmoidal activation function)
	{
		error_signal = (desired_output - current_value) * current_value * (1 - current_value);
	}
}