/*
 * @(#)Chromosome.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers.genetic_algorithms;

/**
 * Objects of this class represent the individuals that are evolved  in a genetic algorithm.
 * The genetic material of each chromosome is represented as an array of shorts, which are set
 * to either 1 or 0.
 *
 * <p>Objects of this class store their fitness, as calculated by an <code>Evaluator</code> object
 * that they are passed when they are instantiated. Evolution of <code>Chromosome</code> objects is
 * controlled by <code>Breeder</code> objects. <code>Chromosome</code> objects have no understanding
 *  of the meaning of these bit strings.
 *
 * <p>Objects of this class can be instantiated so that their array is randomly set, or they can
 * be instantiated so that the array is composed of two given components. Objects of this class
 * contain methods to return the entire array or sub-arrays split at a given point of division.
 *
 * <p>The <code>getDNA</code> method of this class can be called to get the DNA of a Chromosome.
 *
 * <p>The <code>getLeftDNA</code> method of this class can be called to get the first part of
 * the DNA of a Chromosome.
 *
 * <p>The <code>getRightDNA</code> method of this class can be called to get the second part of
 * the DNA of a Chromosome.
 *
 * <p>The <code>getFitness</code> method of this class can be called to get the stored fitness
 * of a Chromosome.
 *
 * <p>The <code>mutate</code> method of this class can be called to cause the Chromosome to
 * undergo mutation.
 *
 * @see	Evaluator
 * @see	Breeder
 */ 
public class Chromosome
{
	/* FIELDS ******************************************************************/


	private short[] bit_string;		// the genetic material of a Chromosome
	private double fitness;			// the fitness of a Chromosome


	/* CONSTRUCTORS ************************************************************/


	/**
	 * Randomly fill the genetic material of a Chromosome with 0's and 1's.
	 *
	 * <p>An exception is thrown if there is a problem evaluating the fitness of the DNA.
	 *
	 * @param	size	The number of bits in the Chromosome's DNA.
	 * @param	eval	The <code>Evaluator</code> object that calculatues the fitness of each Chromosome.
	 */
	public Chromosome (int size, Evaluator eval)
		throws Exception
	{
		// Getnerate bit string 
		bit_string = new short[size];
		for (int i = 0; i < size; i++)
		{
			if (Math.random() < 0.5) bit_string[i] = 0;
			else bit_string[i] = 1;
		}

		// Evaluate fitness of bit string
		fitness = eval.getFitness(bit_string);
	}

	
	/**
	 * Generates a Chromosome with genetic material composed of the combination of the two
	 * given parent pieces of genetic material.
	 *
	 * <p>An exception is thrown if there is a problem evaluating the fitness of the DNA.
	 *
	 * @param	left_DNA	The first part of the Chromosome's DNA.
	 * @param	right_DNA	The second part of the Chromosome's DNA.
	 * @param	eval		The <code>Evaluator</code> object that calculatues the fitness of each Chromosome.
	 */
	public Chromosome (short[] left_DNA, short[] right_DNA, Evaluator eval)
		throws Exception
	{
		// Form bit string
		bit_string = new short[left_DNA.length + right_DNA.length];
		for (int i = 0; i < left_DNA.length; i++)
			bit_string[i] = left_DNA[i];
		for (int i = left_DNA.length; i < left_DNA.length + right_DNA.length; i++)
			bit_string[i] = right_DNA[i - left_DNA.length];

		// Evaluate fitness of bit string
		fitness = eval.getFitness(bit_string);
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the entire genetic material of the Chromosome.
	 */
	public short[] getDNA()
	{
		return bit_string;
	}


	/**
	 * Returns the genetic material of the Chromosome up to and including the given split point
	 * (starting at 1, not 0). null is returned if the given split_point is 0
	 */
	public short[] getLeftDNA(int split_point)
	{
		if (split_point == 0) return null;

		short[] DNA_fragment = new short[split_point];
		
		for (int i = 0; i < split_point ; i++)
			DNA_fragment[i] = bit_string[i];

		return DNA_fragment;
	}


	/**
	 * 	Returns the genetic material of the Chromosome after but not including the given split point
	 * (starting at 1, not 0). null is returned if a split_point greater than or equal to the size
	 * of bit_string is specified.
	 */
	public short[] getRightDNA(int split_point)
	{
		if (split_point >= bit_string.length) return null;

		short[] DNA_fragment = new short[bit_string.length - split_point];
		
		for (int i = split_point; i < bit_string.length; i++)
			DNA_fragment[i - split_point] = bit_string[i];

		return DNA_fragment;
	}


	/**
	 * Returns the stored fitness of the Chromosome.
	 */
	public double getFitness()
	{
		return fitness;
	}


	/**
	 * Flips bits in bit string with a probability per bit given by the <i>probability</i> parameter. 
	 * This probablity must be between 0 and 1. This method is used to implement mutation.
	 */
	public void mutate(double probability)
	{
		for (int i = 0; i < bit_string.length; i++)
		{
			if (Math.random() < probability)
			{
				if (bit_string[i] == 0)
					bit_string[i] = 1;
				else if (bit_string[i] == 1)
					bit_string[i] = 0;
			}
		}
	}
}