/*
 * @(#)ClassifierWeightingEvaluator	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.classifiers.genetic_algorithms;

import bodhidharma.classifiers.bio_k_nearest_neighbour.*;
import bodhidharma.classifiers.ClassificationResultsInterpereter;
import java.util.LinkedList;

/**
 * Objects of this class evaluate and interpret the meaning of bit strings stored in
 * genetic algorithm <code>Chromosome</code> objects. Each bit in the bit string
 * represents part of the representation of a double. Each double represents the
 * weighting for a classifier. 
 *
 * The fitness value is calculated by performing a combined classification using the
 * bit string's weighting for all of the classifiers.
 *
 * @see	Chromosome
 * @see Breeder
 */
public class ClassifierWeightingEvaluator
	extends Evaluator
{
	/* FIELDS ******************************************************************/


	// Test classifier result values that will be classified to test performance. First indice
	// indicates classifier, second indice indicates recording and third indice indicates category.
	// Value of 1 for first indice corresponds to single-dimensional feature classifier. Is null if
	// this classifier not used.
	private	double[][][] test_classifier_results;

	// Model classifications of test_classifier_results. First indice indicates recording.
	// Entries indicate the indices of the categories that are correct classifications.
	private	int[][]	correct_classifications;

	// Whether or not each of the classifiers is a candidate for weighting. Weights for values of
	// false are automatically set to 0.
	private boolean[] ensemble_selection;

	// The possible number of categories into which recordings can be classified.
	private int number_categories;

	// Number of bits used to represent each weighting. More bits means a greater
	// variety of possible values.
	private	int		bits_per_weight;

	// Used to perform actual classifications so that fitnesses can be performed. Default fitness will be used if null.
	private	ClassificationResultsInterpereter results_interpereter;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor.
	 *
	 * @param	test_classifier_results		Test classifier result values that will be classified to test performance. First indice indicates classifier, second indice indicates recording and third indice indicates category. Value of 1 for first indice corresponds to single-dimensional feature classifier. Is null if this classifier not used.
	 * @param	correct_classifications		Model classifications of test_classifier_results. First indice indicates recording. Entries indicate the indices of the categories that are correct classifications.
	 * @param	ensemble_selection			Whether or not each of the classifiers is a candidate for weighting. Weights for values of false are automatically set to 0.
	 * @param	number_categories			The possible number of categories into which recordings can be classified.
	 * @param	bits_per_weight				Number of bits used to represent each weighting. More bits means a greater variety of possible values.
	 * @param	results_interpereter		Used to perform actual classifications so that fitnesses can be performed. Default fitness will be used if null.
	 */
	public ClassifierWeightingEvaluator( double[][][] test_classifier_results, int[][] correct_classifications,
	                                     boolean[] ensemble_selection, int number_categories, int bits_per_weight,
	                                     ClassificationResultsInterpereter results_interpereter )
	{
		bit_string_length = test_classifier_results.length;
		this.test_classifier_results = test_classifier_results;
		this.correct_classifications = correct_classifications;
		this.ensemble_selection = ensemble_selection;
		this.number_categories = number_categories;
		this.bits_per_weight = bits_per_weight;
		this.results_interpereter = results_interpereter;
		bit_string_length = bits_per_weight * test_classifier_results.length;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Return the fitness of a chromosome given its DNA (bit string). The fitness
	 * should vary between 0 (least fit possible) and 1 (most fit possible). 
	 * Bit strings passed to this method must be of the same length as the length 
	 * specified in the constructor of this object.
	 *
	 * <p>The fitness is calculated for each recording and category, and then averaged
	 * over all recordings. 
	 *
	 * <p>There are two ways of calculating fitness. The first is used if the results_interpereter
	 * field is not null, and the second if it is null. The first way gives a fitness
	 * based on the fraction of correct classifications based on the given bit_string.
	 *
	 * <p>The alternative fitness for each recording is set equal to the sum of the
	 * normalized scores returned by the classifier for those categories to which
	 * the recording belongs.
	 *
	 * <p>If previous feature selection set a classifier to not be used, its weighting
	 * is interepereted as 0, irregardless of what it actually is, and its value is
	 * changed to 0 in the bit string. This is done by <code>parseBitString</code> method.
	 *
	 * <p>Throws an exception if there is a problem with the classification or if
	 * the length of the <i>bit_string</i> is not equal to the length of the
	 * <i>bit_string_length</i> field.
	 */
	public double getFitness(short[] bit_string)
		throws Exception
	{
		// Throw an exception if the length of the bit_string is not equal to the
		// length of the bit_string_length> field
		if (bit_string.length != bit_string_length)
			throw new Exception("The given bit string has " + bit_string.length +
			                    " bits, but it should have " + bit_string_length + ".");

		// Set the classifier_weightings based on the value of the passed bit_string
		Double[] interpereted_bit_string = (Double[]) parseBitString(bit_string);
		double[] classifier_weightings = new double[interpereted_bit_string.length];
		for (int clas = 0; clas < classifier_weightings.length; clas++)
			classifier_weightings[clas] = interpereted_bit_string[clas].doubleValue();

		// Calculate the combined ensebmle results using ensemble_selection (because 
		// parseBitString mehtod sets falses to 0) and the appropriate feature weightings
		int number_recordings = correct_classifications.length;
		double[][] combined_results = new double[number_recordings][number_categories];
		for (int rec = 0; rec < number_recordings; rec++)
		{
			for (int cat = 0; cat < number_categories; cat++)
			{
				combined_results[rec][cat] = 0.0;
				for (int clas = 0; clas < test_classifier_results.length; clas++)
					if (test_classifier_results[clas] != null)
						combined_results[rec][cat] += test_classifier_results[clas][rec][cat] * classifier_weightings[clas];
			}
		}

		// The fitness
		double fitness;

		// Base fitness on basic scores
		if (results_interpereter == null)
		{
			// Calculate the model results in score form
			double[][] model_scores = new double[number_recordings][number_categories];
			for (int rec = 0; rec < number_recordings; rec++)
				for (int cat = 0; cat < number_categories; cat++)
					model_scores[rec][cat] = 0.0;
			for (int rec = 0; rec < number_recordings; rec++)
				for (int i = 0; i < correct_classifications[rec].length; i++)
					model_scores[rec][ correct_classifications[rec][i] ] = 1.0;

			// Calculate the errors of the ensemble
			double errors = 0.0;
			for (int rec = 0; rec < number_recordings; rec++)
				for (int cat = 0; cat < number_categories; cat++)
					errors += Math.abs( model_scores[rec][cat] - combined_results[rec][cat] );
			errors = errors / ((double) (number_recordings * number_categories));

			// Calculate the average fitness for all recordings
			fitness = 1.0 - errors;
		}

		// Base fitness on classifications
		else
		{
			// Calculate the model classifications
			boolean[][] model_winners = new boolean[number_recordings][number_categories];
			for (int rec = 0; rec < number_recordings; rec++)
				for (int cat = 0; cat < number_categories; cat++)
					model_winners[rec][cat] = false;
			for (int rec = 0; rec < number_recordings; rec++)
				for (int i = 0; i < correct_classifications[rec].length; i++)
					model_winners[rec][ correct_classifications[rec][i] ] = true;

			// Find the winners based on the given bit_string and classifier results
			boolean[][] winners = results_interpereter.getWinners(combined_results);

			// Find the correct fraction of categories found per recording
			int[] corrects = new int[number_recordings];
			for (int rec = 0; rec < corrects.length; rec++)
				corrects[rec] = correct_classifications[rec].length;
			double[] percentage_right_per_rec = new double[number_recordings];
			for (int rec = 0; rec < corrects.length; rec++)
			{
				int found_so_far = 0;
				for (int cat = 0; cat < winners[rec].length; cat++)
				{
					if (winners[rec][cat] && model_winners[rec][cat])
						found_so_far++;
				}
				percentage_right_per_rec[rec] = ((double) found_so_far) / ((double) corrects[rec]);
			}

			// Set the fitness to the fraction correct
			fitness = bodhidharma.utilities.Statistics.getAverage(percentage_right_per_rec);
		}		

		// Return the fitness
		return fitness;
	}


	/**
	 * Returns an array of Doubles holding the meaning of the given bit string
	 * of a <code>Chromosome</code>. These values represent feature weightings
	 * and fall between 0 and 1. These values are normalized accross the array.
	 *
	 * <p>If previous classifier selection set a classifier to not be used, its
	 * weighting is interepereted as 0, irregardless of what it actually is, and
	 * its value is changed to 0 in the bit string.
	 *
	 * <p>If all of the weightings are 0 for the given bit_string, one of the
	 * weightings selected by previous classifier selection is randomly chosen
	 * to be set to the maximum weighting.
	 *
	 * <p>Throws an exception if the length of the <i>bit_string</i> is not equal
	 * to the length of the <i>bit_string_length</i> field or if it is not evenly
	 * divisible by the word length.
	 */
	public Object[] parseBitString(short[] bit_string)
		throws Exception
	{
		// Throw an exception if the length of the bit_string is not equal to the
		// length of the bit_string_length> field
		if (bit_string.length != bit_string_length)
			throw new Exception("The given bit string has " + bit_string.length +
			                    " bits, but it should have " + bit_string_length + ".");

		// Throw an exception if the length of the bit_string is not evenly divisible
		// by the word length
		if (bit_string_length % bits_per_weight != 0)
			throw new Exception("The given bit string has is not evenly divisible by the " +
			                    "word length.");

		// To hold translated weightings
		double[] classifier_weightings = new double[bit_string_length / bits_per_weight];

		// Check if weighting should be set to 0 because of classifier selection
		// and do so if necessary
		for (int word = 0; word < classifier_weightings.length; word++)
		{
			int starting_indice = word * bits_per_weight;
			if (ensemble_selection != null)
				if (!ensemble_selection[word])
					for (int bit = 0; bit < bits_per_weight; bit++)
						bit_string[ starting_indice + bit ] = 0;
		}

		// Check if all weights in the bit string are set to 0
		boolean a_non_zero_weight_exists = false;
		for (int word = 0; word < classifier_weightings.length; word++)
		{
			int starting_indice = word * bits_per_weight;
			short[] this_word = new short[bits_per_weight];
			for (int bit = 0; bit < bits_per_weight; bit++)
				this_word[bit] = bit_string[ starting_indice + bit ];
			boolean non_zero = false;
			for (int digit = 0; digit < this_word.length; digit++)
				if (this_word[digit] != 0)
					non_zero = true;
			if (non_zero)
			{
				a_non_zero_weight_exists = true;
				word = classifier_weightings.length + 1; // exit loop
			}
		}

		// If all weights in the bit string are set to 0, then randomly set
		// one of them to the highest weight.
		if (!a_non_zero_weight_exists)
		{
			// Find the indices of the elligible classifiers
			LinkedList indices_of_elligible_classifiers_list = new LinkedList();
			if (ensemble_selection == null)
				for (int w = 0; w < classifier_weightings.length; w++)
					indices_of_elligible_classifiers_list.add(new Integer[w]);
			else
				for (int w = 0; w < ensemble_selection.length; w++)
					if (ensemble_selection[w])
						indices_of_elligible_classifiers_list.add(new Integer(w));
			Object[] temp_objs = indices_of_elligible_classifiers_list.toArray();
			int[] indices_of_elligible_classifiers = new int[temp_objs.length];
			for (int w = 0; w < indices_of_elligible_classifiers.length; w++)
				indices_of_elligible_classifiers[w] = ((Integer) temp_objs[w]).intValue();

			// Choose one to set as selected
			int ran = bodhidharma.utilities.Statistics.generateRandomNumber(indices_of_elligible_classifiers.length);
			int selected_indice_to_activate = indices_of_elligible_classifiers[ran];

			// Set the selected weighting to maximum weighting
			int starting_indice = selected_indice_to_activate * bits_per_weight;
			for (int bit = 0; bit < bits_per_weight; bit++)
				bit_string[ starting_indice + bit ] = 1;
		}

		// Translate the weightings
		for (int word = 0; word < classifier_weightings.length; word++)
		{
			// The starting indice of the word in bit_string
			int starting_indice = word * bits_per_weight;

			// Construct each word representing a weighting
			short[] this_word = new short[bits_per_weight];
			for (int bit = 0; bit < bits_per_weight; bit++)
				this_word[bit] = bit_string[ starting_indice + bit ];
		
			// Find the value of the word
			classifier_weightings[word] = translateABitWord(this_word);
		}

		// Normalize the weightings
		classifier_weightings = bodhidharma.utilities.Statistics.normalize(classifier_weightings);

		// Store weightings in double objects
		Double[] classifier_weightings_objects = new Double[classifier_weightings.length];
		for (int clas = 0; clas < classifier_weightings.length; clas++)
			classifier_weightings_objects[clas] = new Double(classifier_weightings[clas]);

		// Return the results
		return classifier_weightings_objects;
	}

	
	/* PRIVATE METHODS *********************************************************/


	// Returns the value of a given segment of a bit string
	// Bit strings of length bits are converted to double form using the formula:
	//	value = ((2^0*string(0) + 2^1*string(1) + ... + 2^(bits-1)*string(bits-1)) * 200 / 2^bits ) - 100
	// This results values between 0 and 1
	// An exception is thrown if the bit string is of an inappropriate length
	private double translateABitWord(short[] bit_string_segment)
		throws Exception
	{
		// Make sure that bit_string_segment is of the appropriate length
		if (bit_string_segment.length != bits_per_weight)
			throw new Exception("Inappropriate bit string size.");

		// Calculate the value of the word
		double value = 0.0;
		for (int i = 0; i < bits_per_weight; i++)
			value += Math.pow(2,i) * bit_string_segment[i];
		value = value / ((double) (Math.pow(2,bits_per_weight) - 1));
		
		// Return the value of the word
		return value;
	}
}