/*
 * @(#)FeatureSet.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma.data_structures;

import	bodhidharma.midi_parsing.*;
import	bodhidharma.midi_parsing.one_dimensional_features.*;
import	bodhidharma.midi_parsing.multi_dimensional_features.*;

/**
 * Part of the Bodhidharma automatic music classification
 * project. Holds the features extracted from a MIDI recording.
 * Features may be one-dimensional or multi-dimensional.
 *
 * @author Cory McKay
 * @see Feature
 * @see MIDIIntermediateRepresentations
 */
public class  FeatureSet
{
	/* FIELD *******************************************************************/

	/**
	 * The array of features for a recording
	 */
	public	Feature[]	features;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * Takes in the name of a MIDI file, extracts the features from it
	 * and stores them in <i>features,/i>. Throws an exception if there
	 * are any problems during parsing of the MIDI file or feature extraction.
	 * This exception contains information about the type of problem (it is
	 * generated by a MIDIIntermediateRepresentations object).
	 *
	 * <p>Only the features whose names are referred to in <i>features_to_extract</i>
	 * are extracted. An exception is thrown if an unidentified feature name
	 * is encountered.
	 *
	 * <p>Exceptions are thrown if there is a problem during parsing or if an
	 * improper feature value is found.
	 *
	 * @param		file_path			The path of a MIDI file
	 * @param		features_to_extract	Names of features to extract.
	 * @exception	Exception
	 */
	public FeatureSet(String file_path, String[] features_to_extract)
		throws Exception
	{
		MIDIIntermediateRepresentations recording_data = new MIDIIntermediateRepresentations(file_path);

		int number_features = features_to_extract.length;
		features = new Feature[number_features];
		
		for (int i = 0; i < features_to_extract.length; i++)
			features[i] = FeatureMaker.makeFeature(features_to_extract[i], recording_data);

		// Make sure that the features have valid values
		for (int feat = 0; feat < features.length; feat++)
		{
			// Check one dimensional features
			if (features[feat] instanceof FeatureOneDimensional)
			{
				if ( ((FeatureOneDimensional) features[feat]).getValue() < 0.0 )
					throw new Exception("The feature " + features[feat].getName() + " has a negative value: " +((FeatureOneDimensional) features[feat]).getValue() + ".");
				if ( ((FeatureOneDimensional) features[feat]).getValue() == Double.NaN )
					throw new Exception("The feature " + features[feat].getName() + " has a value of: not a number.");
				if ( ((FeatureOneDimensional) features[feat]).getValue() == Double.NEGATIVE_INFINITY )
					throw new Exception("The feature " + features[feat].getName() + " has a value of: negative infinity.");
				if ( ((FeatureOneDimensional) features[feat]).getValue() == Double.POSITIVE_INFINITY )
					throw new Exception("The feature " + features[feat].getName() + " has a value of: positive infinity.");
			}
			
			// Check multi-dimensional features
			if (features[feat] instanceof FeatureMultiDimensional)
			{
				double[] value = ((FeatureMultiDimensional) features[feat]).getValue();
				for (int dim = 0; dim < value.length; dim++)
				{
					if (value[dim] < 0.0 )
						throw new Exception("The feature " + features[feat].getName() + " has a negative value: " + value[dim] + ".");
					if (value[dim] == Double.NaN )
						throw new Exception("The feature " + features[feat].getName() + " has a value of: not a number.");
					if (value[dim] == Double.NEGATIVE_INFINITY )
						throw new Exception("The feature " + features[feat].getName() + " has a value of: negative infinity.");
					if (value[dim] == Double.POSITIVE_INFINITY )
						throw new Exception("The feature " + features[feat].getName() + " has a value of: positive infinity.");
				}
			}

		}
	}


	/**
	 * Stores the given array in the <i>features</i> field.
	 */
	public FeatureSet(Feature[] given_features)
	{
		features = given_features;
	}
}