/*
 * @(#)MIDIFileVerifier.java.java	1.1 06/25/04
 *
 * McGill Univarsity
 */

package bodhidharma.midi_parsing;

import  java.io.File;
import  java.io.IOException;
import  javax.sound.midi.*;
import	java.util.LinkedList;
import	bodhidharma.data_structures.Recording;
import	bodhidharma.utilities.StringMethods;

/**
 * A holder class for the verifyMIDIFiles method.
 *
 * @author	Cory McKay
 */
public class MIDIFileVerifier 
{
	/**
	 * Takes in an array of <code>Recording</code> objects. Checks that 
	 * the file referred to in the path of each object exists at the given
	 * paths.
	 *
	 * <p>Returns a <code>String</code> noting that no problems were encountered
	 * if all the files are found parsed succesfully. If not, returns a string
	 * identifying the files that could not be found, and gives some
	 * information as to the cause of the problem.
	 *
	 * <p>Although currently commented out, there is a section of code that
	 * can check that each file can be parsed by javax.sound.midi.
	 *
	 * @param	recordings	The set of recordings to be checked.
	 * @return	The results of the verification.
	 * @see		Recording
	 * @see		DefaultMutableTreeNode
	 */
	public static String verifyMIDIFiles (Recording[] recordings)
	{
		// List of any problems encountered
		LinkedList errors = new LinkedList();

		// Check files one by one
		for (int i = 0; i < recordings.length; i++)
		{
			// Any errors found in the current file
			String current_error = null;
			
			// Name of current file being checked
			String current_file = StringMethods.convertFilePathToFileName(recordings[i].file_path);

			// Check if file exists
			if (!(new File(recordings[i].file_path)).exists())
				current_error = new String(current_file + " does not exist in the specified directory.");

			/* Decommenting this part makes the system also check that the file can be parsed by javax.sound.midi
			// Check if file can be parsed
			else
			{
				Sequence sequence = null;
				try
				{
					sequence = MidiSystem.getSequence(new File(recordings[i].file_path));

					if (sequence == null)
						current_error = new String(current_file + " is not a valid MIDI file with content.");
					Track[] tracks = sequence.getTracks();
					if (sequence.getDivisionType() != Sequence.PPQ)
						current_error = new String(current_file + " is a MIDI file that uses SMPTE time encoding." +
											" Only PPQ time encoding is excepted here.");
					if ( ((double) sequence.getTickLength()) > ((double) Integer.MAX_VALUE) - 1.0)
						current_error = new String(current_file + " is too big to be processed.");
				}
				catch (IOException e)
					{current_error = new String(current_file + " is not a valid file.");}
				catch (InvalidMidiDataException e)
					{current_error = new String(current_file + " is not a valid MIDI file.");}
			}
			*/

			// Provide recording identification information if problem encountered
			if (current_error != null)
			{
				current_error += "\nTITLE: " + recordings[i].title + "    ARTIST: " + recordings[i].artist + "\nLEAF GENRES: ";
				for (int j = 0; j < recordings[i].leaf_categories.length; j++)
					current_error += recordings[i].leaf_categories[j] + " ";
				current_error += "\n\n"; 
			}

			// Store current error in errors linked list
			if (current_error != null)
				errors.add(current_error);
		}

		// Generate final report
		String final_report = new String();
		int number_files = recordings.length;
		final_report += (number_files - errors.size()) + " MIDI files out of " + number_files + " succesfully " +
		                 "found.\n\n";
		if (errors.size() == 0)
			final_report += "VERIFICATION SUCCESFUL";
		else
		{
			final_report += "PROBLEMS ENCOUNTERED:\n\n";
			for (int i = 0; i < errors.size(); i++)
				final_report += (String) errors.get(i);
		}

		// Return the final report
		return final_report;
	}
}