/*
 * @(#)AverageRangeOfGlissandosFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma.midi_parsing.*;
import java.util.LinkedList;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class AverageRangeOfGlissandosFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public AverageRangeOfGlissandosFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// If there are no pitch bends
			if (recording_info.pitch_bends_list.size() == 0)
				value = 0.0;

			else
			{
				// Generate array of pitch bends
				Object[] notes_objects = recording_info.pitch_bends_list.toArray();
				LinkedList[] notes = new LinkedList[notes_objects.length];
				for (int i = 0; i < notes.length; i++)
					notes[i] = (LinkedList) notes_objects[i];
				int[][] pitch_bends = new int[notes.length][];
				for (int i = 0; i < notes.length; i++)
				{
					Object[] this_note_pitch_bends_objects = notes[i].toArray();
					pitch_bends[i] = new int[this_note_pitch_bends_objects.length];
					for (int j = 0; j < pitch_bends[i].length; j++)
						pitch_bends[i][j] = ((Integer) this_note_pitch_bends_objects[j]).intValue();
				}

				// Find the range of the bend for each note
				double[] greatest_differences = new double[pitch_bends.length];
				for (int note = 0; note < greatest_differences.length; note++)
				{
					int greatest_so_far = 0;
					for (int bend = 0; bend < pitch_bends[note].length; bend++)
						if (Math.abs( 64 - pitch_bends[note][bend] ) > greatest_so_far)
							greatest_so_far = Math.abs( 64 - pitch_bends[note][bend] );
					greatest_differences[note] = (double) greatest_so_far;
				}

				// Calculate the value
				value = bodhidharma.utilities.Statistics.getAverage(greatest_differences);
			}
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Average Range of Glissandos");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Average range of Pitch Bends, where range is defined as the\n" + 
		                  "greatest value of the absolute difference between 64 and the\n" +
		                  "second data byte of all MIDI Pitch Bend messages falling between" +
		                  "the Note On and Note Off messages of any note.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("P-24");
	}
}