/*
 * @(#)DurationOfMelodicArcsFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma. midi_parsing.*;
import java.util.LinkedList.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class DurationOfMelodicArcsFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public DurationOfMelodicArcsFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			int total_number_intervening_intervals = 0;
			int number_arcs = 0;
			for (int chan = 0; chan < recording_info.melody_list.length; chan++)
			{
				if (chan != (10 - 1))
				{
					// Convert to array
					Object[] list_contents = recording_info.melody_list[chan].toArray();
					int[] intervals = new int[list_contents.length];
					for (int i = 0; i < intervals.length; i++)
						intervals[i] = ((Integer) list_contents[i]).intValue();

					// Find the number of arcs
					int direction = 0;
					for (int i = 0; i < intervals.length; i++)
					{
						// If arc is currently decending
						if (direction == -1)
						{
							if (intervals[i] < 0)
								total_number_intervening_intervals++;
							else if (intervals[i] > 0)
							{
								total_number_intervening_intervals++;
								number_arcs++;
								direction = 1;
							}
						}

						// If arc is currently ascending
						else if (direction == 1)
						{
							if (intervals[i] > 0)
								total_number_intervening_intervals++;
							else if (intervals[i] < 0)
							{
								total_number_intervening_intervals++;
								number_arcs++;
								direction = -1;
							}
						}

						// If arc is currently stationary
						else if (direction == 0)
						{
							if (intervals[i] > 0)
							{
								direction = 1;
								total_number_intervening_intervals++;
							}
							if (intervals[i] < 0)
							{
								direction = -1;
								total_number_intervening_intervals++;
							}
						}
					}
				}
			}

			// Calculate the value
			value = (double) total_number_intervening_intervals / (double) number_arcs;
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Duration of Melodic Arcs");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Average number of notes that separate melodic peaks and troughs\n" +
		                  "in any channel.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("M-18");
	}
}