/*
 * @(#)ImportanceOfLoudestVoiceFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma. midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class ImportanceOfLoudestVoiceFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public ImportanceOfLoudestVoiceFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// Find the number of channels with no note ons
			int silent_count = 0;
			for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
				if (recording_info.channel_statistics[chan][0] == 0)
					silent_count++;

			// Find the loudest channel
			int max_so_far = 0;
			int loudest_chan = 0;
			for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
				if (recording_info.channel_statistics[chan][0] != 0)
					if (recording_info.channel_statistics[chan][2] > max_so_far)
					{
						max_so_far = recording_info.channel_statistics[chan][2];
						loudest_chan = chan;
					}
			double loudest_average = (double) max_so_far;

			// Find the average of the other channels and set value
			int number_voices = recording_info.channel_statistics.length - silent_count;
			if (number_voices < 2)
				value = 0.0;
			else
			{
				double[] other_averages = new double[number_voices - 1];
				int count = 0;
				for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
					if (recording_info.channel_statistics[chan][0] != 0 && chan != loudest_chan)
					{
						other_averages[count] = (double) recording_info.channel_statistics[chan][2];
						count++;
					}
				double average = bodhidharma.utilities.Statistics.getAverage(other_averages);
				value = loudest_average - average;
			}
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Importance of Loudest Voice");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Difference between the average loudness of the loudest channel and the\n" +
		                  "average loudness of the other channels that contain at least one note.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("T-9");
	}
}