/*
 * @(#)MelodicIntervalsInLowestLineFeature.java	1.0 01/06/04
 *
 * McGill University
 */

package bodhidharma.midi_parsing.one_dimensional_features;

import bodhidharma. midi_parsing.*;

/**
 * Each object of this class calculates the value of a single 1-D feature
 * (described in the <code>getDescription</code> method) and stores it.
 *
 * <p>The constructor takes in an intermediate representation of a recording
 * and derives the value of the feature from it.
 *
 * <p>The <code>getName,</code> <code>getDescription,</code> and <code>getCode</code>
 * methods can be called to get information identifying the feature.
 *
 * @author Cory McKay
 */
public class MelodicIntervalsInLowestLineFeature
	extends FeatureOneDimensional
{
	/* CONSTRUCTOR *************************************************************/


	/**
	 * Take in the intermediate representations of a MIDI file and generate the feature
	 * value based on them. Sets the feature value to -1 if the parameter is null.
	 */
	public MelodicIntervalsInLowestLineFeature(MIDIIntermediateRepresentations recording_info)
	{
		if (recording_info != null)
		{
			// Find the channel with the lowest average pitch
			int min_so_far = 0;
			int lowest_chan = 0;
			for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
				if (recording_info.channel_statistics[chan][6] != 0 && chan != (10 - 1))
					if (recording_info.channel_statistics[chan][6] < min_so_far)
					{
						min_so_far = recording_info.channel_statistics[chan][6];
						lowest_chan = chan;
					}

			// Find the number of channels with no note ons (or that is channel 10 or that is the highest channel)
			int silent_count = 0;
			for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
				if (recording_info.channel_statistics[chan][0] == 0 || chan == (10 - 1))
					silent_count++;

			// Find the average melodic interval of notes in the other channels
			double[] intervals = new double[recording_info.channel_statistics.length - silent_count];
			int count = 0;
			for (int chan = 0; chan < recording_info.channel_statistics.length; chan++)
				if (recording_info.channel_statistics[chan][0] != 0 && chan != (10 - 1))
				{
					intervals[count] = (double) recording_info.channel_statistics[chan][3];
					count++;
				}
			double average = bodhidharma.utilities.Statistics.getAverage(intervals);
			
			// Set value
			value = ((double) recording_info.channel_statistics[lowest_chan][3]) / ((double) average);
		}

		else value = -1.0;
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Returns the name of the feature.
	 */
	public String getName()
	{
		return new String("Melodic Intervals in Lowest Line");
	}


	/**
	 * Returns a sentence or two describing the feature.
	 */
	public String getDescription()
	{
		return new String("Average melodic interval in semitones of the channel with the lowest average\n" +
		                  "pitch divided by the average melodic interval of all channels that contain\n" +
						  "at least two note.");
	}


	/**
	 * Returns the code number of the feature.
	 */
	public String getCode()
	{
		return new String("T-15");
	}
}