/*
 * @(#)LargeProgressBarDialog.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.utilities;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * A general purpose window that displays the progress of a long process
 * with three components as it occurs.
 *
 * <p>The user may cancel processing during processing by pressing on the <i>Cancel</i>
 * button. 
 *
 * <p>The actual processing is done by the <code>LargeProgressBarTaskMonitor</code>
 * object that is passed to the constructor. This object is also provides the
 * information needed to keep this window updated.
 *
 * <p>The parent <coed>JFrame</code> which instantiated this object is made
 * inaccessible to the user during processing in order to avoid concurrency problems.
 *
 * <p>This window includes three progress bars as well as a listing of the current
 * task, the time elapsed and the estimated processing time remaining for each bar.
 *
 * @author	Cory McKay
 * @see		LargeProgressBarTaskMonitor
 * @see		Timer
 * @see		JProgressBar
 */
 public class LargeProgressBarDialog
	extends JFrame
	implements ActionListener
{
	/* FIELDS ******************************************************************/

	// Performs and monitors task that is to be monitored by this JFrame
    private	LargeProgressBarTaskMonitor	task_monitor;

	// Timer that queries task_monitor to keep this JFrame up to date
    private	Timer					timer;

	// Window that is involved in instantiation of this window
	private JFrame					parent_window;
	
	// Components to display progress
	private	JProgressBar			top_progress_bar;
	private	JProgressBar			middle_progress_bar;
	private	JProgressBar			bottom_progress_bar;
	private JTextArea				top_report_text_area;
	private	JTextArea				middle_report_text_area;
	private	JTextArea				bottom_report_text_area;

	// Containers to hold Swing elements
	private Container				content_pane;
	private JPanel					progress_panel;
	private JPanel					button_panel;

	// Buttons
	private JButton					cancel_button;


	/* CONSTRUCTOR *************************************************************/


	/**
	 * A constructor that configures and displays the window, triggers processing
	 * by the given <code>LargeProgressBarTaskMonitor</code> and keeps the window updated
	 * during processing.
	 *
	 * <p>It is assumed that all sub-tasks will take an equal amount of time which, while
	 * incorrect in most cases, is nonetheless close enough for most applications.
	 *
  	 * @param	parent	Name of the <code>JFrame</code> window that instantiated this window.
  	 * @param	task_checker_and_performer	Performs processing and provides information needed for updates to this window.	
  	 * @param	time_between_polls			Time in milliseconds between polling of task_checker_and_performer.	
	 * @see		ProgressBarTaskMonitor
	 * @see		Timer
	 * @see		JProgressBar
	 */
	public LargeProgressBarDialog(JFrame parent, LargeProgressBarTaskMonitor task_checker_and_performer, double time_between_polls)
	{
		// Disable the parent window
		parent_window = parent;
		parent_window.setEnabled(false);

		// Configure overall window settings and instantiate related panels
		setTitle("Progress");
		int horizontal_gap = 6; // horizontal space between GUI elements
		int vertical_gap = 4; // horizontal space between GUI elements
		content_pane = getContentPane();
		content_pane.setLayout(new BorderLayout(horizontal_gap, vertical_gap));
		progress_panel = new JPanel(new GridLayout(6, 1, horizontal_gap, vertical_gap));
		button_panel = new JPanel();

		// Configure progress bars		
		top_progress_bar = new JProgressBar(0, 1);
		top_progress_bar.setValue(0);
		top_progress_bar.setStringPainted(true);
		middle_progress_bar = new JProgressBar(0, 1);
		middle_progress_bar.setValue(0);
		middle_progress_bar.setStringPainted(true);
		bottom_progress_bar = new JProgressBar(0, 1);
		bottom_progress_bar.setValue(0);
		bottom_progress_bar.setStringPainted(true);

		// Configure text areas
		int number_text_columns = 50;
		int number_text_rows = 2;
		top_report_text_area = new JTextArea(number_text_rows, number_text_columns);
		top_report_text_area.setEditable(false);
		top_report_text_area.setLineWrap(false);
		top_report_text_area.setCursor(null); //inherit the panel's cursor (see Java bug 4851758)
		middle_report_text_area = new JTextArea(number_text_rows, number_text_columns);
		middle_report_text_area.setEditable(false);
		middle_report_text_area.setLineWrap(false);
		middle_report_text_area.setCursor(null); //inherit the panel's cursor (see Java bug 4851758)
		bottom_report_text_area = new JTextArea(number_text_rows, number_text_columns);
		bottom_report_text_area.setEditable(false);
		bottom_report_text_area.setLineWrap(false);
		bottom_report_text_area.setCursor(null); //inherit the panel's cursor (see Java bug 4851758)

		// Configure buttons
		cancel_button = new JButton("Cancel");
		cancel_button.addActionListener(this);

		// Add labels and previously created elements to panels and container
		progress_panel.add(top_progress_bar);
		progress_panel.add(top_report_text_area);
		progress_panel.add(middle_progress_bar);
		progress_panel.add(middle_report_text_area);
		progress_panel.add(bottom_progress_bar);
		progress_panel.add(bottom_report_text_area);
		button_panel.add(cancel_button);
		content_pane.add(progress_panel, BorderLayout.CENTER);
		content_pane.add(button_panel, BorderLayout.SOUTH);

		// Cause program to react when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				cancel();
			}
		});

		// Set up task_monitor
		task_monitor = task_checker_and_performer;

		// Create a timer to check progress every time_between_polls and update
		// this window accordingly. Also takes appropriate actions when
		// processing is finished.
		timer = new Timer((new Double(time_between_polls)).intValue(), new ActionListener()
		{
			public void actionPerformed(ActionEvent evt)
			{
				if (task_monitor.startNewTopProgress())
				{
					top_progress_bar.setValue(0);
					top_progress_bar.setMaximum(task_monitor.getTopLengthOfTask());
				}
				top_progress_bar.setValue(task_monitor.getTopCurrentAmountCompleted());
				top_report_text_area.setText(task_monitor.getTopTextMessage());
				top_report_text_area.setCaretPosition(top_report_text_area.getDocument().getLength());
				if (task_monitor.startNewMiddleProgress())
				{
					middle_progress_bar.setValue(0);
					middle_progress_bar.setMaximum(task_monitor.getMiddleLengthOfTask());
				}
				middle_progress_bar.setValue(task_monitor.getMiddleCurrentAmountCompleted());
				middle_report_text_area.setText(task_monitor.getMiddleTextMessage());
				middle_report_text_area.setCaretPosition(middle_report_text_area.getDocument().getLength());
				if (task_monitor.startNewBottomProgress())
				{
					bottom_progress_bar.setValue(0);
					bottom_progress_bar.setMaximum(task_monitor.getBottomLengthOfTask());
				}
				bottom_progress_bar.setValue(task_monitor.getBottomCurrentAmountCompleted());
				bottom_report_text_area.setText(task_monitor.getBottomTextMessage());
				bottom_report_text_area.setCaretPosition(bottom_report_text_area.getDocument().getLength());

				if (task_monitor.isDone())
					done();
			}
		});

		// Prepare to begin the task and monitor it
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		timer.start();

		// Dispay this window
		pack();
		show();
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the cancel_button
		if (event.getSource().equals(cancel_button))
			cancel();
	}


	// React to the close button or the cancel button.
	// Stop the process, reenable the parent_window and hide this window without
	// taking any actions to enact results of processing.
	public void cancel()
	{
		timer.stop();
		task_monitor.stop();
		setCursor(null); //turn off the wait cursor
		parent_window.setEnabled(true);
		this.hide();		
	}


	/* PRIVATE METHODS *********************************************************/


	// React to finishing the task
	// Beeps, stops the timer, reenables the parent window and hides this window.
	private void done()
	{
		Toolkit.getDefaultToolkit().beep(); // beep when done
		timer.stop(); // stop the timer
		setCursor(null); //turn off the wait cursor
		parent_window.setEnabled(true);
		this.hide();
	}
}