/*
 * @(#)LargeProgressBarTaskMonitor.java	1.1 06/25/04
 *
 * McGill University
 */

package bodhidharma.utilities;

/**
 * An abstract class that is used by <code>LargeProgressBarDialog</code> windows
 * in order to perform the processing that is measured by the window as well
 * as keep the window updated.
 *
 * <p>All parts of this class are implemented here except for the <code>go</code>
 * method. This method should use a <code>SwingWorker</code> object to
 * start and perform the task to be performed. It should also keep the
 * fields of this object updated so that they can be queried by the calling
 * <code>LargeProgressBarDialog</code>.
 *
 * <p>The part of the class implemented here essentially provides an interface
 * for the calling <code>LargeProgressBarDialog</code> and calculates the time elapsed
 * and the estimated processing time remaining. The get and is methods can be called to
 * get updates on how the process is progressing. The <code>stop</code> method
 * can be called in order to stop processing if the process is to be cancelled.
 * The <code>markTaskComplete</code> method is used to note that the task has
 * been completed.
 *
 * @author	Cory McKay
 * @see		LargeProgressBarDialog
 * @see		SwingWorker
 */
public abstract class LargeProgressBarTaskMonitor
{
	/* FIELDS ******************************************************************/


	// The number of sub-tasks to be completed before entire task is done
	protected	int				top_length_of_task;
	protected	int				middle_length_of_task;
	protected	int				bottom_length_of_task;

	// Whether or not a new progress bar needs to be started for each task
	protected	boolean			new_top_task_started;
	protected	boolean			new_middle_task_started;
	protected	boolean			new_bottom_task_started;

	// The current number of sub-tasks completed
	protected	int				top_current_amount_completed;
	protected	int				middle_current_amount_completed;
	protected	int				bottom_current_amount_completed;
	
	// The time when the overall task began
	protected	long			top_time_started;
	protected	long			middle_time_started;
	protected	long			bottom_time_started;

	// Messages to be displayed on the LargeProgressBarDialog
	protected	String			top_progress_message;
	protected	String			middle_progress_message;
	protected	String			bottom_progress_message;

	// Whether or not the process has been completed or cancelled
	protected	boolean			task_completed;
	protected	boolean			task_canceled;
	

	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor that sets other private fields to defaluts.
	 */
	public LargeProgressBarTaskMonitor()
	{
		top_length_of_task = 1;
		middle_length_of_task = 1;
		bottom_length_of_task = 1;
		top_current_amount_completed = 0;
		middle_current_amount_completed = 0;
		bottom_current_amount_completed = 0;
		top_time_started = System.currentTimeMillis();
		middle_time_started = System.currentTimeMillis();
		bottom_time_started = System.currentTimeMillis();
		new_top_task_started = false;
		new_middle_task_started = false;
		new_bottom_task_started = false;
		top_progress_message = new String("");
		bottom_progress_message = new String("");
		middle_progress_message = new String("");
		task_completed = false;
		task_canceled = false;
    }


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Called from ProgressBarDialog to start the overall task. Sets up a SwingWorker
	 * object to carry out the thread and sets time_started. Keeps the appropriate
	 * fields updated (see above).
	 */
	public abstract void go();


	/**
	 * Called from a LargeProgressBarDialog to find out how many sub-tasks remain
	 * to be done.
	 */
	public boolean startNewTopProgress()
	{
		boolean temp = new_top_task_started;
		new_top_task_started = false;
		return temp;
	}


	/**
	 * Called from a LargeProgressBarDialog to find out how many sub-tasks remain
	 * to be done.
	 */
	public boolean startNewMiddleProgress()
	{
		boolean temp = new_middle_task_started;
		new_middle_task_started = false;
		return temp;
	}


	/**
	 * Called from a LargeProgressBarDialog to find out how many sub-tasks remain
	 * to be done.
	 */
	public boolean startNewBottomProgress()
	{
		boolean temp = new_bottom_task_started;
		new_bottom_task_started = false;
		return temp;
	}


	/**
	 * Called from a LargeProgressBarDialog to find out how many sub-tasks remain
	 * to be done.
	 */
	public int getTopLengthOfTask()
	{
		return top_length_of_task;
	}


	/**
	 * Called from a LargeProgressBarDialog to find out how many sub-tasks remain
	 * to be done.
	 */
	public int getMiddleLengthOfTask()
	{
		return middle_length_of_task;
	}


	/**
	 * Called from a LargeProgressBarDialog to find out how many sub-tasks remain
	 * to be done.
	 */
	public int getBottomLengthOfTask()
	{
		return bottom_length_of_task;
	}


	/**
	 * Called from a LargeProgressBarDialog to find out how many sub-tasks have been done.
	 */
	public int getTopCurrentAmountCompleted()
	{
		return top_current_amount_completed;
	}


	/**
	 * Called from a LargeProgressBarDialog to find out how many sub-tasks have been done.
	 */
	public int getMiddleCurrentAmountCompleted()
	{
		return middle_current_amount_completed;
	}


	/**
	 * Called from a LargeProgressBarDialog to find out how many sub-tasks have been done.
	 */
	public int getBottomCurrentAmountCompleted()
	{
		return bottom_current_amount_completed;
	}


	/**
	 * Called to find out if the overall task has completed.
	 */
	public boolean isDone()
	{
		return task_completed;
	}


	/**
	 * Called if the overall task has been cancelled.
	 */
	public boolean isCancelled()
	{
		return task_canceled;
	}


	/**
	 * Returns a <code>String</code> indicating how much time has already passed and the
	 * estimated time remaining as well as the task currently being performed.
	 */
	public String getTopTextMessage()
	{
		// Return empty if nothing ready yet
		if (top_length_of_task == 0)
			return new String("");

		// Include the message on the first line
		String return_string = new String(top_progress_message + "\n");
		
		// Include the time passed and estimate
		return_string += getTimeEstimate( (double) top_current_amount_completed,
		                                  (double) top_length_of_task,
		                                  top_time_started );

		// Return the message 
		return return_string;
	}


	/**
	 * Returns a <code>String</code> indicating how much time has already passed and the
	 * estimated time remaining as well as the task currently being performed.
	 */
	public String getMiddleTextMessage()
	{
		// Return empty if nothing ready yet
		if (middle_length_of_task == 0)
			return new String("");

		// Include the message on the first line
		String return_string = new String(middle_progress_message + "\n");
		
		// Include the time passed and estimate
		return_string += getTimeEstimate( (double) middle_current_amount_completed,
		                                  (double) middle_length_of_task,
		                                  middle_time_started );

		// Return the message 
		return return_string;
	}


	/**
	 * Returns a <code>String</code> indicating how much time has already passed and the
	 * estimated time remaining as well as the task currently being performed.
	 */
	public String getBottomTextMessage()
	{
		// Return empty if nothing ready yet
		if (bottom_length_of_task == 0)
			return new String("");

		// Include the message on the first line
		String return_string = new String(bottom_progress_message + "\n");
		
		// Include the time passed and estimate
		return_string += getTimeEstimate( (double) bottom_current_amount_completed,
		                                  (double) bottom_length_of_task,
		                                  bottom_time_started );

		// Return the message 
		return return_string;
	}


	/**
	 * Called if the overall task is cancelled. Stops processing.
	 */
	public void stop()
	{
		task_canceled = true;
	}


	/**
	 * Called if the overall task is complete. Stops processing.
	 */
	public void markTaskComplete()
	{
		task_completed = true;
	}


	/* PRIVATE METHODS *********************************************************/

	
	// Calculate the time passed since the start of task and estimate the remaining time
	private String getTimeEstimate(double current_amount_completed, double length_of_task, long time_started)
	{
		// Calculate the estimated time remaining before task is completed
		double fraction_completed = current_amount_completed / length_of_task;
		double time_already_passed_sec = (double) ((System.currentTimeMillis() - time_started) / 1000);
		double time_total_sec = time_already_passed_sec / fraction_completed;
		double time_remaining_sec = time_total_sec - time_already_passed_sec;
		int time_already_passed = (int) time_already_passed_sec;
		int time_remaining = (int) time_remaining_sec;
		if (time_remaining < 0)
			time_remaining = 0;

		// Devise the string to return
		String time_passed_message = findTimeWithProperUnits(time_already_passed);
		String time_remaining_message = findTimeWithProperUnits(time_remaining);
		if (current_amount_completed < 0.99)
			return new String("Time elapsed: " + time_passed_message);
		else
			return new String("Time elapsed: " + time_passed_message +
							  "        Estimated time remaining: " + time_remaining_message);
	}


	// Takes in a number of seconds and returns the appropriate amount of time
	// corresponding to this based on how many seconds there are. Some rounding
	// takes place.
	protected String findTimeWithProperUnits(int seconds)
	{
		if (seconds > 3601)
		{
			int hours = seconds / 3600;
			int minutes = (seconds - (hours * 3600)) / 60;
			return new String(hours + " hours and " + minutes + " minutes");
		}
		else if (seconds > 120)
		{
			int minutes = seconds / 60;
			return new String(minutes + " minutes");
		}
		else
			return new String(seconds + " seconds");
	}
}