/*
 * @(#)StringMethods.java	1.1.1 08/12/04
 *
 * McGill Univarsity
 */

package bodhidharma.utilities;

import java.io.File;
import java.text.DecimalFormat;

/**
 * A holder class for static methods relating to manipulating strings.
 *
 * @author	Cory McKay
 * @see		String
 */
public class StringMethods 
{
	/**
	 * Returns the name of the file referred to by the given path.
	 *
	 * @param	file_path The file path from which the file name is to be extracted.
	 * @return	The name of the file referred to in the parameter.
	 */
	public static String convertFilePathToFileName(String file_path)
	{
		return file_path.substring(file_path.lastIndexOf(File.separator) + 1, file_path.length());
	}


	/**
	 * Returns the name of the directory that the given filename
	 * is found in. Throws an exception if no valid directory
	 * separator is present.
	 */
	 public static String getDirectoryName(String file_path)
		 throws Exception
	{
		int index_of_last_separator = file_path.lastIndexOf(File.separatorChar);
		if (index_of_last_separator == -1)
			throw new Exception(file_path + " does not contain a valid directory separator.");
		return new String(file_path.substring(0, file_path.lastIndexOf(File.separator)) + File.separator);
	}


	/**
	 * Returns a copy of the given string with the last four characters removed. Returns null
	 * if the fourth to last character is not a period or there are less than five characters
	 * in the string.
	 *
	 * @param	filename The name of the file from which the extension is to be removed.
	 * @return	The name of the file with the extension removed.
	 */
	public static String removeExtension(String filename)
	{
		if (filename.length() < 5)
			return null;
		if (filename.charAt(filename.length() - 4) != '.')
			return null;
		return filename.substring(0, (filename.length() - 4));
	}


	/**
	 * Returns the three letter extension of the given file name. Returns null
	 * if the fourth to last character is not a period or there are less than five characters
	 * in the string.
	 *
	 * @param	filename The name of the file from which the extension is to be returned.
	 * @return	The extension of the file.
	 */
	public static String getExtension(String filename)
	{
		if (filename.length() < 5)
			return null;
		if (filename.charAt(filename.length() - 4) != '.')
			return null;
		return filename.substring((filename.length() - 4), filename.length());
	}


	/**
	 * Returns a copy of the given string with all but the first <i>number_characters</i> eliminated.
	 * If the given string is shorter than <i>number_characters</i>, then blank spaces are
	 * added to the end of the string in order to make it the full length.
	 *
	 * @param	string_to_shorten	The string to be shortened or have spaces added to its end.
	 * @param	number_characters	Number of characters in the new string.
	 * @return	The shortened string.
	 */
	public static String getBeginningOfString(String string_to_shorten, int number_characters)
	{
		String copy = new String(string_to_shorten);
		if (string_to_shorten.length() < number_characters)
		{
			int difference = number_characters - string_to_shorten.length();
			for (int i = 0; i < difference; i++)
				copy += " ";
			return copy;
		}
		else if (string_to_shorten.length() > number_characters)
			return string_to_shorten.substring(0, number_characters);
		else return copy;
	}


	/**
	 * Returns a copy of the given string with all but the first <i>number_characters</i> eliminated.
	 * If the given string is shorter than <i>number_characters</i>, then hyphens are are added
	 * to the end of the string in order to make it the full length (with two blank spaces on each side.
	 *
	 * @param	string_to_shorten	The string to be shortened or have spaces added to its end.
	 * @param	number_characters	Number of characters in the new string.
	 * @return	The shortened string.
	 */
	public static String getBeginningOfStringWithHyphenFiller(String string_to_shorten, int number_characters)
	{
		String copy = new String(string_to_shorten);
		if (string_to_shorten.length() < number_characters)
		{
			int difference = number_characters - string_to_shorten.length();
			for (int i = 0; i < difference; i++)
			{
				if (i == 0 || i == 1 || i == (difference - 2) || i == (difference - 1))
					copy += " ";
				else
					copy += "-";
			}
			return copy;
		}
		else if (string_to_shorten.length() > number_characters)
			return string_to_shorten.substring(0, number_characters);
		else return copy;
	}


	/**
	 * Returns a formatted version of <i>number_to_round</i>. It will always
	 * show a 0 before numbers less than one, and will only show up to
	 * <i>decimal_places</i> decimal places.
	 *
	 * <p>Values of not a number, negative infinity and positive infinity will
	 * be returned as NaN, -Infinity and Infinity respectively.
	 *
	 * @param	number_to_round	The number that is to be rounded.
	 * @param	decimal_places	The maximum number of decimal places that will be displayed.
	 */
	public static String getRoundedDouble(double number_to_round, int decimal_places)
	{
		if (number_to_round == Double.NaN)
			return new String("NaN");
		if (number_to_round == Double.NEGATIVE_INFINITY)
			return new String("-Infinity");
		if (number_to_round == Double.POSITIVE_INFINITY)
			return new String("Infinity");

		String format_pattern = "#0.";
		for (int i = 0; i < decimal_places; i++)
			format_pattern += "#";
		DecimalFormat formatter = new DecimalFormat(format_pattern);

		return formatter.format(number_to_round);
	}


	/**
	 * Returns the index in the possible_names array where the given_name parameter
	 * occurs. Throws an exception if it is not there.
	 *
	 * @param	given_name		The string to search for in possible_names.
	 * @param	possible_names	The array to search for given_names in.
	 * @return	The index in possible_names that contains given_name.
	 */
	public static int getIndexOfString(String given_name, String[] possible_names)
		throws Exception
	{
		for (int i = 0; i < possible_names.length; i++)
			if (given_name.equals(possible_names[i]))
				return i;
		throw new Exception("Unable to find " + given_name + ".");
	}


	public static String giveStringXMLEscapeCharacters(String data)
	{
		data = data.replaceAll("&", "&amp;");
		data = data.replaceAll("\"", "&quot;");
		data = data.replaceAll("\'", "&apos;");
		data = data.replaceAll("<", "&lt;");
		data = data.replaceAll(">", "&gt;");
		return data;
	}
}