/*
 * @(#)ParseFeaturesFileHandler.java	1.1 06/25/04
 *
 * McGill Univarsity
 */


package bodhidharma.xml_parsing;

import org.xml.sax.*;
import java.util.LinkedList;
import bodhidharma.data_structures.*;
import bodhidharma.midi_parsing.*;
import bodhidharma.midi_parsing.multi_dimensional_features.*;
import bodhidharma.midi_parsing.one_dimensional_features.*;


/**
 * An extension of the <i>Xerces</i> <code>XML DefaultHandler</code>
 * class that implements the <i>SAX</i> <code>ContentHandler</code>.
 * The methods of this class are called by an instance of an 
 * <code>XMLReader</code> while it is parsing an XML document.
 *
 * <p>This particular implementation is custom designed to parse
 * XML files of the <i>features_file</i> type used by the Bodhidharma
 * music classification system. A custom exceptions is thrown 
 * if the file is not of this type. At the end of parsing,
 * the contents of the files elements are stored in the
 * <i>parsed_file_contents</i> field.
 *
 * @author Cory McKay
 * @see	Bodhidharma
 * @see ParseFileHandler
 * @see XMLDocumentParser
 * @see DefaultHandler
 * @see ContentHandler
 * @see XMLReader
 */
public class ParseFeaturesFileHandler extends ParseFileHandler
{
	/* FIELDS ******************************************************************/


	/* NOTE: String[] parsed_file_contents:
	 * Holds the data extracted from the XML file.
	 */
	
	// Holds the FeatureSettings objects before they are finally converted to an array
	private	LinkedList		features_settings;

	// The FeatureSet currently being filled
	private	FeatureSettings	current_feature_settings;
	
	// Identifies what tag has been found
	private	int			tag_identifier;

	// A count of the number of start elements encountered
	private	int				count;


	/* PUBLIC METHODS **********************************************************/


	/**
	 * This method is called when the start of the XML file to be
	 * parsed is reached. Instantiates the <i>features_settings</i>
	 * field and sets feature_path_added to false and count to 0.
	 */
	public void startDocument()
	{
		features_settings = new LinkedList();
		count = 0;
	}


	/**
	 * This method is called when the start of an XML element
	 * is encountered. Instantiates a new <code>FeatureSettings</code>
	 * object if a <i>feature</i> tag was encountered. Otherwise lets
	 * the <code>characters</code> method know what kind of action to take.
	 *
	 * @param	name			Name of the element that is encountered.
	 * @throws	SAXException	Exception thrown if is wrong type of XML file.
	 */
	public void startElement(String namespace, String name, String qName, Attributes atts)
		throws SAXException
	{
		// Make sure is correct file type
		if (count == 0)
			if (!name.equals("features_file"))
				throw new SAXException("\n\nIt is in reality of the type " + name + ".");
		count++;

		// Identify the type of tag
		tag_identifier = 0;
		if (name.equals("feature"))
			current_feature_settings = new FeatureSettings();
		else if (name.equals("name"))
			tag_identifier = 1;
		else if (name.equals("extract"))
			tag_identifier = 2;
		else if (name.equals("candidate"))
			tag_identifier = 3;
		else if (name.equals("overide"))
			tag_identifier = 4;
	}


	/**
	 * This method responds to the contents of tags in a way
	 * determined by the name of the tag (as determined by the
	 * <code>startElement</code> method).
	 * 
	 * <p>The <code>FeatureSettings</code> object represented by
	 * <i>current_feature_settings</i> is filled by the contents
	 * of the <i>name</i>, <i>extract</i>, <i>candidate</i> and
	 * <i>overide</i> tags.
	 */
	public void characters(char[] ch, int start, int length)
	{
		String contents = new String(ch, start, length);

		if (tag_identifier == 1)
		{
			try{ current_feature_settings.setFeature(FeatureMaker.makeFeature(contents, null)); }
			catch (Exception e)
			{
				// System.out.println("UNRECOGNIZED FEATURE NAME: ->" + contents + "<-");
			}
		}
		else if (tag_identifier == 2)
		{
			if (contents.equals("false"))
				current_feature_settings.extract_feature = false;
			else
				current_feature_settings.extract_feature = true;
		}
		else if (tag_identifier == 3)
		{
			if (contents.equals("false"))
				current_feature_settings.candidate_feature = false;
			else
				current_feature_settings.candidate_feature = true;
		}
		else if (tag_identifier == 4)
		{
			if (contents.equals("false"))
				current_feature_settings.overide_feature = false;
			else
				current_feature_settings.overide_feature = true;
		}
	}


	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Adds the <i>current_feature_settings</i>
	 * to the <i>features_settings</i> list when the <i>feature</i>
	 * end tag is encountered.
	 *
	 * @param	name			Name of the element that is encountered.
	 */
	public void endElement(String namespace, String name, String qName)
	{
		if (name.equals("feature"))
			features_settings.add(current_feature_settings);
	}

	
	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Fills the <i>parsed_file_contents</i>
	 * field with the <i>features_settings</i> <code>LinkedList</code>.
	 */
	public void endDocument()
	{
		parsed_file_contents = features_settings.toArray();
	}
}