/*
 * @(#)ParseModelClassificationsFileHandler.java	1.1 06/25/04
 *
 * McGill Univarsity
 */


package bodhidharma.xml_parsing;

import org.xml.sax.*;
import java.util.LinkedList;


/**
 * An extension of the <i>Xerces</i> <code>XML DefaultHandler</code>
 * class that implements the <i>SAX</i> <code>ContentHandler</code>.
 * The methods of this class are called by an instance of an 
 * <code>XMLReader</code> while it is parsing an XML document.
 *
 * <p>This particular implementation is custom designed to parse
 * XML files of the <i>model_classifications_file</i> type used by the Bodhidharma
 * music classification system. A custom exceptions is thrown 
 * if the file is not of this type. At the end of parsing,
 * the contents of the files elements are stored in the
 * <i>parsed_file_contents</i> field.
 *
 * @author Cory McKay
 * @see ParseFileHandler
 * @see	Bodhidharma
 * @see XMLDocumentParser
 * @see DefaultHandler
 * @see ContentHandler
 * @see XMLReader
 */
public class ParseModelClassificationsFileHandler
	extends ParseFileHandler
{
	/* FIELDS ******************************************************************/


	/* NOTE: Object[] parsed_file_contents:
	 * Holds the data extracted from the XML file.
	 */
	
	// Counts how many elements have been encountered and read
	private int count;

	// Identifies what tag has been found
	private	int tag_identifier;

	// Stores the data for each ensemble
	private LinkedList ensembles;

	// Stores the recordings in an ensemble
	private LinkedList recordings;

	// Stores the categories in a recording
	private LinkedList categories;


	/* PUBLIC METHODS **********************************************************/


	/**
	 * This method is called when the start of the XML file to be
	 * parsed is reached. Initializes the parsed_file_contents
	 * field to an array of empty strings and initializes the
	 * count to 0.
	 */
	public void startDocument()
	{
		ensembles = new LinkedList();
		count = 0;
	}

	/**
	 * This method is called when the start of an XML element
	 * is encountered. Increments the count when this occurs.
	 *
	 * @param	name			Name of the element that is encountered.
	 * @throws	SAXException	Exception thrown if is wrong type of XML file.
	 */
	public void startElement(String namespace, String name, String qName, Attributes atts)
		throws SAXException
	{
		// Make sure is correct file type
		if (count == 0)
			if (!name.equals("model_classifications_file"))
				throw new SAXException(".\n\nIt is in reality of the type " + name + ".");
		count++;

		// Identify the type of tag
		tag_identifier = 0;
		if (name.equals("ensemble"))
			recordings = new LinkedList();
		else if (name.equals("recording"))
			categories = new LinkedList();
		else if (name.equals("category"))
			tag_identifier = 1;
	}

	/**
	 * This method is called when the end tag of an XML element
	 * is encountered.
	 *
	 * @param	name			Name of the element that is encountered.
	 */
	public void endElement(String namespace, String name, String qName)
	{
		if (name.equals("ensemble"))
			ensembles.add(recordings);
		else if (name.equals("recording"))
			recordings.add(categories);
	}

	/**
	 * This stores the name of a category in contents.
	 */
	public void characters(char[] ch, int start, int length)
	{
		String contents = new String(ch, start, length);
		if (tag_identifier == 1)
			categories.add(contents);
	}


	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Fills the <i>parsed_file_contents</i> field.
	 *
	 * <p>After this mehtod is called, the <i>parsed_file_contents</i> field
	 * will hold an array of Objects. Each Object corresponds to an ensemble,
	 * which is represented by a 2-D array of Strings (first indice identifies
	 * recording and second identifies category).
	 */
	public void endDocument()
	{
		Object[] results = ensembles.toArray();
		parsed_file_contents = new Object[results.length];
		for (int ens = 0; ens < results.length; ens++)
		{
			Object[] recs = ((LinkedList) results[ens]).toArray();
			String[][] cats_of_recs = new String[recs.length][];
			for (int rec = 0; rec < recs.length; rec++)
			{
				Object[] cats = ((LinkedList) recs[rec]).toArray();
				String[] category_names = new String[cats.length];
				for (int cat = 0; cat < category_names.length; cat++)
					category_names[cat] = (String) cats[cat];
				cats_of_recs[rec] = category_names;
			}
			parsed_file_contents[ens] = cats_of_recs;
		}
	}
}