/*
 * @(#)ParseNNBioKNNEnsembleFileHandler.java	1.1 06/25/04
 *
 * McGill Univarsity
 */


package bodhidharma.xml_parsing;

import org.xml.sax.*;
import java.util.LinkedList;


/**
 * An extension of the <i>Xerces</i> <code>XML DefaultHandler</code>
 * class that implements the <i>SAX</i> <code>ContentHandler</code>.
 * The methods of this class are called by an instance of an 
 * <code>XMLReader</code> while it is parsing an XML document.
 *
 * <p>This particular implementation is custom designed to parse
 * XML files of the <i>NN_BioKNN_Ensemble_file</i> type used by the Bodhidharma
 * music classification system. A custom exceptions is thrown 
 * if the file is not of this type. At the end of parsing,
 * the contents of the files elements are stored in the
 * <i>parsed_file_contents</i> field.
 *
 * @author Cory McKay
 * @see ParseFileHandler
 * @see	Bodhidharma
 * @see XMLDocumentParser
 * @see DefaultHandler
 * @see ContentHandler
 * @see XMLReader
 */
public class ParseNNBioKNNEnsembleFileHandler extends ParseFileHandler
{
	/* FIELDS ******************************************************************/


	/* NOTE: Object[] parsed_file_contents:
	 * Holds the data extracted from the XML file.
	 */
	
	// Counts how many elements have been encountered and read
	private int count;

	// Identifies what tag has been found
	private	int tag_identifier;

	// Stores items that are sure to occur once and only once
	private String[] fixed_fields;

	// Stores items of unknown length
	private LinkedList current_list;


	/* PUBLIC METHODS **********************************************************/


	/**
	 * This method is called when the start of the XML file to be
	 * parsed is reached. Initializes the fixed_fields
	 * field to an array of empty strings and initializes the
	 * count to 0. Initializes parsed_file_contents.
	 */
	public void startDocument()
	{
		fixed_fields = new String[16];
		for (int i = 0; i < fixed_fields.length; i++)
			fixed_fields[i] = new String("");
		count = 0;
		parsed_file_contents = new Object[7];
		for (int i = 0; i < parsed_file_contents.length; i++)
			parsed_file_contents[i] = null;
	}

	/**
	 * This method is called when the start of an XML element
	 * is encountered. Increments the count when this occurs.
	 *
	 * @param	name			Name of the element that is encountered.
	 * @throws	SAXException	Exception thrown if is wrong type of XML file.
	 */
	public void startElement(String namespace, String name, String qName, Attributes atts)
		throws SAXException
	{
		// Make sure is correct file type
		if (count == 0)
			if (!name.equals("NN_BioKNN_Ensemble_file"))
				throw new SAXException(".\n\nIt is in reality of the type " + name + ".");
		count++;

		// Identify the type of tag
		tag_identifier = 0;

		// Identify the tags corresponding to fixed_fields
		if (name.equals("identifier"))
			tag_identifier = 1;
		else if (name.equals("BioKNearestNeighbour_file_path"))
			tag_identifier = 2;
		else if (name.equals("already_trained"))
			tag_identifier = 3;
		else if (name.equals("ensemble_using_classifier_selection"))
			tag_identifier = 4;
		else if (name.equals("ensemble_using_classifier_weighting"))
			tag_identifier = 5;
		else if (name.equals("csw_training_fraction"))
			tag_identifier = 6;
		else if (name.equals("using_one_dimensional_features"))
			tag_identifier = 7;
		else if (name.equals("using_multi_dimensional_features"))
			tag_identifier = 8;
		else if (name.equals("bioknn_using_feature_selection"))
			tag_identifier = 9;
		else if (name.equals("bioknn_using_feature_weighting"))
			tag_identifier = 10;
		else if (name.equals("bioknn_iterations"))
			tag_identifier = 11;
		else if (name.equals("bioknn_acceptable_threshold"))
			tag_identifier = 12;
		else if (name.equals("bioknn_consecutive_iterations"))
			tag_identifier = 13;
		else if (name.equals("network_iterations"))
			tag_identifier = 14;
		else if (name.equals("network_acceptable_threshold"))
			tag_identifier = 15;
		else if (name.equals("network_consecutive_iterations"))
			tag_identifier = 16;

		// Identify the tags corresponding to the starts of new linked lists
		// and instantiate the new LinkedList
		else if ( name.equals("FeedForwardNeuralNetwork_file_paths") ||
		          name.equals("ensemble_selection") ||
		          name.equals("ensemble_weightings") ||
		          name.equals("possible_taxonomy_labels") ||
		          name.equals("one_dim_feature_labels") ||
		          name.equals("multi_dim_feature_labels") )
			current_list = new LinkedList();

		// Identify the tags corresponding to adding items to linked lists
		else if ( name.equals("paths") ||
		          name.equals("used") ||
		          name.equals("weight") ||
		          name.equals("label")  )
			tag_identifier = 17;

	}

	/**
	 * This method is called when the end tag of an XML element
	 * is encountered.
	 *
	 * @param	name			Name of the element that is encountered.
	 */
	public void endElement(String namespace, String name, String qName)
	{
		if (name.equals("FeedForwardNeuralNetwork_file_paths"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[1] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				parsed_file_contents[1] = temp;
			}
		}
		else if (name.equals("ensemble_selection"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[2] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				boolean[] results = new boolean[temp.length];
				for (int i = 0; i < results.length; i++)
				{
					if (temp[i].equals("true"))
						results[i] = true;
					else
						results[i] = false;
				}
				parsed_file_contents[2] = results;
			}
		}
		else if (name.equals("ensemble_weightings"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[3] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				double[] results = new double[temp.length];
				for (int i = 0; i < results.length; i++)
					results[i] = (Double.valueOf(temp[i])).doubleValue();
				parsed_file_contents[3] = results;
			}
		}
		else if (name.equals("possible_taxonomy_labels"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[4] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				parsed_file_contents[4] = temp;
			}
		}
		else if (name.equals("one_dim_feature_labels"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[5] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				parsed_file_contents[5] = temp;
			}
		}
		else if (name.equals("multi_dim_feature_labels"))
		{
			Object[] temp_obj = current_list.toArray();
			if (temp_obj[0] == null)
				parsed_file_contents[6] = null;
			else
			{
				String[] temp = new String[temp_obj.length];
				for (int i = 0; i < temp.length; i++)
					temp[i] = (String) temp_obj[i];
				parsed_file_contents[6] = temp;
			}
		}
	}

	/**
	 * This stores the contents of elements in the appropriate
	 * element of the fixed_fields array or in current_list.
	 */
	public void characters(char[] ch, int start, int length)
	{
		String contents = new String(ch, start, length);
		if (contents.equals("null"))
			contents = null;

		// Store items in fixed_fields, if appropriate
		if (tag_identifier == 1)
			fixed_fields[0] = contents;
		else if (tag_identifier == 2)
			fixed_fields[1] = contents;
		else if (tag_identifier == 3)
			fixed_fields[2] = contents;
		else if (tag_identifier == 4)
			fixed_fields[3] = contents;
		else if (tag_identifier == 5)
			fixed_fields[4] = contents;
		else if (tag_identifier == 6)
			fixed_fields[5] = contents;
		else if (tag_identifier == 7)
			fixed_fields[6] = contents;
		else if (tag_identifier == 8)
			fixed_fields[7] = contents;
		else if (tag_identifier == 9)
			fixed_fields[8] = contents;
		else if (tag_identifier == 10)
			fixed_fields[9] = contents;
		else if (tag_identifier == 11)
			fixed_fields[10] = contents;
		else if (tag_identifier == 12)
			fixed_fields[11] = contents;
		else if (tag_identifier == 13)
			fixed_fields[12] = contents;
		else if (tag_identifier == 14)
			fixed_fields[13] = contents;
		else if (tag_identifier == 15)
			fixed_fields[14] = contents;
		else if (tag_identifier == 16)
			fixed_fields[15] = contents;

		// Add items to current_list
		else if (tag_identifier == 17)
			current_list.add(contents);
	}

	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Fills the first entry of the <i>parsed_file_contents</i>
	 * with the <i>fixed_fields</i> array. Other entries of <i>parsed_file_contents</i>
	 * have already been filled.
	 */
	public void endDocument()
	{
		parsed_file_contents[0] = fixed_fields;
	}
}