/*
 * @(#)ParseRecordingsFileHandler.java	1.2.2 07/15/05
 *
 * McGill Univarsity
 */

package bodhidharma.xml_parsing;

import org.xml.sax.*;
import java.util.LinkedList;
import java.io.File;
import bodhidharma.*;
import bodhidharma.data_structures.*;

/**
 * An extension of the <i>Xerces</i> <code>XML DefaultHandler</code>
 * class that implements the <i>SAX</i> <code>ContentHandler</code>.
 * The methods of this class are called by an instance of an 
 * <code>XMLReader</code> while it is parsing an XML document.
 *
 * <p>This particular implementation is custom designed to parse
 * XML files of the <i>recordings_file</i> type used by the Bodhidharma
 * music classification system. A custom exceptions is thrown 
 * if the file is not of this type. At the end of parsing,
 * the contents of the files elements are stored in the
 * <i>parsed_file_contents</i> field.
 *
 * <p>After parsing, the <i>parsed_file_contents</i> field will hold an
 * array of objects. All of the objects will be <code>Recording</code> objects
 * with fields filled based on the tags in the parsed document. 
 *
 * <p>The <i>leaf_categories</i> fields of these <code>Recording</code> objects
 * will be filled, but the <i>ancestor_categories</i> fields will not (set to null).
 * If a lear category is not specified in the file, it will be filled in with "Unknown".
 * The <i>file_path</i> fields will be filled with a concatonation of the
 * <i>file_path</i> tag of the parsed file and the <i>filename</i> tag
 * for each recording, with the result being a path relative to the
 * <i>recordings_file</i> file that has been parsed.
 *
 * @author Cory McKay
 * @see	Bodhidharma
 * @see Recording
 * @see RecordingsPanel
 * @see ParseFileHandler
 * @see XMLDocumentParser
 * @see DefaultHandler
 * @see ContentHandler
 * @see XMLReader
 * @see LinkedList
 */
public class ParseRecordingsFileHandler
	extends ParseFileHandler
{

	
	/* FIELDS ******************************************************************/


	/* NOTE: public Recording[] parsed_file_contents:
	 * Holds the data extracted from the XML file.
	 */
	
	// Holds the Recording objects before they are finally converted to an array
	private	LinkedList	recordings_list;

	// The recording currently being filled
	private	Recording	current_recording;
	
	// The path of the MIDI files (relative to the location of the recordings_file)
	private String		file_path;
	
	// The name so far of the current leaf genre being processed
	private String		this_leaf;

	// Identifies what tag has been found
	private	LinkedList	leaf_categories;

	// Identifies what tag has been found
	private	int			tag_identifier;

	// A count of the number of start elements encountered
	private	int			count;


	/* PUBLIC METHODS **********************************************************/


	/**
	 * This method is called when the start of the XML file to be
	 * parsed is reached. Instantiates the <i>recordings_list</i>
	 * field and the <i>file_path</i> fields and sets  count to 0.
	 */
	public void startDocument()
	{
		recordings_list = new LinkedList();
		file_path = new String("");
		this_leaf = new String("");
		count = 0;
	}


	/**
	 * This method is called when the start of an XML element
	 * is encountered. Instantiates a new Recording if a <i>recording</i>
	 * tag was encountered or a new list of leaf categories if a <i>model_genres</i>
	 * tag was encountered. Otherwise lets the <code>characters</code>
	 * method know what kind of action to take.
	 *
	 * @param	name			Name of the element that is encountered.
	 * @throws	SAXException	Exception thrown if is wrong type of XML file.
	 */
	public void startElement(String namespace, String name, String qName, Attributes atts)
		throws SAXException
	{
		// Make sure is correct file type
		if (count == 0)
			if (!name.equals("recordings_file"))
				throw new SAXException("\n\nIt is in reality of the type " + name + ".");
		count++;

		// Identify the type of tag
		tag_identifier = 0;
		if (name.equals("recording"))
			current_recording = new Recording();
		else if (name.equals("model_genres"))
			leaf_categories = new LinkedList();
		else if (name.equals("filename"))
			tag_identifier = 1;
		else if (name.equals("title"))
			tag_identifier = 2;
		else if (name.equals("artist"))
			tag_identifier = 3;
		else if (name.equals("role"))
			tag_identifier = 4;
		else if (name.equals("genre"))
			tag_identifier = 5;
		else if (name.equals("file_path"))
			tag_identifier = 6;
	}


	/**
	 * This method responds to the contents of tags in a way
	 * determined by the name of the tag (as determined by the
	 * <code>startElement</code> method).
	 * 
	 * <p>Simply stores the names of the piece title or artist in a
	 * <code>Recording</code> object if the <i>title</i> or <i>artist</i>
	 * tags were just encountered. Stores the concatonation of the
	 * contents of the <i>file_path</i> and <i>filename</i> tags in
	 * the <code>Recording</code> object. Sets the appropriate boolean
	 * fields of the <code>Recording</code> object based on the contents
	 * of the <i>role</i> tag (the acceptable values are <i>training</i>,
	 * <i>testing</i> or <i>classification</i>). Stores the contents of
	 * <i>genre</i> tags in a list that is later converted to an array
	 * and stored in the <code>Recording</code> object by the
	 * <code>endElement</code> method.
	 */
	public void characters(char[] ch, int start, int length)
	{
		String contents = new String(ch, start, length);
		if (tag_identifier == 1)
		{
			if (current_recording.file_path == null)
				current_recording.file_path = file_path + File.separator + contents;
			else
				current_recording.file_path += contents;
		}
		else if (tag_identifier == 2)
		{
			if (current_recording.title == null)
				current_recording.title = contents;
			else
				current_recording.title += contents;
		}
		else if (tag_identifier == 3)
		{
			if (current_recording.artist == null)
				current_recording.artist = contents;
			else
				current_recording.artist += contents;
		}
		else if (tag_identifier == 4)
		{
			if (contents.equals("training"))
				current_recording.use_for_training = true;
			else if (contents.equals("testing"))
				current_recording.use_for_testing = true;
			else if (contents.equals("classification"))
				current_recording.use_for_classification = true;
		}
		else if (tag_identifier == 5)
			this_leaf += contents;
		else if (tag_identifier == 6)
			file_path += contents;
	}


	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Adds the current recording to the list
	 * of recordings when its end tag is encountered and adds
	 * the leaf categories to the current recording when the
	 * model_genres end tag is encountered.
	 *
	 * @param	name			Name of the element that is encountered.
	 */
	public void endElement(String namespace, String name, String qName)
	{
		if (name.equals("recording"))
			recordings_list.add(current_recording);
		else if (name.equals("model_genres"))
		{
			Object[] results = leaf_categories.toArray();
			String[] leaf_genre_array = new String[results.length];
			for (int i = 0; i < leaf_genre_array.length; i++)
				leaf_genre_array[i] = (String) results[i];
			current_recording.leaf_categories = leaf_genre_array;
		}
		else if (name.equals("genre"))
		{
			leaf_categories.add(this_leaf);
			this_leaf = new String("");
		}
	}

	
	/**
	 * This method is called when the end tag of an XML element
	 * is encountered. Fills the <i>parsed_file_contents</i>
	 * field with the <i>recordings_list</i> <code>LinkedList</code>.
	 *
	 * <p>If a lear category is not specified, fills it in with "Unknown".
	 */
	public void endDocument()
	{
		parsed_file_contents = recordings_list.toArray();
		for (int i = 1; i < parsed_file_contents.length; i++)
		{
			if (((Recording) (parsed_file_contents[i])).leaf_categories.length == 0)
			{
				String[] unknown_id = new String[1];
				unknown_id[0] = "Unknown";
				((Recording) (parsed_file_contents[i])).leaf_categories = unknown_id;
			}
		}
	}
}