/*
 * @(#)ClassificationPanel.java.java	1.2 06/29/05
 *
 * McGill Univarsity
 */

package bodhidharma;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.text.DecimalFormat;
import java.io.*;
import bodhidharma.data_structures.*;
import bodhidharma.utilities.*;
import bodhidharma.classifiers.*;
import bodhidharma.classifiers.feedforward_neural_networks.*;
import bodhidharma.classifiers.bio_k_nearest_neighbour.BioKNearestNeighbour;
import bodhidharma.classifiers.genetic_algorithms.GeneticAlgorithmJFrame;
import bodhidharma.midi_parsing.multi_dimensional_features.*;
import bodhidharma.midi_parsing.one_dimensional_features.*;
import bodhidharma.xml_parsing.*;

/**
 * Part of the Bodhidharma automatic music classification
 * project. This is the panel of the GUI that allows users
 * to train, classify and see the results of classifications.
 *
 * <p>This panel allows users to assign different functions
 * to buttons with combo box. Each combo box affects the function
 * of the button above it.
 *
 * <p>The way in which much of the operations are performed here
 * depends greatly on the <code>PreferencesPanel</code>.
 *
 * <p>The main display area gives reports.
 *
 * <p>The <i>Train</i> button causes the system to train the
 * classifiers using the currently loaded taxonomy, preferences,
 * features and recordings. Features must have already been
 * extracted from recordings in order for this to work. The
 * particular recordings used for training depends on the
 * the contents of the combo box below the <i>Train</i> button.
 * The <i>Randomly Partitioned Recs</i> option randomly selects
 * a portion of the recordings for training (based on the
 * value of the <i>Fraction of Recordings to use for Training</i>
 * preference) and reserves the rest for testing. The
 * <i>Recs Marked For Training</i> option only uses those
 * recordings marked with a Role of <i>training</i> for training.
 * The <i>All Recs</i> option trains on all currently available
 * recordings. All types of recording auto-generate the training progress
 * and selection / weightings reports (see below). A multi-level
 * progress bar is shown during training which gives an estimate
 * of the training time remaining. Training can be cancelled at any time.
 *
 * <p>The <i>Classify</i> button causes the system to classify
 * a certain set of recordings based on the contents of the following
 * combo box. The <i>Reserved Recs</i> option classifies any recordings
 * that have been reserved for testing, usually as a result of
 * <i>Randomly Partitioned Recs</i> training (see above). The
 * <i>Recs Marked For Classification</i> option classifies the
 * recordings with a role of <i>classification</i>. The
 * <i>Recs Marked For Testing</i> option classifies the recordings
 * with a role of <i>testing</i>. The <i>All Recs</i> option 
 * classifies all currently loaded recordings. All recordings must
 * have had their features already extracted for classification to
 * be performed. Classification only works if training has already been
 * performed. Classification results reports are automatically generated
 * upon completion of training (see below). An overall report is also
 * generated and displayed upon completion of classificaiton. Note
 * that success rates are only shown for the <i>Randomly Partitioned Recs</i>
 * option (although they can still be manually calculated otherwise).
 *
 * <p>The <i>Cross-Validation Test</i> automatically divides the loaded
 * recordings into different folds (with a number determined by the 
 * <i>Fraction of Recordings to use for Training</i> preference. Each
 * fold consists of training and testing partitions, such that each
 * recording serves as a testing recording exactly once accross all folds
 * and as a training recording in the other folds. Each fold is then trained
 * and tested, and the results of each fold are saved to disk. A combined
 * results report is then displayed and saved. Reports of the average feature
 * (one-dimensional and multi-dimensional) weightings and their standard
 * deviations accross folds are also saved for each ensemble. Note that the
 * weightings of individual one-dimensional features are scaled based on the 
 * overall weighting of one-dimensional features. The reported weightings 
 * have been normalized.
 *
 * <p>The <i>Report Training Progress</i>, <i>Report Classification Results</i>
 * and <i>Report Selections / Weightings</i> buttons show the auto-generated
 * reports specified by the combo boxes below them. "DD" refers to hierarchal
 * classifiers and "RR" refers to round robin classifiers.
 *
 * <p>The <i>Save Reports</i> button saves any currently loaded reports to
 * disk.
 *
 * <p>The other load and save buttons are used to save classifiers and
 * any reserved recordings to disk.
 *
 * <p>The <i>clearStoredData</i> and <i>clearMultiFoldStoredData</i>
 * methods are used to reset the classifiers.
 *
 * <p>The <i>getScaledFeatureValues</i>, <i>getMaxFeatureValueCutoffs</i>,
 * <i>getMinFeatureValueCutoffs</i> and <i>getOneDimFeatureNames</i> methods
 * provide information about the scaling of the features.
 *
 * <p>The <i>saveClassifierFiles</i> method is used to save classifiers
 * and associated test recordings to disk. The <i>saveReports</i> method
 * is used to save training, classification and selection/weighting reports
 * to disk.
 *
 * <p>The other public methods are used for training and classification.
 *
 * @author Cory McKay
 * @see	PreferencesPanel
 * @see	ClassificationResults_interpereter
 * @see	NN_BioKNN_Ensemble
 * @see MainJFrame
 */
public class ClassificationPanel
	extends JPanel
	implements ActionListener
{
	/* FIELDS ******************************************************************/


	// The array of classifiers
	private	NN_BioKNN_Ensemble[]	classifier_ensemble;

	// Recordings that have been reserved for testing when trained using recordings
	// randomly partitioned into trainining and testing recordings. Indice
	// indicates the recording. Set to null if no recordings reserved.
	// Unlike the model_testing_classifications for these recordings described
	// below, these do not vary from ensemble to ensemble.
	private Recording[]				testing_recordings;
	
	// The model classifications of the testing_recordings field. First indice
	// indicates the ensesmble, second indice indicates the recording and the
	// third indice indicates the category. There is only an entry for each category
	// to which a recording belongs to. Set to null if no recordings reserved.
	private String[][][]			model_testing_classifications;

	// The classifierss reported progress as they train. First indice corresponds
	// to that of classifier_ensemble, second to the classifier in the ensemble,
	// and third to the iteration.
	private double[][][]			combined_training_errors;

	// The names of recordings used for training (first entry) and their related leaf
	// categories (other entries). The first indice indicates ensemble, the second
	// indicates the name of the recording and the third indicates leaf category(ies). 
	private String[][][]			training_recordings_info;

	// The contents of the report of all training times for a given training run.
	private String					stored_training_times_report;

	// The individual classification results (before classifier selection and weighting have
	// been applied) for each of the classifiers in an ensemble in the classifier_ensemble
	// field. The first indice specifies ensemble, the second specifies the classifier,
	// the third specifies the recording and the fourth specifies the category.
	// Results are in the form of relative scores for each of the possible categories 
	// for each recording (not normalized) for each classifier. Scores should fall in the
	// range between 0.0 and 1.0 and higher scores correspond to a greater certainty that
	// a recording has a corresponding label. The last entry for the first indice holds
	// the combined classification data.
	private double[][][][]			individual_classifier_results_all_ensembles;

	// The combined classification results for all of the classifiers in an ensemble of
	// classifiers in the classifier_ensemble field(after classifier selection and weighting
	// have been applied). Results are in the form of relative scores for each of the possible
	// categories for each recording (not normalized). Scores should fall in the range between
	// 0.0 and 1.0 and higher scores correspond to a greater certainty that a recording has a
	// corresponding label. The first indice corresponds to the ensemble, the second to the 
	// recording and the third to the category. The last entry for the first indice holds
	// the combined classification data.
	private double[][][]			combined_classifier_results_all_ensembles;

	// The names of the candidate categories for the combined classification
	private String[]				candidate_categories_for_combined_classification;

	// The model classifications of the testing_recordings field used for the
	// combined classifiecatoin. First indice indicates the recording and the
	// second indice indicates the category. There is only an entry for each category
	// to which a recording belongs to. Set to null if no recordings reserved.
	private String[][]				model_testing_classifications_for_combined_classifier;

	// Existing fields stored over multiple folds. First indice identifies fold
	private double[][][][]			multi_fold_combined_classifier_results_all_ensembles;
	private String[][]				multi_fold_candidate_categories_for_combined_classification;
	private String[][][]			multi_fold_model_testing_classifications_for_combined_classifier;

	// Used to store weightings accross folds. First indice indicates fold, second indicates
	// ensemble and third indicates feature or classifier.
	private	double[][][]			one_dim_feature_weightings_accross_folds;
	private	double[][][]			classifier_weightings_accross_folds;

	// Used to store names and codes of ensembles, one-dimensional features and classifier ensembles accross
	// folds
	private String[]				ensemble_name_accross_folds;
	private	String[]				one_dim_features_names_accross_folds;
	private	String[]				classifier_names_accross_folds;
	private	String[]				one_dim_feature_codes_accross_folds;
	private	String[]				classifier_codes_accross_folds;

	// Total training time in seconds stored over multiple folds.
	private	int[]					mult_fold_training_times;

	// Directory to save reports during cross-validation
	private	File					 multi_fold_dir_to_save;

	// Use to perform training
	public ProgressBarTaskTrainMonitor training_monitor;

	// Window that holds this JPanel
	private	MainJFrame				parent_window;

	// Display panels
	public	JTextArea				results_display;
	private JScrollPane				results_display_scroll_pane;
	private JPanel					button_panel;

	// Displays general messages
	private JTextField				status_bar;

	// Buttons
	private JButton					train_button;
	private JButton					classify_button;
	private JButton					train_and_classify_with_cross_validation_button;
	private JButton					report_training_errors_button;
	private JButton					report_classification_results_button;
	private JButton					report_selection_weighting_button;
	private JButton					save_reports_button;
	private JButton					load_classifiers_button;
	private JButton					save_classifiers_button;
	private JButton					save_classifiers_as_button;

	// Combo boxes for testing and training
	private	JComboBox				training_combo_box;
	private	JComboBox				classifying_combo_box;

	// Combo boxes for reports
	private	JComboBox				training_errors_ensemble_selector;
	private	JComboBox				classification_results_selector;
	private	JComboBox				selection_weighting_selector;

	/* CONSTRUCTOR *************************************************************/


	/**
	 * Basic constructor that configures and instantiates the panel.
	 *
 	 * @param	parent	Name of the <code>MainJFrame</code> window that holds this panel.
	 * @see		ConfigureFileLocationsJFrame
	 * @see		MainJFrame
	 */
	public ClassificationPanel(MainJFrame parent)
	{
		// Store a pointer to the container that called this RecordingsPanel
		parent_window = parent;

		// Initialize layout settings
		setLayout(new BorderLayout());
		int horizontal_gap = 4; // horizontal space between GUI elements
		int vertical_gap = 4; // horizontal space between GUI elements

		// Set up status bar
		status_bar = new JTextField("");
		status_bar.setEditable(false);
		add(status_bar, BorderLayout.SOUTH);

		// Set up button display panel
		button_panel = new JPanel(new GridLayout(18, 1, horizontal_gap, vertical_gap));
		train_button = new JButton("Train");
		training_combo_box = new JComboBox();
		training_combo_box.setEditable(false);
		classify_button = new JButton("Classify");
		classifying_combo_box = new JComboBox();
		classifying_combo_box.setEditable(false);
		train_and_classify_with_cross_validation_button = new JButton("Cross-Validation Test");
		report_training_errors_button = new JButton("Report Training Progress");
		report_classification_results_button = new JButton("Report Classification Results");
		report_selection_weighting_button = new JButton("Report Selections / Weightings");
		training_errors_ensemble_selector = new JComboBox();
		training_errors_ensemble_selector.setEditable(false);
		classification_results_selector = new JComboBox();
		classification_results_selector.setEditable(false);
		selection_weighting_selector = new JComboBox();
		selection_weighting_selector.setEditable(false);
		save_reports_button = new JButton("Save Reports");
		load_classifiers_button = new JButton("Load Classifiers");
		save_classifiers_button = new JButton("Save Classifiers");
		save_classifiers_as_button = new JButton("Save Classifiers As");
		training_combo_box.addItem("Randomly Partitioned Recs");
		training_combo_box.addItem("Recs Marked For Training");
		training_combo_box.addItem("All Recs");
		classifying_combo_box.addItem("Reserved Recs");
		classifying_combo_box.addItem("Recs Marked For Classification");
		classifying_combo_box.addItem("Recs Marked For Testing");
		classifying_combo_box.addItem("All Recs");
		train_button.addActionListener(this);
		classify_button.addActionListener(this);
		train_and_classify_with_cross_validation_button.addActionListener(this);
		report_training_errors_button.addActionListener(this);
		report_classification_results_button.addActionListener(this);
		report_selection_weighting_button.addActionListener(this);
		save_reports_button.addActionListener(this);
		load_classifiers_button.addActionListener(this);
		save_classifiers_button.addActionListener(this);
		save_classifiers_as_button.addActionListener(this);
		button_panel.add(train_button);
		button_panel.add(training_combo_box);
		button_panel.add(classify_button);
		button_panel.add(classifying_combo_box);
		button_panel.add(train_and_classify_with_cross_validation_button);
		button_panel.add(new JLabel(""));
		button_panel.add(report_training_errors_button);
		button_panel.add(training_errors_ensemble_selector);
		button_panel.add(report_classification_results_button);
		button_panel.add(classification_results_selector);
		button_panel.add(report_selection_weighting_button);
		button_panel.add(selection_weighting_selector);
		button_panel.add(save_reports_button);
		button_panel.add(new JLabel(""));
		button_panel.add(load_classifiers_button);
		button_panel.add(save_classifiers_button);
		button_panel.add(save_classifiers_as_button);
		add(button_panel, BorderLayout.EAST);

		// Set up the results_display
		results_display = new JTextArea();
		results_display.setLineWrap(false);
		results_display.setEditable(false);
		results_display_scroll_pane = new JScrollPane(results_display);
		add(results_display_scroll_pane, BorderLayout.CENTER);

		// Set to no stored classifiers or related information 
		clearStoredData();
		clearMultiFoldStoredData();

		// Read in the starting classifiers, if any
		if (!(parent_window.file_locations.classifiers_path.equals("")))
			parseClassifiersFile(parent_window.file_locations.classifiers_path);

		// Update the status bar
		updateStatusBar();
	}


	/* PUBLIC METHODS **********************************************************/


	/**
	 * Calls the appropriate methods when the buttons are pressed.
	 *
	 * @param	event		The event that is to be reacted to.
	 */
	public void actionPerformed(ActionEvent event)
	{
		// React to the train_button
		if (event.getSource().equals(train_button))
		{
			try
			{
				// Randomly divide all loaded recordings into training and testing groups
				if (((String) training_combo_box.getSelectedItem()).equals("Randomly Partitioned Recs"))
					train(true, true);

				// Train only on recordings marked "training" as their role in the RecordingsPanel
				else if (((String) training_combo_box.getSelectedItem()).equals("Recs Marked For Training"))
					train(false, false);

				// Train on all currently loaded recordings
				else if (((String) training_combo_box.getSelectedItem()).equals("All Recs"))
					train(false, true);
			}
			catch (Exception e)
			{
				// e.printStackTrace();
				JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			}
		}

		// React to the classify_button
		else if (event.getSource().equals(classify_button))
		{
			// Classify all recordings currently reserved for testing
			if (((String) classifying_combo_box.getSelectedItem()).equals("Reserved Recs"))
				classifyReservedRecordings();

			// Classify recordings marked "classification" as their role in the RecordingsPanel
			else if (((String) classifying_combo_box.getSelectedItem()).equals("Recs Marked For Classification"))
				classifyRecordingsWithClassificationRole();

			// Classify recordings marked "testing" as their role in the RecordingsPanel
			else if (((String) classifying_combo_box.getSelectedItem()).equals("Recs Marked For Testing"))
				classifyRecordingsWithTestingRole();

			// Classify all currently loaded recordings
			else if (((String) classifying_combo_box.getSelectedItem()).equals("All Recs"))
				classifyAllRecordings();
		}

		// React to the train_and_classify_with_cross_validation_button
		else if (event.getSource().equals(train_and_classify_with_cross_validation_button))
			train_with_cross_validation();

		// React to the report_training_errors_button
		else if (event.getSource().equals(report_training_errors_button))
			displaySelectedTrainingErrors();

		// React to the report_classification_results_button
		else if (event.getSource().equals(report_classification_results_button))
		{
			if (classification_results_selector.getSelectedIndex() == classifier_ensemble.length)
				displayCombinedClassificationResults();
			else
				displaySelectedClassificationResultsForAnEnsemble();
		}

		// React to the report_selection_weighting_button
		else if (event.getSource().equals(report_selection_weighting_button))
			displaySelectedWeightingsAndSelections();

		// React to the save_reports_button
		else if (event.getSource().equals(save_reports_button))
		{
			try
			{
				File place_to_save = getReportSaveLocation(true);
				if (place_to_save != null)
					saveReports(place_to_save);	
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(null, ("Unable to save reports. " + e.getMessage()), "ERROR", JOptionPane.ERROR_MESSAGE);
				// e.printStackTrace();
			}
		}

		// React to the load_classifiers_button
		else if (event.getSource().equals(load_classifiers_button))
		{
			boolean ok_pressed = parent_window.file_locations.browseClassifiersPath(true);
			if (ok_pressed)
				parseClassifiersFile(parent_window.file_locations.classifiers_path);
		}

		// React to the save_classifiers_button
		else if (event.getSource().equals(save_classifiers_button))
			saveClassifierFiles("", true);

		// React to the save_classifiers_as_button
		else if (event.getSource().equals(save_classifiers_as_button))
			saveClassifierFiles("", false);
	}


	/**
	 * Reset stored classifiers and related information (except for mult-fold fields
	 * (use clearMultiFoldStoredData method for this)).
	 */
	public void clearStoredData()
	{
		classifier_ensemble = null;
		testing_recordings = null;
		model_testing_classifications = null;
		combined_training_errors = null;
		training_recordings_info = null;
		stored_training_times_report = null;
		individual_classifier_results_all_ensembles = null;
		combined_classifier_results_all_ensembles = null;
		candidate_categories_for_combined_classification = null;
		model_testing_classifications_for_combined_classifier = null;
		training_errors_ensemble_selector.removeAllItems();
		classification_results_selector.removeAllItems();
		selection_weighting_selector.removeAllItems();
		updateStatusBar();
	}

	
	/**
	 * Reset multi-fold fields.
	 */
	public void clearMultiFoldStoredData()
	{
		multi_fold_combined_classifier_results_all_ensembles = null;
		multi_fold_candidate_categories_for_combined_classification = null;
		multi_fold_model_testing_classifications_for_combined_classifier = null;
		mult_fold_training_times = null;
		multi_fold_dir_to_save = null;
		one_dim_feature_weightings_accross_folds = null;
		classifier_weightings_accross_folds = null;
		ensemble_name_accross_folds = null;
		one_dim_features_names_accross_folds = null;
		classifier_names_accross_folds = null;
		one_dim_feature_codes_accross_folds = null;
		classifier_codes_accross_folds = null;
	}


	/**
	 * Returns the scaled values of the given <i>recordings</i>. Scaling is performed by 
	 * a trained one-dimensional classifier. An exception is thrown if no such classifier
	 * is available or if there is a problem with the given recordings. Only one-dimensional
	 * feature values are returned, so the <code>getScaledFeatureNames</code> method
	 * should be called to understand what  features the second indice refers to. First
	 * indice refers to recording.
	 */
	public double[][] getScaledFeatureValues(Recording[] recordings_to_scale)
		throws Exception
	{
		if (classifier_ensemble == null)
			throw new Exception("No classifier is currently loaded that has been trained to scale\n" +
			                    "one-dimensional feature values.");
		return classifier_ensemble[0].getScaledOneDimensionalFeatureValues(recordings_to_scale);
	}


	/**
	 * Returns the maximum acceptable one-dimensional feature value for scaling. All featue values above this
	 * are scaled to 1. This is calculated by a trained one-dimensional classifier. An exception is thrown if
	 * no such classifier is availables. The <code>getScaledFeatureNames</code> method should be called in 
	 * order to understand what features the indice refer to.
	 */
	public double[] getMaxFeatureValueCutoffs()
		throws Exception
	{
		if (classifier_ensemble == null)
			throw new Exception("No classifier is currently loaded that has been trained to scale\n" +
			                    "one-dimensional feature values.");
		return classifier_ensemble[0].getMaxFeatureValueCutoffs();
	}


	/**
	 * Returns the minimum acceptable one-dimensional feature value for scaling. All featue values below this
	 * are scaled to 0. This is calculated by a trained one-dimensional classifier. An exception is thrown if
	 * no such classifier is availables. The <code>getScaledFeatureNames</code> method should be called in 
	 * order to understand what features the indice refer to.
	 */
	public double[] getMinFeatureValueCutoffs()
		throws Exception
	{
		if (classifier_ensemble == null)
			throw new Exception("No classifier is currently loaded that has been trained to scale\n" +
			                    "one-dimensional feature values.");
		return classifier_ensemble[0].getMinFeatureValueCutoffs();
	}


	/**
	 * Returns the names of all one-dimensional features in the given recordings in the order that they occur.
	 */
	public String[] getOneDimFeatureNames(Recording[] recordings_to_name_one_D_features_from)
	{
		// Find number one-dimensional features
		int number_one_dim_features = 0;
		for (int feat = 0; feat < recordings_to_name_one_D_features_from[0].feature_set.features.length; feat++)
			if (recordings_to_name_one_D_features_from[0].feature_set.features[feat] instanceof FeatureOneDimensional)
				number_one_dim_features++;

		// Find and return the feature names		
		String[] feature_names = new String[number_one_dim_features];
		int current = 0;
		for (int feat = 0; feat < recordings_to_name_one_D_features_from[0].feature_set.features.length; feat++)
			if (recordings_to_name_one_D_features_from[0].feature_set.features[feat] instanceof FeatureOneDimensional)
			{
				feature_names[current] = recordings_to_name_one_D_features_from[0].feature_set.features[feat].getName();
				current++;
			}
		return feature_names;
	}

	
	/**
	 * Saves a <i>classifiers_file</i> XML file with a related folder and the appropriate
	 * classifier files within it.
	 *
	 * <p>If the <code>is_basic_save</code> parameter is true, then the
	 * file is saved at the path specified in the <code>ConfigureFileLocationsJFrame</code>
	 * object's <code>classifiers_text_field</code> field (automatically overwriting
	 * any existing files with the same path). If the <code>is_basic_save</code>
	 * parameter is false, then the file is saved at a location chosen by the user
	 * with a <code>JFileChooser</code> dialog box.
	 *
	 * <p>Displays an infomative error message if a problem occurs.
	 *
	 * @param	comments		Any comments to be saved inside the comments element of the XML file.
	 * @param	is_basic_save	Whether this is a save (true) or save as (false) operation.
	 * @see		ConfigureFileLocationsJFrame
	 */
	public void saveClassifierFiles(String comments, boolean is_basic_save)
	{
		// Reset testing_recordings if needed
		if (model_testing_classifications == null)
			testing_recordings = null;

		// Get File to save to
		File to_save_to = null;
		if (is_basic_save) // perform a basic save operation
		{
			if (parent_window.file_locations.classifiers_path.equals(""))
			{
				JOptionPane.showMessageDialog(null, "No file name is specified for saving the classifiers.\nPlease enter one in the following dialog box.", "NOTE", JOptionPane.INFORMATION_MESSAGE);
				to_save_to = parent_window.file_locations.saveFileExternally(false, "classifiers_file");
			}
			else
				to_save_to = parent_window.file_locations.getFile(parent_window.file_locations.classifiers_path, true);
		}
		else // perform a save as operation
			to_save_to = parent_window.file_locations.saveFileExternally(false, "classifiers_file");

		// Write the file to disk
		if (to_save_to != null)
		{
			try
			{
				// Find all the appropriate file names
				String this_file_path = to_save_to.getPath();
				String classifier_sub_folder_path = bodhidharma.utilities.StringMethods.removeExtension(this_file_path) + File.separator;
				if (classifier_sub_folder_path == null)
					throw new Exception("File name must have a three letter extension.");
				String[] NN_BioKNN_Ensemble_file_paths;
				if (classifier_ensemble == null)
				{
					NN_BioKNN_Ensemble_file_paths = new String[1];
					NN_BioKNN_Ensemble_file_paths[0] = "null";
				}
				else
				{
					NN_BioKNN_Ensemble_file_paths = new String[classifier_ensemble.length];
					for (int clas = 0; clas < NN_BioKNN_Ensemble_file_paths.length; clas++)
						NN_BioKNN_Ensemble_file_paths[clas] = classifier_sub_folder_path + "NN_BioKNN_Ensemble_" + clas + ".xml";
				}
				String testing_recordings_file_path;
				if (testing_recordings == null)
					testing_recordings_file_path = "null";
				else
					testing_recordings_file_path = classifier_sub_folder_path + "testing_recordings.xml";
				String model_testing_classifications_file_path;
				if (model_testing_classifications == null)
					model_testing_classifications_file_path = "null";
				else
					model_testing_classifications_file_path = classifier_sub_folder_path + "model_testing_classifications.xml";
				String stored_training_times_report_file_path;
				if (stored_training_times_report == null)
					stored_training_times_report_file_path = "null";
				else
					stored_training_times_report_file_path = classifier_sub_folder_path + "stored_training_report.txt";

				// Create directory if does not exist already
				File folder = new File(classifier_sub_folder_path);
				if (folder.isFile() == false && folder.isDirectory() == false)
					folder.mkdir();

				// Write the NN_BioKNN_Ensemble_files
				if (classifier_ensemble != null)
					for (int clas = 0; clas < classifier_ensemble.length; clas++)
					{
						File NN_BioKNN_Ensemble_save_file = new File(NN_BioKNN_Ensemble_file_paths[clas]);
						classifier_ensemble[clas].save(NN_BioKNN_Ensemble_save_file);
					}

				// Write the testing recordings file
				if (testing_recordings != null)
				{
					File testing_recordings_save_file = new File(testing_recordings_file_path);
					FileOutputStream rec_to = new FileOutputStream(testing_recordings_save_file);
					DataOutputStream rec_writer = new DataOutputStream(rec_to);
					String rec_header = new String
					(
						"<?xml version=\"1.0\"?>\n" +
						"<!DOCTYPE recordings_file [\n" +
						"   <!ELEMENT recordings_file (comments, file_path, recording+)>\n" +
						"   <!ELEMENT comments (#PCDATA)>\n" +
						"   <!ELEMENT file_path (#PCDATA)>\n" +
						"   <!ELEMENT recording (filename, role, model_genres, title, artist)>\n" +
						"   <!ELEMENT filename (#PCDATA)>\n" +
						"   <!ELEMENT role (#PCDATA)>\n" +
						"   <!ELEMENT model_genres (genre*)>\n" +
						"   <!ELEMENT genre (#PCDATA)>\n" +
						"   <!ELEMENT title (#PCDATA)>\n" +
						"   <!ELEMENT artist (#PCDATA)>\n" +
						"]>\n\n" +
						"<recordings_file>\n\n" +
						"   <comments></comments>\n\n" + 	
						"   <file_path>" + testing_recordings[0].file_path.substring(0, testing_recordings[0].file_path.lastIndexOf(File.separator)) + "</file_path>\n\n"
					);
					rec_writer.writeBytes(rec_header);
					for (int i = 0; i < testing_recordings.length; i++)
					{
						rec_writer.writeBytes("   <recording>\n");
						String filename = bodhidharma.utilities.StringMethods.convertFilePathToFileName(testing_recordings[i].file_path);
						rec_writer.writeBytes("      <filename>" + filename + "</filename>\n");
						String role;
						if (testing_recordings[i].use_for_training)
							role = new String("training");		
						else if (testing_recordings[i].use_for_testing)
							role = new String("testing");		
						else
							role = new String("classification");	
						rec_writer.writeBytes("      <role>" + role + "</role>\n");
						rec_writer.writeBytes("      <model_genres>\n");
						for (int j = 0; j < testing_recordings[i].leaf_categories.length; j++)
							rec_writer.writeBytes("         <genre>" + testing_recordings[i].leaf_categories[j] + "</genre>\n");
						rec_writer.writeBytes("      </model_genres>\n");
						rec_writer.writeBytes("      <title>" + testing_recordings[i].title + "</title>\n");
						rec_writer.writeBytes("      <artist>" + testing_recordings[i].artist + "</artist>\n");
						rec_writer.writeBytes("   </recording>\n\n");
					}
					rec_writer.writeBytes("</recordings_file>");
					
					// Close the output stream
					rec_writer.close();
				}

				// Write the testing recording feature values file
				if (testing_recordings != null)
				{
					String feature_values_path = testing_recordings_file_path.substring(0, testing_recordings_file_path.length() - 4); // strip of .xml of testing_recordings_file_path
					feature_values_path += "_feature_values.xml";
					File testing_recordings_values_save_file = new File(feature_values_path);
					FileOutputStream rec_to = new FileOutputStream(testing_recordings_values_save_file);
					DataOutputStream writer = new DataOutputStream(rec_to);
					String header = new String
					(
						"<?xml version=\"1.0\"?>\n" +
						"<!DOCTYPE feature_values_file [\n" +
						"   <!ELEMENT feature_values_file (comments, recording*)>\n" +
						"   <!ELEMENT comments (#PCDATA)>\n" +
						"   <!ELEMENT recording (file_path, feature+)>\n" +
						"   <!ELEMENT file_path (#PCDATA)>\n" +
						"   <!ELEMENT feature (name, dimensions, value+)>\n" +
						"   <!ELEMENT name (#PCDATA)>\n" +
						"   <!ELEMENT dimensions (#PCDATA)>\n" +
						"   <!ELEMENT value (#PCDATA)>\n" +
						"]>\n\n" +
						"<feature_values_file>\n\n" +
						"   <comments></comments>\n\n"
					);
					writer.writeBytes(header);
					for (int rec = 0; rec < testing_recordings.length; rec++)
					{
						if (testing_recordings[rec].feature_set != null )
						{
							writer.writeBytes("   <recording>\n");
							writer.writeBytes("      <file_path>" + testing_recordings[rec].file_path + "</file_path>\n");

							for (int feat = 0; feat < testing_recordings[rec].feature_set.features.length; feat++)
							{
								writer.writeBytes("      <feature>\n");
								writer.writeBytes("         <name>" + testing_recordings[rec].feature_set.features[feat].getName() + "</name>\n");
								if (testing_recordings[rec].feature_set.features[feat] instanceof FeatureOneDimensional)
								{
									writer.writeBytes("         <dimensions>1</dimensions>\n");
									writer.writeBytes("         <value>" + ((FeatureOneDimensional) testing_recordings[rec].feature_set.features[feat]).getValue() + "</value>\n");
								}
								else if (testing_recordings[rec].feature_set.features[feat] instanceof FeatureMultiDimensional)
								{
									int number_dimensions = ((FeatureMultiDimensional) testing_recordings[rec].feature_set.features[feat]).getValue().length;
									writer.writeBytes("         <dimensions>" + number_dimensions + "</dimensions>\n");
									for (int val = 0; val < number_dimensions; val++)
										writer.writeBytes("         <value>" + ((FeatureMultiDimensional) testing_recordings[rec].feature_set.features[feat]).getValue()[val] + "</value>\n");
								}
								writer.writeBytes("      </feature>\n");
							}
							writer.writeBytes("   </recording>\n\n");
						}
					}
					writer.writeBytes("</feature_values_file>");
					writer.close();
				}

				// Write the model_testing_classifications_file
				if (model_testing_classifications != null)
				{
					File model_testing_classifications_save_file = new File(model_testing_classifications_file_path);
					FileOutputStream model_to = new FileOutputStream(model_testing_classifications_save_file);
					DataOutputStream model_writer = new DataOutputStream(model_to);
					String header_model_clas = new String
					(
						"<?xml version=\"1.0\"?>\n" +
						"<!DOCTYPE model_classifications_file [\n" +
						"   <!ELEMENT model_classifications_file (ensemble+)>\n" +
						"   <!ELEMENT ensemble (recording+)>\n" +
						"   <!ELEMENT recording (category?)>\n" +
						"   <!ELEMENT category (#PCDATA)>\n" +
						"]>\n\n" +
						"<model_classifications_file>\n\n"
					);
					model_writer.writeBytes(header_model_clas);
					for (int ens = 0; ens < model_testing_classifications.length; ens++)
					{
						model_writer.writeBytes("   <ensemble>\n");
						for (int rec = 0; rec < model_testing_classifications[ens].length; rec++)
						{
							model_writer.writeBytes("      <recording>\n");
							for (int cat = 0; cat < model_testing_classifications[ens][rec].length; cat++)
								model_writer.writeBytes("         <category>" + model_testing_classifications[ens][rec][cat] + "</category>\n");
							model_writer.writeBytes("      </recording>\n");
						}
						model_writer.writeBytes("   </ensemble>\n");
					}
					model_writer.writeBytes("</model_classifications_file>");
					model_writer.close();
				}

				// Write the stored_training_times_report file
				if (stored_training_times_report != null)
				{
					File stored_training_times_report_save_file = new File(stored_training_times_report_file_path);
					FileOutputStream times_to = new FileOutputStream(stored_training_times_report_save_file);
					DataOutputStream times_writer = new DataOutputStream(times_to);
					times_writer.writeBytes(stored_training_times_report);
					times_writer.close();
				}

				// Write the file that holds the links to the other files
				FileOutputStream to = new FileOutputStream(to_save_to);
				DataOutputStream writer = new DataOutputStream(to);
				String header_main = new String
				(
					"<?xml version=\"1.0\"?>\n" +
					"<!DOCTYPE classifiers_file [\n" +
					"   <!ELEMENT classifiers_file (classifiers_comments, classifier_sub_folder_path, NN_BioKNN_Ensemble_file_path+, testing_recordings_file_path, model_testing_classifications_file_path, stored_training_times_report_file_path)>\n" +
					"   <!ELEMENT classifiers_comments (#PCDATA)>\n" +
					"   <!ELEMENT classifier_sub_folder_path (#PCDATA)>\n" +
					"   <!ELEMENT NN_BioKNN_Ensemble_file_path (#PCDATA)>\n" +
					"   <!ELEMENT testing_recordings_file_path (#PCDATA)>\n" +
					"   <!ELEMENT model_testing_classifications_file_path (#PCDATA)>\n" +
					"   <!ELEMENT stored_training_times_report_file_path (#PCDATA)>\n" +
					"]>\n\n" +
					"<classifiers_file>\n\n" +
					"   <classifiers_comments>" + comments + "</classifiers_comments>\n\n"	
				);
				writer.writeBytes(header_main);
				writer.writeBytes("   <classifier_sub_folder_path>" + classifier_sub_folder_path + "</classifier_sub_folder_path>\n");
				for (int clas = 0; clas < NN_BioKNN_Ensemble_file_paths.length; clas++)
					writer.writeBytes("   <NN_BioKNN_Ensemble_file_path>" + NN_BioKNN_Ensemble_file_paths[clas] + "</NN_BioKNN_Ensemble_file_path>\n");
				writer.writeBytes("   <testing_recordings_file_path>" + testing_recordings_file_path + "</testing_recordings_file_path>\n");
				writer.writeBytes("   <model_testing_classifications_file_path>" + model_testing_classifications_file_path + "</model_testing_classifications_file_path>\n");
				writer.writeBytes("   <stored_training_times_report_file_path>" + stored_training_times_report_file_path + "</stored_training_times_report_file_path>\n");
				writer.writeBytes("</classifiers_file>");
				writer.close();
			}
			catch (Exception e)
			{
				JOptionPane.showMessageDialog(null, ("Unable to write file " + to_save_to.getName() + ". " + e.getMessage()), "ERROR", JOptionPane.ERROR_MESSAGE);
				// e.printStackTrace();
			}
		}
	}


	/**
	 * Saves all currently available reports in the given new directory.
	 */
	public void saveReports(File place_to_save)
		throws Exception
	{
		// Basic save path
		String save_path = place_to_save.getPath() + File.separator;

		// Save all training reports
		String training_path = save_path + File.separator + "TrainingReports" + File.separator;
		File training_directory = new File(training_path);
		training_directory.mkdir();
		for (int rep = 0; rep < training_errors_ensemble_selector.getItemCount(); rep++)
		{
			// Get the contents for an entry
			training_errors_ensemble_selector.setSelectedIndex(rep);
			displaySelectedTrainingErrors();
			String save_contents = results_display.getText();

			// Save the contents to disk
			String path_name = training_path + ((String) training_errors_ensemble_selector.getSelectedItem()) + ".txt";
			File save_file = new File(path_name);
			FileOutputStream to = new FileOutputStream(save_file);
			DataOutputStream writer = new DataOutputStream(to);
			writer.writeBytes(save_contents);
			writer.close();
		}

		// Save all classification reports
		String classification_path = save_path + File.separator + "ClassificationReports" + File.separator;
		File classification_directory = new File(classification_path);
		classification_directory.mkdir();
		for (int rep = 0; rep < classification_results_selector.getItemCount(); rep++)
		{
			// Get the contents for an entry
			classification_results_selector.setSelectedIndex(rep);
			if (rep == classifier_ensemble.length)
				displayCombinedClassificationResults();
			else
				displaySelectedClassificationResultsForAnEnsemble();
			String save_contents = results_display.getText();

			// Save the contents to disk
			String path_name = classification_path + ((String) classification_results_selector.getSelectedItem()) + ".txt";
			File save_file = new File(path_name);
			FileOutputStream to = new FileOutputStream(save_file);
			DataOutputStream writer = new DataOutputStream(to);
			writer.writeBytes(save_contents);
			writer.close();
		}

		// Save all weightings and selections reports
		String selection_path = save_path + File.separator + "SelectionReports" + File.separator;
		File selection_directory = new File(selection_path);
		selection_directory.mkdir();
		for (int rep = 0; rep < selection_weighting_selector.getItemCount(); rep++)
		{
			// Get the contents for an entry
			selection_weighting_selector.setSelectedIndex(rep);
			displaySelectedWeightingsAndSelections();
			String save_contents = results_display.getText();

			// Save the contents to disk
			String path_name = selection_path + ((String) selection_weighting_selector.getSelectedItem()) + ".txt";
			File save_file = new File(path_name);
			FileOutputStream to = new FileOutputStream(save_file);
			DataOutputStream writer = new DataOutputStream(to);
			writer.writeBytes(save_contents);
			writer.close();
		}

		// Clear screen
		results_display.setText("Reports succesfully saved.");
	}

	
	/**
	 * Perorms a multi-fold cross-validation. The number of folds depends on the results
	 * of the <code>getTrainingFraction</code> method of a <code>PreferencesPanel</code>
	 * object.
	 *
	 * <p>All reports are saved in a location specified by the user.
	 */
	public void train_with_cross_validation()
	{
		try
		{
			// Find the fraction of recordings to use for training in each fold
			double training_fraction = parent_window.preferences_panel.getTrainingFraction();
			int training_fraction_int = (new Long(Math.round(100.0 * training_fraction))).intValue();

			// Make sure can have equal sized folds
			if ( training_fraction_int != 50 && training_fraction_int != 67 && training_fraction_int != 75 &&
				 training_fraction_int != 80 &&training_fraction_int != 83 && training_fraction_int != 86 &&
				  training_fraction_int != 87 && training_fraction_int != 89 && training_fraction_int != 90 )
				throw new Exception("You have specified a Fraction of Recordings to use for Training\n" +
				                    "that does not allow an even number of folds. Please adjust this\n" +
				                    "in the Preferences Panel. Only fractions 0f 0.5, 0.67, 0.75, 0.83,\n" +
			                        "0.86, 0.87, 0.89 and 0.9 are permitted for cross-validation.");
			
			// Get the training_keys
			Recording[] all_eligible_recordings = parent_window.recordings_panel.recordings_list;
			String[][] temp_model_classification_names = new String[all_eligible_recordings.length][];
			for (int rec = 0; rec < all_eligible_recordings.length; rec++)
				temp_model_classification_names[rec] = all_eligible_recordings[rec].leaf_categories;
			String[] temp_elligible_categories = parent_window.taxonomy_panel.getLeafLabels();
			boolean[][] training_keys = getRandomlyChosenTrainingAndTestingSamplesForMultipleFolds( temp_model_classification_names,
			                                                                                        temp_elligible_categories,
			                                                                                        training_fraction );

			// Train, classify and save results one fold at a time
			performAFoldOfTrainingAndClassification(training_keys, 0, 0);
		}
		catch (Exception e)
		{
			clearMultiFoldStoredData();
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}


	/**
	 * Recursively performs the cross-validation training and classifying.
	 * Saves the reports for each fold and displays and saves an overall
	 * report. Also saves the averages and standard deviations of 
	 * feature and classifier weightings accross all cross-validation folds.
	 *
	 * <p>Works through a <code>ProgressBarTaskTrainMonitor</code> object.
	 *
	 * @param	training_keys		Which recordings to train for each fold. First indice is fold, second is recording.
	 * @param	current_fold		The current fold.
	 * @param	training_time_taken	The time taken to train the last fold.
	 */
	public void performAFoldOfTrainingAndClassification(boolean[][] training_keys, int current_fold, int training_time_taken)
		throws Exception
	{
		// During the first fold
		if (current_fold == 0)
		{
			// Remove any previously stored multi-fold data
			clearMultiFoldStoredData();

			// Instantiate multi-fold fields
			multi_fold_combined_classifier_results_all_ensembles = new double[training_keys.length][][][];
			multi_fold_candidate_categories_for_combined_classification = new String[training_keys.length][];
			multi_fold_model_testing_classifications_for_combined_classifier = new String[training_keys.length][][];
			mult_fold_training_times = new int[training_keys.length];
			one_dim_feature_weightings_accross_folds = new double[training_keys.length][][];
			classifier_weightings_accross_folds = new double[training_keys.length][][];
		
			// Ask where want save reports and create directory
			multi_fold_dir_to_save = getReportSaveLocation(false);
			if (multi_fold_dir_to_save == null)
				throw new Exception("Valid directory not specified. Cross-validation aborted.");
		}
		
		// During all folds but the first
		else
		{
			// Classify recordings for this fold
			classifyReservedRecordings();

			// Save reports in a new folder for this fold
			String save_path = multi_fold_dir_to_save.getPath() + File.separator + "Fold_" + current_fold;
			File new_directory = new File(save_path);
			new_directory.mkdir();
			saveReports(new_directory);

			// Update multi-fold fields
			multi_fold_combined_classifier_results_all_ensembles[current_fold - 1] = combined_classifier_results_all_ensembles;
			multi_fold_candidate_categories_for_combined_classification[current_fold - 1] = candidate_categories_for_combined_classification;
			multi_fold_model_testing_classifications_for_combined_classifier[current_fold - 1] = model_testing_classifications_for_combined_classifier;
			mult_fold_training_times[current_fold - 1] = training_time_taken;
		}

		// During all but the last go around
		if (current_fold < training_keys.length)
		{
			// Clear all stored data from previous training or classifications
			clearStoredData();

			// Include all currently loaded recordings
			Recording[] all_eligible_recordings = parent_window.recordings_panel.recordings_list;

			// Train this fold (this method will eventually be called again to reach further folds)
			performHierarchicalTrainingThroughProgressBar( all_eligible_recordings,
			                                               training_keys[current_fold],
			                                               false,
			                                               training_keys,
			                                               current_fold );
		}
		
		// End
		else
		{
			// Clear all non-multi fold data
			clearStoredData();

			// Save the average feature weightings accross folds in a new directory
			saveAverageWeightings();

			// Display and save the multi-fold report
			displayResultsAveragedOverFolds();
			String save_path = multi_fold_dir_to_save.getPath() + File.separator + "Overall_Report.txt";
			File new_path = new File(save_path);
			FileOutputStream to = new FileOutputStream(new_path);
			DataOutputStream writer = new DataOutputStream(to);
			writer.writeBytes(results_display.getText());
			writer.close();

			// Clear all multi fold data
			clearMultiFoldStoredData();
		}
	}


	/**
	 * Records the names and weightings of all one-dimensional features and classifiers in the given
	 * fold.
	 *
	 * @param	fold	The current fold.
	 */
	public void recordWeightingsAccrossFolds(int fold)
		throws Exception
	{
		try
		{
			// Find the names and codes of the one-dimensional features and classifiers (assumes
			// that all folds use the same ordering of features).
			if (fold == 0)
			{
				// Get names 
				one_dim_features_names_accross_folds = classifier_ensemble[0].getNamesOfFeaturesUsedByOneDimensionalClassifier();
				classifier_names_accross_folds = classifier_ensemble[0].getClassifierNames();

				// Get the name of the classifier
				ensemble_name_accross_folds = new String[selection_weighting_selector.getItemCount()];
				for (int ens = 0; ens < selection_weighting_selector.getItemCount(); ens++)
					ensemble_name_accross_folds[ens] = classifier_ensemble[ens].getClassifierName();

				// Get codes of features and classifiers
				FeatureSettings[] feature_settings = parent_window.features_panel.getFeatureSettings();
				String model_feature_codes[] = new String[feature_settings.length];
				String model_feature_names[] = new String[feature_settings.length];
				for (int feat = 0; feat < model_feature_codes.length; feat++)
				{
					model_feature_codes[feat] = feature_settings[feat].getFeature().getCode();
					model_feature_names[feat] = feature_settings[feat].getFeature().getName();
				}
				if (one_dim_features_names_accross_folds != null)
				{
					one_dim_feature_codes_accross_folds = new String[one_dim_features_names_accross_folds.length];
					for (int i = 0; i < one_dim_feature_codes_accross_folds.length; i++)
					{
						one_dim_feature_codes_accross_folds[i] = " ";
						for (int j = 0; j < model_feature_codes.length; j++)
							if (one_dim_features_names_accross_folds[i].equals(model_feature_names[j]))
							{
								one_dim_feature_codes_accross_folds[i] = model_feature_codes[j];
								j = model_feature_codes.length + 1; // exit loop
							}
					}
				}
				classifier_codes_accross_folds = new String[classifier_names_accross_folds.length];
				for (int i = 0; i < classifier_codes_accross_folds.length; i++)
				{
					classifier_codes_accross_folds[i] = " ";
					for (int j = 0; j < model_feature_codes.length; j++)
						if (classifier_names_accross_folds[i].equals(model_feature_names[j]))
						{
							classifier_codes_accross_folds[i] = model_feature_codes[j];
							j = model_feature_codes.length + 1; // exit loop
						}
				}
			}
			
			// Record all weightings for this fold
			one_dim_feature_weightings_accross_folds[fold] = new double[selection_weighting_selector.getItemCount()][];
			classifier_weightings_accross_folds[fold] = new double[selection_weighting_selector.getItemCount()][];
			for (int ens = 0; ens < selection_weighting_selector.getItemCount(); ens++)
			{
				// Get the weightings (null means that features were not used)
				one_dim_feature_weightings_accross_folds[fold][ens] = classifier_ensemble[ens].getOneDimensionalFeatureWeightings();
				classifier_weightings_accross_folds[fold][ens] = classifier_ensemble[ens].getEnsembleWeightings();
				
				// Use 1-D selection if weightings were not performed
				if (one_dim_features_names_accross_folds != null)
				{
					if (one_dim_feature_weightings_accross_folds[fold][ens] == null)
					{
						boolean[] selections = classifier_ensemble[ens].getOneDimensionalFeatureSelection();
						if (selections != null) // set to 1 if selected and to 0 if not
						{
							one_dim_feature_weightings_accross_folds[fold][ens] = new double[selections.length];
							for (int feat = 0; feat < selections.length; feat++)
							{
								if (selections[feat])
									one_dim_feature_weightings_accross_folds[fold][ens][feat] = 1.0;
								else
									one_dim_feature_weightings_accross_folds[fold][ens][feat] = 0.0;
							}
							
							one_dim_feature_weightings_accross_folds[fold][ens] = bodhidharma.utilities.Statistics.normalize(one_dim_feature_weightings_accross_folds[fold][ens]);
						}
						else if (selections == null) // set all to 1 if no selection was performed
						{
							one_dim_feature_weightings_accross_folds[fold][ens] = new double[classifier_ensemble[ens].getNamesOfFeaturesUsedByOneDimensionalClassifier().length];
							for (int feat = 0; feat < one_dim_feature_weightings_accross_folds[fold][ens].length; feat++)
								one_dim_feature_weightings_accross_folds[fold][ens][feat] = 1.0;
							one_dim_feature_weightings_accross_folds[fold][ens] = bodhidharma.utilities.Statistics.normalize(one_dim_feature_weightings_accross_folds[fold][ens]);
						}
					}
				}
				
				// Normalize weightings
				if (one_dim_features_names_accross_folds != null)
					one_dim_feature_weightings_accross_folds[fold][ens] = bodhidharma.utilities.Statistics.normalize(one_dim_feature_weightings_accross_folds[fold][ens]);
				classifier_weightings_accross_folds[fold][ens] = bodhidharma.utilities.Statistics.normalize(classifier_weightings_accross_folds[fold][ens]);
			}
		}
		catch (Exception e)
		{
			e.printStackTrace();
			throw new Exception("Could not collect weighting information for fold " + (fold + 1) + " .\nDETAILS: " + e.getMessage());
		}
	}


	/**
	 * Trains the array of NN_BioKNN_Ensembles in the classifier_ensemble field and records 
	 * the recordings which are reserved for testing them in the testing_recordings and
	 * model_testing_classifications fields. Also stores the names of the recordings
	 * used to train each ensemble in the training_recordings_info field and the training
	 * errors for each ensemble in the combined_training_errors field. This information
	 * can be accessed using the report_training_errors_button and the training_errors_ensemble_selector
	 * GUI elements. Information about the trained classifiers can be accessed using the
	 * report_selection_weighting_button and selection_weighting_selector fields.
	 *
	 * <p>If the training_key parameter is null, then all_eligible_recordings will be randomly
	 * partitioned into training and testing sets as specified by the getTrainingFraction method
	 * of the PrefencesPanel. If this parameter is not null, it will control this partitioning.
	 *
	 * <p>Any combination of four types of ensembles of classifiers will be trained, based on the selections
	 * in the PreferencesPanel. The four ensemble types, the order they appear in the arrays mentioned above
	 * (including NN_BioKNN_Ensembles) used and the subset of the recordings chosen for training are:
	 *
	 * <p><b>Leaf Classifier:</b> Classifies recordings into one or more of any of the leaf categories.
	 * Trained on all available training recordings. Ensemble is labeled "Leaf Classifier". Only
	 * one ensemble of this type is generated.
	 *
	 * <br><b>Root Classifier:</b> Classifies recordings into one or more of any of the root categories
	 * (the ones at the top of the taxonomy tree). Trained on all available training recordings. Ensemble
	 * is labeled "Root Classifier". Only one ensemble of this type is generated.
	 *
	 * <br><b>Direct Descendant Classifiers:</b> Classify recordings into one or more of any of the categories
	 * that are the direct descendants of any of a category node with children. Trained only on
	 * the subset of training recordings that belong, at least partially, to this parent category.
	 * One such ensemble is generated for each category node with children. Ensembles are labeled
	 * "DD <name of parent category>". 
	 *
	 * <br><b>Round Robin Classifiers:</b> Classify recordings into one or more of any pair of
	 * leaf categories. Trained only on the subset of training recordings that belong, at least
	 * partially, to one of these two leaf categories. One such ensemble is generated for every possible pair.
	 * Ensembles are labeled "RR <name of first category> / <name of second category>". 
	 *
	 * @param	all_eligible_recordings	Recordings that will be broken up into training and testing groups.
	 * @param	training_key			Which recordings are to be used for testing and which are to be used for training. Indice corresponds to that of the all_eligible_recordings parameter. Value of true means use for training. A value of null means that breakdown is randomly selected.
	 * @param	report_training_results	Whether or not training errors (i.e. classifier convergence) are reported. Overrides preferences settings.
	 * @param	training_monitor		ProgressBarTaskTrainMonitor that is monitoring progress of training. Can be null if no ProgressBarTaskTrainMonitor is used.
	 */
	public void performHierarchicalTraining( Recording[] all_eligible_recordings,
	                                         boolean[] training_key,
	                                         boolean report_training_results,
	                                         ProgressBarTaskTrainMonitor training_monitor )
		throws Exception
	{
		// Find user preferences
		PreferencesPanel preferences_panel = parent_window.preferences_panel;

		// Find information about the taxonomy
		String[] leaf_labels = parent_window.taxonomy_panel.getLeafLabels();
		String[] root_labels = parent_window.taxonomy_panel.getParentLabels();
		String[] intermediate_labels = parent_window.taxonomy_panel.getMiddleLabels();

		// Calculate the number of classification ensembles
		int number_ensembles = 0;
		if (preferences_panel.shouldClassifyIntoAllLeafCategories())
			number_ensembles++;
		if (preferences_panel.shouldClassifyIntoRootCategories())
			number_ensembles++;
		if (preferences_panel.shouldClassifyIntoDirectDescendantsOfParentCategories())
			number_ensembles += (intermediate_labels.length + root_labels.length);
		if (preferences_panel.shouldClassifyUsingRoundRobin())
			for (int i = 0; i < (leaf_labels.length - 1); i++)
				for (int j = i + 1; j < leaf_labels.length; j++)
					number_ensembles++;

		// Throw an exception if none of the three types of training are enabled
		if ( number_ensembles == 0 )
			throw new Exception("None of the three types of training are enabled. Please change your preferences.");

		// Prepare the array holding each ensemble of classifiers
		// The first classifies among all leaf categories and the second among all root
		// categories (if present). The next ones classify among the direct decendants of each
		// root or intermediate category (if present).
		classifier_ensemble = new NN_BioKNN_Ensemble[number_ensembles];

		// Prepare the field to store training errors for each ensemble
		combined_training_errors = new double[number_ensembles][][];

		// Prepare the field to store the names of training recordings used for each ensemble
		training_recordings_info = new String[number_ensembles][][];

		// To hold the names of the model categories for the testing_recordings for all ensembles
		model_testing_classifications = new String[number_ensembles][][];

		// Prepare the choices of training recordings and testing recordings if they are not
		// already known. If they are not, then randomly determine which recordings are to be
		// used for training and which for testing, using the percentage set in the preferences.
		// This percentage is applied separately to each leaf category.
		if (training_key == null)
		{
			double training_fraction = parent_window.preferences_panel.getTrainingFraction();
			String[][] temp_model_classification_names = new String[all_eligible_recordings.length][];
			for (int rec = 0; rec < all_eligible_recordings.length; rec++)
				temp_model_classification_names[rec] = all_eligible_recordings[rec].leaf_categories;
			String[] temp_elligible_categories = parent_window.taxonomy_panel.getLeafLabels();
			training_key = getRandomlyChosenTrainingAndTestingSamples(temp_model_classification_names, temp_elligible_categories, training_fraction);
		}

		// Find the number of recordings to be used for testing and for training
		int number_training_recordings = 0;
		int number_testing_recordings = 0;
		for (int rec = 0; rec < training_key.length; rec++)
		{
			if (training_key[rec]) number_training_recordings++;
			else number_testing_recordings++;
		}

		// To hold the recordings to be used for testing for all ensembles
		testing_recordings = new Recording[number_testing_recordings];

		// To hold the recordings to be used as candidatesfor training for all ensembles
		Recording[] pot_training_recordings = new Recording[number_training_recordings];

		// Find the recordings to be used for testing and for training
		int current_training_indice = 0;
		int current_testing_indice = 0;
		for (int rec = 0; rec < training_key.length; rec++)
		{
			if (training_key[rec])
			{
				pot_training_recordings[current_training_indice] = all_eligible_recordings[rec];
				current_training_indice++;
			}
			else
			{
				testing_recordings[current_testing_indice] = all_eligible_recordings[rec];
				current_testing_indice++;
			}
		}

		// Variables that will vary from ensemble to ensemble
		int ensemble_indice = -1; // To hold the indice of the classifier currently being configured
		String ensemble_label; // To hold the name of the ensemble currently being configured
		String[] elligible_categories; // To hold the elligible classification categories for a given ensemble
		boolean[] elligible_recordings; // The recordings in pot_training_recordings to use in specific_training_recordings for a given ensemble
		Recording[] specific_training_recordings; // The recordings used to train a given ensemble
		String[][] model_training_classifications; // To hold the model classifications of the specific_training_recordings in the context of given elligible_categories

		// Prepare and train leaf classifier ensemble, where all recordings are classified directly into leaf genres
		if (preferences_panel.shouldClassifyIntoAllLeafCategories())
		{
			ensemble_indice++;
			ensemble_label = "Leaf Classifier";
			elligible_categories = leaf_labels;
			model_training_classifications = new String[pot_training_recordings.length][];
			for (int rec = 0; rec < pot_training_recordings.length; rec++)
				model_training_classifications[rec] = pot_training_recordings[rec].leaf_categories;
			elligible_recordings = getRecordingsBelongingToElligibleCategories(model_training_classifications, elligible_categories);
			model_training_classifications = getReducedModelClassifications(model_training_classifications, elligible_recordings);
			specific_training_recordings = getReducedRecordings(pot_training_recordings, elligible_recordings);
			model_testing_classifications[ensemble_indice] = new String[testing_recordings.length][];
			for (int rec = 0; rec < testing_recordings.length; rec++)
				model_testing_classifications[ensemble_indice][rec] = testing_recordings[rec].leaf_categories;
			trainAnEnsembleAndReportErrors( ensemble_indice, ensemble_label, elligible_categories,
											specific_training_recordings, model_training_classifications,
			                                report_training_results, training_monitor );
		}

		// Prepare and train root classifier ensemble, where all recordings are classified into one of the root categories.
		if (preferences_panel.shouldClassifyIntoRootCategories())
		{
			ensemble_indice++;
			ensemble_label = "Root Classifier";
			elligible_categories = root_labels;
			model_training_classifications = getModelClassifications(pot_training_recordings, elligible_categories);
			elligible_recordings = getRecordingsBelongingToElligibleCategories(model_training_classifications, elligible_categories);
			model_training_classifications = getReducedModelClassifications(model_training_classifications, elligible_recordings);
			specific_training_recordings = getReducedRecordings(pot_training_recordings, elligible_recordings);
			model_testing_classifications[ensemble_indice] = getModelClassifications(testing_recordings, elligible_categories);
			trainAnEnsembleAndReportErrors( ensemble_indice, ensemble_label, elligible_categories,
											specific_training_recordings, model_training_classifications,
			                                report_training_results, training_monitor );
		}

		// Prepare and train the direct descendant ensembles, each of which classifies among the direct
		// descendants of a node with children. There is an ensemble for every node
		// with children.
		if (preferences_panel.shouldClassifyIntoDirectDescendantsOfParentCategories())
		{
			// Find the elligible categories for each ensemble (descendant_names)
			String[][] parents_and_their_children = parent_window.taxonomy_panel.getAllParentsAndTheirDirectChildren();
			String[] parent_names = new String[parents_and_their_children.length];
			String[][] descendant_names = new String[parents_and_their_children.length][];
			for (int node = 0; node < parents_and_their_children.length; node++)
			{
				parent_names[node] = parents_and_their_children[node][0];
				descendant_names[node] = new String[parents_and_their_children[node].length - 1];
				for (int i = 0; i < descendant_names[node].length; i++)
					descendant_names[node][i] = parents_and_their_children[node][i + 1];
			}

			// Train ensembles one by one
			for (int ensemble = 0; ensemble < parent_names.length; ensemble++)
			{
				ensemble_indice++;
				ensemble_label = "DD " + parent_names[ensemble] + " Classifier";
				elligible_categories = descendant_names[ensemble];
				model_training_classifications = getModelClassifications(pot_training_recordings, elligible_categories);
				elligible_recordings = getRecordingsBelongingToElligibleCategories(model_training_classifications, elligible_categories);
				model_training_classifications = getReducedModelClassifications(model_training_classifications, elligible_recordings);
				specific_training_recordings = getReducedRecordings(pot_training_recordings, elligible_recordings);
				model_testing_classifications[ensemble_indice] = getModelClassifications(testing_recordings, elligible_categories);
				trainAnEnsembleAndReportErrors( ensemble_indice, ensemble_label, elligible_categories,
												specific_training_recordings, model_training_classifications,
				                                report_training_results, training_monitor );
			}
		}

		// Prepare the train the round robin ensembles of classifiers, each of which classifies between a pair
		// of leaf categories. Every pair is represented by one ensemble.
		if (preferences_panel.shouldClassifyUsingRoundRobin())
		{
			// Find the names of all of the pairs of leaf categories
			int number_round_robins = 0;
			for (int i = 0; i < (leaf_labels.length - 1); i++)
				for (int j = i + 1; j < leaf_labels.length; j++)
					number_round_robins++;
			int current_indice = 0;
			String[][] round_robin_pairs = new String[number_round_robins][2];
			for (int i = 0; i < (leaf_labels.length - 1); i++)
				for (int j = i + 1; j < leaf_labels.length; j++)
				{
					round_robin_pairs[current_indice][0] = leaf_labels[i];
					round_robin_pairs[current_indice][1] = leaf_labels[j];
					current_indice++;
				}
		
			// Train ensembles one by one
			for (int ensemble = 0; ensemble < round_robin_pairs.length; ensemble++)
			{
				ensemble_indice++;
				ensemble_label = "RR " + round_robin_pairs[ensemble][0] + " AND " + round_robin_pairs[ensemble][1] + " Classifier";
				elligible_categories = round_robin_pairs[ensemble];
				model_training_classifications = getModelClassifications(pot_training_recordings, elligible_categories);
				elligible_recordings = getRecordingsBelongingToElligibleCategories(model_training_classifications, elligible_categories);
				model_training_classifications = getReducedModelClassifications(model_training_classifications, elligible_recordings);
				specific_training_recordings = getReducedRecordings(pot_training_recordings, elligible_recordings);
				model_testing_classifications[ensemble_indice] = getModelClassifications(testing_recordings, elligible_categories);
				trainAnEnsembleAndReportErrors( ensemble_indice, ensemble_label, elligible_categories,
												specific_training_recordings, model_training_classifications,
				                                report_training_results, training_monitor );
			}
		}

		// Stop monitoring training, if appropriate
		if (training_monitor != null)
			training_monitor.markTaskComplete();

		// Set testing_recordings and model_testing_classifications to null if appropriate
		if (testing_recordings == null)
			model_testing_classifications = null;
		else if (testing_recordings.length == 0)
		{
			testing_recordings = null;
			model_testing_classifications = null;
		}

		// Note that training has completed
		results_display.append("TRAINING COMPLETE\n");

		// Update the status bar
		updateStatusBar();

		// Gather training times and display them and make them available
		if (training_monitor != null)
			if (training_monitor.isTrainingTimesAvailable())
				displayAndStoreTrainingTimes(training_monitor, preferences_panel.shouldDisplayTrainingTimesInRawSeconds());
	}


	/**
	 * Classifies the specified recordings and displays the classification results.
	 * Any included model categories are ignored (i.e. no success rates are given).
	 * The <i>testing_recordings</i> field is set to the given <i>recordings_to_classify</i>
	 * parameter and the <i>model_testing_classifications</i> field is set to null.
	 */
	public void classifyRecordings(Recording[] recordings_to_classify)
	{
		testing_recordings = recordings_to_classify;
		model_testing_classifications = null;
		classifyReservedRecordings();
	}


	/**
	 * Train on all currently loaded recordings and save the results
	 * in the locations specified in the ConfigureFileLocationsJFrame.
	 * This method is designed to be used with the command line
	 * interface of Bodhidharma.
	 */
	public void trainFromCommandLine()
	{
		// Set to train on all available recordings
		training_combo_box.setSelectedIndex(2);

		// Perform training
		train_button.doClick();
		while (!training_monitor.isDone() && !training_monitor.isCancelled())
		{
		}

		// Save the trained classifiers
		if (training_monitor.isDone())
			save_classifiers_button.doClick();
	}


	/**
	 * Classify all currently loaded recordings and save the results
	 * to the specified file. This method is designed to be used with the command line
	 * interface of Bodhidharma.
	 *
	 * @param	output_classifications_path	The path to save the output
	 *										classifications.
	 */
	public void classifyFromCommandLine(String output_classifications_path)
		throws Exception
	{
		// Prepare to save results
		File save_file = new File(output_classifications_path);
		DataOutputStream writer = new DataOutputStream(new FileOutputStream(save_file));
		
		// Set to train on all available recordings
		classifying_combo_box.setSelectedIndex(3);

		// Perform classification
		classify_button.doClick();

		// The source of prefence values on what to report
		PreferencesPanel preferences_panel = parent_window.preferences_panel;

		// Used to interperet results
		ClassificationResultsInterpereter results_interpereter = new ClassificationResultsInterpereter(preferences_panel);
			
		// Find the model classifications
		String[][] temp_model_classifications;
		if (preferences_panel.shouldDisplayModelResultsAlongWithResults())
			temp_model_classifications = model_testing_classifications_for_combined_classifier;
		else
			temp_model_classifications = null;

		// Find the winning classification results
		String[] combined_labels_of_results = results_interpereter.getWinnerLabels( combined_classifier_results_all_ensembles[combined_classifier_results_all_ensembles.length - 1],
																					candidate_categories_for_combined_classification,
																					temp_model_classifications, 
																					preferences_panel.reportScoresWithWinningCategories(),
																					preferences_panel.reportSecondaryWinningCategories() );

		// Print the results to standard out
		for (int i = 0; i < combined_labels_of_results.length; i++)
			writer.writeBytes(testing_recordings[i].file_path + "\t" + combined_labels_of_results[i] + "\n");

		// Close the writer
		writer.close();
	}


	/* PRIVATE METHODS *********************************************************/


	// Update the status bar to indicate if classification ensembles are available
	private void updateStatusBar()
	{
		if (classifier_ensemble == null)
			status_bar.setText("No classification ensembles currently loaded.");
		else
			status_bar.setText("Classification ensembles currently loaded.");
	}


	// Parses the classifiers_file referred to in the given parameter and any related
	// files it refers to. Displays an error message if a problem occurs during parsing.
	// if it does, then restores original state.
	// If parsing is succesful, stores the loaded data in the classifier_ensemble, 
	// testing_recordings and model_testing_classifications fields and deletes the
	// contents of other fields. Allows access to the appropriate ensembles by 
	// population the selection_weighting_selector combo box.
	private void parseClassifiersFile(String classifiers_file_path)
	{
		// Variables for temporarily holding data to be stored in fields
		NN_BioKNN_Ensemble[] classifier_ensemble_temp;
		Recording[] testing_recordings_temp;
		String[][][] model_testing_classifications_temp;
		String stored_training_times_report_temp = null;

		// Perform parsing and perform appropriate updates
		try
		{
			// Parse the contents of the classifiers_file
			String[] classifiers_file_results = (String[]) XMLDocumentParser.parseXMLDocument(classifiers_file_path, "classifiers_file");
			String classifier_sub_folder_path = classifiers_file_results[0];
			String testing_recordings_file_path = classifiers_file_results[1];
			String model_testing_classifications_file_path = classifiers_file_results[2];
			String stored_training_times_report_file_path = classifiers_file_results[3];
			String[] NN_BioKNN_Ensemble_file_paths = new String[classifiers_file_results.length - 4];
			for (int i = 4; i < classifiers_file_results.length; i++)
				NN_BioKNN_Ensemble_file_paths[i - 4] = classifiers_file_results[i];

			// Parse the contents of the testing recordings file, the testing recordings feature values file
			// and the model_testing_classifications_file if they are present
			if (testing_recordings_file_path.equals("null") || model_testing_classifications_file_path.equals(null))
			{
				testing_recordings_temp = null;
				model_testing_classifications_temp = null;
			}
			else
			{
				// Parse the contents of the testing_recordings_file
				Object[] testing_recordings_file_results = (Object[]) XMLDocumentParser.parseXMLDocument(testing_recordings_file_path, "recordings_file");
				testing_recordings_temp = new Recording[testing_recordings_file_results.length];
				for (int i = 0; i < testing_recordings_temp.length; i++)
					testing_recordings_temp[i] = (Recording) testing_recordings_file_results[i];
				for (int i = 0; i < testing_recordings_temp.length; i++)
				{
					if (testing_recordings_temp[i].title == null)
						testing_recordings_temp[i].title = "Unknown";
					if (testing_recordings_temp[i].artist == null)
						testing_recordings_temp[i].artist = "Unknown";
				}
				String[][] leaves_and_ancestors = parent_window.taxonomy_panel.getLeafsAndTheirParents();
				for (int i = 0; i < testing_recordings_temp.length; i++)
				{
					if (!testing_recordings_temp[i].leaf_categories[0].equals("Unknown"))
					{
						for(int j = 0; j < testing_recordings_temp[i].leaf_categories.length; j++)
						{
							int k = 0;
							while (k <= leaves_and_ancestors.length)
							{
								if (k == leaves_and_ancestors.length)
									throw new Exception("The file " + testing_recordings_file_path + " contains the leaf " +
														"category " + testing_recordings_temp[i].leaf_categories[j] +
														". This category is not a part of the currently " +
														"loaded taxonomy.\n\nPlease correct the recordings" +
														" file or adjust your taxonomy. An empty list of " +
														" recordings will now be generated.");
								else if (testing_recordings_temp[i].leaf_categories[j].equals(leaves_and_ancestors[k][0]))
									k = leaves_and_ancestors.length + 1;
								k++;
							}
						}
					}
				}
				for (int i = 0; i < testing_recordings_temp.length; i++)
				{
					if (testing_recordings_temp[i].leaf_categories[0].equals("Unknown"))
					{
						String[] unknown_id = new String[1];
						unknown_id[0] = "Unknown";
						testing_recordings_temp[i].ancestor_categories = unknown_id;
					}
					else
					{
						int j = 0;
						while(j < testing_recordings_temp[i].leaf_categories.length)
						{
							int k = 0;
							while(k < leaves_and_ancestors.length)
							{
								if (testing_recordings_temp[i].leaf_categories[j].equals(leaves_and_ancestors[k][0]))
								{
									String[] ancestors = new String[leaves_and_ancestors[k].length - 1];
									for (int m = 0; m < ancestors.length; m++)
										ancestors[m] = leaves_and_ancestors[k][m + 1];
									if (testing_recordings_temp[i].ancestor_categories != null)
									{
										String[] existing_ancestors = testing_recordings_temp[i].ancestor_categories;
										int repetition_count = 0;
										for (int a = 0; a < ancestors.length; a++)
										{
											for (int b = 0; b < existing_ancestors.length; b++)
											{
												if (ancestors[a] != null)
												{
												if (ancestors[a].equals(existing_ancestors[b]))
													{
													repetition_count++;
													ancestors[a] = null;
													}
												}
											}
										}
										if (repetition_count != 0)
										{
											String[] temp = new String[ancestors.length];
											for (int a = 0; a < ancestors.length; a++)
												temp[a] = ancestors[a];
											ancestors = new String[ancestors.length - repetition_count];
											int ancestors_index = 0;
											for (int a = 0; a < temp.length; a++)
											{
												if (temp[a] != null)
												{
													ancestors[ancestors_index] = temp[a];
													ancestors_index++;
												}
											}
										}
										String[] combined_ancestors = new String[existing_ancestors.length + ancestors.length];
										for (int m = 0; m < existing_ancestors.length; m++)
											combined_ancestors[m] = existing_ancestors[m];
										for (int m = existing_ancestors.length; m < existing_ancestors.length + ancestors.length; m++)
											combined_ancestors[m] = ancestors[m - existing_ancestors.length];
										testing_recordings_temp[i].ancestor_categories = combined_ancestors;
									}
									else
										testing_recordings_temp[i].ancestor_categories = ancestors;
								}
								k++;
							}
							j++;
						}
					}
				}

				// Parse the contents of the testing recordings feature values file
				String values_file_path = testing_recordings_file_path.substring(0, testing_recordings_file_path.length() - 4) +  "_feature_values.xml";
				Object[] temp = (Object[]) XMLDocumentParser.parseXMLDocument(values_file_path, "feature_values_file");
				Recording[] feature_value_holders = new Recording[temp.length];
				for (int i = 0; i < feature_value_holders.length; i++)
					feature_value_holders[i] = (Recording) temp[i];
				for (int i = 0; i < testing_recordings_temp.length; i++)
				{
					for (int j = 0; j < feature_value_holders.length; j++)
					{
						if ((testing_recordings_temp[i].file_path).equals(feature_value_holders[j].file_path))
						{
							testing_recordings_temp[i].feature_set = new FeatureSet(feature_value_holders[j].feature_set.features);
							j = feature_value_holders.length + 1; // exit the loop
						}
					}
				}		

				// Parse the contents of the model_testing_classifications_file
				Object[] model_testing_classifications_file_results = (Object[]) XMLDocumentParser.parseXMLDocument(model_testing_classifications_file_path, "model_classifications_file");
				model_testing_classifications_temp = new String[model_testing_classifications_file_results.length][][];
				for (int i = 0; i < model_testing_classifications_file_results.length; i++)
					model_testing_classifications_temp[i] = (String[][]) model_testing_classifications_file_results[i];

				// Parse the contents of the stored_training_times_report_file_path
				try
				{
					File src = new File(stored_training_times_report_file_path);
					char[] chars = new char[(int) src.length()];
					src = null;
					InputStream in = new FileInputStream(stored_training_times_report_file_path);
					int ch;
					for (int i = 0; (ch = in.read()) != -1; i++)
						chars[i] = (char)ch;
					in.close();
					stored_training_times_report_temp = new String(chars);
				}
				catch (Exception e){}
			}

			// Parse the contents of the NN_BioKNN_Ensemble_files
			if (!NN_BioKNN_Ensemble_file_paths[0].equals("null"))
			{
				classifier_ensemble_temp = new NN_BioKNN_Ensemble[NN_BioKNN_Ensemble_file_paths.length];
				for (int ens = 0; ens < NN_BioKNN_Ensemble_file_paths.length; ens++)
					classifier_ensemble_temp[ens] = new NN_BioKNN_Ensemble( NN_BioKNN_Ensemble_file_paths[ens],
					                                                        parent_window.preferences_panel.ga_settings,
					                                                        parent_window.preferences_panel.network_settings );
			}
			else classifier_ensemble_temp = null;

			// Clear the unused fields and GUI elemnents
			clearStoredData();

			// Set the fields to their loaded values
			classifier_ensemble = classifier_ensemble_temp;
			testing_recordings = testing_recordings_temp;
			model_testing_classifications = model_testing_classifications_temp;
			stored_training_times_report = stored_training_times_report_temp;
			if (stored_training_times_report != null)
				if (stored_training_times_report.length() != 0)
					training_errors_ensemble_selector.addItem("All Times");
			
			// Add the appropriate labels to the selection_weighting_selector
			if (classifier_ensemble != null)
				for (int i = 0; i < classifier_ensemble.length; i++)
					selection_weighting_selector.addItem(classifier_ensemble[i].getClassifierName());

			// Update the status bar and display message showing loading was succesful
			results_display.setText("Classifiers loaded succesfully.");
			updateStatusBar();
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage() + "\n\nAny pre-existing training information has not been changed.",
			                              "ERROR", JOptionPane.ERROR_MESSAGE);
			//e.printStackTrace();
		}
	}


	// Allows the user to choose a new directory to be created to store reports.
	// Returns the created directory created as a File. Returns null if cancelled
	// of if there is a problem.
	private File getReportSaveLocation(boolean reports_must_already_be_available)
		throws Exception
	{
		// Make sure that there is information available to save
		if ( reports_must_already_be_available &&
			 training_errors_ensemble_selector.getItemCount() == 0 &&
			 classification_results_selector.getItemCount() == 0 &&
			 selection_weighting_selector.getItemCount() == 0 )
			throw new Exception("No results to save.");
		
		// Choose what directory to start in
		JFileChooser save_dialog;
		if (!parent_window.file_locations.classifiers_path.equals(""))
			save_dialog = new JFileChooser();
		else
			save_dialog = new JFileChooser(new File(parent_window.file_locations.current_directory));

		// Only work with directories
		save_dialog.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);

		// Display dialog box, make directory and return path
		int save_return = save_dialog.showSaveDialog(this);
		if (save_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
		{
			File place_to_save = save_dialog.getSelectedFile();
			if (place_to_save.exists())
				throw new Exception("Directory already exists.");
			place_to_save.mkdir();
			return place_to_save;
		}
		else
			return null;
	}


	/**
	 * Train as described in the performHierarchicalTraining method.
	 * Displays a dialog box if any problems are encountered.
	 */
	private void train(boolean reserve_recordings_for_testing, boolean overide_selected_trainers)
		throws Exception
	{
		Recording[] all_recordings = parent_window.recordings_panel.recordings_list;
		
		// Give error if no recordings to classify
		if (all_recordings == null)
			throw new Exception("No recordings available to train with.");
		if (all_recordings.length == 0)
			throw new Exception("No recordings available to train with.");

		// Give error if features not extracted yet
		for (int rec = 0; rec < all_recordings.length; rec++)
			if (all_recordings[rec].feature_set == null)
				throw new Exception( "Cannot train because features have not been\n" +
									 "extracted from " + all_recordings[rec].title + " yet." );

		// Recordings that will be elligible for training
		Recording[] all_eligible_recordings;
		
		// Clear all stored data from previous training or classifications
		clearStoredData();

		// States which recordings are for training
		boolean[] training_key;

		// Perform training
		try
		{
			// Clear the display and note that training
			results_display.setText("TRAINING ENSEMBLE...\n");

			// Case where all loaded recordings are candidates for training and a subset of
			// these are reserved for testing
			if (overide_selected_trainers && reserve_recordings_for_testing)
			{
				all_eligible_recordings = parent_window.recordings_panel.recordings_list;
				training_key = null;
				performHierarchicalTrainingThroughProgressBar(all_eligible_recordings, training_key, false, null, 0);
			}

			// Case where all loaded recordings are used for training and none are
			// are reserved for testing
			else if (overide_selected_trainers && !reserve_recordings_for_testing)
			{
				all_eligible_recordings = parent_window.recordings_panel.recordings_list;
				training_key = new boolean[all_eligible_recordings.length];
				for (int rec = 0; rec < training_key.length; rec++)
					training_key[rec] = true;
				performHierarchicalTrainingThroughProgressBar(all_eligible_recordings, training_key, false, null, 0);
			}

			// Case where only recordings marked as "training" in the Recordings Panel are 
			// used for training and none are are reserved for testing
			else if (!overide_selected_trainers && !reserve_recordings_for_testing)
			{
				all_eligible_recordings = getRecordingsMarkedForPurpose("training");
				training_key = new boolean[all_eligible_recordings.length];
				for (int rec = 0; rec < training_key.length; rec++)
					training_key[rec] = true;
				performHierarchicalTrainingThroughProgressBar(all_eligible_recordings, training_key, false, null, 0);
			}

			// Case where only recordings marked as "training" in the Recordings Panel are 
			// used for training and the ones marked "testing" are reserved for testing
			else
				throw new Exception("Cannot reserve recordings for testing if limited to\n" +
				                    "recordings marked for \"training.\"");
		}

		// Catch any problems and report them in an error dialog
		catch (Throwable t)
		{
			// React to the Java Runtime running out of memory
			if (t.toString().equals("java.lang.OutOfMemoryError"))
				JOptionPane.showMessageDialog(null, "The Java Runtime ran out of memory. Please rerun this program with a higher amount of memory assigned to the Java Runtime heap.", "ERROR", JOptionPane.ERROR_MESSAGE);
			else
				JOptionPane.showMessageDialog(null, t.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);

			// Note that an error occured
			results_display.append("COULD NOT COMPLETE TRAINING DUE TO AN ERROR.");

			// Clear all stored classifier ensembles and information about them
			clearStoredData();

			// t.printStackTrace();
		}

		// Update the status bar
		updateStatusBar();
	}


	// Creates a LargeProgressBar for the performHierarchicalTraining method and initiates
	// the performHierarchicalTraining method through the go method of the LargeProgressBar.
	//
	// The last two parameters are only needed if multi-fold cross-validation is being 
	// performed. If only one fold is being processed in total, then multi_fold_training_keys
	// should be null and multi_fold_current_fold is irrelevant.
	private void performHierarchicalTrainingThroughProgressBar( Recording[] all_eligible_recordings,
	                                                            boolean[] training_key,
	                                                            boolean report_training_results,
	                                                            boolean[][] multi_fold_training_keys,
	                                                            int multi_fold_current_fold )
		throws Exception
	{
		// Find user preferences
		PreferencesPanel preferences_panel = parent_window.preferences_panel;

		// Find information about the taxonomy
		String[] leaf_labels = parent_window.taxonomy_panel.getLeafLabels();
		String[] root_labels = parent_window.taxonomy_panel.getParentLabels();
		String[] intermediate_labels = parent_window.taxonomy_panel.getMiddleLabels();

		// Calculate the number of classification ensembles
		int number_ensembles = 0;
		if (preferences_panel.shouldClassifyIntoAllLeafCategories())
			number_ensembles++;
		if (preferences_panel.shouldClassifyIntoRootCategories())
			number_ensembles++;
		if (preferences_panel.shouldClassifyIntoDirectDescendantsOfParentCategories())
			number_ensembles += (intermediate_labels.length + root_labels.length);
		if (preferences_panel.shouldClassifyUsingRoundRobin())
			for (int i = 0; i < (leaf_labels.length - 1); i++)
				for (int j = i + 1; j < leaf_labels.length; j++)
					number_ensembles++;

		// Prepare training progress bar
		training_monitor = new ProgressBarTaskTrainMonitor( all_eligible_recordings,
	                                                        training_key,
	                                                        report_training_results,
	                                                        this,
	                                                        multi_fold_training_keys,
	                                                        multi_fold_current_fold );
		LargeProgressBarDialog training_indicator = new LargeProgressBarDialog( parent_window,
		                                                                        training_monitor,
		                                                                        preferences_panel.getTimeBetweenProgressBarPolls() );
		training_monitor.setProgressDialog(training_indicator);
		
		// Initiate performHierarchicalTraining method through the go method of the LargeProgressBar
		training_monitor.startTopTask(number_ensembles, new String("Training a total of " + number_ensembles + " ensembles."));
		training_monitor.go();
	}
	

	// Reports the training times for all ensembles and their classifiers and 
	// stores the results in the stored_training_times_report field
	private void displayAndStoreTrainingTimes(ProgressBarTaskTrainMonitor training_monitor, boolean raw_seconds)
	{
		// Set the font and clear the screen
		results_display.setFont( new Font( "Monospaced", Font.PLAIN, 12 ) );
		results_display.setText("");

		// Set up main heading
		results_display.append("<<<<<< TRAINING TIMES ");
		if (raw_seconds)
			results_display.append("(in seconds) ");
		results_display.append(">>>>>>\n\n\n\n");

		// Gather training times
		String[][] ensemble_training_info = training_monitor.getMiddleTaskDurations(raw_seconds);
		String[][][] classifier_training_info = training_monitor.getBottomTaskDurations(raw_seconds);
		String overall_duration = training_monitor.getOverallTaskTime(raw_seconds);

		// Remove "Training ", " Classifier." and "." from training names
		for (int ens = 0; ens < classifier_training_info.length; ens++)
		{
			ensemble_training_info[ens][0] = ensemble_training_info[ens][0].substring(9, ensemble_training_info[ens][0].length() - 12);
			for (int clas = 0; clas < classifier_training_info[ens].length; clas++)
				classifier_training_info[ens][clas][0] = classifier_training_info[ens][clas][0].substring(9, classifier_training_info[ens][clas][0].length() - 1);
		}

		// Give overall training time
		appendTableTitle("OVERALL TRAININING TIME: " + overall_duration);

		// Give training time for each ensemble
		appendTableTitle("ENSEMBLE TRAINING TIMES:");
		for (int ens = 0; ens < ensemble_training_info.length; ens++)
			results_display.append(ensemble_training_info[ens][0] + "    " + ensemble_training_info[ens][1] + "\n");
		results_display.append("\n\n\n");
		
		// Give training time for each classifier
		for (int ens = 0; ens < classifier_training_info.length; ens++)
		{
			appendTableTitle("CLASSIFIER TRAINING TIMES FOR <" + ensemble_training_info[ens][0] + "> ENSEMBLE:");
			for (int clas = 0; clas < classifier_training_info[ens].length; clas++)
				results_display.append(classifier_training_info[ens][clas][0] + "    " + classifier_training_info[ens][clas][1] + "\n");
			results_display.append("\n\n\n");
		}

		// Update the training_errors_ensemble_selector and store the training times
		training_errors_ensemble_selector.addItem("All Times");
		stored_training_times_report = results_display.getText();
	}


	/**
	 * Trains the specified NN_BioKNN_Ensemble in the classifier_ensemble field.
	 * Reports training errors if this has been selected in preferences. The errors
	 * will be stored in the combined_training_errors field, in any event. The feature
	 * and classifier weighting and selection will also be stored in selection_weighting_selector.
	 * Also records the names and categories of the training_recordings in the training_recordings_info
	 * field.
	 *
	 * @param	ensemble_indice					The indice of the classifier_ensemble field to be trained.
	 * @param	ensemble_label					The label of the classifier_ensemble field to be trained.
	 * @param	elligible_categories			The elligible classification categories for the ensemble.
	 * @param	training_recordings				The recordings to be used for training for the ensemble.
	 * @param	model_training_classifications	The names of the model categories for the training_recordings. First indice indicates recording. There is only an entry for categories that the recording belongs to.
	 * @param	report_training_results			Whether or not to print training results immediately.
	 * @param	training_monitor				ProgressBarTaskTrainMonitor that is monitoring progress of training. Can be null if no ProgressBarTaskTrainMonitor is used.
	 */
	private void trainAnEnsembleAndReportErrors( int ensemble_indice,
	                                             String ensemble_label,
	                                             String[] elligible_categories,
	                                             Recording[] training_recordings,
	                                             String[][] model_training_classifications,
	                                             boolean report_training_results,
	                                             ProgressBarTaskTrainMonitor training_monitor )
		throws Exception
	{
		// Verify that there are sufficient training recordings and throw
		// and exception if there are not
		verifySufficientTrainingExamples(elligible_categories, model_training_classifications );

		// Find user preferences
		PreferencesPanel preferences_panel = parent_window.preferences_panel;

		// Instantiate the ensemble
		classifier_ensemble[ensemble_indice] = 
			new NN_BioKNN_Ensemble( ensemble_label,
		                            elligible_categories,
		                            parent_window.features_panel.getFeatureSettings(),
		                            preferences_panel.shouldUseFeaturesMarkedAsSelectionCandidates(),
		                            preferences_panel.shouldSelectClassifiersInEachEnsemble(),
		                            preferences_panel.shouldWeightClassifiersInEachEnsemble(),
		                            preferences_panel.getInternalTrainingFraction(),
		                            preferences_panel.shouldUseOneDimensionalFeaturesInTraining(),
		                            preferences_panel.shouldUseMultiDimensionalFeaturesInTraining(),
		                            preferences_panel.ga_settings,
		                            preferences_panel.shouldBioKNNsUseFeatureSelection(),
		                            preferences_panel.shouldBioKNNsUseFeatureWeighting(),
		                            preferences_panel.getNumberGAIterations(),
			                        preferences_panel.getClassifierVsFeatureGAMultiplier(),
		                            preferences_panel.getMaxGATrainingErrorForEarlyEnding(),
		                            preferences_panel.getMinGATrainingIterationsForEarlyEnding(),
		                            preferences_panel.network_settings,
		                            preferences_panel.getNumberNNIterations(),
		                            preferences_panel.getMaxNNTrainingErrorForEarlyEnding(),
		                            preferences_panel.getMinNNTrainingIterationsForEarlyEnding() );

		// Find number of trainings needing to be done (for use with progress bar)
		int number_trainings = 0;
		if (preferences_panel.shouldUseOneDimensionalFeaturesInTraining())
		{
			if (preferences_panel.shouldBioKNNsUseFeatureSelection())
				number_trainings++;
			if (preferences_panel.shouldBioKNNsUseFeatureWeighting())
				number_trainings++;
		}
		if (preferences_panel.shouldUseMultiDimensionalFeaturesInTraining())
			number_trainings += classifier_ensemble[ensemble_indice].getNumberMultiDimFeatures();
		if (preferences_panel.shouldSelectClassifiersInEachEnsemble() && preferences_panel.shouldUseMultiDimensionalFeaturesInTraining())
			number_trainings++;
		if (preferences_panel.shouldWeightClassifiersInEachEnsemble() && preferences_panel.shouldUseMultiDimensionalFeaturesInTraining())
			number_trainings++;
		
		// Update the progress bar
		if (training_monitor != null)
			training_monitor.startMiddleTask(number_trainings, new String("Training " + ensemble_label + "."));
		
		// Train the ensemble
		double[][] training_errors = classifier_ensemble[ensemble_indice].train( training_recordings,
		                                                                         model_training_classifications,
		                                                                         training_monitor,
		                                                                         new ClassificationResultsInterpereter(preferences_panel) );

		// Update the training_errors_ensemble_selector and selection_weighting_selector
		// so that the errors can be accessed in the future
		training_errors_ensemble_selector.addItem(ensemble_label);
		selection_weighting_selector.addItem(ensemble_label);

		// Store the training errors
		combined_training_errors[ensemble_indice] = training_errors;

		// Record the names and categories of recordings used to train this ensemble
		training_recordings_info[ensemble_indice] = new String[training_recordings.length][];
		for (int rec = 0; rec < training_recordings.length; rec++)
		{
			training_recordings_info[ensemble_indice][rec] = new String[1 + training_recordings[rec].leaf_categories.length];
			training_recordings_info[ensemble_indice][rec][0] = training_recordings[rec].title;
			for (int i = 0; i < training_recordings[rec].leaf_categories.length; i++)
				training_recordings_info[ensemble_indice][rec][i + 1] = training_recordings[rec].leaf_categories[i];
		}

		// Report training results if requested
		if (report_training_results)
		{
			results_display.append("\n\n>>>>>>>>>> TRAINING RESULTS FOR " + ensemble_label + " ENSEMBLE <<<<<<<<<< \n\n");
			displayTrainingStatisticsForAnEnsemble( training_errors, training_recordings_info[ensemble_indice], classifier_ensemble[ensemble_indice] );
		}
	}


	/**
	 * Classifies all the recordings currently loaded in the <code>RecordingsPanel</code>.
	 * 
	 * <p>Any included model categories are ignored (i.e. no success rates are given).
	 * The <i>testing_recordings</i> field is set to the given <i>recordings_to_classify</i>
	 * parameter and the <i>model_testing_classifications</i> field is set to null.
	 */
	private void classifyAllRecordings()
	{
		try
		{
			// Find the recordings to classify
			Recording[] recordings_to_classify = parent_window.recordings_panel.recordings_list;
			
			// Give an error if the taxonomy is empty
			if (parent_window.taxonomy_panel.isTreeEmpty())
				throw new Exception("The currently loaded taxonomy is empty, so recordings may not be classified.");

			// Give an error if no recordings are marked for classification
			if (recordings_to_classify == null)
				throw new Exception("No recordings available for classification.");

			// Give error if features not extracted yet
			for (int rec = 0; rec < recordings_to_classify.length; rec++)
				if (recordings_to_classify[rec].feature_set == null)
					throw new Exception( "Cannot classify recordings because features have not\n" +
					                     "been extracted from " + recordings_to_classify[rec].title + " yet." );

			// Classify the recordings
			classifyRecordings(recordings_to_classify);
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}


	/**
	 * Classifies the recordings marked with a <i>role</i> of <i>classification</i> in
	 * the <code>RecordingsPanel</code>.
	 * 
	 * <p>Any included model categories are ignored (i.e. no success rates are given).
	 * The <i>testing_recordings</i> field is set to the given <i>recordings_to_classify</i>
	 * parameter and the <i>model_testing_classifications</i> field is set to null.
	 */
	private void classifyRecordingsWithClassificationRole()
	{
		try
		{
			// Find the recordings to classify
			Recording[] recordings_to_classify = parent_window.recordings_panel.getRecordingsMarkedForClassification();
			
			// Give an error if the taxonomy is empty
			if (parent_window.taxonomy_panel.isTreeEmpty())
				throw new Exception("The currently loaded taxonomy is empty, so recordings may not be classified.");

			// Give an error if no recordings are marked for classification
			if (recordings_to_classify == null)
				throw new Exception("No recordings marked for classification.");

			// Give error if features not extracted yet
			for (int rec = 0; rec < recordings_to_classify.length; rec++)
				if (recordings_to_classify[rec].feature_set == null)
					throw new Exception( "Cannot classify recordings because features have not\n" +
					                     "been extracted from " + recordings_to_classify[rec].title + " yet." );

			// Classify the recordings
			classifyRecordings(recordings_to_classify);
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}


	/**
	 * Classifies the recordings marked with a <i>role</i> of <i>testing</i> in
	 * the <code>RecordingsPanel</code>.
	 * 
	 * <p>Any included model categories are ignored (i.e. no success rates are given).
	 * The <i>testing_recordings</i> field is set to the given <i>recordings_to_classify</i>
	 * parameter and the <i>model_testing_classifications</i> field is set to null.
	 */
	private void classifyRecordingsWithTestingRole()
	{
		try
		{
			// Find the recordings to classify
			Recording[] recordings_to_classify = parent_window.recordings_panel.getRecordingsMarkedForTesting();

			// Give an error if the taxonomy is empty
			if (parent_window.taxonomy_panel.isTreeEmpty())
				throw new Exception("The currently loaded taxonomy is empty, so recordings may not be classified.");

			// Give an error if no recordings are marked for classification
			if (recordings_to_classify == null)
				throw new Exception("No recordings marked for testing.");

			// Give error if features not extracted yet
			for (int rec = 0; rec < recordings_to_classify.length; rec++)
				if (recordings_to_classify[rec].feature_set == null)
					throw new Exception( "Cannot classify recordings because features have not\n" +
					                     "been extracted from " + recordings_to_classify[rec].title + " yet." );

			// Classify the recordings
			classifyRecordings(recordings_to_classify);
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}


	// Classify the recordings stored in testing_recordings and store the
	// results in the individual_classifier_results_all_ensembles
	// and combined_classifier_results_all_ensembles fields.
	// Update the classification_results_selector combo box to make
	// the classification reports available.
	private void classifyReservedRecordings()
	{
		try
		{
			// Clear the display and note that training
			results_display.setText("CLASSIFYING ENSEMBLE...\n");

			// Clear previous classification information
			individual_classifier_results_all_ensembles = null;
			combined_classifier_results_all_ensembles = null;
			candidate_categories_for_combined_classification = null;
			model_testing_classifications_for_combined_classifier = null;
			classification_results_selector.removeAllItems();
			
			// Check that classification can be performed
			if (classifier_ensemble == null)
				throw new Exception("Cannot perform classification. No trained classification ensembles available.");
			if (testing_recordings == null)
				throw new Exception("Cannot classify reserved recordings. No reserved recordings available.");
			
			// Instantiate individual_classifier_results_all_ensembles and combined_classifier_results_all_ensembles
			individual_classifier_results_all_ensembles = new double[classifier_ensemble.length][][][];
			combined_classifier_results_all_ensembles = new double[classifier_ensemble.length + 1][][];

			// Perform the classification for all ensembles individually
			for (int ens = 0; ens < classifier_ensemble.length; ens++)
				classifyWithAnEnsemble( ens, testing_recordings );

			// Perform the combined classification
			performCombinedClassification(testing_recordings, true);

			// Clear the display and note that training
			results_display.append("CLASSIFICATION COMPLETE.\n");
		}
		catch (Exception e)
		{
			// Note that an error occured
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			results_display.append("COULD NOT COMPLETE CLASSIFICATION.");

			// Clear all stored classifier ensembles and information about them
			testing_recordings = null;
			model_testing_classifications = null;
			individual_classifier_results_all_ensembles = null;
			combined_classifier_results_all_ensembles = null;
			candidate_categories_for_combined_classification = null;
			model_testing_classifications_for_combined_classifier = null;
			classification_results_selector.removeAllItems();
			updateStatusBar();

			// e.printStackTrace();
		}
	}


	// Classifies the given recordings with the ensemble of classifiers at
	// the given indice of the classifier_ensemble field and stores the
	// results in the individual_classifier_results_all_ensembles and
	// combined_classifier_results_all_ensembles fields.
	// Updates the classification_results_selector field to include the
	// name of this classifier so that the results can be displayed.
	private void classifyWithAnEnsemble( int ensemble_indice, Recording[] recordings_to_test_with )
		throws Exception
	{
		// Perform the classification and record the results in the appropriate fields
		individual_classifier_results_all_ensembles[ensemble_indice] = new double[(classifier_ensemble[ensemble_indice].getEnsembleSelection()).length][][];
		combined_classifier_results_all_ensembles[ensemble_indice] = classifier_ensemble[ensemble_indice].classify( recordings_to_test_with, individual_classifier_results_all_ensembles[ensemble_indice]);

		// Update the classification_results_selector so that the results can be accessed in the future
		classification_results_selector.addItem( classifier_ensemble[ensemble_indice].getClassifierName() );
	}


	// Classify the recordings using the combined ensembles, in whatever orientation
	// is selected in the preferences. Stores the results in the last entry of
	// combined_classifier_results_all_ensembles and the classification_results_selector
	// combo_box. Also displays the report of these results if print_classification_results
	// is true. model_testing_classifications_for_combined_classifier are also
	// stored, if appropriate
	//
	// IMPORTANT: It is essential that the taxonomy currently loaded in the taxonomy panel
	// corresponds to that used during training. This is not checked here.
	private void performCombinedClassification( Recording[] recordings_to_classify, boolean print_classification_results )
		throws Exception
	{
		// Access combination of ensembles preferences
		PreferencesPanel preferences = parent_window.preferences_panel;
		boolean should_classify_into_root_categories = preferences.shouldClassifyIntoRootCategories();
		boolean should_classify_hierarchically    = preferences.shouldClassifyIntoDirectDescendantsOfParentCategories();
		boolean should_classify_into_all_leaf_categories = preferences.shouldClassifyIntoAllLeafCategories();
		boolean should_classify_using_round_robin = preferences.shouldClassifyUsingRoundRobin();
		double hierarchal_weighting = preferences.getHierarchalRelativeWeighing();
		double leaf_weighting = preferences.getAllLeafRelativeWeighing();
		double round_robin_weighting = preferences.getRoundRobinRelativeWeighing();

		// Throw exception if incorrect types of classification selected
		if ( !should_classify_into_root_categories && !should_classify_hierarchically &&
			 !should_classify_into_all_leaf_categories && !should_classify_using_round_robin ) 
			throw new Exception("You must select at least one of the possible types of classification\n" +
			                    "listed in the Preferences Panel.");
		if ( should_classify_hierarchically && !should_classify_into_root_categories )
			throw new Exception("If you wish to perform a hierarchal classification, you must also enable\n" +
			                    "the all root option.");

		// Calculate absolute weightings from relative weightings
		if (!should_classify_into_root_categories && !should_classify_hierarchically)
			hierarchal_weighting = 0.0;
		if (!should_classify_into_all_leaf_categories)
			leaf_weighting = 0.0;
		if (!should_classify_using_round_robin)
			round_robin_weighting = 0.0;
		double relative_weighting_total = hierarchal_weighting + leaf_weighting + round_robin_weighting;
		hierarchal_weighting = hierarchal_weighting / relative_weighting_total;
		leaf_weighting = leaf_weighting / relative_weighting_total;
		round_robin_weighting = round_robin_weighting / relative_weighting_total;

		// The classification results
		double[][] classification_results = new double[recordings_to_classify.length][];

		// Perform hierarchal classification of any kind
		if (should_classify_hierarchically)
		{
			// The candidate leaf categories
			String[] leaf_labels = parent_window.taxonomy_panel.getLeafLabels();

			// Prepare the results
			classification_results = new double[recordings_to_classify.length][leaf_labels.length];
			for (int i = 0; i < classification_results.length; i++)
				for (int j = 0; j < classification_results[i].length; j++)
					classification_results[i][j] = 0.0;

			// Classsify recordings one by one
			for (int rec = 0; rec < recordings_to_classify.length; rec++)
			{
				performHierarchicalClassification( recordings_to_classify[rec],
				                                   leaf_labels,
				                                   classification_results[rec],
				                                   null,
				                                   "Root Classifier",
				                                   should_classify_using_round_robin,
				                                   should_classify_into_all_leaf_categories,
			                                       hierarchal_weighting,
				                                   leaf_weighting,
				                                   round_robin_weighting );
			}

			// Store the categories and model resulsts for the combined classification
			candidate_categories_for_combined_classification = leaf_labels;
			if (model_testing_classifications == null)
				model_testing_classifications_for_combined_classifier = null;
			else
			{
				model_testing_classifications_for_combined_classifier = new String[testing_recordings.length][];
				for (int rec = 0; rec < testing_recordings.length; rec++)
					model_testing_classifications_for_combined_classifier[rec] = testing_recordings[rec].leaf_categories;
			}
		}

		// Perform all non-hierarchal classifications except just root categories
		else if (should_classify_into_all_leaf_categories || should_classify_using_round_robin)
		{
			// The candidate leaf categories
			String[] leaf_labels = parent_window.taxonomy_panel.getLeafLabels();

			// Prepare the results
			classification_results = new double[recordings_to_classify.length][leaf_labels.length];
			for (int i = 0; i < classification_results.length; i++)
				for (int j = 0; j < classification_results[i].length; j++)
					classification_results[i][j] = 0.0;

			// Perform round robin classification, if appropriate
			double[][] round_robin_results = new double[recordings_to_classify.length][leaf_labels.length];
			for (int i = 0; i < round_robin_results.length; i++)
				for (int j = 0; j < round_robin_results[i].length; j++)
					round_robin_results[i][j] = 0.0;
			if (should_classify_using_round_robin)
				for (int rec = 0; rec < recordings_to_classify.length; rec++)
					round_robin_results[rec] = performRoundRobinClassification(recordings_to_classify[rec], leaf_labels);	

			// Perform all leaf category classification classification, if appropriate
			double[][] all_leaf_results = new double[recordings_to_classify.length][leaf_labels.length];
			for (int i = 0; i < all_leaf_results.length; i++)
				for (int j = 0; j < all_leaf_results[i].length; j++)
					all_leaf_results[i][j] = 0.0;
			if (should_classify_into_all_leaf_categories)
			{
				int current_classifier_indice = getIndiceOfEnsembleWithGivenName("Leaf Classifier");
				all_leaf_results = classifier_ensemble[current_classifier_indice].classify( recordings_to_classify, null );
			}

			// Perorm root category classification, if appropriate, and bring it down to leaf categories
			double[][] root_results = new double[recordings_to_classify.length][leaf_labels.length];
			for (int i = 0; i < root_results.length; i++)
				for (int j = 0; j < root_results[i].length; j++)
					root_results[i][j] = 0.0;
			if (should_classify_into_root_categories)
			{
				int current_classifier_indice = getIndiceOfEnsembleWithGivenName("Root Classifier");
				double[][] root_level_results = classifier_ensemble[current_classifier_indice].classify( recordings_to_classify, null );
				String[] root_names = classifier_ensemble[current_classifier_indice].getCategories();
				for (int rec = 0; rec < recordings_to_classify.length; rec++)
				{
					int[] number_hits = new int[leaf_labels.length];
					for (int i = 0; i < number_hits.length; i++)
						number_hits[i] = 0;
					for (int root_cat = 0; root_cat < root_names.length; root_cat++)
					{
						boolean[] children = getLeafsBelongingToAncestor(root_names[root_cat], leaf_labels);
						for (int leaf_cat = 0; leaf_cat < children.length; leaf_cat++)
						{
							if (children[leaf_cat])
							{
								root_results[rec][leaf_cat] += root_level_results[rec][root_cat];
								number_hits[leaf_cat]++;
							}
						}
					}
					for (int i = 0; i < root_results[rec].length; i++)
						if (number_hits[i] != 0)
							root_results[rec][i] = root_results[rec][i] / ((double) number_hits[i]);
				}
			}

			// Combine into classification_results using weightings
			for (int rec = 0; rec < classification_results.length; rec++)
			{
				for (int cat = 0; cat < classification_results[rec].length; cat++)
				{
					if (should_classify_using_round_robin)
						classification_results[rec][cat] += round_robin_results[rec][cat] * round_robin_weighting;
					if (should_classify_into_all_leaf_categories)
						classification_results[rec][cat] += all_leaf_results[rec][cat] * leaf_weighting;
					if (should_classify_into_root_categories)
						classification_results[rec][cat] += root_results[rec][cat] * hierarchal_weighting;
				}
			}

			// Store the categories and model resulsts for the combined classification
			candidate_categories_for_combined_classification = leaf_labels;
			if (model_testing_classifications == null)
				model_testing_classifications_for_combined_classifier = null;
			else
			{
				model_testing_classifications_for_combined_classifier = new String[testing_recordings.length][];
				for (int rec = 0; rec < testing_recordings.length; rec++)
					model_testing_classifications_for_combined_classifier[rec] = testing_recordings[rec].leaf_categories;
			}
		}

		// Perform only root level classification
		else
		{
			// Find the results
			int current_classifier_indice = getIndiceOfEnsembleWithGivenName("Root Classifier");
			classification_results = classifier_ensemble[current_classifier_indice].classify( recordings_to_classify, null );

			// Store the categories and model resulsts for the combined classification
			candidate_categories_for_combined_classification =  classifier_ensemble[current_classifier_indice].getCategories();
			if (model_testing_classifications == null)
				model_testing_classifications_for_combined_classifier = null;
			else
				model_testing_classifications_for_combined_classifier = model_testing_classifications[current_classifier_indice];
		}

		// Store classification results, make them available on the selector and display them
		combined_classifier_results_all_ensembles[combined_classifier_results_all_ensembles.length - 1] = classification_results;
		classification_results_selector.addItem("Combined Classifier");
		if (print_classification_results)
			displayCombinedClassificationResults();
	}


	// Fill the the given classification_results array (one entry for each entry in leaf_category_names parameter
	// and should have all entries set to 0) with the summed ensemble of classifiers scores for each category in the
	// given leaf_category_names for the given recording. The classifiers_used parameter should be null if this method 
	// is being called by a method other than itself. This parameter is used internally to calculate the final
	// classification_results. The leaf_category_names parameter should include all possible leaves in the taxonomy
	// that the classifeirs were trained on.
	//
	// The ensemble_name parameter specifies what level of the taxonomical hierarchy is currently being operated at.
	// The should_classify_using_round_robin specifies whether round robin classification is being used and the 
	// should_classify_into_all_leaf_categories paramet specifies whether flat all leaf classification is being used.
	// The hierarchal_weighting, leaf_weighting and round_robin_weighting parameters specify what
	// weighting to give each of these components. The three weightings should add up to 1.
	//
	// This method calls itself recursively to descend down the hierarchy.
	//
	// This method requires that the taxonomy used in the classifiers matches the one in the current TaxonomyPanel.
	// This is not checked here, and must be ensured by the calling method.
	private void performHierarchicalClassification( Recording recording,
	                                                String[] leaf_category_names,
	                                                double[] classification_results,
	                                                int[] classifiers_used,
	                                                String ensemble_name,
	                                                boolean should_classify_using_round_robin,
	                                                boolean should_classify_into_all_leaf_categories,
	                                                double hierarchal_weighting,
	                                                double leaf_weighting,
	                                                double round_robin_weighting )
		throws Exception
	{
		// Record whether this method was called externally and initialize classifiers_used
		// if it was
		boolean called_externally;
		if (classifiers_used == null)
		{
			called_externally = true;
			classifiers_used = new int[classification_results.length];
			for (int i = 0; i < classifiers_used.length; i++)
				classifiers_used[i] = 0;
		}
		else
			called_externally = false;
		
		// To identify different classifier ensembles
		int current_classifier_indice;

		// The names of all categories at this level of the hierarchy
		current_classifier_indice = getIndiceOfEnsembleWithGivenName(ensemble_name);
		String[] this_level_category_names = classifier_ensemble[current_classifier_indice].getCategories();

		// Prepare the recording for classification
		Recording[] recording_array = new Recording[1];
		recording_array[0] = recording;

		// Perform the hierarchal classification with the enemble_name ensemble of classifiers on recording
		current_classifier_indice = getIndiceOfEnsembleWithGivenName(ensemble_name);
		double[][] hierarchal_scores_array = classifier_ensemble[current_classifier_indice].classify( recording_array, null );
		double[] hierarchal_scores = hierarchal_scores_array[0];

		// Perform flat leaf classification, if appropriate
		double[] combined_leaf_scores = null;
		if (should_classify_into_all_leaf_categories)
		{
			// Perform the all leaves classification into leaf categories
			current_classifier_indice = getIndiceOfEnsembleWithGivenName("Leaf Classifier");
			String[] leaf_classifier_category_names = classifier_ensemble[current_classifier_indice].getCategories();
			double[][] raw_leaf_scores_array = classifier_ensemble[current_classifier_indice].classify( recording_array, null );
			double[] raw_leaf_scores = raw_leaf_scores_array[0];
		
			// Combine the raw leaf scores into the candidate categories at this level of the hierarchy.
			// Discard those leaves that aren't descended from any of these categories.
			// combined_leaf_scores will hold the results for the leaf classifier
			combined_leaf_scores = getAncestorScoresFromLeafScores( raw_leaf_scores,
			                                                        leaf_classifier_category_names,
			                                                        this_level_category_names );
		}

		// Perform the round robin (if requested) classification
		double[] combined_round_robin_scores = null;
		if (should_classify_using_round_robin)
		{
			// Perform the round robin classification into all leaf scores
			double[] raw_leaf_scores = performRoundRobinClassification(recording, leaf_category_names);

			// Combine the raw leaf scores into the candidate categories at this level of the hierarchy.
			// Discard those leaves that aren't descended from any of these categories.
			// combined_leaf_scores will hold the results for the leaf classifier
			combined_round_robin_scores = getAncestorScoresFromLeafScores( raw_leaf_scores,
			                                                               leaf_category_names,
			                                                               this_level_category_names );
		}

		// Combine the results for the three types of classifications with weightings at this level
		// of the taxonomy
		double[] combined_scores_at_this_level = new double[this_level_category_names.length];
		for (int i = 0; i < combined_scores_at_this_level.length; i++)
		{
			combined_scores_at_this_level[i] = hierarchal_scores[i] * hierarchal_weighting;
			if (should_classify_into_all_leaf_categories)
				combined_scores_at_this_level[i] += combined_leaf_scores[i] * leaf_weighting;
			if (should_classify_using_round_robin)
				combined_scores_at_this_level[i] += combined_round_robin_scores[i] * round_robin_weighting;
		}

		// Determine which of the categories at this level of the taxonomy are leaves
		boolean[] is_leaf = isLeaf(this_level_category_names, leaf_category_names);
	
		// Store the classification certainty for those categories that are leaves and
		// determine if there are any non-leaf categories.
		boolean are_non_leaves = false;
		for (int i = 0; i < combined_scores_at_this_level.length; i++)
		{
			if (is_leaf[i])
			{
				int index_of_leaf = bodhidharma.utilities.StringMethods.getIndexOfString(this_level_category_names[i], leaf_category_names);
				classification_results[index_of_leaf] = combined_scores_at_this_level[i];
				classifiers_used[index_of_leaf]++;
			}
			else
				are_non_leaves = true;
		}

		// Find if there are any winning non-leaf categories, and recurse down them
		// (descend down the hierarchy) if there are.
		if (are_non_leaves)
		{
			// Fine the non-leaf winners, and store them in winners array
			double[][] combined_scores_at_this_level_array = new double[1][];
			combined_scores_at_this_level_array[0] = combined_scores_at_this_level;
			ClassificationResultsInterpereter results_interpereter = new ClassificationResultsInterpereter(parent_window.preferences_panel);
			float[][] descendant_scores =  results_interpereter.getWinningScores(combined_scores_at_this_level_array);
			boolean[] winners = new boolean[combined_scores_at_this_level.length];
			for (int i = 0; i < winners.length; i++)
			{
				if (descendant_scores[0][i] > 0 && !is_leaf[i])
					winners[i] = true;
				else
					winners[i] = false;
			}

			// Recurse down non-leaf winners
			for (int i = 0; i < winners.length; i++)
			{
				if (winners[i])
				{
					String name_of_descendant = "DD " + this_level_category_names[i] + " Classifier";
					performHierarchicalClassification( recording,
													   leaf_category_names,
													   classification_results,
													   classifiers_used,
													   name_of_descendant,
													   should_classify_using_round_robin,
													   should_classify_into_all_leaf_categories,
													   hierarchal_weighting,
													   leaf_weighting,
													   round_robin_weighting );
				}
			}
		}

		// Divide the final scores for each leaf category by the number of classifiers that assigned
		// it scores.
		if (called_externally)
			for (int i = 0; i < classification_results.length; i++)
			{
				if (classifiers_used[i] == 0)
					classification_results[i] = 0;
				else
					classification_results[i] = classification_results[i] / ((double) classifiers_used[i]);
			}
	}


	// Classifies the given recording using all of the round robin classifiers combined. The returned
	// array has one entry for each of the possible leaf categories, in the same order as
	// specified in leaf_category_names. Each entry of the returned array has the certainty
	// score (from 0 to 1, non-normalized) as produced by the classifiers. Throws an exception
	// if there is no trained classifer round robin for each of the entries of leaf_category_names.
	private double[] performRoundRobinClassification(Recording recording, String[] leaf_category_names)
		throws Exception
	{
		// Make sure that there is at least one round robin classifier
		boolean is_at_least_one = false;
		for (int ens = 0; ens < classifier_ensemble.length; ens++)
			if (classifier_ensemble[ens].getClassifierName().substring(0, 2).equals("RR"))
			{
				is_at_least_one = true;
				ens = classifier_ensemble.length + 1; // exit loop
			}
		if (!is_at_least_one)
			throw new Exception("Could not perform classification because no round robin\n" +
			                    "classifiers are available.");
				
		// Prepare to hold combined scores for all leaf categories
		double[] results = new double[leaf_category_names.length];
		int[] hits = new int[results.length];
		for (int i = 0; i < hits.length; i++)
			hits[i] = 0;

		// Prepare recording for classification
		Recording[] recording_array = new Recording[1];
		recording_array[0] = recording;

		// Classify the recording with each trained round robin classifier
		for (int ens = 0; ens < classifier_ensemble.length; ens++)
		{
			// Only include round robin classifier ensembles
			if (classifier_ensemble[ens].getClassifierName().substring(0, 2).equals("RR"))
			{
				// Find the indices of the two leaf categories this round robin classifier
				// deals with
				String first_category = (classifier_ensemble[ens].getCategories())[0];
				String second_category = (classifier_ensemble[ens].getCategories())[1];
				int first_category_indice = -1;
				int second_category_indice = -1;
				for (int cat = 0; cat < leaf_category_names.length; cat++)
				{
					if (first_category.equals(leaf_category_names[cat]))
						first_category_indice = cat;
					if (second_category.equals(leaf_category_names[cat]))
						second_category_indice = cat;
				}

				// Throw an exception if neither category recognized
				if (first_category_indice == -1)
					throw new Exception("Round robin classifier unable to find category " + first_category);
				if (second_category_indice == -1)
					throw new Exception("Round robin classifier unable to find category " + second_category);

				// Perform classification for this round robin ensemble
				double[][] temp_results =  classifier_ensemble[ens].classify(recording_array, null);

				// Store results for this round robin ensemble
				results[first_category_indice] += temp_results[0][0];
				results[second_category_indice] += temp_results[0][1];
				hits[first_category_indice]++;
				hits[second_category_indice]++;
			}
		}

		// Prepare the combined classifcation of all combined round robin classifiers
		for (int i = 0; i < results.length; i++)
			results[i] = results[i] / ((double) hits[i]);

		// If selected, divide all values by the value in the preferences
		if (parent_window.preferences_panel.getRoundRobinDivisor() != 1.0)
			for (int i = 0; i < results.length; i++)
				results[i] = results[i] / parent_window.preferences_panel.getRoundRobinDivisor();

		// If selected in the preferences, only consider the highest round robin value
		if (parent_window.preferences_panel.shouldOnlyUseBestForRoundRobin())
		{
			int index_of_highest = bodhidharma.utilities.Statistics.getIndexOfLargest(results);
			for (int i = 0; i < results.length; i++)
				if (i != index_of_highest)
					results[i] = 0.0;
		}

		// Return the results
		return results;
	}


	// Returns the last index of classifier_ensemble that has a classifier with the 
	// given name. Throws an exception if no such classifier is in classifier_ensemble.
	private int getIndiceOfEnsembleWithGivenName(String name)
		throws Exception
	{
		int index = -1;
		for (int i = 0; i < classifier_ensemble.length; i++)
			if (classifier_ensemble[i].getClassifierName().equals(name))
				index = i;

		if (index == -1)
			throw new Exception("Unable to perform classification because ensemble called " +
			                    name + " has not been trained.");
		
		return index;
	}


	// Takes in the leaf_scores, which are scores, each between 0 and 1, for each category
	// for a recording. Each entry of leaf_scores corresponds to a leaf category in leaf_categories.
	// The scores are combined into a set of categories higher up the hierarchy, based on the
	// current taxonomy, that are in ancestor_categories. The returned array stores a score
	// between 0 and 1 for each category in ancestor_categories.
	private double[] getAncestorScoresFromLeafScores( double[] leaf_scores,
	                                                  String[] leaf_categories,
	                                                  String[] ancestor_categories )
	{
		double[] ancestor_scores = new double[ancestor_categories.length];
		for (int i = 0; i < ancestor_scores.length; i++)
			ancestor_scores[i] = 0.0;
		int[] combined_hits = new int[ancestor_scores.length];
		for (int i = 0; i < combined_hits.length; i++)
			combined_hits[i] = 0;
		for (int i = 0; i < leaf_categories.length; i++)
		{
			boolean[] elligibles = getAncestorsLeafBelongsTo(ancestor_categories, leaf_categories[i]);
			for (int j = 0; j < elligibles.length; j++)
				if (elligibles[j])
				{
					ancestor_scores[j] += leaf_scores[i];
					combined_hits[j]++;
				}
		}
		for (int i = 0; i < ancestor_scores.length; i++)
			if (combined_hits[i] != 0)
				ancestor_scores[i] = ancestor_scores[i] / ((double) combined_hits[i]);
		return ancestor_scores;
	}


	// Returns an array whose entries correspond to those of the elligible_ancestors. Each entry
	// is set to true if the given leaf paramter is a decendant of the given elligible_ancestor
	// and to false if it is not, based on the currently loaded taxonomy.
	private boolean[] getAncestorsLeafBelongsTo(String[] elligible_ancestors, String leaf)
	{
		String[][] leaves_and_their_parents = parent_window.taxonomy_panel.getLeafsAndTheirParents();
		boolean[] membership = new boolean[elligible_ancestors.length];
		for (int i = 0; i < membership.length; i++)
			membership[i] = false;
		boolean done = false;
		for (int i = 0; i < elligible_ancestors.length; i++)
			for (int j = 0; j < leaves_and_their_parents.length; j++)
			{
				if (leaf.equals(leaves_and_their_parents[j][0]))
				{
					for (int k = 0; k < leaves_and_their_parents[j].length; k++)
					{
						if (elligible_ancestors[i].equals(leaves_and_their_parents[j][k]))
						{
							membership[i] = true;
							k = leaves_and_their_parents[j].length + 1; // exit loop
						}
					}
					j = leaves_and_their_parents.length + 1; // exit loop
				}
			}
		return membership;
	}


	// Returns an array whose entries correspond to the leaf categories in leaf_categories, and
	// are set to true if the ancestor parameter is an ancestor of the corresponding leaf category
	// under the current taxonomy.
	private boolean[] getLeafsBelongingToAncestor(String ancestor, String[] leaf_categories)
	{
		boolean[] results = new boolean[leaf_categories.length];
		for (int cat = 0; cat < results.length; cat++)
			results[cat] = false;

		String[][] leaves_and_ancestors = parent_window.taxonomy_panel.getLeafsAndTheirParents();
		for (int j = 0; j < leaves_and_ancestors.length; j++)
			for (int k = 0; k < leaves_and_ancestors[j].length; k++)
			{
				if (ancestor.equals( leaves_and_ancestors[j][k] ))
				{
					// Mark the correct leaf in results
					for (int m = 0; m < leaf_categories.length; m++)
					{
						if (leaf_categories[m].equals( leaves_and_ancestors[j][0] ))
						{
							results[m] = true;
							m = leaf_categories.length + 1; // exit the loop
						}
					}
					k = leaves_and_ancestors[j].length + 1; // exit loop
				}
			}
		
		return results;
	}
	

	// Returns an array with one entry for each entry in the candidate_categories parameter.
	// Each entry is set to true if the corresponding entry in candidate_categories is
	// contained in the leaf_categories parameter
	private boolean[] isLeaf(String[] candidate_categories, String[] leaf_categories)
	{
		boolean[] is_leaf = new boolean[candidate_categories.length];
		for (int cand = 0; cand < is_leaf.length; cand++)
		{
			is_leaf[cand] = false;
			for (int leaf = 0; leaf < leaf_categories.length; leaf++)
				if (candidate_categories[cand].equals(leaf_categories[leaf]))
				{
					is_leaf[cand] = true;
					leaf = leaf_categories.length + 1; // exit the loop
				}
		}
		return is_leaf;
	}
	
	
	// Returns a subset of all recordings loaded in the recordings panel that are
	// marked with the purpose given in the parameter. Throws an exception
	// if an invalid purpose is given.
	private Recording[] getRecordingsMarkedForPurpose(String purpose)
		throws Exception
	{
		// Make sure valid purpose is given
		if (!purpose.equals("training") && !purpose.equals("testing") && !purpose.equals("classification"))
			throw new Exception("Invalid recording purpose specified.");
		
		// Cause recording panel to reflect updates to roles
		parent_window.recordings_panel.updateRoles();

		// All loaded recordings
		Recording[] all_recordings = parent_window.recordings_panel.recordings_list;

		// Find number of recordings marked for the purpose in the RecordingsPanel
		int number_hits = 0;
		if (purpose.equals("training"))
		{
			for (int rec = 0; rec < all_recordings.length; rec++)
				if (all_recordings[rec].use_for_training)
					number_hits++;
		}
		else if (purpose.equals("testing"))
		{
			for (int rec = 0; rec < all_recordings.length; rec++)
				if (all_recordings[rec].use_for_testing)
					number_hits++;
		}
		else if (purpose.equals("classification"))
		{
			for (int rec = 0; rec < all_recordings.length; rec++)
				if (all_recordings[rec].use_for_classification)
					number_hits++;
		}

		// Return the subset of all_recordings
		Recording[] subset = new Recording[number_hits];
		int current_indice = 0;
		if (purpose.equals("training"))
		{
			for (int rec = 0; rec < all_recordings.length; rec++)
				if (all_recordings[rec].use_for_training)
				{
					subset[current_indice] = all_recordings[rec];
					current_indice++;
				}
		}
		else if (purpose.equals("testing"))
		{
			for (int rec = 0; rec < all_recordings.length; rec++)
				if (all_recordings[rec].use_for_testing)
				{
					subset[current_indice] = all_recordings[rec];
					current_indice++;
				}
		}
		else if (purpose.equals("classification"))
		{
			for (int rec = 0; rec < all_recordings.length; rec++)
				if (all_recordings[rec].use_for_classification)
				{
					subset[current_indice] = all_recordings[rec];
					current_indice++;
				}
		}

		// Return the appropriate subset of recordings
		return subset;
	}


	// Takes the given recordings, and returns an array containing the subset of elligible_categories
	// that each recording belongs to. The first indice of the returned array indicates recording
	// and the second indice indicates category. There is one entry for each category that each
	// recording belongs to. Categories at all levels of the classification hiearchy of each 
	// recording are considered.
	private String[][] getModelClassifications(Recording[] recordings, String[] elligible_categories)
	{
		// Find all of the categories that each recording belongs to, at all
		// levels of the hierarchy
		String[][] categories_belonged_to = new String[recordings.length][];
		for (int rec = 0; rec < categories_belonged_to.length; rec++)
		{
			categories_belonged_to[rec] = new String[ recordings[rec].leaf_categories.length + recordings[rec].ancestor_categories.length ];
			for (int cat = 0; cat < categories_belonged_to[rec].length; cat++)
			{
				if (cat < recordings[rec].leaf_categories.length)
					categories_belonged_to[rec][cat] = recordings[rec].leaf_categories[cat];
				else
					categories_belonged_to[rec][cat] = recordings[rec].ancestor_categories[cat - recordings[rec].leaf_categories.length];
			}
		}

		// Fill in the model categories
		String[][] model_classifications = new String[recordings.length][];
		for (int rec = 0; rec < model_classifications.length; rec++)
		{
			// Find the number of ellibible categories that the recording belongs to
			int number_belonged_to = 0;
			for (int ecat = 0; ecat < elligible_categories.length; ecat++)
				for (int rcat = 0; rcat < categories_belonged_to[rec].length; rcat++)
					if ( elligible_categories[ecat].equals( categories_belonged_to[rec][rcat] ) )
						number_belonged_to++;
			
			// Set the model categories for the recording	
			model_classifications[rec] = new String[number_belonged_to];
			int current_indice = 0;
			for (int ecat = 0; ecat < elligible_categories.length; ecat++)
				for (int rcat = 0; rcat < categories_belonged_to[rec].length; rcat++)
					if ( elligible_categories[ecat].equals( categories_belonged_to[rec][rcat] ) )
					{
						model_classifications[rec][current_indice] = elligible_categories[ecat];
						current_indice++;
					}
		}

		// Return the model classifications
		return model_classifications;
	}
	

	/**
	 * Returns an array with an entry for each recording referred to in model_training_classifications.
	 * Entries are set to true if the referred to recording belongs to at least one category in
	 * elligible_categories. An exception is thrown if a recoding does not belong to any categories
	 * in elligible_categories.
	 *
	 * @param	model_training_classifications	Model classifications of recordings. First indice indicates recording and second category(ies).
	 * @param	elligible_categories			Categories that are used by a classifier.
	 */
	private boolean[] getRecordingsBelongingToElligibleCategories( String[][] model_training_classifications,
	                                                               String[] elligible_categories )
		throws Exception
	{
		// Find the subset of recordings that belong to at least one category in descendant_names
		// and make sure that each elligible category in elligible_categories is present in at
		// least one recording
		boolean[] recording_contains_a_category = new boolean[model_training_classifications.length];
		boolean[] elligible_category_present_at_least_once = new boolean[elligible_categories.length];
		for (int cat = 0; cat < elligible_category_present_at_least_once.length; cat++)
			elligible_category_present_at_least_once[cat] = false;
		for (int rec = 0; rec < recording_contains_a_category.length; rec++)
		{
			if (model_training_classifications[rec].length != 0)
			{
				// Mark that the recording has a category that is a subset of elligible_categories
				recording_contains_a_category[rec] = true;
			
				// Mark that at least one of the elligible categories is present in training_recordings
				for (int ecat = 0; ecat < elligible_categories.length; ecat++)
					for (int mcat = 0; mcat < model_training_classifications[rec].length; mcat++)
						if (elligible_categories[ecat].equals(model_training_classifications[rec][mcat]))
							elligible_category_present_at_least_once[ecat] = true;
			}
			else
				recording_contains_a_category[rec] = false;
		}

		// Throw an exception if not every elligible category has at
		// least one training recording
		for (int cat = 0; cat < elligible_category_present_at_least_once.length; cat++)
			if (!elligible_category_present_at_least_once[cat])
				throw new Exception("No training recordings are available in the " +
									elligible_categories[cat] + " category.\nAll categories in the " +
									"taxonomy must be represented during training.\n" +
		                            "Please make sure that there are a sufficient number of\n" + 
									"recordings from each leaf category available to be used\n" +
									"as training examples. Remember that only a fraction of " + parent_window.preferences_panel.getTrainingFraction() + "\n" +
			                        "of recordings per category are used for training under your current\n" +
			                        "settings if you have set recordings to be partitioned into training\n" +
			                        "and testing examples.");

		// Return the information
		return recording_contains_a_category;
	}


	// Verifies that there are a sufficient number of training examples per category
	// in order for training to proceed, given the current preferences. Throws an exception
	// if there is not a sufficient number.
	private void verifySufficientTrainingExamples( String[] elligible_categories, 
	                                               String[][] model_training_classifications )
		throws Exception
	{
		// Reference user preferences
		PreferencesPanel preferences = parent_window.preferences_panel;
		
		// Find the number of recordings belonging to each elligible category (can
		// have more than the number of recordings, since each recording can belong
		// to more than one category)
		int[] category_count = new int[elligible_categories.length];
		for (int cat = 0; cat < category_count.length; cat++)
			category_count[cat] = 0;
		for (int rec = 0; rec < model_training_classifications.length; rec++)
			for (int rec_cat = 0; rec_cat < model_training_classifications[rec].length; rec_cat++)
				for (int cat = 0; cat < elligible_categories.length; cat++)
					if ( elligible_categories[cat].equals( model_training_classifications[rec][rec_cat] ) )
						category_count[cat]++;

		// Find the minimum number of recordings needed per category
		// (based on the Selection/Weighting Training Fraction option in the preferences
		double min_required_frac = 1 / preferences.getInternalTrainingFraction();
		double integer_part = Math.floor(min_required_frac);
		int min_required_int = 1;
		if ( ( preferences.shouldUseOneDimensionalFeaturesInTraining() && (preferences.shouldBioKNNsUseFeatureSelection() || preferences.shouldBioKNNsUseFeatureWeighting()) ) ||
			 ( preferences.shouldUseMultiDimensionalFeaturesInTraining() && (preferences.shouldSelectClassifiersInEachEnsemble() || preferences.shouldWeightClassifiersInEachEnsemble()) ) )
		{
			if ( (min_required_frac - integer_part) == 0)
				min_required_int = (new Double(min_required_frac)).intValue();
			else
				min_required_int = (new Double(min_required_frac)).intValue() + 1;
		}

		// Report an error if there are an insufficient number of recordings belonging
		// to any category
		for (int cat = 0; cat < category_count.length; cat++)
			if (category_count[cat] < min_required_int)
			{
				// Find the total amount of recordings needed per category (taking into account
				// the number of recordings that were reserved for testing)
				int total_min_required_int = min_required_int;
				if ( ( preferences.shouldUseOneDimensionalFeaturesInTraining() && (preferences.shouldBioKNNsUseFeatureSelection() || preferences.shouldBioKNNsUseFeatureWeighting()) ) ||
					 ( preferences.shouldUseMultiDimensionalFeaturesInTraining() && (preferences.shouldSelectClassifiersInEachEnsemble() || preferences.shouldWeightClassifiersInEachEnsemble()) ) )
				{
					double total_min_required_frac = (new Double(min_required_int)).doubleValue() / preferences.getTrainingFraction();
					double total_integer_part = Math.floor(total_min_required_frac);
					total_min_required_int = 1;
					if ( (total_min_required_frac - total_integer_part) == 0)
						total_min_required_int = (new Double(total_min_required_frac)).intValue();
					else
						total_min_required_int = (new Double(total_min_required_frac)).intValue() + 1;
				}

				// Throw an informative exception
				throw new Exception("Insufficient training recordings available belonging to the " + elligible_categories[cat] + " category.\n" +
				                    "At least " + min_required_int + " recordings reserved for testing and " + +total_min_required_int + " recordings in all\n" +
				                    "belonging to this category must be available given the current preferences.\n" +
				                    "In addition, it is preferable to have significantly more recordings belonging\n" +
				                    "to each category, if possible.");
			}
	}


	// Returns a smaller version of candidates, that only contains the corresponding
	// entries marked true in qualifers.
	private Recording[] getReducedRecordings(Recording[] candidates, boolean[] qualifiers)
	{
		int number_qualifiers = 0;
		for (int rec = 0; rec < qualifiers.length; rec++)
			if (qualifiers[rec])
				number_qualifiers++;

		Recording[] results = new Recording[number_qualifiers];
		int current_indice = 0;
		for (int rec = 0; rec < qualifiers.length; rec++)
			if (qualifiers[rec])
			{
				results[current_indice] = candidates[rec];
				current_indice++;
			}

		return results;
	}


	// Returns a smaller version of candidates, that only contains the corresponding
	// entries marked true in qualifers.
	private String[][] getReducedModelClassifications(String[][] candidates, boolean[] qualifiers)
	{
		int number_qualifiers = 0;
		for (int rec = 0; rec < qualifiers.length; rec++)
			if (qualifiers[rec])
				number_qualifiers++;

		String[][] results = new String[number_qualifiers][];
		int current_indice = 0;
		for (int rec = 0; rec < qualifiers.length; rec++)
			if (qualifiers[rec])
			{
				results[current_indice] = candidates[rec];
				current_indice++;
			}

		return results;
	}


	// Returns an array whose indice identifies the recording (in the same order
	// as the first indice of the model_results parameter). Each entry is set
	// to true if the recording is to be used for training and to false if it
	// is to be used for testing.
	// 
	// The model_clssification_names parameter gives the names of the correct
	// category(ies) for each recording. First indicie indicates recording.
	// There is one entry for each correct category for each recording.
	//
	// The elligible_categories parameter identifies the categories to which
	// ar recording can belong.
	//
	// The training_fraction identifies the fraction of all samples that should
	// be used for training. This value must be between 0 and 1.
	//
	// The fraction of samples belonging to each category given by training_sample
	// is used for training. In other words, steps are taken to ensure that the
	// number of samples belonging to each category used for training is proportionately 
	// the same as the fraction of the total number of samples belonging to each category.
	private boolean[] getRandomlyChosenTrainingAndTestingSamples( String[][] model_classification_names,
	                                                              String[] elligible_categories,
	                                                              double training_fraction )
	{
		// The total number of recordings
		int num_recordings = model_classification_names.length;

		// The total number of categories
		int num_categories = elligible_categories.length;
		
		// Generate the model results in boolean form
		boolean[][] model_results = new boolean[num_recordings][num_categories];
		for (int rec = 0; rec < model_results.length; rec++)
			for (int ecat = 0; ecat < model_results[rec].length; ecat++)
			{
				model_results[rec][ecat] = false;
				for (int scat = 0; scat < model_classification_names[rec].length; scat++)
					if ( model_classification_names[rec][scat].equals(elligible_categories[ecat]) )
						model_results[rec][ecat] = true;
			}

		// Identifies which recordings are to be used for training
		boolean[] reduced_training_list = new boolean[num_recordings];
		for (int samp = 0; samp < reduced_training_list.length; samp++)
			reduced_training_list[samp] = false;
		
		// Choose the training samples
		for (int cat = 0; cat < num_categories; cat++)
		{
			// Find the number of samples belonging to this category that
			// have not already been selected for training
			int number_belonging_to_cat = 0;
			for (int samp = 0; samp < num_recordings; samp++)
				if (model_results[samp][cat] && !reduced_training_list[samp])
					number_belonging_to_cat++;

			// Store the sample indices of samples belonging to this category that
			// have not already been selected for training
			int[] indices_belonging_to_category = new int[number_belonging_to_cat];
			int current_sample = 0;
			for (int samp = 0; samp < num_recordings; samp++)
				if (model_results[samp][cat] && !reduced_training_list[samp])
				{
					indices_belonging_to_category[current_sample] = samp;
					current_sample++;
				}
			
			// Randomly select training_fraction of the category for training by randomly
			// generating a number for each sample that belongs to cat and has not already
			// been chosen, and choosing the highest number (as set by training_fraction).
			int random_numbers[] = new int[indices_belonging_to_category.length];
			for (int i = 0; i < indices_belonging_to_category.length; i++)
				random_numbers[i] = bodhidharma.utilities.Statistics.generateRandomNumber(indices_belonging_to_category.length * 3);
			int num_to_train = (int) (((double) number_belonging_to_cat) * training_fraction);
			for (int i = 0; i < num_to_train; i++)
			{
				int index_of_largest = bodhidharma.utilities.Statistics.getIndexOfLargest(random_numbers);
				random_numbers[index_of_largest] = 0; // to avoid double counting
				reduced_training_list[ indices_belonging_to_category[index_of_largest] ] = true;
			}
		}

		// Return the information identifying which samples should be used for training
		return reduced_training_list;
	}


	// Returns a 2-D array whose first indice identifies the a fold in a cross-
	// validation and whose second indice identifies the recording (in the same order
	// as the first indice of the model_results parameter). Each entry is set
	// to true if the recording is to be used for training and to false if it
	// is to be used for testing during a given fold.
	// 
	// The model_clssification_names parameter gives the names of the correct
	// category(ies) for each recording. First indicie indicates recording.
	// There is one entry for each correct category for each recording.
	//
	// The elligible_categories parameter identifies the categories to which
	// ar recording can belong.
	//
	// The training_fraction parameter identifies the fraction of all samples that should
	// be used for training. This value must be between 0 and 1.
	//
	// The fraction of samples belonging to each category given by training_fraction
	// is used for training. In other words, steps are taken to ensure that the
	// number of samples belonging to each category used for training is proportionately 
	// the same as the fraction of the total number of samples belonging to each category.
	//
	// Each recording can only be used for testing in one of the folds.
	private boolean[][] getRandomlyChosenTrainingAndTestingSamplesForMultipleFolds( String[][] model_classification_names,
	                                                                                String[] elligible_categories,
	                                                                                double training_fraction )
	{
		// Find the number of folds to use			
		int number_of_folds = 1 + new Long(Math.round(training_fraction / (1.0 - training_fraction))).intValue();

		// The total number of recordings
		int num_recordings = model_classification_names.length;

		// The total number of categories
		int num_categories = elligible_categories.length;
		
		// Generate the model results in boolean form
		boolean[][] model_results = new boolean[num_recordings][num_categories];
		for (int rec = 0; rec < model_results.length; rec++)
			for (int ecat = 0; ecat < model_results[rec].length; ecat++)
			{
				model_results[rec][ecat] = false;
				for (int scat = 0; scat < model_classification_names[rec].length; scat++)
					if ( model_classification_names[rec][scat].equals(elligible_categories[ecat]) )
						model_results[rec][ecat] = true;
			}

		// Identifies in which fold recordings are to be used for testing
		int[] testing_list = new int[num_recordings];
		for (int samp = 0; samp < testing_list.length; samp++)
			testing_list[samp] = -1;

		// Choose the training samples
		int current_fold = 0;
		for (int cat = 0; cat < num_categories; cat++)
		{
			// Find the number of samples belonging to this category that
			// have not already been selected for testing
			int number_belonging_to_cat = 0;
			for (int samp = 0; samp < num_recordings; samp++)
				if (model_results[samp][cat] && testing_list[samp] == -1)
					number_belonging_to_cat++;

			// Store the sample indices of samples belonging to this category that
			// have not already been selected for testing
			int[] indices_belonging_to_category = new int[number_belonging_to_cat];
			int current_sample = 0;
			for (int samp = 0; samp < num_recordings; samp++)
				if (model_results[samp][cat] && testing_list[samp] == -1)
				{
					indices_belonging_to_category[current_sample] = samp;
					current_sample++;
				}

			// Randomly select testing fraction of the category for testing for each fold
			// by randomly generating a number for each sample that belongs to cat and has
			// not already been chosen, and choosing the highest number (as set by training_fraction).
			int random_numbers[] = new int[indices_belonging_to_category.length];
			for (int i = 0; i < indices_belonging_to_category.length; i++)
				random_numbers[i] = bodhidharma.utilities.Statistics.generateRandomNumber(indices_belonging_to_category.length * 100);
			for (int i = 0; i < random_numbers.length; i++)
			{
				int index_of_largest = bodhidharma.utilities.Statistics.getIndexOfLargest(random_numbers);
				random_numbers[index_of_largest] = -1; // to avoid double counting
				testing_list[ indices_belonging_to_category[index_of_largest] ] = current_fold;
				current_fold++;
				if (current_fold == number_of_folds)
					current_fold = 0;
			}
		}

		// Prepare the training list in boolean form
		boolean[][] training_keys = new boolean[number_of_folds][num_recordings];
		for (int samp = 0; samp < num_recordings; samp++)
			for (int fold = 0; fold < number_of_folds; fold++)
			{
				if (fold == testing_list[samp])
					training_keys[fold][samp] = false;
				else
					training_keys[fold][samp] = true;
			}

		// Return the information identifying which samples should be used for training
		return training_keys;
	}


	// Generates and saves reports on the average weightings of one-dimensional features and multi-dimensional
	// features accross folds. Generates a directory in the projects directory called AverageWeightings.
	// The following files are generated in this directory, with the following contents:
	//    - Overall_Weightings_Report.txt: gives the normalized combined average weighting and standard deviation of
	//      each feature group as well as the ranked normalized averages and standard deviations of each feature (one and
	//      multi-dimensional). Note that the standard deviations are calculated accross classifier ensembles, not folds.
	//      Note that the average weightings for the one-dimensional features are scaled by the overall weighting of
	//      the one-dimensional classifier.
	//    - A file for each classifier ensemble: gives the normalized combined average weighting each feature group for the
	//      ensemble, the ranked normalized averages of each feature (one and multi-dimensional) (with the average 
	//      weightings for the one-dimensional features are scaled by the overall weighting of the one-dimensional classifier) 
	//      and separate normalized unranked weightings and standard deviations (accross folds) for the classifers and the
	//      one-dimensional features (both scaled and unscaled by the ranking of the one-dimensional classifier).
	// NOTE: Only displays information if one-dimensional features were used
	private void saveAverageWeightings()
		throws Exception
	{
		if (one_dim_features_names_accross_folds != null)
		{
			// Find the averages and standard deviations for each feature and classifier accross the folds for each
			// ensemble of classifiers
			double[][] one_dim_averages = new double[ensemble_name_accross_folds.length][one_dim_features_names_accross_folds.length];
			double[][] one_dim_stdev = new double[ensemble_name_accross_folds.length][one_dim_features_names_accross_folds.length];
			double[][] classifier_averages = new double[ensemble_name_accross_folds.length][classifier_names_accross_folds.length];
			double[][] classifier_stdev = new double[ensemble_name_accross_folds.length][classifier_names_accross_folds.length];
			for (int ens = 0; ens < ensemble_name_accross_folds.length; ens++)
			{
				// Do for one dimensional features
				for (int feat = 0; feat < one_dim_features_names_accross_folds.length; feat++)
				{
					double[] temp = new double[one_dim_feature_weightings_accross_folds.length];
					for (int fold = 0; fold < one_dim_feature_weightings_accross_folds.length; fold++)
						temp[fold] = one_dim_feature_weightings_accross_folds[fold][ens][feat];
					one_dim_averages[ens][feat] = bodhidharma.utilities.Statistics.getAverage(temp);
					one_dim_stdev[ens][feat] = bodhidharma.utilities.Statistics.getStandardDeviation(temp);
				}

				// Do for classifiers
				for (int feat = 0; feat < classifier_names_accross_folds.length; feat++)
				{
					double[] temp = new double[classifier_weightings_accross_folds.length];
					for (int fold = 0; fold < classifier_weightings_accross_folds.length; fold++)
						temp[fold] = classifier_weightings_accross_folds[fold][ens][feat];
					classifier_averages[ens][feat] = bodhidharma.utilities.Statistics.getAverage(temp);
					classifier_stdev[ens][feat] = bodhidharma.utilities.Statistics.getStandardDeviation(temp);
				}
			}

			// Find the one-dimensional feature weightings scaled by the One Dimensional
			// Feature Classifier weighting
			double[][] scaled_one_dim_averages = new double[one_dim_averages.length][one_dim_averages[0].length];
			for (int ens = 0; ens < ensemble_name_accross_folds.length; ens++)
			{
				double scaling_factor = 1.0;
				for (int i = 0; i < classifier_names_accross_folds.length; i++)
					if (classifier_names_accross_folds[i].equals("One Dimensional Feature Classifier"))
					{
						scaling_factor = classifier_averages[ens][i];
						i = classifier_names_accross_folds.length + 1; // exit loop
					}
				for (int feat = 0; feat < scaled_one_dim_averages[ens].length; feat++)
					scaled_one_dim_averages[ens][feat] = one_dim_averages[ens][feat] * scaling_factor;
			}


			// Find the ordered list of features (one-dimensional as well as multi-dimensional)
			int total_size = one_dim_averages[0].length + classifier_averages[0].length;
			for (int feat = 0; feat < classifier_codes_accross_folds.length; feat++)
				if (classifier_codes_accross_folds[feat].equals(" "))
					total_size--;
			double[][] ranked_weightings = new double[one_dim_averages.length][total_size];
			String[][] ranked_feature_names = new String[one_dim_averages.length][total_size];
			String[][] ranked_feature_codes = new String[one_dim_averages.length][total_size];
			int cur_pos = 0;
			for (int ens = 0; ens < ranked_weightings.length; ens++) // combine one and multi dimensional features
			{
				for (int o_feat = 0; o_feat < scaled_one_dim_averages[ens].length; o_feat++)
				{
					ranked_weightings[ens][o_feat] = scaled_one_dim_averages[ens][o_feat];
					ranked_feature_names[ens][o_feat] = one_dim_features_names_accross_folds[o_feat];
					ranked_feature_codes[ens][o_feat] = one_dim_feature_codes_accross_folds[o_feat];
					cur_pos = o_feat;
				}
				cur_pos++;
				for (int m_feat = 0; m_feat < classifier_averages[ens].length; m_feat++)
				{
					if (!classifier_codes_accross_folds[m_feat].equals(" "))
					{
						ranked_weightings[ens][cur_pos] = classifier_averages[ens][m_feat];
						ranked_feature_names[ens][cur_pos] = classifier_names_accross_folds[m_feat];
						ranked_feature_codes[ens][cur_pos] = classifier_codes_accross_folds[m_feat];
						cur_pos++;
					}
				}
			}
			for (int ens = 0; ens < ranked_weightings.length; ens++) // sort
			{
				int[] sort_key = bodhidharma.utilities.Sorter.getDoubleArraySortKey(ranked_weightings[ens]);
				double[] temp_ranked_weightings = new double[total_size];
				String[] temp_ranked_feature_names = new String[total_size];
				String[] temp_ranked_feature_codes = new String[total_size];
				for (int feat = 0; feat < total_size; feat++)
					for (int pos = 0; pos < sort_key.length; pos++)
						if (sort_key[pos] == feat)
						{
							temp_ranked_weightings[feat] = ranked_weightings[ens][pos];
							temp_ranked_feature_names[feat] = ranked_feature_names[ens][pos];
							temp_ranked_feature_codes[feat] = ranked_feature_codes[ens][pos];
							pos = sort_key.length + 1; // exit loop
						}
				ranked_weightings[ens] = temp_ranked_weightings;
				ranked_feature_names[ens] = temp_ranked_feature_names;
				ranked_feature_codes[ens] = temp_ranked_feature_codes;
			}
			for (int ens = 0; ens < ranked_weightings.length; ens++) // reverse list order
			{
				double[] temp_ranked_weightings = new double[total_size];
				String[] temp_ranked_feature_names = new String[total_size];
				String[] temp_ranked_feature_codes = new String[total_size];
				for (int feat = 0; feat < total_size; feat++)
				{
					temp_ranked_weightings[feat] = ranked_weightings[ens][total_size - feat - 1];
					temp_ranked_feature_names[feat] = ranked_feature_names[ens][total_size - feat - 1];
					temp_ranked_feature_codes[feat] = ranked_feature_codes[ens][total_size - feat - 1];
				}
				ranked_weightings[ens] = temp_ranked_weightings;
				ranked_feature_names[ens] = temp_ranked_feature_names;
				ranked_feature_codes[ens] = temp_ranked_feature_codes;
			}

			// Find the combined score for feature groups
			String[] feature_group_names = { "Instrumentation", "Texture", "Rhythm", "Dynamics", "Pitch", "Melody", "Chords" };
			String[] feature_group_initials = { "I", "T", "R", "D", "P", "M", "C" };
			double[][] combined_scores = new double[ranked_weightings.length][feature_group_names.length];
			int[][] feature_group_count = new int[ranked_weightings.length][feature_group_names.length];
			for (int ens = 0; ens < combined_scores.length; ens++)
				for (int group = 0; group < combined_scores[ens].length; group++)
				{
					combined_scores[ens][group] = 0.0;
					feature_group_count[ens][group] = 0;
				}
			for (int ens = 0; ens < combined_scores.length; ens++)
				for (int feat = 0; feat < ranked_feature_codes[ens].length; feat++)
					for (int group = 0; group < feature_group_initials.length; group++)
						if ( feature_group_initials[group].equals( bodhidharma.utilities.StringMethods.getBeginningOfString(ranked_feature_codes[ens][feat], 1) ) )
						{
							combined_scores[ens][group] += ranked_weightings[ens][feat];
							feature_group_count[ens][group]++;
							group = feature_group_initials.length + 1; // exit loop
						}

			// Create directory for saving weighting files if does not exist already
			String average_weightings_save_path = multi_fold_dir_to_save.getPath() + File.separator + "AverageWeightings" + File.separator;
			File weightings_directory = new File(average_weightings_save_path);
			if (weightings_directory.isFile() == false && weightings_directory.isDirectory() == false)
				weightings_directory.mkdir();

			// Write a file for each classifier ensemble
			for (int ens = 0; ens < ensemble_name_accross_folds.length; ens++)
			{
				// Open output stream for writing file
				String file_save_path = average_weightings_save_path + ensemble_name_accross_folds[ens] + ".txt";
				File save_file = new File(file_save_path);
				FileOutputStream save_to = new FileOutputStream(save_file);
				DataOutputStream save_writer = new DataOutputStream(save_to);

				// Write the file
				int heading_length = 60;
				save_writer.writeBytes("<<<<<< FEATURE & CLASSIFIER AVERAGE WEIGHTING AND STDEV FOR " + ensemble_name_accross_folds[ens] + " ACCROSS " + one_dim_feature_weightings_accross_folds.length + " FOLDS >>>>>>\n");
				save_writer.writeBytes("\n\n\n");
				save_writer.writeBytes("------------------------------\n" +
									   "FEATURE GROUP COMBINED SCORES:\n" +
									   "------------------------------\n\n");
				for (int group = 0; group < feature_group_names.length; group++)
				{
					save_writer.writeBytes( bodhidharma.utilities.StringMethods.getBeginningOfStringWithHyphenFiller(feature_group_names[group], 21) +
											bodhidharma.utilities.StringMethods.getRoundedDouble(combined_scores[ens][group], 4) + 
											" (combined total of " + feature_group_count[ens][group] + " features in the group)\n" );
				}
				save_writer.writeBytes("\n\n\n");
				save_writer.writeBytes("------------------------------------------------------\n" +
									   "RANKED COMBINED WEIGHTINGS (1-D AND MULTI-D FEATURES):\n" +
									   "------------------------------------------------------\n\n");
				for (int feat = 0; feat < ranked_feature_codes[ens].length; feat++)
				{
					save_writer.writeBytes( (feat + 1) + "." + "\t" +
											ranked_feature_codes[ens][feat] + "\t" +
											bodhidharma.utilities.StringMethods.getBeginningOfStringWithHyphenFiller(ranked_feature_names[ens][feat], heading_length) +
											bodhidharma.utilities.StringMethods.getRoundedDouble(ranked_weightings[ens][feat], 5) + "\n" );
				}
				save_writer.writeBytes("\n\n\n");
				save_writer.writeBytes("---------------------\n" +
									   "CLASSIFIER WEIGHTINGS:\n" +
									   "----------------------\n\n");
				for (int feat = 0; feat < classifier_names_accross_folds.length; feat++)
				{
					save_writer.writeBytes( classifier_codes_accross_folds[feat] + "\t" +
											bodhidharma.utilities.StringMethods.getBeginningOfStringWithHyphenFiller(classifier_names_accross_folds[feat], heading_length) +
											bodhidharma.utilities.StringMethods.getRoundedDouble(classifier_averages[ens][feat], 4) +
											"  (stdev: " + bodhidharma.utilities.StringMethods.getRoundedDouble(classifier_stdev[ens][feat], 4) + ")\n" );
				}
				save_writer.writeBytes("\n\n\n");
				save_writer.writeBytes("-----------------------------------------------------------------------\n" +
									   "ONE-DIMENSIONAL FEATURE WEIGHTINGS (SCALED BY 1-D CLASSIFIER WEIGHTING):\n" +
									   "-----------------------------------------------------------------------\n\n");
				for (int feat = 0; feat < one_dim_features_names_accross_folds.length; feat++)
				{
					save_writer.writeBytes( one_dim_feature_codes_accross_folds[feat] + "\t" +
											bodhidharma.utilities.StringMethods.getBeginningOfStringWithHyphenFiller(one_dim_features_names_accross_folds[feat], heading_length) +
											bodhidharma.utilities.StringMethods.getRoundedDouble(scaled_one_dim_averages[ens][feat], 6) + "\n" );
				}
				save_writer.writeBytes("\n\n\n");
				save_writer.writeBytes("----------------------------------------------\n" +
									   "ONE-DIMENSIONAL FEATURE WEIGHTINGS (UNSCALED):\n" +
									   "----------------------------------------------\n\n");
				for (int feat = 0; feat < one_dim_features_names_accross_folds.length; feat++)
				{
					save_writer.writeBytes( one_dim_feature_codes_accross_folds[feat] + "\t" +
											bodhidharma.utilities.StringMethods.getBeginningOfStringWithHyphenFiller(one_dim_features_names_accross_folds[feat], heading_length) +
											bodhidharma.utilities.StringMethods.getRoundedDouble(one_dim_averages[ens][feat], 4) +
											"  (stdev: " + bodhidharma.utilities.StringMethods.getRoundedDouble(one_dim_stdev[ens][feat], 4) + ")\n" );
				}
				save_writer.writeBytes("\n");

				// Close the output stream
				save_writer.close();
			}

			// Find the ranked combined averages accross all classifier ensembles (one-dimensional
			// and multi-dimensional). Also find standard deviation (accross classifier ensembles,
			// has nothing to do with accross folds).
			double[] ranked_weightings_accross_ensembles = new double[total_size];
			double[] stdev_of_ranked_weightings_accross_ensembles = new double[total_size];
			String[] ranked_feature_names_accross_ensembles = new String[total_size];
			String[] ranked_feature_codes_accross_ensembles = new String[total_size];
			cur_pos = 0;
			for (int o_feat = 0; o_feat < scaled_one_dim_averages[0].length; o_feat++) // for 1-dimensional features
			{
				double[] temp = new double[scaled_one_dim_averages.length];
				for (int ens = 0; ens < scaled_one_dim_averages.length; ens++)
					temp[ens] = scaled_one_dim_averages[ens][o_feat];
				ranked_weightings_accross_ensembles[o_feat] = bodhidharma.utilities.Statistics.getAverage(temp);
				stdev_of_ranked_weightings_accross_ensembles[o_feat] = bodhidharma.utilities.Statistics.getStandardDeviation(temp);
				ranked_feature_names_accross_ensembles[o_feat] = one_dim_features_names_accross_folds[o_feat];
				ranked_feature_codes_accross_ensembles[o_feat] = one_dim_feature_codes_accross_folds[o_feat];
				cur_pos = o_feat;
			}
			cur_pos++;
			for (int m_feat = 0; m_feat < classifier_averages[0].length; m_feat++) // for multi-dimensional features
			{
				if (!classifier_codes_accross_folds[m_feat].equals(" "))
				{
					double[] temp = new double[classifier_averages.length];
					for (int ens = 0; ens < classifier_averages.length; ens++)
						temp[ens] = classifier_averages[ens][m_feat];
					ranked_weightings_accross_ensembles[cur_pos] = bodhidharma.utilities.Statistics.getAverage(temp);
					stdev_of_ranked_weightings_accross_ensembles[cur_pos] = bodhidharma.utilities.Statistics.getStandardDeviation(temp);
					ranked_feature_names_accross_ensembles[cur_pos] = classifier_names_accross_folds[m_feat];
					ranked_feature_codes_accross_ensembles[cur_pos] = classifier_codes_accross_folds[m_feat];
					cur_pos++;
				}
			}
			int[] sort_key = bodhidharma.utilities.Sorter.getDoubleArraySortKey(ranked_weightings_accross_ensembles); // sort
			double[] temp_ranked_weightings = new double[total_size];
			double[] temp_stdev_ranked_weightings = new double[total_size];
			String[] temp_ranked_feature_names = new String[total_size];
			String[] temp_ranked_feature_codes = new String[total_size];
			for (int feat = 0; feat < total_size; feat++)
				for (int pos = 0; pos < sort_key.length; pos++)
					if (sort_key[pos] == feat)
					{
						temp_ranked_weightings[feat] = ranked_weightings_accross_ensembles[pos];
						temp_stdev_ranked_weightings[feat] = stdev_of_ranked_weightings_accross_ensembles[pos];
						temp_ranked_feature_names[feat] = ranked_feature_names_accross_ensembles[pos];
						temp_ranked_feature_codes[feat] = ranked_feature_codes_accross_ensembles[pos];
						pos = sort_key.length + 1; // exit loop
					}
			ranked_weightings_accross_ensembles = temp_ranked_weightings;
			stdev_of_ranked_weightings_accross_ensembles = temp_stdev_ranked_weightings;
			ranked_feature_names_accross_ensembles = temp_ranked_feature_names;
			ranked_feature_codes_accross_ensembles = temp_ranked_feature_codes;
			temp_ranked_weightings = new double[total_size];
			temp_stdev_ranked_weightings = new double[total_size];
			temp_ranked_feature_names = new String[total_size];
			temp_ranked_feature_codes = new String[total_size];
			for (int feat = 0; feat < total_size; feat++) // reverse order
			{
				temp_ranked_weightings[feat] = ranked_weightings_accross_ensembles[total_size - feat - 1];
				temp_stdev_ranked_weightings[feat] = stdev_of_ranked_weightings_accross_ensembles[total_size - feat - 1];
				temp_ranked_feature_names[feat] = ranked_feature_names_accross_ensembles[total_size - feat - 1];
				temp_ranked_feature_codes[feat] = ranked_feature_codes_accross_ensembles[total_size - feat - 1];
			}
			ranked_weightings_accross_ensembles = temp_ranked_weightings;
			stdev_of_ranked_weightings_accross_ensembles = temp_stdev_ranked_weightings;
			ranked_feature_names_accross_ensembles = temp_ranked_feature_names;
			ranked_feature_codes_accross_ensembles = temp_ranked_feature_codes;

			// Find the combined scores for feature groups accross all classifier ensembles
			double[] combined_group_scores_accross_ensembles = new double[feature_group_names.length];
			double[] combined_stdev_group_scores_accross_ensembles = new double[feature_group_names.length];
			for (int group = 0; group < combined_group_scores_accross_ensembles.length; group++)
			{
				double[] temp = new double[combined_scores.length];
				for (int ens = 0; ens < combined_scores.length; ens++)
					temp[ens] = combined_scores[ens][group];
				combined_group_scores_accross_ensembles[group] = bodhidharma.utilities.Statistics.getAverage(temp);
				combined_stdev_group_scores_accross_ensembles[group] = bodhidharma.utilities.Statistics.getStandardDeviation(temp);
			}
		
			// Write a file for averages accross all ensembles
			String overall_file_save_path = average_weightings_save_path + "Overall_Weightings_Report.txt";
			File overall_save_file = new File(overall_file_save_path);
			FileOutputStream overall_save_to = new FileOutputStream(overall_save_file);
			DataOutputStream overall_save_writer = new DataOutputStream(overall_save_to);
			overall_save_writer.writeBytes("<<<<<< OVERALL FEATURE & CLASSIFIER AVERAGE WEIGHTING AND STDEV FOR ALL CLASSIFIER ENSEMBLES ACCROSS " + one_dim_feature_weightings_accross_folds.length + " FOLDS >>>>>>\n");
			overall_save_writer.writeBytes("\n\n\n");
			overall_save_writer.writeBytes("------------------------------\n" +
										   "FEATURE GROUP COMBINED SCORES:\n" +
										   "------------------------------\n\n");
			for (int group = 0; group < feature_group_names.length; group++)
			{
				overall_save_writer.writeBytes( bodhidharma.utilities.StringMethods.getBeginningOfStringWithHyphenFiller(feature_group_names[group], 21) +
												bodhidharma.utilities.StringMethods.getRoundedDouble(combined_group_scores_accross_ensembles[group], 4) +
												" (stdev: " + bodhidharma.utilities.StringMethods.getRoundedDouble(combined_stdev_group_scores_accross_ensembles[group], 4) + ")\n" );
			}
			overall_save_writer.writeBytes("\n\n\n");
			overall_save_writer.writeBytes("------------------------------------------------------\n" +
										   "RANKED COMBINED WEIGHTINGS (1-D AND MULTI-D FEATURES):\n" +
										   "------------------------------------------------------\n\n");
			for (int feat = 0; feat < ranked_feature_codes_accross_ensembles.length; feat++)
			{
				overall_save_writer.writeBytes( (feat + 1) + "." + "\t" +
												ranked_feature_codes_accross_ensembles[feat] + "\t" +
												bodhidharma.utilities.StringMethods.getBeginningOfStringWithHyphenFiller(ranked_feature_names_accross_ensembles[feat], 60) +
												bodhidharma.utilities.StringMethods.getRoundedDouble(ranked_weightings_accross_ensembles[feat], 5) +
												" (stdev: " + bodhidharma.utilities.StringMethods.getRoundedDouble(stdev_of_ranked_weightings_accross_ensembles[feat], 5) + ")\n" );
			}
			overall_save_writer.close();
		}
	}


	// Return how often the recordings were correctly classified as a descendant
	// of the correct root category (i.e. how often is the root category right). 
	private double getRootSuccessRate( double[][] combined_classifier_results,
	                                   String[][] model_classifications, 
	                                   String[] classifier_categories,
									   ClassificationResultsInterpereter results_interpereter )
		throws Exception
	{
		boolean[][] winners = results_interpereter.getWinners(combined_classifier_results);
		String[] root_categories = parent_window.taxonomy_panel.getParentLabels();
		double[] success_rates = new double[winners.length];
		for (int rec = 0; rec < winners.length; rec++)
		{
			// Find the correct ancestors
			boolean[] correct_ancestors = new boolean[root_categories.length];
			for (int anc = 0; anc < correct_ancestors.length; anc++)
				correct_ancestors[anc] = false;
			int number_correct_ancestors = 0;
			for (int leaf = 0; leaf < model_classifications[rec].length; leaf++)
			{
				boolean[] ancestors = getAncestorsLeafBelongsTo(root_categories, model_classifications[rec][leaf]);
				for (int i = 0; i < ancestors.length; i++)
					if (ancestors[i])
					{
						if (!correct_ancestors[i])
							number_correct_ancestors++;
						correct_ancestors[i] = true;
					}
			}

			// Find the classified ancestors
			boolean[] found_ancestors = new boolean[root_categories.length];
			for (int anc = 0; anc < found_ancestors.length; anc++)
				found_ancestors[anc] = false;
			int number_found_ancestors = 0;
			for (int leaf = 0; leaf < winners[rec].length; leaf++)
			{
				if (winners[rec][leaf])
				{
					boolean[] ancestors = getAncestorsLeafBelongsTo(root_categories, classifier_categories[leaf]);
					for (int i = 0; i < ancestors.length; i++)
						if (ancestors[i])
						{
							if (!found_ancestors[i])
								number_found_ancestors++;
							found_ancestors[i] = true;
						}
				}
			}

			// Penalize for false positives
			for (int anc = 0; anc < found_ancestors.length; anc++)
			{
				if (found_ancestors[anc] && !correct_ancestors[anc])
					number_found_ancestors--;
			}

			// Store success rate
			success_rates[rec] = (double) number_found_ancestors / (double) number_correct_ancestors;
		}

		// Return the success rate for the root
		return 100.0 * bodhidharma.utilities.Statistics.getAverage(success_rates);
	}


	// Display the stored classification results for the ensemble specified in classification_results_selector.
	// These results are for the recordings stored in the testing_recordings field.
	// Displays an error dialog if a problem occurs.
	private void displaySelectedClassificationResultsForAnEnsemble()
	{
		try
		{
			// Clear the display
			results_display.setText("");

			// Find the correct indice of the ensemble
			int ensemble_indice = classification_results_selector.getSelectedIndex();
				
			// If no ensemble is available
			if (ensemble_indice == -1)
				throw new Exception("No classification results are currently available.");
				
			// Display the heading
			results_display.append("<<<<<< CLASSIFICATION RESULTS FOR " +
								   classifier_ensemble[ensemble_indice].getClassifierName() +
								   " >>>>>>\n\n\n\n");

			// Display the results of stuff
			String[][] model_classifications = null;
			if (model_testing_classifications != null)
				model_classifications = model_testing_classifications[ensemble_indice];
			displayClassificationResultsForAnEnsemble( combined_classifier_results_all_ensembles[ensemble_indice],
													   individual_classifier_results_all_ensembles[ensemble_indice],
													   classifier_ensemble[ensemble_indice],
													   model_classifications,
													   testing_recordings );
			// Start at top of display
			results_display.setCaretPosition(0);
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}
 

	/**
	 * Displays classification results for stored combined classification results for all of the ensembles in
	 * classification_results_selector in the display area. Which results are reported depends on the user's
	 * report settings. These results are for the recordings stored in the testing_recordings field.
	 *
	 * <p>It should be noted that the scores reported here are rounded for display purposes only. Scores vary
	 * from 0 to 1, with 1 meaning certainty that a certain category is correct and 0 meaning certainty that
	 * it is not.
	 *
	 * <p>Displays a message if a problem occurs
	 */
	private void displayCombinedClassificationResults()
	{
		try
		{
			// Clear the display
			results_display.setText("");

			// Display the heading
			results_display.append("<<<<<< COMBINED RESULTS OF ALL ENSEMBLES OF CLASSIFIERS >>>>>>\n\n");

			// The source of prefence values on what to report
			PreferencesPanel preferences_panel = parent_window.preferences_panel;

			// Used to interperet results
			ClassificationResultsInterpereter results_interpereter = new ClassificationResultsInterpereter(preferences_panel);
			
			// Some useful constants
			int tab_size = 15;
			int many_column_max_decimal_places = 5;
			int one_column_max_decimal_places = 10;

			// Set the tab size
			results_display.setTabSize(tab_size);
			results_display.setFont( new Font( "Monospaced", Font.PLAIN, 12 ) );

			// Find the scores of the combined classification
			double[][] combined_classifier_results = combined_classifier_results_all_ensembles[combined_classifier_results_all_ensembles.length - 1];

			// Find the categories into which the classifiers classified test rerordings
			String[] classifier_categories = candidate_categories_for_combined_classification;

			// Find the model classifications
			String[][] model_classifications = model_testing_classifications_for_combined_classifier;

			// Construct the list of recording names
			String[] recording_names = new String[testing_recordings.length];
			for (int rec = 0; rec < recording_names.length; rec++)
				recording_names[rec] = testing_recordings[rec].title;

			// Used to store table titles
			String table_title = new String("");

			// Show success rates for combined classification
			if (preferences_panel.shouldDisplaySuccessRatesForClassifiers())
			{
				double[] combined_stats = results_interpereter.getSuccessRate( combined_classifier_results,
																			   model_classifications,
																			   classifier_categories );
				if (combined_stats != null) // in case no model results provided
				{
					table_title = new String("SUCCESS RATES:");
					
					String[] row_names = new String[1];
					row_names[0] = "Combined Classifiers";
					
					String[] column_names = new String[3];
					column_names[0] = "First Choice";
					column_names[1] = "Second Choice";
					column_names[2] = "Found Too Many";
					
					double[][] results = new double[1][];
					results[0] = combined_stats;
					printTable(table_title, column_names, row_names, results, tab_size, many_column_max_decimal_places);
				}
			}

			// Show the success rates of the root categories
			if (preferences_panel.shouldDisplayRootCategorySuccessRates() && model_classifications != null)
			{
				double root_success_rate = getRootSuccessRate( combined_classifier_results,
	                                                           model_classifications, 
	                                                           classifier_categories,
									                           results_interpereter );
				appendTableTitle("Root Category Success Rate: " + ((float) root_success_rate) + "%");
				results_display.append("\n\n");
			}

			// Show confusion matrix and percent of errors that were classified as unknown
			if (preferences_panel.shouldDisplayConfusionMatrix() && model_classifications != null)
			{
				table_title = new String("CONFUSION MATRIX:");
				double[][] confusion_matrix = results_interpereter.getConfusionMatrix( combined_classifier_results,
				                                                                       model_classifications,
				                                                                       classifier_categories );
				String[] row_names = new String[classifier_categories.length + 1];			
				for (int i = 0; i < row_names.length; i++)
				{
					if (i == classifier_categories.length)
						row_names[i] = "Unknown";
					else
						row_names[i] = classifier_categories[i];
				}
				printTable( table_title, classifier_categories, row_names,
	                        confusion_matrix, tab_size, many_column_max_decimal_places );

				// Display percentage of wrong guesses that are unknown
				displatyUnknownWrongPercentage(confusion_matrix);
			}

			// Display classification settings, if requested
			if (preferences_panel.shouldDisplayClassifierSettings())
				displayPreferenceSettings();

			// Show the order of the categories used by the classifier
			if (preferences_panel.shouldDisplayListOfElligibleCategories())
			{
				table_title = new String("ELLIGIBLE CATEGORY LIST USED BY CLASSIFIERS:");
				appendTableTitle(table_title);
				for (int cat = 0; cat < classifier_categories.length; cat++)
					results_display.append("Category " + cat + ":\t" + classifier_categories[cat] + "\n");
				results_display.append("\n\n");
			}

			// Show the results in list form
			if (preferences_panel.shouldDisplayCombinedClassificationResultsForEachRecording())
			{
				String[][] temp_model_classifications;
				if (preferences_panel.shouldDisplayModelResultsAlongWithResults())
					temp_model_classifications = model_classifications;
				else
					temp_model_classifications = null;
				String[] combined_labels_of_results = results_interpereter.getWinnerLabels( combined_classifier_results,
																							classifier_categories,
																							temp_model_classifications, 
																							preferences_panel.reportScoresWithWinningCategories(),
																							preferences_panel.reportSecondaryWinningCategories() );
				table_title = new String("CLASSIFICATION RESULTS FOR EACH RECORDING:");
				printTable(table_title, "Classification Results", recording_names, combined_labels_of_results, ((int) (tab_size * 1.5)) );
			}

			// Show the combined results in table form (scores for each category)
			if (preferences_panel.shouldDisplayCombinedClassificationScoresForEachRecording())
			{
				table_title = new String("CLASSIFICATION SCORES FOR EACH CATEGORY OF EACH RECORDING:");
				printTable( table_title, classifier_categories, recording_names, combined_classifier_results,
							tab_size, many_column_max_decimal_places );
			}

			// Show the model results in list form
			if (preferences_panel.shouldDisplayModelResultsForEeachRecordingInSeparateTable() && model_classifications != null)
			{
				table_title = new String("MODEL RESULTS FOR EACH RECORDING:");
				printTable(table_title, "Model Results", recording_names, model_classifications, ((int) (tab_size * 1.5)) );
			}

			// Start at top of display
			results_display.setCaretPosition(0);
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}


	/**
	 * Displays averages of results after multi-fold training and testing.
	 *
	 * <p>It should be noted that the scores reported here are rounded for display purposes only. Scores vary
	 * from 0 to 1, with 1 meaning certainty that a certain category is correct and 0 meaning certainty that
	 * it is not.
	 *
	 * <p>Displays a message if a problem occurs
	 */
	private void displayResultsAveragedOverFolds()
	{
		try
		{
			// Clear the display
			results_display.setText("");

			// Display the heading
			results_display.append("<<<<<< AVERAGED RESULTS OVER ALL FOLDS >>>>>>\n\n");

			// The source of prefence values on what to report
			PreferencesPanel preferences_panel = parent_window.preferences_panel;

			// The number of folds
			int number_of_folds = multi_fold_combined_classifier_results_all_ensembles.length;

			// Used to interperet results
			ClassificationResultsInterpereter results_interpereter = new ClassificationResultsInterpereter(preferences_panel);
			
			// Some useful constants
			int tab_size = 15;
			int many_column_max_decimal_places = 5;
			int one_column_max_decimal_places = 10;
			String table_title = new String("");

			// Set the tab size
			results_display.setTabSize(tab_size);
			results_display.setFont( new Font( "Monospaced", Font.PLAIN, 12 ) );

			// Find the scores of the combined classification for each fold
			double[][][] multi_fold_combined_classifier_results = new double[number_of_folds][][];
			for (int fold = 0; fold < number_of_folds; fold++)
				multi_fold_combined_classifier_results[fold] = multi_fold_combined_classifier_results_all_ensembles[fold][multi_fold_combined_classifier_results_all_ensembles[fold].length - 1];

			// Display statistics of rates over all folds
			if (preferences_panel.shouldDisplaySuccessRatesForClassifiers() && multi_fold_candidate_categories_for_combined_classification[0] != null)
			{
				// Find the success rates for each fold
				double[][] success_rates = new double[number_of_folds][];
				for (int fold = 0; fold < success_rates.length; fold++)
					success_rates[fold] = results_interpereter.getSuccessRate( multi_fold_combined_classifier_results[fold], 
																			   multi_fold_model_testing_classifications_for_combined_classifier[fold],
																			   multi_fold_candidate_categories_for_combined_classification[fold] );

				// Find the average and standard deviation stats over all folds
				double[] first_guesses = new double[number_of_folds];
				double[] second_guesses = new double[number_of_folds];
				double[] extra_choices = new double[number_of_folds];
				for (int fold = 0; fold < success_rates.length; fold++)
				{
					first_guesses[fold] = success_rates[fold][0];
					second_guesses[fold] = success_rates[fold][1];
					extra_choices[fold] = success_rates[fold][2];
				}
				double first_guess_average = bodhidharma.utilities.Statistics.getAverage(first_guesses);
				double first_guess_STD = bodhidharma.utilities.Statistics.getStandardDeviation(first_guesses);
				double second_guess_average = bodhidharma.utilities.Statistics.getAverage(second_guesses);
				double second_guess_STD = bodhidharma.utilities.Statistics.getStandardDeviation(second_guesses);
				double extra_choice_average = bodhidharma.utilities.Statistics.getAverage(extra_choices);
				double extra_choice_STD = bodhidharma.utilities.Statistics.getStandardDeviation(extra_choices);
				double[][] rate_averages = new double[1][3];
				rate_averages[0][0] = first_guess_average;
				rate_averages[0][1] = second_guess_average;
				rate_averages[0][2] = extra_choice_average;
				double[][] rate_STDs = new double[1][3];
				rate_STDs[0][0] = first_guess_STD;
				rate_STDs[0][1] = second_guess_STD;
				rate_STDs[0][2] = extra_choice_STD;

				// Prepare tables
				String[] row_names = new String[1];
				row_names[0] = "Combined Classifiers";
				String[] column_names = new String[3];
				column_names[0] = "First Choice";
				column_names[1] = "Second Choice";
				column_names[2] = "Found Too Many";
				
				// Display averages of stats
				table_title = new String("AVERAGE SUCCESS RATES ACCROSS FOLDS:");
				printTable(table_title, column_names, row_names, rate_averages, tab_size, many_column_max_decimal_places);

				// Display standard deviatios of stats
				table_title = new String("STANDARD DEVIATIONS OF SUCCESS RATES ACCROSS FOLDS:");
				printTable(table_title, column_names, row_names, rate_STDs, tab_size, many_column_max_decimal_places);
			}

			// Show training times
			double average_training_time = bodhidharma.utilities.Statistics.getAverage(mult_fold_training_times);
			double STD_training_time = bodhidharma.utilities.Statistics.getStandardDeviation(mult_fold_training_times);
			appendTableTitle("TRAINING TIMES ACCROSS FOLDS:");
			results_display.append("Average:\t\t" + average_training_time + " seconds\n");
			results_display.append("Standard Deviation:\t" + STD_training_time + " seconds\n\n\n\n");

			// Show the success rates of the root categories
			if (preferences_panel.shouldDisplayRootCategorySuccessRates() && multi_fold_model_testing_classifications_for_combined_classifier[0] != null)
			{
				double[] root_success_rate = new double[number_of_folds];
				
				// Go through one fold at a time
				for (int fold = 0; fold < root_success_rate.length; fold++)
				{
					double[][] combined_classifier_results = multi_fold_combined_classifier_results[fold];
					String[][] model_classifications = multi_fold_model_testing_classifications_for_combined_classifier[fold];
					String[] classifier_categories = multi_fold_candidate_categories_for_combined_classification[fold];

					root_success_rate[fold] = getRootSuccessRate( combined_classifier_results,
	                                                              model_classifications, 
	                                                              classifier_categories,
									                              results_interpereter );
				}

				// Display the table
				float average_root_success = (float) bodhidharma.utilities.Statistics.getAverage(root_success_rate);
				float STD_root = (float) bodhidharma.utilities.Statistics.getStandardDeviation(root_success_rate);
				appendTableTitle("CORRECTLY CHOSEN ROOT SUCCESS RATES:");
				results_display.append("Average:\t\t" + average_root_success + "%\n");
				results_display.append("Standard Deviation:\t" + STD_root + "%\n\n\n\n");
			}

			// Show confusion matrix and percent of those wrong classified as unknown
			if (preferences_panel.shouldDisplayConfusionMatrix() && multi_fold_candidate_categories_for_combined_classification[0] != null)
			{
				// Find the confusion matrix for each fold
				double[][][] confusion_matrices = new double[number_of_folds][][];
				for (int fold = 0; fold < confusion_matrices.length; fold++)
					confusion_matrices[fold] = results_interpereter.getConfusionMatrix( multi_fold_combined_classifier_results[fold], 
																		                multi_fold_model_testing_classifications_for_combined_classifier[fold],
																		                multi_fold_candidate_categories_for_combined_classification[fold] );

				// Find the average and standard deviation confusion matrix over all the folds
				double[][] averages_confusion_matrix = new double[confusion_matrices[0].length][confusion_matrices[0][0].length];
				double[][] STDs_confusion_matrix = new double[confusion_matrices[0].length][confusion_matrices[0][0].length];
				for (int row = 0; row < averages_confusion_matrix.length; row++)
				{
					for (int col = 0; col < averages_confusion_matrix[row].length; col++)
					{
						double[] values_accross_folds = new double[number_of_folds];
						for (int fold = 0; fold < values_accross_folds.length; fold++)
							values_accross_folds[fold] = confusion_matrices[fold][row][col];
						averages_confusion_matrix[row][col] = bodhidharma.utilities.Statistics.getAverage(values_accross_folds);
						STDs_confusion_matrix[row][col] = bodhidharma.utilities.Statistics.getStandardDeviation(values_accross_folds);
					}
				}

				// Prepare tables
				String[] row_names = new String[multi_fold_candidate_categories_for_combined_classification[0].length + 1];			
				for (int i = 0; i < row_names.length; i++)
				{
					if (i == multi_fold_candidate_categories_for_combined_classification[0].length)
						row_names[i] = "Unknown";
					else
						row_names[i] = multi_fold_candidate_categories_for_combined_classification[0][i];
				}
				
				// Display averages of stats
				table_title = new String("AVERAGE CONFUSION MATRIX ACCROSS FOLDS:");
				printTable( table_title, multi_fold_candidate_categories_for_combined_classification[0], row_names,
							averages_confusion_matrix, tab_size, many_column_max_decimal_places );

				// Display percentage of wrong guesses that are unknown
				displatyUnknownWrongPercentage(averages_confusion_matrix);

				// Display standard deviatios of stats
				table_title = new String("STANDARD DEVIATIONS OF CONFUSION MATRIX VALUES ACCROSS FOLDS:");
				printTable( table_title, multi_fold_candidate_categories_for_combined_classification[0], row_names,
							STDs_confusion_matrix, tab_size, many_column_max_decimal_places );
			}

			// Display classification settings, if requested
			if (preferences_panel.shouldDisplayClassifierSettings())
				displayPreferenceSettings();

			// Start at top of display
			results_display.setCaretPosition(0);
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}


	// Append the preferences have to do with training and testing to the results_display
	private void displayPreferenceSettings()
		throws Exception
	{
		// The source of prefence values on what to report
		PreferencesPanel preferences_panel = parent_window.preferences_panel;

		// Display the training options
		appendTableTitle("TRAINING OPTIONS");
		results_display.append("Use One-D Features:\t" + preferences_panel.shouldUseOneDimensionalFeaturesInTraining() + "\n");
		results_display.append("Use Multi-D Features:\t" + preferences_panel.shouldUseMultiDimensionalFeaturesInTraining() + "\n");
		results_display.append("Perform Feature Selection:\t" + preferences_panel.shouldBioKNNsUseFeatureSelection() + "\n");
		results_display.append("Perform Feature Weighting:\t" + preferences_panel.shouldBioKNNsUseFeatureWeighting() + "\n");
		results_display.append("Perform Classifier Selection:\t" + preferences_panel.shouldSelectClassifiersInEachEnsemble() + "\n");
		results_display.append("Perform Classifier Weighting:\t" + preferences_panel.shouldWeightClassifiersInEachEnsemble() + "\n");
		int number_of_folds = 1 + new Long(Math.round(preferences_panel.getTrainingFraction() / (1.0 - preferences_panel.getTrainingFraction()))).intValue();
		results_display.append("Training Fraction:\t" + preferences_panel.getTrainingFraction() + " (" + number_of_folds + " folds)\n");
		results_display.append("Maximum GA Epochs:\t" + Math.abs(preferences_panel.getNumberGAIterations()));
		if (preferences_panel.getNumberGAIterations() < 0)
			results_display.append(" (early ending permitted)\n");
		else
			results_display.append(" (early ending not permitted)\n");
		results_display.append("Maximum NN Epochs:\t" + Math.abs(preferences_panel.getNumberNNIterations()));
		if (preferences_panel.getNumberNNIterations() < 0)
			results_display.append(" (early ending permitted)\n");
		else
			results_display.append(" (early ending not permitted)\n");
		results_display.append("\n\n");

		// Calculate absolute weightings from relative weightings
		boolean should_classify_hierarchically =  preferences_panel.shouldClassifyIntoDirectDescendantsOfParentCategories();
		boolean should_classify_into_root_categories =  preferences_panel.shouldClassifyIntoRootCategories();
		boolean should_classify_into_all_leaf_categories =  preferences_panel.shouldClassifyIntoAllLeafCategories();
		boolean should_classify_using_round_robin =  preferences_panel.shouldClassifyUsingRoundRobin();
		double hierarchal_weighting = preferences_panel.getHierarchalRelativeWeighing();
		double leaf_weighting = preferences_panel.getAllLeafRelativeWeighing();
		double round_robin_weighting = preferences_panel.getRoundRobinRelativeWeighing();
		if (!should_classify_into_root_categories && !should_classify_hierarchically)
			hierarchal_weighting = 0.0;
		if (!should_classify_into_all_leaf_categories)
			leaf_weighting = 0.0;
		if (!should_classify_using_round_robin)
			round_robin_weighting = 0.0;
		double relative_weighting_total = hierarchal_weighting + leaf_weighting + round_robin_weighting;
		hierarchal_weighting = hierarchal_weighting / relative_weighting_total;
		leaf_weighting = leaf_weighting / relative_weighting_total;
		round_robin_weighting = round_robin_weighting / relative_weighting_total;

		// Display the classifiers used and their absolute weightings
		appendTableTitle("CLASSIFIERS USED AND THEIR WEIGHTINGS");
		if (should_classify_hierarchically)
			results_display.append("Hierarchal Classifier:\t" + hierarchal_weighting + "\n");
		if (should_classify_into_root_categories && !should_classify_hierarchically)
			results_display.append("Root Classifier:\t" + hierarchal_weighting + "\n");
		if (should_classify_into_all_leaf_categories)
			results_display.append("Flat Leaf Classifier:\t" + leaf_weighting + "\n");
		if (should_classify_using_round_robin)
			results_display.append("Round Robin Classifier:\t" + round_robin_weighting + "\n");
		if (should_classify_using_round_robin)
		{
			if (preferences_panel.shouldOnlyUseBestForRoundRobin())
				results_display.append("NOTE: Only single winners were produced by round robin classification.\n");
			if (preferences_panel.getRoundRobinDivisor() != 1.0)
				results_display.append("NOTE: Round robin classifier scores were divided by " + preferences_panel.getRoundRobinDivisor() + " during classifier coordination.\n");
		}
		results_display.append("\n\n");

		// Display the winner selection rules
		appendTableTitle("RULES USED TO SELECT WINNERS");
		results_display.append("- All scores above " + preferences_panel.getMinimumAutomaticQualificationScore() + " give an automatic win.\n");
		results_display.append("- Only categories with scores of at least " + preferences_panel.getMinimumScoreToQualify() + " can be winners.\n");
		results_display.append("- Categories with scores of at least " + preferences_panel.getPercentageBelowHighestToQualify() + "% of highest can be winners.\n");
		if (preferences_panel.shouldOnlyPickOneWinnerForCombinedClassification())
			results_display.append("- Only the best performing category for each recording chosen as winner.\n\n\n");
		else
			results_display.append("- Multiple categories may be chosen as winner for each recording.\n\n\n");
	}


	// Display percentage of wrong guesses that were classified as unknown
	// The confusion_matrix parameter gives the percentages right for each
	// category. The rows of the parameter correspond to how the recordings
	// were classified (with one extra at the bottom for unknown) and the
	// columns indicate how they should have been classified.
	//
	// Is normalized accross categories (i.e. unknown percentage for every
	// category is treated equally, even if some categories have more recordings
	// than others.
	private void displatyUnknownWrongPercentage(double[][] confusion_matrix)
	{
		// Find the percentage of those categories that were categorized 
		// wrongly that were set to Unknown and output it
		double wrong = 0.0;
		double unknown = 0.0;
		for (int col = 0; col < confusion_matrix[0].length; col++)
		{
			for (int row = 0; row < confusion_matrix.length; row++)
				if (col != row)
				{
					if (row != col)
						wrong += confusion_matrix[row][col];
					if (row == (confusion_matrix.length - 1))
						unknown += confusion_matrix[row][col];
				}
		}

		// Display overall unknown error percentage
		float unknown_average = (float) (100.0 * (unknown / wrong));
		appendTableTitle("NORM % OF WRONG CLASSIFICATIONS THAT WERE ASSIGNED UNKNOWN: " + unknown_average + "%");
		results_display.append("\n\n");
	}


	// Display the stored training errors for the ensemble specified in training_errors_ensemble_selector.
	// Displays an error dialog if a problem occurs. Also can didsplay all training times together.
	private void displaySelectedTrainingErrors()
	{
		try
		{
			// Clear the display
			results_display.setText("");
			
			// Find the correct indice of the ensemble
			int ensemble_indice = training_errors_ensemble_selector.getSelectedIndex();
			
			// If no ensemble is available
			if (ensemble_indice == -1)
				throw new Exception("No training reports are currently available.");
			
			// Display all training times
			if (((String) training_errors_ensemble_selector.getSelectedItem()).equals("All Times"))
				results_display.setText(stored_training_times_report);

			// Display training stats for an ensemble
			else
			{
				// Display the heading
				results_display.append("<<<<<< TRAINING ERROR HISTORY FOR " +
									   classifier_ensemble[ensemble_indice].getClassifierName() +
									   " >>>>>>\n\n\n\n");

				// Display the training data
				displayTrainingStatisticsForAnEnsemble( combined_training_errors[ensemble_indice],
														training_recordings_info[ensemble_indice],
														classifier_ensemble[ensemble_indice] );
			}

			// Start at top of display
			results_display.setCaretPosition(0);
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}


	// Displays feature and classifier selection and weighting information.
	// Displays an error dialog if a problem occurs.
	private void displaySelectedWeightingsAndSelections()
	{
		try
		{
			// Clear the display
			results_display.setText("");
			
			// Find the correct indice of the ensemble
			int ensemble_indice = selection_weighting_selector.getSelectedIndex();

			// If no ensemble is available
			if (ensemble_indice == -1)
				throw new Exception("No selections or weighings are currently available.");
			
			// Display the heading
			results_display.append("<<<<<< FEATURE & CLASSIFIER SELECTION & WEIGHTING FOR " +
			                       classifier_ensemble[ensemble_indice].getClassifierName() +
			                       " >>>>>>\n\n\n\n");

			// Display the training data
			displaySelectionWeightingForAnEnsemble( classifier_ensemble[ensemble_indice] );

			// Start at top of display
			results_display.setCaretPosition(0);
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null, e.getMessage(), "ERROR", JOptionPane.ERROR_MESSAGE);
			// e.printStackTrace();
		}
	}


	/**
	 * Appends classification results for the given classifier ensemble to the display area. Which results are
	 * reported depends on the user's report settings. The following information can be reported by this method:
	 *
	 * <p>- The candidate categories and their associated number codes.
	 * <p>- The categories that the combined classifiers claim each recording belongs to. This can or cannot
	 * include associated classification scores, second choices and model results, depending on the user's
	 * preferences.
	 * <p>- The above information for each classifier in the ensemble.
	 * <p>- A table of classifier scores for each candidate category for each recording, as reported by the
	 * combined ensemble.
	 * <p>- The above information for each classifier in the ensemble.
	 *
	 * <p>It should be noted that the scores reported here are rounded for display purposes only. Scores vary
	 * from 0 to 1, with 1 meaning certainty that a certain category is correct and 0 meaning certainty that
	 * it is not.
	 *
	 * <p>This method may also be used if model classification are not present (pass null to the parameter).
	 *
	 * @param	combined_classifier_results		The combined classification results for all of the classifiers in the <i>ensemble_of_clasifiers</i> after classifier selection and weighting have been applied. Results are in the form of relative scores for each of the possible categories for each recording (not normalized). Scores should fall in the range between 0.0 and 1.0 and higher scores correspond to a greater certainty that a recording has a corresponding label. The first indice corresponds to recording and the second corresponds to the category.
	 * @param	individual_classifier_results	The individual classification results for each of the classifiers in the <i>ensemble_of_clasifiers</i> (before classifier selection and weighting been applied). Results are in the form of relative scores for each of the possible categories for each recording (not normalized) for each classifier. Scores should fall in the range between 0.0 and 1.0 and higher scores correspond to a greater certainty that a recording has a corresponding label. The first indice specifies the classifier, the second specifies the recording and the third specifies the category.
	 * @param	ensemble_of_clasifiers			The <code>NN_BioKNN_Ensemble</code> used for training and classification.
	 * @param	model_classifications			The correct category(ies) for each of the recordings. The indice identifies the recording that the category(ies) correspond to. A value of null indicates that model results should not be printed.
	 * @param	classified_recordings			The <code>Recordings<code> that were classified by the <i>ensemble_of_clasifiers</i>.
	 */
	private void displayClassificationResultsForAnEnsemble( double[][] combined_classifier_results,
	                                                        double[][][] individual_classifier_results,
	                                                        NN_BioKNN_Ensemble ensemble_of_clasifiers,
	                                                        String[][] model_classifications,
	                                                        Recording[] classified_recordings )
		throws Exception
	{
		// The source of prefence values on what to report
		PreferencesPanel preferences_panel = parent_window.preferences_panel;

		// Used to interperet results
		ClassificationResultsInterpereter results_interpereter = new ClassificationResultsInterpereter(preferences_panel);
		
		// Some useful constants
		int tab_size = 15;
		int many_column_max_decimal_places = 5;
		int one_column_max_decimal_places = 10;

		// Set the tab size
		results_display.setTabSize(tab_size);
		results_display.setFont( new Font( "Monospaced", Font.PLAIN, 12 ) );
		
		// Find the categories into which the classifiers classified test rerordings
		String[] classifier_categories = ensemble_of_clasifiers.getCategories();

		// Find the names of the classifiers stored in the classifier
		String[] classifier_names = ensemble_of_clasifiers.getClassifierNames();

		// Construct the list of recording names
		String[] recording_names = new String[classified_recordings.length];
		for (int rec = 0; rec < classified_recordings.length; rec++)
			recording_names[rec] = classified_recordings[rec].title;

		// Used to store table titles
		String table_title = new String("");

		// Show success rates for combined classification as well as individual classifiers
		if (preferences_panel.shouldDisplaySuccessRatesForClassifiers())
		{
			double[] combined_stats = results_interpereter.getSuccessRate(combined_classifier_results, model_classifications, classifier_categories);
			if (combined_stats != null) // in case no model results provided
			{
				table_title = new String("SUCCESS RATES:");
				
				String[] row_names = new String[classifier_names.length + 1];
				row_names[0] = "Combined Classifiers";
				for (int i = 0; i < classifier_names.length; i++)
					row_names[i + 1] = classifier_names[i];
				
				String[] column_names = new String[3];
				column_names[0] = "First Choice";
				column_names[1] = "Second Choice";
				column_names[2] = "Found Too Many";
				
				double[][] results = new double[row_names.length][column_names.length];
				results[0] = combined_stats;
				for (int clas = 0; clas < individual_classifier_results.length; clas++)
				{
					if (individual_classifier_results[clas] != null)
						results[clas + 1] = results_interpereter.getSuccessRate(individual_classifier_results[clas], model_classifications, classifier_categories);
					else
					{
						results[clas + 1][0] = 0.0; 
						results[clas + 1][1] = 0.0; 
						results[clas + 1][2] = 0.0; 
					}
				}
				printTable(table_title, column_names, row_names, results, tab_size, many_column_max_decimal_places);
			}
		}

		// Show confusion matrix
		if (preferences_panel.shouldDisplayConfusionMatrix() && model_classifications != null)
		{
			table_title = new String("CONFUSION MATRIX:");
			double[][] confusion_matrix = results_interpereter.getConfusionMatrix( combined_classifier_results,
																				   model_classifications,
																				   classifier_categories );
			String[] row_names = new String[classifier_categories.length + 1];			
			for (int i = 0; i < row_names.length; i++)
			{
				if (i == classifier_categories.length)
					row_names[i] = "Unknown";
				else
					row_names[i] = classifier_categories[i];
			}
			printTable( table_title, classifier_categories, row_names,
						confusion_matrix, tab_size, many_column_max_decimal_places );
		}

		// Show the order of the categories used by the classifier
		if (preferences_panel.shouldDisplayListOfElligibleCategories())
		{
			table_title = new String("ELLIGIBLE CATEGORY LIST USED BY CLASSIFIERS:");
			appendTableTitle(table_title);
			for (int cat = 0; cat < classifier_categories.length; cat++)
				results_display.append("Category " + cat + ":\t" + classifier_categories[cat] + "\n");
			results_display.append("\n\n");
		}

		// Show the combined results in list form
		if (preferences_panel.shouldDisplayCombinedClassificationResultsForEachRecording())
		{
			String[][] temp_model_classifications;
			if (preferences_panel.shouldDisplayModelResultsAlongWithResults())
				temp_model_classifications = model_classifications;
			else
				temp_model_classifications = null;
			String[] combined_labels_of_results = results_interpereter.getWinnerLabels( combined_classifier_results,
																						classifier_categories,
																					    temp_model_classifications, 
																					    preferences_panel.reportScoresWithWinningCategories(),
																					    preferences_panel.reportSecondaryWinningCategories() );
			table_title = new String("COMBINED CLASSIFICATION RESULTS FOR EACH RECORDING:");
			printTable(table_title, "Classification Results", recording_names, combined_labels_of_results, ((int) (tab_size * 1.5)) );
		}

		// Show the results in list form for each classifier
		if (preferences_panel.shouldDisplayIndividualClassifierResultsForEachRecording())
		{
			String[][] temp_model_classifications;
			if (preferences_panel.shouldDisplayModelResultsAlongWithResults())
				temp_model_classifications = model_classifications;
			else
				temp_model_classifications = null;
			for (int clas = 0; clas < individual_classifier_results.length; clas++)
			{
				if (individual_classifier_results[clas] != null)
				{
					String[] individual_classifier_labels_of_results = results_interpereter.getWinnerLabels( individual_classifier_results[clas],
																											 classifier_categories,
																											 temp_model_classifications, 
																											 preferences_panel.reportScoresWithWinningCategories(),
																											 preferences_panel.reportSecondaryWinningCategories() );
					table_title = new String("CLASSIFICATION RESULTS FOR " + classifier_names[clas] + " CLASSIFIER:");
					printTable(table_title, "Classification Results", recording_names, individual_classifier_labels_of_results, ((int) (tab_size * 1.5)) );
				}
				else results_display.append("\n");
			}
		}

		// Show the combined results in table form (scores for each category)
		if (preferences_panel.shouldDisplayCombinedClassificationScoresForEachRecording())
		{
			table_title = new String("COMBINED CLASSIFICATION SCORES FOR EACH CATEGORY OF EACH RECORDING:");
			printTable( table_title, classifier_categories, recording_names, combined_classifier_results,
						tab_size, many_column_max_decimal_places );
		}

		// Show the individual classifier results in table form (scores for each category)
		if (preferences_panel.shouldDisplayIndividualClassifierScoresForEachRecording())
		{
			for (int clas = 0; clas < individual_classifier_results.length; clas++)
			{
				table_title = new String("CLASSIFICATION SCORES FOR " + classifier_names[clas] + " CLASSIFIER:");
				if (individual_classifier_results[clas] != null)
					printTable( table_title, classifier_categories, recording_names, individual_classifier_results[clas],
								tab_size, many_column_max_decimal_places );
				else results_display.append("\n");
			}
		}

		// Show the model results in list form
		if (preferences_panel.shouldDisplayModelResultsForEeachRecordingInSeparateTable() && model_classifications != null)
		{
			table_title = new String("MODEL RESULTS FOR EACH RECORDING:");
			printTable(table_title, "Model Results", recording_names, model_classifications, ((int) (tab_size * 1.5)) );
		}
	}


	/**
	 * Appends training statistics for the given classifier ensemble to the display area. Which results are
	 * reported depends on the user's report settings. The following information can be reported by this method:
	 *
	 * <p>- The training error for feature selection OR weighting (only weighting error is reported if both
	 * are performed. This is effectively the training for the one-dimensional feature classifier.
	 * <p>- The training error for each of the multi-dimensional classifiers.
	 * <p>- The training error for classifier selection.
	 * <p>- The training error for classifier weighting.
	 *
	 * <p>It should be noted that the errors reported here are rounded for display purposes only. Lower values
	 * indicate increasing convergence of the classifiers. The user can control how many training iterations
	 * are skipped between reported training errors.
	 *
	 * @param	training_errors					The training errors for each of the classifiers in <i>ensemble_of_clasifiers</i> and for classifier selection and weighting. In the format returned by the <code>train</code> method of the <code>NN_BioKNN_Ensemble</code> class.
	 * @param	training_recordings_details		The names and model leaf categories of all training recordings.
	 * @param	ensemble_of_clasifiers			The <code>NN_BioKNN_Ensemble</code> used for training and classification.
	 */
	private void displayTrainingStatisticsForAnEnsemble( double[][] training_errors,
	                                                     String[][] training_recordings_details,
	                                                     NN_BioKNN_Ensemble ensemble_of_clasifiers )
		throws Exception
	{
		// Some useful constants
		int tab_size = 15;
		int many_column_max_decimal_places = 5;
		int one_column_max_decimal_places = 10;

		// Set the tab size
		results_display.setTabSize(tab_size);
		results_display.setFont( new Font( "Monospaced", Font.PLAIN, 12 ) );
		
		// Find the names of the classifiers stored in the classifier
		String[] classifier_names = ensemble_of_clasifiers.getClassifierNames();

		// The source of prefence values on what to report
		PreferencesPanel preferences_panel = parent_window.preferences_panel;

		// Find the number of rows to skip when reporting errors
		int rows_to_skip = preferences_panel.getLinesToSkipWhenReportingTrainingErrors();

		// Used to store table titles
		String table_title = new String("");

		// Show the names of recordings used to train this classifier
		appendTableTitle("NAMES AND MODEL LEAF CATEGORIES OF RECORDINGS USED FOR TRAININ:");
		if (preferences_panel.shouldDisplayTrainingRecordingNames())
		{
			for (int rec = 0; rec < training_recordings_details.length; rec++)
			{
				results_display.append(training_recordings_details[rec][0] + ":  ");
				for (int cat = 1; cat < training_recordings_details[rec].length; cat++)
				{
					results_display.append(training_recordings_details[rec][cat] + " | ");
					if (cat != training_recordings_details[rec].length - 1)
						results_display.append(", ");
				}
				results_display.append("\n");
			}
			results_display.append("\n\n\n");
		}

		// Show the training error for the feature selection or weighting (both are not recorded)
		if (preferences_panel.shouldDisplayFeatureSelectionWeightingTrainingError())
		{
			double[] feature_selection_weighting_training_errors = training_errors[0];
			if (feature_selection_weighting_training_errors == null)
				results_display.append("CANNOT REPORT ONE-DIMENSIONAL FEATURE SELECTION/WEIGHTING TRAINING ERRORS (SELECTION/WEIGHTING NOT PERFORMED)\n\n\n");
			else
			{
				if (preferences_panel.shouldBioKNNsUseFeatureWeighting())
					table_title = new String("TRAINING ERRORS FOR ONE-DIMENSIONAL FEATURE WEIGHTING (" + feature_selection_weighting_training_errors.length + " ITERATIONS):");
				else
					table_title = new String("TRAINING ERRORS FOR ONE-DIMENSIONAL FEATURE SELECTION (" + feature_selection_weighting_training_errors.length + " ITERATIONS):");
				printTableSkippingRows( table_title, "Training Error", feature_selection_weighting_training_errors, rows_to_skip, tab_size, one_column_max_decimal_places );
			}
		}

		// Show the training errors for each of the multi-dimensional classifiers
		if (preferences_panel.shouldDisplayMultiDimClassifierTrainingError())
		{
			for (int clas = 1; clas < classifier_names.length; clas++)
			{
				double[] training_error = training_errors[clas];
				table_title = new String("TRAINING ERRORS FOR FOR " + classifier_names[clas] + " CLASSIFIER:");
				printTableSkippingRows( table_title, "Training Error", training_error, rows_to_skip, tab_size, one_column_max_decimal_places );
			}
		}

		// Show the training error for the classifier selection
		if (preferences_panel.shouldDisplayClassifierSelectionTrainingError())
		{
			double[] classification_selection_errors = training_errors[training_errors.length - 2];
			table_title = new String("TRAINING ERRORS FOR CLASSIFIER SELECTION (" + classification_selection_errors.length + " ITERATIONS):");
			printTableSkippingRows( table_title, "Training Error", classification_selection_errors, rows_to_skip, tab_size, one_column_max_decimal_places );
		}

		// Show the training error for the classifier weighting
		if (preferences_panel.shouldDisplayClassifierWeightingTrainingError())
		{
			double[] classification_weighting_errors = training_errors[training_errors.length - 1];
			table_title = new String("TRAINING ERRORS FOR CLASSIFIER WEIGHTING (" + classification_weighting_errors.length + " ITERATIONS):");
			printTableSkippingRows( table_title, "Training Error", classification_weighting_errors, rows_to_skip, tab_size, one_column_max_decimal_places );
		}
	}


	/**
	 * Appends the one-dimensional feature selection, one-dimensional feature weighting, classifier
	 * selection and classifier weighting used by the given ensemble of classifiers. Which results are
	 * reported depends on the user's report settings. The following information can be reported by this method:
	 *
	 * <p>The weightingss are normalized for display purposes only. The original values
	 * may not be normalized.
	 *
	 * <p>- One-dimensional feature selection
	 * <p>- One-dimensional feature weighting
	 * <p>- Classifier selection
	 * <p>- Classifier weighting
	 *
	 * @param	ensemble_of_clasifiers			The <code>NN_BioKNN_Ensemble</code> used for training and classification.
	 */
	private void displaySelectionWeightingForAnEnsemble( NN_BioKNN_Ensemble ensemble_of_clasifiers )
		throws Exception
	{
		// Some useful constants
		int tab_size = 15;
		int many_column_max_decimal_places = 5;
		int one_column_max_decimal_places = 10;

		// Set the tab size
		results_display.setTabSize(tab_size);
		results_display.setFont( new Font( "Monospaced", Font.PLAIN, 12 ) );
		
		// The source of prefence values on what to report
		PreferencesPanel preferences_panel = parent_window.preferences_panel;

		// Find the names of the one-dimensional features used by the BioKNN (null if BioKNN not used)
		String[] one_dim_features_names = ensemble_of_clasifiers.getNamesOfFeaturesUsedByOneDimensionalClassifier();

		// Find the names of the classifiers stored in the classifier
		String[] classifier_names = ensemble_of_clasifiers.getClassifierNames();

		// Used to store table titles
		String table_title = new String("");

		// Show the feature selection for the one-dimensional features
		if (preferences_panel.shouldDisplayOneDimensionalFeaturesSelected())
		{
			boolean[] one_dim_feature_selection = ensemble_of_clasifiers.getOneDimensionalFeatureSelection();
			if (one_dim_feature_selection == null)
			{
				if (ensemble_of_clasifiers.getOneDimensionalFeatureWeightings() == null)
					results_display.append("CANNOT REPORT ONE-DIMENSIONAL FEATURES USED (NO ONE-DIMENSIONAL CLASSIFIER USED)\n\n");
				else
					results_display.append("SEE FEATURE WEIGHTING FOR INFORMATION ON ONE-DIMENSIONAL FEATURES SELECTED\n\n");
			}
			else
			{
				table_title = new String("FEATURE SELECTION (ONE-DIMENSIONAL):");
				printTable( table_title, "Used", one_dim_features_names, one_dim_feature_selection, ((int) (tab_size * 1.5)) );
			}	
		}

		// Show the feature weighting for the one-dimensional features
		// These values are normalized for display purposes only. The original values
		// are not normalized.
		if (preferences_panel.shouldDisplayOneDimensionalFeatureWeightings())
		{
			double[] one_dim_feature_weightings = ensemble_of_clasifiers.getOneDimensionalFeatureWeightings();
			if (one_dim_feature_weightings == null)
				results_display.append("CANNOT REPORT ONE-DIMENSIONAL FEATURE WEIGHTINGS (NO ONE-DIMENSIONAL CLASSIFIER USED)\n\n\n");
			else
			{
				double[] norm_one_dim_feature_weightings = bodhidharma.utilities.Statistics.normalize(one_dim_feature_weightings);
				table_title = new String("FEATURE WEIGHTINGS (ONE-DIMENSIONAL):");
				printTable( table_title, "Used", one_dim_features_names, norm_one_dim_feature_weightings,
							((int) (tab_size * 1.5)), one_column_max_decimal_places );
			}	
		}

		// Show the classifier selection
		if (preferences_panel.shouldDisplayClassifiersSelected())
		{
			boolean[] classifier_selection = ensemble_of_clasifiers.getEnsembleSelection();
			table_title = new String("CLASSIFIER SELECTION:");
			printTable( table_title, "Used", classifier_names, classifier_selection, ((int) (tab_size * 1.5)) );
		}

		// Show the classifier weighting
		// These values are normalized for display purposes only. The original values
		// may not be normalized.
		if (preferences_panel.shouldDisplayClassifiersWeightings())
		{
			double[] classifier_weightings = ensemble_of_clasifiers.getEnsembleWeightings();
			double[] norm_classifier_weightings = bodhidharma.utilities.Statistics.normalize(classifier_weightings);
			table_title = new String("CLASSIFIER WEIGHTINGS:");
			printTable( table_title, "Weighting", classifier_names, norm_classifier_weightings,
						((int) (tab_size * 1.5)), one_column_max_decimal_places );
		}
	}


	// Appends a table to the results_display field. This table contains the contents of the
	// table_data parameter (first indice inicates row, second indicates column). The
	// column_headings and row_headings parameters set the labels of the columns and rows
	// respectively. All displays of the column and row labels will be shortened to the
	// number of characters specified in the column_widths parameter less one. All entries
	// in the table_data parameter are rounded to a maximum of max_decimal_places digits.
	private void printTable( String table_heading,
	                         String[] column_headings,
	                         String[] row_headings,
	                         double[][] table_data,
	                         int column_widths,
	                         int max_decimal_places )
	{
		// Print out the table heading
		appendTableTitle(table_heading);

		// Print out the column headings
		for (int col = 0; col < column_headings.length; col++)
		{
			String column_name = column_headings[col];
			column_name = bodhidharma.utilities.StringMethods.getBeginningOfString( column_name, (column_widths - 1) );
			results_display.append("\t" + column_name);
		}
		results_display.append("\n");
		
		// Print out the table
		for (int row = 0; row < row_headings.length; row++)
		{
			// Print the row headings
			String row_name = row_headings[row];
			row_name = bodhidharma.utilities.StringMethods.getBeginningOfString( row_name, (column_widths - 1) );
			results_display.append(row_name + "\t");
			
			// Print out table data
			for (int col = 0; col < column_headings.length; col++)
			{
				String formatted_value =
					bodhidharma.utilities.StringMethods.getRoundedDouble( table_data[row][col], max_decimal_places );
				results_display.append( formatted_value + "\t" );
			}
			results_display.append("\n");
		}
		results_display.append("\n\n\n");
	}


	// Appends a one column table to the results_display field. This table contains the
	// contents of the table_data parameter (indice inicates row). The column_headings
	// and row_headings parameters set the labels of the columns and rows respectively.
	// All displays of the column and row labels as well as entries will be shortened to the
	// number of characters specified in the column_widths parameter less one.
	private void printTable( String table_heading,
	                         String column_name,
	                         String[] row_headings,
	                         boolean[] table_data,
	                         int column_widths )
	{
		// Print out the table heading
		appendTableTitle(table_heading);

		// Print out the column heading
		column_name = bodhidharma.utilities.StringMethods.getBeginningOfString( column_name, (column_widths - 1) );
		results_display.append("\t\t" + column_name);
		results_display.append("\n");
		
		// Print out the table
		for (int row = 0; row < row_headings.length; row++)
		{
			// Print the row headings
			String row_name = row_headings[row];
			row_name = bodhidharma.utilities.StringMethods.getBeginningOfString( row_name, (column_widths - 1) );
			results_display.append(row_name + "\t");
			
			// Print out table data
			results_display.append( table_data[row] + "\n" );
		}
		results_display.append("\n\n\n");
	}


	// Appends a one column table to the results_display field. This table contains the
	// contents of the table_data parameter (indice inicates row). The column_headings
	// and row_headings parameters set the labels of the columns and rows respectively.
	// All displays of the column and row labels as well as entries will be shortened to the
	// number of characters specified in the column_widths parameter less one. All entries
	// in the table_data parameter are rounded to a maximum of max_decimal_places digits.
	private void printTable( String table_heading,
	                         String column_name,
	                         String[] row_headings,
	                         double[] table_data,
	                         int column_widths,
	                         int max_decimal_places )
	{
		// Print out the table heading
		appendTableTitle(table_heading);

		// Print out the column heading
		column_name = bodhidharma.utilities.StringMethods.getBeginningOfString( column_name, (column_widths - 1) );
		results_display.append("\t\t" + column_name);
		results_display.append("\n");
		
		// Print out the table
		for (int row = 0; row < row_headings.length; row++)
		{
			// Print the row headings
			String row_name = row_headings[row];
			row_name = bodhidharma.utilities.StringMethods.getBeginningOfString( row_name, (column_widths - 1) );
			results_display.append(row_name + "\t");
			
			// Print out table data
			String formatted_value =
				bodhidharma.utilities.StringMethods.getRoundedDouble( table_data[row], max_decimal_places );
			results_display.append( formatted_value + "\n" );
		}
		results_display.append("\n\n\n");
	}


	// Appends a one column table to the results_display field. This table contains the
	// contents of the table_data parameter (indice inicates row). The column_heading
	// parameters setthe label of the column. The row heading is the indice of the row.
	// Only every rows_to_skip rows will be displayed, however.
	// All displays of the column and row labels as well as entries will be shortened to the
	// number of characters specified in the column_widths parameter less one. All entries
	// in the table_data parameter are rounded to a maximum of max_decimal_places digits.
	private void printTableSkippingRows( String table_heading,
	                                     String column_name,
	                                     double[] table_data,
	                                     int rows_to_skip,
	                                     int column_widths,
	                                     int max_decimal_places )
	{
		// Print out the table heading
		appendTableTitle(table_heading);

		// Print out the column heading
		column_name = bodhidharma.utilities.StringMethods.getBeginningOfString( column_name, (column_widths - 1) );
		results_display.append("\t\t" + column_name);
		results_display.append("\n");
		
		// Print out the table
		for (int row = 0; row < table_data.length; row++)
		{
			// Skip some rows
			if (row % rows_to_skip == 0)
			{
				// Set the format for row headings
				String format_pattern = new String("");
				for (int i = 0; i < (column_widths - 1); i++)
					format_pattern += "0";
				DecimalFormat formatter = new DecimalFormat(format_pattern);

				// Print the row headings
				String row_name = formatter.format(row + 1);
				results_display.append(row_name + "\t");
				
				// Print out table data
				String formatted_value =
					bodhidharma.utilities.StringMethods.getRoundedDouble( table_data[row], max_decimal_places );
				results_display.append( formatted_value + "\n" );
			}
		}
		results_display.append("\n\n\n");
	}


	// Appends a one column table to the results_display field. This table contains the
	// contents of the table_data parameter (indice inicates row). The column_headings
	// and row_headings parameters set the labels of the columns and rows respectively.
	// All displays of the row labels will be shortened to the number of characters 
	// specified in the column_widths parameter less one.
	private void printTable( String table_heading,
	                         String column_name,
	                         String[] row_headings,
	                         String[] table_data,
	                         int column_widths )
	{
		// Print out the table heading
		appendTableTitle(table_heading);

		// Print out the column heading
		column_name = bodhidharma.utilities.StringMethods.getBeginningOfString( column_name, (column_widths - 1) );
		results_display.append("\t\t" + column_name);
		results_display.append("\n");
		
		// Print out the table
		for (int row = 0; row < row_headings.length; row++)
		{
			// Print the row headings
			String row_name = row_headings[row];
			row_name = bodhidharma.utilities.StringMethods.getBeginningOfString( row_name, (column_widths - 1) );
			results_display.append(row_name + "\t");
			
			// Print out table data
			results_display.append( table_data[row] + "\n" );
		}
		results_display.append("\n\n\n");
	}


	// Appends a table whose number of columns depend on the row to the results_display field.
	// This table contains the contents of the table_data parameter (indice inicates row). The
	// column_headings and row_headings parameters set the labels of the columns and rows respectively.
	// All displays of the row labels will be shortened to the number of characters 
	// specified in the column_widths parameter less one.
	private void printTable( String table_heading,
	                         String column_name,
	                         String[] row_headings,
	                         String[][] table_data,
	                         int column_widths )
	{
		// Print out the table heading
		appendTableTitle(table_heading);

		// Print out the column heading
		column_name = bodhidharma.utilities.StringMethods.getBeginningOfString( column_name, (column_widths - 1) );
		results_display.append("\t\t" + column_name);
		results_display.append("\n");
		
		// Print out the table
		for (int row = 0; row < row_headings.length; row++)
		{
			// Print the row headings
			String row_name = row_headings[row];
			row_name = bodhidharma.utilities.StringMethods.getBeginningOfString( row_name, (column_widths - 1) );
			results_display.append(row_name + "\t");
			
			// Print out table data
			for (int col = 0; col < table_data[row].length; col++)
			{
				results_display.append(table_data[row][col]);
				if (col < (table_data[row].length - 1))
					results_display.append(", ");
			}
			results_display.append("\n");
		}
		results_display.append("\n\n\n");
	}


	// Appends the given title to results_display with a properly sized underline
	// on top and on botton
	private void appendTableTitle(String title)
	{
		String underline = new String();
		for (int i = 0; i < title.length(); i++)
			underline += "-";
		results_display.append(underline + "\n" + title + "\n" + underline + "\n\n");
	}
}